#include "php_gtk.h"

#if HAVE_PHP_GTK
#include "php_gtk_api.h"
#include "ext/standard/file.h"
#include "ext/gtk+/phpg_custom_tree_model.h"

static void phpg_about_dialog_activate_link_func_marshal (GtkAboutDialog *about, const gchar *link, gpointer data);
static gboolean phpg_entry_completion_match_func_marshal (GtkEntryCompletion *completion, const gchar *key, GtkTreeIter *iter, gpointer data);
static void phpg_file_filter_add_custom_marshal_add_file(const gchar *filename, int flags, int flag, zval *php_filter_info);
static gboolean phpg_file_filter_add_custom_marshal(GtkFileFilterInfo *filter_info, gpointer data);
static gboolean phpg_tree_view_row_separator_func_marshal (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void phpg_icon_view_foreach_func_marshal (GtkIconView *icon_view, GtkTreePath *path, gpointer data);
static gboolean phpg_toolbar_signal_func_marshal(GtkWidget *widget, gpointer data);
static void phpg_color_selection_change_palette_with_screen_func_marshal(GdkScreen *screen, const GdkColor *colors, gint n_colors);

/* forward declaration */
static PHP_METHOD(GtkImage, get_image);

#undef GDK_DISPLAY
#define GDK_DISPLAY(object) (GDK_DISPLAY_OBJECT(object))

#ifndef GDK_TYPE_REGION
extern phpg_region_get_type();
#define GDK_TYPE_REGION (phpg_region_get_type())
#endif

static ZEND_METHOD(GtkContainer, get_children);

extern PHP_GTK_EXPORT_CE(gdkdragcontext_ce);
static GtkTargetEntry* phpg_parse_target_entries(zval *php_targets, int *n_targets TSRMLS_DC);


static void phpg_text_tag_table_foreach_func_marshal (GtkTextTagTable *table, gpointer data);
static gboolean phpg_text_iter_char_predicate_func_marshal (gunichar ch, gpointer data);


#include "zend_interfaces.h"

static void phpg_cell_data_func_marshal(GtkCellLayout *cell_layout, GtkCellRenderer *renderer, GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void phpg_store_set_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool is_tree);
static gint phpg_tree_iter_compare_func_marshal(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data);
static gboolean phpg_tree_model_filter_visible_func_marshal(GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_view_row_separator_func_marshal (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_view_column_drop_func_marshal (GtkTreeView *tree_view, GtkTreeViewColumn *column, GtkTreeViewColumn *prev_column, GtkTreeViewColumn *next_column, gpointer data);
static gboolean phpg_tree_view_search_equal_func_marshal (GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *iter, gpointer data);
static void phpg_tree_selection_foreach_func_marshal (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_selection_func_marshal (GtkTreeSelection *selection, GtkTreeModel *model, GtkTreePath *path, gboolean path_currently_selected, gpointer data);

static void phpg_link_button_set_uri_hook_func_marshal(GtkLinkButton *button, const gchar *link, gpointer data);


typedef struct {
	zval *instance;
	zval *extra;
#ifdef ZTS
	TSRMLS_D;
#endif
} phpg_gtkbuilder_t;

static void phpg_scale_button_set_icons_assist(GtkScaleButton *, zval * TSRMLS_DC);

PHP_GTK_EXPORT_CE(gtktoolshell_ce);
PHP_GTK_EXPORT_CE(gtkbuildable_ce);
PHP_GTK_EXPORT_CE(gtkprintoperationpreview_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooser_ce);
PHP_GTK_EXPORT_CE(gtkcelleditable_ce);
PHP_GTK_EXPORT_CE(gtkcelllayout_ce);
PHP_GTK_EXPORT_CE(gtkeditable_ce);
PHP_GTK_EXPORT_CE(gtkfilechooser_ce);
PHP_GTK_EXPORT_CE(gtktreedragdest_ce);
PHP_GTK_EXPORT_CE(gtktreedragsource_ce);
PHP_GTK_EXPORT_CE(gtktreemodel_ce);
PHP_GTK_EXPORT_CE(gtktreesortable_ce);
PHP_GTK_EXPORT_CE(gtkactivatable_ce);
PHP_GTK_EXPORT_CE(gtkorientable_ce);
PHP_GTK_EXPORT_CE(gtkmountoperation_ce);
PHP_GTK_EXPORT_CE(gtkbuilder_ce);
PHP_GTK_EXPORT_CE(gtktooltip_ce);
PHP_GTK_EXPORT_CE(gtkpagesetup_ce);
PHP_GTK_EXPORT_CE(gtkprintcontext_ce);
PHP_GTK_EXPORT_CE(gtkprintoperation_ce);
PHP_GTK_EXPORT_CE(gtkprintsettings_ce);
PHP_GTK_EXPORT_CE(gtkrecentmanager_ce);
PHP_GTK_EXPORT_CE(gtkstatusicon_ce);
PHP_GTK_EXPORT_CE(phpgtkcustomtreemodel_ce);
PHP_GTK_EXPORT_CE(gtkaccelgroup_ce);
PHP_GTK_EXPORT_CE(gtkaccelmap_ce);
PHP_GTK_EXPORT_CE(gtkaccessible_ce);
PHP_GTK_EXPORT_CE(gtkaction_ce);
PHP_GTK_EXPORT_CE(gtkactiongroup_ce);
PHP_GTK_EXPORT_CE(gtkrecentaction_ce);
PHP_GTK_EXPORT_CE(gtkclipboard_ce);
PHP_GTK_EXPORT_CE(gtkentrycompletion_ce);
PHP_GTK_EXPORT_CE(gtkiconfactory_ce);
PHP_GTK_EXPORT_CE(gtkicontheme_ce);
PHP_GTK_EXPORT_CE(gtkliststore_ce);
PHP_GTK_EXPORT_CE(gtkobject_ce);
PHP_GTK_EXPORT_CE(gtkitemfactory_ce);
PHP_GTK_EXPORT_CE(gtkimcontext_ce);
PHP_GTK_EXPORT_CE(gtkfilefilter_ce);
PHP_GTK_EXPORT_CE(gtkimmulticontext_ce);
PHP_GTK_EXPORT_CE(gtkimcontextsimple_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderer_ce);
PHP_GTK_EXPORT_CE(gtkadjustment_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderertoggle_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderertext_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererprogress_ce);
PHP_GTK_EXPORT_CE(gtkcellrendereraccel_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererspin_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderercombo_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererpixbuf_ce);
PHP_GTK_EXPORT_CE(gtkrecentfilter_ce);
PHP_GTK_EXPORT_CE(gtkrcstyle_ce);
PHP_GTK_EXPORT_CE(gtksettings_ce);
PHP_GTK_EXPORT_CE(gtksizegroup_ce);
PHP_GTK_EXPORT_CE(gtkstyle_ce);
PHP_GTK_EXPORT_CE(gtktextbuffer_ce);
PHP_GTK_EXPORT_CE(gtktextchildanchor_ce);
PHP_GTK_EXPORT_CE(gtktextmark_ce);
PHP_GTK_EXPORT_CE(gtktexttag_ce);
PHP_GTK_EXPORT_CE(gtktexttagtable_ce);
PHP_GTK_EXPORT_CE(gtktoggleaction_ce);
PHP_GTK_EXPORT_CE(gtkradioaction_ce);
PHP_GTK_EXPORT_CE(gtktooltips_ce);
PHP_GTK_EXPORT_CE(gtktreemodelfilter_ce);
PHP_GTK_EXPORT_CE(gtktreemodelsort_ce);
PHP_GTK_EXPORT_CE(gtktreeselection_ce);
PHP_GTK_EXPORT_CE(gtktreestore_ce);
PHP_GTK_EXPORT_CE(gtktreeviewcolumn_ce);
PHP_GTK_EXPORT_CE(gtkuimanager_ce);
PHP_GTK_EXPORT_CE(gtkwidget_ce);
PHP_GTK_EXPORT_CE(gtkseparator_ce);
PHP_GTK_EXPORT_CE(gtkruler_ce);
PHP_GTK_EXPORT_CE(gtkvseparator_ce);
PHP_GTK_EXPORT_CE(gtkhseparator_ce);
PHP_GTK_EXPORT_CE(gtkvruler_ce);
PHP_GTK_EXPORT_CE(gtkrange_ce);
PHP_GTK_EXPORT_CE(gtkhruler_ce);
PHP_GTK_EXPORT_CE(gtkscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkprogress_ce);
PHP_GTK_EXPORT_CE(gtkvscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkhscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkscale_ce);
PHP_GTK_EXPORT_CE(gtkpreview_ce);
PHP_GTK_EXPORT_CE(gtkvscale_ce);
PHP_GTK_EXPORT_CE(gtkhscale_ce);
PHP_GTK_EXPORT_CE(gtkprogressbar_ce);
PHP_GTK_EXPORT_CE(gtkoldeditable_ce);
PHP_GTK_EXPORT_CE(gtkmisc_ce);
PHP_GTK_EXPORT_CE(gtkinvisible_ce);
PHP_GTK_EXPORT_CE(gtkpixmap_ce);
PHP_GTK_EXPORT_CE(gtkarrow_ce);
PHP_GTK_EXPORT_CE(gtkimage_ce);
PHP_GTK_EXPORT_CE(gtklabel_ce);
PHP_GTK_EXPORT_CE(gtkentry_ce);
PHP_GTK_EXPORT_CE(gtktipsquery_ce);
PHP_GTK_EXPORT_CE(gtkaccellabel_ce);
PHP_GTK_EXPORT_CE(gtkspinbutton_ce);
PHP_GTK_EXPORT_CE(gtkdrawingarea_ce);
PHP_GTK_EXPORT_CE(gtkcontainer_ce);
PHP_GTK_EXPORT_CE(gtkcurve_ce);
PHP_GTK_EXPORT_CE(gtktreeview_ce);
PHP_GTK_EXPORT_CE(gtktoolbar_ce);
PHP_GTK_EXPORT_CE(gtktextview_ce);
PHP_GTK_EXPORT_CE(gtktable_ce);
PHP_GTK_EXPORT_CE(gtksocket_ce);
PHP_GTK_EXPORT_CE(gtkpaned_ce);
PHP_GTK_EXPORT_CE(gtknotebook_ce);
PHP_GTK_EXPORT_CE(gtkvpaned_ce);
PHP_GTK_EXPORT_CE(gtkhpaned_ce);
PHP_GTK_EXPORT_CE(gtkmenushell_ce);
PHP_GTK_EXPORT_CE(gtklist_ce);
PHP_GTK_EXPORT_CE(gtkmenu_ce);
PHP_GTK_EXPORT_CE(gtkmenubar_ce);
PHP_GTK_EXPORT_CE(gtkrecentchoosermenu_ce);
PHP_GTK_EXPORT_CE(gtklayout_ce);
PHP_GTK_EXPORT_CE(gtkiconview_ce);
PHP_GTK_EXPORT_CE(gtkfixed_ce);
PHP_GTK_EXPORT_CE(gtkcellview_ce);
PHP_GTK_EXPORT_CE(gtkbin_ce);
PHP_GTK_EXPORT_CE(gtkbox_ce);
PHP_GTK_EXPORT_CE(gtkwindow_ce);
PHP_GTK_EXPORT_CE(gtkviewport_ce);
PHP_GTK_EXPORT_CE(gtkdialog_ce);
PHP_GTK_EXPORT_CE(gtkplug_ce);
PHP_GTK_EXPORT_CE(gtkcolorselectiondialog_ce);
PHP_GTK_EXPORT_CE(gtkaboutdialog_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooserdialog_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserdialog_ce);
PHP_GTK_EXPORT_CE(gtkfileselection_ce);
PHP_GTK_EXPORT_CE(gtkfontselectiondialog_ce);
PHP_GTK_EXPORT_CE(gtkinputdialog_ce);
PHP_GTK_EXPORT_CE(gtkmessagedialog_ce);
PHP_GTK_EXPORT_CE(gtktoolitem_ce);
PHP_GTK_EXPORT_CE(gtkvbox_ce);
PHP_GTK_EXPORT_CE(gtkscrolledwindow_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooserwidget_ce);
PHP_GTK_EXPORT_CE(gtkcolorselection_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserwidget_ce);
PHP_GTK_EXPORT_CE(gtkfontselection_ce);
PHP_GTK_EXPORT_CE(gtkgammacurve_ce);
PHP_GTK_EXPORT_CE(gtkseparatortoolitem_ce);
PHP_GTK_EXPORT_CE(gtktoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkitem_ce);
PHP_GTK_EXPORT_CE(gtktoggletoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkmenutoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkradiotoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkhandlebox_ce);
PHP_GTK_EXPORT_CE(gtkmenuitem_ce);
PHP_GTK_EXPORT_CE(gtklistitem_ce);
PHP_GTK_EXPORT_CE(gtktearoffmenuitem_ce);
PHP_GTK_EXPORT_CE(gtkseparatormenuitem_ce);
PHP_GTK_EXPORT_CE(gtkcheckmenuitem_ce);
PHP_GTK_EXPORT_CE(gtkimagemenuitem_ce);
PHP_GTK_EXPORT_CE(gtkradiomenuitem_ce);
PHP_GTK_EXPORT_CE(gtkhbox_ce);
PHP_GTK_EXPORT_CE(gtkframe_ce);
PHP_GTK_EXPORT_CE(gtkstatusbar_ce);
PHP_GTK_EXPORT_CE(gtkcombo_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserbutton_ce);
PHP_GTK_EXPORT_CE(gtkexpander_ce);
PHP_GTK_EXPORT_CE(gtkaspectframe_ce);
PHP_GTK_EXPORT_CE(gtkeventbox_ce);
PHP_GTK_EXPORT_CE(gtkalignment_ce);
PHP_GTK_EXPORT_CE(gtkbutton_ce);
PHP_GTK_EXPORT_CE(gtkcombobox_ce);
PHP_GTK_EXPORT_CE(gtktogglebutton_ce);
PHP_GTK_EXPORT_CE(gtkoptionmenu_ce);
PHP_GTK_EXPORT_CE(gtkcheckbutton_ce);
PHP_GTK_EXPORT_CE(gtkfontbutton_ce);
PHP_GTK_EXPORT_CE(gtkradiobutton_ce);
PHP_GTK_EXPORT_CE(gtklinkbutton_ce);
PHP_GTK_EXPORT_CE(gtkscalebutton_ce);
PHP_GTK_EXPORT_CE(gtkcolorbutton_ce);
PHP_GTK_EXPORT_CE(gtkvolumebutton_ce);
PHP_GTK_EXPORT_CE(gtkclist_ce);
PHP_GTK_EXPORT_CE(gtkcomboboxentry_ce);
PHP_GTK_EXPORT_CE(gtkbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkcalendar_ce);
PHP_GTK_EXPORT_CE(gtkvbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkhbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkctree_ce);
PHP_GTK_EXPORT_CE(gtkhsv_ce);
PHP_GTK_EXPORT_CE(gtkassistant_ce);
PHP_GTK_EXPORT_CE(gtkwindowgroup_ce);
PHP_GTK_EXPORT_CE(gtkinfobar_ce);
PHP_GTK_EXPORT_CE(gtkentrybuffer_ce);
PHP_GTK_EXPORT_CE(gtktoolitemgroup_ce);
PHP_GTK_EXPORT_CE(gtktoolpalette_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererspinner_ce);
PHP_GTK_EXPORT_CE(gtkoffscreenwindow_ce);
PHP_GTK_EXPORT_CE(gtkspinner_ce);
PHP_GTK_EXPORT_CE(gtkcomboboxtext_ce);
PHP_GTK_EXPORT_CE(gtkpapersize_ce);
PHP_GTK_EXPORT_CE(gtkrecentinfo_ce);
PHP_GTK_EXPORT_CE(gtktargetlist_ce);
PHP_GTK_EXPORT_CE(gtkborder_ce);
PHP_GTK_EXPORT_CE(gtkiconinfo_ce);
PHP_GTK_EXPORT_CE(gtkiconset_ce);
PHP_GTK_EXPORT_CE(gtkiconsource_ce);
PHP_GTK_EXPORT_CE(gtkrequisition_ce);
PHP_GTK_EXPORT_CE(gtkselectiondata_ce);
PHP_GTK_EXPORT_CE(gtktextattributes_ce);
PHP_GTK_EXPORT_CE(gtktextiter_ce);
PHP_GTK_EXPORT_CE(gtktreeiter_ce);
PHP_GTK_EXPORT_CE(gtktreerowreference_ce);
PHP_GTK_EXPORT_CE(gtkctreenode_ce);
PHP_GTK_EXPORT_CE(gtk_ce);
void phpg_gtk0_register_classes(TSRMLS_D);
void phpg_gtk1_register_classes(TSRMLS_D);

static PHP_METHOD(GtkAssistant, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAssistant);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAssistant);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkAssistant, get_current_page)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_assistant_get_current_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAssistant, set_current_page)
{
	long page_num;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

    gtk_assistant_set_current_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), (gint)page_num);

}


static PHP_METHOD(GtkAssistant, get_n_pages)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_assistant_get_n_pages(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAssistant, get_nth_page)
{
	long page_num;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

    php_retval = gtk_assistant_get_nth_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), (gint)page_num);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAssistant, prepend_page)
{
	zval *page;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_prepend_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAssistant, append_page)
{
	zval *page;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_append_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAssistant, insert_page)
{
	zval *page;
	long position, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &page, gtkwidget_ce, &position))
		return;

    php_retval = gtk_assistant_insert_page(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), (gint)position);
	RETVAL_LONG(php_retval);
}

#line 585 "ext/gtk+/gtk-2.10.overrides"

static gint
phpg_assistant_forward_page_func_marshal(gint cur, gpointer data)
{
    char *callback;
    int argc = 0, status = 0;
    zval ***argv = NULL, *retval = NULL;
    zval *zcur;
    phpg_cb_data_t *cbd = (phpg_cb_data_t *)data;
    
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(
            E_WARNING,
            "Unable to invoke callback '%s' specified in %s on line %ld",
            callback,
            cbd->src_filename,
            cbd->src_lineno
        );
        
        efree(callback);
        return 0;
    }

    MAKE_STD_ZVAL(zcur);
    ZVAL_LONG(zcur,cur);

    argv = php_gtk_hash_as_array_offset(cbd->user_args,1,&argc);
    argv[0] = &zcur;

    call_user_function_ex(
        EG(function_table),
        NULL,
        cbd->callback,
        &retval,
        argc,
        argv,
        0,
        NULL TSRMLS_CC
    );

    if(retval) {
       convert_to_long(retval);
       status = Z_LVAL_P(retval);
       zval_ptr_dtor(&retval);
    } else {
       status = 0;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback);
    efree(argv);

    return status;
}

static PHP_METHOD(GtkAssistant, set_forward_page_func)
{
    zval *callback, *vararg;
    phpg_cb_data_t *calldata;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_varargs(ZEND_NUM_ARGS(),1,&vararg,"V",&callback)) {
        return;
    }

    zval_add_ref(&callback);
    calldata = phpg_cb_data_new(callback,vararg TSRMLS_CC);

    gtk_assistant_set_forward_page_func(
        GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)),
        (GtkAssistantPageFunc)phpg_assistant_forward_page_func_marshal,
        calldata,
        phpg_cb_data_destroy
    );
}

#line 210 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAssistant, set_page_type)
{
	zval *page, *php_type = NULL;
	GtkAssistantPageType type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &page, gtkwidget_ce, &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_ASSISTANT_PAGE_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_assistant_set_page_type(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), type);

}


static PHP_METHOD(GtkAssistant, get_page_type)
{
	zval *page;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_get_page_type(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAssistant, set_page_title)
{
	zval *page;
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &page, gtkwidget_ce, &title, &free_title))
		return;

    gtk_assistant_set_page_title(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkAssistant, get_page_title)
{
	zval *page;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_get_page_title(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAssistant, set_page_header_image)
{
	zval *page, *php_pixbuf = NULL;
	GdkPixbuf *pixbuf = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &page, gtkwidget_ce, &php_pixbuf, gdkpixbuf_ce))
		return;
    if (php_pixbuf) {
        if (Z_TYPE_P(php_pixbuf) == IS_NULL)
            pixbuf = NULL;
        else
            pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));
    }

    gtk_assistant_set_page_header_image(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), pixbuf);

}


static PHP_METHOD(GtkAssistant, get_page_header_image)
{
	zval *page;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_get_page_header_image(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAssistant, set_page_side_image)
{
	zval *page, *php_pixbuf = NULL;
	GdkPixbuf *pixbuf = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &page, gtkwidget_ce, &php_pixbuf, gdkpixbuf_ce))
		return;
    if (php_pixbuf) {
        if (Z_TYPE_P(php_pixbuf) == IS_NULL)
            pixbuf = NULL;
        else
            pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));
    }

    gtk_assistant_set_page_side_image(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), pixbuf);

}


static PHP_METHOD(GtkAssistant, get_page_side_image)
{
	zval *page;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_get_page_side_image(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAssistant, set_page_complete)
{
	zval *page;
	zend_bool complete;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &page, gtkwidget_ce, &complete))
		return;

    gtk_assistant_set_page_complete(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)), (gboolean)complete);

}


static PHP_METHOD(GtkAssistant, get_page_complete)
{
	zval *page;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &page, gtkwidget_ce))
		return;

    php_retval = gtk_assistant_get_page_complete(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(page)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAssistant, add_action_widget)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_assistant_add_action_widget(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkAssistant, remove_action_widget)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_assistant_remove_action_widget(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkAssistant, update_buttons_state)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_assistant_update_buttons_state(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAssistant, commit)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_assistant_commit(GTK_ASSISTANT(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_set_current_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_nth_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_prepend_page, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_append_page, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_insert_page, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_set_forward_page_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_set_page_type, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_page_type, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_set_page_title, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_page_title, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkassistant_set_page_header_image, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_page_header_image, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkassistant_set_page_side_image, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_page_side_image, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_set_page_complete, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
    ZEND_ARG_INFO(0, complete)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_get_page_complete, 0)
    ZEND_ARG_OBJ_INFO(0, page, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_add_action_widget, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkassistant_remove_action_widget, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkassistant_methods[] = {
	PHP_ME(GtkAssistant, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, add_action_widget,    arginfo_gtk_gtkassistant_add_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, append_page,          arginfo_gtk_gtkassistant_append_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, commit,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_current_page,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_n_pages,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_nth_page,         arginfo_gtk_gtkassistant_get_nth_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_page_complete,    arginfo_gtk_gtkassistant_get_page_complete, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_page_header_image, arginfo_gtk_gtkassistant_get_page_header_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_page_side_image,  arginfo_gtk_gtkassistant_get_page_side_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_page_title,       arginfo_gtk_gtkassistant_get_page_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, get_page_type,        arginfo_gtk_gtkassistant_get_page_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, insert_page,          arginfo_gtk_gtkassistant_insert_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, prepend_page,         arginfo_gtk_gtkassistant_prepend_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, remove_action_widget, arginfo_gtk_gtkassistant_remove_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_current_page,     arginfo_gtk_gtkassistant_set_current_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_forward_page_func, arginfo_gtk_gtkassistant_set_forward_page_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_page_complete,    arginfo_gtk_gtkassistant_set_page_complete, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_page_header_image, arginfo_gtk_gtkassistant_set_page_header_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_page_side_image,  arginfo_gtk_gtkassistant_set_page_side_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_page_title,       arginfo_gtk_gtkassistant_set_page_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, set_page_type,        arginfo_gtk_gtkassistant_set_page_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAssistant, update_buttons_state, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkWindowGroup, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkWindowGroup);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkWindowGroup);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkWindowGroup, add_window)
{
	zval *window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gtkwindow_ce))
		return;

    gtk_window_group_add_window(GTK_WINDOW_GROUP(PHPG_GOBJECT(this_ptr)), GTK_WINDOW(PHPG_GOBJECT(window)));

}


static PHP_METHOD(GtkWindowGroup, remove_window)
{
	zval *window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gtkwindow_ce))
		return;

    gtk_window_group_remove_window(GTK_WINDOW_GROUP(PHPG_GOBJECT(this_ptr)), GTK_WINDOW(PHPG_GOBJECT(window)));

}


static PHP_METHOD(GtkWindowGroup, get_current_grab)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_group_get_current_grab(GTK_WINDOW_GROUP(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindowgroup_add_window, 0)
    ZEND_ARG_OBJ_INFO(0, window, GtkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindowgroup_remove_window, 0)
    ZEND_ARG_OBJ_INFO(0, window, GtkWindow, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkwindowgroup_methods[] = {
	PHP_ME(GtkWindowGroup, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindowGroup, add_window,           arginfo_gtk_gtkwindowgroup_add_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindowGroup, get_current_grab,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindowGroup, remove_window,        arginfo_gtk_gtkwindowgroup_remove_window, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkInfoBar, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInfoBar);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInfoBar);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkInfoBar, add_action_widget)
{
	zval *child;
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &response_id))
		return;

    gtk_info_bar_add_action_widget(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)response_id);

}


static PHP_METHOD(GtkInfoBar, add_button)
{
	char *button_text;
	zend_bool free_button_text = FALSE;
	long response_id;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ui", &button_text, &free_button_text, &response_id))
		return;

    php_retval = gtk_info_bar_add_button(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), button_text, (gint)response_id);
	if (free_button_text) g_free(button_text);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 54 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkInfoBar, add_buttons)
{
    zval *buttons = NULL;
    zval **text, **response;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &buttons))
        return;


    if (zend_hash_num_elements(Z_ARRVAL_P(buttons)) % 2) {
        php_error(E_WARNING,
                    "%s::%s(): button list has to contain pairs of items",
                        get_active_class_name(NULL TSRMLS_CC),
                            get_active_function_name(TSRMLS_C));
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(buttons));
    while (zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&text) == SUCCESS) {
        zend_hash_move_forward(Z_ARRVAL_P(buttons));
        zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&response); /* safe */
        zend_hash_move_forward(Z_ARRVAL_P(buttons));

        if (Z_TYPE_PP(text) != IS_STRING || Z_TYPE_PP(response) != IS_LONG) {
            php_error(E_WARNING, "%s::%s(): each pair in button list has to be string/number",
                        get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }

        gtk_info_bar_add_button(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), Z_STRVAL_PP(text), Z_LVAL_PP(response));
    }
}

#line 749 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkInfoBar, get_action_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_info_bar_get_action_area(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkInfoBar, get_content_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_info_bar_get_content_area(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkInfoBar, get_message_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_info_bar_get_message_type(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkInfoBar, response)
{
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &response_id))
		return;

    gtk_info_bar_response(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), (gint)response_id);

}


static PHP_METHOD(GtkInfoBar, set_default_response)
{
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &response_id))
		return;

    gtk_info_bar_set_default_response(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), (gint)response_id);

}


static PHP_METHOD(GtkInfoBar, set_message_type)
{
	GtkMessageType message_type;
	zval *php_message_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_message_type))
		return;

	if (php_message_type && phpg_gvalue_get_enum(GTK_TYPE_MESSAGE_TYPE, php_message_type, (gint *)&message_type) == FAILURE) {
		return;
	}

    gtk_info_bar_set_message_type(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), message_type);

}


static PHP_METHOD(GtkInfoBar, set_response_sensitive)
{
	long response_id;
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &response_id, &sensitive))
		return;

    gtk_info_bar_set_response_sensitive(GTK_INFO_BAR(PHPG_GOBJECT(this_ptr)), (gint)response_id, (gboolean)sensitive);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_add_action_widget, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_add_button, 0)
    ZEND_ARG_INFO(0, button_text)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_add_buttons, 0)
    ZEND_ARG_INFO(0, buttons)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_response, 0)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_set_default_response, 0)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_set_message_type, 0)
    ZEND_ARG_INFO(0, message_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinfobar_set_response_sensitive, 0)
    ZEND_ARG_INFO(0, response_id)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

static zend_function_entry gtkinfobar_methods[] = {
	PHP_ME(GtkInfoBar, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, add_action_widget,    arginfo_gtk_gtkinfobar_add_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, add_button,           arginfo_gtk_gtkinfobar_add_button, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, add_buttons,          arginfo_gtk_gtkinfobar_add_buttons, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, get_action_area,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, get_content_area,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, get_message_type,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, response,             arginfo_gtk_gtkinfobar_response, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, set_default_response, arginfo_gtk_gtkinfobar_set_default_response, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, set_message_type,     arginfo_gtk_gtkinfobar_set_message_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInfoBar, set_response_sensitive, arginfo_gtk_gtkinfobar_set_response_sensitive, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkEntryBuffer, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntryBuffer);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntryBuffer);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkEntryBuffer, delete_text)
{
	long position, n_chars, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &position, &n_chars))
		return;

    php_retval = gtk_entry_buffer_delete_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), (guint)position, (gint)n_chars);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryBuffer, emit_deleted_text)
{
	long position, n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &position, &n_chars))
		return;

    gtk_entry_buffer_emit_deleted_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), (guint)position, (gint)n_chars);

}


static PHP_METHOD(GtkEntryBuffer, emit_inserted_text)
{
	long position, n_chars;
	char *chars;
	zend_bool free_chars = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iui", &position, &chars, &free_chars, &n_chars))
		return;

    gtk_entry_buffer_emit_inserted_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), (guint)position, chars, (guint)n_chars);
	if (free_chars) g_free(chars);

}


static PHP_METHOD(GtkEntryBuffer, get_bytes)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_buffer_get_bytes(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryBuffer, get_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_buffer_get_length(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryBuffer, get_max_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_buffer_get_max_length(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryBuffer, get_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_buffer_get_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkEntryBuffer, insert_text)
{
	long position, n_chars, php_retval;
	char *chars;
	zend_bool free_chars = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iui", &position, &chars, &free_chars, &n_chars))
		return;

    php_retval = gtk_entry_buffer_insert_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), (guint)position, chars, (guint)n_chars);
	if (free_chars) g_free(chars);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryBuffer, set_max_length)
{
	long max_length;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &max_length))
		return;

    gtk_entry_buffer_set_max_length(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), (gint)max_length);

}


static PHP_METHOD(GtkEntryBuffer, set_text)
{
	char *chars;
	zend_bool free_chars = FALSE;
	long n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ui", &chars, &free_chars, &n_chars))
		return;

    gtk_entry_buffer_set_text(GTK_ENTRY_BUFFER(PHPG_GOBJECT(this_ptr)), chars, (guint)n_chars);
	if (free_chars) g_free(chars);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_delete_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_emit_deleted_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_emit_inserted_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, chars)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_insert_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, chars)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_set_max_length, 0)
    ZEND_ARG_INFO(0, max_length)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrybuffer_set_text, 0)
    ZEND_ARG_INFO(0, chars)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

static zend_function_entry gtkentrybuffer_methods[] = {
	PHP_ME(GtkEntryBuffer, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, delete_text,          arginfo_gtk_gtkentrybuffer_delete_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, emit_deleted_text,    arginfo_gtk_gtkentrybuffer_emit_deleted_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, emit_inserted_text,   arginfo_gtk_gtkentrybuffer_emit_inserted_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, get_bytes,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, get_length,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, get_max_length,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, insert_text,          arginfo_gtk_gtkentrybuffer_insert_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, set_max_length,       arginfo_gtk_gtkentrybuffer_set_max_length, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryBuffer, set_text,             arginfo_gtk_gtkentrybuffer_set_text, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolItemGroup, __construct)
{
	char *label;
	zend_bool free_label = FALSE;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolItemGroup);
	}

	wrapped_obj = (GObject *) gtk_tool_item_group_new(label);
	if (free_label) g_free(label);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolItemGroup);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToolItemGroup, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_tool_item_group_set_label(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkToolItemGroup, set_label_widget)
{
	zval *label_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &label_widget, gtkwidget_ce))
		return;

    gtk_tool_item_group_set_label_widget(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(label_widget)));

}


static PHP_METHOD(GtkToolItemGroup, set_collapsed)
{
	zend_bool collapsed;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &collapsed))
		return;

    gtk_tool_item_group_set_collapsed(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), (gboolean)collapsed);

}


static PHP_METHOD(GtkToolItemGroup, set_ellipsize)
{
	PangoEllipsizeMode ellipsize;
	zval *php_ellipsize = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_ellipsize))
		return;

	if (php_ellipsize && phpg_gvalue_get_enum(PANGO_TYPE_ELLIPSIZE_MODE, php_ellipsize, (gint *)&ellipsize) == FAILURE) {
		return;
	}

    gtk_tool_item_group_set_ellipsize(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), ellipsize);

}


static PHP_METHOD(GtkToolItemGroup, set_header_relief)
{
	GtkReliefStyle style;
	zval *php_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_style))
		return;

	if (php_style && phpg_gvalue_get_enum(GTK_TYPE_RELIEF_STYLE, php_style, (gint *)&style) == FAILURE) {
		return;
	}

    gtk_tool_item_group_set_header_relief(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), style);

}


static PHP_METHOD(GtkToolItemGroup, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_label(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkToolItemGroup, get_label_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_label_widget(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolItemGroup, get_collapsed)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_collapsed(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItemGroup, get_ellipsize)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_ellipsize(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItemGroup, get_header_relief)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_header_relief(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItemGroup, insert)
{
	zval *item;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &item, gtktoolitem_ce, &position))
		return;

    gtk_tool_item_group_insert(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM(PHPG_GOBJECT(item)), (gint)position);

}


static PHP_METHOD(GtkToolItemGroup, set_item_position)
{
	zval *item;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &item, gtktoolitem_ce, &position))
		return;

    gtk_tool_item_group_set_item_position(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM(PHPG_GOBJECT(item)), (gint)position);

}


static PHP_METHOD(GtkToolItemGroup, get_item_position)
{
	zval *item;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &item, gtktoolitem_ce))
		return;

    php_retval = gtk_tool_item_group_get_item_position(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM(PHPG_GOBJECT(item)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItemGroup, get_n_items)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_group_get_n_items(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItemGroup, get_nth_item)
{
	long index;
	GtkToolItem* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    php_retval = gtk_tool_item_group_get_nth_item(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), (guint)index);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolItemGroup, get_drop_item)
{
	long x, y;
	GtkToolItem* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_tool_item_group_get_drop_item(GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_gtk_tool_item_group_new, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_label_widget, 0)
    ZEND_ARG_OBJ_INFO(0, label_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_collapsed, 0)
    ZEND_ARG_INFO(0, collapsed)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_ellipsize, 0)
    ZEND_ARG_INFO(0, ellipsize)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_header_relief, 0)
    ZEND_ARG_INFO(0, style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_insert, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkToolItem, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_set_item_position, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkToolItem, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_get_item_position, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkToolItem, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_get_nth_item, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitemgroup_get_drop_item, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoolitemgroup_methods[] = {
	PHP_ME(GtkToolItemGroup, __construct,          arginfo_gtk_gtktoolitemgroup_gtk_tool_item_group_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_collapsed,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_drop_item,        arginfo_gtk_gtktoolitemgroup_get_drop_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_ellipsize,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_header_relief,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_item_position,    arginfo_gtk_gtktoolitemgroup_get_item_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_label_widget,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_n_items,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, get_nth_item,         arginfo_gtk_gtktoolitemgroup_get_nth_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, insert,               arginfo_gtk_gtktoolitemgroup_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_collapsed,        arginfo_gtk_gtktoolitemgroup_set_collapsed, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_ellipsize,        arginfo_gtk_gtktoolitemgroup_set_ellipsize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_header_relief,    arginfo_gtk_gtktoolitemgroup_set_header_relief, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_item_position,    arginfo_gtk_gtktoolitemgroup_set_item_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_label,            arginfo_gtk_gtktoolitemgroup_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItemGroup, set_label_widget,     arginfo_gtk_gtktoolitemgroup_set_label_widget, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolPalette, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolPalette);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolPalette);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToolPalette, set_group_position)
{
	zval *group;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &group, gtktoolitemgroup_ce, &position))
		return;

    gtk_tool_palette_set_group_position(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)), (gint)position);

}


static PHP_METHOD(GtkToolPalette, set_exclusive)
{
	zval *group;
	zend_bool exclusive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &group, gtktoolitemgroup_ce, &exclusive))
		return;

    gtk_tool_palette_set_exclusive(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)), (gboolean)exclusive);

}


static PHP_METHOD(GtkToolPalette, set_expand)
{
	zval *group;
	zend_bool expand;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &group, gtktoolitemgroup_ce, &expand))
		return;

    gtk_tool_palette_set_expand(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)), (gboolean)expand);

}


static PHP_METHOD(GtkToolPalette, get_group_position)
{
	zval *group;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &group, gtktoolitemgroup_ce))
		return;

    php_retval = gtk_tool_palette_get_group_position(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolPalette, get_exclusive)
{
	zval *group;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &group, gtktoolitemgroup_ce))
		return;

    php_retval = gtk_tool_palette_get_exclusive(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolPalette, get_expand)
{
	zval *group;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &group, gtktoolitemgroup_ce))
		return;

    php_retval = gtk_tool_palette_get_expand(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM_GROUP(PHPG_GOBJECT(group)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolPalette, set_icon_size)
{
	GtkIconSize icon_size;
	zval *php_icon_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_size))
		return;

	if (php_icon_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_icon_size, (gint *)&icon_size) == FAILURE) {
		return;
	}

    gtk_tool_palette_set_icon_size(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), icon_size);

}


static PHP_METHOD(GtkToolPalette, unset_icon_size)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tool_palette_unset_icon_size(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolPalette, set_toolbar_style)
{
	GtkToolbarStyle style;
	zval *php_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_style))
		return;

	if (php_style && phpg_gvalue_get_enum(GTK_TYPE_TOOLBAR_STYLE, php_style, (gint *)&style) == FAILURE) {
		return;
	}

    gtk_tool_palette_set_style(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), style);

}


static PHP_METHOD(GtkToolPalette, unset_toolbar_style)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tool_palette_unset_style(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolPalette, get_icon_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_palette_get_icon_size(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolPalette, get_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_palette_get_style(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolPalette, get_drop_item)
{
	long x, y;
	GtkToolItem* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_tool_palette_get_drop_item(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolPalette, get_drop_group)
{
	long x, y;
	GtkToolItemGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_tool_palette_get_drop_group(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolPalette, get_drag_item)
{
	GtkSelectionData *selection = NULL;
	zval *php_selection;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_selection, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_selection, GTK_TYPE_SELECTION_DATA, FALSE TSRMLS_CC)) {
        selection = (GtkSelectionData *) PHPG_GBOXED(php_selection);
    } else {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GtkSelectionData object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tool_palette_get_drag_item(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), selection);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolPalette, set_drag_source)
{
	GtkToolPaletteDragTargets targets;
	zval *php_targets = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_targets))
		return;

	if (php_targets && phpg_gvalue_get_flags(GTK_TYPE_TOOL_PALETTE_DRAG_TARGETS, php_targets, (gint *)&targets) == FAILURE) {
		return;
	}

    gtk_tool_palette_set_drag_source(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), targets);

}


static PHP_METHOD(GtkToolPalette, add_drag_dest)
{
	zval *widget, *php_flags = NULL, *php_targets = NULL, *php_actions = NULL;
	GtkDestDefaults flags;
	GtkToolPaletteDragTargets targets;
	GdkDragAction actions;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVV", &widget, gtkwidget_ce, &php_flags, &php_targets, &php_actions))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_DEST_DEFAULTS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

	if (php_targets && phpg_gvalue_get_flags(GTK_TYPE_TOOL_PALETTE_DRAG_TARGETS, php_targets, (gint *)&targets) == FAILURE) {
		return;
	}

	if (php_actions && phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
	}

    gtk_tool_palette_add_drag_dest(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), flags, targets, actions);

}


static PHP_METHOD(GtkToolPalette, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_palette_get_hadjustment(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolPalette, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_palette_get_vadjustment(GTK_TOOL_PALETTE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_group_position, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_exclusive, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
    ZEND_ARG_INFO(0, exclusive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_expand, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_group_position, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_exclusive, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_expand, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkToolItemGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_icon_size, 0)
    ZEND_ARG_INFO(0, icon_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_toolbar_style, 0)
    ZEND_ARG_INFO(0, style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_drop_item, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_drop_group, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_get_drag_item, 0)
    ZEND_ARG_OBJ_INFO(0, selection, GtkSelectionData, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_set_drag_source, 0)
    ZEND_ARG_OBJ_INFO(0, targets, GtkToolPaletteDragTargets, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolpalette_add_drag_dest, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, flags, GtkDestDefaults, 1)
    ZEND_ARG_OBJ_INFO(0, targets, GtkToolPaletteDragTargets, 1)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoolpalette_methods[] = {
	PHP_ME(GtkToolPalette, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, add_drag_dest,        arginfo_gtk_gtktoolpalette_add_drag_dest, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_drag_item,        arginfo_gtk_gtktoolpalette_get_drag_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_drop_group,       arginfo_gtk_gtktoolpalette_get_drop_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_drop_item,        arginfo_gtk_gtktoolpalette_get_drop_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_exclusive,        arginfo_gtk_gtktoolpalette_get_exclusive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_expand,           arginfo_gtk_gtktoolpalette_get_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_group_position,   arginfo_gtk_gtktoolpalette_get_group_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_icon_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_style,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_drag_source,      arginfo_gtk_gtktoolpalette_set_drag_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_exclusive,        arginfo_gtk_gtktoolpalette_set_exclusive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_expand,           arginfo_gtk_gtktoolpalette_set_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_group_position,   arginfo_gtk_gtktoolpalette_set_group_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_icon_size,        arginfo_gtk_gtktoolpalette_set_icon_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, set_toolbar_style,    arginfo_gtk_gtktoolpalette_set_toolbar_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, unset_icon_size,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolPalette, unset_toolbar_style,  NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererSpinner, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererSpinner);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererSpinner);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrendererspinner_methods[] = {
	PHP_ME(GtkCellRendererSpinner, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkOffscreenWindow, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkOffscreenWindow);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkOffscreenWindow);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkOffscreenWindow, get_pixmap)
{
	GdkPixmap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_offscreen_window_get_pixmap(GTK_OFFSCREEN_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkOffscreenWindow, get_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_offscreen_window_get_pixbuf(GTK_OFFSCREEN_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static zend_function_entry gtkoffscreenwindow_methods[] = {
	PHP_ME(GtkOffscreenWindow, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOffscreenWindow, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOffscreenWindow, get_pixmap,           NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSpinner, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinner);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinner);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkSpinner, start)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_spinner_start(GTK_SPINNER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkSpinner, stop)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_spinner_stop(GTK_SPINNER(PHPG_GOBJECT(this_ptr)));

}


static zend_function_entry gtkspinner_methods[] = {
	PHP_ME(GtkSpinner, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinner, start,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinner, stop,                 NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkComboBoxText, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBoxText);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBoxText);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkComboBoxText, append_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBoxText, insert_text)
{
	long position;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &position, &text, &free_text))
		return;

    gtk_combo_box_text_insert_text(GTK_COMBO_BOX_TEXT(PHPG_GOBJECT(this_ptr)), (gint)position, text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBoxText, prepend_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_combo_box_text_prepend_text(GTK_COMBO_BOX_TEXT(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBoxText, remove_text)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_combo_box_text_remove(GTK_COMBO_BOX_TEXT(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkComboBoxText, get_active_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcomboboxtext_append_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcomboboxtext_insert_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcomboboxtext_prepend_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcomboboxtext_remove_text, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcomboboxtext_methods[] = {
	PHP_ME(GtkComboBoxText, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxText, append_text,          arginfo_gtk_gtkcomboboxtext_append_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxText, get_active_text,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxText, insert_text,          arginfo_gtk_gtkcomboboxtext_insert_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxText, prepend_text,         arginfo_gtk_gtkcomboboxtext_prepend_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxText, remove_text,          arginfo_gtk_gtkcomboboxtext_remove_text, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPaperSize, __construct)
{
	char *name = NULL;
	zend_bool free_name = FALSE;

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|u", &name, &free_name)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_PAPER_SIZE;
    pobj->boxed = gtk_paper_size_new(name);
	if (free_name) g_free(name);

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}
    pobj->free_on_destroy = TRUE;
}


static PHP_METHOD(GtkPaperSize, new_from_ppd)
{
	char *ppd_name, *ppd_display_name;
	zend_bool free_ppd_name = FALSE, free_ppd_display_name = FALSE;
	double width, height;
	GtkPaperSize *wrapped_obj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uudd", &ppd_name, &free_ppd_name, &ppd_display_name, &free_ppd_display_name, &width, &height)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}

    wrapped_obj = gtk_paper_size_new_from_ppd(ppd_name, ppd_display_name, width, height);
	if (free_ppd_name) g_free(ppd_name);
	if (free_ppd_display_name) g_free(ppd_display_name);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}
    phpg_gboxed_new(&return_value, GTK_TYPE_PAPER_SIZE, wrapped_obj, FALSE, TRUE TSRMLS_CC);
}


static PHP_METHOD(GtkPaperSize, new_custom)
{
	char *name, *display_name;
	zend_bool free_name = FALSE, free_display_name = FALSE;
	double width, height;
	GtkUnit unit;
	zval *php_unit = NULL;
	GtkPaperSize *wrapped_obj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uuddV", &name, &free_name, &display_name, &free_display_name, &width, &height, &php_unit)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}

    wrapped_obj = gtk_paper_size_new_custom(name, display_name, width, height, unit);
	if (free_name) g_free(name);
	if (free_display_name) g_free(display_name);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPaperSize);
	}
    phpg_gboxed_new(&return_value, GTK_TYPE_PAPER_SIZE, wrapped_obj, FALSE, TRUE TSRMLS_CC);
}


static PHP_METHOD(GtkPaperSize, copy)
{
	GtkPaperSize *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_copy((GtkPaperSize *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_PAPER_SIZE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkPaperSize, is_equal)
{
	GtkPaperSize *size2 = NULL;
	zval *php_size2;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_size2, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_size2, GTK_TYPE_PAPER_SIZE, FALSE TSRMLS_CC)) {
        size2 = (GtkPaperSize *) PHPG_GBOXED(php_size2);
    } else {
        php_error(E_WARNING, "%s::%s() expects size2 argument to be a valid GtkPaperSize object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_paper_size_is_equal((GtkPaperSize *)PHPG_GBOXED(this_ptr), size2);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_get_name((GtkPaperSize *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPaperSize, get_display_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_get_display_name((GtkPaperSize *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPaperSize, get_ppd_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_get_ppd_name((GtkPaperSize *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPaperSize, get_width)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_width((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_height)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_height((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, is_custom)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_is_custom((GtkPaperSize *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPaperSize, set_size)
{
	double width, height;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddV", &width, &height, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_paper_size_set_size((GtkPaperSize *)PHPG_GBOXED(this_ptr), width, height, unit);

}


static PHP_METHOD(GtkPaperSize, get_default_top_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_default_top_margin((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_default_bottom_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_default_bottom_margin((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_default_left_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_default_left_margin((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_default_right_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_paper_size_get_default_right_margin((GtkPaperSize *)PHPG_GBOXED(this_ptr), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPaperSize, get_default)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paper_size_get_default();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkpapersize_paper_size_new, 0, 0, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_paper_size_new_from_ppd, 0)
    ZEND_ARG_INFO(0, ppd_name)
    ZEND_ARG_INFO(0, ppd_display_name)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_paper_size_new_custom, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, display_name)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_is_equal, 0)
    ZEND_ARG_OBJ_INFO(0, size2, GtkPaperSize, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_width, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_height, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_set_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_default_top_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_default_bottom_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_default_left_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpapersize_get_default_right_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

static zend_function_entry gtkpapersize_methods[] = {
	PHP_ME(GtkPaperSize, __construct,          arginfo_gtk_gtkpapersize_paper_size_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, new_from_ppd,         arginfo_gtk_gtkpapersize_paper_size_new_from_ppd, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPaperSize, new_custom,           arginfo_gtk_gtkpapersize_paper_size_new_custom, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPaperSize, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_default,          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPaperSize, get_default_bottom_margin, arginfo_gtk_gtkpapersize_get_default_bottom_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_default_left_margin, arginfo_gtk_gtkpapersize_get_default_left_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_default_right_margin, arginfo_gtk_gtkpapersize_get_default_right_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_default_top_margin, arginfo_gtk_gtkpapersize_get_default_top_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_display_name,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_height,           arginfo_gtk_gtkpapersize_get_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_ppd_name,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, get_width,            arginfo_gtk_gtkpapersize_get_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, is_custom,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, is_equal,             arginfo_gtk_gtkpapersize_is_equal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaperSize, set_size,             arginfo_gtk_gtkpapersize_set_size, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 237 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentInfo, get_uri)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    
    php_retval = gtk_recent_info_get_uri((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 2721 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentInfo, get_display_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_display_name((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkRecentInfo, get_description)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_description((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkRecentInfo, get_mime_type)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_mime_type((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkRecentInfo, get_added)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_added((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRecentInfo, get_modified)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_modified((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRecentInfo, get_visited)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_visited((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRecentInfo, get_private_hint)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_private_hint((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}

#line 267 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentInfo, get_application_info)
{
    gchar *app_name, *app_exec = NULL;
    gboolean ret;
    guint count = 0;
    time_t time_;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "s", &app_name))
        return;

    ret = gtk_recent_info_get_application_info(PHPG_GBOXED(this_ptr), app_name, &app_exec, &count, &time_);

    if (ret) {
        php_gtk_build_value(&return_value, "(sii)", app_exec, count, time_);
        g_free(app_exec);
    }
}
#line 2884 "ext/gtk+/gen_gtk.c"


#line 288 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentInfo, get_applications)
{
    gchar **apps;
    gsize length, i;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    apps = gtk_recent_info_get_applications(PHPG_GBOXED(this_ptr), &length);
    if (!apps)
        length = 0;
    array_init(return_value);
    for (i = 0; i < length; i++) {
        add_next_index_string(return_value, apps[i], 1);
    }

    g_strfreev(apps);
}
#line 2908 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentInfo, last_application)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_last_application((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkRecentInfo, has_application)
{
	char *app_name;
	zend_bool free_app_name = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &app_name, &free_app_name))
		return;

    php_retval = gtk_recent_info_has_application((GtkRecentInfo *)PHPG_GBOXED(this_ptr), app_name);
	if (free_app_name) g_free(app_name);
	RETVAL_BOOL(php_retval);
}

#line 204 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentInfo, get_groups)
{
    gchar **groups;
    gsize length, i;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    groups = gtk_recent_info_get_groups(PHPG_GBOXED(this_ptr), &length);

    if (!groups)
        length = 0;

    array_init(return_value);
    for (i = 0; i < length; i++) {
        add_next_index_string(return_value, groups[i], 1);
    }

    g_strfreev(groups);

}

#line 2980 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentInfo, has_group)
{
	char *group_name;
	zend_bool free_group_name = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &group_name, &free_group_name))
		return;

    php_retval = gtk_recent_info_has_group((GtkRecentInfo *)PHPG_GBOXED(this_ptr), group_name);
	if (free_group_name) g_free(group_name);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentInfo, get_icon)
{
	long size;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &size))
		return;

    php_retval = gtk_recent_info_get_icon((GtkRecentInfo *)PHPG_GBOXED(this_ptr), (gint)size);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkRecentInfo, get_short_name)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_short_name((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkRecentInfo, get_uri_display)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_uri_display((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkRecentInfo, get_age)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_get_age((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRecentInfo, is_local)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_is_local((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentInfo, exists)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_info_exists((GtkRecentInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentInfo, match)
{
	GtkRecentInfo *info_b = NULL;
	zval *php_info_b;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_info_b, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_info_b, GTK_TYPE_RECENT_INFO, FALSE TSRMLS_CC)) {
        info_b = (GtkRecentInfo *) PHPG_GBOXED(php_info_b);
    } else {
        php_error(E_WARNING, "%s::%s() expects info_b argument to be a valid GtkRecentInfo object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_recent_info_match((GtkRecentInfo *)PHPG_GBOXED(this_ptr), info_b);
	RETVAL_BOOL(php_retval);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentinfo_get_application_info, 0)
    ZEND_ARG_INFO(0, app_name)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentinfo_has_application, 0)
    ZEND_ARG_INFO(0, app_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentinfo_has_group, 0)
    ZEND_ARG_INFO(0, group_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentinfo_get_icon, 0)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentinfo_match, 0)
    ZEND_ARG_OBJ_INFO(0, info_b, GtkRecentInfo, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentinfo_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkRecentInfo, exists,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_added,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_age,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_application_info, arginfo_gtk_gtkrecentinfo_get_application_info, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_applications,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_description,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_display_name,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_groups,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_icon,             arginfo_gtk_gtkrecentinfo_get_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_mime_type,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_modified,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_private_hint,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_short_name,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_uri,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_uri_display,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, get_visited,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, has_application,      arginfo_gtk_gtkrecentinfo_has_application, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, has_group,            arginfo_gtk_gtkrecentinfo_has_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, is_local,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, last_application,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentInfo, match,                arginfo_gtk_gtkrecentinfo_match, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTargetList, add)
{
	GdkAtom target;
	zval *php_target = NULL;
	long flags, info;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vii", &php_target, &flags, &info))
		return;

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_target_list_add((GtkTargetList *)PHPG_GBOXED(this_ptr), target, (guint)flags, (guint)info);

}


static PHP_METHOD(GtkTargetList, add_image_targets)
{
	long info;
	zend_bool writable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &info, &writable))
		return;

    gtk_target_list_add_image_targets((GtkTargetList *)PHPG_GBOXED(this_ptr), (guint)info, (gboolean)writable);

}


static PHP_METHOD(GtkTargetList, add_text_targets)
{
	long info;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &info))
		return;

    gtk_target_list_add_text_targets((GtkTargetList *)PHPG_GBOXED(this_ptr), (guint)info);

}


static PHP_METHOD(GtkTargetList, add_uri_targets)
{
	long info;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &info))
		return;

    gtk_target_list_add_uri_targets((GtkTargetList *)PHPG_GBOXED(this_ptr), (guint)info);

}


static PHP_METHOD(GtkTargetList, remove)
{
	GdkAtom target;
	zval *php_target = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_target))
		return;

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_target_list_remove((GtkTargetList *)PHPG_GBOXED(this_ptr), target);

}


static PHP_METHOD(GtkTargetList, add_rich_text_targets)
{
	long info;
	zend_bool deserializable;
	zval *buffer;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ibO", &info, &deserializable, &buffer, gtktextbuffer_ce))
		return;

    gtk_target_list_add_rich_text_targets((GtkTargetList *)PHPG_GBOXED(this_ptr), (guint)info, (gboolean)deserializable, GTK_TEXT_BUFFER(PHPG_GOBJECT(buffer)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_add, 0)
    ZEND_ARG_INFO(0, target)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_INFO(0, info)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_add_image_targets, 0)
    ZEND_ARG_INFO(0, info)
    ZEND_ARG_INFO(0, writable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_add_text_targets, 0)
    ZEND_ARG_INFO(0, info)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_add_uri_targets, 0)
    ZEND_ARG_INFO(0, info)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_remove, 0)
    ZEND_ARG_INFO(0, target)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktargetlist_add_rich_text_targets, 0)
    ZEND_ARG_INFO(0, info)
    ZEND_ARG_INFO(0, deserializable)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkTextBuffer, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktargetlist_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkTargetList, add,                  arginfo_gtk_gtktargetlist_add, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTargetList, add_image_targets,    arginfo_gtk_gtktargetlist_add_image_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTargetList, add_rich_text_targets, arginfo_gtk_gtktargetlist_add_rich_text_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTargetList, add_text_targets,     arginfo_gtk_gtktargetlist_add_text_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTargetList, add_uri_targets,      arginfo_gtk_gtktargetlist_add_uri_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTargetList, remove,               arginfo_gtk_gtktargetlist_remove, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkBorder, __construct)
{

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkBorder);
	}

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_BORDER;
    pobj->boxed = gtk_border_new();

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkBorder);
	}
    pobj->free_on_destroy = TRUE;
}


static PHP_METHOD(GtkBorder, copy)
{
	GtkBorder *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_border_copy((GtkBorder *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_BORDER, php_retval, FALSE, TRUE TSRMLS_CC);

}


static zend_function_entry gtkborder_methods[] = {
	PHP_ME(GtkBorder, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBorder, copy,                 NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIconInfo, copy)
{
	GtkIconInfo *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_copy((GtkIconInfo *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_INFO, php_retval, FALSE, TRUE TSRMLS_CC);

}

#line 3787 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconInfo, get_attach_points)
{
    GdkPoint *points;
    int n_points;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (gtk_icon_info_get_attach_points((GtkIconInfo *)PHPG_GBOXED(this_ptr), &points,
                                        &n_points)) {
        int i;
        zval *item = NULL;

        array_init(return_value);
        for (i = 0; i < n_points; i++) {
            MAKE_STD_ZVAL(item);
            array_init(item);
            add_next_index_long(item, points[i].x);
            add_next_index_long(item, points[i].y);
            add_next_index_zval(return_value, item);
        }
        efree(points);
    }
}

#line 3428 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconInfo, get_base_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_get_base_size((GtkIconInfo *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconInfo, get_builtin_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_get_builtin_pixbuf((GtkIconInfo *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconInfo, get_display_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_get_display_name((GtkIconInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkIconInfo, get_embedded_rect)
{
	GdkRectangle rectangle = { 0, 0, 0, 0 };
	zval *php_rectangle;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_rectangle))
		return;

    if (phpg_rectangle_from_zval(php_rectangle, (GdkRectangle*)&rectangle TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects rectangle argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    php_retval = gtk_icon_info_get_embedded_rect((GtkIconInfo *)PHPG_GBOXED(this_ptr), &rectangle);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconInfo, get_filename)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_get_filename((GtkIconInfo *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkIconInfo, load_icon)
{
	GError *error = NULL;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_info_load_icon((GtkIconInfo *)PHPG_GBOXED(this_ptr), &error);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconInfo, set_raw_coordinates)
{
	zend_bool raw_coordinates;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &raw_coordinates))
		return;

    gtk_icon_info_set_raw_coordinates((GtkIconInfo *)PHPG_GBOXED(this_ptr), (gboolean)raw_coordinates);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconinfo_get_embedded_rect, 0)
    ZEND_ARG_OBJ_INFO(0, rectangle, GdkRectangle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconinfo_load_icon, 0)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconinfo_set_raw_coordinates, 0)
    ZEND_ARG_INFO(0, raw_coordinates)
ZEND_END_ARG_INFO();

static zend_function_entry gtkiconinfo_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkIconInfo, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_attach_points,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_base_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_builtin_pixbuf,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_display_name,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_embedded_rect,    arginfo_gtk_gtkiconinfo_get_embedded_rect, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, get_filename,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, load_icon,            arginfo_gtk_gtkiconinfo_load_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconInfo, set_raw_coordinates,  arginfo_gtk_gtkiconinfo_set_raw_coordinates, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIconSet, __construct)
{

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSet);
	}

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_ICON_SET;
    pobj->boxed = gtk_icon_set_new();

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSet);
	}
    pobj->free_on_destroy = TRUE;
}


static PHP_METHOD(GtkIconSet, new_from_pixbuf)
{
	zval *pixbuf;
	GtkIconSet *wrapped_obj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSet);
	}

    wrapped_obj = gtk_icon_set_new_from_pixbuf(GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSet);
	}
    phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SET, wrapped_obj, FALSE, TRUE TSRMLS_CC);
}


static PHP_METHOD(GtkIconSet, add_source)
{
	GtkIconSource *source = NULL;
	zval *php_source;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_source, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_source, GTK_TYPE_ICON_SOURCE, FALSE TSRMLS_CC)) {
        source = (GtkIconSource *) PHPG_GBOXED(php_source);
    } else {
        php_error(E_WARNING, "%s::%s() expects source argument to be a valid GtkIconSource object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_set_add_source((GtkIconSet *)PHPG_GBOXED(this_ptr), source);

}


static PHP_METHOD(GtkIconSet, copy)
{
	GtkIconSet *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_set_copy((GtkIconSet *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SET, php_retval, FALSE, TRUE TSRMLS_CC);

}

#line 3647 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconSet, get_sizes)
{
    GtkIconSize *sizes;
    gint n_sizes, i;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_icon_set_get_sizes((GtkIconSet *)PHPG_GBOXED(this_ptr), &sizes, &n_sizes);
    array_init(return_value);

    for (i = 0; i < n_sizes; i++) {
        add_next_index_long(return_value, sizes[i]);
    }

    g_free(sizes);
}
#line 3696 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconSet, render_icon)
{
	zval *style, *php_direction = NULL, *php_state = NULL, *php_size = NULL, *php_widget = NULL;
	GtkTextDirection direction;
	GtkStateType state;
	GtkIconSize size;
	GtkWidget *widget = NULL;
	char *detail = NULL;
	zend_bool free_detail = FALSE;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVV|Nu", &style, gtkstyle_ce, &php_direction, &php_state, &php_size, &php_widget, gtkwidget_ce, &detail, &free_detail))
		return;

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}
    if (php_widget) {
        if (Z_TYPE_P(php_widget) == IS_NULL)
            widget = NULL;
        else
            widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));
    }

    php_retval = gtk_icon_set_render_icon((GtkIconSet *)PHPG_GBOXED(this_ptr), GTK_STYLE(PHPG_GOBJECT(style)), direction, state, size, widget, detail);
	if (free_detail) g_free(detail);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconset_icon_set_new_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconset_add_source, 0)
    ZEND_ARG_OBJ_INFO(0, source, GtkIconSource, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkiconset_render_icon, 0, 0, 4)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
    ZEND_ARG_INFO(0, direction)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
ZEND_END_ARG_INFO();

static zend_function_entry gtkiconset_methods[] = {
	PHP_ME(GtkIconSet, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSet, new_from_pixbuf,      arginfo_gtk_gtkiconset_icon_set_new_from_pixbuf, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkIconSet, add_source,           arginfo_gtk_gtkiconset_add_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSet, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSet, get_sizes,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSet, render_icon,          arginfo_gtk_gtkiconset_render_icon, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIconSource, __construct)
{

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSource);
	}

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_ICON_SOURCE;
    pobj->boxed = gtk_icon_source_new();

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconSource);
	}
    pobj->free_on_destroy = TRUE;
}


static PHP_METHOD(GtkIconSource, copy)
{
	GtkIconSource *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_copy((GtkIconSource *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SOURCE, php_retval, FALSE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkIconSource, get_direction)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_direction((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconSource, get_direction_wildcarded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_direction_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconSource, get_filename)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_filename((GtkIconSource *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkIconSource, get_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_icon_name((GtkIconSource *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkIconSource, get_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_pixbuf((GtkIconSource *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconSource, get_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_size((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconSource, get_size_wildcarded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_size_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconSource, get_state)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_state((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconSource, get_state_wildcarded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_source_get_state_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconSource, set_direction)
{
	GtkTextDirection direction;
	zval *php_direction = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_direction))
		return;

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

    gtk_icon_source_set_direction((GtkIconSource *)PHPG_GBOXED(this_ptr), direction);

}


static PHP_METHOD(GtkIconSource, set_direction_wildcarded)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_icon_source_set_direction_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr), (gboolean)setting);

}


static PHP_METHOD(GtkIconSource, set_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_icon_source_set_filename((GtkIconSource *)PHPG_GBOXED(this_ptr), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkIconSource, set_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_icon_source_set_icon_name((GtkIconSource *)PHPG_GBOXED(this_ptr), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkIconSource, set_pixbuf)
{
	zval *pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce))
		return;

    gtk_icon_source_set_pixbuf((GtkIconSource *)PHPG_GBOXED(this_ptr), GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

}


static PHP_METHOD(GtkIconSource, set_size)
{
	GtkIconSize size;
	zval *php_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_icon_source_set_size((GtkIconSource *)PHPG_GBOXED(this_ptr), size);

}


static PHP_METHOD(GtkIconSource, set_size_wildcarded)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_icon_source_set_size_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr), (gboolean)setting);

}


static PHP_METHOD(GtkIconSource, set_state)
{
	GtkStateType state;
	zval *php_state = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_state))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    gtk_icon_source_set_state((GtkIconSource *)PHPG_GBOXED(this_ptr), state);

}


static PHP_METHOD(GtkIconSource, set_state_wildcarded)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_icon_source_set_state_wildcarded((GtkIconSource *)PHPG_GBOXED(this_ptr), (gboolean)setting);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_direction, 0)
    ZEND_ARG_INFO(0, direction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_direction_wildcarded, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_size, 0)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_size_wildcarded, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_state, 0)
    ZEND_ARG_INFO(0, state)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconsource_set_state_wildcarded, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

static zend_function_entry gtkiconsource_methods[] = {
	PHP_ME(GtkIconSource, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_direction,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_direction_wildcarded, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_filename,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_size_wildcarded,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_state,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, get_state_wildcarded, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_direction,        arginfo_gtk_gtkiconsource_set_direction, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_direction_wildcarded, arginfo_gtk_gtkiconsource_set_direction_wildcarded, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_filename,         arginfo_gtk_gtkiconsource_set_filename, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_icon_name,        arginfo_gtk_gtkiconsource_set_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_pixbuf,           arginfo_gtk_gtkiconsource_set_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_size,             arginfo_gtk_gtkiconsource_set_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_size_wildcarded,  arginfo_gtk_gtkiconsource_set_size_wildcarded, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_state,            arginfo_gtk_gtkiconsource_set_state, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconSource, set_state_wildcarded, arginfo_gtk_gtkiconsource_set_state_wildcarded, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

PHPG_PROP_READER(GtkRequisition, width)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkRequisition *)((phpg_gboxed_t *)object)->boxed)->width;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkRequisition, height)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkRequisition *)((phpg_gboxed_t *)object)->boxed)->height;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtkrequisition_prop_info[] = {
	{ "width", PHPG_PROP_READ_FN(GtkRequisition, width), NULL },
	{ "height", PHPG_PROP_READ_FN(GtkRequisition, height), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkRequisition, copy)
{
	GtkRequisition *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_requisition_copy((GtkRequisition *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_REQUISITION, php_retval, FALSE, TRUE TSRMLS_CC);

}


static zend_function_entry gtkrequisition_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkRequisition, copy,                 NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

PHPG_PROP_READER(GtkSelectionData, selection)
{
	GdkAtom php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkSelectionData *)((phpg_gboxed_t *)object)->boxed)->selection;
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkSelectionData, target)
{
	GdkAtom php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkSelectionData *)((phpg_gboxed_t *)object)->boxed)->target;
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkSelectionData, type)
{
	GdkAtom php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkSelectionData *)((phpg_gboxed_t *)object)->boxed)->type;
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkSelectionData, format)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkSelectionData *)((phpg_gboxed_t *)object)->boxed)->format;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkSelectionData, data)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkSelectionData *)((phpg_gboxed_t *)object)->boxed)->data;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


static prop_info_t gtkselectiondata_prop_info[] = {
	{ "selection", PHPG_PROP_READ_FN(GtkSelectionData, selection), NULL },
	{ "target", PHPG_PROP_READ_FN(GtkSelectionData, target), NULL },
	{ "type", PHPG_PROP_READ_FN(GtkSelectionData, type), NULL },
	{ "format", PHPG_PROP_READ_FN(GtkSelectionData, format), NULL },
	{ "data", PHPG_PROP_READ_FN(GtkSelectionData, data), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkSelectionData, copy)
{
	GtkSelectionData *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_copy((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_SELECTION_DATA, php_retval, FALSE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkSelectionData, get_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_pixbuf((GtkSelectionData *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSelectionData, get_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_text((GtkSelectionData *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 5231 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkSelectionData, get_uris)
{
    gchar **uris, **tmp;
    gchar *cp_str;
    gsize cp_len;
    zend_bool free_cp_str;
    GtkSelectionData *selection_data = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    array_init(return_value);

    selection_data = (GtkSelectionData *) PHPG_GBOXED(this_ptr);
    uris = gtk_selection_data_get_uris(selection_data);
    tmp = uris;
    while (*tmp) {
        cp_str = phpg_from_utf8(*tmp, strlen(*tmp), &cp_len, &free_cp_str TSRMLS_CC);
        if (cp_str) {
            add_next_index_stringl(return_value, cp_str, cp_len, 1);
            if (free_cp_str) {
                g_free(cp_str);
            }
        } else {
            zval_dtor(return_value);
            RETURN_NULL();
        }
        tmp++;
    }
    g_strfreev(uris);
}

#line 4416 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkSelectionData, set)
{
	GdkAtom type;
	zval *php_type = NULL;
	long format, length;
	char *data;
	zend_bool free_data = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Viui", &php_type, &format, &data, &free_data, &length))
		return;

    type = phpg_gdkatom_from_zval(php_type TSRMLS_CC);
    if (type == NULL) {
        php_error(E_WARNING, "%s::%s() expects type argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_selection_data_set((GtkSelectionData *)PHPG_GBOXED(this_ptr), type, (gint)format, data, (gint)length);
	if (free_data) g_free(data);

}


static PHP_METHOD(GtkSelectionData, set_pixbuf)
{
	zval *pixbuf;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce))
		return;

    php_retval = gtk_selection_data_set_pixbuf((GtkSelectionData *)PHPG_GBOXED(this_ptr), GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, set_text)
{
	char *str;
	zend_bool free_str = FALSE;
	long len = -1;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u|i", &str, &free_str, &len))
		return;

    php_retval = gtk_selection_data_set_text((GtkSelectionData *)PHPG_GBOXED(this_ptr), str, (gint)len);
	if (free_str) g_free(str);
	RETVAL_BOOL(php_retval);
}

#line 5272 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkSelectionData, set_uris)
{
    gchar **uris;
    zval *arr = NULL;
    zval **data;
    int i, arr_count;
    gboolean result;
    GtkSelectionData *selection_data = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a/", &arr)) {
        return;
    }

    arr_count = zend_hash_num_elements(Z_ARRVAL_P(arr));
    uris = safe_emalloc(arr_count+1, sizeof(gchar *), 0);

    for (i=0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(arr));
         zend_hash_get_current_data(Z_ARRVAL_P(arr), (void **)&data) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(arr))) {

        convert_to_string_ex(data);
        uris[i++] = Z_STRVAL_PP(data);

    }

    uris[i] = NULL;
    selection_data = (GtkSelectionData *) PHPG_GBOXED(this_ptr);
    result = gtk_selection_data_set_uris(selection_data, (gchar **)uris);
    efree(uris);
    RETURN_BOOL(result);
}

#line 4511 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkSelectionData, targets_include_image)
{
	zend_bool writable;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &writable))
		return;

    php_retval = gtk_selection_data_targets_include_image((GtkSelectionData *)PHPG_GBOXED(this_ptr), (gboolean)writable);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, targets_include_text)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_targets_include_text((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, tree_set_row_drag_data)
{
	zval *tree_model, *php_path;
	GtkTreePath *path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &tree_model, gtktreemodel_ce, &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_set_row_drag_data((GtkSelectionData *)PHPG_GBOXED(this_ptr), GTK_TREE_MODEL(PHPG_GOBJECT(tree_model)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, targets_include_rich_text)
{
	zval *buffer;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &buffer, gtktextbuffer_ce))
		return;

    php_retval = gtk_selection_data_targets_include_rich_text((GtkSelectionData *)PHPG_GBOXED(this_ptr), GTK_TEXT_BUFFER(PHPG_GOBJECT(buffer)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, targets_include_uri)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_targets_include_uri((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSelectionData, get_target)
{
	GdkAtom php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_target((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSelectionData, get_data_type)
{
	GdkAtom php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_data_type((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSelectionData, get_format)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_format((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSelectionData, get_data)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_data((GtkSelectionData *)PHPG_GBOXED(this_ptr));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkSelectionData, get_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_length((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSelectionData, get_display)
{
	GdkDisplay* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_display((GtkSelectionData *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSelectionData, get_selection)
{
	GdkAtom php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_selection_data_get_selection((GtkSelectionData *)PHPG_GBOXED(this_ptr));
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_set, 0)
    ZEND_ARG_INFO(0, type)
    ZEND_ARG_INFO(0, format)
    ZEND_ARG_INFO(0, data)
    ZEND_ARG_INFO(0, length)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_set_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkselectiondata_set_text, 0, 0, 1)
    ZEND_ARG_INFO(0, str)
    ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_set_uris, 0)
    ZEND_ARG_INFO(0, uris)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_targets_include_image, 0)
    ZEND_ARG_INFO(0, writable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_tree_set_row_drag_data, 0)
    ZEND_ARG_INFO(0, tree_model)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkselectiondata_targets_include_rich_text, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkTextBuffer, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkselectiondata_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkSelectionData, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_data,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_data_type,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_display,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_format,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_length,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_selection,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_target,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, get_uris,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, set,                  arginfo_gtk_gtkselectiondata_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, set_pixbuf,           arginfo_gtk_gtkselectiondata_set_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, set_text,             arginfo_gtk_gtkselectiondata_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, set_uris,             arginfo_gtk_gtkselectiondata_set_uris, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, targets_include_image, arginfo_gtk_gtkselectiondata_targets_include_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, targets_include_rich_text, arginfo_gtk_gtkselectiondata_targets_include_rich_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, targets_include_text, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, targets_include_uri,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSelectionData, tree_set_row_drag_data, arginfo_gtk_gtkselectiondata_tree_set_row_drag_data, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextAttributes, __construct)
{

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextAttributes);
	}

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_TEXT_ATTRIBUTES;
    pobj->boxed = gtk_text_attributes_new();

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextAttributes);
	}
    pobj->free_on_destroy = TRUE;
}

#line 5370 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, bg_color)
{
    GdkColor php_retval;

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.bg_color;
    phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}

#line 4808 "ext/gtk+/gen_gtk.c"


#line 5382 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, fg_color)
{
    GdkColor php_retval;

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.fg_color;
    phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}

#line 4822 "ext/gtk+/gen_gtk.c"


#line 5394 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, bg_stipple)
{
    GdkBitmap *php_retval;

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.bg_stipple;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);

    return SUCCESS;
}

#line 4836 "ext/gtk+/gen_gtk.c"


#line 5406 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, fg_stipple)
{
    GdkBitmap *php_retval;

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.fg_stipple;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);

    return SUCCESS;
}

#line 4850 "ext/gtk+/gen_gtk.c"


#line 5418 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, rise)
{
    RETVAL_LONG(((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.rise);

    return SUCCESS;
}

#line 4861 "ext/gtk+/gen_gtk.c"


#line 5427 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, underline)
{
    RETVAL_LONG(((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.underline);

    return SUCCESS;
}

#line 4872 "ext/gtk+/gen_gtk.c"


#line 5436 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, strikethrough)
{
    RETVAL_LONG(((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.strikethrough);

    return SUCCESS;
}

#line 4883 "ext/gtk+/gen_gtk.c"


#line 5445 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTextAttributes, draw_bg)
{
    RETVAL_BOOL(((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->appearance.draw_bg);

    return SUCCESS;
}
#line 4893 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkTextAttributes, justification)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->justification;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, direction)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->direction;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, font)
{
	PangoFontDescription *php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->font;
	phpg_gboxed_new(&return_value, PANGO_TYPE_FONT_DESCRIPTION, php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, font_scale)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->font_scale;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, left_margin)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->left_margin;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, indent)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->indent;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, right_margin)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->right_margin;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pixels_above_lines)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pixels_above_lines;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pixels_below_lines)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pixels_below_lines;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pixels_inside_wrap)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pixels_inside_wrap;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, tabs)
{
	PangoTabArray *php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->tabs;
	phpg_gboxed_new(&return_value, PANGO_TYPE_TAB_ARRAY, php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, wrap_mode)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->wrap_mode;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, language)
{
	PangoLanguage *php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->language;
	phpg_gboxed_new(&return_value, PANGO_TYPE_LANGUAGE, php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, invisible)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->invisible;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, bg_full_height)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->bg_full_height;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, editable)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->editable;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, realized)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->realized;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pad1)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pad1;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pad2)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pad2;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pad3)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pad3;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTextAttributes, pad4)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = ((GtkTextAttributes *)((phpg_gboxed_t *)object)->boxed)->pad4;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtktextattributes_prop_info[] = {
	{ "bg_color", PHPG_PROP_READ_FN(GtkTextAttributes, bg_color), NULL },
	{ "fg_color", PHPG_PROP_READ_FN(GtkTextAttributes, fg_color), NULL },
	{ "bg_stipple", PHPG_PROP_READ_FN(GtkTextAttributes, bg_stipple), NULL },
	{ "fg_stipple", PHPG_PROP_READ_FN(GtkTextAttributes, fg_stipple), NULL },
	{ "rise", PHPG_PROP_READ_FN(GtkTextAttributes, rise), NULL },
	{ "underline", PHPG_PROP_READ_FN(GtkTextAttributes, underline), NULL },
	{ "strikethrough", PHPG_PROP_READ_FN(GtkTextAttributes, strikethrough), NULL },
	{ "draw_bg", PHPG_PROP_READ_FN(GtkTextAttributes, draw_bg), NULL },
	{ "justification", PHPG_PROP_READ_FN(GtkTextAttributes, justification), NULL },
	{ "direction", PHPG_PROP_READ_FN(GtkTextAttributes, direction), NULL },
	{ "font", PHPG_PROP_READ_FN(GtkTextAttributes, font), NULL },
	{ "font_scale", PHPG_PROP_READ_FN(GtkTextAttributes, font_scale), NULL },
	{ "left_margin", PHPG_PROP_READ_FN(GtkTextAttributes, left_margin), NULL },
	{ "indent", PHPG_PROP_READ_FN(GtkTextAttributes, indent), NULL },
	{ "right_margin", PHPG_PROP_READ_FN(GtkTextAttributes, right_margin), NULL },
	{ "pixels_above_lines", PHPG_PROP_READ_FN(GtkTextAttributes, pixels_above_lines), NULL },
	{ "pixels_below_lines", PHPG_PROP_READ_FN(GtkTextAttributes, pixels_below_lines), NULL },
	{ "pixels_inside_wrap", PHPG_PROP_READ_FN(GtkTextAttributes, pixels_inside_wrap), NULL },
	{ "tabs", PHPG_PROP_READ_FN(GtkTextAttributes, tabs), NULL },
	{ "wrap_mode", PHPG_PROP_READ_FN(GtkTextAttributes, wrap_mode), NULL },
	{ "language", PHPG_PROP_READ_FN(GtkTextAttributes, language), NULL },
	{ "invisible", PHPG_PROP_READ_FN(GtkTextAttributes, invisible), NULL },
	{ "bg_full_height", PHPG_PROP_READ_FN(GtkTextAttributes, bg_full_height), NULL },
	{ "editable", PHPG_PROP_READ_FN(GtkTextAttributes, editable), NULL },
	{ "realized", PHPG_PROP_READ_FN(GtkTextAttributes, realized), NULL },
	{ "pad1", PHPG_PROP_READ_FN(GtkTextAttributes, pad1), NULL },
	{ "pad2", PHPG_PROP_READ_FN(GtkTextAttributes, pad2), NULL },
	{ "pad3", PHPG_PROP_READ_FN(GtkTextAttributes, pad3), NULL },
	{ "pad4", PHPG_PROP_READ_FN(GtkTextAttributes, pad4), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkTextAttributes, copy)
{
	GtkTextAttributes *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_attributes_copy((GtkTextAttributes *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ATTRIBUTES, php_retval, FALSE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextAttributes, copy_values)
{
	GtkTextAttributes *dest = NULL;
	zval *php_dest;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_dest, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_dest, GTK_TYPE_TEXT_ATTRIBUTES, FALSE TSRMLS_CC)) {
        dest = (GtkTextAttributes *) PHPG_GBOXED(php_dest);
    } else {
        php_error(E_WARNING, "%s::%s() expects dest argument to be a valid GtkTextAttributes object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_attributes_copy_values((GtkTextAttributes *)PHPG_GBOXED(this_ptr), dest);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextattributes_copy_values, 0)
    ZEND_ARG_OBJ_INFO(0, dest, GtkTextAttributes, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktextattributes_methods[] = {
	PHP_ME(GtkTextAttributes, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextAttributes, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextAttributes, copy_values,          arginfo_gtk_gtktextattributes_copy_values, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextIter, backward_char)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_char((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_chars)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_chars((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_cursor_position)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_cursor_position((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_cursor_positions)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_cursor_positions((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}

#line 484 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextIter, backward_find_char)
{
    zval *php_callback, *extra, *limit_iter = NULL;
    phpg_cb_data_t *cb_data;
    GtkTextCharPredicate callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "VN", &php_callback, &limit_iter, gtktextiter_ce))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            php_error(E_WARNING, "%s::%s(): You must specify a callback!", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTextCharPredicate)phpg_text_iter_char_predicate_func_marshal;
        }
    }

    if (Z_TYPE_P(limit_iter) != IS_NULL)
        gtk_text_iter_backward_find_char((GtkTextIter *)PHPG_GBOXED(this_ptr),
                                            callback, cb_data, (GtkTextIter *)PHPG_GBOXED(limit_iter));
    else
        gtk_text_iter_backward_find_char((GtkTextIter *)PHPG_GBOXED(this_ptr),
                                            callback, cb_data, NULL); 
}

#line 5366 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, backward_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_lines)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_lines((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_search)
{
	char *str;
	zend_bool free_str = FALSE;
	GtkTextSearchFlags flags;
	zval *php_flags = NULL, *php_match_start, *php_match_end, *php_limit;
	GtkTextIter *match_start = NULL, *match_end = NULL, *limit = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uVOON", &str, &free_str, &php_flags, &php_match_start, gboxed_ce, &php_match_end, gboxed_ce, &php_limit, gboxed_ce))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_TEXT_SEARCH_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    if (phpg_gboxed_check(php_match_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        match_start = (GtkTextIter *) PHPG_GBOXED(php_match_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects match_start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_match_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        match_end = (GtkTextIter *) PHPG_GBOXED(php_match_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects match_end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_limit) != IS_NULL) {
        if (phpg_gboxed_check(php_limit, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
            limit = (GtkTextIter *) PHPG_GBOXED(php_limit);
        } else {
            php_error(E_WARNING, "%s::%s() expects limit argument to be a valid GtkTextIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    php_retval = gtk_text_iter_backward_search((GtkTextIter *)PHPG_GBOXED(this_ptr), str, flags, match_start, match_end, limit);
	if (free_str) g_free(str);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_sentence_start)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_sentence_start((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_sentence_starts)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_sentence_starts((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_to_tag_toggle)
{
	GtkTextTag *tag = NULL;
	zval *php_tag;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_tag, gtktexttag_ce))
		return;
    if (Z_TYPE_P(php_tag) != IS_NULL)
        tag = GTK_TEXT_TAG(PHPG_GOBJECT(php_tag));

    php_retval = gtk_text_iter_backward_to_tag_toggle((GtkTextIter *)PHPG_GBOXED(this_ptr), tag);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_cursor_position)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_visible_cursor_position((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_cursor_positions)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_visible_cursor_positions((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_word_start)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_visible_word_start((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_word_starts)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_visible_word_starts((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_word_start)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_word_start((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_word_starts)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_word_starts((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, begins_tag)
{
	GtkTextTag *tag = NULL;
	zval *php_tag = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_tag, gtktexttag_ce))
		return;
    if (php_tag) {
        if (Z_TYPE_P(php_tag) == IS_NULL)
            tag = NULL;
        else
            tag = GTK_TEXT_TAG(PHPG_GOBJECT(php_tag));
    }

    php_retval = gtk_text_iter_begins_tag((GtkTextIter *)PHPG_GBOXED(this_ptr), tag);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, can_insert)
{
	zend_bool default_editability;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &default_editability))
		return;

    php_retval = gtk_text_iter_can_insert((GtkTextIter *)PHPG_GBOXED(this_ptr), (gboolean)default_editability);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, compare)
{
	GtkTextIter *rhs = NULL;
	zval *php_rhs;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_rhs, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_rhs, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        rhs = (GtkTextIter *) PHPG_GBOXED(php_rhs);
    } else {
        php_error(E_WARNING, "%s::%s() expects rhs argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_compare((GtkTextIter *)PHPG_GBOXED(this_ptr), rhs);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, copy)
{
	GtkTextIter *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_copy((GtkTextIter *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, php_retval, FALSE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextIter, editable)
{
	zend_bool default_setting;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &default_setting))
		return;

    php_retval = gtk_text_iter_editable((GtkTextIter *)PHPG_GBOXED(this_ptr), (gboolean)default_setting);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, ends_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_ends_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, ends_sentence)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_ends_sentence((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, ends_tag)
{
	GtkTextTag *tag = NULL;
	zval *php_tag = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_tag, gtktexttag_ce))
		return;
    if (php_tag) {
        if (Z_TYPE_P(php_tag) == IS_NULL)
            tag = NULL;
        else
            tag = GTK_TEXT_TAG(PHPG_GOBJECT(php_tag));
    }

    php_retval = gtk_text_iter_ends_tag((GtkTextIter *)PHPG_GBOXED(this_ptr), tag);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, ends_word)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_ends_word((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, equal)
{
	GtkTextIter *rhs = NULL;
	zval *php_rhs;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_rhs, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_rhs, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        rhs = (GtkTextIter *) PHPG_GBOXED(php_rhs);
    } else {
        php_error(E_WARNING, "%s::%s() expects rhs argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_equal((GtkTextIter *)PHPG_GBOXED(this_ptr), rhs);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_char)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_char((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_chars)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_chars((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_cursor_position)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_cursor_position((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_cursor_positions)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_cursor_positions((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}

#line 388 "ext/gtk+/gtktextview.overrides"

static gboolean phpg_text_iter_char_predicate_func_marshal (gunichar ch, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0; gint length;
    char *callback_name;
    zval *php_char;
    gchar *param = safe_emalloc(6, sizeof(gchar *), 0);
    gchar *cp_ret; gsize cp_len;
    zend_bool toreturn, free_result = FALSE;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    length = g_unichar_to_utf8(ch, param);
    cp_ret = phpg_from_utf8(param, length, &cp_len, &free_result TSRMLS_CC);
    
    if (cp_ret) {
        args = php_gtk_hash_as_array_offset(cbd->user_args, 1, &n_args);
        MAKE_STD_ZVAL(php_char);
        ZVAL_STRINGL(php_char, (char*)cp_ret, cp_len, 1);
        args[0] = &php_char;
    } else {
        php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return FALSE;
    }

    if (free_result)
        g_free(cp_ret);

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    if (retval) {
        toreturn = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        toreturn = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return toreturn;
}

static PHP_METHOD(GtkTextIter, forward_find_char)
{
    zval *php_callback, *extra, *limit_iter = NULL;
    phpg_cb_data_t *cb_data;
    GtkTextCharPredicate callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "VN", &php_callback, &limit_iter, gtktextiter_ce))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            php_error(E_WARNING, "%s::%s(): You must specify a callback!", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTextCharPredicate)phpg_text_iter_char_predicate_func_marshal;
        }
    }

    if (Z_TYPE_P(limit_iter) != IS_NULL)
        gtk_text_iter_forward_find_char((GtkTextIter *)PHPG_GBOXED(this_ptr),
                                            callback, cb_data, (GtkTextIter *)PHPG_GBOXED(limit_iter));
    else
        gtk_text_iter_forward_find_char((GtkTextIter *)PHPG_GBOXED(this_ptr),
                                            callback, cb_data, NULL); 
}

#line 5905 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, forward_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_lines)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_lines((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_search)
{
	char *str;
	zend_bool free_str = FALSE;
	GtkTextSearchFlags flags;
	zval *php_flags = NULL, *php_match_start, *php_match_end, *php_limit;
	GtkTextIter *match_start = NULL, *match_end = NULL, *limit = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uVOON", &str, &free_str, &php_flags, &php_match_start, gboxed_ce, &php_match_end, gboxed_ce, &php_limit, gboxed_ce))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_TEXT_SEARCH_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    if (phpg_gboxed_check(php_match_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        match_start = (GtkTextIter *) PHPG_GBOXED(php_match_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects match_start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_match_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        match_end = (GtkTextIter *) PHPG_GBOXED(php_match_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects match_end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_limit) != IS_NULL) {
        if (phpg_gboxed_check(php_limit, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
            limit = (GtkTextIter *) PHPG_GBOXED(php_limit);
        } else {
            php_error(E_WARNING, "%s::%s() expects limit argument to be a valid GtkTextIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    php_retval = gtk_text_iter_forward_search((GtkTextIter *)PHPG_GBOXED(this_ptr), str, flags, match_start, match_end, limit);
	if (free_str) g_free(str);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_sentence_end)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_sentence_end((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_sentence_ends)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_sentence_ends((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_to_end)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_text_iter_forward_to_end((GtkTextIter *)PHPG_GBOXED(this_ptr));

}


static PHP_METHOD(GtkTextIter, forward_to_line_end)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_to_line_end((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_to_tag_toggle)
{
	GtkTextTag *tag = NULL;
	zval *php_tag;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_tag, gtktexttag_ce))
		return;
    if (Z_TYPE_P(php_tag) != IS_NULL)
        tag = GTK_TEXT_TAG(PHPG_GOBJECT(php_tag));

    php_retval = gtk_text_iter_forward_to_tag_toggle((GtkTextIter *)PHPG_GBOXED(this_ptr), tag);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_cursor_position)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_visible_cursor_position((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_cursor_positions)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_visible_cursor_positions((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_word_end)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_visible_word_end((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_word_ends)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_visible_word_ends((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_word_end)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_word_end((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_word_ends)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_word_ends((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, get_attributes)
{
	GtkTextAttributes *values = NULL;
	zval *php_values;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_values, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_values, GTK_TYPE_TEXT_ATTRIBUTES, FALSE TSRMLS_CC)) {
        values = (GtkTextAttributes *) PHPG_GBOXED(php_values);
    } else {
        php_error(E_WARNING, "%s::%s() expects values argument to be a valid GtkTextAttributes object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_get_attributes((GtkTextIter *)PHPG_GBOXED(this_ptr), values);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, get_buffer)
{
	GtkTextBuffer* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_buffer((GtkTextIter *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextIter, get_bytes_in_line)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_bytes_in_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}

#line 278 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextIter, get_char)
{
    gunichar ichar;
    gint len;
    gchar *ret = safe_emalloc(6, sizeof(gchar *), 0);
    gchar *cp_ret;
    gsize cp_len;
    zend_bool free_result = FALSE;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    ichar = gtk_text_iter_get_char((GtkTextIter *)PHPG_GBOXED(this_ptr));

    if (ichar) {
        len = g_unichar_to_utf8(ichar, ret);
        cp_ret = phpg_from_utf8(ret, len, &cp_len, &free_result TSRMLS_CC);

        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }

        if (free_result)
            g_free(cp_ret);
        else
            RETVAL_STRINGL((char *)ichar, len, 1);
    } else {
        RETVAL_NULL();
    }

    efree(ret);
}

#line 6236 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, get_chars_in_line)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_chars_in_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_child_anchor)
{
	GtkTextChildAnchor* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_child_anchor((GtkTextIter *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextIter, get_language)
{
	PangoLanguage *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_language((GtkTextIter *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, PANGO_TYPE_LANGUAGE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextIter, get_line)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_line_index)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_line_index((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_line_offset)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_line_offset((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}

#line 337 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextIter, get_marks)
{
    GSList *marks, *mark;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    marks = gtk_text_iter_get_marks((GtkTextIter *)PHPG_GBOXED(this_ptr));

    array_init(return_value);
    for (mark = marks; mark; mark = mark->next) {
        zval *php_mark = NULL;
        phpg_gobject_new(&php_mark, G_OBJECT(mark->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_mark);
    }
    g_slist_free(marks);
}

#line 6344 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, get_offset)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_offset((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_pixbuf((GtkTextIter *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextIter, get_slice)
{
	GtkTextIter *end = NULL;
	zval *php_end;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_get_slice((GtkTextIter *)PHPG_GBOXED(this_ptr), end);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 316 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextIter, get_tags)
{
    GSList *tags, *tag;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    tags = gtk_text_iter_get_tags((GtkTextIter *)PHPG_GBOXED(this_ptr));

    array_init(return_value);
    for (tag = tags; tag; tag = tag->next) {
        zval *php_tag = NULL;
        phpg_gobject_new(&php_tag, G_OBJECT(tag->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_tag);
    }
    g_slist_free(tags);
}

#line 6432 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, get_text)
{
	GtkTextIter *end = NULL;
	zval *php_end;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_get_text((GtkTextIter *)PHPG_GBOXED(this_ptr), end);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 358 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextIter, get_toggled_tags)
{
    GSList *tags, *tag;
    zend_bool toggled_on = 1;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &toggled_on))
        return;

    tags = gtk_text_iter_get_toggled_tags((GtkTextIter *)PHPG_GBOXED(this_ptr), toggled_on);

    array_init(return_value);
    for (tag = tags; tag; tag = tag->next) {
        zval *php_tag = NULL;
        phpg_gobject_new(&php_tag, G_OBJECT(tag->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_tag);
    }
    g_slist_free(tags);
}

#line 6493 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextIter, get_visible_line_index)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_visible_line_index((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_visible_line_offset)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_get_visible_line_offset((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextIter, get_visible_slice)
{
	GtkTextIter *end = NULL;
	zval *php_end;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_get_visible_slice((GtkTextIter *)PHPG_GBOXED(this_ptr), end);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkTextIter, get_visible_text)
{
	GtkTextIter *end = NULL;
	zval *php_end;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_get_visible_text((GtkTextIter *)PHPG_GBOXED(this_ptr), end);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkTextIter, has_tag)
{
	zval *tag;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &tag, gtktexttag_ce))
		return;

    php_retval = gtk_text_iter_has_tag((GtkTextIter *)PHPG_GBOXED(this_ptr), GTK_TEXT_TAG(PHPG_GOBJECT(tag)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, in_range)
{
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_iter_in_range((GtkTextIter *)PHPG_GBOXED(this_ptr), start, end);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, inside_sentence)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_inside_sentence((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, inside_word)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_inside_word((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, is_cursor_position)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_is_cursor_position((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, is_end)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_is_end((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, is_start)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_is_start((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, order)
{
	GtkTextIter *second = NULL;
	zval *php_second;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_second, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_second, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        second = (GtkTextIter *) PHPG_GBOXED(php_second);
    } else {
        php_error(E_WARNING, "%s::%s() expects second argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_iter_order((GtkTextIter *)PHPG_GBOXED(this_ptr), second);

}


static PHP_METHOD(GtkTextIter, set_line)
{
	long line_number;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &line_number))
		return;

    gtk_text_iter_set_line((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)line_number);

}


static PHP_METHOD(GtkTextIter, set_line_index)
{
	long byte_on_line;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &byte_on_line))
		return;

    gtk_text_iter_set_line_index((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)byte_on_line);

}


static PHP_METHOD(GtkTextIter, set_line_offset)
{
	long char_on_line;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &char_on_line))
		return;

    gtk_text_iter_set_line_offset((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)char_on_line);

}


static PHP_METHOD(GtkTextIter, set_offset)
{
	long char_offset;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &char_offset))
		return;

    gtk_text_iter_set_offset((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)char_offset);

}


static PHP_METHOD(GtkTextIter, set_visible_line_index)
{
	long byte_on_line;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &byte_on_line))
		return;

    gtk_text_iter_set_visible_line_index((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)byte_on_line);

}


static PHP_METHOD(GtkTextIter, set_visible_line_offset)
{
	long char_on_line;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &char_on_line))
		return;

    gtk_text_iter_set_visible_line_offset((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)char_on_line);

}


static PHP_METHOD(GtkTextIter, starts_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_starts_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, starts_sentence)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_starts_sentence((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, starts_word)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_starts_word((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, toggles_tag)
{
	GtkTextTag *tag = NULL;
	zval *php_tag = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_tag, gtktexttag_ce))
		return;
    if (php_tag) {
        if (Z_TYPE_P(php_tag) == IS_NULL)
            tag = NULL;
        else
            tag = GTK_TEXT_TAG(PHPG_GOBJECT(php_tag));
    }

    php_retval = gtk_text_iter_toggles_tag((GtkTextIter *)PHPG_GBOXED(this_ptr), tag);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_backward_visible_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, backward_visible_lines)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_backward_visible_lines((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_line)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_iter_forward_visible_line((GtkTextIter *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextIter, forward_visible_lines)
{
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &count))
		return;

    php_retval = gtk_text_iter_forward_visible_lines((GtkTextIter *)PHPG_GBOXED(this_ptr), (gint)count);
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_chars, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_cursor_positions, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_find_char, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_lines, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_search, 0)
    ZEND_ARG_INFO(0, str)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_OBJ_INFO(0, match_start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, match_end, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, limit, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_sentence_starts, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_to_tag_toggle, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_visible_cursor_positions, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_visible_word_starts, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_word_starts, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextiter_begins_tag, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_can_insert, 0)
    ZEND_ARG_INFO(0, default_editability)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_compare, 0)
    ZEND_ARG_OBJ_INFO(0, rhs, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_editable, 0)
    ZEND_ARG_INFO(0, default_setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextiter_ends_tag, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_equal, 0)
    ZEND_ARG_OBJ_INFO(0, rhs, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_chars, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_cursor_positions, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_find_char, 0)
    ZEND_ARG_INFO(0, callback)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_lines, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_search, 0)
    ZEND_ARG_INFO(0, str)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_OBJ_INFO(0, match_start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, match_end, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, limit, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_sentence_ends, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_to_tag_toggle, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_visible_cursor_positions, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_visible_word_ends, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_word_ends, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_get_attributes, 0)
    ZEND_ARG_OBJ_INFO(0, values, GtkTextAttributes, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_get_slice, 0)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_get_text, 0)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_get_visible_slice, 0)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_get_visible_text, 0)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_has_tag, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_in_range, 0)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_order, 0)
    ZEND_ARG_OBJ_INFO(0, second, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_line, 0)
    ZEND_ARG_INFO(0, line_number)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_line_index, 0)
    ZEND_ARG_INFO(0, byte_on_line)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_line_offset, 0)
    ZEND_ARG_INFO(0, char_on_line)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_offset, 0)
    ZEND_ARG_INFO(0, char_offset)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_visible_line_index, 0)
    ZEND_ARG_INFO(0, byte_on_line)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_set_visible_line_offset, 0)
    ZEND_ARG_INFO(0, char_on_line)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextiter_toggles_tag, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_backward_visible_lines, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextiter_forward_visible_lines, 0)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

static zend_function_entry gtktextiter_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkTextIter, backward_char,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_chars,       arginfo_gtk_gtktextiter_backward_chars, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_cursor_position, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_cursor_positions, arginfo_gtk_gtktextiter_backward_cursor_positions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_find_char,   arginfo_gtk_gtktextiter_backward_find_char, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_line,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_lines,       arginfo_gtk_gtktextiter_backward_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_search,      arginfo_gtk_gtktextiter_backward_search, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_sentence_start, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_sentence_starts, arginfo_gtk_gtktextiter_backward_sentence_starts, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_to_tag_toggle, arginfo_gtk_gtktextiter_backward_to_tag_toggle, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_cursor_position, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_cursor_positions, arginfo_gtk_gtktextiter_backward_visible_cursor_positions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_line, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_lines, arginfo_gtk_gtktextiter_backward_visible_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_word_start, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_visible_word_starts, arginfo_gtk_gtktextiter_backward_visible_word_starts, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_word_start,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, backward_word_starts, arginfo_gtk_gtktextiter_backward_word_starts, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, begins_tag,           arginfo_gtk_gtktextiter_begins_tag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, can_insert,           arginfo_gtk_gtktextiter_can_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, compare,              arginfo_gtk_gtktextiter_compare, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, editable,             arginfo_gtk_gtktextiter_editable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, ends_line,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, ends_sentence,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, ends_tag,             arginfo_gtk_gtktextiter_ends_tag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, ends_word,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, equal,                arginfo_gtk_gtktextiter_equal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_char,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_chars,        arginfo_gtk_gtktextiter_forward_chars, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_cursor_position, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_cursor_positions, arginfo_gtk_gtktextiter_forward_cursor_positions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_find_char,    arginfo_gtk_gtktextiter_forward_find_char, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_line,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_lines,        arginfo_gtk_gtktextiter_forward_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_search,       arginfo_gtk_gtktextiter_forward_search, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_sentence_end, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_sentence_ends, arginfo_gtk_gtktextiter_forward_sentence_ends, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_to_end,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_to_line_end,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_to_tag_toggle, arginfo_gtk_gtktextiter_forward_to_tag_toggle, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_cursor_position, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_cursor_positions, arginfo_gtk_gtktextiter_forward_visible_cursor_positions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_line, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_lines, arginfo_gtk_gtktextiter_forward_visible_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_word_end, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_visible_word_ends, arginfo_gtk_gtktextiter_forward_visible_word_ends, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_word_end,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, forward_word_ends,    arginfo_gtk_gtktextiter_forward_word_ends, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_attributes,       arginfo_gtk_gtktextiter_get_attributes, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_buffer,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_bytes_in_line,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_char,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_chars_in_line,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_child_anchor,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_language,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_line,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_line_index,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_line_offset,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_marks,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_offset,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_slice,            arginfo_gtk_gtktextiter_get_slice, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_tags,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_text,             arginfo_gtk_gtktextiter_get_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_toggled_tags,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_visible_line_index, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_visible_line_offset, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_visible_slice,    arginfo_gtk_gtktextiter_get_visible_slice, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, get_visible_text,     arginfo_gtk_gtktextiter_get_visible_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, has_tag,              arginfo_gtk_gtktextiter_has_tag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, in_range,             arginfo_gtk_gtktextiter_in_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, inside_sentence,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, inside_word,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, is_cursor_position,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, is_end,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, is_start,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, order,                arginfo_gtk_gtktextiter_order, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_line,             arginfo_gtk_gtktextiter_set_line, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_line_index,       arginfo_gtk_gtktextiter_set_line_index, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_line_offset,      arginfo_gtk_gtktextiter_set_line_offset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_offset,           arginfo_gtk_gtktextiter_set_offset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_visible_line_index, arginfo_gtk_gtktextiter_set_visible_line_index, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, set_visible_line_offset, arginfo_gtk_gtktextiter_set_visible_line_offset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, starts_line,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, starts_sentence,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, starts_word,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextIter, toggles_tag,          arginfo_gtk_gtktextiter_toggles_tag, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static zend_function_entry gtktreeiter_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeRowReference, __construct)
{
	zval *model, *php_path;
	GtkTreePath *path;

    phpg_gboxed_t *pobj = NULL;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &model, gtktreemodel_ce, &php_path)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeRowReference);
	}

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeRowReference);
    }

    pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
    pobj->gtype = GTK_TYPE_TREE_ROW_REFERENCE;
    pobj->boxed = gtk_tree_row_reference_new(GTK_TREE_MODEL(PHPG_GOBJECT(model)), path);

    if (path)
        gtk_tree_path_free(path);

	if (!pobj->boxed) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeRowReference);
	}
    pobj->free_on_destroy = TRUE;
}


static PHP_METHOD(GtkTreeRowReference, copy)
{
	GtkTreeRowReference *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_row_reference_copy((GtkTreeRowReference *)PHPG_GBOXED(this_ptr));
	phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ROW_REFERENCE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTreeRowReference, get_path)
{
	GtkTreePath *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_row_reference_get_path((GtkTreeRowReference *)PHPG_GBOXED(this_ptr));

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
        gtk_tree_path_free(php_retval);
    }

}


static PHP_METHOD(GtkTreeRowReference, valid)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_row_reference_valid((GtkTreeRowReference *)PHPG_GBOXED(this_ptr));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeRowReference, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_row_reference_get_model((GtkTreeRowReference *)PHPG_GBOXED(this_ptr));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreerowreference_gtk_tree_row_reference_new, 0)
    ZEND_ARG_INFO(0, model)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreerowreference_methods[] = {
	PHP_ME(GtkTreeRowReference, __construct,          arginfo_gtk_gtktreerowreference_gtk_tree_row_reference_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeRowReference, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeRowReference, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeRowReference, get_path,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeRowReference, valid,                NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static zend_function_entry gtkctreenode_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	{ NULL, NULL, NULL }
};

void phpg_gtk1_register_classes(TSRMLS_D)
{

	gtkhsv_ce = phpg_register_class("GtkHsv", NULL, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_HSV TSRMLS_CC);

	gtkassistant_ce = phpg_register_class("GtkAssistant", gtkassistant_methods, gtkwindow_ce, 0, NULL, NULL, GTK_TYPE_ASSISTANT TSRMLS_CC);

	gtkwindowgroup_ce = phpg_register_class("GtkWindowGroup", gtkwindowgroup_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_WINDOW_GROUP TSRMLS_CC);

	gtkinfobar_ce = phpg_register_class("GtkInfoBar", gtkinfobar_methods, gtkhbox_ce, 0, NULL, NULL, GTK_TYPE_INFO_BAR TSRMLS_CC);

	gtkentrybuffer_ce = phpg_register_class("GtkEntryBuffer", gtkentrybuffer_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ENTRY_BUFFER TSRMLS_CC);

	gtktoolitemgroup_ce = phpg_register_class("GtkToolItemGroup", gtktoolitemgroup_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TOOL_ITEM_GROUP TSRMLS_CC);

	gtktoolpalette_ce = phpg_register_class("GtkToolPalette", gtktoolpalette_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TOOL_PALETTE TSRMLS_CC);

	gtkcellrendererspinner_ce = phpg_register_class("GtkCellRendererSpinner", gtkcellrendererspinner_methods, gtkcellrenderer_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_SPINNER TSRMLS_CC);

	gtkoffscreenwindow_ce = phpg_register_class("GtkOffscreenWindow", gtkoffscreenwindow_methods, gtkwindow_ce, 0, NULL, NULL, GTK_TYPE_OFFSCREEN_WINDOW TSRMLS_CC);

	gtkspinner_ce = phpg_register_class("GtkSpinner", gtkspinner_methods, gtkdrawingarea_ce, 0, NULL, NULL, GTK_TYPE_SPINNER TSRMLS_CC);

	gtkcomboboxtext_ce = phpg_register_class("GtkComboBoxText", gtkcomboboxtext_methods, gtkcombobox_ce, 0, NULL, NULL, GTK_TYPE_COMBO_BOX_TEXT TSRMLS_CC);

    gtkpapersize_ce = phpg_register_boxed("GtkPaperSize", gtkpapersize_methods, NULL, NULL, GTK_TYPE_PAPER_SIZE TSRMLS_CC);

    gtkrecentinfo_ce = phpg_register_boxed("GtkRecentInfo", gtkrecentinfo_methods, NULL, NULL, GTK_TYPE_RECENT_INFO TSRMLS_CC);

    gtktargetlist_ce = phpg_register_boxed("GtkTargetList", gtktargetlist_methods, NULL, NULL, GTK_TYPE_TARGET_LIST TSRMLS_CC);

    gtkborder_ce = phpg_register_boxed("GtkBorder", gtkborder_methods, NULL, NULL, GTK_TYPE_BORDER TSRMLS_CC);

    gtkiconinfo_ce = phpg_register_boxed("GtkIconInfo", gtkiconinfo_methods, NULL, NULL, GTK_TYPE_ICON_INFO TSRMLS_CC);

    gtkiconset_ce = phpg_register_boxed("GtkIconSet", gtkiconset_methods, NULL, NULL, GTK_TYPE_ICON_SET TSRMLS_CC);

    gtkiconsource_ce = phpg_register_boxed("GtkIconSource", gtkiconsource_methods, NULL, NULL, GTK_TYPE_ICON_SOURCE TSRMLS_CC);

    gtkrequisition_ce = phpg_register_boxed("GtkRequisition", gtkrequisition_methods, gtkrequisition_prop_info, NULL, GTK_TYPE_REQUISITION TSRMLS_CC);

    gtkselectiondata_ce = phpg_register_boxed("GtkSelectionData", gtkselectiondata_methods, gtkselectiondata_prop_info, NULL, GTK_TYPE_SELECTION_DATA TSRMLS_CC);

    gtktextattributes_ce = phpg_register_boxed("GtkTextAttributes", gtktextattributes_methods, gtktextattributes_prop_info, NULL, GTK_TYPE_TEXT_ATTRIBUTES TSRMLS_CC);

    gtktextiter_ce = phpg_register_boxed("GtkTextIter", gtktextiter_methods, NULL, NULL, GTK_TYPE_TEXT_ITER TSRMLS_CC);

    gtktreeiter_ce = phpg_register_boxed("GtkTreeIter", gtktreeiter_methods, NULL, NULL, GTK_TYPE_TREE_ITER TSRMLS_CC);

    gtktreerowreference_ce = phpg_register_boxed("GtkTreeRowReference", gtktreerowreference_methods, NULL, NULL, GTK_TYPE_TREE_ROW_REFERENCE TSRMLS_CC);

    gtkctreenode_ce = phpg_register_pointer("GtkCTreeNode", gtkctreenode_methods, NULL, GTK_TYPE_CTREE_NODE TSRMLS_CC);
}

void phpg_gtk_register_classes(TSRMLS_D)
{
	phpg_gtk0_register_classes(TSRMLS_C);
	phpg_gtk1_register_classes(TSRMLS_C);
}

void phpg_gtk_register_constants(const char *strip_prefix TSRMLS_DC)
{
	phpg_register_enum(GTK_TYPE_NUMBER_UP_LAYOUT, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_BUILDER_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ASSISTANT_PAGE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CELL_RENDERER_ACCEL_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PAGE_ORIENTATION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PAGE_SET, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_DUPLEX, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_OPERATION_ACTION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_OPERATION_RESULT, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_PAGES, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_QUALITY, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PRINT_STATUS, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RECENT_CHOOSER_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RECENT_MANAGER_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RECENT_SORT_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SENSITIVITY_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TEXT_BUFFER_TARGET_INFO, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TREE_VIEW_GRID_LINES, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_UNIT, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_RECENT_FILTER_FLAGS, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_FILE_CHOOSER_CONFIRMATION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ICON_VIEW_DROP_POSITION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PACK_DIRECTION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ANCHOR_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ARROW_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_BUTTON_BOX_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_BUTTONS_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CELL_RENDERER_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CELL_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CLIST_DRAG_POS, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CORNER_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CTREE_EXPANDER_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CTREE_EXPANSION_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CTREE_LINE_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CTREE_POS, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_CURVE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_DELETE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_DIRECTION_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_EXPANDER_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_FILE_CHOOSER_ACTION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_FILE_CHOOSER_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ICON_SIZE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ICON_THEME_ERROR, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_IMAGE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_IM_PREEDIT_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_IM_STATUS_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_JUSTIFICATION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_MATCH_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_MENU_DIRECTION_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_MESSAGE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_METRIC_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_MOVEMENT_STEP, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_NOTEBOOK_TAB, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ORIENTATION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PACK_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PATH_PRIORITY_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PATH_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_POLICY_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_POSITION_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PREVIEW_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PROGRESS_BAR_ORIENTATION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_PROGRESS_BAR_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RC_TOKEN_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RELIEF_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RESIZE_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_RESPONSE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SCROLL_STEP, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SCROLL_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SELECTION_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SHADOW_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SIDE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SIZE_GROUP_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SORT_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SPIN_BUTTON_UPDATE_POLICY, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SPIN_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_STATE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SUBMENU_DIRECTION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_SUBMENU_PLACEMENT, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TEXT_DIRECTION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TEXT_WINDOW_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TOOLBAR_CHILD_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TOOLBAR_SPACE_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TOOLBAR_STYLE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TREE_VIEW_COLUMN_SIZING, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TREE_VIEW_DROP_POSITION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_TREE_VIEW_MODE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_UPDATE_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_VISIBILITY, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_WIDGET_HELP_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_WINDOW_POSITION, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_WINDOW_TYPE, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_WRAP_MODE, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_ACCEL_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_ARG_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_ATTACH_OPTIONS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_BUTTON_ACTION, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_CALENDAR_DISPLAY_OPTIONS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_CELL_RENDERER_STATE, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_DEBUG_FLAG, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_DEST_DEFAULTS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_DIALOG_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_FILE_FILTER_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_ICON_LOOKUP_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_OBJECT_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_PRIVATE_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_RC_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_TARGET_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_TEXT_SEARCH_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_TREE_MODEL_FLAGS, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_UI_MANAGER_ITEM_TYPE, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_WIDGET_FLAGS, strip_prefix, gtk_ce);
	phpg_register_enum(GTK_TYPE_ENTRY_ICON_POSITION, strip_prefix, gtk_ce);
	phpg_register_flags(GTK_TYPE_TOOL_PALETTE_DRAG_TARGETS, strip_prefix, gtk_ce);


    /* register gtype constants for all classes */

	phpg_register_int_constant(gtkmountoperation_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MOUNT_OPERATION);
	phpg_register_int_constant(gtkbuilder_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BUILDER);
	phpg_register_int_constant(gtktooltip_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOLTIP);
	phpg_register_int_constant(gtkpagesetup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PAGE_SETUP);
	phpg_register_int_constant(gtkprintcontext_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PRINT_CONTEXT);
	phpg_register_int_constant(gtkprintoperation_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PRINT_OPERATION);
	phpg_register_int_constant(gtkprintsettings_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PRINT_SETTINGS);
	phpg_register_int_constant(gtkrecentmanager_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_MANAGER);
	phpg_register_int_constant(gtkstatusicon_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_STATUS_ICON);
	phpg_register_int_constant(phpgtkcustomtreemodel_ce, "gtype", sizeof("gtype")-1, PHPG_TYPE_CUSTOM_TREE_MODEL);
	phpg_register_int_constant(gtkaccelgroup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACCEL_GROUP);
	phpg_register_int_constant(gtkaccelmap_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACCEL_MAP);
	phpg_register_int_constant(gtkaccessible_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACCESSIBLE);
	phpg_register_int_constant(gtkaction_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACTION);
	phpg_register_int_constant(gtkactiongroup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACTION_GROUP);
	phpg_register_int_constant(gtkrecentaction_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_ACTION);
	phpg_register_int_constant(gtkclipboard_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CLIPBOARD);
	phpg_register_int_constant(gtkentrycompletion_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ENTRY_COMPLETION);
	phpg_register_int_constant(gtkiconfactory_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_FACTORY);
	phpg_register_int_constant(gtkicontheme_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_THEME);
	phpg_register_int_constant(gtkliststore_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LIST_STORE);
	phpg_register_int_constant(gtkobject_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_OBJECT);
	phpg_register_int_constant(gtkitemfactory_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ITEM_FACTORY);
	phpg_register_int_constant(gtkimcontext_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_IM_CONTEXT);
	phpg_register_int_constant(gtkfilefilter_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FILE_FILTER);
	phpg_register_int_constant(gtkimmulticontext_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_IM_MULTICONTEXT);
	phpg_register_int_constant(gtkimcontextsimple_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_IM_CONTEXT_SIMPLE);
	phpg_register_int_constant(gtkcellrenderer_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER);
	phpg_register_int_constant(gtkadjustment_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ADJUSTMENT);
	phpg_register_int_constant(gtkcellrenderertoggle_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_TOGGLE);
	phpg_register_int_constant(gtkcellrenderertext_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_TEXT);
	phpg_register_int_constant(gtkcellrendererprogress_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_PROGRESS);
	phpg_register_int_constant(gtkcellrendereraccel_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_ACCEL);
	phpg_register_int_constant(gtkcellrendererspin_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_SPIN);
	phpg_register_int_constant(gtkcellrenderercombo_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_COMBO);
	phpg_register_int_constant(gtkcellrendererpixbuf_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_PIXBUF);
	phpg_register_int_constant(gtkrecentfilter_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_FILTER);
	phpg_register_int_constant(gtkrcstyle_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RC_STYLE);
	phpg_register_int_constant(gtksettings_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SETTINGS);
	phpg_register_int_constant(gtksizegroup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SIZE_GROUP);
	phpg_register_int_constant(gtkstyle_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_STYLE);
	phpg_register_int_constant(gtktextbuffer_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_BUFFER);
	phpg_register_int_constant(gtktextchildanchor_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_CHILD_ANCHOR);
	phpg_register_int_constant(gtktextmark_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_MARK);
	phpg_register_int_constant(gtktexttag_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_TAG);
	phpg_register_int_constant(gtktexttagtable_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_TAG_TABLE);
	phpg_register_int_constant(gtktoggleaction_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOGGLE_ACTION);
	phpg_register_int_constant(gtkradioaction_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RADIO_ACTION);
	phpg_register_int_constant(gtktooltips_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOLTIPS);
	phpg_register_int_constant(gtktreemodelfilter_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_MODEL_FILTER);
	phpg_register_int_constant(gtktreemodelsort_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_MODEL_SORT);
	phpg_register_int_constant(gtktreeselection_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_SELECTION);
	phpg_register_int_constant(gtktreestore_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_STORE);
	phpg_register_int_constant(gtktreeviewcolumn_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_VIEW_COLUMN);
	phpg_register_int_constant(gtkuimanager_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_UI_MANAGER);
	phpg_register_int_constant(gtkwidget_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_WIDGET);
	phpg_register_int_constant(gtkseparator_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SEPARATOR);
	phpg_register_int_constant(gtkruler_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RULER);
	phpg_register_int_constant(gtkvseparator_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VSEPARATOR);
	phpg_register_int_constant(gtkhseparator_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HSEPARATOR);
	phpg_register_int_constant(gtkvruler_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VRULER);
	phpg_register_int_constant(gtkrange_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RANGE);
	phpg_register_int_constant(gtkhruler_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HRULER);
	phpg_register_int_constant(gtkscrollbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SCROLLBAR);
	phpg_register_int_constant(gtkprogress_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PROGRESS);
	phpg_register_int_constant(gtkvscrollbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VSCROLLBAR);
	phpg_register_int_constant(gtkhscrollbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HSCROLLBAR);
	phpg_register_int_constant(gtkscale_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SCALE);
	phpg_register_int_constant(gtkpreview_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PREVIEW);
	phpg_register_int_constant(gtkvscale_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VSCALE);
	phpg_register_int_constant(gtkhscale_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HSCALE);
	phpg_register_int_constant(gtkprogressbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PROGRESS_BAR);
	phpg_register_int_constant(gtkoldeditable_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_OLD_EDITABLE);
	phpg_register_int_constant(gtkmisc_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MISC);
	phpg_register_int_constant(gtkinvisible_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_INVISIBLE);
	phpg_register_int_constant(gtkpixmap_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PIXMAP);
	phpg_register_int_constant(gtkarrow_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ARROW);
	phpg_register_int_constant(gtkimage_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_IMAGE);
	phpg_register_int_constant(gtklabel_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LABEL);
	phpg_register_int_constant(gtkentry_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ENTRY);
	phpg_register_int_constant(gtktipsquery_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TIPS_QUERY);
	phpg_register_int_constant(gtkaccellabel_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ACCEL_LABEL);
	phpg_register_int_constant(gtkspinbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SPIN_BUTTON);
	phpg_register_int_constant(gtkdrawingarea_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_DRAWING_AREA);
	phpg_register_int_constant(gtkcontainer_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CONTAINER);
	phpg_register_int_constant(gtkcurve_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CURVE);
	phpg_register_int_constant(gtktreeview_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_VIEW);
	phpg_register_int_constant(gtktoolbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOLBAR);
	phpg_register_int_constant(gtktextview_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_VIEW);
	phpg_register_int_constant(gtktable_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TABLE);
	phpg_register_int_constant(gtksocket_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SOCKET);
	phpg_register_int_constant(gtkpaned_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PANED);
	phpg_register_int_constant(gtknotebook_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_NOTEBOOK);
	phpg_register_int_constant(gtkvpaned_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VPANED);
	phpg_register_int_constant(gtkhpaned_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HPANED);
	phpg_register_int_constant(gtkmenushell_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MENU_SHELL);
	phpg_register_int_constant(gtklist_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LIST);
	phpg_register_int_constant(gtkmenu_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MENU);
	phpg_register_int_constant(gtkmenubar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MENU_BAR);
	phpg_register_int_constant(gtkrecentchoosermenu_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_CHOOSER_MENU);
	phpg_register_int_constant(gtklayout_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LAYOUT);
	phpg_register_int_constant(gtkiconview_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_VIEW);
	phpg_register_int_constant(gtkfixed_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FIXED);
	phpg_register_int_constant(gtkcellview_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_VIEW);
	phpg_register_int_constant(gtkbin_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BIN);
	phpg_register_int_constant(gtkbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BOX);
	phpg_register_int_constant(gtkwindow_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_WINDOW);
	phpg_register_int_constant(gtkviewport_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VIEWPORT);
	phpg_register_int_constant(gtkdialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_DIALOG);
	phpg_register_int_constant(gtkplug_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PLUG);
	phpg_register_int_constant(gtkcolorselectiondialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COLOR_SELECTION_DIALOG);
	phpg_register_int_constant(gtkaboutdialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ABOUT_DIALOG);
	phpg_register_int_constant(gtkrecentchooserdialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_CHOOSER_DIALOG);
	phpg_register_int_constant(gtkfilechooserdialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FILE_CHOOSER_DIALOG);
	phpg_register_int_constant(gtkfileselection_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FILE_SELECTION);
	phpg_register_int_constant(gtkfontselectiondialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FONT_SELECTION_DIALOG);
	phpg_register_int_constant(gtkinputdialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_INPUT_DIALOG);
	phpg_register_int_constant(gtkmessagedialog_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MESSAGE_DIALOG);
	phpg_register_int_constant(gtktoolitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOL_ITEM);
	phpg_register_int_constant(gtkvbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VBOX);
	phpg_register_int_constant(gtkscrolledwindow_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SCROLLED_WINDOW);
	phpg_register_int_constant(gtkrecentchooserwidget_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_CHOOSER_WIDGET);
	phpg_register_int_constant(gtkcolorselection_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COLOR_SELECTION);
	phpg_register_int_constant(gtkfilechooserwidget_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FILE_CHOOSER_WIDGET);
	phpg_register_int_constant(gtkfontselection_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FONT_SELECTION);
	phpg_register_int_constant(gtkgammacurve_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_GAMMA_CURVE);
	phpg_register_int_constant(gtkseparatortoolitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SEPARATOR_TOOL_ITEM);
	phpg_register_int_constant(gtktoolbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOL_BUTTON);
	phpg_register_int_constant(gtkitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ITEM);
	phpg_register_int_constant(gtktoggletoolbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOGGLE_TOOL_BUTTON);
	phpg_register_int_constant(gtkmenutoolbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MENU_TOOL_BUTTON);
	phpg_register_int_constant(gtkradiotoolbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RADIO_TOOL_BUTTON);
	phpg_register_int_constant(gtkhandlebox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HANDLE_BOX);
	phpg_register_int_constant(gtkmenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_MENU_ITEM);
	phpg_register_int_constant(gtklistitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LIST_ITEM);
	phpg_register_int_constant(gtktearoffmenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEAROFF_MENU_ITEM);
	phpg_register_int_constant(gtkseparatormenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SEPARATOR_MENU_ITEM);
	phpg_register_int_constant(gtkcheckmenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CHECK_MENU_ITEM);
	phpg_register_int_constant(gtkimagemenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_IMAGE_MENU_ITEM);
	phpg_register_int_constant(gtkradiomenuitem_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RADIO_MENU_ITEM);
	phpg_register_int_constant(gtkhbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HBOX);
	phpg_register_int_constant(gtkframe_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FRAME);
	phpg_register_int_constant(gtkstatusbar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_STATUSBAR);
	phpg_register_int_constant(gtkcombo_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COMBO);
	phpg_register_int_constant(gtkfilechooserbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FILE_CHOOSER_BUTTON);
	phpg_register_int_constant(gtkexpander_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_EXPANDER);
	phpg_register_int_constant(gtkaspectframe_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ASPECT_FRAME);
	phpg_register_int_constant(gtkeventbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_EVENT_BOX);
	phpg_register_int_constant(gtkalignment_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ALIGNMENT);
	phpg_register_int_constant(gtkbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BUTTON);
	phpg_register_int_constant(gtkcombobox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COMBO_BOX);
	phpg_register_int_constant(gtktogglebutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOGGLE_BUTTON);
	phpg_register_int_constant(gtkoptionmenu_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_OPTION_MENU);
	phpg_register_int_constant(gtkcheckbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CHECK_BUTTON);
	phpg_register_int_constant(gtkfontbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_FONT_BUTTON);
	phpg_register_int_constant(gtkradiobutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RADIO_BUTTON);
	phpg_register_int_constant(gtklinkbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_LINK_BUTTON);
	phpg_register_int_constant(gtkscalebutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SCALE_BUTTON);
	phpg_register_int_constant(gtkcolorbutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COLOR_BUTTON);
	phpg_register_int_constant(gtkvolumebutton_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VOLUME_BUTTON);
	phpg_register_int_constant(gtkclist_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CLIST);
	phpg_register_int_constant(gtkcomboboxentry_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COMBO_BOX_ENTRY);
	phpg_register_int_constant(gtkbuttonbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BUTTON_BOX);
	phpg_register_int_constant(gtkcalendar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CALENDAR);
	phpg_register_int_constant(gtkvbuttonbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_VBUTTON_BOX);
	phpg_register_int_constant(gtkhbuttonbox_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HBUTTON_BOX);
	phpg_register_int_constant(gtkctree_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CTREE);
	phpg_register_int_constant(gtkhsv_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_HSV);
	phpg_register_int_constant(gtkassistant_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ASSISTANT);
	phpg_register_int_constant(gtkwindowgroup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_WINDOW_GROUP);
	phpg_register_int_constant(gtkinfobar_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_INFO_BAR);
	phpg_register_int_constant(gtkentrybuffer_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ENTRY_BUFFER);
	phpg_register_int_constant(gtktoolitemgroup_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOL_ITEM_GROUP);
	phpg_register_int_constant(gtktoolpalette_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TOOL_PALETTE);
	phpg_register_int_constant(gtkcellrendererspinner_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CELL_RENDERER_SPINNER);
	phpg_register_int_constant(gtkoffscreenwindow_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_OFFSCREEN_WINDOW);
	phpg_register_int_constant(gtkspinner_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SPINNER);
	phpg_register_int_constant(gtkcomboboxtext_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_COMBO_BOX_TEXT);
	phpg_register_int_constant(gtkpapersize_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_PAPER_SIZE);
	phpg_register_int_constant(gtkrecentinfo_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_RECENT_INFO);
	phpg_register_int_constant(gtktargetlist_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TARGET_LIST);
	phpg_register_int_constant(gtkborder_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_BORDER);
	phpg_register_int_constant(gtkiconinfo_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_INFO);
	phpg_register_int_constant(gtkiconset_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_SET);
	phpg_register_int_constant(gtkiconsource_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_ICON_SOURCE);
	phpg_register_int_constant(gtkrequisition_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_REQUISITION);
	phpg_register_int_constant(gtkselectiondata_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_SELECTION_DATA);
	phpg_register_int_constant(gtktextattributes_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_ATTRIBUTES);
	phpg_register_int_constant(gtktextiter_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TEXT_ITER);
	phpg_register_int_constant(gtktreeiter_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_ITER);
	phpg_register_int_constant(gtktreerowreference_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_TREE_ROW_REFERENCE);
	phpg_register_int_constant(gtkctreenode_ce, "gtype", sizeof("gtype")-1, GTK_TYPE_CTREE_NODE);

}

#endif /* HAVE_PHP_GTK */
