#include "php_gtk.h"

#if HAVE_PHP_GTK
#include "php_gtk_api.h"
#include "ext/standard/file.h"
#include "ext/gtk+/phpg_custom_tree_model.h"

static void phpg_about_dialog_activate_link_func_marshal (GtkAboutDialog *about, const gchar *link, gpointer data);
static gboolean phpg_entry_completion_match_func_marshal (GtkEntryCompletion *completion, const gchar *key, GtkTreeIter *iter, gpointer data);
static void phpg_file_filter_add_custom_marshal_add_file(const gchar *filename, int flags, int flag, zval *php_filter_info);
static gboolean phpg_file_filter_add_custom_marshal(GtkFileFilterInfo *filter_info, gpointer data);
static gboolean phpg_tree_view_row_separator_func_marshal (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void phpg_icon_view_foreach_func_marshal (GtkIconView *icon_view, GtkTreePath *path, gpointer data);
static gboolean phpg_toolbar_signal_func_marshal(GtkWidget *widget, gpointer data);
static void phpg_color_selection_change_palette_with_screen_func_marshal(GdkScreen *screen, const GdkColor *colors, gint n_colors);

/* forward declaration */
static PHP_METHOD(GtkImage, get_image);

#undef GDK_DISPLAY
#define GDK_DISPLAY(object) (GDK_DISPLAY_OBJECT(object))

#ifndef GDK_TYPE_REGION
extern phpg_region_get_type();
#define GDK_TYPE_REGION (phpg_region_get_type())
#endif

static ZEND_METHOD(GtkContainer, get_children);

extern PHP_GTK_EXPORT_CE(gdkdragcontext_ce);
static GtkTargetEntry* phpg_parse_target_entries(zval *php_targets, int *n_targets TSRMLS_DC);


static void phpg_text_tag_table_foreach_func_marshal (GtkTextTagTable *table, gpointer data);
static gboolean phpg_text_iter_char_predicate_func_marshal (gunichar ch, gpointer data);


#include "zend_interfaces.h"

static void phpg_cell_data_func_marshal(GtkCellLayout *cell_layout, GtkCellRenderer *renderer, GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void phpg_store_set_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool is_tree);
static gint phpg_tree_iter_compare_func_marshal(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data);
static gboolean phpg_tree_model_filter_visible_func_marshal(GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_view_row_separator_func_marshal (GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_view_column_drop_func_marshal (GtkTreeView *tree_view, GtkTreeViewColumn *column, GtkTreeViewColumn *prev_column, GtkTreeViewColumn *next_column, gpointer data);
static gboolean phpg_tree_view_search_equal_func_marshal (GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *iter, gpointer data);
static void phpg_tree_selection_foreach_func_marshal (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static gboolean phpg_tree_selection_func_marshal (GtkTreeSelection *selection, GtkTreeModel *model, GtkTreePath *path, gboolean path_currently_selected, gpointer data);

static void phpg_link_button_set_uri_hook_func_marshal(GtkLinkButton *button, const gchar *link, gpointer data);


typedef struct {
	zval *instance;
	zval *extra;
#ifdef ZTS
	TSRMLS_D;
#endif
} phpg_gtkbuilder_t;

static void phpg_scale_button_set_icons_assist(GtkScaleButton *, zval * TSRMLS_DC);

PHP_GTK_EXPORT_CE(gtktoolshell_ce);
PHP_GTK_EXPORT_CE(gtkbuildable_ce);
PHP_GTK_EXPORT_CE(gtkprintoperationpreview_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooser_ce);
PHP_GTK_EXPORT_CE(gtkcelleditable_ce);
PHP_GTK_EXPORT_CE(gtkcelllayout_ce);
PHP_GTK_EXPORT_CE(gtkeditable_ce);
PHP_GTK_EXPORT_CE(gtkfilechooser_ce);
PHP_GTK_EXPORT_CE(gtktreedragdest_ce);
PHP_GTK_EXPORT_CE(gtktreedragsource_ce);
PHP_GTK_EXPORT_CE(gtktreemodel_ce);
PHP_GTK_EXPORT_CE(gtktreesortable_ce);
PHP_GTK_EXPORT_CE(gtkactivatable_ce);
PHP_GTK_EXPORT_CE(gtkorientable_ce);
PHP_GTK_EXPORT_CE(gtkmountoperation_ce);
PHP_GTK_EXPORT_CE(gtkbuilder_ce);
PHP_GTK_EXPORT_CE(gtktooltip_ce);
PHP_GTK_EXPORT_CE(gtkpagesetup_ce);
PHP_GTK_EXPORT_CE(gtkprintcontext_ce);
PHP_GTK_EXPORT_CE(gtkprintoperation_ce);
PHP_GTK_EXPORT_CE(gtkprintsettings_ce);
PHP_GTK_EXPORT_CE(gtkrecentmanager_ce);
PHP_GTK_EXPORT_CE(gtkstatusicon_ce);
PHP_GTK_EXPORT_CE(phpgtkcustomtreemodel_ce);
PHP_GTK_EXPORT_CE(gtkaccelgroup_ce);
PHP_GTK_EXPORT_CE(gtkaccelmap_ce);
PHP_GTK_EXPORT_CE(gtkaccessible_ce);
PHP_GTK_EXPORT_CE(gtkaction_ce);
PHP_GTK_EXPORT_CE(gtkactiongroup_ce);
PHP_GTK_EXPORT_CE(gtkrecentaction_ce);
PHP_GTK_EXPORT_CE(gtkclipboard_ce);
PHP_GTK_EXPORT_CE(gtkentrycompletion_ce);
PHP_GTK_EXPORT_CE(gtkiconfactory_ce);
PHP_GTK_EXPORT_CE(gtkicontheme_ce);
PHP_GTK_EXPORT_CE(gtkliststore_ce);
PHP_GTK_EXPORT_CE(gtkobject_ce);
PHP_GTK_EXPORT_CE(gtkitemfactory_ce);
PHP_GTK_EXPORT_CE(gtkimcontext_ce);
PHP_GTK_EXPORT_CE(gtkfilefilter_ce);
PHP_GTK_EXPORT_CE(gtkimmulticontext_ce);
PHP_GTK_EXPORT_CE(gtkimcontextsimple_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderer_ce);
PHP_GTK_EXPORT_CE(gtkadjustment_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderertoggle_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderertext_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererprogress_ce);
PHP_GTK_EXPORT_CE(gtkcellrendereraccel_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererspin_ce);
PHP_GTK_EXPORT_CE(gtkcellrenderercombo_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererpixbuf_ce);
PHP_GTK_EXPORT_CE(gtkrecentfilter_ce);
PHP_GTK_EXPORT_CE(gtkrcstyle_ce);
PHP_GTK_EXPORT_CE(gtksettings_ce);
PHP_GTK_EXPORT_CE(gtksizegroup_ce);
PHP_GTK_EXPORT_CE(gtkstyle_ce);
PHP_GTK_EXPORT_CE(gtktextbuffer_ce);
PHP_GTK_EXPORT_CE(gtktextchildanchor_ce);
PHP_GTK_EXPORT_CE(gtktextmark_ce);
PHP_GTK_EXPORT_CE(gtktexttag_ce);
PHP_GTK_EXPORT_CE(gtktexttagtable_ce);
PHP_GTK_EXPORT_CE(gtktoggleaction_ce);
PHP_GTK_EXPORT_CE(gtkradioaction_ce);
PHP_GTK_EXPORT_CE(gtktooltips_ce);
PHP_GTK_EXPORT_CE(gtktreemodelfilter_ce);
PHP_GTK_EXPORT_CE(gtktreemodelsort_ce);
PHP_GTK_EXPORT_CE(gtktreeselection_ce);
PHP_GTK_EXPORT_CE(gtktreestore_ce);
PHP_GTK_EXPORT_CE(gtktreeviewcolumn_ce);
PHP_GTK_EXPORT_CE(gtkuimanager_ce);
PHP_GTK_EXPORT_CE(gtkwidget_ce);
PHP_GTK_EXPORT_CE(gtkseparator_ce);
PHP_GTK_EXPORT_CE(gtkruler_ce);
PHP_GTK_EXPORT_CE(gtkvseparator_ce);
PHP_GTK_EXPORT_CE(gtkhseparator_ce);
PHP_GTK_EXPORT_CE(gtkvruler_ce);
PHP_GTK_EXPORT_CE(gtkrange_ce);
PHP_GTK_EXPORT_CE(gtkhruler_ce);
PHP_GTK_EXPORT_CE(gtkscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkprogress_ce);
PHP_GTK_EXPORT_CE(gtkvscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkhscrollbar_ce);
PHP_GTK_EXPORT_CE(gtkscale_ce);
PHP_GTK_EXPORT_CE(gtkpreview_ce);
PHP_GTK_EXPORT_CE(gtkvscale_ce);
PHP_GTK_EXPORT_CE(gtkhscale_ce);
PHP_GTK_EXPORT_CE(gtkprogressbar_ce);
PHP_GTK_EXPORT_CE(gtkoldeditable_ce);
PHP_GTK_EXPORT_CE(gtkmisc_ce);
PHP_GTK_EXPORT_CE(gtkinvisible_ce);
PHP_GTK_EXPORT_CE(gtkpixmap_ce);
PHP_GTK_EXPORT_CE(gtkarrow_ce);
PHP_GTK_EXPORT_CE(gtkimage_ce);
PHP_GTK_EXPORT_CE(gtklabel_ce);
PHP_GTK_EXPORT_CE(gtkentry_ce);
PHP_GTK_EXPORT_CE(gtktipsquery_ce);
PHP_GTK_EXPORT_CE(gtkaccellabel_ce);
PHP_GTK_EXPORT_CE(gtkspinbutton_ce);
PHP_GTK_EXPORT_CE(gtkdrawingarea_ce);
PHP_GTK_EXPORT_CE(gtkcontainer_ce);
PHP_GTK_EXPORT_CE(gtkcurve_ce);
PHP_GTK_EXPORT_CE(gtktreeview_ce);
PHP_GTK_EXPORT_CE(gtktoolbar_ce);
PHP_GTK_EXPORT_CE(gtktextview_ce);
PHP_GTK_EXPORT_CE(gtktable_ce);
PHP_GTK_EXPORT_CE(gtksocket_ce);
PHP_GTK_EXPORT_CE(gtkpaned_ce);
PHP_GTK_EXPORT_CE(gtknotebook_ce);
PHP_GTK_EXPORT_CE(gtkvpaned_ce);
PHP_GTK_EXPORT_CE(gtkhpaned_ce);
PHP_GTK_EXPORT_CE(gtkmenushell_ce);
PHP_GTK_EXPORT_CE(gtklist_ce);
PHP_GTK_EXPORT_CE(gtkmenu_ce);
PHP_GTK_EXPORT_CE(gtkmenubar_ce);
PHP_GTK_EXPORT_CE(gtkrecentchoosermenu_ce);
PHP_GTK_EXPORT_CE(gtklayout_ce);
PHP_GTK_EXPORT_CE(gtkiconview_ce);
PHP_GTK_EXPORT_CE(gtkfixed_ce);
PHP_GTK_EXPORT_CE(gtkcellview_ce);
PHP_GTK_EXPORT_CE(gtkbin_ce);
PHP_GTK_EXPORT_CE(gtkbox_ce);
PHP_GTK_EXPORT_CE(gtkwindow_ce);
PHP_GTK_EXPORT_CE(gtkviewport_ce);
PHP_GTK_EXPORT_CE(gtkdialog_ce);
PHP_GTK_EXPORT_CE(gtkplug_ce);
PHP_GTK_EXPORT_CE(gtkcolorselectiondialog_ce);
PHP_GTK_EXPORT_CE(gtkaboutdialog_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooserdialog_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserdialog_ce);
PHP_GTK_EXPORT_CE(gtkfileselection_ce);
PHP_GTK_EXPORT_CE(gtkfontselectiondialog_ce);
PHP_GTK_EXPORT_CE(gtkinputdialog_ce);
PHP_GTK_EXPORT_CE(gtkmessagedialog_ce);
PHP_GTK_EXPORT_CE(gtktoolitem_ce);
PHP_GTK_EXPORT_CE(gtkvbox_ce);
PHP_GTK_EXPORT_CE(gtkscrolledwindow_ce);
PHP_GTK_EXPORT_CE(gtkrecentchooserwidget_ce);
PHP_GTK_EXPORT_CE(gtkcolorselection_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserwidget_ce);
PHP_GTK_EXPORT_CE(gtkfontselection_ce);
PHP_GTK_EXPORT_CE(gtkgammacurve_ce);
PHP_GTK_EXPORT_CE(gtkseparatortoolitem_ce);
PHP_GTK_EXPORT_CE(gtktoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkitem_ce);
PHP_GTK_EXPORT_CE(gtktoggletoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkmenutoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkradiotoolbutton_ce);
PHP_GTK_EXPORT_CE(gtkhandlebox_ce);
PHP_GTK_EXPORT_CE(gtkmenuitem_ce);
PHP_GTK_EXPORT_CE(gtklistitem_ce);
PHP_GTK_EXPORT_CE(gtktearoffmenuitem_ce);
PHP_GTK_EXPORT_CE(gtkseparatormenuitem_ce);
PHP_GTK_EXPORT_CE(gtkcheckmenuitem_ce);
PHP_GTK_EXPORT_CE(gtkimagemenuitem_ce);
PHP_GTK_EXPORT_CE(gtkradiomenuitem_ce);
PHP_GTK_EXPORT_CE(gtkhbox_ce);
PHP_GTK_EXPORT_CE(gtkframe_ce);
PHP_GTK_EXPORT_CE(gtkstatusbar_ce);
PHP_GTK_EXPORT_CE(gtkcombo_ce);
PHP_GTK_EXPORT_CE(gtkfilechooserbutton_ce);
PHP_GTK_EXPORT_CE(gtkexpander_ce);
PHP_GTK_EXPORT_CE(gtkaspectframe_ce);
PHP_GTK_EXPORT_CE(gtkeventbox_ce);
PHP_GTK_EXPORT_CE(gtkalignment_ce);
PHP_GTK_EXPORT_CE(gtkbutton_ce);
PHP_GTK_EXPORT_CE(gtkcombobox_ce);
PHP_GTK_EXPORT_CE(gtktogglebutton_ce);
PHP_GTK_EXPORT_CE(gtkoptionmenu_ce);
PHP_GTK_EXPORT_CE(gtkcheckbutton_ce);
PHP_GTK_EXPORT_CE(gtkfontbutton_ce);
PHP_GTK_EXPORT_CE(gtkradiobutton_ce);
PHP_GTK_EXPORT_CE(gtklinkbutton_ce);
PHP_GTK_EXPORT_CE(gtkscalebutton_ce);
PHP_GTK_EXPORT_CE(gtkcolorbutton_ce);
PHP_GTK_EXPORT_CE(gtkvolumebutton_ce);
PHP_GTK_EXPORT_CE(gtkclist_ce);
PHP_GTK_EXPORT_CE(gtkcomboboxentry_ce);
PHP_GTK_EXPORT_CE(gtkbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkcalendar_ce);
PHP_GTK_EXPORT_CE(gtkvbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkhbuttonbox_ce);
PHP_GTK_EXPORT_CE(gtkctree_ce);
PHP_GTK_EXPORT_CE(gtkhsv_ce);
PHP_GTK_EXPORT_CE(gtkassistant_ce);
PHP_GTK_EXPORT_CE(gtkwindowgroup_ce);
PHP_GTK_EXPORT_CE(gtkinfobar_ce);
PHP_GTK_EXPORT_CE(gtkentrybuffer_ce);
PHP_GTK_EXPORT_CE(gtktoolitemgroup_ce);
PHP_GTK_EXPORT_CE(gtktoolpalette_ce);
PHP_GTK_EXPORT_CE(gtkcellrendererspinner_ce);
PHP_GTK_EXPORT_CE(gtkoffscreenwindow_ce);
PHP_GTK_EXPORT_CE(gtkspinner_ce);
PHP_GTK_EXPORT_CE(gtkcomboboxtext_ce);
PHP_GTK_EXPORT_CE(gtkpapersize_ce);
PHP_GTK_EXPORT_CE(gtkrecentinfo_ce);
PHP_GTK_EXPORT_CE(gtktargetlist_ce);
PHP_GTK_EXPORT_CE(gtkborder_ce);
PHP_GTK_EXPORT_CE(gtkiconinfo_ce);
PHP_GTK_EXPORT_CE(gtkiconset_ce);
PHP_GTK_EXPORT_CE(gtkiconsource_ce);
PHP_GTK_EXPORT_CE(gtkrequisition_ce);
PHP_GTK_EXPORT_CE(gtkselectiondata_ce);
PHP_GTK_EXPORT_CE(gtktextattributes_ce);
PHP_GTK_EXPORT_CE(gtktextiter_ce);
PHP_GTK_EXPORT_CE(gtktreeiter_ce);
PHP_GTK_EXPORT_CE(gtktreerowreference_ce);
PHP_GTK_EXPORT_CE(gtkctreenode_ce);
PHP_GTK_EXPORT_CE(gtk_ce);
void phpg_gtk0_register_classes(TSRMLS_D);
void phpg_gtk1_register_classes(TSRMLS_D);


static PHP_METHOD(Gtk, accel_groups_activate)
{
	zval *object, *php_accel_mods = NULL;
	long accel_key;
	GdkModifierType accel_mods;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiV", &object, gobject_ce, &accel_key, &php_accel_mods))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_accel_groups_activate(G_OBJECT(PHPG_GOBJECT(object)), (guint)accel_key, accel_mods);
	RETVAL_BOOL(php_retval);
}

#line 774 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, accel_groups_from_object)
{
    zval *php_obj;
    GSList *list, *tmp;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_obj, gobject_ce)) {
        return;
    }

    list = gtk_accel_groups_from_object(PHPG_GOBJECT(php_obj));

    array_init(return_value);

    for (tmp = list; tmp; tmp = tmp->next) {
        zval *item = NULL;
        phpg_gobject_new(&item, G_OBJECT(tmp->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}

#line 311 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, accelerator_get_default_mod_mask)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accelerator_get_default_mod_mask();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, accelerator_get_label)
{
	long accelerator_key;
	GdkModifierType accelerator_mods;
	zval *php_accelerator_mods = NULL;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &accelerator_key, &php_accelerator_mods))
		return;

	if (php_accelerator_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accelerator_mods, (gint *)&accelerator_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_accelerator_get_label((guint)accelerator_key, accelerator_mods);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(Gtk, accelerator_name)
{
	long accelerator_key;
	GdkModifierType accelerator_mods;
	zval *php_accelerator_mods = NULL;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &accelerator_key, &php_accelerator_mods))
		return;

	if (php_accelerator_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accelerator_mods, (gint *)&accelerator_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_accelerator_name((guint)accelerator_key, accelerator_mods);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 726 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, accelerator_parse)
{
    char *accel;
    guint keyval;
    GdkModifierType modifier;
    zend_bool free_accel;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel, &free_accel))
        return;

    gtk_accelerator_parse((const gchar *)accel, &keyval, &modifier);
    if (free_accel)
        g_free(accel);

    php_gtk_build_value(&return_value, "(ii)", keyval, modifier);
}

#line 408 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, accelerator_set_default_mod_mask)
{
	GdkModifierType default_mod_mask;
	zval *php_default_mod_mask = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_default_mod_mask))
		return;

	if (php_default_mod_mask && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_default_mod_mask, (gint *)&default_mod_mask) == FAILURE) {
		return;
	}

    gtk_accelerator_set_default_mod_mask(default_mod_mask);

}


static PHP_METHOD(Gtk, accelerator_valid)
{
	long keyval;
	GdkModifierType modifiers;
	zval *php_modifiers = NULL;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &keyval, &php_modifiers))
		return;

	if (php_modifiers && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_modifiers, (gint *)&modifiers) == FAILURE) {
		return;
	}

    php_retval = gtk_accelerator_valid((guint)keyval, modifiers);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, check_version)
{
	long required_major, required_minor, required_micro;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iii", &required_major, &required_minor, &required_micro))
		return;

    php_retval = gtk_check_version((guint)required_major, (guint)required_minor, (guint)required_micro);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(Gtk, alternative_dialog_button_order)
{
	GdkScreen *screen = NULL;
	zval *php_screen = NULL;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_screen, gdkscreen_ce))
		return;
    if (php_screen) {
        if (Z_TYPE_P(php_screen) == IS_NULL)
            screen = NULL;
        else
            screen = GDK_SCREEN(PHPG_GOBJECT(php_screen));
    }

    php_retval = gtk_alternative_dialog_button_order(screen);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, disable_setlocale)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_disable_setlocale();

}


static PHP_METHOD(Gtk, drag_set_default_icon)
{
	zval *colormap, *pixmap, *mask;
	long hot_x, hot_y;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOOii", &colormap, gdkcolormap_ce, &pixmap, gdkpixmap_ce, &mask, gdkpixmap_ce, &hot_x, &hot_y))
		return;

	phpg_warn_deprecated("change the stock pixbuf for GTK_STOCK_DND" TSRMLS_CC);

    gtk_drag_set_default_icon(GDK_COLORMAP(PHPG_GOBJECT(colormap)), GDK_PIXMAP(PHPG_GOBJECT(pixmap)), GDK_PIXMAP(PHPG_GOBJECT(mask)), (gint)hot_x, (gint)hot_y);

}


static PHP_METHOD(Gtk, draw_insertion_cursor)
{
	zval *widget, *drawable, *php_area = NULL, *php_location, *php_direction = NULL;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area, location = { 0, 0, 0, 0 };
	zend_bool is_primary, draw_arrow;
	GtkTextDirection direction;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOVVbVb", &widget, gtkwidget_ce, &drawable, gdkdrawable_ce, &php_area, &php_location, &is_primary, &php_direction, &draw_arrow))
		return;

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (phpg_rectangle_from_zval(php_location, (GdkRectangle*)&location TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects location argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

    gtk_draw_insertion_cursor(GTK_WIDGET(PHPG_GOBJECT(widget)), GDK_DRAWABLE(PHPG_GOBJECT(drawable)), area, &location, (gboolean)is_primary, direction, (gboolean)draw_arrow);

}


static PHP_METHOD(Gtk, events_pending)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_events_pending();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, get_current_event)
{
	GdkEvent *php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_get_current_event();
	phpg_gboxed_new(&return_value, GDK_TYPE_EVENT, php_retval, TRUE, TRUE TSRMLS_CC);

}

#line 685 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, get_current_event_state)
{
    GdkModifierType state = 0;

    if (gtk_get_current_event_state(&state)) {
        RETURN_LONG(state);
    }
}

#line 586 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, get_current_event_time)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_get_current_event_time();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, get_default_language)
{
	PangoLanguage *php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_get_default_language();
	phpg_gboxed_new(&return_value, PANGO_TYPE_LANGUAGE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(Gtk, grab_get_current)
{
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_grab_get_current();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, icon_size_from_name)
{
	char *name;
	zend_bool free_name = FALSE;
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    php_retval = gtk_icon_size_from_name(name);
	if (free_name) g_free(name);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, icon_size_get_name)
{
	GtkIconSize size;
	zval *php_size = NULL;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    php_retval = gtk_icon_size_get_name(size);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 546 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, icon_size_lookup)
{
    gint width, height;
    zval *php_size = NULL;
    GtkIconSize size;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_size))
        return;

    if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
        return;
    }

    gtk_icon_size_lookup(size, &width, &height);

    php_gtk_build_value(&return_value, "(ii)", width, height);
}


#line 693 "ext/gtk+/gen_gtk.c"


#line 575 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, icon_size_lookup_for_settings)
{
    gint width, height;
    zval *php_settings;
    zval *php_size = NULL;
    GtkIconSize size;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &php_settings, gtksettings_ce, &php_size))
        return;

    if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE)
        return;


    if (gtk_icon_size_lookup_for_settings(GTK_SETTINGS(PHPG_GOBJECT(php_settings)), size, &width, &height)) {
        php_gtk_build_value(&return_value, "(ii)", width, height);
    } else {
        RETURN_FALSE;
    }
}


#line 719 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, icon_size_register)
{
	char *name;
	zend_bool free_name = FALSE;
	long width, height, php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uii", &name, &free_name, &width, &height))
		return;

    php_retval = gtk_icon_size_register(name, (gint)width, (gint)height);
	if (free_name) g_free(name);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, icon_size_register_alias)
{
	char *alias;
	zend_bool free_alias = FALSE;
	GtkIconSize target;
	zval *php_target = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &alias, &free_alias, &php_target))
		return;

	if (php_target && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_target, (gint *)&target) == FAILURE) {
		return;
	}

    gtk_icon_size_register_alias(alias, target);
	if (free_alias) g_free(alias);

}

#line 299 "ext/gtk+/gtk.overrides"

static void phpg_gtk_idle_add_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool with_priority)
{
    gint priority = G_PRIORITY_DEFAULT;
    zval *callback = NULL;
    zval *extra = NULL;
    zval *data = NULL;
    char *callback_filename;
    uint callback_lineno;
    guint handler_id;
    int req_args = with_priority ? 2 : 1;

    if (ZEND_NUM_ARGS() < req_args) {
        php_error(E_WARNING, "%s::%s() requires at least %d arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), req_args, ZEND_NUM_ARGS());
        return;
    }

    if (with_priority) {
        if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), req_args, &extra, "iV", &priority, &callback))
            return;
    } else {
        if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), req_args, &extra, "V", &callback))
            return;
    }

    callback_filename = zend_get_executed_filename(TSRMLS_C);
    callback_lineno = zend_get_executed_lineno(TSRMLS_C);
    if (!extra) {
        MAKE_STD_ZVAL(extra);
        array_init(extra);
    }
    php_gtk_build_value(&data, "(VNsi)", callback, extra, callback_filename, callback_lineno);

    handler_id = g_idle_add_full(priority, phpg_handler_marshal, data, phpg_destroy_notify);
    RETURN_LONG(handler_id);
}

static PHP_METHOD(Gtk, idle_add)
{
    phpg_gtk_idle_add_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
}

#line 802 "ext/gtk+/gen_gtk.c"


#line 345 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, idle_add_priority)
{
    phpg_gtk_idle_add_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
}

#line 811 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, idle_remove)
{
	long idle_handler_id;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &idle_handler_id))
		return;

    gtk_idle_remove((guint)idle_handler_id);

}


static PHP_METHOD(Gtk, input_remove)
{
	long input_handler_id;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &input_handler_id))
		return;

    gtk_input_remove((guint)input_handler_id);

}


static PHP_METHOD(Gtk, main)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_main();

}


static PHP_METHOD(Gtk, main_do_event)
{
	GdkEvent *event = NULL;
	zval *php_event;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_main_do_event(event);

}


static PHP_METHOD(Gtk, main_iteration)
{
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_main_iteration();
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, main_iteration_do)
{
	zend_bool blocking = TRUE;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &blocking))
		return;

    php_retval = gtk_main_iteration_do((gboolean)blocking);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, main_level)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_main_level();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(Gtk, main_quit)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_main_quit();

}

#line 378 "ext/gtk+/gtk.overrides"

static PHP_METHOD(Gtk, quit_add)
{
    guint main_level;
    zval *callback = NULL;
    zval *extra;
    zval *data = NULL;
    char *callback_filename;
    uint callback_lineno;
    guint handler_id;

    if (ZEND_NUM_ARGS() < 2) {
        php_error(E_WARNING, "%s::%s() requires at least 2 arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), ZEND_NUM_ARGS());
        return;
    }

    if (!php_gtk_parse_args(2, "iV", &main_level, &callback))
        return;

    callback_filename = zend_get_executed_filename(TSRMLS_C);
    callback_lineno = zend_get_executed_lineno(TSRMLS_C);
    extra = php_gtk_func_args_as_hash(ZEND_NUM_ARGS(), 2, ZEND_NUM_ARGS());
    if (!extra) {
        MAKE_STD_ZVAL(extra);
        array_init(extra);
    }
    php_gtk_build_value(&data, "(VNsi)", callback, extra, callback_filename, callback_lineno);

    handler_id = gtk_quit_add_full(main_level, phpg_handler_marshal, NULL, data, phpg_destroy_notify);
    RETURN_LONG(handler_id);
}

#line 953 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, quit_remove)
{
	long quit_handler_id;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &quit_handler_id))
		return;

    gtk_quit_remove((guint)quit_handler_id);

}


static PHP_METHOD(Gtk, rc_add_default_file)
{
	char *filename;
	zend_bool free_filename = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_rc_add_default_file(filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(Gtk, rc_find_module_in_path)
{
	char *module_file;
	zend_bool free_module_file = FALSE, free_result;
	gchar *php_retval, *cp_ret;
	gsize cp_len;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &module_file, &free_module_file))
		return;

    php_retval = gtk_rc_find_module_in_path(module_file);
	if (free_module_file) g_free(module_file);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 599 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, rc_get_default_files)
{
    gchar **ret;
    gchar *cp_str;
    gsize cp_len;
    zend_bool free_cp_str;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    ret = gtk_rc_get_default_files();
    if (ret) {
        array_init(return_value);
        while (*ret) {
            cp_str = phpg_from_utf8(*ret, strlen(*ret), &cp_len, &free_cp_str TSRMLS_CC);
            if (cp_str) {
                add_next_index_stringl(return_value, cp_str, cp_len, 1);
                if (free_cp_str) {
                    g_free(cp_str);
                }
            } else {
                zval_dtor(return_value);
                RETURN_NULL();
            }
            *ret++;
        }
    } else {
        RETURN_NULL();
    }
}


#line 1042 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, rc_get_im_module_file)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_get_im_module_file();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(Gtk, rc_get_im_module_path)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_get_im_module_path();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(Gtk, rc_get_module_dir)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_get_module_dir();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(Gtk, rc_get_style_by_paths)
{
	zval *settings, *php_type = NULL;
	char *widget_path, *class_path;
	zend_bool free_widget_path = FALSE, free_class_path = FALSE;
	GType type;
	GtkStyle* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OuuV", &settings, gtksettings_ce, &widget_path, &free_widget_path, &class_path, &free_class_path, &php_type))
		return;
    if ((type = phpg_gtype_from_zval(php_type TSRMLS_CC)) == 0) {
        return;
    }

    php_retval = gtk_rc_get_style_by_paths(GTK_SETTINGS(PHPG_GOBJECT(settings)), widget_path, class_path, type);
	if (free_widget_path) g_free(widget_path);
	if (free_class_path) g_free(class_path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, rc_get_theme_dir)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_get_theme_dir();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(Gtk, rc_parse)
{
	char *filename;
	zend_bool free_filename = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_rc_parse(filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(Gtk, rc_parse_string)
{
	char *rc_string;
	zend_bool free_rc_string = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &rc_string, &free_rc_string))
		return;

    gtk_rc_parse_string(rc_string);
	if (free_rc_string) g_free(rc_string);

}


static PHP_METHOD(Gtk, rc_reparse_all)
{
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_reparse_all();
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, rc_reparse_all_for_settings)
{
	zval *settings;
	zend_bool force_load;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &settings, gtksettings_ce, &force_load))
		return;

    php_retval = gtk_rc_reparse_all_for_settings(GTK_SETTINGS(PHPG_GOBJECT(settings)), (gboolean)force_load);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, rc_reset_styles)
{
	zval *settings;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &settings, gtksettings_ce))
		return;

    gtk_rc_reset_styles(GTK_SETTINGS(PHPG_GOBJECT(settings)));

}


static PHP_METHOD(Gtk, selection_owner_set_for_display)
{
	zval *display, *php_widget, *php_selection = NULL;
	GtkWidget *widget = NULL;
	GdkAtom selection;
	long time = GDK_CURRENT_TIME;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ONV|i", &display, gdkdisplay_ce, &php_widget, gtkwidget_ce, &php_selection, &time))
		return;
    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_selection_owner_set_for_display(GDK_DISPLAY(PHPG_GOBJECT(display)), widget, selection, (guint32)time);
	RETVAL_BOOL(php_retval);
}

#line 633 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, stock_list_ids)
{
    GSList *ids;

    array_init(return_value);
    for (ids = gtk_stock_list_ids(); ids; ids = ids->next) {
        add_next_index_string(return_value, (char *) ids->data, 1);
        g_free(ids->data);
    }
    g_slist_free(ids);
}

#line 1270 "ext/gtk+/gen_gtk.c"


#line 654 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, stock_lookup)
{
    char *stock_id;
    GtkStockItem item;
    gchar *cp = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "s", &stock_id))
        return;

    if (gtk_stock_lookup(stock_id, &item)) {
        cp = phpg_from_utf8(item.label, strlen(item.label), &cp_len, &free_cp TSRMLS_CC);
        if (cp) {
            item.label = cp;
        } else {
            php_error(E_WARNING, "Could not convert item label from UTF-8");
            return;
        }
        php_gtk_build_value(&return_value, "(ssiis)", item.stock_id, item.label,
                            item.modifier, item.keyval, item.translation_domain);
        if (free_cp) {
            g_free(cp);
        }
    } else {
        return;
    }
}

#line 1303 "ext/gtk+/gen_gtk.c"


#line 235 "ext/gtk+/gtk.overrides"

static void phpg_gtk_timeout_add_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool with_priority)
{
    guint interval;
    gint priority = G_PRIORITY_DEFAULT;
    zval *callback = NULL;
    zval *extra;
    zval *data = NULL;
    char *callback_filename;
    uint callback_lineno;
    guint handler_id;
    int req_args = with_priority ? 3 : 2;

    if (ZEND_NUM_ARGS() < req_args) {
        php_error(E_WARNING, "%s::%s() requires at least %d arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), req_args, ZEND_NUM_ARGS());
        return;
    }

    if (with_priority) {
        if (!php_gtk_parse_args(req_args, "iiV", &interval, &priority, &callback))
            return;
    } else {
        if (!php_gtk_parse_args(req_args, "iV", &interval, &callback))
            return;
    }

    callback_filename = zend_get_executed_filename(TSRMLS_C);
    callback_lineno = zend_get_executed_lineno(TSRMLS_C);
    extra = php_gtk_func_args_as_hash(ZEND_NUM_ARGS(), req_args, ZEND_NUM_ARGS());
    if (!extra) {
        MAKE_STD_ZVAL(extra);
        array_init(extra);
    }
    php_gtk_build_value(&data, "(VNsi)", callback, extra, callback_filename, callback_lineno);

    handler_id = g_timeout_add_full(priority, interval, phpg_handler_marshal, data, phpg_destroy_notify);
    RETURN_LONG(handler_id);
}

static PHP_METHOD(Gtk, timeout_add)
{
    phpg_gtk_timeout_add_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
}

#line 1353 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, timeout_remove)
{
	long timeout_handler_id;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &timeout_handler_id))
		return;

    gtk_timeout_remove((guint)timeout_handler_id);

}

#line 696 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, tooltips_data_get)
{
    zval *widget;
    GtkTooltipsData *data;
    zval *tooltips = NULL, *tt_widget = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce)) {
        return;
    }

    data = gtk_tooltips_data_get(GTK_WIDGET(PHPG_GOBJECT(widget)));
    if (!data) {
        return;
    }

    phpg_gobject_new(&tooltips, (GObject *) data->tooltips TSRMLS_CC);
    phpg_gobject_new(&tt_widget, (GObject *) data->widget TSRMLS_CC);
    php_gtk_build_value(&return_value, "(NNuu)", tooltips, tt_widget, data->tip_text,
                        data->tip_private);
}

#line 1390 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(Gtk, link_button_new)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_link_button_new(uri);
	if (free_uri) g_free(uri);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, recent_chooser_menu_new)
{
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_menu_new();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, recent_chooser_widget_new)
{
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_widget_new();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, icon_info_new_for_pixbuf)
{
	zval *icon_theme, *pixbuf;
	GtkIconInfo *php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &icon_theme, gtkicontheme_ce, &pixbuf, gdkpixbuf_ce))
		return;

    php_retval = gtk_icon_info_new_for_pixbuf(GTK_ICON_THEME(PHPG_GOBJECT(icon_theme)), GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_INFO, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(Gtk, show_uri)
{
	GdkScreen *screen = NULL;
	zval *php_screen;
	char *uri;
	zend_bool free_uri = FALSE;
	long timestamp;
	GError *error = NULL;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Nui", &php_screen, gdkscreen_ce, &uri, &free_uri, &timestamp))
		return;
    if (Z_TYPE_P(php_screen) != IS_NULL)
        screen = GDK_SCREEN(PHPG_GOBJECT(php_screen));

    php_retval = gtk_show_uri(screen, uri, (guint32)timestamp, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(Gtk, window_get_default_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_default_icon_name();
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(Gtk, combo_box_new_with_entry)
{
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_new_with_entry();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, combo_box_new_with_model_and_entry)
{
	zval *model;
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &model, gtktreemodel_ce))
		return;

    php_retval = gtk_combo_box_new_with_model_and_entry(GTK_TREE_MODEL(PHPG_GOBJECT(model)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(Gtk, combo_box_text_new_with_entry)
{
	GtkWidget* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_text_new_with_entry();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 292 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, timeout_add_priority)
{
    phpg_gtk_timeout_add_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
}

#line 1539 "ext/gtk+/gen_gtk.c"


#line 352 "ext/gtk+/gtk.overrides"
static PHP_METHOD(Gtk, get_version)
{
    guint major;
    guint minor;
    guint micro;
    char version[256];

    major = GTK_MAJOR_VERSION;
    minor = GTK_MINOR_VERSION;
    micro = GTK_MICRO_VERSION;

    snprintf(version, sizeof(version), "GTK %d.%d.%d", major, minor, micro);

    RETURN_STRING(version, 1);
}

#line 1559 "ext/gtk+/gen_gtk.c"


#line 414 "ext/gtk+/gtk.overrides"

static gboolean phpg_iowatch_marshal(GIOChannel *source, GIOCondition condition, gpointer user_data)
{
    zval *callback_data = (zval *) user_data;
    zval **callback, **extra = NULL, **stream_rsrc;
    zval **callback_filename = NULL, **callback_lineno = NULL;
    zval ***args = NULL, *z_condition = NULL;
    int n_args = 0;
    zval *retval = NULL;
    char *callback_name;
    gboolean result;
    TSRMLS_FETCH();

    /* Callback is always passed as the first element. */
    zend_hash_index_find(Z_ARRVAL_P(callback_data), 0, (void **)&callback);
    zend_hash_index_find(Z_ARRVAL_P(callback_data), 1, (void **)&stream_rsrc);
    zend_hash_index_find(Z_ARRVAL_P(callback_data), 2, (void **)&extra);
    zend_hash_index_find(Z_ARRVAL_P(callback_data), 3, (void **)&callback_filename);
    zend_hash_index_find(Z_ARRVAL_P(callback_data), 4, (void **)&callback_lineno);

    if (!zend_is_callable(*callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke handler callback '%s' specified in %s on line %ld", callback_name, Z_STRVAL_PP(callback_filename), Z_LVAL_PP(callback_lineno));
        efree(callback_name);
        return 0;
    }

    args = php_gtk_hash_as_array_offset(*extra, 2, &n_args);
    args[0] = stream_rsrc;
    MAKE_STD_ZVAL(z_condition);
    ZVAL_LONG(z_condition, condition);
    args[1] = &z_condition;

    call_user_function_ex(EG(function_table), NULL, *callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&z_condition);

    result = FALSE;
    if (retval) {
        result = zval_is_true(retval);
        zval_ptr_dtor(&retval);
    }

    efree(callback_name);
    if (args)
        efree(args);

    phpg_handle_marshaller_exception(TSRMLS_C);
    return result;
}

static void phpg_io_add_watch_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool with_priority)
{
    gint priority = G_PRIORITY_DEFAULT;
    GIOCondition condition;
    zval *stream_rsrc = NULL;
    zval *callback = NULL;
    zval *extra = NULL;
    zval *data = NULL;
    char *callback_filename;
    uint callback_lineno;
    guint handler_id;
    GIOChannel *channel = NULL;
    php_stream *stream = NULL;
    int fd;
    int req_args = with_priority ? 4 : 3;

    if (ZEND_NUM_ARGS() < req_args) {
        php_error(E_WARNING, "%s::%s() requires at least %d arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), req_args, ZEND_NUM_ARGS());
        return;
    }

    if (with_priority) {
        if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), req_args, &extra, "riVi", &stream_rsrc, &condition, &callback, &priority))
            return;
    } else {
        if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), req_args, &extra, "riV", &stream_rsrc, &condition, &callback))
            return;
    }

    php_stream_from_zval(stream, &stream_rsrc);
    if (php_stream_can_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT) == SUCCESS) {
        php_stream_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT, (void*)&fd, 0);
    } else {
        php_error(E_WARNING, "%s::%s() could not use stream of type '%s'",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), stream->ops->label);
        return;
    }

    if (!extra) {
        MAKE_STD_ZVAL(extra);
        array_init(extra);
    }

    channel = g_io_channel_unix_new(fd);
    callback_filename = zend_get_executed_filename(TSRMLS_C);
    callback_lineno = zend_get_executed_lineno(TSRMLS_C);
    php_gtk_build_value(&data, "(VVNsi)", callback, stream_rsrc, extra, callback_filename, callback_lineno);
    handler_id = g_io_add_watch_full(channel, priority, condition, phpg_iowatch_marshal, data, phpg_destroy_notify);
    g_io_channel_unref(channel);

    RETURN_LONG(handler_id);
}

static PHP_METHOD(Gtk, io_add_watch)
{
    phpg_io_add_watch_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
}

#line 1674 "ext/gtk+/gen_gtk.c"


#line 527 "ext/gtk+/gtk.overrides"

static PHP_METHOD(Gtk, io_add_watch_priority)
{
    phpg_io_add_watch_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
}

#line 1684 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accel_groups_activate, 0)
    ZEND_ARG_OBJ_INFO(0, object, GObject, 1)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
ZEND_END_ARG_INFO();
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accel_groups_from_object, 0)
    ZEND_ARG_OBJ_INFO(0, object, GObject, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accelerator_get_label, 0)
    ZEND_ARG_INFO(0, accelerator_key)
    ZEND_ARG_INFO(0, accelerator_mods)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accelerator_name, 0)
    ZEND_ARG_INFO(0, accelerator_key)
    ZEND_ARG_INFO(0, accelerator_mods)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accelerator_parse, 0)
    ZEND_ARG_INFO(0, accel_string)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accelerator_set_default_mod_mask, 0)
    ZEND_ARG_INFO(0, default_mod_mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_accelerator_valid, 0)
    ZEND_ARG_INFO(0, keyval)
    ZEND_ARG_INFO(0, modifiers)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_check_version, 0)
    ZEND_ARG_INFO(0, required_major)
    ZEND_ARG_INFO(0, required_minor)
    ZEND_ARG_INFO(0, required_micro)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtk_alternative_dialog_button_order, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_drag_set_default_icon, 0)
    ZEND_ARG_OBJ_INFO(0, colormap, GdkColormap, 1)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
    ZEND_ARG_INFO(0, hot_x)
    ZEND_ARG_INFO(0, hot_y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_draw_insertion_cursor, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, drawable, GdkDrawable, 1)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, location, GdkRectangle, 1)
    ZEND_ARG_INFO(0, is_primary)
    ZEND_ARG_INFO(0, direction)
    ZEND_ARG_INFO(0, draw_arrow)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_from_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_get_name, 0)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_lookup, 0)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_lookup_for_settings, 0)
    ZEND_ARG_OBJ_INFO(0, settings, GtkSettings, 0)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_register, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_size_register_alias, 0)
    ZEND_ARG_INFO(0, alias)
    ZEND_ARG_INFO(0, target)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_idle_remove, 0)
    ZEND_ARG_INFO(0, idle_handler_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_input_remove, 0)
    ZEND_ARG_INFO(0, input_handler_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_main_do_event, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtk_main_iteration_do, 0, 0, 0)
    ZEND_ARG_INFO(0, blocking)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_quit_add, 0)
    ZEND_ARG_INFO(0, main_level)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_quit_remove, 0)
    ZEND_ARG_INFO(0, quit_handler_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_add_default_file, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_find_module_in_path, 0)
    ZEND_ARG_INFO(0, module_file)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_get_style_by_paths, 0)
    ZEND_ARG_OBJ_INFO(0, settings, GtkSettings, 1)
    ZEND_ARG_INFO(0, widget_path)
    ZEND_ARG_INFO(0, class_path)
    ZEND_ARG_OBJ_INFO(0, type, GType, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_parse, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_parse_string, 0)
    ZEND_ARG_INFO(0, rc_string)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_reparse_all_for_settings, 0)
    ZEND_ARG_OBJ_INFO(0, settings, GtkSettings, 1)
    ZEND_ARG_INFO(0, force_load)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_rc_reset_styles, 0)
    ZEND_ARG_OBJ_INFO(0, settings, GtkSettings, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtk_selection_owner_set_for_display, 0, 0, 3)
    ZEND_ARG_OBJ_INFO(0, display, GdkDisplay, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, selection)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_stock_lookup, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_timeout_add, 0)
    ZEND_ARG_INFO(0, interval)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_timeout_remove, 0)
    ZEND_ARG_INFO(0, timeout_handler_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_link_button_new, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_icon_info_new_for_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, icon_theme, GtkIconTheme, 1)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_show_uri, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, timestamp)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_combo_box_new_with_model_and_entry, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtk_timeout_add_priority, 0)
    ZEND_ARG_INFO(0, interval)
    ZEND_ARG_INFO(0, priority)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtk_methods[] = {
	PHP_ME(Gtk, accel_groups_activate, arginfo_gtk_gtk_accel_groups_activate, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accel_groups_from_object, arginfo_gtk_gtk_accel_groups_from_object, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_get_default_mod_mask, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_get_label, arginfo_gtk_gtk_accelerator_get_label, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_name,     arginfo_gtk_gtk_accelerator_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_parse,    arginfo_gtk_gtk_accelerator_parse, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_set_default_mod_mask, arginfo_gtk_gtk_accelerator_set_default_mod_mask, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, accelerator_valid,    arginfo_gtk_gtk_accelerator_valid, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, alternative_dialog_button_order, arginfo_gtk_gtk_alternative_dialog_button_order, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, check_version,        arginfo_gtk_gtk_check_version, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, combo_box_new_with_entry, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, combo_box_new_with_model_and_entry, arginfo_gtk_gtk_combo_box_new_with_model_and_entry, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, combo_box_text_new_with_entry, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, disable_setlocale,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, drag_set_default_icon, arginfo_gtk_gtk_drag_set_default_icon, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, draw_insertion_cursor, arginfo_gtk_gtk_draw_insertion_cursor, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, events_pending,       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, get_current_event,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, get_current_event_state, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, get_current_event_time, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, get_default_language, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, get_version,          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, grab_get_current,     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_info_new_for_pixbuf, arginfo_gtk_gtk_icon_info_new_for_pixbuf, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_from_name,  arginfo_gtk_gtk_icon_size_from_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_get_name,   arginfo_gtk_gtk_icon_size_get_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_lookup,     arginfo_gtk_gtk_icon_size_lookup, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_lookup_for_settings, arginfo_gtk_gtk_icon_size_lookup_for_settings, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_register,   arginfo_gtk_gtk_icon_size_register, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, icon_size_register_alias, arginfo_gtk_gtk_icon_size_register_alias, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, idle_add,             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, idle_add_priority,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, idle_remove,          arginfo_gtk_gtk_idle_remove, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, input_remove,         arginfo_gtk_gtk_input_remove, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, io_add_watch,         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, io_add_watch_priority, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, link_button_new,      arginfo_gtk_gtk_link_button_new, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main,                 NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main_do_event,        arginfo_gtk_gtk_main_do_event, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main_iteration,       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main_iteration_do,    arginfo_gtk_gtk_main_iteration_do, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main_level,           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, main_quit,            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, quit_add,             arginfo_gtk_gtk_quit_add, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, quit_remove,          arginfo_gtk_gtk_quit_remove, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_add_default_file,  arginfo_gtk_gtk_rc_add_default_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_find_module_in_path, arginfo_gtk_gtk_rc_find_module_in_path, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_default_files, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_im_module_file, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_im_module_path, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_module_dir,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_style_by_paths, arginfo_gtk_gtk_rc_get_style_by_paths, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_get_theme_dir,     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_parse,             arginfo_gtk_gtk_rc_parse, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_parse_string,      arginfo_gtk_gtk_rc_parse_string, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_reparse_all,       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_reparse_all_for_settings, arginfo_gtk_gtk_rc_reparse_all_for_settings, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, rc_reset_styles,      arginfo_gtk_gtk_rc_reset_styles, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, recent_chooser_menu_new, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, recent_chooser_widget_new, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, selection_owner_set_for_display, arginfo_gtk_gtk_selection_owner_set_for_display, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, show_uri,             arginfo_gtk_gtk_show_uri, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, stock_list_ids,       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, stock_lookup,         arginfo_gtk_gtk_stock_lookup, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, timeout_add,          arginfo_gtk_gtk_timeout_add, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, timeout_add_priority, arginfo_gtk_gtk_timeout_add_priority, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, timeout_remove,       arginfo_gtk_gtk_timeout_remove, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, tooltips_data_get,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(Gtk, window_get_default_icon_name, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolShell, get_icon_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_icon_size(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_orientation(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, get_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_style(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, get_relief_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_relief_style(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, rebuild_menu)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tool_shell_rebuild_menu(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolShell, get_text_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_text_orientation(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, get_text_alignment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_text_alignment(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkToolShell, get_ellipsize_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_ellipsize_mode(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolShell, get_text_size_group)
{
	GtkSizeGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_shell_get_text_size_group(GTK_TOOL_SHELL(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static zend_function_entry gtktoolshell_methods[] = {
	ZEND_FENTRY(get_ellipsize_mode, ZEND_MN(GtkToolShell_get_ellipsize_mode),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_icon_size, ZEND_MN(GtkToolShell_get_icon_size),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_orientation, ZEND_MN(GtkToolShell_get_orientation),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_relief_style, ZEND_MN(GtkToolShell_get_relief_style),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_style, ZEND_MN(GtkToolShell_get_style),            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_text_alignment, ZEND_MN(GtkToolShell_get_text_alignment),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_text_orientation, ZEND_MN(GtkToolShell_get_text_orientation), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_text_size_group, ZEND_MN(GtkToolShell_get_text_size_group),  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(rebuild_menu, ZEND_MN(GtkToolShell_rebuild_menu),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkBuildable, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_buildable_set_name(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkBuildable, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_buildable_get_name(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkBuildable, add_child)
{
	zval *builder, *child;
	char *type;
	zend_bool free_type = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOu", &builder, gtkbuilder_ce, &child, gobject_ce, &type, &free_type))
		return;

    gtk_buildable_add_child(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)), GTK_BUILDER(PHPG_GOBJECT(builder)), G_OBJECT(PHPG_GOBJECT(child)), type);
	if (free_type) g_free(type);

}


static PHP_METHOD(GtkBuildable, construct_child)
{
	zval *builder;
	char *name;
	zend_bool free_name = FALSE;
	GObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &builder, gtkbuilder_ce, &name, &free_name))
		return;

    php_retval = gtk_buildable_construct_child(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)), GTK_BUILDER(PHPG_GOBJECT(builder)), name);
	if (free_name) g_free(name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkBuildable, parser_finished)
{
	zval *builder;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &builder, gtkbuilder_ce))
		return;

    gtk_buildable_parser_finished(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)), GTK_BUILDER(PHPG_GOBJECT(builder)));

}


static PHP_METHOD(GtkBuildable, get_internal_child)
{
	zval *builder;
	char *childname;
	zend_bool free_childname = FALSE;
	GObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &builder, gtkbuilder_ce, &childname, &free_childname))
		return;

    php_retval = gtk_buildable_get_internal_child(GTK_BUILDABLE(PHPG_GOBJECT(this_ptr)), GTK_BUILDER(PHPG_GOBJECT(builder)), childname);
	if (free_childname) g_free(childname);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuildable_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuildable_add_child, 0)
    ZEND_ARG_OBJ_INFO(0, builder, GtkBuilder, 1)
    ZEND_ARG_OBJ_INFO(0, child, GObject, 1)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuildable_construct_child, 0)
    ZEND_ARG_OBJ_INFO(0, builder, GtkBuilder, 1)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuildable_parser_finished, 0)
    ZEND_ARG_OBJ_INFO(0, builder, GtkBuilder, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuildable_get_internal_child, 0)
    ZEND_ARG_OBJ_INFO(0, builder, GtkBuilder, 1)
    ZEND_ARG_INFO(0, childname)
ZEND_END_ARG_INFO();

static zend_function_entry gtkbuildable_methods[] = {
	ZEND_FENTRY(add_child, ZEND_MN(GtkBuildable_add_child),            arginfo_gtk_gtkbuildable_add_child, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(construct_child, ZEND_MN(GtkBuildable_construct_child),      arginfo_gtk_gtkbuildable_construct_child, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_internal_child, ZEND_MN(GtkBuildable_get_internal_child),   arginfo_gtk_gtkbuildable_get_internal_child, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_name, ZEND_MN(GtkBuildable_get_name),             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(parser_finished, ZEND_MN(GtkBuildable_parser_finished),      arginfo_gtk_gtkbuildable_parser_finished, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_name, ZEND_MN(GtkBuildable_set_name),             arginfo_gtk_gtkbuildable_set_name, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPrintOperationPreview, render_page)
{
	long page_nr;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_nr))
		return;

    gtk_print_operation_preview_render_page(GTK_PRINT_OPERATION_PREVIEW(PHPG_GOBJECT(this_ptr)), (gint)page_nr);

}


static PHP_METHOD(GtkPrintOperationPreview, end_preview)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_print_operation_preview_end_preview(GTK_PRINT_OPERATION_PREVIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkPrintOperationPreview, is_selected)
{
	long page_nr;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_nr))
		return;

    php_retval = gtk_print_operation_preview_is_selected(GTK_PRINT_OPERATION_PREVIEW(PHPG_GOBJECT(this_ptr)), (gint)page_nr);
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperationpreview_render_page, 0)
    ZEND_ARG_INFO(0, page_nr)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperationpreview_is_selected, 0)
    ZEND_ARG_INFO(0, page_nr)
ZEND_END_ARG_INFO();

static zend_function_entry gtkprintoperationpreview_methods[] = {
	ZEND_FENTRY(end_preview, ZEND_MN(GtkPrintOperationPreview_end_preview),          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(is_selected, ZEND_MN(GtkPrintOperationPreview_is_selected),          arginfo_gtk_gtkprintoperationpreview_is_selected, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(render_page, ZEND_MN(GtkPrintOperationPreview_render_page),          arginfo_gtk_gtkprintoperationpreview_render_page, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentChooser, set_show_private)
{
	zend_bool show_private;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_private))
		return;

    gtk_recent_chooser_set_show_private(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_private);

}


static PHP_METHOD(GtkRecentChooser, get_show_private)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_show_private(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_show_not_found)
{
	zend_bool show_not_found;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_not_found))
		return;

    gtk_recent_chooser_set_show_not_found(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_not_found);

}


static PHP_METHOD(GtkRecentChooser, get_show_not_found)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_show_not_found(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_select_multiple)
{
	zend_bool select_multiple;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &select_multiple))
		return;

    gtk_recent_chooser_set_select_multiple(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)select_multiple);

}


static PHP_METHOD(GtkRecentChooser, get_select_multiple)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_select_multiple(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_limit)
{
	long limit;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &limit))
		return;

    gtk_recent_chooser_set_limit(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gint)limit);

}


static PHP_METHOD(GtkRecentChooser, get_limit)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_limit(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_local_only)
{
	zend_bool local_only;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &local_only))
		return;

    gtk_recent_chooser_set_local_only(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)local_only);

}


static PHP_METHOD(GtkRecentChooser, get_local_only)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_local_only(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_show_tips)
{
	zend_bool show_tips;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_tips))
		return;

    gtk_recent_chooser_set_show_tips(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_tips);

}


static PHP_METHOD(GtkRecentChooser, get_show_tips)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_show_tips(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_show_numbers)
{
	zend_bool show_numbers;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_numbers))
		return;

    gtk_recent_chooser_set_show_numbers(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_numbers);

}


static PHP_METHOD(GtkRecentChooser, get_show_numbers)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_show_numbers(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_show_icons)
{
	zend_bool show_icons;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_icons))
		return;

    gtk_recent_chooser_set_show_icons(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_icons);

}


static PHP_METHOD(GtkRecentChooser, get_show_icons)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_show_icons(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, set_sort_type)
{
	GtkRecentSortType sort_type;
	zval *php_sort_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sort_type))
		return;

	if (php_sort_type && phpg_gvalue_get_enum(GTK_TYPE_RECENT_SORT_TYPE, php_sort_type, (gint *)&sort_type) == FAILURE) {
		return;
	}

    gtk_recent_chooser_set_sort_type(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), sort_type);

}


static PHP_METHOD(GtkRecentChooser, get_sort_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_sort_type(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 321 "ext/gtk+/gtk-2.10.overrides"

static gint
phpg_recent_chooser_sort_func_marshal(GtkRecentInfo *a, GtkRecentInfo *b, gpointer data)
{
    char *callback;
    int argc = 0, status = 0;
    zval ***argv = NULL, *retval = NULL; 
    zval *zinfo_a = NULL, *zinfo_b = NULL;
    phpg_cb_data_t *cbd = (phpg_cb_data_t *)data;
  
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(
            E_WARNING,
            "Unable to invoke callback '%s' specified in %s on line %ld",
            callback,
            cbd->src_filename,
            cbd->src_lineno
        );
        
        efree(callback);
        return 0;
    }

    phpg_gboxed_new(&zinfo_a,GTK_TYPE_RECENT_INFO,a,TRUE,TRUE TSRMLS_CC);
    phpg_gboxed_new(&zinfo_b,GTK_TYPE_RECENT_INFO,b,TRUE,TRUE TSRMLS_CC);

    argv = php_gtk_hash_as_array_offset(cbd->user_args,2,&argc);
    argv[0] = &zinfo_a;
    argv[1] = &zinfo_b;

    call_user_function_ex(
        EG(function_table),
        NULL,
        cbd->callback,
        &retval,
        argc,
        argv,
        0,
        NULL TSRMLS_CC
    );

    zval_ptr_dtor(&zinfo_a);
    zval_ptr_dtor(&zinfo_b);

    if(retval) {
        convert_to_long(retval);
        status = Z_LVAL_P(retval);
        zval_ptr_dtor(&retval);
    } else {
        status = 0;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback);
    efree(argv);

    return status;
}

static PHP_METHOD(GtkRecentChooser, set_sort_func)
{
    zval *callback, *vararg;
    phpg_cb_data_t *calldata;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_varargs(ZEND_NUM_ARGS(),1,&vararg,"V",&callback)) {
        return;
    }

    zval_add_ref(&callback);
    calldata = phpg_cb_data_new(callback,vararg TSRMLS_CC);

    gtk_recent_chooser_set_sort_func(
        GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)),
        (GtkRecentSortFunc)phpg_recent_chooser_sort_func_marshal,
        calldata,
        phpg_cb_data_destroy
    );
}

#line 2681 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentChooser, set_current_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_chooser_set_current_uri(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, get_current_uri)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_current_uri(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkRecentChooser, get_current_item)
{
	GtkRecentInfo *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_current_item(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_RECENT_INFO, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkRecentChooser, select_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_chooser_select_uri(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooser, unselect_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    gtk_recent_chooser_unselect_uri(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);

}


static PHP_METHOD(GtkRecentChooser, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_recent_chooser_select_all(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkRecentChooser, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_recent_chooser_unselect_all(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));

}

#line 409 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentChooser, get_items)
{
    GList *list = NULL, *element = NULL;
    zval *current = NULL;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"")) {
        return;
    }

    list = gtk_recent_chooser_get_items(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);

    for(element = list; element; element = element->next) {
        MAKE_STD_ZVAL(current);
        phpg_gboxed_new(
            &current,
            GTK_TYPE_RECENT_INFO,
            element->data,
            TRUE,
            TRUE TSRMLS_CC
        );

        add_next_index_zval(return_value,current);
    }

    g_list_foreach(list,(GFunc)gtk_recent_info_unref,NULL);
    g_list_free(list);
}

#line 2843 "ext/gtk+/gen_gtk.c"


#line 443 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentChooser, get_uris)
{
    gchar **list;
    gsize size, iter;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"")) {
        return;
    }

    list = gtk_recent_chooser_get_uris(
        GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)),
        &size
    );

    array_init(return_value);

    iter = 0;
    while(iter < size) {
        add_next_index_string(return_value,list[iter++],1);
    }

    g_strfreev(list);	
}

#line 2873 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentChooser, add_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkrecentfilter_ce))
		return;

    gtk_recent_chooser_add_filter(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_RECENT_FILTER(PHPG_GOBJECT(filter)));

}


static PHP_METHOD(GtkRecentChooser, remove_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkrecentfilter_ce))
		return;

    gtk_recent_chooser_remove_filter(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_RECENT_FILTER(PHPG_GOBJECT(filter)));

}

#line 471 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentChooser, list_filters)
{
    GSList *list, *element;
    zval *current;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"")) {
        return;
    }

    list = gtk_recent_chooser_list_filters(
        GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr))
    );

    array_init(return_value);

    for(element = list; element; element = element->next) {
        MAKE_STD_ZVAL(current);
        phpg_gobject_new(&current,element->data TSRMLS_CC);
        add_next_index_zval(return_value,current);
    }	

    g_slist_free(list);
}


#line 2932 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentChooser, set_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkrecentfilter_ce))
		return;

    gtk_recent_chooser_set_filter(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_RECENT_FILTER(PHPG_GOBJECT(filter)));

}


static PHP_METHOD(GtkRecentChooser, get_filter)
{
	GtkRecentFilter* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_get_filter(GTK_RECENT_CHOOSER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_show_private, 0)
    ZEND_ARG_INFO(0, show_private)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_show_not_found, 0)
    ZEND_ARG_INFO(0, show_not_found)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_select_multiple, 0)
    ZEND_ARG_INFO(0, select_multiple)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_limit, 0)
    ZEND_ARG_INFO(0, limit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_local_only, 0)
    ZEND_ARG_INFO(0, local_only)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_show_tips, 0)
    ZEND_ARG_INFO(0, show_tips)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_show_numbers, 0)
    ZEND_ARG_INFO(0, show_numbers)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_show_icons, 0)
    ZEND_ARG_INFO(0, show_icons)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_sort_type, 0)
    ZEND_ARG_INFO(0, sort_type)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_sort_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_current_uri, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_select_uri, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_unselect_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_add_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkRecentFilter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_remove_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkRecentFilter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchooser_set_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkRecentFilter, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentchooser_methods[] = {
	ZEND_FENTRY(add_filter, ZEND_MN(GtkRecentChooser_add_filter),           arginfo_gtk_gtkrecentchooser_add_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_current_item, ZEND_MN(GtkRecentChooser_get_current_item),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_current_uri, ZEND_MN(GtkRecentChooser_get_current_uri),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_filter, ZEND_MN(GtkRecentChooser_get_filter),           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_items, ZEND_MN(GtkRecentChooser_get_items),            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_limit, ZEND_MN(GtkRecentChooser_get_limit),            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_local_only, ZEND_MN(GtkRecentChooser_get_local_only),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_select_multiple, ZEND_MN(GtkRecentChooser_get_select_multiple),  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_icons, ZEND_MN(GtkRecentChooser_get_show_icons),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_not_found, ZEND_MN(GtkRecentChooser_get_show_not_found),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_numbers, ZEND_MN(GtkRecentChooser_get_show_numbers),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_private, ZEND_MN(GtkRecentChooser_get_show_private),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_tips, ZEND_MN(GtkRecentChooser_get_show_tips),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_sort_type, ZEND_MN(GtkRecentChooser_get_sort_type),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_uris, ZEND_MN(GtkRecentChooser_get_uris),             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(list_filters, ZEND_MN(GtkRecentChooser_list_filters),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(remove_filter, ZEND_MN(GtkRecentChooser_remove_filter),        arginfo_gtk_gtkrecentchooser_remove_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_all, ZEND_MN(GtkRecentChooser_select_all),           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_uri, ZEND_MN(GtkRecentChooser_select_uri),           arginfo_gtk_gtkrecentchooser_select_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_current_uri, ZEND_MN(GtkRecentChooser_set_current_uri),      arginfo_gtk_gtkrecentchooser_set_current_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_filter, ZEND_MN(GtkRecentChooser_set_filter),           arginfo_gtk_gtkrecentchooser_set_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_limit, ZEND_MN(GtkRecentChooser_set_limit),            arginfo_gtk_gtkrecentchooser_set_limit, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_local_only, ZEND_MN(GtkRecentChooser_set_local_only),       arginfo_gtk_gtkrecentchooser_set_local_only, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_select_multiple, ZEND_MN(GtkRecentChooser_set_select_multiple),  arginfo_gtk_gtkrecentchooser_set_select_multiple, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_icons, ZEND_MN(GtkRecentChooser_set_show_icons),       arginfo_gtk_gtkrecentchooser_set_show_icons, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_not_found, ZEND_MN(GtkRecentChooser_set_show_not_found),   arginfo_gtk_gtkrecentchooser_set_show_not_found, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_numbers, ZEND_MN(GtkRecentChooser_set_show_numbers),     arginfo_gtk_gtkrecentchooser_set_show_numbers, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_private, ZEND_MN(GtkRecentChooser_set_show_private),     arginfo_gtk_gtkrecentchooser_set_show_private, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_tips, ZEND_MN(GtkRecentChooser_set_show_tips),        arginfo_gtk_gtkrecentchooser_set_show_tips, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_sort_func, ZEND_MN(GtkRecentChooser_set_sort_func),        arginfo_gtk_gtkrecentchooser_set_sort_func, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_sort_type, ZEND_MN(GtkRecentChooser_set_sort_type),        arginfo_gtk_gtkrecentchooser_set_sort_type, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unselect_all, ZEND_MN(GtkRecentChooser_unselect_all),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unselect_uri, ZEND_MN(GtkRecentChooser_unselect_uri),         arginfo_gtk_gtkrecentchooser_unselect_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellEditable, editing_done)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_cell_editable_editing_done(GTK_CELL_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCellEditable, remove_widget)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_cell_editable_remove_widget(GTK_CELL_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCellEditable, start_editing)
{
	GdkEvent *event = NULL;
	zval *php_event;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_cell_editable_start_editing(GTK_CELL_EDITABLE(PHPG_GOBJECT(this_ptr)), event);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcelleditable_start_editing, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcelleditable_methods[] = {
	ZEND_FENTRY(editing_done, ZEND_MN(GtkCellEditable_editing_done),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(remove_widget, ZEND_MN(GtkCellEditable_remove_widget),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(start_editing, ZEND_MN(GtkCellEditable_start_editing),        arginfo_gtk_gtkcelleditable_start_editing, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellLayout, add_attribute)
{
	zval *cell;
	char *attribute;
	zend_bool free_attribute = FALSE;
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oui", &cell, gtkcellrenderer_ce, &attribute, &free_attribute, &column))
		return;

    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), attribute, (gint)column);
	if (free_attribute) g_free(attribute);

}


static PHP_METHOD(GtkCellLayout, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_cell_layout_clear(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCellLayout, clear_attributes)
{
	zval *cell;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &cell, gtkcellrenderer_ce))
		return;

    gtk_cell_layout_clear_attributes(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)));

}


static PHP_METHOD(GtkCellLayout, pack_end)
{
	zval *cell;
	zend_bool expand = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|b", &cell, gtkcellrenderer_ce, &expand))
		return;

    gtk_cell_layout_pack_end(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), (gboolean)expand);

}


static PHP_METHOD(GtkCellLayout, pack_start)
{
	zval *cell;
	zend_bool expand = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|b", &cell, gtkcellrenderer_ce, &expand))
		return;

    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), (gboolean)expand);

}


static PHP_METHOD(GtkCellLayout, reorder)
{
	zval *cell;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &cell, gtkcellrenderer_ce, &position))
		return;

    gtk_cell_layout_reorder(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), (gint)position);

}

#line 75 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkCellLayout, set_attributes)
{
    zval *php_cell = NULL;
    GtkCellRenderer *cell = NULL;
    GtkCellLayout *layout = NULL;
    zval ***args = NULL;
    int i, argc = ZEND_NUM_ARGS();

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(1, "O", &php_cell, gtkcellrenderer_ce))
        return;

    if ((argc-1)% 2) {
        php_error(E_WARNING, "%s::%s() requires arguments 3-n to be attribute/column pairs",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    layout = GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr));
    cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));
    gtk_cell_layout_clear_attributes(layout, cell);
    args = php_gtk_func_args(argc);
    for (i = 1; i < argc; i += 2) {
        zval *attr = *args[i];
        zval *column = *args[i+1];

        if (Z_TYPE_P(attr) != IS_STRING) {
            php_error(E_WARNING, "%s::%s() requires argument %d to be a string, %s given",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                      i, zend_zval_type_name(attr));
            efree(args);
            return;
        }
        if (Z_TYPE_P(column) != IS_LONG) {
            php_error(E_WARNING, "%s::%s() requires argument %d to be an integer, %s given",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                      i, zend_zval_type_name(column));
            efree(args);
            return;
        }

        gtk_cell_layout_add_attribute(layout, cell, Z_STRVAL_P(attr), Z_LVAL_P(column));
    }
    efree(args);
}

#line 3281 "ext/gtk+/gen_gtk.c"


#line 132 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkCellLayout, set_cell_data_func)
{
    zval *php_cell;
    GtkCellRenderer *cell;
    GtkCellLayout *layout = NULL;
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "OV", &php_cell, gtkcellrenderer_ce, &callback))
        return;

    layout = GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr));
    cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_cell_layout_set_cell_data_func(layout, cell, (GtkCellLayoutDataFunc)phpg_cell_data_func_marshal,
                                       cb_data, phpg_cb_data_destroy);
}
#line 3307 "ext/gtk+/gen_gtk.c"


#line 682 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkCellLayout, get_cells)
{
    GList *list, *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")){
        return;
    }

    list = gtk_cell_layout_get_cells(GTK_CELL_LAYOUT(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (item = list; item; item = item->next) {
        zval *php_item = NULL;
        GtkWidget *renderer = item->data;
        phpg_gobject_new(&php_item, G_OBJECT(renderer) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }

    g_list_free(list);
}

#line 3334 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcelllayout_add_attribute, 0)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, attribute)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcelllayout_clear_attributes, 0)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcelllayout_pack_end, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcelllayout_pack_start, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcelllayout_reorder, 0)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcelllayout_set_attributes, 0, 0, 1)
    ZEND_ARG_INFO(0, cell)
    ZEND_ARG_INFO(0, attribute)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcelllayout_set_cell_data_func, 0)
    ZEND_ARG_OBJ_INFO(0, cellrenderer, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtkcelllayout_methods[] = {
	ZEND_FENTRY(add_attribute, ZEND_MN(GtkCellLayout_add_attribute),        arginfo_gtk_gtkcelllayout_add_attribute, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(clear, ZEND_MN(GtkCellLayout_clear),                NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(clear_attributes, ZEND_MN(GtkCellLayout_clear_attributes),     arginfo_gtk_gtkcelllayout_clear_attributes, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_cells, ZEND_MN(GtkCellLayout_get_cells),            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(pack_end, ZEND_MN(GtkCellLayout_pack_end),             arginfo_gtk_gtkcelllayout_pack_end, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(pack_start, ZEND_MN(GtkCellLayout_pack_start),           arginfo_gtk_gtkcelllayout_pack_start, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(reorder, ZEND_MN(GtkCellLayout_reorder),              arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_attributes, ZEND_MN(GtkCellLayout_set_attributes),       arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_cell_data_func, ZEND_MN(GtkCellLayout_set_cell_data_func),   arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkEditable, copy_clipboard)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_editable_copy_clipboard(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEditable, cut_clipboard)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_editable_cut_clipboard(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEditable, delete_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_editable_delete_selection(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEditable, delete_text)
{
	long start_pos, end_pos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start_pos, &end_pos))
		return;

    gtk_editable_delete_text(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), (gint)start_pos, (gint)end_pos);

}


static PHP_METHOD(GtkEditable, get_chars)
{
	long start_pos, end_pos;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start_pos, &end_pos))
		return;

    php_retval = gtk_editable_get_chars(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), (gint)start_pos, (gint)end_pos);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkEditable, get_editable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_editable_get_editable(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEditable, get_position)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_editable_get_position(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 2709 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEditable, get_selection_bounds)
{
    gint start, end;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (!gtk_editable_get_selection_bounds(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), &start, &end)) {
        RETURN_FALSE;
    } else {
        php_gtk_build_value(&return_value, "(ii)", start, end);
    }
}


#line 3520 "ext/gtk+/gen_gtk.c"


#line 2736 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEditable, insert_text)
{
    char *text;
    gint pos;
    int text_len;
    zend_bool free_text;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu#", &pos, &text, &text_len, &free_text))
        return;

    gtk_editable_insert_text(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), text, text_len, &pos);

    if (free_text) g_free(text);

    RETURN_LONG(pos);
}

#line 3543 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEditable, paste_clipboard)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_editable_paste_clipboard(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEditable, select_region)
{
	long start, end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start, &end))
		return;

    gtk_editable_select_region(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), (gint)start, (gint)end);

}


static PHP_METHOD(GtkEditable, set_editable)
{
	zend_bool is_editable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_editable))
		return;

    gtk_editable_set_editable(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), (gboolean)is_editable);

}


static PHP_METHOD(GtkEditable, set_position)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_editable_set_position(GTK_EDITABLE(PHPG_GOBJECT(this_ptr)), (gint)position);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_delete_text, 0)
    ZEND_ARG_INFO(0, start_pos)
    ZEND_ARG_INFO(0, end_pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_get_chars, 0)
    ZEND_ARG_INFO(0, start_pos)
    ZEND_ARG_INFO(0, end_pos)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_insert_text, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_select_region, 0)
    ZEND_ARG_INFO(0, start)
    ZEND_ARG_INFO(0, end)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_set_editable, 0)
    ZEND_ARG_INFO(0, is_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeditable_set_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

static zend_function_entry gtkeditable_methods[] = {
	ZEND_FENTRY(copy_clipboard, ZEND_MN(GtkEditable_copy_clipboard),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(cut_clipboard, ZEND_MN(GtkEditable_cut_clipboard),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(delete_selection, ZEND_MN(GtkEditable_delete_selection),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(delete_text, ZEND_MN(GtkEditable_delete_text),          arginfo_gtk_gtkeditable_delete_text, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_chars, ZEND_MN(GtkEditable_get_chars),            arginfo_gtk_gtkeditable_get_chars, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_editable, ZEND_MN(GtkEditable_get_editable),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_position, ZEND_MN(GtkEditable_get_position),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_selection_bounds, ZEND_MN(GtkEditable_get_selection_bounds), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(insert_text, ZEND_MN(GtkEditable_insert_text),          arginfo_gtk_gtkeditable_insert_text, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(paste_clipboard, ZEND_MN(GtkEditable_paste_clipboard),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_region, ZEND_MN(GtkEditable_select_region),        arginfo_gtk_gtkeditable_select_region, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_editable, ZEND_MN(GtkEditable_set_editable),         arginfo_gtk_gtkeditable_set_editable, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_position, ZEND_MN(GtkEditable_set_position),         arginfo_gtk_gtkeditable_set_position, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFileChooser, add_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkfilefilter_ce))
		return;

    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_FILE_FILTER(PHPG_GOBJECT(filter)));

}


static PHP_METHOD(GtkFileChooser, add_shortcut_folder)
{
	char *folder;
	zend_bool free_folder = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &folder, &free_folder))
		return;

    php_retval = gtk_file_chooser_add_shortcut_folder(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), folder, &error);
	if (free_folder) g_free(folder);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, add_shortcut_folder_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_file_chooser_add_shortcut_folder_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_action)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_action(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_current_folder)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFileChooser, get_current_folder_uri)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_current_folder_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFileChooser, get_extra_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_extra_widget(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 3314 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, get_filename)
{
    gchar *php_retval, *cp_ret, *fn;
    gsize cp_len;
    zend_bool free_result;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    php_retval = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        if (convert) {
            fn = g_filename_to_utf8(php_retval, strlen(php_retval), NULL, NULL, NULL);
            cp_ret = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_result TSRMLS_CC);
            if (cp_ret) {
                RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
            } else {
                php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            }
        } else {
            RETVAL_STRINGL((char *)php_retval, strlen(php_retval), 1);
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 3824 "ext/gtk+/gen_gtk.c"


#line 3356 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, get_filenames)
{
    GSList *files;
    gchar *cp = NULL;
    gchar *fn = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    array_init(return_value);
    for (files = gtk_file_chooser_get_filenames(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr))); files; files = files->next) {
        if (convert) {
            if (files->data) {
                fn = g_filename_to_utf8(files->data, strlen(files->data), NULL, NULL, NULL);
                cp = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_cp TSRMLS_CC);
                if (cp) {
                    add_next_index_string(return_value, (char *)cp, 1);
                } else {
                    php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                }
                if (free_cp)
                    g_free(cp);
            }
        } else {
            add_next_index_string(return_value, (char *)files->data, 1);
        }
        g_free(files->data);
    }
    g_slist_free(files);
}

#line 3864 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileChooser, get_filter)
{
	GtkFileFilter* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_filter(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFileChooser, get_local_only)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_local_only(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_preview_filename)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_preview_filename(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFileChooser, get_preview_uri)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_preview_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFileChooser, get_preview_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_preview_widget(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFileChooser, get_preview_widget_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_preview_widget_active(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_select_multiple)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_select_multiple(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_show_hidden)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_show_hidden(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, get_uri)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 3394 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, get_uris)
{
    GSList *files;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    array_init(return_value);
    for (files = gtk_file_chooser_get_uris(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr))); files; files = files->next) {
        add_next_index_string(return_value, (char *) files->data, 1);
        g_free(files->data);
    }
    g_slist_free(files);
}

#line 4050 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileChooser, get_use_preview_label)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_use_preview_label(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 3413 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, list_filters)
{
    GSList *filters, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    filters = gtk_file_chooser_list_filters(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = filters; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
    g_slist_free(filters);
}

#line 4089 "ext/gtk+/gen_gtk.c"


#line 3488 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, list_shortcut_folder_uris)
{
    GSList *files;
    gchar *cp = NULL;
    gchar *fn = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    array_init(return_value);
    for (files = gtk_file_chooser_list_shortcut_folder_uris(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr))); files; files = files->next) {
        if (convert) {
            if (files->data) {
                fn = g_filename_to_utf8(files->data, strlen(files->data), NULL, NULL, NULL);
                cp = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_cp TSRMLS_CC);
                if (cp) {
                    add_next_index_string(return_value, (char *)cp, 1);
                } else {
                    php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                }
                if (free_cp)
                    g_free(cp);
            }
        } else {
            add_next_index_string(return_value, (char *)files->data, 1);
        }
        g_free(files->data);
    }
    g_slist_free(files);
}

#line 4129 "ext/gtk+/gen_gtk.c"


#line 3443 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooser, list_shortcut_folders)
{
    GSList *files;
    gchar *cp = NULL;
    gchar *fn = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    array_init(return_value);
    for (files = gtk_file_chooser_list_shortcut_folders(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr))); files; files = files->next) {
        if (convert) {
            if (files->data) {
                fn = g_filename_to_utf8(files->data, strlen(files->data), NULL, NULL, NULL);
                cp = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_cp TSRMLS_CC);
                if (cp) {
                    add_next_index_string(return_value, (char *)cp, 1);
                } else {
                    php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                }
                if (free_cp)
                    g_free(cp);
            }
        } else {
            add_next_index_string(return_value, (char *)files->data, 1);
        }
        g_free(files->data);
    }
    g_slist_free(files);
}

#line 4169 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileChooser, remove_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkfilefilter_ce))
		return;

    gtk_file_chooser_remove_filter(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_FILE_FILTER(PHPG_GOBJECT(filter)));

}


static PHP_METHOD(GtkFileChooser, remove_shortcut_folder)
{
	char *folder;
	zend_bool free_folder = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &folder, &free_folder))
		return;

    php_retval = gtk_file_chooser_remove_shortcut_folder(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), folder, &error);
	if (free_folder) g_free(folder);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, remove_shortcut_folder_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_file_chooser_remove_shortcut_folder_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_file_chooser_select_all(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkFileChooser, select_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    php_retval = gtk_file_chooser_select_filename(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, select_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_file_chooser_select_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_action)
{
	GtkFileChooserAction action;
	zval *php_action = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_action))
		return;

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
		return;
	}

    gtk_file_chooser_set_action(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), action);

}


static PHP_METHOD(GtkFileChooser, set_current_folder)
{
	char *filename;
	zend_bool free_filename = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    php_retval = gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_current_folder_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_file_chooser_set_current_folder_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_current_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkFileChooser, set_extra_widget)
{
	zval *extra_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &extra_widget, gtkwidget_ce))
		return;

    gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(extra_widget)));

}


static PHP_METHOD(GtkFileChooser, set_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    php_retval = gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_filter)
{
	zval *filter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &filter, gtkfilefilter_ce))
		return;

    gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_FILE_FILTER(PHPG_GOBJECT(filter)));

}


static PHP_METHOD(GtkFileChooser, set_local_only)
{
	zend_bool local_only;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &local_only))
		return;

    gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)local_only);

}


static PHP_METHOD(GtkFileChooser, set_preview_widget)
{
	zval *preview_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &preview_widget, gtkwidget_ce))
		return;

    gtk_file_chooser_set_preview_widget(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(preview_widget)));

}


static PHP_METHOD(GtkFileChooser, set_preview_widget_active)
{
	zend_bool active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &active))
		return;

    gtk_file_chooser_set_preview_widget_active(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)active);

}


static PHP_METHOD(GtkFileChooser, set_select_multiple)
{
	zend_bool select_multiple;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &select_multiple))
		return;

    gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)select_multiple);

}


static PHP_METHOD(GtkFileChooser, set_show_hidden)
{
	zend_bool show_hidden;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_hidden))
		return;

    gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)show_hidden);

}


static PHP_METHOD(GtkFileChooser, set_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_file_chooser_set_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_use_preview_label)
{
	zend_bool use_label;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_label))
		return;

    gtk_file_chooser_set_use_preview_label(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)use_label);

}


static PHP_METHOD(GtkFileChooser, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_file_chooser_unselect_all(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkFileChooser, unselect_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_file_chooser_unselect_filename(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkFileChooser, unselect_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    gtk_file_chooser_unselect_uri(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);

}


static PHP_METHOD(GtkFileChooser, set_do_overwrite_confirmation)
{
	zend_bool do_overwrite_confirmation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &do_overwrite_confirmation))
		return;

    gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)do_overwrite_confirmation);

}


static PHP_METHOD(GtkFileChooser, get_do_overwrite_confirmation)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_do_overwrite_confirmation(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooser, set_create_folders)
{
	zend_bool create_folders;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &create_folders))
		return;

    gtk_file_chooser_set_create_folders(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)), (gboolean)create_folders);

}


static PHP_METHOD(GtkFileChooser, get_create_folders)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_get_create_folders(GTK_FILE_CHOOSER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_add_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkFileFilter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_add_shortcut_folder, 0)
    ZEND_ARG_INFO(0, folder)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_add_shortcut_folder_uri, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfilechooser_get_filename, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfilechooser_get_filenames, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfilechooser_list_shortcut_folder_uris, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfilechooser_list_shortcut_folders, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_remove_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkFileFilter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_remove_shortcut_folder, 0)
    ZEND_ARG_INFO(0, folder)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_remove_shortcut_folder_uri, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_select_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_select_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_action, 0)
    ZEND_ARG_INFO(0, action)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_current_folder, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_current_folder_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_current_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_extra_widget, 0)
    ZEND_ARG_OBJ_INFO(0, extra_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_filter, 0)
    ZEND_ARG_OBJ_INFO(0, filter, GtkFileFilter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_local_only, 0)
    ZEND_ARG_INFO(0, local_only)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_preview_widget, 0)
    ZEND_ARG_OBJ_INFO(0, preview_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_preview_widget_active, 0)
    ZEND_ARG_INFO(0, active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_select_multiple, 0)
    ZEND_ARG_INFO(0, select_multiple)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_show_hidden, 0)
    ZEND_ARG_INFO(0, show_hidden)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_use_preview_label, 0)
    ZEND_ARG_INFO(0, use_label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_unselect_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_unselect_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_do_overwrite_confirmation, 0)
    ZEND_ARG_INFO(0, do_overwrite_confirmation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooser_set_create_folders, 0)
    ZEND_ARG_INFO(0, create_folders)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfilechooser_methods[] = {
	ZEND_FENTRY(add_filter, ZEND_MN(GtkFileChooser_add_filter),           arginfo_gtk_gtkfilechooser_add_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(add_shortcut_folder, ZEND_MN(GtkFileChooser_add_shortcut_folder),  arginfo_gtk_gtkfilechooser_add_shortcut_folder, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(add_shortcut_folder_uri, ZEND_MN(GtkFileChooser_add_shortcut_folder_uri), arginfo_gtk_gtkfilechooser_add_shortcut_folder_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_action, ZEND_MN(GtkFileChooser_get_action),           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_create_folders, ZEND_MN(GtkFileChooser_get_create_folders),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_current_folder, ZEND_MN(GtkFileChooser_get_current_folder),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_current_folder_uri, ZEND_MN(GtkFileChooser_get_current_folder_uri), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_do_overwrite_confirmation, ZEND_MN(GtkFileChooser_get_do_overwrite_confirmation), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_extra_widget, ZEND_MN(GtkFileChooser_get_extra_widget),     NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_filename, ZEND_MN(GtkFileChooser_get_filename),         arginfo_gtk_gtkfilechooser_get_filename, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_filenames, ZEND_MN(GtkFileChooser_get_filenames),        arginfo_gtk_gtkfilechooser_get_filenames, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_filter, ZEND_MN(GtkFileChooser_get_filter),           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_local_only, ZEND_MN(GtkFileChooser_get_local_only),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_preview_filename, ZEND_MN(GtkFileChooser_get_preview_filename), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_preview_uri, ZEND_MN(GtkFileChooser_get_preview_uri),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_preview_widget, ZEND_MN(GtkFileChooser_get_preview_widget),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_preview_widget_active, ZEND_MN(GtkFileChooser_get_preview_widget_active), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_select_multiple, ZEND_MN(GtkFileChooser_get_select_multiple),  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_show_hidden, ZEND_MN(GtkFileChooser_get_show_hidden),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_uri, ZEND_MN(GtkFileChooser_get_uri),              NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_uris, ZEND_MN(GtkFileChooser_get_uris),             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_use_preview_label, ZEND_MN(GtkFileChooser_get_use_preview_label), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(list_filters, ZEND_MN(GtkFileChooser_list_filters),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(list_shortcut_folder_uris, ZEND_MN(GtkFileChooser_list_shortcut_folder_uris), arginfo_gtk_gtkfilechooser_list_shortcut_folder_uris, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(list_shortcut_folders, ZEND_MN(GtkFileChooser_list_shortcut_folders), arginfo_gtk_gtkfilechooser_list_shortcut_folders, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(remove_filter, ZEND_MN(GtkFileChooser_remove_filter),        arginfo_gtk_gtkfilechooser_remove_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(remove_shortcut_folder, ZEND_MN(GtkFileChooser_remove_shortcut_folder), arginfo_gtk_gtkfilechooser_remove_shortcut_folder, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(remove_shortcut_folder_uri, ZEND_MN(GtkFileChooser_remove_shortcut_folder_uri), arginfo_gtk_gtkfilechooser_remove_shortcut_folder_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_all, ZEND_MN(GtkFileChooser_select_all),           NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_filename, ZEND_MN(GtkFileChooser_select_filename),      arginfo_gtk_gtkfilechooser_select_filename, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(select_uri, ZEND_MN(GtkFileChooser_select_uri),           arginfo_gtk_gtkfilechooser_select_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_action, ZEND_MN(GtkFileChooser_set_action),           arginfo_gtk_gtkfilechooser_set_action, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_create_folders, ZEND_MN(GtkFileChooser_set_create_folders),   arginfo_gtk_gtkfilechooser_set_create_folders, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_current_folder, ZEND_MN(GtkFileChooser_set_current_folder),   arginfo_gtk_gtkfilechooser_set_current_folder, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_current_folder_uri, ZEND_MN(GtkFileChooser_set_current_folder_uri), arginfo_gtk_gtkfilechooser_set_current_folder_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_current_name, ZEND_MN(GtkFileChooser_set_current_name),     arginfo_gtk_gtkfilechooser_set_current_name, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_do_overwrite_confirmation, ZEND_MN(GtkFileChooser_set_do_overwrite_confirmation), arginfo_gtk_gtkfilechooser_set_do_overwrite_confirmation, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_extra_widget, ZEND_MN(GtkFileChooser_set_extra_widget),     arginfo_gtk_gtkfilechooser_set_extra_widget, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_filename, ZEND_MN(GtkFileChooser_set_filename),         arginfo_gtk_gtkfilechooser_set_filename, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_filter, ZEND_MN(GtkFileChooser_set_filter),           arginfo_gtk_gtkfilechooser_set_filter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_local_only, ZEND_MN(GtkFileChooser_set_local_only),       arginfo_gtk_gtkfilechooser_set_local_only, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_preview_widget, ZEND_MN(GtkFileChooser_set_preview_widget),   arginfo_gtk_gtkfilechooser_set_preview_widget, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_preview_widget_active, ZEND_MN(GtkFileChooser_set_preview_widget_active), arginfo_gtk_gtkfilechooser_set_preview_widget_active, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_select_multiple, ZEND_MN(GtkFileChooser_set_select_multiple),  arginfo_gtk_gtkfilechooser_set_select_multiple, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_show_hidden, ZEND_MN(GtkFileChooser_set_show_hidden),      arginfo_gtk_gtkfilechooser_set_show_hidden, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_uri, ZEND_MN(GtkFileChooser_set_uri),              arginfo_gtk_gtkfilechooser_set_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_use_preview_label, ZEND_MN(GtkFileChooser_set_use_preview_label), arginfo_gtk_gtkfilechooser_set_use_preview_label, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unselect_all, ZEND_MN(GtkFileChooser_unselect_all),         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unselect_filename, ZEND_MN(GtkFileChooser_unselect_filename),    arginfo_gtk_gtkfilechooser_unselect_filename, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unselect_uri, ZEND_MN(GtkFileChooser_unselect_uri),         arginfo_gtk_gtkfilechooser_unselect_uri, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeDragDest, drag_data_received)
{
	GtkTreePath *dest;
	zval *php_dest, *php_selection_data;
	GtkSelectionData *selection_data = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_dest, &php_selection_data, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_dest, &dest TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects dest to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_selection_data, GTK_TYPE_SELECTION_DATA, FALSE TSRMLS_CC)) {
        selection_data = (GtkSelectionData *) PHPG_GBOXED(php_selection_data);
    } else {
        php_error(E_WARNING, "%s::%s() expects selection_data argument to be a valid GtkSelectionData object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_drag_dest_drag_data_received(GTK_TREE_DRAG_DEST(PHPG_GOBJECT(this_ptr)), dest, selection_data);

    if (dest)
        gtk_tree_path_free(dest);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeDragDest, row_drop_possible)
{
	GtkTreePath *dest_path;
	zval *php_dest_path, *php_selection_data;
	GtkSelectionData *selection_data = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_dest_path, &php_selection_data, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_dest_path, &dest_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects dest_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_selection_data, GTK_TYPE_SELECTION_DATA, FALSE TSRMLS_CC)) {
        selection_data = (GtkSelectionData *) PHPG_GBOXED(php_selection_data);
    } else {
        php_error(E_WARNING, "%s::%s() expects selection_data argument to be a valid GtkSelectionData object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_drag_dest_row_drop_possible(GTK_TREE_DRAG_DEST(PHPG_GOBJECT(this_ptr)), dest_path, selection_data);

    if (dest_path)
        gtk_tree_path_free(dest_path);
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreedragdest_drag_data_received, 0)
    ZEND_ARG_INFO(0, dest)
    ZEND_ARG_OBJ_INFO(0, selection_data, GtkSelectionData, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreedragdest_row_drop_possible, 0)
    ZEND_ARG_INFO(0, dest_path)
    ZEND_ARG_OBJ_INFO(0, selection_data, GtkSelectionData, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreedragdest_methods[] = {
	ZEND_FENTRY(drag_data_received, ZEND_MN(GtkTreeDragDest_drag_data_received),   arginfo_gtk_gtktreedragdest_drag_data_received, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_drop_possible, ZEND_MN(GtkTreeDragDest_row_drop_possible),    arginfo_gtk_gtktreedragdest_row_drop_possible, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeDragSource, drag_data_delete)
{
	GtkTreePath *path;
	zval *php_path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_drag_source_drag_data_delete(GTK_TREE_DRAG_SOURCE(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeDragSource, drag_data_get)
{
	GtkTreePath *path;
	zval *php_path, *php_selection_data;
	GtkSelectionData *selection_data = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_selection_data, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_selection_data, GTK_TYPE_SELECTION_DATA, FALSE TSRMLS_CC)) {
        selection_data = (GtkSelectionData *) PHPG_GBOXED(php_selection_data);
    } else {
        php_error(E_WARNING, "%s::%s() expects selection_data argument to be a valid GtkSelectionData object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_drag_source_drag_data_get(GTK_TREE_DRAG_SOURCE(PHPG_GOBJECT(this_ptr)), path, selection_data);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeDragSource, row_draggable)
{
	GtkTreePath *path;
	zval *php_path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_drag_source_row_draggable(GTK_TREE_DRAG_SOURCE(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreedragsource_drag_data_delete, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreedragsource_drag_data_get, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, selection_data, GtkSelectionData, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreedragsource_row_draggable, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreedragsource_methods[] = {
	ZEND_FENTRY(drag_data_delete, ZEND_MN(GtkTreeDragSource_drag_data_delete),     arginfo_gtk_gtktreedragsource_drag_data_delete, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(drag_data_get, ZEND_MN(GtkTreeDragSource_drag_data_get),        arginfo_gtk_gtktreedragsource_drag_data_get, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_draggable, ZEND_MN(GtkTreeDragSource_row_draggable),        arginfo_gtk_gtktreedragsource_row_draggable, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};
#line 1450 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get)
{
    GtkTreeModel *model;
    GtkTreeIter *iter;
    zval *columns = NULL, *php_iter, **item;
    int n_columns;
    zend_bool varargs = FALSE;

    NOT_STATIC_METHOD();

    if (php_gtk_parse_args_quiet(ZEND_NUM_ARGS(), "Oa", &php_iter, gtktreeiter_ce, &columns)) {
        varargs = FALSE;
    } else if (php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &columns, "O", &php_iter, gtktreeiter_ce)) {
        varargs = TRUE;
    }  else {
        return;
    }

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));
    iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);

    array_init(return_value);
    n_columns = gtk_tree_model_get_n_columns(model);

    if (columns) {
        for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(columns));
             zend_hash_get_current_data(Z_ARRVAL_P(columns), (void**)&item) == SUCCESS;
             zend_hash_move_forward(Z_ARRVAL_P(columns))) {

            GValue value = { 0, };
            zval *zv = NULL;

            convert_to_long_ex(item);
            if (Z_LVAL_PP(item) < 0 || Z_LVAL_PP(item) >= n_columns) {
                php_error(E_WARNING, "%s::%s(): column number is out of range - model has %d columns", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C), n_columns);
                if (varargs) zval_ptr_dtor(&columns);
                zval_dtor(return_value);
                RETURN_NULL();
            }

            gtk_tree_model_get_value(model, iter, Z_LVAL_PP(item), &value);
            if (phpg_gvalue_to_zval(&value, &zv, TRUE, TRUE TSRMLS_CC) == FAILURE) {
                g_value_unset(&value);
                if (varargs) zval_ptr_dtor(&columns);
                zval_dtor(return_value);
                RETURN_NULL();
            }
            add_next_index_zval(return_value, zv);
            g_value_unset(&value);
        }

        if (varargs) zval_ptr_dtor(&columns);
    }
}

#line 5043 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModel, get_column_type)
{
	long index;
	GType php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    php_retval = gtk_tree_model_get_column_type(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), (gint)index);
    phpg_gtype_new(return_value, php_retval TSRMLS_CC);

}


static PHP_METHOD(GtkTreeModel, get_flags)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_model_get_flags(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 1138 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get_iter)
{
    zval *php_path = NULL;
    GtkTreeIter iter;
    GtkTreePath *path;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
        return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path argument to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (gtk_tree_model_get_iter(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, path)) {
        gtk_tree_path_free(path);
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    } else {
        gtk_tree_path_free(path);
        php_error(E_WARNING, "%s::%s(): invalid tree path", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
    }
}

#line 5102 "ext/gtk+/gen_gtk.c"


#line 1165 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get_iter_first)
{
    GtkTreeIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    }
}

#line 5120 "ext/gtk+/gen_gtk.c"


#line 1165 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get_iter_root)
{
    GtkTreeIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    }
}

#line 5138 "ext/gtk+/gen_gtk.c"


#line 1188 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get_iter_from_string)
{
    char *path_string = NULL;
    GtkTreeIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "s", &path_string))
        return;

    if (!gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, path_string)) {
        return;
    }

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 5159 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModel, get_n_columns)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_model_get_n_columns(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeModel, get_path)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	GtkTreePath *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_get_path(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
        gtk_tree_path_free(php_retval);
    }

}


static PHP_METHOD(GtkTreeModel, get_string_from_iter)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_get_string_from_iter(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 1215 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, get_value)
{
    zval *php_iter;
    GtkTreeIter *iter;
    GtkTreeModel *model;
    gint column;
    GValue value = { 0 };

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_iter, gtktreeiter_ce, &column))
        return;

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));
    iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);

    if (column < 0 || column >= gtk_tree_model_get_n_columns(model)) {
        php_error(E_WARNING, "%s::%s(): column  is out of range - model has %d column(s)",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                  gtk_tree_model_get_n_columns(model));
        return;
    }

    gtk_tree_model_get_value(model, iter, column, &value);
    phpg_gvalue_to_zval(&value, &return_value, TRUE, TRUE TSRMLS_CC);
    g_value_unset(&value);
}

#line 5271 "ext/gtk+/gen_gtk.c"


#line 1252 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, iter_children)
{
    GtkTreeIter iter;
    zval *php_parent_iter = NULL;
    GtkTreeIter *parent_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_parent_iter, gtktreeiter_ce))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }

    if (gtk_tree_model_iter_children(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, parent_iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    } else {
        RETURN_NULL();
    }
}

#line 5297 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModel, iter_has_child)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_iter_has_child(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeModel, iter_n_children)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_iter, gboxed_ce))
		return;

    if (Z_TYPE_P(php_iter) != IS_NULL) {
        if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
            iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
        } else {
            php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    php_retval = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_LONG(php_retval);
}

#line 1348 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, iter_next)
{
    GtkTreeIter iter;
    zval *php_iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gtktreeiter_ce))
        return;

    iter = *(GtkTreeIter *) PHPG_GBOXED(php_iter);

    if (gtk_tree_model_iter_next(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    } else {
        RETURN_NULL();
    }
}

#line 5369 "ext/gtk+/gen_gtk.c"


#line 1284 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, iter_nth_child)
{
    GtkTreeIter iter;
    zval *php_parent_iter = NULL;
    GtkTreeIter *parent_iter = NULL;
    int n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ni", &php_parent_iter, gtktreeiter_ce, &n))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }

    if (gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, parent_iter, n)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    } else {
        RETURN_NULL();
    }
}

#line 5396 "ext/gtk+/gen_gtk.c"


#line 1317 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModel, iter_parent)
{
    GtkTreeIter iter;
    zval *php_child_iter = NULL;
    GtkTreeIter *child_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_child_iter, gtktreeiter_ce))
        return;

    if (php_child_iter && IS_OBJECT == Z_TYPE_P(php_child_iter)) {
        child_iter = (GtkTreeIter *) PHPG_GBOXED(php_child_iter);
    }

    if (gtk_tree_model_iter_parent(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, child_iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    } else {
        RETURN_NULL();
    }
}

#line 5422 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModel, ref_node)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_ref_node(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);

}


static PHP_METHOD(GtkTreeModel, row_changed)
{
	GtkTreePath *path;
	zval *php_path, *php_iter;
	GtkTreeIter *iter = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_iter, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_row_changed(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), path, iter);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeModel, row_deleted)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_row_deleted(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeModel, row_has_child_toggled)
{
	GtkTreePath *path;
	zval *php_path, *php_iter;
	GtkTreeIter *iter = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_iter, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_row_has_child_toggled(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), path, iter);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeModel, row_inserted)
{
	GtkTreePath *path;
	zval *php_path, *php_iter;
	GtkTreeIter *iter = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_iter, gboxed_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_row_inserted(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), path, iter);

    if (path)
        gtk_tree_path_free(path);

}

#line 1507 "ext/gtk+/gtktreeview.overrides"

static PHP_METHOD(GtkTreeModel, rows_reordered)
{
    zval *php_path, *php_iter, *php_new_order, **elem;
    GtkTreePath *path = NULL;
    GtkTreeIter *iter = NULL;
    int n_children, n_elems, i;
    gint *new_order;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO!a/", &php_path, &php_iter, gtktreeiter_ce, &php_new_order))
        return;

    if (Z_TYPE_P(php_path) == IS_NULL ||
        (Z_TYPE_P(php_path) == IS_STRING && Z_STRLEN_P(php_path) == 0)) {
        path = gtk_tree_path_new();
    } else {
        phpg_tree_path_from_zval(php_path, &path TSRMLS_CC);
    }
    if (path == NULL) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "path argument is not a valid tree path specification");
        return;
    }

    if (php_iter) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    }

    n_children = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);
    n_elems = zend_hash_num_elements(Z_ARRVAL_P(php_new_order));
    if (n_children != n_elems) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "new order array size (%d) not the same as the number of iter children (%d)", n_elems, n_children);
        gtk_tree_path_free(path);
        return;
    }

    new_order = safe_emalloc(n_elems, sizeof(gint), 0);
    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_new_order)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_new_order), (void **) &elem) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_new_order)), i++) {

        convert_to_long(*elem);
        if (Z_LVAL_PP(elem) < 0 || Z_LVAL_PP(elem) >= n_children) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING, "position index out of range 0-%d", n_children);
            gtk_tree_path_free(path);
            efree(new_order);
            return;
        }
        new_order[i] = Z_LVAL_PP(elem);
    }

    gtk_tree_model_rows_reordered(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), path, iter, new_order);

    gtk_tree_path_free(path);
    efree(new_order);
}

#line 5626 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModel, unref_node)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_unref_node(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), iter);

}

#line 1376 "ext/gtk+/gtktreeview.overrides"

gboolean phpg_tree_model_foreach_marshal(GtkTreeModel *model,
                                         GtkTreePath  *path,
                                         GtkTreeIter  *iter,
                                         gpointer      data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_path = NULL, *php_iter = NULL;
    gboolean ret;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return TRUE;
    }

    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 3, &n_args);
    args[0] = &php_model;
    args[1] = &php_path;
    args[2] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_path);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        ret = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        ret = TRUE;
    }

    efree(callback_name);
    efree(args);

    phpg_handle_marshaller_exception(TSRMLS_C);

    return ret;
}

static PHP_METHOD(GtkTreeModel, for_each)
{
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &callback)) {
        return;
    }

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_tree_model_foreach(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)),
                           phpg_tree_model_foreach_marshal, cb_data);
    phpg_cb_data_destroy(cb_data);
}

#line 5725 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_column_type, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_iter, 0)
    ZEND_ARG_INFO(0, treepath)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_iter_from_string, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_path, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_string_from_iter, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_get_value, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_iter_children, 0)
    ZEND_ARG_OBJ_INFO(0, parent_iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_iter_has_child, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_iter_n_children, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_iter_next, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_iter_nth_child, 0)
    ZEND_ARG_OBJ_INFO(0, parent_iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, n)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreemodel_iter_parent, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_ref_node, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_row_changed, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_row_deleted, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_row_has_child_toggled, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_row_inserted, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_unref_node, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodel_for_each, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtktreemodel_methods[] = {
	ZEND_FENTRY(for_each, ZEND_MN(GtkTreeModel_for_each),             arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get, ZEND_MN(GtkTreeModel_get),                  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_column_type, ZEND_MN(GtkTreeModel_get_column_type),      arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_flags, ZEND_MN(GtkTreeModel_get_flags),            NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_iter, ZEND_MN(GtkTreeModel_get_iter),             arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_iter_first, ZEND_MN(GtkTreeModel_get_iter_first),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_iter_from_string, ZEND_MN(GtkTreeModel_get_iter_from_string), arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_iter_root, ZEND_MN(GtkTreeModel_get_iter_root),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_n_columns, ZEND_MN(GtkTreeModel_get_n_columns),        NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_path, ZEND_MN(GtkTreeModel_get_path),             arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_string_from_iter, ZEND_MN(GtkTreeModel_get_string_from_iter), arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_value, ZEND_MN(GtkTreeModel_get_value),            arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_children, ZEND_MN(GtkTreeModel_iter_children),        arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_has_child, ZEND_MN(GtkTreeModel_iter_has_child),       arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_n_children, ZEND_MN(GtkTreeModel_iter_n_children),      arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_next, ZEND_MN(GtkTreeModel_iter_next),            arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_nth_child, ZEND_MN(GtkTreeModel_iter_nth_child),       arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(iter_parent, ZEND_MN(GtkTreeModel_iter_parent),          arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(ref_node, ZEND_MN(GtkTreeModel_ref_node),             arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_changed, ZEND_MN(GtkTreeModel_row_changed),          arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_deleted, ZEND_MN(GtkTreeModel_row_deleted),          arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_has_child_toggled, ZEND_MN(GtkTreeModel_row_has_child_toggled), arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(row_inserted, ZEND_MN(GtkTreeModel_row_inserted),         arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(rows_reordered, ZEND_MN(GtkTreeModel_rows_reordered),       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(unref_node, ZEND_MN(GtkTreeModel_unref_node),           arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};
#line 2051 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeSortable, get_sort_column_id)
{
    gint sort_column_id;
    GtkSortType order;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)), &sort_column_id, &order);

    php_gtk_build_value(&return_value, "(ii)", sort_column_id, (int)order);
}

#line 5873 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSortable, has_default_sort_func)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_sortable_has_default_sort_func(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 2074 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeSortable, set_default_sort_func)
{
    zval *callback = NULL, *extra;
    phpg_cb_data_t *cb_data = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V!", &callback))
        return;

    if (callback) {
        zval_add_ref(&callback);
        cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);
        gtk_tree_sortable_set_default_sort_func(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)),
                                                (GtkTreeIterCompareFunc)phpg_tree_iter_compare_func_marshal,
                                                cb_data, phpg_cb_data_destroy);
    } else {
        gtk_tree_sortable_set_default_sort_func(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)),
                                                NULL, NULL, NULL);
    }
}

#line 5913 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSortable, set_sort_column_id)
{
	long sort_column_id;
	GtkSortType order;
	zval *php_order = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &sort_column_id, &php_order))
		return;

	if (php_order && phpg_gvalue_get_enum(GTK_TYPE_SORT_TYPE, php_order, (gint *)&order) == FAILURE) {
		return;
	}

    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)), (gint)sort_column_id, order);

}

#line 2106 "ext/gtk+/gtktreeview.overrides"

static gint phpg_tree_iter_compare_func_marshal(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_iter_a = NULL, *php_iter_b = NULL;
    gint compared = 0;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_gboxed_new(&php_iter_a, GTK_TYPE_TREE_ITER, a, TRUE, TRUE TSRMLS_CC);
    phpg_gboxed_new(&php_iter_b, GTK_TYPE_TREE_ITER, b, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 3, &n_args);
    args[0] = &php_model;
    args[1] = &php_iter_a;
    args[2] = &php_iter_b;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_iter_a);
    zval_ptr_dtor(&php_iter_b);

    if (retval) {
        convert_to_long(retval);
        compared = Z_LVAL_P(retval);
        zval_ptr_dtor(&retval);
    } else {
        compared = 0;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return compared;
}

static PHP_METHOD(GtkTreeSortable, set_sort_func)
{
    gint sort_column_id;
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "iV", &sort_column_id, &callback))
        return;

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)),
                                                           sort_column_id,
                                                           (GtkTreeIterCompareFunc)phpg_tree_iter_compare_func_marshal,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 6008 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSortable, sort_column_changed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_sortable_sort_column_changed(GTK_TREE_SORTABLE(PHPG_GOBJECT(this_ptr)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreesortable_set_default_sort_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreesortable_set_sort_column_id, 0)
    ZEND_ARG_INFO(0, sort_column_id)
    ZEND_ARG_INFO(0, order)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreesortable_set_sort_func, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtktreesortable_methods[] = {
	ZEND_FENTRY(get_sort_column_id, ZEND_MN(GtkTreeSortable_get_sort_column_id),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(has_default_sort_func, ZEND_MN(GtkTreeSortable_has_default_sort_func), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_default_sort_func, ZEND_MN(GtkTreeSortable_set_default_sort_func), arginfo_gtk_gtktreesortable_set_default_sort_func, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_sort_column_id, ZEND_MN(GtkTreeSortable_set_sort_column_id),   arginfo_gtk_gtktreesortable_set_sort_column_id, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_sort_func, ZEND_MN(GtkTreeSortable_set_sort_func),        arginfo_gtk_gtktreesortable_set_sort_func, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(sort_column_changed, ZEND_MN(GtkTreeSortable_sort_column_changed),  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkActivatable, sync_action_properties)
{
	GtkAction *action = NULL;
	zval *php_action = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_action, gtkaction_ce))
		return;
    if (php_action) {
        if (Z_TYPE_P(php_action) == IS_NULL)
            action = NULL;
        else
            action = GTK_ACTION(PHPG_GOBJECT(php_action));
    }

    gtk_activatable_sync_action_properties(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)), action);

}


static PHP_METHOD(GtkActivatable, set_related_action)
{
	zval *action;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &action, gtkaction_ce))
		return;

    gtk_activatable_set_related_action(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)), GTK_ACTION(PHPG_GOBJECT(action)));

}


static PHP_METHOD(GtkActivatable, get_related_action)
{
	GtkAction* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_activatable_get_related_action(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkActivatable, set_use_action_appearance)
{
	zend_bool use_appearance;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_appearance))
		return;

    gtk_activatable_set_use_action_appearance(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)), (gboolean)use_appearance);

}


static PHP_METHOD(GtkActivatable, get_use_action_appearance)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_activatable_get_use_action_appearance(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkActivatable, do_set_related_action)
{
	zval *action;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &action, gtkaction_ce))
		return;

    gtk_activatable_do_set_related_action(GTK_ACTIVATABLE(PHPG_GOBJECT(this_ptr)), GTK_ACTION(PHPG_GOBJECT(action)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkactivatable_sync_action_properties, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactivatable_set_related_action, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactivatable_set_use_action_appearance, 0)
    ZEND_ARG_INFO(0, use_appearance)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactivatable_do_set_related_action, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkactivatable_methods[] = {
	ZEND_FENTRY(do_set_related_action, ZEND_MN(GtkActivatable_do_set_related_action), arginfo_gtk_gtkactivatable_do_set_related_action, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_related_action, ZEND_MN(GtkActivatable_get_related_action),   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(get_use_action_appearance, ZEND_MN(GtkActivatable_get_use_action_appearance), NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_related_action, ZEND_MN(GtkActivatable_set_related_action),   arginfo_gtk_gtkactivatable_set_related_action, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_use_action_appearance, ZEND_MN(GtkActivatable_set_use_action_appearance), arginfo_gtk_gtkactivatable_set_use_action_appearance, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(sync_action_properties, ZEND_MN(GtkActivatable_sync_action_properties), arginfo_gtk_gtkactivatable_sync_action_properties, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkOrientable, set_orientation)
{
	GtkOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_orientable_set_orientation(GTK_ORIENTABLE(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkOrientable, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_orientable_get_orientation(GTK_ORIENTABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkorientable_set_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

static zend_function_entry gtkorientable_methods[] = {
	ZEND_FENTRY(get_orientation, ZEND_MN(GtkOrientable_get_orientation),      NULL, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	ZEND_FENTRY(set_orientation, ZEND_MN(GtkOrientable_set_orientation),      arginfo_gtk_gtkorientable_set_orientation, ZEND_ACC_PUBLIC|ZEND_ACC_ABSTRACT)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkMountOperation, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMountOperation);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMountOperation);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkMountOperation, is_showing)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_mount_operation_is_showing(GTK_MOUNT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkMountOperation, set_parent)
{
	zval *parent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &parent, gtkwindow_ce))
		return;

    gtk_mount_operation_set_parent(GTK_MOUNT_OPERATION(PHPG_GOBJECT(this_ptr)), GTK_WINDOW(PHPG_GOBJECT(parent)));

}


static PHP_METHOD(GtkMountOperation, get_parent)
{
	GtkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_mount_operation_get_parent(GTK_MOUNT_OPERATION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMountOperation, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_mount_operation_set_screen(GTK_MOUNT_OPERATION(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}


static PHP_METHOD(GtkMountOperation, get_screen)
{
	GdkScreen* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_mount_operation_get_screen(GTK_MOUNT_OPERATION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmountoperation_set_parent, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmountoperation_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmountoperation_methods[] = {
	PHP_ME(GtkMountOperation, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMountOperation, get_parent,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMountOperation, get_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMountOperation, is_showing,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMountOperation, set_parent,           arginfo_gtk_gtkmountoperation_set_parent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMountOperation, set_screen,           arginfo_gtk_gtkmountoperation_set_screen, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkBuilder, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkBuilder);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkBuilder);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkBuilder, add_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GError *error = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    php_retval = gtk_builder_add_from_file(GTK_BUILDER(PHPG_GOBJECT(this_ptr)), filename, &error);
	if (free_filename) g_free(filename);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_LONG(php_retval);
}

#line 56 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkBuilder, add_from_string)
{
	char *buffer;
	zend_bool free_buffer;
	int buffer_len;
	GError **error = NULL;
	guint retval;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u#", &buffer, &buffer_len, &free_buffer)) {
		return;
	}

	retval = gtk_builder_add_from_string(GTK_BUILDER(PHPG_GOBJECT(this_ptr)), buffer, buffer_len, error);

	if (phpg_handle_gerror(error TSRMLS_CC)) {
		return;
	}

	if (free_buffer) g_free(buffer);
	RETURN_BOOL(retval);
}

#line 6400 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkBuilder, get_object)
{
	char *name;
	zend_bool free_name = FALSE;
	GObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    php_retval = gtk_builder_get_object(GTK_BUILDER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 30 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkBuilder, get_objects)
{
	GSList *list, *current;

	NOT_STATIC_METHOD();

	array_init(return_value);
	list = gtk_builder_get_objects(GTK_BUILDER(PHPG_GOBJECT(this_ptr)));
	for (current = list; current; current = current->next)
	{
		zval *item = NULL;
		phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
		add_next_index_zval(return_value, item);
	}
	g_slist_free(list);
}

#line 6438 "ext/gtk+/gen_gtk.c"


#line 82 "ext/gtk+/gtk-2.12.overrides"
static void phpg_gtkbuilder_connect_signals_func(GtkBuilder *builder, GObject *object,
												  const gchar *signal_name, const gchar *handler_name,
												  GObject *connect_object, GConnectFlags flags,
												  phpg_gtkbuilder_t *data)
{
#ifdef ZTS
	TSRMLS_D;
#endif
	zval *callback = NULL, *php_object = NULL, *user_args = NULL;
	zval **callback_data = NULL, **callback_ptr = NULL;
	GClosure *closure = NULL;
	int handler_id, len = 0, connect_normal = 1;
	char *php_class = NULL, *php_method = NULL;
	zend_class_entry *ce = NULL, **pce;
#ifdef ZTS
	TSRMLS_C = data->TSRMLS_C;
#endif

	/*
	 * First, check if the handler represents a static method. This can be used to invoke
	 * Gtk/Gdk functions, such as Gtk::main_quit(). It may refer to a user class, of
	 * course, but those connections are much less useful.
	 */
	if (strstr(handler_name, "::")) {
		len = strcspn(handler_name, ":");
		php_class = safe_estrndup(handler_name, len);
		php_method = strstr(handler_name, "::") + 2;
		if (zend_lookup_class(php_class, strlen(php_class), &pce TSRMLS_CC) == SUCCESS) {
			ce = *pce;
			if (ce->type == ZEND_INTERNAL_CLASS) {
				connect_normal = 0;
			}
		}
		MAKE_STD_ZVAL(callback);
		array_init(callback);
		add_next_index_string(callback, php_class, 1);
		add_next_index_string(callback, php_method, 1);
		efree(php_class);
	} else {
		/*
		 * Secondly, check if we are supposed to connect handlers to methods of the
		 * specified object - connect_signals_instance.
		 */
		if (data->instance) {
			MAKE_STD_ZVAL(callback);
			array_init(callback);
			zval_add_ref(&data->instance);
			add_next_index_zval(callback, data->instance);
			add_next_index_string(callback, (char *)handler_name, 1);
			user_args = data->extra;

			/*
			 * We verify callability right away, because classes cannot acquire new
			 * methods at runtime (unless you decide to stab yourself in the foot with
			 * runkit).
			 */ 
			if (!zend_is_callable(callback, 0, NULL PHPGTK_ZEND_IS_CALLABLE)) {
				php_error(E_NOTICE, "Unable to auto-connect handler '%s': no such method in the specified object of class '%s'", handler_name, Z_OBJCE_P(data->instance)->name);
				return;
			}
		} else {
			/*
			 * Thirdly, if we have no object, we must have a map representing non-default
			 * handler mappings of the form:
			 *
			 *   array('handler' => <callback>)
			 *   array('handler' => array(<callback> [, $arg1, $arg2, ...]))
			 *   array('handler' => array(null [, $arg1, $arg2, ...]))
			 *
			 * where <callback> is the standard PHP callback form. If null is specified
			 * instead of callback, the connection will be made to the function of the
			 * same name as the handler.
			 */
			if (data->extra && zend_hash_find(Z_ARRVAL_P(data->extra), (char *)handler_name, strlen(handler_name) + 1, (void **)&callback_data) == SUCCESS) {
				/* Check if the value is callable. */
				if (zend_is_callable(*callback_data, 0, NULL PHPGTK_ZEND_IS_CALLABLE)) {
					zval_add_ref(callback_data);
					callback = *callback_data;
				} else {
					/*
					 * If not, check that it is an array with at least one entry, which is
					 * assumed to be the callback reference.
					 */ 
					if (Z_TYPE_PP(callback_data) != IS_ARRAY ||
						zend_hash_index_find(Z_ARRVAL_PP(callback_data), 0, (void **)&callback_ptr) == FAILURE) {
						php_error(E_NOTICE, "Unable to auto-connect handler '%s': invalid callback data structure", handler_name);
						return; 
					}
					if (Z_TYPE_PP(callback_ptr) != IS_NULL) {
						zval_add_ref(callback_ptr);
						callback = *callback_ptr;
					}

					/* Shift callback data - the rest are extra user arguments. */
					zend_hash_index_del(Z_ARRVAL_PP(callback_data), 0);
					user_args = *callback_data;
				}
			}
		}
	}

	/*
	 * Make a simple function callback if everything else failed. 
	 */
	if (callback == NULL) {
		MAKE_STD_ZVAL(callback);
		ZVAL_STRING(callback, (char *)handler_name, 1);
	}

	/*
	 * If we have a swapped connection object, make a PHP wrapper for it and use it with
	 * 'replace' type of connection.
	 */
	if (connect_object) {
		zval *php_connect_object = NULL;
		phpg_gobject_new(&php_connect_object, (GObject*)connect_object TSRMLS_CC);
		closure = phpg_closure_new(callback, user_args, PHPG_CONNECT_REPLACE, php_connect_object TSRMLS_CC);
		/* it's okay to decrement the only reference to the wrapper here - the GObject will save it! */
		zval_ptr_dtor(&php_connect_object);
	} else if (!connect_normal) {
		closure = phpg_closure_new(callback, user_args, PHPG_CONNECT_SIMPLE, NULL TSRMLS_CC);
	} else {
		closure = phpg_closure_new(callback, user_args, PHPG_CONNECT_NORMAL, NULL TSRMLS_CC);
	}

	phpg_gobject_new(&php_object, (GObject*)object TSRMLS_CC);
	phpg_gobject_watch_closure(php_object, closure TSRMLS_CC);
	handler_id = g_signal_connect_closure(object, signal_name, closure, flags);

	/* same as above - okay to decrement the reference here */
	zval_ptr_dtor(&php_object);
	zval_ptr_dtor(&callback);
	if (user_args) {
		zval_ptr_dtor(&user_args);
	}
}

static PHP_METHOD(GtkBuilder, connect_signals)
{
	zval *map = NULL;
	phpg_gtkbuilder_t data;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|a", &map))
		return;

	data.instance = NULL;
	data.extra = map;
#ifdef ZTS
	data.TSRMLS_C = TSRMLS_C;
#endif
	gtk_builder_connect_signals_full(GTK_BUILDER(PHPG_GOBJECT(this_ptr)),
									 (GtkBuilderConnectFunc)phpg_gtkbuilder_connect_signals_func,
									 &data);
}

#line 6599 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkBuilder, set_translation_domain)
{
	char *domain;
	zend_bool free_domain = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &domain, &free_domain))
		return;

    gtk_builder_set_translation_domain(GTK_BUILDER(PHPG_GOBJECT(this_ptr)), domain);
	if (free_domain) g_free(domain);

}


static PHP_METHOD(GtkBuilder, get_translation_domain)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_builder_get_translation_domain(GTK_BUILDER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkBuilder, get_type_from_name)
{
	char *type_name;
	zend_bool free_type_name = FALSE;
	GType php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &type_name, &free_type_name))
		return;

    php_retval = gtk_builder_get_type_from_name(GTK_BUILDER(PHPG_GOBJECT(this_ptr)), type_name);
	if (free_type_name) g_free(type_name);
    phpg_gtype_new(return_value, php_retval TSRMLS_CC);

}

#line 248 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkBuilder, connect_signals_instance)
{
	zval *instance = NULL;
	zval *extra = NULL;
	phpg_gtkbuilder_t data;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "o", &instance))
		return;

	data.instance = instance;
	data.extra = extra;
#ifdef ZTS
	data.TSRMLS_C = TSRMLS_C;
#endif

	gtk_builder_connect_signals_full(GTK_BUILDER(PHPG_GOBJECT(this_ptr)),
									 (GtkBuilderConnectFunc)phpg_gtkbuilder_connect_signals_func,
									 &data);
}

#line 6687 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuilder_add_from_file, 0)
    ZEND_ARG_INFO(0, filename)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkbuilder_add_from_string, 0, 0, 1)
	ZEND_ARG_INFO(0, string)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuilder_get_object, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuilder_set_translation_domain, 0)
    ZEND_ARG_INFO(0, domain)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuilder_get_type_from_name, 0)
    ZEND_ARG_INFO(0, type_name)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuilder_connect_signals_instance, 0)
	ZEND_ARG_INFO(0, object)
ZEND_END_ARG_INFO();


static zend_function_entry gtkbuilder_methods[] = {
	PHP_ME(GtkBuilder, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, add_from_file,        arginfo_gtk_gtkbuilder_add_from_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, add_from_string,      arginfo_gtk_gtkbuilder_add_from_string, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, connect_signals,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, connect_signals_instance, arginfo_gtk_gtkbuilder_connect_signals_instance, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, get_object,           arginfo_gtk_gtkbuilder_get_object, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, get_objects,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, get_translation_domain, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, get_type_from_name,   arginfo_gtk_gtkbuilder_get_type_from_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBuilder, set_translation_domain, arginfo_gtk_gtkbuilder_set_translation_domain, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTooltip, set_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup))
		return;

    gtk_tooltip_set_markup(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkTooltip, set_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_tooltip_set_text(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkTooltip, set_icon)
{
	GdkPixbuf *pixbuf = NULL;
	zval *php_pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_pixbuf, gdkpixbuf_ce))
		return;
    if (Z_TYPE_P(php_pixbuf) != IS_NULL)
        pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));

    gtk_tooltip_set_icon(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), pixbuf);

}


static PHP_METHOD(GtkTooltip, set_icon_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &stock_id, &free_stock_id, &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_tooltip_set_icon_from_stock(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), stock_id, size);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkTooltip, set_custom)
{
	zval *custom_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &custom_widget, gtkwidget_ce))
		return;

    gtk_tooltip_set_custom(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(custom_widget)));

}


static PHP_METHOD(GtkTooltip, trigger_tooltip_query)
{
	zval *display;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &display, gdkdisplay_ce))
		return;

    gtk_tooltip_trigger_tooltip_query(GDK_DISPLAY(PHPG_GOBJECT(display)));

}


static PHP_METHOD(GtkTooltip, set_tip_area)
{
	GdkRectangle rect = { 0, 0, 0, 0 };
	zval *php_rect;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_rect))
		return;

    if (phpg_rectangle_from_zval(php_rect, (GdkRectangle*)&rect TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects rect argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    gtk_tooltip_set_tip_area(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), &rect);

}


static PHP_METHOD(GtkTooltip, set_icon_from_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &icon_name, &free_icon_name, &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_tooltip_set_icon_from_icon_name(GTK_TOOLTIP(PHPG_GOBJECT(this_ptr)), icon_name, size);
	if (free_icon_name) g_free(icon_name);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_icon, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_icon_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_custom, 0)
    ZEND_ARG_OBJ_INFO(0, custom_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_trigger_tooltip_query, 0)
    ZEND_ARG_OBJ_INFO(0, display, GdkDisplay, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_tip_area, 0)
    ZEND_ARG_OBJ_INFO(0, rect, GdkRectangle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltip_set_icon_from_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

static zend_function_entry gtktooltip_methods[] = {
	PHP_ME(GtkTooltip, set_custom,           arginfo_gtk_gtktooltip_set_custom, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_icon,             arginfo_gtk_gtktooltip_set_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_icon_from_icon_name, arginfo_gtk_gtktooltip_set_icon_from_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_icon_from_stock,  arginfo_gtk_gtktooltip_set_icon_from_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_markup,           arginfo_gtk_gtktooltip_set_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_text,             arginfo_gtk_gtktooltip_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, set_tip_area,         arginfo_gtk_gtktooltip_set_tip_area, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltip, trigger_tooltip_query, arginfo_gtk_gtktooltip_trigger_tooltip_query, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPageSetup, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPageSetup);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPageSetup);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkPageSetup, new_from_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPageSetup);
	}

	wrapped_obj = (GObject *) gtk_page_setup_new_from_file(file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPageSetup);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkPageSetup, copy)
{
	GtkPageSetup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_page_setup_copy(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPageSetup, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_page_setup_get_orientation(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPageSetup, set_orientation)
{
	GtkPageOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_PAGE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_page_setup_set_orientation(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkPageSetup, get_paper_size)
{
	GtkPaperSize *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_page_setup_get_paper_size(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_PAPER_SIZE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkPageSetup, set_paper_size)
{
	GtkPaperSize *size = NULL;
	zval *php_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_size, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_size, GTK_TYPE_PAPER_SIZE, FALSE TSRMLS_CC)) {
        size = (GtkPaperSize *) PHPG_GBOXED(php_size);
    } else {
        php_error(E_WARNING, "%s::%s() expects size argument to be a valid GtkPaperSize object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_page_setup_set_paper_size(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), size);

}


static PHP_METHOD(GtkPageSetup, get_top_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_top_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, set_top_margin)
{
	double margin;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &margin, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_page_setup_set_top_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), margin, unit);

}


static PHP_METHOD(GtkPageSetup, get_bottom_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_bottom_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, set_bottom_margin)
{
	double margin;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &margin, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_page_setup_set_bottom_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), margin, unit);

}


static PHP_METHOD(GtkPageSetup, get_left_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_left_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, set_left_margin)
{
	double margin;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &margin, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_page_setup_set_left_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), margin, unit);

}


static PHP_METHOD(GtkPageSetup, get_right_margin)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_right_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, set_right_margin)
{
	double margin;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &margin, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_page_setup_set_right_margin(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), margin, unit);

}


static PHP_METHOD(GtkPageSetup, set_paper_size_and_default_margins)
{
	GtkPaperSize *size = NULL;
	zval *php_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_size, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_size, GTK_TYPE_PAPER_SIZE, FALSE TSRMLS_CC)) {
        size = (GtkPaperSize *) PHPG_GBOXED(php_size);
    } else {
        php_error(E_WARNING, "%s::%s() expects size argument to be a valid GtkPaperSize object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_page_setup_set_paper_size_and_default_margins(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), size);

}


static PHP_METHOD(GtkPageSetup, get_paper_width)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_paper_width(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, get_paper_height)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_paper_height(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, get_page_width)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_page_width(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, get_page_height)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_page_setup_get_page_height(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPageSetup, run_dialog)
{
	GtkWindow *parent = NULL;
	zval *php_parent, *php_page_setup, *php_settings;
	GtkPageSetup *page_setup = NULL;
	GtkPrintSettings *settings = NULL;
	GtkPageSetup* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "NNN", &php_parent, gtkwindow_ce, &php_page_setup, gtkpagesetup_ce, &php_settings, gtkprintsettings_ce))
		return;
    if (Z_TYPE_P(php_parent) != IS_NULL)
        parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
    if (Z_TYPE_P(php_page_setup) != IS_NULL)
        page_setup = GTK_PAGE_SETUP(PHPG_GOBJECT(php_page_setup));
    if (Z_TYPE_P(php_settings) != IS_NULL)
        settings = GTK_PRINT_SETTINGS(PHPG_GOBJECT(php_settings));

    php_retval = gtk_print_run_page_setup_dialog(parent, page_setup, settings);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPageSetup, to_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    php_retval = gtk_page_setup_to_file(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPageSetup, load_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    php_retval = gtk_page_setup_load_file(GTK_PAGE_SETUP(PHPG_GOBJECT(this_ptr)), file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_page_setup_new_from_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_paper_size, 0)
    ZEND_ARG_OBJ_INFO(0, size, GtkPaperSize, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_top_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_top_margin, 0)
    ZEND_ARG_INFO(0, margin)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_bottom_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_bottom_margin, 0)
    ZEND_ARG_INFO(0, margin)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_left_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_left_margin, 0)
    ZEND_ARG_INFO(0, margin)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_right_margin, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_right_margin, 0)
    ZEND_ARG_INFO(0, margin)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_set_paper_size_and_default_margins, 0)
    ZEND_ARG_OBJ_INFO(0, size, GtkPaperSize, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_paper_width, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_paper_height, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_page_width, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_get_page_height, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_run_dialog, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_OBJ_INFO(0, page_setup, GtkPageSetup, 1)
    ZEND_ARG_OBJ_INFO(0, settings, GtkPrintSettings, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_to_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpagesetup_load_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

static zend_function_entry gtkpagesetup_methods[] = {
	PHP_ME(GtkPageSetup, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, new_from_file,        arginfo_gtk_gtkpagesetup_page_setup_new_from_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPageSetup, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_bottom_margin,    arginfo_gtk_gtkpagesetup_get_bottom_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_left_margin,      arginfo_gtk_gtkpagesetup_get_left_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_orientation,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_page_height,      arginfo_gtk_gtkpagesetup_get_page_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_page_width,       arginfo_gtk_gtkpagesetup_get_page_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_paper_height,     arginfo_gtk_gtkpagesetup_get_paper_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_paper_size,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_paper_width,      arginfo_gtk_gtkpagesetup_get_paper_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_right_margin,     arginfo_gtk_gtkpagesetup_get_right_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, get_top_margin,       arginfo_gtk_gtkpagesetup_get_top_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, load_file,            arginfo_gtk_gtkpagesetup_load_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, run_dialog,           arginfo_gtk_gtkpagesetup_run_dialog, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPageSetup, set_bottom_margin,    arginfo_gtk_gtkpagesetup_set_bottom_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_left_margin,      arginfo_gtk_gtkpagesetup_set_left_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_orientation,      arginfo_gtk_gtkpagesetup_set_orientation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_paper_size,       arginfo_gtk_gtkpagesetup_set_paper_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_paper_size_and_default_margins, arginfo_gtk_gtkpagesetup_set_paper_size_and_default_margins, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_right_margin,     arginfo_gtk_gtkpagesetup_set_right_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, set_top_margin,       arginfo_gtk_gtkpagesetup_set_top_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPageSetup, to_file,              arginfo_gtk_gtkpagesetup_to_file, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 673 "ext/gtk+/gtk-2.10.overrides"

static PHP_METHOD(GtkPrintContext, get_cairo_context)
{
    cairo_context_object *context;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    object_init_ex(return_value, php_cairo_get_context_ce());
    context = zend_object_store_get_object(return_value TSRMLS_CC);
    context->context = gtk_print_context_get_cairo_context(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
}

#line 7533 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkPrintContext, get_page_setup)
{
	GtkPageSetup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_page_setup(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintContext, get_width)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_width(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintContext, get_height)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_height(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintContext, get_dpi_x)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_dpi_x(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintContext, get_dpi_y)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_dpi_y(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintContext, get_pango_fontmap)
{
	PangoFontMap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_get_pango_fontmap(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintContext, create_pango_context)
{
	PangoContext* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_create_pango_context(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintContext, create_pango_layout)
{
	PangoLayout* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_context_create_pango_layout(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 697 "ext/gtk+/gtk-2.10.overrides"

static PHP_METHOD(GtkPrintContext, set_cairo_context)
{
    double dpi_x, dpi_y;
    zval *context_zval;
    cairo_context_object *context;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Odd", &context_zval, php_cairo_get_context_ce(), 
                            &dpi_x, &dpi_y))
        return;
    
    context = zend_object_store_get_object(context_zval TSRMLS_CC);

    gtk_print_context_set_cairo_context(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)),
        context->context, dpi_x, dpi_y);
}

#line 7666 "ext/gtk+/gen_gtk.c"


#line 30 "ext/gtk+/gtk-2.20.overrides"
static PHP_METHOD(GtkPrintContext, get_hard_margins)
{
    gdouble top, bottom, left, right;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_print_context_get_hard_margins(GTK_PRINT_CONTEXT(PHPG_GOBJECT(this_ptr)),
            &top, &bottom, &left, &right);

    array_init(return_value);
    add_next_index_long(return_value, top);
    add_next_index_long(return_value, bottom);
    add_next_index_long(return_value, left);
    add_next_index_long(return_value, right);
}

#line 7689 "ext/gtk+/gen_gtk.c"



ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkprintcontext_set_cairo_context, ZEND_SEND_BY_VAL, ZEND_RETURN_VAL, 3)
    ZEND_ARG_OBJ_INFO(0, context, CairoContext, 0)
    ZEND_ARG_INFO(0, dpi_x)
    ZEND_ARG_INFO(0, dpi_y)
ZEND_END_ARG_INFO()


static zend_function_entry gtkprintcontext_methods[] = {
	PHP_ME(GtkPrintContext, create_pango_context, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, create_pango_layout,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_cairo_context,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_dpi_x,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_dpi_y,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_hard_margins,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_height,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_page_setup,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_pango_fontmap,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, get_width,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintContext, set_cairo_context,    arginfo_gtk_gtkprintcontext_set_cairo_context, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPrintOperation, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintOperation);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintOperation);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkPrintOperation, set_default_page_setup)
{
	GtkPageSetup *default_page_setup = NULL;
	zval *php_default_page_setup = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_default_page_setup, gtkpagesetup_ce))
		return;
    if (php_default_page_setup) {
        if (Z_TYPE_P(php_default_page_setup) == IS_NULL)
            default_page_setup = NULL;
        else
            default_page_setup = GTK_PAGE_SETUP(PHPG_GOBJECT(php_default_page_setup));
    }

    gtk_print_operation_set_default_page_setup(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), default_page_setup);

}


static PHP_METHOD(GtkPrintOperation, get_default_page_setup)
{
	GtkPageSetup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_default_page_setup(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintOperation, set_print_settings)
{
	GtkPrintSettings *print_settings = NULL;
	zval *php_print_settings = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_print_settings, gtkprintsettings_ce))
		return;
    if (php_print_settings) {
        if (Z_TYPE_P(php_print_settings) == IS_NULL)
            print_settings = NULL;
        else
            print_settings = GTK_PRINT_SETTINGS(PHPG_GOBJECT(php_print_settings));
    }

    gtk_print_operation_set_print_settings(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), print_settings);

}


static PHP_METHOD(GtkPrintOperation, get_print_settings)
{
	GtkPrintSettings* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_print_settings(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintOperation, set_job_name)
{
	char *job_name;
	zend_bool free_job_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &job_name, &free_job_name))
		return;

    gtk_print_operation_set_job_name(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), job_name);
	if (free_job_name) g_free(job_name);

}


static PHP_METHOD(GtkPrintOperation, set_n_pages)
{
	long n_pages;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n_pages))
		return;

    gtk_print_operation_set_n_pages(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gint)n_pages);

}


static PHP_METHOD(GtkPrintOperation, set_current_page)
{
	long current_page;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &current_page))
		return;

    gtk_print_operation_set_current_page(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gint)current_page);

}


static PHP_METHOD(GtkPrintOperation, set_use_full_page)
{
	zend_bool full_page;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &full_page))
		return;

    gtk_print_operation_set_use_full_page(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)full_page);

}


static PHP_METHOD(GtkPrintOperation, set_unit)
{
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_print_operation_set_unit(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), unit);

}


static PHP_METHOD(GtkPrintOperation, set_export_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_print_operation_set_export_filename(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkPrintOperation, set_track_print_status)
{
	zend_bool track_status;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &track_status))
		return;

    gtk_print_operation_set_track_print_status(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)track_status);

}


static PHP_METHOD(GtkPrintOperation, set_show_progress)
{
	zend_bool show_progress;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_progress))
		return;

    gtk_print_operation_set_show_progress(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)show_progress);

}


static PHP_METHOD(GtkPrintOperation, set_allow_async)
{
	zend_bool allow_async;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &allow_async))
		return;

    gtk_print_operation_set_allow_async(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)allow_async);

}


static PHP_METHOD(GtkPrintOperation, set_custom_tab_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_print_operation_set_custom_tab_label(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkPrintOperation, run)
{
	GtkPrintOperationAction action;
	zval *php_action = NULL, *php_parent = NULL;
	GtkWindow *parent = NULL;
	GError *error = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|N", &php_action, &php_parent, gtkwindow_ce))
		return;

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_PRINT_OPERATION_ACTION, php_action, (gint *)&action) == FAILURE) {
		return;
	}
    if (php_parent) {
        if (Z_TYPE_P(php_parent) == IS_NULL)
            parent = NULL;
        else
            parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
    }

    php_retval = gtk_print_operation_run(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), action, parent, &error);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintOperation, get_error)
{
	GError *error = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_print_operation_get_error(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), &error);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }

}


static PHP_METHOD(GtkPrintOperation, get_status)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_status(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintOperation, get_status_string)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_status_string(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintOperation, is_finished)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_is_finished(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintOperation, cancel)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_print_operation_cancel(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkPrintOperation, draw_page_finish)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_print_operation_draw_page_finish(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkPrintOperation, set_defer_drawing)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_print_operation_set_defer_drawing(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkPrintOperation, get_embed_page_setup)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_embed_page_setup(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintOperation, get_has_selection)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_has_selection(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintOperation, get_n_pages_to_print)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_n_pages_to_print(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintOperation, get_support_selection)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_operation_get_support_selection(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintOperation, set_embed_page_setup)
{
	zend_bool embed;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &embed))
		return;

    gtk_print_operation_set_embed_page_setup(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)embed);

}


static PHP_METHOD(GtkPrintOperation, set_has_selection)
{
	zend_bool has_selection;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_selection))
		return;

    gtk_print_operation_set_has_selection(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)has_selection);

}


static PHP_METHOD(GtkPrintOperation, set_support_selection)
{
	zend_bool support_selection;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &support_selection))
		return;

    gtk_print_operation_set_support_selection(GTK_PRINT_OPERATION(PHPG_GOBJECT(this_ptr)), (gboolean)support_selection);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkprintoperation_set_default_page_setup, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, default_page_setup, GtkPageSetup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkprintoperation_set_print_settings, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, print_settings, GtkPrintSettings, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_job_name, 0)
    ZEND_ARG_INFO(0, job_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_n_pages, 0)
    ZEND_ARG_INFO(0, n_pages)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_current_page, 0)
    ZEND_ARG_INFO(0, current_page)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_use_full_page, 0)
    ZEND_ARG_INFO(0, full_page)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_unit, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_export_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_track_print_status, 0)
    ZEND_ARG_INFO(0, track_status)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_show_progress, 0)
    ZEND_ARG_INFO(0, show_progress)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_allow_async, 0)
    ZEND_ARG_INFO(0, allow_async)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_custom_tab_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkprintoperation_run, 0, 0, 1)
    ZEND_ARG_INFO(0, action)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_get_error, 0)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_embed_page_setup, 0)
    ZEND_ARG_INFO(0, embed)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_has_selection, 0)
    ZEND_ARG_INFO(0, has_selection)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintoperation_set_support_selection, 0)
    ZEND_ARG_INFO(0, support_selection)
ZEND_END_ARG_INFO();

static zend_function_entry gtkprintoperation_methods[] = {
	PHP_ME(GtkPrintOperation, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, cancel,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, draw_page_finish,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_default_page_setup, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_embed_page_setup, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_error,            arginfo_gtk_gtkprintoperation_get_error, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_has_selection,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_n_pages_to_print, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_print_settings,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_status,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_status_string,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, get_support_selection, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, is_finished,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, run,                  arginfo_gtk_gtkprintoperation_run, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_allow_async,      arginfo_gtk_gtkprintoperation_set_allow_async, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_current_page,     arginfo_gtk_gtkprintoperation_set_current_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_custom_tab_label, arginfo_gtk_gtkprintoperation_set_custom_tab_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_default_page_setup, arginfo_gtk_gtkprintoperation_set_default_page_setup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_defer_drawing,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_embed_page_setup, arginfo_gtk_gtkprintoperation_set_embed_page_setup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_export_filename,  arginfo_gtk_gtkprintoperation_set_export_filename, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_has_selection,    arginfo_gtk_gtkprintoperation_set_has_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_job_name,         arginfo_gtk_gtkprintoperation_set_job_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_n_pages,          arginfo_gtk_gtkprintoperation_set_n_pages, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_print_settings,   arginfo_gtk_gtkprintoperation_set_print_settings, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_show_progress,    arginfo_gtk_gtkprintoperation_set_show_progress, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_support_selection, arginfo_gtk_gtkprintoperation_set_support_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_track_print_status, arginfo_gtk_gtkprintoperation_set_track_print_status, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_unit,             arginfo_gtk_gtkprintoperation_set_unit, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintOperation, set_use_full_page,    arginfo_gtk_gtkprintoperation_set_use_full_page, ZEND_ACC_PUBLIC)

	/***   GtkPrintOperationPreview interface implementations   ***/

	PHP_MALIAS(GtkPrintOperationPreview, render_page, render_page, arginfo_gtk_gtkprintoperationpreview_render_page, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkPrintOperationPreview, end_preview, end_preview, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkPrintOperationPreview, is_selected, is_selected, arginfo_gtk_gtkprintoperationpreview_is_selected, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPrintSettings, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintSettings);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintSettings);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkPrintSettings, new_from_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintSettings);
	}

	wrapped_obj = (GObject *) gtk_print_settings_new_from_file(file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPrintSettings);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkPrintSettings, copy)
{
	GtkPrintSettings* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_copy(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPrintSettings, has_key)
{
	char *key;
	zend_bool free_key = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    php_retval = gtk_print_settings_has_key(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, get)
{
	char *key;
	zend_bool free_key = FALSE, free_result;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    php_retval = gtk_print_settings_get(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set)
{
	char *key, *value;
	zend_bool free_key = FALSE, free_value = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uu", &key, &free_key, &value, &free_value))
		return;

    gtk_print_settings_set(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, value);
	if (free_key) g_free(key);
	if (free_value) g_free(value);

}


static PHP_METHOD(GtkPrintSettings, unset)
{
	char *key;
	zend_bool free_key = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    gtk_print_settings_unset(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);

}


static PHP_METHOD(GtkPrintSettings, get_bool)
{
	char *key;
	zend_bool free_key = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    php_retval = gtk_print_settings_get_bool(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_bool)
{
	char *key;
	zend_bool free_key = FALSE, value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ub", &key, &free_key, &value))
		return;

    gtk_print_settings_set_bool(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, (gboolean)value);
	if (free_key) g_free(key);

}


static PHP_METHOD(GtkPrintSettings, get_double)
{
	char *key;
	zend_bool free_key = FALSE;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    php_retval = gtk_print_settings_get_double(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, get_double_with_default)
{
	char *key;
	zend_bool free_key = FALSE;
	double def, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ud", &key, &free_key, &def))
		return;

    php_retval = gtk_print_settings_get_double_with_default(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, def);
	if (free_key) g_free(key);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_double)
{
	char *key;
	zend_bool free_key = FALSE;
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ud", &key, &free_key, &value))
		return;

    gtk_print_settings_set_double(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, value);
	if (free_key) g_free(key);

}


static PHP_METHOD(GtkPrintSettings, get_length)
{
	char *key;
	zend_bool free_key = FALSE;
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &key, &free_key, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_print_settings_get_length(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, unit);
	if (free_key) g_free(key);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_length)
{
	char *key;
	zend_bool free_key = FALSE;
	double value;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "udV", &key, &free_key, &value, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_print_settings_set_length(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, value, unit);
	if (free_key) g_free(key);

}


static PHP_METHOD(GtkPrintSettings, get_int)
{
	char *key;
	zend_bool free_key = FALSE;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &key, &free_key))
		return;

    php_retval = gtk_print_settings_get_int(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key);
	if (free_key) g_free(key);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, get_int_with_default)
{
	char *key;
	zend_bool free_key = FALSE;
	long def, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ui", &key, &free_key, &def))
		return;

    php_retval = gtk_print_settings_get_int_with_default(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, (gint)def);
	if (free_key) g_free(key);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_int)
{
	char *key;
	zend_bool free_key = FALSE;
	long value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ui", &key, &free_key, &value))
		return;

    gtk_print_settings_set_int(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), key, (gint)value);
	if (free_key) g_free(key);

}


static PHP_METHOD(GtkPrintSettings, get_printer)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_printer(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_printer)
{
	char *printer;
	zend_bool free_printer = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &printer, &free_printer))
		return;

    gtk_print_settings_set_printer(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), printer);
	if (free_printer) g_free(printer);

}


static PHP_METHOD(GtkPrintSettings, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_orientation(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_orientation)
{
	GtkPageOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_PAGE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_print_settings_set_orientation(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkPrintSettings, get_paper_size)
{
	GtkPaperSize *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_paper_size(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_PAPER_SIZE, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkPrintSettings, set_paper_size)
{
	GtkPaperSize *paper_size = NULL;
	zval *php_paper_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_paper_size, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_paper_size, GTK_TYPE_PAPER_SIZE, FALSE TSRMLS_CC)) {
        paper_size = (GtkPaperSize *) PHPG_GBOXED(php_paper_size);
    } else {
        php_error(E_WARNING, "%s::%s() expects paper_size argument to be a valid GtkPaperSize object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_print_settings_set_paper_size(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), paper_size);

}


static PHP_METHOD(GtkPrintSettings, get_paper_width)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_print_settings_get_paper_width(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_paper_width)
{
	double width;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &width, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_print_settings_set_paper_width(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), width, unit);

}


static PHP_METHOD(GtkPrintSettings, get_paper_height)
{
	GtkUnit unit;
	zval *php_unit = NULL;
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    php_retval = gtk_print_settings_get_paper_height(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), unit);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_paper_height)
{
	double height;
	GtkUnit unit;
	zval *php_unit = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dV", &height, &php_unit))
		return;

	if (php_unit && phpg_gvalue_get_enum(GTK_TYPE_UNIT, php_unit, (gint *)&unit) == FAILURE) {
		return;
	}

    gtk_print_settings_set_paper_height(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), height, unit);

}


static PHP_METHOD(GtkPrintSettings, get_use_color)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_use_color(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_use_color)
{
	zend_bool use_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_color))
		return;

    gtk_print_settings_set_use_color(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gboolean)use_color);

}


static PHP_METHOD(GtkPrintSettings, get_collate)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_collate(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_collate)
{
	zend_bool collate;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &collate))
		return;

    gtk_print_settings_set_collate(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gboolean)collate);

}


static PHP_METHOD(GtkPrintSettings, get_reverse)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_reverse(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_reverse)
{
	zend_bool reverse;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reverse))
		return;

    gtk_print_settings_set_reverse(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gboolean)reverse);

}


static PHP_METHOD(GtkPrintSettings, get_duplex)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_duplex(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_duplex)
{
	GtkPrintDuplex duplex;
	zval *php_duplex = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_duplex))
		return;

	if (php_duplex && phpg_gvalue_get_enum(GTK_TYPE_PRINT_DUPLEX, php_duplex, (gint *)&duplex) == FAILURE) {
		return;
	}

    gtk_print_settings_set_duplex(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), duplex);

}


static PHP_METHOD(GtkPrintSettings, get_quality)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_quality(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_quality)
{
	GtkPrintQuality quality;
	zval *php_quality = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_quality))
		return;

	if (php_quality && phpg_gvalue_get_enum(GTK_TYPE_PRINT_QUALITY, php_quality, (gint *)&quality) == FAILURE) {
		return;
	}

    gtk_print_settings_set_quality(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), quality);

}


static PHP_METHOD(GtkPrintSettings, get_n_copies)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_n_copies(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_n_copies)
{
	long num_copies;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &num_copies))
		return;

    gtk_print_settings_set_n_copies(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gint)num_copies);

}


static PHP_METHOD(GtkPrintSettings, get_number_up)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_number_up(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_number_up)
{
	long number_up;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &number_up))
		return;

    gtk_print_settings_set_number_up(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gint)number_up);

}


static PHP_METHOD(GtkPrintSettings, get_resolution)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_resolution(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_resolution)
{
	long resolution;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &resolution))
		return;

    gtk_print_settings_set_resolution(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gint)resolution);

}


static PHP_METHOD(GtkPrintSettings, get_scale)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_scale(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_scale)
{
	double scale;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &scale))
		return;

    gtk_print_settings_set_scale(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), scale);

}


static PHP_METHOD(GtkPrintSettings, get_print_pages)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_print_pages(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_print_pages)
{
	GtkPrintPages pages;
	zval *php_pages = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_pages))
		return;

	if (php_pages && phpg_gvalue_get_enum(GTK_TYPE_PRINT_PAGES, php_pages, (gint *)&pages) == FAILURE) {
		return;
	}

    gtk_print_settings_set_print_pages(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), pages);

}


static PHP_METHOD(GtkPrintSettings, get_page_set)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_page_set(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_page_set)
{
	GtkPageSet page_set;
	zval *php_page_set = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_page_set))
		return;

	if (php_page_set && phpg_gvalue_get_enum(GTK_TYPE_PAGE_SET, php_page_set, (gint *)&page_set) == FAILURE) {
		return;
	}

    gtk_print_settings_set_page_set(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), page_set);

}


static PHP_METHOD(GtkPrintSettings, get_default_source)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_default_source(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_default_source)
{
	char *default_source;
	zend_bool free_default_source = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &default_source, &free_default_source))
		return;

    gtk_print_settings_set_default_source(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), default_source);
	if (free_default_source) g_free(default_source);

}


static PHP_METHOD(GtkPrintSettings, get_media_type)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_media_type(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_media_type)
{
	char *media_type;
	zend_bool free_media_type = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &media_type, &free_media_type))
		return;

    gtk_print_settings_set_media_type(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), media_type);
	if (free_media_type) g_free(media_type);

}


static PHP_METHOD(GtkPrintSettings, get_dither)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_dither(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_dither)
{
	char *dither;
	zend_bool free_dither = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &dither, &free_dither))
		return;

    gtk_print_settings_set_dither(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), dither);
	if (free_dither) g_free(dither);

}


static PHP_METHOD(GtkPrintSettings, get_finishings)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_finishings(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_finishings)
{
	char *finishings;
	zend_bool free_finishings = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &finishings, &free_finishings))
		return;

    gtk_print_settings_set_finishings(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), finishings);
	if (free_finishings) g_free(finishings);

}


static PHP_METHOD(GtkPrintSettings, get_output_bin)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_output_bin(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkPrintSettings, set_output_bin)
{
	char *output_bin;
	zend_bool free_output_bin = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &output_bin, &free_output_bin))
		return;

    gtk_print_settings_set_output_bin(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), output_bin);
	if (free_output_bin) g_free(output_bin);

}


static PHP_METHOD(GtkPrintSettings, to_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    php_retval = gtk_print_settings_to_file(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, load_file)
{
	char *file_name;
	zend_bool free_file_name = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    php_retval = gtk_print_settings_load_file(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), file_name, &error);
	if (free_file_name) g_free(file_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPrintSettings, get_number_up_layout)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_number_up_layout(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_number_up_layout)
{
	GtkNumberUpLayout number_up_layout;
	zval *php_number_up_layout = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_number_up_layout))
		return;

	if (php_number_up_layout && phpg_gvalue_get_enum(GTK_TYPE_NUMBER_UP_LAYOUT, php_number_up_layout, (gint *)&number_up_layout) == FAILURE) {
		return;
	}

    gtk_print_settings_set_number_up_layout(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), number_up_layout);

}


static PHP_METHOD(GtkPrintSettings, get_resolution_x)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_resolution_x(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, get_resolution_y)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_resolution_y(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_resolution_xy)
{
	long resolution_x, resolution_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &resolution_x, &resolution_y))
		return;

    gtk_print_settings_set_resolution_xy(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), (gint)resolution_x, (gint)resolution_y);

}


static PHP_METHOD(GtkPrintSettings, get_printer_lpi)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_print_settings_get_printer_lpi(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkPrintSettings, set_printer_lpi)
{
	double lpi;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &lpi))
		return;

    gtk_print_settings_set_printer_lpi(GTK_PRINT_SETTINGS(PHPG_GOBJECT(this_ptr)), lpi);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_print_settings_new_from_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_has_key, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_unset, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_bool, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_bool, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_double, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_double_with_default, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, def)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_double, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_length, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_length, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_int, 0)
    ZEND_ARG_INFO(0, key)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_int_with_default, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, def)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_int, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_printer, 0)
    ZEND_ARG_INFO(0, printer)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_paper_size, 0)
    ZEND_ARG_OBJ_INFO(0, paper_size, GtkPaperSize, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_paper_width, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_paper_width, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_get_paper_height, 0)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_paper_height, 0)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, unit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_use_color, 0)
    ZEND_ARG_INFO(0, use_color)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_collate, 0)
    ZEND_ARG_INFO(0, collate)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_reverse, 0)
    ZEND_ARG_INFO(0, reverse)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_duplex, 0)
    ZEND_ARG_INFO(0, duplex)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_quality, 0)
    ZEND_ARG_INFO(0, quality)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_n_copies, 0)
    ZEND_ARG_INFO(0, num_copies)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_number_up, 0)
    ZEND_ARG_INFO(0, number_up)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_resolution, 0)
    ZEND_ARG_INFO(0, resolution)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_scale, 0)
    ZEND_ARG_INFO(0, scale)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_print_pages, 0)
    ZEND_ARG_INFO(0, pages)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_page_set, 0)
    ZEND_ARG_INFO(0, page_set)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_default_source, 0)
    ZEND_ARG_INFO(0, default_source)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_media_type, 0)
    ZEND_ARG_INFO(0, media_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_dither, 0)
    ZEND_ARG_INFO(0, dither)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_finishings, 0)
    ZEND_ARG_INFO(0, finishings)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_output_bin, 0)
    ZEND_ARG_INFO(0, output_bin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_to_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_load_file, 0)
    ZEND_ARG_INFO(0, file_name)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_number_up_layout, 0)
    ZEND_ARG_OBJ_INFO(0, number_up_layout, GtkNumberUpLayout, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_resolution_xy, 0)
    ZEND_ARG_INFO(0, resolution_x)
    ZEND_ARG_INFO(0, resolution_y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprintsettings_set_printer_lpi, 0)
    ZEND_ARG_INFO(0, lpi)
ZEND_END_ARG_INFO();

static zend_function_entry gtkprintsettings_methods[] = {
	PHP_ME(GtkPrintSettings, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, new_from_file,        arginfo_gtk_gtkprintsettings_print_settings_new_from_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPrintSettings, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get,                  arginfo_gtk_gtkprintsettings_get, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_bool,             arginfo_gtk_gtkprintsettings_get_bool, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_collate,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_default_source,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_dither,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_double,           arginfo_gtk_gtkprintsettings_get_double, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_double_with_default, arginfo_gtk_gtkprintsettings_get_double_with_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_duplex,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_finishings,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_int,              arginfo_gtk_gtkprintsettings_get_int, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_int_with_default, arginfo_gtk_gtkprintsettings_get_int_with_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_length,           arginfo_gtk_gtkprintsettings_get_length, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_media_type,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_n_copies,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_number_up,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_number_up_layout, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_orientation,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_output_bin,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_page_set,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_paper_height,     arginfo_gtk_gtkprintsettings_get_paper_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_paper_size,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_paper_width,      arginfo_gtk_gtkprintsettings_get_paper_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_print_pages,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_printer,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_printer_lpi,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_quality,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_resolution,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_resolution_x,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_resolution_y,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_reverse,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_scale,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, get_use_color,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, has_key,              arginfo_gtk_gtkprintsettings_has_key, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, load_file,            arginfo_gtk_gtkprintsettings_load_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set,                  arginfo_gtk_gtkprintsettings_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_bool,             arginfo_gtk_gtkprintsettings_set_bool, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_collate,          arginfo_gtk_gtkprintsettings_set_collate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_default_source,   arginfo_gtk_gtkprintsettings_set_default_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_dither,           arginfo_gtk_gtkprintsettings_set_dither, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_double,           arginfo_gtk_gtkprintsettings_set_double, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_duplex,           arginfo_gtk_gtkprintsettings_set_duplex, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_finishings,       arginfo_gtk_gtkprintsettings_set_finishings, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_int,              arginfo_gtk_gtkprintsettings_set_int, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_length,           arginfo_gtk_gtkprintsettings_set_length, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_media_type,       arginfo_gtk_gtkprintsettings_set_media_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_n_copies,         arginfo_gtk_gtkprintsettings_set_n_copies, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_number_up,        arginfo_gtk_gtkprintsettings_set_number_up, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_number_up_layout, arginfo_gtk_gtkprintsettings_set_number_up_layout, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_orientation,      arginfo_gtk_gtkprintsettings_set_orientation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_output_bin,       arginfo_gtk_gtkprintsettings_set_output_bin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_page_set,         arginfo_gtk_gtkprintsettings_set_page_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_paper_height,     arginfo_gtk_gtkprintsettings_set_paper_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_paper_size,       arginfo_gtk_gtkprintsettings_set_paper_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_paper_width,      arginfo_gtk_gtkprintsettings_set_paper_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_print_pages,      arginfo_gtk_gtkprintsettings_set_print_pages, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_printer,          arginfo_gtk_gtkprintsettings_set_printer, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_printer_lpi,      arginfo_gtk_gtkprintsettings_set_printer_lpi, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_quality,          arginfo_gtk_gtkprintsettings_set_quality, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_resolution,       arginfo_gtk_gtkprintsettings_set_resolution, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_resolution_xy,    arginfo_gtk_gtkprintsettings_set_resolution_xy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_reverse,          arginfo_gtk_gtkprintsettings_set_reverse, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_scale,            arginfo_gtk_gtkprintsettings_set_scale, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, set_use_color,        arginfo_gtk_gtkprintsettings_set_use_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, to_file,              arginfo_gtk_gtkprintsettings_to_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPrintSettings, unset,                arginfo_gtk_gtkprintsettings_unset, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentManager, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentManager);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentManager);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkRecentManager, get_default)
{
	GtkRecentManager* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_manager_get_default();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


static PHP_METHOD(GtkRecentManager, get_for_screen)
{
	zval *screen;
	GtkRecentManager* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    php_retval = gtk_recent_manager_get_for_screen(GDK_SCREEN(PHPG_GOBJECT(screen)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


static PHP_METHOD(GtkRecentManager, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_recent_manager_set_screen(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}


static PHP_METHOD(GtkRecentManager, add_item)
{
	char *uri;
	zend_bool free_uri = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_manager_add_item(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentManager, remove_item)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_manager_remove_item(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentManager, lookup_item)
{
	char *uri;
	zend_bool free_uri = FALSE;
	GError *error = NULL;
	GtkRecentInfo *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_manager_lookup_item(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), uri, &error);
	if (free_uri) g_free(uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	phpg_gboxed_new(&return_value, GTK_TYPE_RECENT_INFO, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkRecentManager, has_item)
{
	char *uri;
	zend_bool free_uri = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    php_retval = gtk_recent_manager_has_item(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentManager, move_item)
{
	char *uri, *new_uri;
	zend_bool free_uri = FALSE, free_new_uri = FALSE;
	GError *error = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uu", &uri, &free_uri, &new_uri, &free_new_uri))
		return;

    php_retval = gtk_recent_manager_move_item(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), uri, new_uri, &error);
	if (free_uri) g_free(uri);
	if (free_new_uri) g_free(new_uri);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentManager, set_limit)
{
	long limit;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &limit))
		return;

    gtk_recent_manager_set_limit(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), (gint)limit);

}


static PHP_METHOD(GtkRecentManager, get_limit)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_manager_get_limit(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 167 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentManager, get_items)
{
    GList *list = NULL, *element = NULL;
    zval *current = NULL;

    NOT_STATIC_METHOD();

    if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"")) {
        return;
    }

    list = gtk_recent_manager_get_items(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);

    for(element = list; element; element = element->next) {
        MAKE_STD_ZVAL(current);
        phpg_gboxed_new(
            &current,
            GTK_TYPE_RECENT_INFO,
            element->data,
            TRUE,
            TRUE TSRMLS_CC
        );

        add_next_index_zval(return_value,current);
    }

    g_list_foreach(list,(GFunc)gtk_recent_info_unref,NULL);
    g_list_free(list);
}

#line 10070 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRecentManager, purge_items)
{
	GError *error = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_manager_purge_items(GTK_RECENT_MANAGER(PHPG_GOBJECT(this_ptr)), &error);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_get_for_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_add_item, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_remove_item, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_lookup_item, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_has_item, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_move_item, 0)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, new_uri)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_set_limit, 0)
    ZEND_ARG_INFO(0, limit)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentmanager_purge_items, 0)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentmanager_methods[] = {
	PHP_ME(GtkRecentManager, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, add_item,             arginfo_gtk_gtkrecentmanager_add_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, get_default,          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkRecentManager, get_for_screen,       arginfo_gtk_gtkrecentmanager_get_for_screen, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkRecentManager, get_items,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, get_limit,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, has_item,             arginfo_gtk_gtkrecentmanager_has_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, lookup_item,          arginfo_gtk_gtkrecentmanager_lookup_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, move_item,            arginfo_gtk_gtkrecentmanager_move_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, purge_items,          arginfo_gtk_gtkrecentmanager_purge_items, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, remove_item,          arginfo_gtk_gtkrecentmanager_remove_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, set_limit,            arginfo_gtk_gtkrecentmanager_set_limit, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentManager, set_screen,           arginfo_gtk_gtkrecentmanager_set_screen, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkStatusIcon, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkStatusIcon, new_from_pixbuf)
{
	zval *pixbuf;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}

	wrapped_obj = (GObject *) gtk_status_icon_new_from_pixbuf(GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkStatusIcon, new_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}

	wrapped_obj = (GObject *) gtk_status_icon_new_from_file(filename);
	if (free_filename) g_free(filename);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkStatusIcon, new_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}

	wrapped_obj = (GObject *) gtk_status_icon_new_from_stock(stock_id);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkStatusIcon, new_from_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}

	wrapped_obj = (GObject *) gtk_status_icon_new_from_icon_name(icon_name);
	if (free_icon_name) g_free(icon_name);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusIcon);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkStatusIcon, set_from_pixbuf)
{
	GdkPixbuf *pixbuf = NULL;
	zval *php_pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_pixbuf, gdkpixbuf_ce))
		return;
    if (Z_TYPE_P(php_pixbuf) != IS_NULL)
        pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));

    gtk_status_icon_set_from_pixbuf(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), pixbuf);

}


static PHP_METHOD(GtkStatusIcon, set_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_status_icon_set_from_file(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkStatusIcon, set_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    gtk_status_icon_set_from_stock(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkStatusIcon, set_from_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_status_icon_set_from_icon_name(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkStatusIcon, get_storage_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_storage_type(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkStatusIcon, get_pixbuf)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_pixbuf(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkStatusIcon, get_stock)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_stock(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkStatusIcon, get_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_icon_name(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkStatusIcon, get_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_size(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkStatusIcon, set_tooltip)
{
	char *tooltip_text;
	zend_bool free_tooltip_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &tooltip_text, &free_tooltip_text))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, Use GtkStatusIcon::set_tooltip_text() instead" TSRMLS_CC);

    gtk_status_icon_set_tooltip(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), tooltip_text);
	if (free_tooltip_text) g_free(tooltip_text);

}


static PHP_METHOD(GtkStatusIcon, set_visible)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_status_icon_set_visible(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


static PHP_METHOD(GtkStatusIcon, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_visible(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkStatusIcon, set_blinking)
{
	zend_bool blinking;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &blinking))
		return;

    gtk_status_icon_set_blinking(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), (gboolean)blinking);

}


static PHP_METHOD(GtkStatusIcon, get_blinking)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_blinking(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkStatusIcon, is_embedded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_is_embedded(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 120 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkStatusIcon, position_menu)
{
    zval *menu = NULL, *statusicon = NULL;
    gint x, y;
    gboolean push_in;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &menu, gtkmenu_ce, &statusicon, gtkstatusicon_ce))
        return;

    gtk_status_icon_position_menu(GTK_MENU(PHPG_GOBJECT(menu)), &x, &y, &push_in, PHPG_GOBJECT(statusicon));

    php_gtk_build_value(&return_value, "(iii)", x, y, push_in);
}

#line 10529 "ext/gtk+/gen_gtk.c"


#line 136 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkStatusIcon, get_geometry)
{
    GdkScreen *screen;
    GdkRectangle rect;
    GtkOrientation orientation;
    gboolean ret;
    zval *php_screen = NULL, *php_rect = NULL, *php_orient = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    ret = gtk_status_icon_get_geometry(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)),
                                       &screen, &rect, &orientation);

    if (ret)
    {
      phpg_gobject_new(&php_screen, (GObject *) screen TSRMLS_CC);
      phpg_gboxed_new(&php_rect, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
      phpg_gobject_new(&php_orient, (GObject *) orientation TSRMLS_CC);
      php_gtk_build_value(&return_value, "(NNN)", screen, rect, orientation);
    }
}
#line 10558 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkStatusIcon, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_status_icon_set_screen(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}


static PHP_METHOD(GtkStatusIcon, get_screen)
{
	GdkScreen* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_screen(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkStatusIcon, get_x11_window_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_x11_window_id(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkStatusIcon, get_has_tooltip)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_has_tooltip(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkStatusIcon, get_tooltip_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_tooltip_text(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkStatusIcon, get_tooltip_markup)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_tooltip_markup(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkStatusIcon, set_has_tooltip)
{
	zend_bool has_tooltip;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_tooltip))
		return;

    gtk_status_icon_set_has_tooltip(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), (gboolean)has_tooltip);

}


static PHP_METHOD(GtkStatusIcon, set_tooltip_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_status_icon_set_tooltip_text(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkStatusIcon, set_tooltip_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup))
		return;

    gtk_status_icon_set_tooltip_markup(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkStatusIcon, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_status_icon_get_title(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkStatusIcon, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_status_icon_set_title(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkStatusIcon, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_status_icon_set_name(GTK_STATUS_ICON(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_status_icon_new_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_status_icon_new_from_file, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_status_icon_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_status_icon_new_from_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_from_file, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_from_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_tooltip, 0)
    ZEND_ARG_INFO(0, tooltip_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_blinking, 0)
    ZEND_ARG_INFO(0, blinking)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_position_menu, 0)
    ZEND_ARG_OBJ_INFO(0, menu, GtkMenu, 1)
    ZEND_ARG_OBJ_INFO(0, statusicon, GtkStatusIcon, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_has_tooltip, 0)
    ZEND_ARG_INFO(0, has_tooltip)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_tooltip_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_tooltip_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusicon_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

static zend_function_entry gtkstatusicon_methods[] = {
	PHP_ME(GtkStatusIcon, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, new_from_pixbuf,      arginfo_gtk_gtkstatusicon_status_icon_new_from_pixbuf, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkStatusIcon, new_from_file,        arginfo_gtk_gtkstatusicon_status_icon_new_from_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkStatusIcon, new_from_stock,       arginfo_gtk_gtkstatusicon_status_icon_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkStatusIcon, new_from_icon_name,   arginfo_gtk_gtkstatusicon_status_icon_new_from_icon_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkStatusIcon, get_blinking,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_geometry,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_has_tooltip,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_stock,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_storage_type,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_tooltip_markup,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_tooltip_text,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, get_x11_window_id,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, is_embedded,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, position_menu,        arginfo_gtk_gtkstatusicon_position_menu, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkStatusIcon, set_blinking,         arginfo_gtk_gtkstatusicon_set_blinking, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_from_file,        arginfo_gtk_gtkstatusicon_set_from_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_from_icon_name,   arginfo_gtk_gtkstatusicon_set_from_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_from_pixbuf,      arginfo_gtk_gtkstatusicon_set_from_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_from_stock,       arginfo_gtk_gtkstatusicon_set_from_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_has_tooltip,      arginfo_gtk_gtkstatusicon_set_has_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_name,             arginfo_gtk_gtkstatusicon_set_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_screen,           arginfo_gtk_gtkstatusicon_set_screen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_title,            arginfo_gtk_gtkstatusicon_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_tooltip,          arginfo_gtk_gtkstatusicon_set_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_tooltip_markup,   arginfo_gtk_gtkstatusicon_set_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_tooltip_text,     arginfo_gtk_gtkstatusicon_set_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusIcon, set_visible,          arginfo_gtk_gtkstatusicon_set_visible, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(PhpGtkCustomTreeModel, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(PhpGtkCustomTreeModel);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(PhpGtkCustomTreeModel);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 3082 "ext/gtk+/gtktreeview.overrides"
static zval* phpg_phpgtkcustomtreemodel_read_dimension_handler(zval *object, zval *offset, int type TSRMLS_DC)
{
    return phpg_gtktreemodel_read_dimension_handler(object, offset, type TSRMLS_CC);
}

#line 10936 "ext/gtk+/gen_gtk.c"


#line 3089 "ext/gtk+/gtktreeview.overrides"
static void phpg_phpgtkcustomtreemodel_write_dimension_handler(zval *object, zval *offset, zval *value TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, value TSRMLS_CC);
}

#line 10945 "ext/gtk+/gen_gtk.c"


#line 3075 "ext/gtk+/gtktreeview.overrides"
static int phpg_phpgtkcustomtreemodel_has_dimension_handler(zval *object, zval *offset, int check_empty TSRMLS_DC)
{
    return phpg_gtktreemodel_has_dimension_handler(object, offset, check_empty TSRMLS_CC);
}

#line 10954 "ext/gtk+/gen_gtk.c"


#line 3096 "ext/gtk+/gtktreeview.overrides"
static void phpg_phpgtkcustomtreemodel_unset_dimension_handler(zval *object, zval *offset TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, NULL TSRMLS_CC);
}

#line 10963 "ext/gtk+/gen_gtk.c"


#line 3068 "ext/gtk+/gtktreeview.overrides"
static int phpg_phpgtkcustomtreemodel_count_elements_handler(zval *object, long *count TSRMLS_DC)
{
    return phpg_gtktreemodel_count_elements_handler(object, count TSRMLS_CC);
}

#line 10972 "ext/gtk+/gen_gtk.c"



static zend_object_handlers phpgtkcustomtreemodel_object_handlers;

static zend_object_value phpg_create_phpgtkcustomtreemodel(zend_class_entry *ce TSRMLS_DC)
{
    zend_object_value zov;

    zov = phpg_create_gobject(ce TSRMLS_CC);
    zov.handlers = &phpgtkcustomtreemodel_object_handlers;

    return zov;
}

static zend_function_entry phpgtkcustomtreemodel_methods[] = {
	PHP_ME(PhpGtkCustomTreeModel, __construct,          NULL, ZEND_ACC_PUBLIC)

	/***   GtkTreeModel interface implementations   ***/

	PHP_MALIAS(GtkTreeModel, get, get, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_column_type, get_column_type, arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_flags, get_flags, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter, get_iter, arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_first, get_iter_first, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_root, get_iter_root, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_from_string, get_iter_from_string, arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_n_columns, get_n_columns, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_path, get_path, arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_string_from_iter, get_string_from_iter, arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_value, get_value, arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_children, iter_children, arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_has_child, iter_has_child, arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_n_children, iter_n_children, arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_next, iter_next, arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_nth_child, iter_nth_child, arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_parent, iter_parent, arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, ref_node, ref_node, arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_changed, row_changed, arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_deleted, row_deleted, arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_has_child_toggled, row_has_child_toggled, arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_inserted, row_inserted, arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, rows_reordered, rows_reordered, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, unref_node, unref_node, arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, for_each, for_each, arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAccelGroup, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAccelGroup);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAccelGroup);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkAccelGroup, disconnect_key)
{
	long accel_key;
	GdkModifierType accel_mods;
	zval *php_accel_mods = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &accel_key, &php_accel_mods))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_accel_group_disconnect_key(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)), (guint)accel_key, accel_mods);
	RETVAL_BOOL(php_retval);
}

#line 807 "ext/gtk+/gtk.overrides"
static gboolean phpg_accelgroup_find_func_marshal(GtkAccelKey *key, GClosure *closure, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_key = NULL, *retval = NULL;
    gboolean return_value;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return (gboolean) 0;
    }
    /* TODO - check closure to see if it's invalidated? If so error and return false? */
    
    /* Make the key useable */
    MAKE_STD_ZVAL(php_key);
    array_init(php_key);
    add_next_index_long(php_key, key->accel_key);
    add_next_index_long(php_key, key->accel_mods);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 1, &n_args);
    args[0] = &php_key;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_key);

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
    
    convert_to_boolean(retval);
    return_value = Z_BVAL_P(retval);
    zval_ptr_dtor(&retval);
    return return_value;
}

static PHP_METHOD(GtkAccelGroup, find)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkAccelGroupFindFunc callback;
    GtkAccelKey *result;

	NOT_STATIC_METHOD();
 
    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (Z_TYPE_P(php_callback) == IS_NULL) {
        cb_data  = NULL;
        callback = NULL;
    } else {
        zval_add_ref(&php_callback);
        cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
        callback = (GtkAccelGroupFindFunc)phpg_accelgroup_find_func_marshal;
    }

    result = gtk_accel_group_find(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)), callback, cb_data);
    /* only if we have a result */
    if (result) {
		array_init(return_value);
		add_next_index_long(return_value, result->accel_key);
		add_next_index_long(return_value, result->accel_mods);
	}
	
	if (cb_data) {
		phpg_cb_data_destroy(cb_data);
	}
}

#line 11144 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAccelGroup, lock)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_accel_group_lock(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAccelGroup, unlock)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_accel_group_unlock(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAccelGroup, get_is_locked)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_group_get_is_locked(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAccelGroup, get_modifier_mask)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_group_get_modifier_mask(GTK_ACCEL_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelgroup_disconnect_key, 0)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelgroup_find, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtkaccelgroup_methods[] = {
	PHP_ME(GtkAccelGroup, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, disconnect_key,       arginfo_gtk_gtkaccelgroup_disconnect_key, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, find,                 arginfo_gtk_gtkaccelgroup_find, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, get_is_locked,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, get_modifier_mask,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, lock,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelGroup, unlock,               NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAccelMap, add_entry)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE;
	long accel_key;
	GdkModifierType accel_mods;
	zval *php_accel_mods = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiV", &accel_path, &free_accel_path, &accel_key, &php_accel_mods))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

    gtk_accel_map_add_entry(accel_path, (guint)accel_key, accel_mods);
	if (free_accel_path) g_free(accel_path);

}


static PHP_METHOD(GtkAccelMap, add_filter)
{
	char *filter_pattern;
	zend_bool free_filter_pattern = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filter_pattern, &free_filter_pattern))
		return;

    gtk_accel_map_add_filter(filter_pattern);
	if (free_filter_pattern) g_free(filter_pattern);

}


static PHP_METHOD(GtkAccelMap, change_entry)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE, replace;
	long accel_key;
	GdkModifierType accel_mods;
	zval *php_accel_mods = NULL;
	gboolean php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiVb", &accel_path, &free_accel_path, &accel_key, &php_accel_mods, &replace))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_accel_map_change_entry(accel_path, (guint)accel_key, accel_mods, (gboolean)replace);
	if (free_accel_path) g_free(accel_path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAccelMap, get)
{
	GtkAccelMap* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_map_get();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAccelMap, load)
{
	char *file_name;
	zend_bool free_file_name = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    gtk_accel_map_load(file_name);
	if (free_file_name) g_free(file_name);

}


static PHP_METHOD(GtkAccelMap, load_fd)
{
	long fd;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &fd))
		return;

    gtk_accel_map_load_fd((gint)fd);

}


static PHP_METHOD(GtkAccelMap, lock_path)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
		return;

    gtk_accel_map_lock_path(accel_path);
	if (free_accel_path) g_free(accel_path);

}

#line 751 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAccelMap, lookup_entry)
{
    char *accel_path;
    zend_bool free_accel_path;
    GtkAccelKey accel_key;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
        return;

    if (gtk_accel_map_lookup_entry((const gchar *)accel_path, &accel_key)) {
        php_gtk_build_value(&return_value, "(ii)", accel_key.accel_key,
                            accel_key.accel_mods);
    }
}

#line 11348 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAccelMap, save)
{
	char *file_name;
	zend_bool free_file_name = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &file_name, &free_file_name))
		return;

    gtk_accel_map_save(file_name);
	if (free_file_name) g_free(file_name);

}


static PHP_METHOD(GtkAccelMap, save_fd)
{
	long fd;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &fd))
		return;

    gtk_accel_map_save_fd((gint)fd);

}


static PHP_METHOD(GtkAccelMap, unlock_path)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
		return;

    gtk_accel_map_unlock_path(accel_path);
	if (free_accel_path) g_free(accel_path);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_add_entry, 0)
    ZEND_ARG_INFO(0, accel_path)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_add_filter, 0)
    ZEND_ARG_INFO(0, filter_pattern)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_change_entry, 0)
    ZEND_ARG_INFO(0, accel_path)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
    ZEND_ARG_INFO(0, replace)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_load, 0)
    ZEND_ARG_INFO(0, file_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_load_fd, 0)
    ZEND_ARG_INFO(0, fd)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_lock_path, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_lookup_entry, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_save, 0)
    ZEND_ARG_INFO(0, file_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_save_fd, 0)
    ZEND_ARG_INFO(0, fd)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccelmap_unlock_path, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaccelmap_methods[] = {
	PHP_ME(GtkAccelMap, add_entry,            arginfo_gtk_gtkaccelmap_add_entry, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, add_filter,           arginfo_gtk_gtkaccelmap_add_filter, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, change_entry,         arginfo_gtk_gtkaccelmap_change_entry, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, get,                  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, load,                 arginfo_gtk_gtkaccelmap_load, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, load_fd,              arginfo_gtk_gtkaccelmap_load_fd, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, lock_path,            arginfo_gtk_gtkaccelmap_lock_path, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, lookup_entry,         arginfo_gtk_gtkaccelmap_lookup_entry, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, save,                 arginfo_gtk_gtkaccelmap_save, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, save_fd,              arginfo_gtk_gtkaccelmap_save_fd, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAccelMap, unlock_path,          arginfo_gtk_gtkaccelmap_unlock_path, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAccessible, connect_widget_destroyed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_accessible_connect_widget_destroyed(GTK_ACCESSIBLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAccessible, set_widget)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_accessible_set_widget(GTK_ACCESSIBLE(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkAccessible, get_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accessible_get_widget(GTK_ACCESSIBLE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccessible_set_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaccessible_methods[] = {
	PHP_ME(GtkAccessible, connect_widget_destroyed, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccessible, get_widget,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccessible, set_widget,           arginfo_gtk_gtkaccessible_set_widget, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAction, __construct)
{
	GParameter params[4];
	zval *php_args[5] = { NULL, };
	char *prop_names[] = { "name", "label", "tooltip", "stock_id",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u^u^u^u", &php_args[0], &php_args[1], &php_args[2], &php_args[3], &php_args[4], &php_args[5], &php_args[6], &php_args[7])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAction);
	}

	memset(&params, 0, 4 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAction);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAction);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkAction, activate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_action_activate(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAction, block_activate_from)
{
	zval *proxy;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &proxy, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, GtkActivatables are now responsible for activating the action directly" TSRMLS_CC);

    gtk_action_block_activate_from(GTK_ACTION(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(proxy)));

}


static PHP_METHOD(GtkAction, connect_accelerator)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_action_connect_accelerator(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAction, connect_proxy)
{
	zval *proxy;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &proxy, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, Use GtkActivatable::set_related_action() instead" TSRMLS_CC);

    gtk_action_connect_proxy(GTK_ACTION(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(proxy)));

}


static PHP_METHOD(GtkAction, create_icon)
{
	GtkIconSize icon_size;
	zval *php_icon_size = NULL;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_size))
		return;

	if (php_icon_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_icon_size, (gint *)&icon_size) == FAILURE) {
		return;
	}

    php_retval = gtk_action_create_icon(GTK_ACTION(PHPG_GOBJECT(this_ptr)), icon_size);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAction, create_menu_item)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_create_menu_item(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAction, create_tool_item)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_create_tool_item(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAction, disconnect_accelerator)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_action_disconnect_accelerator(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAction, disconnect_proxy)
{
	zval *proxy;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &proxy, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, Use GtkActivatable::set_related_action() instead" TSRMLS_CC);

    gtk_action_disconnect_proxy(GTK_ACTION(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(proxy)));

}


static PHP_METHOD(GtkAction, get_accel_path)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_accel_path(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_name(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 891 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAction, get_proxies)
{
    GSList *widgets, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    widgets = gtk_action_get_proxies(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = widgets; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}
#line 11760 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAction, get_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_sensitive(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_visible(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, is_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_is_sensitive(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, is_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_is_visible(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, set_accel_group)
{
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_group, gtkaccelgroup_ce))
		return;

    gtk_action_set_accel_group(GTK_ACTION(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));

}


static PHP_METHOD(GtkAction, set_accel_path)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
		return;

    gtk_action_set_accel_path(GTK_ACTION(PHPG_GOBJECT(this_ptr)), accel_path);
	if (free_accel_path) g_free(accel_path);

}


static PHP_METHOD(GtkAction, set_sensitive)
{
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &sensitive))
		return;

    gtk_action_set_sensitive(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)sensitive);

}


static PHP_METHOD(GtkAction, set_visible)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_action_set_visible(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


static PHP_METHOD(GtkAction, unblock_activate_from)
{
	zval *proxy;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &proxy, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, GtkActivatables are now responsible for activating the action directly" TSRMLS_CC);

    gtk_action_unblock_activate_from(GTK_ACTION(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(proxy)));

}


static PHP_METHOD(GtkAction, create_menu)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_create_menu(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAction, block_activate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_action_block_activate(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAction, unblock_activate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_action_unblock_activate(GTK_ACTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAction, set_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_action_set_icon_name(GTK_ACTION(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkAction, get_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_icon_name(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, set_visible_horizontal)
{
	zend_bool visible_horizontal;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible_horizontal))
		return;

    gtk_action_set_visible_horizontal(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)visible_horizontal);

}


static PHP_METHOD(GtkAction, get_visible_horizontal)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_visible_horizontal(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, set_visible_vertical)
{
	zend_bool visible_vertical;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible_vertical))
		return;

    gtk_action_set_visible_vertical(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)visible_vertical);

}


static PHP_METHOD(GtkAction, get_visible_vertical)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_visible_vertical(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, set_is_important)
{
	zend_bool is_important;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_important))
		return;

    gtk_action_set_is_important(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)is_important);

}


static PHP_METHOD(GtkAction, get_is_important)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_is_important(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_action_set_label(GTK_ACTION(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkAction, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_label(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, set_short_label)
{
	char *short_label;
	zend_bool free_short_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &short_label, &free_short_label))
		return;

    gtk_action_set_short_label(GTK_ACTION(PHPG_GOBJECT(this_ptr)), short_label);
	if (free_short_label) g_free(short_label);

}


static PHP_METHOD(GtkAction, get_short_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_short_label(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, set_tooltip)
{
	char *tooltip;
	zend_bool free_tooltip = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &tooltip, &free_tooltip))
		return;

    gtk_action_set_tooltip(GTK_ACTION(PHPG_GOBJECT(this_ptr)), tooltip);
	if (free_tooltip) g_free(tooltip);

}


static PHP_METHOD(GtkAction, get_tooltip)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_tooltip(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, set_stock_id)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    gtk_action_set_stock_id(GTK_ACTION(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkAction, get_stock_id)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_stock_id(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAction, get_always_show_image)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_get_always_show_image(GTK_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAction, set_always_show_image)
{
	zend_bool always_show;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &always_show))
		return;

    gtk_action_set_always_show_image(GTK_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)always_show);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_gtk_action_new, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, tooltip)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_block_activate_from, 0)
    ZEND_ARG_OBJ_INFO(0, proxy, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_connect_proxy, 0)
    ZEND_ARG_OBJ_INFO(0, proxy, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_create_icon, 0)
    ZEND_ARG_INFO(0, icon_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_disconnect_proxy, 0)
    ZEND_ARG_OBJ_INFO(0, proxy, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_accel_group, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_accel_path, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_sensitive, 0)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_unblock_activate_from, 0)
    ZEND_ARG_OBJ_INFO(0, proxy, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_visible_horizontal, 0)
    ZEND_ARG_INFO(0, visible_horizontal)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_visible_vertical, 0)
    ZEND_ARG_INFO(0, visible_vertical)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_is_important, 0)
    ZEND_ARG_INFO(0, is_important)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_short_label, 0)
    ZEND_ARG_INFO(0, short_label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_tooltip, 0)
    ZEND_ARG_INFO(0, tooltip)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_stock_id, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaction_set_always_show_image, 0)
    ZEND_ARG_INFO(0, always_show)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaction_methods[] = {
	PHP_ME(GtkAction, __construct,          arginfo_gtk_gtkaction_gtk_action_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, activate,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, block_activate,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, block_activate_from,  arginfo_gtk_gtkaction_block_activate_from, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, connect_accelerator,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, connect_proxy,        arginfo_gtk_gtkaction_connect_proxy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, create_icon,          arginfo_gtk_gtkaction_create_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, create_menu,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, create_menu_item,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, create_tool_item,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, disconnect_accelerator, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, disconnect_proxy,     arginfo_gtk_gtkaction_disconnect_proxy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_accel_path,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_always_show_image, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_is_important,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_proxies,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_sensitive,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_short_label,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_stock_id,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_tooltip,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_visible_horizontal, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, get_visible_vertical, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, is_sensitive,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, is_visible,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_accel_group,      arginfo_gtk_gtkaction_set_accel_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_accel_path,       arginfo_gtk_gtkaction_set_accel_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_always_show_image, arginfo_gtk_gtkaction_set_always_show_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_icon_name,        arginfo_gtk_gtkaction_set_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_is_important,     arginfo_gtk_gtkaction_set_is_important, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_label,            arginfo_gtk_gtkaction_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_sensitive,        arginfo_gtk_gtkaction_set_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_short_label,      arginfo_gtk_gtkaction_set_short_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_stock_id,         arginfo_gtk_gtkaction_set_stock_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_tooltip,          arginfo_gtk_gtkaction_set_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_visible,          arginfo_gtk_gtkaction_set_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_visible_horizontal, arginfo_gtk_gtkaction_set_visible_horizontal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, set_visible_vertical, arginfo_gtk_gtkaction_set_visible_vertical, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, unblock_activate,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAction, unblock_activate_from, arginfo_gtk_gtkaction_unblock_activate_from, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkActionGroup, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "name",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkActionGroup);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkActionGroup);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkActionGroup);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkActionGroup, add_action)
{
	zval *action;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &action, gtkaction_ce))
		return;

    gtk_action_group_add_action(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), GTK_ACTION(PHPG_GOBJECT(action)));

}


static PHP_METHOD(GtkActionGroup, add_action_with_accel)
{
	zval *action;
	char *accelerator;
	zend_bool free_accelerator = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &action, gtkaction_ce, &accelerator, &free_accelerator))
		return;

    gtk_action_group_add_action_with_accel(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), GTK_ACTION(PHPG_GOBJECT(action)), accelerator);
	if (free_accelerator) g_free(accelerator);

}


static PHP_METHOD(GtkActionGroup, get_action)
{
	char *action_name;
	zend_bool free_action_name = FALSE;
	GtkAction* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &action_name, &free_action_name))
		return;

    php_retval = gtk_action_group_get_action(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), action_name);
	if (free_action_name) g_free(action_name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkActionGroup, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_group_get_name(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkActionGroup, get_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_group_get_sensitive(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkActionGroup, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_action_group_get_visible(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 916 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkActionGroup, list_actions)
{
    GList *actions, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    actions = gtk_action_group_list_actions(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = actions; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }

}
#line 12571 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkActionGroup, remove_action)
{
	zval *action;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &action, gtkaction_ce))
		return;

    gtk_action_group_remove_action(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), GTK_ACTION(PHPG_GOBJECT(action)));

}


static PHP_METHOD(GtkActionGroup, set_sensitive)
{
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &sensitive))
		return;

    gtk_action_group_set_sensitive(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), (gboolean)sensitive);

}


static PHP_METHOD(GtkActionGroup, set_translation_domain)
{
	char *domain;
	zend_bool free_domain = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &domain, &free_domain))
		return;

    gtk_action_group_set_translation_domain(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), domain);
	if (free_domain) g_free(domain);

}


static PHP_METHOD(GtkActionGroup, set_visible)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_action_group_set_visible(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


static PHP_METHOD(GtkActionGroup, translate_string)
{
	char *string;
	zend_bool free_string = FALSE, free_result;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &string, &free_string))
		return;

    php_retval = gtk_action_group_translate_string(GTK_ACTION_GROUP(PHPG_GOBJECT(this_ptr)), string);
	if (free_string) g_free(string);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_gtk_action_group_new, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_add_action, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_add_action_with_accel, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
    ZEND_ARG_INFO(0, accelerator)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_get_action, 0)
    ZEND_ARG_INFO(0, action_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_remove_action, 0)
    ZEND_ARG_OBJ_INFO(0, action, GtkAction, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_set_sensitive, 0)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_set_translation_domain, 0)
    ZEND_ARG_INFO(0, domain)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkactiongroup_translate_string, 0)
    ZEND_ARG_INFO(0, string)
ZEND_END_ARG_INFO();

static zend_function_entry gtkactiongroup_methods[] = {
	PHP_ME(GtkActionGroup, __construct,          arginfo_gtk_gtkactiongroup_gtk_action_group_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, add_action,           arginfo_gtk_gtkactiongroup_add_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, add_action_with_accel, arginfo_gtk_gtkactiongroup_add_action_with_accel, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, get_action,           arginfo_gtk_gtkactiongroup_get_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, get_sensitive,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, list_actions,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, remove_action,        arginfo_gtk_gtkactiongroup_remove_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, set_sensitive,        arginfo_gtk_gtkactiongroup_set_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, set_translation_domain, arginfo_gtk_gtkactiongroup_set_translation_domain, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, set_visible,          arginfo_gtk_gtkactiongroup_set_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkActionGroup, translate_string,     arginfo_gtk_gtkactiongroup_translate_string, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentAction, __construct)
{
	char *name, *label, *tooltip, *stock_id;
	zend_bool free_name = FALSE, free_label = FALSE, free_tooltip = FALSE, free_stock_id = FALSE;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uuuu", &name, &free_name, &label, &free_label, &tooltip, &free_tooltip, &stock_id, &free_stock_id)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentAction);
	}

	wrapped_obj = (GObject *) gtk_recent_action_new(name, label, tooltip, stock_id);
	if (free_name) g_free(name);
	if (free_label) g_free(label);
	if (free_tooltip) g_free(tooltip);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentAction);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkRecentAction, new_for_manager)
{
	char *name, *label, *tooltip, *stock_id;
	zend_bool free_name = FALSE, free_label = FALSE, free_tooltip = FALSE, free_stock_id = FALSE;
	zval *manager;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uuuuO", &name, &free_name, &label, &free_label, &tooltip, &free_tooltip, &stock_id, &free_stock_id, &manager, gtkrecentmanager_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentAction);
	}

	wrapped_obj = (GObject *) gtk_recent_action_new_for_manager(name, label, tooltip, stock_id, GTK_RECENT_MANAGER(PHPG_GOBJECT(manager)));
	if (free_name) g_free(name);
	if (free_label) g_free(label);
	if (free_tooltip) g_free(tooltip);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentAction);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkRecentAction, get_show_numbers)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_action_get_show_numbers(GTK_RECENT_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentAction, set_show_numbers)
{
	zend_bool show_numbers;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_numbers))
		return;

    gtk_recent_action_set_show_numbers(GTK_RECENT_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)show_numbers);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentaction_gtk_recent_action_new, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, tooltip)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentaction_recent_action_new_for_manager, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, tooltip)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_OBJ_INFO(0, manager, GtkRecentManager, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentaction_set_show_numbers, 0)
    ZEND_ARG_INFO(0, show_numbers)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentaction_methods[] = {
	PHP_ME(GtkRecentAction, __construct,          arginfo_gtk_gtkrecentaction_gtk_recent_action_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentAction, new_for_manager,      arginfo_gtk_gtkrecentaction_recent_action_new_for_manager, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkRecentAction, get_show_numbers,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentAction, set_show_numbers,     arginfo_gtk_gtkrecentaction_set_show_numbers, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkClipboard, __construct)
{
	GdkDisplay *display = gdk_display_get_default();
	zval *php_display = NULL, *php_selection = NULL;
	GdkAtom selection = GDK_SELECTION_CLIPBOARD;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|OV", &php_display, gdkdisplay_ce, &php_selection)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkClipboard);
	}
    if (php_display)
        display = GDK_DISPLAY(PHPG_GOBJECT(php_display));

    if (php_selection) {
        selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
        if (selection == NULL) {
            php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkClipboard);
        }
    }

	wrapped_obj = (GObject *) gtk_clipboard_get_for_display(display, selection);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkClipboard);
	}
	g_object_ref(wrapped_obj);
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkClipboard, get)
{
	GdkAtom selection = GDK_SELECTION_CLIPBOARD;
	zval *php_selection = NULL;
	GtkClipboard* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|V", &php_selection))
		return;

    if (php_selection) {
        selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
        if (selection == NULL) {
            php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    php_retval = gtk_clipboard_get(selection);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkClipboard, get_display)
{
	GdkDisplay* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_get_display(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 280 "ext/gtk+/gtkclipboard.overrides"
typedef struct {
    phpg_cb_data_t *get_cb_data;
    phpg_cb_data_t *clear_cb_data;
} phpg_clipboard_set_info;

static void phpg_clipboard_get_func_marshal(GtkClipboard *clipboard,
                                            GtkSelectionData *selection_data,
                                            guint info,
                                            gpointer user_data)
{
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_clipboard = NULL;
    zval *php_selection_data = NULL;
    zval *php_info = NULL;
    phpg_clipboard_set_info *set_info = (phpg_clipboard_set_info *) user_data;

    TSRMLS_FETCH();

    if (!zend_is_callable(set_info->get_cb_data->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, set_info->get_cb_data->src_filename, set_info->get_cb_data->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_clipboard, (GObject*)clipboard TSRMLS_CC);
    phpg_gboxed_new(&php_selection_data, GTK_TYPE_SELECTION_DATA, selection_data, FALSE, FALSE TSRMLS_CC);
    MAKE_STD_ZVAL(php_info);
    ZVAL_LONG(php_info, info);

    args = php_gtk_hash_as_array_offset(set_info->get_cb_data->user_args, 3, &n_args);
    args[0] = &php_clipboard;
    args[1] = &php_selection_data;
    args[2] = &php_info;

    call_user_function_ex(EG(function_table), NULL, set_info->get_cb_data->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_clipboard);
    zval_ptr_dtor(&php_selection_data);
    zval_ptr_dtor(&php_info);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static void phpg_clipboard_clear_func_marshal(GtkClipboard *clipboard,
                                              gpointer user_data)
{
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_clipboard = NULL;
    phpg_clipboard_set_info *set_info = (phpg_clipboard_set_info *) user_data;

    TSRMLS_FETCH();

    if (!zend_is_callable(set_info->clear_cb_data->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, set_info->clear_cb_data->src_filename, set_info->clear_cb_data->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_clipboard, (GObject*)clipboard TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(set_info->clear_cb_data->user_args, 1, &n_args);
    args[0] = &php_clipboard;

    call_user_function_ex(EG(function_table), NULL, set_info->clear_cb_data->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_clipboard);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    phpg_cb_data_destroy(set_info->get_cb_data);
    phpg_cb_data_destroy(set_info->clear_cb_data);
    efree(set_info);
}

static PHP_METHOD(GtkClipboard, set_with_data)
{
    zval *get_cb, *clear_cb, *php_targets, *extra;
    phpg_clipboard_set_info *set_info;
	GtkTargetEntry *entries;
    gboolean ret;
    int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 3, &extra, "aVV", &php_targets, &get_cb, &clear_cb))
		return;

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
    if (!entries) return;

    zval_add_ref(&get_cb);
    zval_add_ref(&clear_cb);
    if (extra) {
        zval_add_ref(&extra);
    }

    set_info = emalloc(sizeof(phpg_clipboard_set_info));
    set_info->get_cb_data  = phpg_cb_data_new(get_cb, extra TSRMLS_CC);
    set_info->clear_cb_data  = phpg_cb_data_new(clear_cb, extra TSRMLS_CC);

    ret = gtk_clipboard_set_with_data(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), entries, n,
                                      phpg_clipboard_get_func_marshal,
                                      phpg_clipboard_clear_func_marshal,
                                      set_info);
    efree(entries);

    RETURN_BOOL(ret);
}

#line 13041 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkClipboard, get_owner)
{
	GObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_get_owner(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkClipboard, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clipboard_clear(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkClipboard, set_text)
{
	char *text;
	zend_bool free_text = FALSE;
	long len = -1;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u|i", &text, &free_text, &len))
		return;

    gtk_clipboard_set_text(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), text, (gint)len);
	if (free_text) g_free(text);

}

#line 15 "ext/gtk+/gtkclipboard.overrides"

static void phpg_clipboard_request_contents_marshal(GtkClipboard *clipboard,
                                                    GtkSelectionData *selection_data,
                                                    gpointer user_data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) user_data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_clipboard = NULL;
    zval *php_sel_data = NULL;
    
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_clipboard, (GObject*)clipboard TSRMLS_CC);
    phpg_gboxed_new(&php_sel_data, GTK_TYPE_SELECTION_DATA, selection_data, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_clipboard;
    args[1] = &php_sel_data;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_clipboard);
    zval_ptr_dtor(&php_sel_data);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
    phpg_cb_data_destroy(cbd);
}

static PHP_METHOD(GtkClipboard, request_contents)
{
    zval *php_target, *callback, *extra;
    phpg_cb_data_t *cbd;
    GdkAtom target;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "VV", &php_target, &callback))
		return;
    
    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    zval_add_ref(&callback);
    cbd = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_clipboard_request_contents(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), target,
                                   phpg_clipboard_request_contents_marshal, cbd);

    phpg_cb_data_destroy(cbd);
    
    RETURN_TRUE;
}

#line 13164 "ext/gtk+/gen_gtk.c"


#line 185 "ext/gtk+/gtkclipboard.overrides"

static void phpg_clipboard_request_text_marshal(GtkClipboard *clipboard,
                                                const gchar *text,
                                                gpointer user_data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) user_data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_clipboard = NULL;
    zval *php_text = NULL;
    gchar *cp_str;
    gsize cp_len;
    zend_bool free_cp_str;
    
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_clipboard, (GObject*)clipboard TSRMLS_CC);
    MAKE_STD_ZVAL(php_text);
    if (text) {
        cp_str = phpg_from_utf8(text, strlen(text), &cp_len, &free_cp_str TSRMLS_CC);
        if (cp_str) {
            ZVAL_STRINGL(php_text, cp_str, cp_len, 1);
            if (free_cp_str) {
                g_free(cp_str);
            }
        } else {
            ZVAL_NULL(php_text);
        }
    } else {
        ZVAL_NULL(php_text);
    }

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_clipboard;
    args[1] = &php_text;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_clipboard);
    zval_ptr_dtor(&php_text);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
    phpg_cb_data_destroy(cbd);
}

static PHP_METHOD(GtkClipboard, request_text)
{
    zval *callback, *extra;
    phpg_cb_data_t *cbd;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &callback))
		return;

    zval_add_ref(&callback);
    cbd = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_clipboard_request_text(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)),
                               phpg_clipboard_request_text_marshal, cbd);
    RETURN_TRUE;
}

#line 13248 "ext/gtk+/gen_gtk.c"


#line 101 "ext/gtk+/gtkclipboard.overrides"

static void phpg_clipboard_request_targets_marshal(GtkClipboard *clipboard,
                                                   GdkAtom *atoms,
                                                   gint n_atoms,
                                                   gpointer user_data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) user_data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_clipboard = NULL;
    zval *php_atoms = NULL;
    gchar *name;
    int i;
    
    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_clipboard, (GObject*)clipboard TSRMLS_CC);
    MAKE_STD_ZVAL(php_atoms);
    array_init(php_atoms);
    for (i = 0; i < n_atoms; i++) {
        name = gdk_atom_name(atoms[i]);
        add_next_index_string(php_atoms, name, 1);
        g_free(name);
    }

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_clipboard;
    args[1] = &php_atoms;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_clipboard);
    zval_ptr_dtor(&php_atoms);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
    phpg_cb_data_destroy(cbd);
}

static PHP_METHOD(GtkClipboard, request_targets)
{
    zval *callback, *extra;
    phpg_cb_data_t *cbd;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &callback))
		return;

    zval_add_ref(&callback);
    cbd = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_clipboard_request_targets(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)),
                                  phpg_clipboard_request_targets_marshal, cbd);
    RETURN_TRUE;
}

#line 13323 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkClipboard, wait_for_contents)
{
	GdkAtom target;
	zval *php_target = NULL;
	GtkSelectionData *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_target))
		return;

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_clipboard_wait_for_contents(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), target);
	phpg_gboxed_new(&return_value, GTK_TYPE_SELECTION_DATA, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkClipboard, wait_for_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_wait_for_text(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkClipboard, wait_is_text_available)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_wait_is_text_available(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 416 "ext/gtk+/gtkclipboard.overrides"
static PHP_METHOD(GtkClipboard, wait_for_targets)
{
    GtkSelectionData *data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    /* have to work around GTK+ bug #139883 */
    data = gtk_clipboard_wait_for_contents(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)),
                                           gdk_atom_intern("TARGETS", FALSE));

    if (data) {
        gint n_targets = 0;
        GdkAtom *targets = NULL;

        if (gtk_selection_data_get_targets(data, &targets, &n_targets)) {
            gchar *name;
            int i;

            array_init(return_value);
            for (i = 0; i < n_targets; i++) {
                name = gdk_atom_name(targets[i]);
                add_next_index_string(return_value, name, 1);
                g_free(name);
            }
            g_free(targets);
        }

        gtk_selection_data_free(data);
    }
}

#line 13427 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkClipboard, wait_for_image)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_wait_for_image(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkClipboard, set_image)
{
	zval *pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce))
		return;

    gtk_clipboard_set_image(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

}


static PHP_METHOD(GtkClipboard, wait_is_image_available)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_wait_is_image_available(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 460 "ext/gtk+/gtkclipboard.overrides"
static PHP_METHOD(GtkClipboard, set_can_store)
{
    zval *php_targets = NULL;
	GtkTargetEntry *entries;
    int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a!", &php_targets))
		return;

    if (php_targets) {
        entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
    } else {
        entries = NULL;
        n = 0;
    }

    gtk_clipboard_set_can_store(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), entries, n);
    efree(entries);
    RETURN_TRUE;
}

#line 13496 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkClipboard, store)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clipboard_store(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkClipboard, wait_is_target_available)
{
	GdkAtom target;
	zval *php_target = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_target))
		return;

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_clipboard_wait_is_target_available(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), target);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkClipboard, wait_is_rich_text_available)
{
	zval *buffer;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &buffer, gtktextbuffer_ce))
		return;

    php_retval = gtk_clipboard_wait_is_rich_text_available(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)), GTK_TEXT_BUFFER(PHPG_GOBJECT(buffer)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkClipboard, wait_is_uris_available)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clipboard_wait_is_uris_available(GTK_CLIPBOARD(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkclipboard_clipboard_get_for_display, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, display, GdkDisplay, 1)
    ZEND_ARG_INFO(0, selection)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkclipboard_get, 0, 0, 0)
    ZEND_ARG_INFO(0, selection)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_set_with_data, 0)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, get_callback)
    ZEND_ARG_INFO(0, clear_callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkclipboard_set_text, 0, 0, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_request_contents, 0)
    ZEND_ARG_INFO(0, target)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_request_text, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_request_targets, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_wait_for_contents, 0)
    ZEND_ARG_INFO(0, target)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_set_image, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_set_can_store, 0)
    ZEND_ARG_INFO(0, targets)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_wait_is_target_available, 0)
    ZEND_ARG_INFO(0, target)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclipboard_wait_is_rich_text_available, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkTextBuffer, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkclipboard_methods[] = {
	PHP_ME(GtkClipboard, __construct,          arginfo_gtk_gtkclipboard_clipboard_get_for_display, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, clear,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, get,                  arginfo_gtk_gtkclipboard_get, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkClipboard, get_display,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, get_owner,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, request_contents,     arginfo_gtk_gtkclipboard_request_contents, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, request_targets,      arginfo_gtk_gtkclipboard_request_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, request_text,         arginfo_gtk_gtkclipboard_request_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, set_can_store,        arginfo_gtk_gtkclipboard_set_can_store, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, set_image,            arginfo_gtk_gtkclipboard_set_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, set_text,             arginfo_gtk_gtkclipboard_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, set_with_data,        arginfo_gtk_gtkclipboard_set_with_data, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, store,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_for_contents,    arginfo_gtk_gtkclipboard_wait_for_contents, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_for_image,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_for_targets,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_for_text,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_is_image_available, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_is_rich_text_available, arginfo_gtk_gtkclipboard_wait_is_rich_text_available, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_is_target_available, arginfo_gtk_gtkclipboard_wait_is_target_available, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_is_text_available, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkClipboard, wait_is_uris_available, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkEntryCompletion, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntryCompletion);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntryCompletion);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkEntryCompletion, complete)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_entry_completion_complete(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEntryCompletion, delete_action)
{
	long index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    gtk_entry_completion_delete_action(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gint)index);

}


static PHP_METHOD(GtkEntryCompletion, get_entry)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_entry(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntryCompletion, get_inline_completion)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_inline_completion(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, get_minimum_key_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_minimum_key_length(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_model(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntryCompletion, get_popup_completion)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_popup_completion(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, get_text_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_text_column(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, insert_action_markup)
{
	long index;
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &index, &markup, &free_markup))
		return;

    gtk_entry_completion_insert_action_markup(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gint)index, markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkEntryCompletion, insert_action_text)
{
	long index;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &index, &text, &free_text))
		return;

    gtk_entry_completion_insert_action_text(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gint)index, text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkEntryCompletion, insert_prefix)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_entry_completion_insert_prefix(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEntryCompletion, set_inline_completion)
{
	zend_bool inline_completion;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &inline_completion))
		return;

    gtk_entry_completion_set_inline_completion(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gboolean)inline_completion);

}

#line 2884 "ext/gtk+/gtk.overrides"
static gboolean phpg_entry_completion_match_func_marshal (GtkEntryCompletion *completion, const gchar *key, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_completion = NULL, *php_iter = NULL, *php_key;
    gboolean matches = FALSE;
    gchar *cp_key;
    gsize cp_len;
    zend_bool free_result;
    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_completion,       (GObject*)completion   TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    cp_key = phpg_from_utf8(key, strlen(key), &cp_len, &free_result TSRMLS_CC);
    if (!cp_key) {
        php_error(E_WARNING, "Could not convert key from UTF-8");
        return FALSE;
    }
    MAKE_STD_ZVAL(php_key);
    ZVAL_STRINGL(php_key, (char*)cp_key, cp_len, 1);
    if (free_result) {
        g_free(cp_key);
    }

    args = php_gtk_hash_as_array_offset(cbd->user_args, 3, &n_args);
    args[0] = &php_completion;
    args[1] = &php_key;
    args[2] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_completion);
    zval_ptr_dtor(&php_key);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        matches = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        matches = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return matches;
}

static PHP_METHOD(GtkEntryCompletion, set_match_func)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkEntryCompletionMatchFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkEntryCompletionMatchFunc)phpg_entry_completion_match_func_marshal;
        }
    }

    gtk_entry_completion_set_match_func(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}


#line 13940 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEntryCompletion, set_minimum_key_length)
{
	long length;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &length))
		return;

    gtk_entry_completion_set_minimum_key_length(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gint)length);

}


static PHP_METHOD(GtkEntryCompletion, set_model)
{
	zval *model;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &model, gtktreemodel_ce))
		return;

    gtk_entry_completion_set_model(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), GTK_TREE_MODEL(PHPG_GOBJECT(model)));

}


static PHP_METHOD(GtkEntryCompletion, set_popup_completion)
{
	zend_bool popup_completion;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &popup_completion))
		return;

    gtk_entry_completion_set_popup_completion(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gboolean)popup_completion);

}


static PHP_METHOD(GtkEntryCompletion, set_text_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_entry_completion_set_text_column(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkEntryCompletion, set_popup_set_width)
{
	zend_bool popup_set_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &popup_set_width))
		return;

    gtk_entry_completion_set_popup_set_width(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gboolean)popup_set_width);

}


static PHP_METHOD(GtkEntryCompletion, get_popup_set_width)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_popup_set_width(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, set_popup_single_match)
{
	zend_bool popup_single_match;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &popup_single_match))
		return;

    gtk_entry_completion_set_popup_single_match(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gboolean)popup_single_match);

}


static PHP_METHOD(GtkEntryCompletion, get_popup_single_match)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_popup_single_match(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntryCompletion, get_completion_prefix)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_completion_prefix(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkEntryCompletion, set_inline_selection)
{
	zend_bool inline_selection;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &inline_selection))
		return;

    gtk_entry_completion_set_inline_selection(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)), (gboolean)inline_selection);

}


static PHP_METHOD(GtkEntryCompletion, get_inline_selection)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_completion_get_inline_selection(GTK_ENTRY_COMPLETION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_delete_action, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_insert_action_markup, 0)
    ZEND_ARG_INFO(0, index)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_insert_action_text, 0)
    ZEND_ARG_INFO(0, index)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_inline_completion, 0)
    ZEND_ARG_INFO(0, inline_completion)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_match_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_minimum_key_length, 0)
    ZEND_ARG_INFO(0, length)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_model, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_popup_completion, 0)
    ZEND_ARG_INFO(0, popup_completion)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_text_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_popup_set_width, 0)
    ZEND_ARG_INFO(0, popup_set_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_popup_single_match, 0)
    ZEND_ARG_INFO(0, popup_single_match)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentrycompletion_set_inline_selection, 0)
    ZEND_ARG_INFO(0, inline_selection)
ZEND_END_ARG_INFO();

static zend_function_entry gtkentrycompletion_methods[] = {
	PHP_ME(GtkEntryCompletion, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, complete,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, delete_action,        arginfo_gtk_gtkentrycompletion_delete_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_completion_prefix, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_entry,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_inline_completion, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_inline_selection, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_minimum_key_length, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_popup_completion, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_popup_set_width,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_popup_single_match, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, get_text_column,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, insert_action_markup, arginfo_gtk_gtkentrycompletion_insert_action_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, insert_action_text,   arginfo_gtk_gtkentrycompletion_insert_action_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, insert_prefix,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_inline_completion, arginfo_gtk_gtkentrycompletion_set_inline_completion, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_inline_selection, arginfo_gtk_gtkentrycompletion_set_inline_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_match_func,       arginfo_gtk_gtkentrycompletion_set_match_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_minimum_key_length, arginfo_gtk_gtkentrycompletion_set_minimum_key_length, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_model,            arginfo_gtk_gtkentrycompletion_set_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_popup_completion, arginfo_gtk_gtkentrycompletion_set_popup_completion, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_popup_set_width,  arginfo_gtk_gtkentrycompletion_set_popup_set_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_popup_single_match, arginfo_gtk_gtkentrycompletion_set_popup_single_match, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntryCompletion, set_text_column,      arginfo_gtk_gtkentrycompletion_set_text_column, ZEND_ACC_PUBLIC)

	/***   GtkCellLayout interface implementations   ***/

	PHP_MALIAS(GtkCellLayout, add_attribute, add_attribute, arginfo_gtk_gtkcelllayout_add_attribute, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear, clear, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear_attributes, clear_attributes, arginfo_gtk_gtkcelllayout_clear_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_end, pack_end, arginfo_gtk_gtkcelllayout_pack_end, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_start, pack_start, arginfo_gtk_gtkcelllayout_pack_start, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, reorder, reorder, arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_attributes, set_attributes, arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_cell_data_func, set_cell_data_func, arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, get_cells, get_cells, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIconFactory, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconFactory);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconFactory);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkIconFactory, add)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSet *icon_set = NULL;
	zval *php_icon_set;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uO", &stock_id, &free_stock_id, &php_icon_set, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_icon_set, GTK_TYPE_ICON_SET, FALSE TSRMLS_CC)) {
        icon_set = (GtkIconSet *) PHPG_GBOXED(php_icon_set);
    } else {
        php_error(E_WARNING, "%s::%s() expects icon_set argument to be a valid GtkIconSet object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_factory_add(GTK_ICON_FACTORY(PHPG_GOBJECT(this_ptr)), stock_id, icon_set);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkIconFactory, add_default)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_factory_add_default(GTK_ICON_FACTORY(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIconFactory, lookup)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSet *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    php_retval = gtk_icon_factory_lookup(GTK_ICON_FACTORY(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SET, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkIconFactory, lookup_default)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSet *php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    php_retval = gtk_icon_factory_lookup_default(stock_id);
	if (free_stock_id) g_free(stock_id);
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SET, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkIconFactory, remove_default)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_factory_remove_default(GTK_ICON_FACTORY(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconfactory_add, 0)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_OBJ_INFO(0, icon_set, GtkIconSet, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconfactory_lookup, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconfactory_lookup_default, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

static zend_function_entry gtkiconfactory_methods[] = {
	PHP_ME(GtkIconFactory, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconFactory, add,                  arginfo_gtk_gtkiconfactory_add, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconFactory, add_default,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconFactory, lookup,               arginfo_gtk_gtkiconfactory_lookup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconFactory, lookup_default,       arginfo_gtk_gtkiconfactory_lookup_default, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkIconFactory, remove_default,       NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIconTheme, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconTheme);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconTheme);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkIconTheme, add_builtin_icon)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	long size;
	zval *pixbuf;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiO", &icon_name, &free_icon_name, &size, &pixbuf, gdkpixbuf_ce))
		return;

    gtk_icon_theme_add_builtin_icon(icon_name, (gint)size, GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkIconTheme, append_search_path)
{
	char *path;
	zend_bool free_path = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    gtk_icon_theme_append_search_path(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);

}


static PHP_METHOD(GtkIconTheme, get_default)
{
	GtkIconTheme* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_theme_get_default();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconTheme, get_example_icon_name)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_theme_get_example_icon_name(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkIconTheme, get_for_screen)
{
	zval *screen;
	GtkIconTheme* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    php_retval = gtk_icon_theme_get_for_screen(GDK_SCREEN(PHPG_GOBJECT(screen)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 3727 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconTheme, get_search_path)
{
    gchar **paths;
    gint n_items = 0;
    int i;
    
    NOT_STATIC_METHOD();
    
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;
    
    gtk_icon_theme_get_search_path(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)),
				   &paths, &n_items);

    array_init(return_value);
    for (i = 0; i < n_items; i++)
    {
		gchar *string;
		gsize len;
		zend_bool free_result;
        string = phpg_from_utf8(paths[i], strlen(paths[i]), &len, &free_result TSRMLS_CC);
        if (string) {
            add_next_index_stringl(return_value,(char *)string, len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(string);
    }

    g_strfreev(paths);
}

#line 14492 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconTheme, has_icon)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    php_retval = gtk_icon_theme_has_icon(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);
	RETVAL_BOOL(php_retval);
}

#line 3762 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconTheme, list_icons)
{
    GList *icons, *item;
    gchar *context = NULL;
    zend_bool free_context = FALSE;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|u", &context, &free_context))
        return;

    icons = gtk_icon_theme_list_icons(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), context);

    array_init(return_value);
    for (item = icons; item; item = item->next) {
          add_next_index_string(return_value, item->data, 1);
          g_free(item->data);
    }
    g_list_free(icons);

    if (free_context) g_free(context);
}

#line 14536 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconTheme, load_icon)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	long size;
	GtkIconLookupFlags flags;
	zval *php_flags = NULL;
	GError *error = NULL;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiV", &icon_name, &free_icon_name, &size, &php_flags))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_ICON_LOOKUP_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    php_retval = gtk_icon_theme_load_icon(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), icon_name, (gint)size, flags, &error);
	if (free_icon_name) g_free(icon_name);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconTheme, lookup_icon)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	long size;
	GtkIconLookupFlags flags;
	zval *php_flags = NULL;
	GtkIconInfo *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiV", &icon_name, &free_icon_name, &size, &php_flags))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_ICON_LOOKUP_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    php_retval = gtk_icon_theme_lookup_icon(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), icon_name, (gint)size, flags);
	if (free_icon_name) g_free(icon_name);
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_INFO, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkIconTheme, prepend_search_path)
{
	char *path;
	zend_bool free_path = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    gtk_icon_theme_prepend_search_path(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);

}


static PHP_METHOD(GtkIconTheme, rescan_if_needed)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_theme_rescan_if_needed(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconTheme, set_custom_theme)
{
	char *theme_name;
	zend_bool free_theme_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &theme_name, &free_theme_name))
		return;

    gtk_icon_theme_set_custom_theme(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), theme_name);
	if (free_theme_name) g_free(theme_name);

}


static PHP_METHOD(GtkIconTheme, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_icon_theme_set_screen(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}

#line 3679 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconTheme, set_search_path)
{
	zval *php_paths, **path;
	gchar **paths;
	int i;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_paths))
        return;
        
    paths = g_new(gchar *, zend_hash_num_elements(Z_ARRVAL_P(php_paths)));

	for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_paths)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_paths), (void **)&path) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_paths)), i++) {

         gchar *utf8 = NULL;
         gsize utf8_len = 0;
         zend_bool free_utf8 = 0;
         
         convert_to_string_ex(path);
         utf8 = phpg_to_utf8(Z_STRVAL_PP(path), Z_STRLEN_PP(path), &utf8_len, &free_utf8 TSRMLS_CC);

        if (!utf8) {
            g_free(paths);
            php_error(E_WARNING, "%s::%s() Could not convert path string to UTF-8",
                    get_active_class_name(NULL TSRMLS_CC),
                    get_active_function_name(TSRMLS_C));
            return;
        }

        if (free_utf8) {
            paths[i] = utf8;
        } else {
            /* Use GTK+ memory function here since gtk will own this */
            paths[i] = g_strdup(utf8);
        }
   }
        
	gtk_icon_theme_set_search_path(GTK_ICON_THEME(PHPG_GOBJECT(this_ptr)),
				   (const gchar **)paths, i);

    g_free(paths);
}

#line 14699 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_add_builtin_icon, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_append_search_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_get_for_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_has_icon, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_load_icon, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_lookup_icon, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_prepend_search_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_set_custom_theme, 0)
    ZEND_ARG_INFO(0, theme_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkicontheme_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkicontheme_set_search_path, 0, 0, 1)
    ZEND_ARG_ARRAY_INFO(0, search_paths, 0)
ZEND_END_ARG_INFO();


static zend_function_entry gtkicontheme_methods[] = {
	PHP_ME(GtkIconTheme, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, add_builtin_icon,     arginfo_gtk_gtkicontheme_add_builtin_icon, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkIconTheme, append_search_path,   arginfo_gtk_gtkicontheme_append_search_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, get_default,          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkIconTheme, get_example_icon_name, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, get_for_screen,       arginfo_gtk_gtkicontheme_get_for_screen, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkIconTheme, get_search_path,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, has_icon,             arginfo_gtk_gtkicontheme_has_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, list_icons,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, load_icon,            arginfo_gtk_gtkicontheme_load_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, lookup_icon,          arginfo_gtk_gtkicontheme_lookup_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, prepend_search_path,  arginfo_gtk_gtkicontheme_prepend_search_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, rescan_if_needed,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, set_custom_theme,     arginfo_gtk_gtkicontheme_set_custom_theme, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, set_screen,           arginfo_gtk_gtkicontheme_set_screen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconTheme, set_search_path,      arginfo_gtk_gtkicontheme_set_search_path, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 274 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, __construct)
{
    zval ***args;
    GType *column_types;
    GObject *wrapped_obj;
    int i, argc = ZEND_NUM_ARGS();

    if (argc > 0) {
        args = php_gtk_func_args(argc);
        column_types = emalloc(argc * sizeof(GType));

        for (i = 0; i < argc; i++) {
            column_types[i] = phpg_gtype_from_zval(*args[i] TSRMLS_CC);
            if (column_types[i] == 0) {
                efree(column_types);
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkListStore);
            }
        }

        wrapped_obj = g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);
        gtk_list_store_set_column_types(GTK_LIST_STORE(wrapped_obj), argc, column_types);

        efree(column_types);
        efree(args);
    } else {
        wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkListStore);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 14815 "ext/gtk+/gen_gtk.c"


#line 732 "ext/gtk+/gtktreeview.overrides"
static zval* phpg_gtkliststore_read_dimension_handler(zval *object, zval *offset, int type TSRMLS_DC)
{
    return phpg_gtktreemodel_read_dimension_handler(object, offset, type TSRMLS_CC);
}

#line 14824 "ext/gtk+/gen_gtk.c"


#line 739 "ext/gtk+/gtktreeview.overrides"
static void phpg_gtkliststore_write_dimension_handler(zval *object, zval *offset, zval *value TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, value TSRMLS_CC);
}

#line 14833 "ext/gtk+/gen_gtk.c"


#line 725 "ext/gtk+/gtktreeview.overrides"
static int phpg_gtkliststore_has_dimension_handler(zval *object, zval *offset, int check_empty TSRMLS_DC)
{
    return phpg_gtktreemodel_has_dimension_handler(object, offset, check_empty TSRMLS_CC);
}

#line 14842 "ext/gtk+/gen_gtk.c"


#line 746 "ext/gtk+/gtktreeview.overrides"
static void phpg_gtkliststore_unset_dimension_handler(zval *object, zval *offset TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, NULL TSRMLS_CC);
}

#line 14851 "ext/gtk+/gen_gtk.c"


#line 718 "ext/gtk+/gtktreeview.overrides"
static int phpg_gtkliststore_count_elements_handler(zval *object, long *count TSRMLS_DC)
{
    return phpg_gtktreemodel_count_elements_handler(object, count TSRMLS_CC);
}

#line 14860 "ext/gtk+/gen_gtk.c"



static zend_object_handlers gtkliststore_object_handlers;

static zend_object_value phpg_create_gtkliststore(zend_class_entry *ce TSRMLS_DC)
{
    zend_object_value zov;

    zov = phpg_create_gobject(ce TSRMLS_CC);
    zov.handlers = &gtkliststore_object_handlers;

    return zov;
}
#line 357 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, append)
{
    GtkTreeIter iter;
    zval *items = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|a", &items))
        return;

    gtk_list_store_append(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), &iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 14894 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkListStore, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_list_store_clear(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)));

}

#line 413 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, insert)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    gint position;
    zval *items = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i|a", &position, &items))
        return;

    if (position < 0) {
        php_error(E_WARNING, "%s::%s() requires argument 1 to be greater than zero, %d given",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C), position);
        return;
    }

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));

    if (items) {
        gint n_cols, i;
        gint *columns;
        GValue *values;
        zval **item;

        n_cols = gtk_tree_model_get_n_columns(model);
        if (zend_hash_num_elements(Z_ARRVAL_P(items)) != n_cols) {
            php_error(E_WARNING, "Cannot set row: number of row elements does not match the model");
            return;
        }

        columns = ecalloc(n_cols, sizeof(gint));
        values  = ecalloc(n_cols, sizeof(GValue));

        for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(items)), i = 0;
             zend_hash_get_current_data(Z_ARRVAL_P(items), (void **)&item) == SUCCESS;
             zend_hash_move_forward(Z_ARRVAL_P(items)), i++) {

            g_value_init(&values[i], gtk_tree_model_get_column_type(model, i));

            if (phpg_gvalue_from_zval(&values[i], item, TRUE TSRMLS_CC) == FAILURE) {
                php_error(E_WARNING, "Cannot set row: type of element %d does not match the model", i);
                for ( ; i >= 0; i--) {
                    g_value_unset(&values[i]);
                }
                efree(columns);
                efree(values);
                return;
            }
            columns[i] = i;
        }

        gtk_list_store_insert_with_valuesv(GTK_LIST_STORE(model), &iter, position,
                                           columns, values, n_cols);

        for (i = 0; i < n_cols; i++) {
            g_value_unset(&values[i]);
        }
        efree(columns);
        efree(values);
    } else {
        gtk_list_store_insert(GTK_LIST_STORE(model), &iter, position);
    }

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 14979 "ext/gtk+/gen_gtk.c"


#line 483 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, insert_after)
{
    GtkTreeIter iter;
    GtkTreeIter *sibling = NULL;
    GtkTreeModel *model;
    zval *items = NULL;
    zval *php_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|a", &php_iter, gtktreeiter_ce, &items))
        return;

    sibling = (GtkTreeIter *) PHPG_GBOXED(php_iter);

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));

    gtk_list_store_insert_after(GTK_LIST_STORE(model), &iter, sibling);

    if (items) {
        gint n_cols, i;
        GValue value = { 0 };
        zval **item;

        n_cols = gtk_tree_model_get_n_columns(model);
        if (zend_hash_num_elements(Z_ARRVAL_P(items)) != n_cols) {
            php_error(E_WARNING, "Cannot set row: number of row elements does not match the model");
            return;
        }

        for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(items)), i = 0;
             zend_hash_get_current_data(Z_ARRVAL_P(items), (void **)&item) == SUCCESS;
             zend_hash_move_forward(Z_ARRVAL_P(items)), i++) {

            g_value_init(&value, gtk_tree_model_get_column_type(model, i));

            if (phpg_gvalue_from_zval(&value, item, TRUE TSRMLS_CC) == FAILURE) {
                php_error(E_WARNING, "Cannot set row: type of element %d does not match the model", i);
                g_value_unset(&value);
                return;
            }
            gtk_list_store_set_value(GTK_LIST_STORE(model), &iter, i, &value);
        }
        g_value_unset(&value);
    }

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 15032 "ext/gtk+/gen_gtk.c"


#line 534 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, insert_before)
{
    GtkTreeIter iter;
    GtkTreeIter *sibling = NULL;
    GtkTreeModel *model;
    zval *items = NULL;
    zval *php_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|a", &php_iter, gtktreeiter_ce, &items))
        return;

    sibling = (GtkTreeIter *) PHPG_GBOXED(php_iter);

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));

    gtk_list_store_insert_before(GTK_LIST_STORE(model), &iter, sibling);

    if (items) {
        gint n_cols, i;
        GValue value = { 0 };
        zval **item;

        n_cols = gtk_tree_model_get_n_columns(model);
        if (zend_hash_num_elements(Z_ARRVAL_P(items)) != n_cols) {
            php_error(E_WARNING, "Cannot set row: number of row elements does not match the model");
            return;
        }

        for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(items)), i = 0;
             zend_hash_get_current_data(Z_ARRVAL_P(items), (void **)&item) == SUCCESS;
             zend_hash_move_forward(Z_ARRVAL_P(items)), i++) {

            g_value_init(&value, gtk_tree_model_get_column_type(model, i));

            if (phpg_gvalue_from_zval(&value, item, TRUE TSRMLS_CC) == FAILURE) {
                php_error(E_WARNING, "Cannot set row: type of element %d does not match the model", i);
                g_value_unset(&value);
                return;
            }
            gtk_list_store_set_value(GTK_LIST_STORE(model), &iter, i, &value);
        }
        g_value_unset(&value);
    }

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 15085 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkListStore, iter_is_valid)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_list_store_iter_is_valid(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkListStore, move_after)
{
	GtkTreeIter *iter = NULL, *position = NULL;
	zval *php_iter, *php_position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &php_iter, gboxed_ce, &php_position, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_position) != IS_NULL) {
        if (phpg_gboxed_check(php_position, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
            position = (GtkTreeIter *) PHPG_GBOXED(php_position);
        } else {
            php_error(E_WARNING, "%s::%s() expects position argument to be a valid GtkTreeIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_list_store_move_after(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), iter, position);

}


static PHP_METHOD(GtkListStore, move_before)
{
	GtkTreeIter *iter = NULL, *position = NULL;
	zval *php_iter, *php_position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &php_iter, gboxed_ce, &php_position, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_position) != IS_NULL) {
        if (phpg_gboxed_check(php_position, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
            position = (GtkTreeIter *) PHPG_GBOXED(php_position);
        } else {
            php_error(E_WARNING, "%s::%s() expects position argument to be a valid GtkTreeIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_list_store_move_before(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), iter, position);

}

#line 384 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, prepend)
{
    GtkTreeIter iter;
    zval *items = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|a", &items))
        return;

    gtk_list_store_prepend(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), &iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 15195 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkListStore, remove)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_list_store_remove(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}

#line 677 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, reorder)
{
    zval *php_order, **elem;
    GtkListStore *store;
    gint *order;
    int i, n_elems, store_size;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a/", &php_order))
        return;

    store = GTK_LIST_STORE(PHPG_GOBJECT(this_ptr));
    store_size = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(store), NULL);
    n_elems = zend_hash_num_elements(Z_ARRVAL_P(php_order));

    if (n_elems != store_size) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "new order array size (%d) not the same the number of items in the store (%d)", n_elems, store_size);
        return;
    }

    order = safe_emalloc(store_size, sizeof(gint), 0);
    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_order)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_order), (void **) &elem) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_order)), i++) {

        convert_to_long(*elem);
        if (Z_LVAL_PP(elem) < 0 || Z_LVAL_PP(elem) >= store_size) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING, "position index out of range 0-%d", store_size);
            efree(order);
            return;
        }
        order[i] = Z_LVAL_PP(elem);
    }

    gtk_list_store_reorder(store, order);
    efree(order);
}

#line 15262 "ext/gtk+/gen_gtk.c"


#line 596 "ext/gtk+/gtktreeview.overrides"
static void phpg_store_set_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool is_tree)
{
    zval ***items;
    zval *php_iter;
    GtkTreeIter *iter = NULL;
    GtkTreeModel *model;
    int i, argc = ZEND_NUM_ARGS();

    NOT_STATIC_METHOD();

    if (argc < 3) {
        php_error(E_WARNING, "%s::%s() requires at least 3 arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C), argc);
        return;
    }

    if (!php_gtk_parse_args(1, "O", &php_iter, gtktreeiter_ce))
        return;

    iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);

    if ((argc-1) % 2) {
        php_error(E_WARNING, "%s::%s() requires arguments 2-n to be column/value pairs",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    model = GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr));
    items = php_gtk_func_args(argc);
    for (i = 1; i < argc; i += 2) {
        zval *zcolumn = *items[i];
        zval *zvalue = *items[i+1];
        gint column;
        GValue value = { 0 };

        if (Z_TYPE_P(zcolumn) != IS_LONG) {
            php_error(E_WARNING, "%s::%s() requires argument %d to be an integer, %s given",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                      i, zend_zval_type_name(zcolumn));
            efree(items);
            return;
        }

        column = Z_LVAL_P(zcolumn);
        if (column < 0 || column >= gtk_tree_model_get_n_columns(model)) {
            php_error(E_WARNING, "%s::%s(): argument %d is out of range - model has %d column(s)",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                      i, gtk_tree_model_get_n_columns(model));
            efree(items);
            return;
        }

        g_value_init(&value, gtk_tree_model_get_column_type(model, column));

        if (phpg_gvalue_from_zval(&value, &zvalue, TRUE TSRMLS_CC) == FAILURE) {
            php_error(E_WARNING, "%s::%s(): argument %d is of the wrong type for column %d",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                      i+1, column);
            efree(items);
            return;
        }

        if (is_tree) {
            gtk_tree_store_set_value(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter, column, &value);
        } else {
            gtk_list_store_set_value(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), iter, column, &value);
        }

        g_value_unset(&value);
    }

    efree(items);
}

static PHP_METHOD(GtkListStore, set)
{
    phpg_store_set_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);
}

#line 15345 "ext/gtk+/gen_gtk.c"


#line 311 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkListStore, set_column_types)
{
    zval *php_types, **item;
    GType *column_types;
    int i, n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_types)) {
        return;
    }

    n = zend_hash_num_elements(Z_ARRVAL_P(php_types));

    if (n == 0) {
        php_error(E_WARNING, "number of columns has to be > 0");
        return;
    }

    column_types = safe_emalloc(n, sizeof(GType), 0);
    for (i = 0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_types));
         zend_hash_get_current_data(Z_ARRVAL_P(php_types), (void**)&item) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_types)), i++) {

        column_types[i] = phpg_gtype_from_zval(*item TSRMLS_CC);
        if (column_types[i] == 0) {
            efree(column_types);
            php_error(E_WARNING, "could not set column types for GtkListStore");
            return;
        }
    }

    gtk_list_store_set_column_types(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), i, column_types);

    efree(column_types);
}

#line 15386 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkListStore, swap)
{
	GtkTreeIter *a = NULL, *b = NULL;
	zval *php_a, *php_b;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_a, gboxed_ce, &php_b, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_a, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        a = (GtkTreeIter *) PHPG_GBOXED(php_a);
    } else {
        php_error(E_WARNING, "%s::%s() expects a argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_b, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        b = (GtkTreeIter *) PHPG_GBOXED(php_b);
    } else {
        php_error(E_WARNING, "%s::%s() expects b argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_list_store_swap(GTK_LIST_STORE(PHPG_GOBJECT(this_ptr)), a, b);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkliststore_gtk_list_store_new, 0, 0, 0)
    ZEND_ARG_INFO(0, type_col_0)
    ZEND_ARG_INFO(0, type_col_1)
    ZEND_ARG_INFO(0, ...)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkliststore_append, 0, 0, 0)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkliststore_insert, 0, 0, 1)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkliststore_iter_is_valid, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkliststore_move_after, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, position, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkliststore_move_before, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, position, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkliststore_prepend, 0, 0, 0)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkliststore_remove, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkliststore_set, 0, 0, 3)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkliststore_swap, 0)
    ZEND_ARG_OBJ_INFO(0, a, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, b, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkliststore_methods[] = {
	PHP_ME(GtkListStore, __construct,          arginfo_gtk_gtkliststore_gtk_list_store_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, append,               arginfo_gtk_gtkliststore_append, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, clear,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, insert,               arginfo_gtk_gtkliststore_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, insert_after,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, insert_before,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, iter_is_valid,        arginfo_gtk_gtkliststore_iter_is_valid, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, move_after,           arginfo_gtk_gtkliststore_move_after, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, move_before,          arginfo_gtk_gtkliststore_move_before, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, prepend,              arginfo_gtk_gtkliststore_prepend, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, remove,               arginfo_gtk_gtkliststore_remove, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, reorder,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, set,                  arginfo_gtk_gtkliststore_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, set_column_types,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListStore, swap,                 arginfo_gtk_gtkliststore_swap, ZEND_ACC_PUBLIC)

	/***   GtkTreeModel interface implementations   ***/

	PHP_MALIAS(GtkTreeModel, get, get, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_column_type, get_column_type, arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_flags, get_flags, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter, get_iter, arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_first, get_iter_first, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_root, get_iter_root, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_from_string, get_iter_from_string, arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_n_columns, get_n_columns, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_path, get_path, arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_string_from_iter, get_string_from_iter, arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_value, get_value, arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_children, iter_children, arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_has_child, iter_has_child, arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_n_children, iter_n_children, arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_next, iter_next, arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_nth_child, iter_nth_child, arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_parent, iter_parent, arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, ref_node, ref_node, arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_changed, row_changed, arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_deleted, row_deleted, arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_has_child_toggled, row_has_child_toggled, arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_inserted, row_inserted, arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, rows_reordered, rows_reordered, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, unref_node, unref_node, arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, for_each, for_each, arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC)

	/***   GtkTreeDragSource interface implementations   ***/

	PHP_MALIAS(GtkTreeDragSource, drag_data_delete, drag_data_delete, arginfo_gtk_gtktreedragsource_drag_data_delete, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, drag_data_get, drag_data_get, arginfo_gtk_gtktreedragsource_drag_data_get, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, row_draggable, row_draggable, arginfo_gtk_gtktreedragsource_row_draggable, ZEND_ACC_PUBLIC)

	/***   GtkTreeDragDest interface implementations   ***/

	PHP_MALIAS(GtkTreeDragDest, drag_data_received, drag_data_received, arginfo_gtk_gtktreedragdest_drag_data_received, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragDest, row_drop_possible, row_drop_possible, arginfo_gtk_gtktreedragdest_row_drop_possible, ZEND_ACC_PUBLIC)

	/***   GtkTreeSortable interface implementations   ***/

	PHP_MALIAS(GtkTreeSortable, get_sort_column_id, get_sort_column_id, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, has_default_sort_func, has_default_sort_func, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_default_sort_func, set_default_sort_func, arginfo_gtk_gtktreesortable_set_default_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_column_id, set_sort_column_id, arginfo_gtk_gtktreesortable_set_sort_column_id, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_func, set_sort_func, arginfo_gtk_gtktreesortable_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, sort_column_changed, sort_column_changed, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkObject, flags)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = GTK_OBJECT_FLAGS(GTK_OBJECT(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkObject, set_flags)
{
	long flags;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &flags))
		return;

    GTK_OBJECT_SET_FLAGS(GTK_OBJECT(PHPG_GOBJECT(this_ptr)), (guint)flags);

}


static PHP_METHOD(GtkObject, unset_flags)
{
	long flags;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &flags))
		return;

    GTK_OBJECT_UNSET_FLAGS(GTK_OBJECT(PHPG_GOBJECT(this_ptr)), (guint)flags);

}


static PHP_METHOD(GtkObject, destroy)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_object_destroy(GTK_OBJECT(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkobject_set_flags, 0)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkobject_unset_flags, 0)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

static zend_function_entry gtkobject_methods[] = {
	PHP_ME(GtkObject, destroy,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkObject, flags,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkObject, set_flags,            arginfo_gtk_gtkobject_set_flags, ZEND_ACC_PUBLIC)
	PHP_ME(GtkObject, unset_flags,          arginfo_gtk_gtkobject_unset_flags, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkItemFactory, __construct)
{
	GType container_type;
	zval *php_container_type = NULL, *php_accel_group = NULL;
	char *path;
	zend_bool free_path = FALSE;
	GtkAccelGroup *accel_group = NULL;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu|N", &php_container_type, &path, &free_path, &php_accel_group, gtkaccelgroup_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkItemFactory);
	}
    if ((container_type = phpg_gtype_from_zval(php_container_type TSRMLS_CC)) == 0) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkItemFactory);
    }
    if (php_accel_group) {
        if (Z_TYPE_P(php_accel_group) == IS_NULL)
            accel_group = NULL;
        else
            accel_group = GTK_ACCEL_GROUP(PHPG_GOBJECT(php_accel_group));
    }

	phpg_warn_deprecated("use GTKUIManager" TSRMLS_CC);

	wrapped_obj = (GObject *) gtk_item_factory_new(container_type, path, accel_group);
	if (free_path) g_free(path);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkItemFactory);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, path_delete)
{
	char *ifactory_path, *path;
	zend_bool free_ifactory_path = FALSE, free_path = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uu", &ifactory_path, &free_ifactory_path, &path, &free_path))
		return;

    gtk_item_factories_path_delete(ifactory_path, path);
	if (free_ifactory_path) g_free(ifactory_path);
	if (free_path) g_free(path);

}


static PHP_METHOD(GtkItemFactory, add_foreign)
{
	zval *accel_widget, *accel_group, *php_modifiers = NULL;
	char *full_path;
	zend_bool free_full_path = FALSE;
	long keyval;
	GdkModifierType modifiers;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OuOiV", &accel_widget, gtkwidget_ce, &full_path, &free_full_path, &accel_group, gtkaccelgroup_ce, &keyval, &php_modifiers))
		return;

	if (php_modifiers && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_modifiers, (gint *)&modifiers) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkMenuItem::set_accel_path() and GtkWidget::set_accel_path()" TSRMLS_CC);

    gtk_item_factory_add_foreign(GTK_WIDGET(PHPG_GOBJECT(accel_widget)), full_path, GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)), (guint)keyval, modifiers);
	if (free_full_path) g_free(full_path);

}


static PHP_METHOD(GtkItemFactory, construct)
{
	GType container_type;
	zval *php_container_type = NULL, *accel_group;
	char *path;
	zend_bool free_path = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VuO", &php_container_type, &path, &free_path, &accel_group, gtkaccelgroup_ce))
		return;
    if ((container_type = phpg_gtype_from_zval(php_container_type TSRMLS_CC)) == 0) {
        return;
    }

    gtk_item_factory_construct(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), container_type, path, GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));
	if (free_path) g_free(path);

}


static PHP_METHOD(GtkItemFactory, delete_item)
{
	char *path;
	zend_bool free_path = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    gtk_item_factory_delete_item(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);

}


static PHP_METHOD(GtkItemFactory, from_path)
{
	char *path;
	zend_bool free_path = FALSE;
	GtkItemFactory* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    php_retval = gtk_item_factory_from_path(path);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, from_widget)
{
	zval *widget;
	GtkItemFactory* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    php_retval = gtk_item_factory_from_widget(GTK_WIDGET(PHPG_GOBJECT(widget)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, get_item)
{
	char *path;
	zend_bool free_path = FALSE;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    php_retval = gtk_item_factory_get_item(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, get_item_by_action)
{
	long action;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &action))
		return;

    php_retval = gtk_item_factory_get_item_by_action(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), (guint)action);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, get_widget)
{
	char *path;
	zend_bool free_path = FALSE;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    php_retval = gtk_item_factory_get_widget(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, get_widget_by_action)
{
	long action;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &action))
		return;

    php_retval = gtk_item_factory_get_widget_by_action(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), (guint)action);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkItemFactory, path_from_widget)
{
	zval *widget;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    php_retval = gtk_item_factory_path_from_widget(GTK_WIDGET(PHPG_GOBJECT(widget)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkItemFactory, popup)
{
	long x, y, mouse_button, time = GDK_CURRENT_TIME;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iii|i", &x, &y, &mouse_button, &time))
		return;

    gtk_item_factory_popup(GTK_ITEM_FACTORY(PHPG_GOBJECT(this_ptr)), (guint)x, (guint)y, (guint)mouse_button, (guint32)time);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkitemfactory_gtk_item_factory_new, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, container_type, GType, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_path_delete, 0)
    ZEND_ARG_INFO(0, ifactory_path)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_add_foreign, 0)
    ZEND_ARG_OBJ_INFO(0, accel_widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, full_path)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
    ZEND_ARG_INFO(0, keyval)
    ZEND_ARG_INFO(0, modifiers)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_construct, 0)
    ZEND_ARG_OBJ_INFO(0, container_type, GType, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_delete_item, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_from_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_from_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_get_item, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_get_item_by_action, 0)
    ZEND_ARG_INFO(0, action)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_get_widget, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_get_widget_by_action, 0)
    ZEND_ARG_INFO(0, action)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkitemfactory_path_from_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkitemfactory_popup, 0, 0, 3)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, mouse_button)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();

static zend_function_entry gtkitemfactory_methods[] = {
	PHP_ME(GtkItemFactory, __construct,          arginfo_gtk_gtkitemfactory_gtk_item_factory_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, add_foreign,          arginfo_gtk_gtkitemfactory_add_foreign, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkItemFactory, construct,            arginfo_gtk_gtkitemfactory_construct, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, delete_item,          arginfo_gtk_gtkitemfactory_delete_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, from_path,            arginfo_gtk_gtkitemfactory_from_path, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkItemFactory, from_widget,          arginfo_gtk_gtkitemfactory_from_widget, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkItemFactory, get_item,             arginfo_gtk_gtkitemfactory_get_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, get_item_by_action,   arginfo_gtk_gtkitemfactory_get_item_by_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, get_widget,           arginfo_gtk_gtkitemfactory_get_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, get_widget_by_action, arginfo_gtk_gtkitemfactory_get_widget_by_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItemFactory, path_delete,          arginfo_gtk_gtkitemfactory_path_delete, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkItemFactory, path_from_widget,     arginfo_gtk_gtkitemfactory_path_from_widget, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkItemFactory, popup,                arginfo_gtk_gtkitemfactory_popup, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIMContext, delete_surrounding)
{
	long offset, n_chars;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &offset, &n_chars))
		return;

    php_retval = gtk_im_context_delete_surrounding(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), (gint)offset, (gint)n_chars);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIMContext, filter_keypress)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_im_context_filter_keypress(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), (GdkEventKey *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIMContext, focus_in)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_im_context_focus_in(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIMContext, focus_out)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_im_context_focus_out(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIMContext, reset)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_im_context_reset(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIMContext, set_client_window)
{
	zval *window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gdkwindow_ce))
		return;

    gtk_im_context_set_client_window(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)));

}


static PHP_METHOD(GtkIMContext, set_cursor_location)
{
	GdkRectangle area = { 0, 0, 0, 0 };
	zval *php_area;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_area))
		return;

    if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    gtk_im_context_set_cursor_location(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), &area);

}


static PHP_METHOD(GtkIMContext, set_surrounding)
{
	char *text;
	zend_bool free_text = FALSE;
	long len, cursor_index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uii", &text, &free_text, &len, &cursor_index))
		return;

    gtk_im_context_set_surrounding(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), text, (gint)len, (gint)cursor_index);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkIMContext, set_use_preedit)
{
	zend_bool use_preedit;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_preedit))
		return;

    gtk_im_context_set_use_preedit(GTK_IM_CONTEXT(PHPG_GOBJECT(this_ptr)), (gboolean)use_preedit);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_delete_surrounding, 0)
    ZEND_ARG_INFO(0, offset)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_filter_keypress, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_set_client_window, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_set_cursor_location, 0)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_set_surrounding, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
    ZEND_ARG_INFO(0, cursor_index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimcontext_set_use_preedit, 0)
    ZEND_ARG_INFO(0, use_preedit)
ZEND_END_ARG_INFO();

static zend_function_entry gtkimcontext_methods[] = {
	PHP_ME(GtkIMContext, delete_surrounding,   arginfo_gtk_gtkimcontext_delete_surrounding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, filter_keypress,      arginfo_gtk_gtkimcontext_filter_keypress, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, focus_in,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, focus_out,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, reset,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, set_client_window,    arginfo_gtk_gtkimcontext_set_client_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, set_cursor_location,  arginfo_gtk_gtkimcontext_set_cursor_location, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, set_surrounding,      arginfo_gtk_gtkimcontext_set_surrounding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMContext, set_use_preedit,      arginfo_gtk_gtkimcontext_set_use_preedit, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFileFilter, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileFilter);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileFilter);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 2986 "ext/gtk+/gtk.overrides"
static void phpg_file_filter_add_custom_marshal_add_file(const gchar *filename, int flags, int flag, zval *php_filter_info)
{
    gchar *cp = NULL;
    gchar *fn = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;
    TSRMLS_FETCH();

    if (filename && (flags & flag)
    ) {
        fn = g_filename_to_utf8(filename, strlen(filename), NULL, NULL, NULL);
        cp = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_cp TSRMLS_CC);
        if (cp) {
            add_next_index_string(php_filter_info, (char *)cp, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_cp)
            g_free(cp);
    } else {
        add_next_index_null(php_filter_info);
    }
}
static gboolean phpg_file_filter_add_custom_marshal(GtkFileFilterInfo *filter_info, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_filter_info = NULL;
    gboolean show = TRUE;
    gchar *cp = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;

    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    MAKE_STD_ZVAL(php_filter_info);
    array_init(php_filter_info);

    phpg_file_filter_add_custom_marshal_add_file(
        filter_info->filename,
        filter_info->contains,
        GTK_FILE_FILTER_FILENAME,
        php_filter_info
    );
    phpg_file_filter_add_custom_marshal_add_file(
        filter_info->uri,
        filter_info->contains,
        GTK_FILE_FILTER_URI,
        php_filter_info
    );
    phpg_file_filter_add_custom_marshal_add_file(
        filter_info->display_name,
        filter_info->contains,
        GTK_FILE_FILTER_DISPLAY_NAME,
        php_filter_info
    );

    //mime_type
    if (filter_info->mime_type
        && (filter_info->contains & GTK_FILE_FILTER_MIME_TYPE)
    ) {
        cp = phpg_from_utf8(filter_info->mime_type, strlen(filter_info->mime_type), &cp_len, &free_cp TSRMLS_CC);
        if (cp) {
            add_next_index_string(php_filter_info, (char *)cp, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_cp)
            g_free(cp);
    } else {
        add_next_index_null(php_filter_info);
    }

    args = php_gtk_hash_as_array_offset(cbd->user_args, 1, &n_args);
    args[0] = &php_filter_info;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_filter_info);

    if (retval) {
        show = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        show = TRUE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return show;
}

static PHP_METHOD(GtkFileFilter, add_custom)
{
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkFileFilterFlags flags = 0;
    zval *php_flags = NULL;


    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &extra, "VV", &php_flags, &callback))
        return;

    if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_FILE_FILTER_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
        return;
    }

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_file_filter_add_custom(
            GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)),
            flags,
            (GtkFileFilterFunc)phpg_file_filter_add_custom_marshal,
            cb_data, phpg_cb_data_destroy
    );
}


#line 16309 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileFilter, add_mime_type)
{
	char *mime_type;
	zend_bool free_mime_type = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &mime_type, &free_mime_type))
		return;

    gtk_file_filter_add_mime_type(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)), mime_type);
	if (free_mime_type) g_free(mime_type);

}


static PHP_METHOD(GtkFileFilter, add_pattern)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_file_filter_add_pattern(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkFileFilter, add_pixbuf_formats)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_file_filter_add_pixbuf_formats(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)));

}

#line 3128 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileFilter, filter)
{
    GtkFileFilterInfo filter_info = { 0, NULL, NULL, NULL, NULL };
    zval *php_filter_info = NULL, **temp_value;
    gchar **values;
    gboolean result;
    int i = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_filter_info)) {
        return;
    }

    if (zend_hash_num_elements(Z_ARRVAL_P(php_filter_info)) != 4) {
        php_error(E_WARNING, "%s::%s() file info array requires 4 values of type string or NULL",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C));
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_filter_info));
    values = safe_emalloc(4, sizeof(gchar *), 0);

    while (zend_hash_get_current_data(Z_ARRVAL_P(php_filter_info), (void **)&temp_value) == SUCCESS) {
        gchar *utf8 = NULL;
        gsize utf8_len = 0;
        zend_bool free_utf8 = 0;

        if (Z_TYPE_P(*temp_value) == IS_NULL) {
            values[i++] = NULL;
        } else {
            convert_to_string_ex(temp_value);
            utf8 = phpg_to_utf8(Z_STRVAL_PP(temp_value), Z_STRLEN_PP(temp_value), &utf8_len, &free_utf8 TSRMLS_CC);

            if (!utf8) {
                efree(values);
                php_error(E_WARNING, "%s::%s() Could not convert filter info string to UTF-8",
                        get_active_class_name(NULL TSRMLS_CC),
                        get_active_function_name(TSRMLS_C));
                return;
            }

            if (free_utf8) {
                values[i++] = utf8;
            } else {
                /* Use GTK+ memory function here since utf8 may also allocated by it */
                values[i++] = g_strdup(utf8);
            }
        }

        zend_hash_move_forward(Z_ARRVAL_P(php_filter_info));
    }

    if (values[0] != NULL) {
        filter_info.filename     = values[0];
        filter_info.contains     |= GTK_FILE_FILTER_FILENAME;
    }
    if (values[1] != NULL) {
        filter_info.uri          = values[1];
        filter_info.contains     |= GTK_FILE_FILTER_URI;
    }
    if (values[2] != NULL) {
        filter_info.display_name = values[2];
        filter_info.contains     |= GTK_FILE_FILTER_DISPLAY_NAME;
    }
    if (values[3] != NULL) {
        filter_info.mime_type    = values[3];
        filter_info.contains     |= GTK_FILE_FILTER_MIME_TYPE;
    }

    efree(values);

    result = gtk_file_filter_filter(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)), &filter_info);
    RETURN_BOOL(result);
}

#line 16435 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileFilter, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_filter_get_name(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFileFilter, get_needed)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_filter_get_needed(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkFileFilter, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_file_filter_set_name(GTK_FILE_FILTER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilefilter_add_custom, 0)
    ZEND_ARG_INFO(0, flags_needed)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilefilter_add_mime_type, 0)
    ZEND_ARG_INFO(0, mime_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilefilter_add_pattern, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilefilter_filter, 0)
    ZEND_ARG_INFO(0, filterinfo)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilefilter_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfilefilter_methods[] = {
	PHP_ME(GtkFileFilter, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, add_custom,           arginfo_gtk_gtkfilefilter_add_custom, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, add_mime_type,        arginfo_gtk_gtkfilefilter_add_mime_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, add_pattern,          arginfo_gtk_gtkfilefilter_add_pattern, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, add_pixbuf_formats,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, filter,               arginfo_gtk_gtkfilefilter_filter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, get_needed,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileFilter, set_name,             arginfo_gtk_gtkfilefilter_set_name, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIMMulticontext, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIMMulticontext);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIMMulticontext);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkIMMulticontext, append_menuitems)
{
	zval *menushell;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &menushell, gtkmenushell_ce))
		return;

    gtk_im_multicontext_append_menuitems(GTK_IM_MULTICONTEXT(PHPG_GOBJECT(this_ptr)), GTK_MENU_SHELL(PHPG_GOBJECT(menushell)));

}


static PHP_METHOD(GtkIMMulticontext, get_context_id)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_im_multicontext_get_context_id(GTK_IM_MULTICONTEXT(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkIMMulticontext, set_context_id)
{
	char *context_id;
	zend_bool free_context_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &context_id, &free_context_id))
		return;

    gtk_im_multicontext_set_context_id(GTK_IM_MULTICONTEXT(PHPG_GOBJECT(this_ptr)), context_id);
	if (free_context_id) g_free(context_id);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimmulticontext_append_menuitems, 0)
    ZEND_ARG_OBJ_INFO(0, menushell, GtkMenuShell, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimmulticontext_set_context_id, 0)
    ZEND_ARG_INFO(0, context_id)
ZEND_END_ARG_INFO();

static zend_function_entry gtkimmulticontext_methods[] = {
	PHP_ME(GtkIMMulticontext, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMMulticontext, append_menuitems,     arginfo_gtk_gtkimmulticontext_append_menuitems, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMMulticontext, get_context_id,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIMMulticontext, set_context_id,       arginfo_gtk_gtkimmulticontext_set_context_id, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkIMContextSimple, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIMContextSimple);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIMContextSimple);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkimcontextsimple_methods[] = {
	PHP_ME(GtkIMContextSimple, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRenderer, activate)
{
	GdkEvent *event = NULL;
	zval *php_event, *widget, *php_background_area, *php_cell_area, *php_flags = NULL;
	char *path;
	zend_bool free_path = FALSE;
	GdkRectangle background_area = { 0, 0, 0, 0 }, cell_area = { 0, 0, 0, 0 };
	GtkCellRendererState flags;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOuVVV", &php_event, gboxed_ce, &widget, gtkwidget_ce, &path, &free_path, &php_background_area, &php_cell_area, &php_flags))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_rectangle_from_zval(php_background_area, (GdkRectangle*)&background_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects background_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (phpg_rectangle_from_zval(php_cell_area, (GdkRectangle*)&cell_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects cell_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_CELL_RENDERER_STATE, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    php_retval = gtk_cell_renderer_activate(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), event, GTK_WIDGET(PHPG_GOBJECT(widget)), path, &background_area, &cell_area, flags);
	if (free_path) g_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCellRenderer, editing_canceled)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.6, please use GtkCellRenderer::stop_editing()" TSRMLS_CC);

    gtk_cell_renderer_editing_canceled(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)));

}

#line 159 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkCellRenderer, get_fixed_size)
{
    int width, height = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_cell_renderer_get_fixed_size(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), &width, &height);
    php_gtk_build_value(&return_value, "(ii)", width, height);
}

#line 16734 "ext/gtk+/gen_gtk.c"


#line 181 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkCellRenderer, get_size)
{
    int x_offset, y_offset, width, height = 0;
    zval *widget = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
        return;

    gtk_cell_renderer_get_size(
        GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)),
        GTK_WIDGET(PHPG_GOBJECT(widget)),
        NULL,
        &x_offset, &y_offset, &width, &height
    );
    php_gtk_build_value(&return_value, "(iiii)", x_offset, y_offset, width, height);
}

#line 16757 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCellRenderer, render)
{
	zval *window, *widget, *php_background_area, *php_cell_area, *php_expose_area, *php_flags = NULL;
	GdkRectangle background_area = { 0, 0, 0, 0 }, cell_area = { 0, 0, 0, 0 }, expose_area = { 0, 0, 0, 0 };
	GtkCellRendererState flags;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOVVVV", &window, gdkwindow_ce, &widget, gtkwidget_ce, &php_background_area, &php_cell_area, &php_expose_area, &php_flags))
		return;

    if (phpg_rectangle_from_zval(php_background_area, (GdkRectangle*)&background_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects background_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (phpg_rectangle_from_zval(php_cell_area, (GdkRectangle*)&cell_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects cell_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (phpg_rectangle_from_zval(php_expose_area, (GdkRectangle*)&expose_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects expose_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_CELL_RENDERER_STATE, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    gtk_cell_renderer_render(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), GTK_WIDGET(PHPG_GOBJECT(widget)), &background_area, &cell_area, &expose_area, flags);

}


static PHP_METHOD(GtkCellRenderer, set_fixed_size)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_cell_renderer_set_fixed_size(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


static PHP_METHOD(GtkCellRenderer, start_editing)
{
	GdkEvent *event = NULL;
	zval *php_event, *widget, *php_background_area, *php_cell_area, *php_flags = NULL;
	char *path;
	zend_bool free_path = FALSE;
	GdkRectangle background_area = { 0, 0, 0, 0 }, cell_area = { 0, 0, 0, 0 };
	GtkCellRendererState flags;
	GtkCellEditable* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOuVVV", &php_event, gboxed_ce, &widget, gtkwidget_ce, &path, &free_path, &php_background_area, &php_cell_area, &php_flags))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_rectangle_from_zval(php_background_area, (GdkRectangle*)&background_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects background_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (phpg_rectangle_from_zval(php_cell_area, (GdkRectangle*)&cell_area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects cell_area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_CELL_RENDERER_STATE, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    php_retval = gtk_cell_renderer_start_editing(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), event, GTK_WIDGET(PHPG_GOBJECT(widget)), path, &background_area, &cell_area, flags);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCellRenderer, stop_editing)
{
	zend_bool canceled;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &canceled))
		return;

    gtk_cell_renderer_stop_editing(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (gboolean)canceled);

}

#line 7 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkCellRenderer, get_alignment)
{
    gfloat xalign = 0, yalign = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_cell_renderer_get_alignment(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), &xalign, &yalign);

    array_init(return_value);
    add_next_index_double(return_value, xalign);
    add_next_index_double(return_value, yalign);
}

#line 16878 "ext/gtk+/gen_gtk.c"


#line 25 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkCellRenderer, get_padding)
{
    gint xpad = 0, ypad = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_cell_renderer_get_padding(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), &xpad, &ypad);

    array_init(return_value);
    add_next_index_long(return_value, xpad);
    add_next_index_long(return_value, ypad);
}

#line 16898 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCellRenderer, get_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_renderer_get_sensitive(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCellRenderer, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_renderer_get_visible(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCellRenderer, set_alignment)
{
	double xalign, yalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &xalign, &yalign))
		return;

    gtk_cell_renderer_set_alignment(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign);

}


static PHP_METHOD(GtkCellRenderer, set_padding)
{
	long xpad, ypad;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &xpad, &ypad))
		return;

    gtk_cell_renderer_set_padding(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (gint)xpad, (gint)ypad);

}


static PHP_METHOD(GtkCellRenderer, set_sensitive)
{
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &sensitive))
		return;

    gtk_cell_renderer_set_sensitive(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (gboolean)sensitive);

}


static PHP_METHOD(GtkCellRenderer, set_visible)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_cell_renderer_set_visible(GTK_CELL_RENDERER(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_activate, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, background_area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, cell_area, GdkRectangle, 1)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_get_size, 0)
    ZEND_ARG_INFO(0, widget)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_render, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, background_area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, cell_area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, expose_area, GdkRectangle, 1)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_set_fixed_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_start_editing, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, background_area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, cell_area, GdkRectangle, 1)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_stop_editing, 0)
    ZEND_ARG_INFO(0, canceled)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_set_alignment, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_set_padding, 0)
    ZEND_ARG_INFO(0, xpad)
    ZEND_ARG_INFO(0, ypad)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_set_sensitive, 0)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderer_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcellrenderer_methods[] = {
	PHP_ME(GtkCellRenderer, activate,             arginfo_gtk_gtkcellrenderer_activate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, editing_canceled,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_alignment,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_fixed_size,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_padding,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_sensitive,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_size,             arginfo_gtk_gtkcellrenderer_get_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, render,               arginfo_gtk_gtkcellrenderer_render, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, set_alignment,        arginfo_gtk_gtkcellrenderer_set_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, set_fixed_size,       arginfo_gtk_gtkcellrenderer_set_fixed_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, set_padding,          arginfo_gtk_gtkcellrenderer_set_padding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, set_sensitive,        arginfo_gtk_gtkcellrenderer_set_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, set_visible,          arginfo_gtk_gtkcellrenderer_set_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, start_editing,        arginfo_gtk_gtkcellrenderer_start_editing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRenderer, stop_editing,         arginfo_gtk_gtkcellrenderer_stop_editing, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1287 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAdjustment, __construct)
{
    gdouble value = 0;
    gdouble lower = 0;
    gdouble upper = 0;
    gdouble step_incr = 0;
    gdouble page_incr = 0;
    gdouble page_size = 0;
    GObject *wrapped_obj;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|dddddd", &value, &lower, &upper,
                            &step_incr, &page_incr, &page_size))
        return;

    wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC),
                               "value", value, "lower", lower,
                               "upper", upper, "step_increment", step_incr,
                               "page_increment", page_incr, "page_size", page_size, NULL);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAdjustment);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 17100 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkAdjustment, value)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->value;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkAdjustment, lower)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->lower;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkAdjustment, upper)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->upper;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkAdjustment, step_increment)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->step_increment;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkAdjustment, page_increment)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->page_increment;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkAdjustment, page_size)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_ADJUSTMENT(((phpg_gobject_t *)object)->obj)->page_size;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


static prop_info_t gtkadjustment_prop_info[] = {
	{ "value", PHPG_PROP_READ_FN(GtkAdjustment, value), NULL },
	{ "lower", PHPG_PROP_READ_FN(GtkAdjustment, lower), NULL },
	{ "upper", PHPG_PROP_READ_FN(GtkAdjustment, upper), NULL },
	{ "step_increment", PHPG_PROP_READ_FN(GtkAdjustment, step_increment), NULL },
	{ "page_increment", PHPG_PROP_READ_FN(GtkAdjustment, page_increment), NULL },
	{ "page_size", PHPG_PROP_READ_FN(GtkAdjustment, page_size), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkAdjustment, changed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_adjustment_changed(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAdjustment, clamp_page)
{
	double lower, upper;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &lower, &upper))
		return;

    gtk_adjustment_clamp_page(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), lower, upper);

}


static PHP_METHOD(GtkAdjustment, get_value)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_value(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_value)
{
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), value);

}


static PHP_METHOD(GtkAdjustment, value_changed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_adjustment_value_changed(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkAdjustment, get_lower)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_lower(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_lower)
{
	double lower;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &lower))
		return;

    gtk_adjustment_set_lower(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), lower);

}


static PHP_METHOD(GtkAdjustment, get_upper)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_upper(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_upper)
{
	double upper;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &upper))
		return;

    gtk_adjustment_set_upper(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), upper);

}


static PHP_METHOD(GtkAdjustment, get_step_increment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_step_increment(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_step_increment)
{
	double step_increment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &step_increment))
		return;

    gtk_adjustment_set_step_increment(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), step_increment);

}


static PHP_METHOD(GtkAdjustment, get_page_increment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_page_increment(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_page_increment)
{
	double page_increment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &page_increment))
		return;

    gtk_adjustment_set_page_increment(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), page_increment);

}


static PHP_METHOD(GtkAdjustment, get_page_size)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_adjustment_get_page_size(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkAdjustment, set_page_size)
{
	double page_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &page_size))
		return;

    gtk_adjustment_set_page_size(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), page_size);

}


static PHP_METHOD(GtkAdjustment, configure)
{
	double value, lower, upper, step_increment, page_increment, page_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dddddd", &value, &lower, &upper, &step_increment, &page_increment, &page_size))
		return;

    gtk_adjustment_configure(GTK_ADJUSTMENT(PHPG_GOBJECT(this_ptr)), value, lower, upper, step_increment, page_increment, page_size);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_gtk_adjustment_new, 0)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, lower)
    ZEND_ARG_INFO(0, upper)
    ZEND_ARG_INFO(0, step_incr)
    ZEND_ARG_INFO(0, page_incr)
    ZEND_ARG_INFO(0, page_size)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_clamp_page, 0)
    ZEND_ARG_INFO(0, lower)
    ZEND_ARG_INFO(0, upper)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_lower, 0)
    ZEND_ARG_INFO(0, lower)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_upper, 0)
    ZEND_ARG_INFO(0, upper)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_step_increment, 0)
    ZEND_ARG_INFO(0, step_increment)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_page_increment, 0)
    ZEND_ARG_INFO(0, page_increment)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_set_page_size, 0)
    ZEND_ARG_INFO(0, page_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkadjustment_configure, 0)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, lower)
    ZEND_ARG_INFO(0, upper)
    ZEND_ARG_INFO(0, step_increment)
    ZEND_ARG_INFO(0, page_increment)
    ZEND_ARG_INFO(0, page_size)
ZEND_END_ARG_INFO();

static zend_function_entry gtkadjustment_methods[] = {
	PHP_ME(GtkAdjustment, __construct,          arginfo_gtk_gtkadjustment_gtk_adjustment_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, changed,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, clamp_page,           arginfo_gtk_gtkadjustment_clamp_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, configure,            arginfo_gtk_gtkadjustment_configure, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_lower,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_page_increment,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_page_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_step_increment,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_upper,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, get_value,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_lower,            arginfo_gtk_gtkadjustment_set_lower, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_page_increment,   arginfo_gtk_gtkadjustment_set_page_increment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_page_size,        arginfo_gtk_gtkadjustment_set_page_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_step_increment,   arginfo_gtk_gtkadjustment_set_step_increment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_upper,            arginfo_gtk_gtkadjustment_set_upper, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, set_value,            arginfo_gtk_gtkadjustment_set_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAdjustment, value_changed,        NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererToggle, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererToggle);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererToggle);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkCellRendererToggle, get_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_renderer_toggle_get_active(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCellRendererToggle, get_radio)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_renderer_toggle_get_radio(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCellRendererToggle, set_active)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_cell_renderer_toggle_set_active(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkCellRendererToggle, set_radio)
{
	zend_bool radio;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &radio))
		return;

    gtk_cell_renderer_toggle_set_radio(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)), (gboolean)radio);

}


static PHP_METHOD(GtkCellRendererToggle, set_activatable)
{
	zend_bool create_folders;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &create_folders))
		return;

    gtk_cell_renderer_toggle_set_activatable(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)), (gboolean)create_folders);

}


static PHP_METHOD(GtkCellRendererToggle, get_activatable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_renderer_toggle_get_activatable(GTK_CELL_RENDERER_TOGGLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderertoggle_set_active, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderertoggle_set_radio, 0)
    ZEND_ARG_INFO(0, radio)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderertoggle_set_activatable, 0)
    ZEND_ARG_INFO(0, create_folders)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcellrenderertoggle_methods[] = {
	PHP_ME(GtkCellRendererToggle, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, get_activatable,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, get_radio,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, set_activatable,      arginfo_gtk_gtkcellrenderertoggle_set_activatable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, set_active,           arginfo_gtk_gtkcellrenderertoggle_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererToggle, set_radio,            arginfo_gtk_gtkcellrenderertoggle_set_radio, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererText, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererText);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererText);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkCellRendererText, set_fixed_height_from_font)
{
	long number_of_rows;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &number_of_rows))
		return;

    gtk_cell_renderer_text_set_fixed_height_from_font(GTK_CELL_RENDERER_TEXT(PHPG_GOBJECT(this_ptr)), (gint)number_of_rows);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellrenderertext_set_fixed_height_from_font, 0)
    ZEND_ARG_INFO(0, number_of_rows)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcellrenderertext_methods[] = {
	PHP_ME(GtkCellRendererText, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellRendererText, set_fixed_height_from_font, arginfo_gtk_gtkcellrenderertext_set_fixed_height_from_font, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererProgress, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererProgress);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererProgress);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrendererprogress_methods[] = {
	PHP_ME(GtkCellRendererProgress, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererAccel, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererAccel);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererAccel);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrendereraccel_methods[] = {
	PHP_ME(GtkCellRendererAccel, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererSpin, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererSpin);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererSpin);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrendererspin_methods[] = {
	PHP_ME(GtkCellRendererSpin, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererCombo, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererCombo);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererCombo);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrenderercombo_methods[] = {
	PHP_ME(GtkCellRendererCombo, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellRendererPixbuf, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererPixbuf);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellRendererPixbuf);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkcellrendererpixbuf_methods[] = {
	PHP_ME(GtkCellRendererPixbuf, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentFilter, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentFilter);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentFilter);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkRecentFilter, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_recent_filter_set_name(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkRecentFilter, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_filter_get_name(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkRecentFilter, add_mime_type)
{
	char *mime_type;
	zend_bool free_mime_type = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &mime_type, &free_mime_type))
		return;

    gtk_recent_filter_add_mime_type(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), mime_type);
	if (free_mime_type) g_free(mime_type);

}


static PHP_METHOD(GtkRecentFilter, add_pattern)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_recent_filter_add_pattern(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkRecentFilter, add_pixbuf_formats)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_recent_filter_add_pixbuf_formats(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkRecentFilter, add_application)
{
	char *application;
	zend_bool free_application = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &application, &free_application))
		return;

    gtk_recent_filter_add_application(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), application);
	if (free_application) g_free(application);

}


static PHP_METHOD(GtkRecentFilter, add_group)
{
	char *group;
	zend_bool free_group = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &group, &free_group))
		return;

    gtk_recent_filter_add_group(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), group);
	if (free_group) g_free(group);

}


static PHP_METHOD(GtkRecentFilter, add_age)
{
	long days;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &days))
		return;

    gtk_recent_filter_add_age(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)), (gint)days);

}


static PHP_METHOD(GtkRecentFilter, get_needed)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_filter_get_needed(GTK_RECENT_FILTER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_add_mime_type, 0)
    ZEND_ARG_INFO(0, mime_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_add_pattern, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_add_application, 0)
    ZEND_ARG_INFO(0, application)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_add_group, 0)
    ZEND_ARG_INFO(0, group)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentfilter_add_age, 0)
    ZEND_ARG_INFO(0, days)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentfilter_methods[] = {
	PHP_ME(GtkRecentFilter, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_age,              arginfo_gtk_gtkrecentfilter_add_age, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_application,      arginfo_gtk_gtkrecentfilter_add_application, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_group,            arginfo_gtk_gtkrecentfilter_add_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_mime_type,        arginfo_gtk_gtkrecentfilter_add_mime_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_pattern,          arginfo_gtk_gtkrecentfilter_add_pattern, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, add_pixbuf_formats,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, get_needed,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentFilter, set_name,             arginfo_gtk_gtkrecentfilter_set_name, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRcStyle, rc_add_class_style)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_rc_add_class_style(GTK_RC_STYLE(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkRcStyle, rc_add_widget_class_style)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_rc_add_widget_class_style(GTK_RC_STYLE(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkRcStyle, rc_add_widget_name_style)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_rc_add_widget_name_style(GTK_RC_STYLE(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkRcStyle, copy)
{
	GtkRcStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_style_copy(GTK_RC_STYLE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrcstyle_rc_add_class_style, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrcstyle_rc_add_widget_class_style, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrcstyle_rc_add_widget_name_style, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrcstyle_methods[] = {
	PHP_ME(GtkRcStyle, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRcStyle, rc_add_class_style,   arginfo_gtk_gtkrcstyle_rc_add_class_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRcStyle, rc_add_widget_class_style, arginfo_gtk_gtkrcstyle_rc_add_widget_class_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRcStyle, rc_add_widget_name_style, arginfo_gtk_gtkrcstyle_rc_add_widget_name_style, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSettings, get_default)
{
	GtkSettings* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_settings_get_default();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSettings, get_for_screen)
{
	zval *screen;
	GtkSettings* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    php_retval = gtk_settings_get_for_screen(GDK_SCREEN(PHPG_GOBJECT(screen)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSettings, set_double_property)
{
	char *name, *origin;
	zend_bool free_name = FALSE, free_origin = FALSE;
	double v_double;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "udu", &name, &free_name, &v_double, &origin, &free_origin))
		return;

    gtk_settings_set_double_property(GTK_SETTINGS(PHPG_GOBJECT(this_ptr)), name, v_double, origin);
	if (free_name) g_free(name);
	if (free_origin) g_free(origin);

}


static PHP_METHOD(GtkSettings, set_long_property)
{
	char *name, *origin;
	zend_bool free_name = FALSE, free_origin = FALSE;
	long v_long;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiu", &name, &free_name, &v_long, &origin, &free_origin))
		return;

    gtk_settings_set_long_property(GTK_SETTINGS(PHPG_GOBJECT(this_ptr)), name, (glong)v_long, origin);
	if (free_name) g_free(name);
	if (free_origin) g_free(origin);

}


static PHP_METHOD(GtkSettings, set_string_property)
{
	char *name, *v_string, *origin;
	zend_bool free_name = FALSE, free_v_string = FALSE, free_origin = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uuu", &name, &free_name, &v_string, &free_v_string, &origin, &free_origin))
		return;

    gtk_settings_set_string_property(GTK_SETTINGS(PHPG_GOBJECT(this_ptr)), name, v_string, origin);
	if (free_name) g_free(name);
	if (free_v_string) g_free(v_string);
	if (free_origin) g_free(origin);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksettings_get_for_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksettings_set_double_property, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, v_double)
    ZEND_ARG_INFO(0, origin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksettings_set_long_property, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, v_long)
    ZEND_ARG_INFO(0, origin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksettings_set_string_property, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, v_string)
    ZEND_ARG_INFO(0, origin)
ZEND_END_ARG_INFO();

static zend_function_entry gtksettings_methods[] = {
	PHP_ME(GtkSettings, get_default,          NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkSettings, get_for_screen,       arginfo_gtk_gtksettings_get_for_screen, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkSettings, set_double_property,  arginfo_gtk_gtksettings_set_double_property, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSettings, set_long_property,    arginfo_gtk_gtksettings_set_long_property, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSettings, set_string_property,  arginfo_gtk_gtksettings_set_string_property, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSizeGroup, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "mode",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^V", &php_args[0])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSizeGroup);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSizeGroup);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSizeGroup);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkSizeGroup, add_widget)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_size_group_add_widget(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkSizeGroup, get_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_size_group_get_mode(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSizeGroup, remove_widget)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_size_group_remove_widget(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkSizeGroup, set_mode)
{
	GtkSizeGroupMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(GTK_TYPE_SIZE_GROUP_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_size_group_set_mode(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkSizeGroup, get_ignore_hidden)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_size_group_get_ignore_hidden(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSizeGroup, set_ignore_hidden)
{
	zend_bool ignore_hidden;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &ignore_hidden))
		return;

    gtk_size_group_set_ignore_hidden(GTK_SIZE_GROUP(PHPG_GOBJECT(this_ptr)), (gboolean)ignore_hidden);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksizegroup_size_group_new, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksizegroup_add_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksizegroup_remove_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksizegroup_set_mode, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksizegroup_set_ignore_hidden, 0)
    ZEND_ARG_INFO(0, ignore_hidden)
ZEND_END_ARG_INFO();

static zend_function_entry gtksizegroup_methods[] = {
	PHP_ME(GtkSizeGroup, __construct,          arginfo_gtk_gtksizegroup_size_group_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, add_widget,           arginfo_gtk_gtksizegroup_add_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, get_ignore_hidden,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, get_mode,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, remove_widget,        arginfo_gtk_gtksizegroup_remove_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, set_ignore_hidden,    arginfo_gtk_gtksizegroup_set_ignore_hidden, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSizeGroup, set_mode,             arginfo_gtk_gtksizegroup_set_mode, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkStyle, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStyle);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStyle);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 61 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, fg)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->fg TSRMLS_CC);

	return SUCCESS;
}

#line 18439 "ext/gtk+/gen_gtk.c"


#line 72 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, bg)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->bg TSRMLS_CC);

	return SUCCESS;
}

#line 18452 "ext/gtk+/gen_gtk.c"


#line 83 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, light)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->light TSRMLS_CC);

	return SUCCESS;
}

#line 18465 "ext/gtk+/gen_gtk.c"


#line 94 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, dark)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->dark TSRMLS_CC);

	return SUCCESS;
}

#line 18478 "ext/gtk+/gen_gtk.c"


#line 105 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, mid)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->mid TSRMLS_CC);

	return SUCCESS;
}

#line 18491 "ext/gtk+/gen_gtk.c"


#line 116 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, text)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->text TSRMLS_CC);

	return SUCCESS;
}

#line 18504 "ext/gtk+/gen_gtk.c"


#line 138 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, base)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->base TSRMLS_CC);

	return SUCCESS;
}

#line 18517 "ext/gtk+/gen_gtk.c"


#line 127 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, text_aa)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_COLOR_ARRAY, style->text_aa TSRMLS_CC);

	return SUCCESS;
}

#line 18530 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkStyle, white)
{
	GdkColor php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->white;
	phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkStyle, black)
{
	GdkColor php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->black;
	phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkStyle, font_desc)
{
	PangoFontDescription *php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->font_desc;
	phpg_gboxed_new(&return_value, PANGO_TYPE_FONT_DESCRIPTION, php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkStyle, xthickness)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->xthickness;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkStyle, ythickness)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->ythickness;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}

#line 149 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, fg_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->fg_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18616 "ext/gtk+/gen_gtk.c"


#line 160 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, bg_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->bg_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18629 "ext/gtk+/gen_gtk.c"


#line 171 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, light_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->light_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18642 "ext/gtk+/gen_gtk.c"


#line 182 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, dark_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->dark_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18655 "ext/gtk+/gen_gtk.c"


#line 193 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, mid_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->mid_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18668 "ext/gtk+/gen_gtk.c"


#line 204 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, text_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->text_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18681 "ext/gtk+/gen_gtk.c"


#line 226 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, base_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->base_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18694 "ext/gtk+/gen_gtk.c"


#line 215 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, text_aa_gc)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_GC_ARRAY, style->text_aa_gc TSRMLS_CC);

	return SUCCESS;
}

#line 18707 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkStyle, white_gc)
{
	GdkGC* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->white_gc;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkStyle, black_gc)
{
	GdkGC* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_STYLE(((phpg_gobject_t *)object)->obj)->black_gc;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}

#line 237 "ext/gtk+/gtkstyle.overrides"
PHPG_PROP_READER(GtkStyle, bg_pixmap)
{
	GtkStyle *style = GTK_STYLE(((phpg_gobject_t *)object)->obj);

	phpg_create_style_helper(&return_value, style, STYLE_PIXMAP_ARRAY, style->bg_pixmap TSRMLS_CC);

	return SUCCESS;
}

#line 18748 "ext/gtk+/gen_gtk.c"



static prop_info_t gtkstyle_prop_info[] = {
	{ "fg", PHPG_PROP_READ_FN(GtkStyle, fg), NULL },
	{ "bg", PHPG_PROP_READ_FN(GtkStyle, bg), NULL },
	{ "light", PHPG_PROP_READ_FN(GtkStyle, light), NULL },
	{ "dark", PHPG_PROP_READ_FN(GtkStyle, dark), NULL },
	{ "mid", PHPG_PROP_READ_FN(GtkStyle, mid), NULL },
	{ "text", PHPG_PROP_READ_FN(GtkStyle, text), NULL },
	{ "base", PHPG_PROP_READ_FN(GtkStyle, base), NULL },
	{ "text_aa", PHPG_PROP_READ_FN(GtkStyle, text_aa), NULL },
	{ "white", PHPG_PROP_READ_FN(GtkStyle, white), NULL },
	{ "black", PHPG_PROP_READ_FN(GtkStyle, black), NULL },
	{ "font_desc", PHPG_PROP_READ_FN(GtkStyle, font_desc), NULL },
	{ "xthickness", PHPG_PROP_READ_FN(GtkStyle, xthickness), NULL },
	{ "ythickness", PHPG_PROP_READ_FN(GtkStyle, ythickness), NULL },
	{ "fg_gc", PHPG_PROP_READ_FN(GtkStyle, fg_gc), NULL },
	{ "bg_gc", PHPG_PROP_READ_FN(GtkStyle, bg_gc), NULL },
	{ "light_gc", PHPG_PROP_READ_FN(GtkStyle, light_gc), NULL },
	{ "dark_gc", PHPG_PROP_READ_FN(GtkStyle, dark_gc), NULL },
	{ "mid_gc", PHPG_PROP_READ_FN(GtkStyle, mid_gc), NULL },
	{ "text_gc", PHPG_PROP_READ_FN(GtkStyle, text_gc), NULL },
	{ "base_gc", PHPG_PROP_READ_FN(GtkStyle, base_gc), NULL },
	{ "text_aa_gc", PHPG_PROP_READ_FN(GtkStyle, text_aa_gc), NULL },
	{ "white_gc", PHPG_PROP_READ_FN(GtkStyle, white_gc), NULL },
	{ "black_gc", PHPG_PROP_READ_FN(GtkStyle, black_gc), NULL },
	{ "bg_pixmap", PHPG_PROP_READ_FN(GtkStyle, bg_pixmap), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkStyle, paint_arrow)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_arrow_type = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE, fill;
	GtkArrowType arrow_type;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuVbiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &php_arrow_type, &fill, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_arrow_type && phpg_gvalue_get_enum(GTK_TYPE_ARROW_TYPE, php_arrow_type, (gint *)&arrow_type) == FAILURE) {
		return;
	}

    gtk_paint_arrow(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, arrow_type, (gboolean)fill, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_box)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_box(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_box_gap)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_gap_side = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height, gap_x, gap_width;
	GtkPositionType gap_side;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiiiVii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height, &php_gap_side, &gap_x, &gap_width))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_gap_side && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_gap_side, (gint *)&gap_side) == FAILURE) {
		return;
	}

    gtk_paint_box_gap(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height, gap_side, (gint)gap_x, (gint)gap_width);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_check)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_check(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_diamond)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_diamond(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_expander)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget, *php_expander_style = NULL;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y;
	GtkExpanderStyle expander_style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuiiV", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &php_expander_style))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_expander_style && phpg_gvalue_get_enum(GTK_TYPE_EXPANDER_STYLE, php_expander_style, (gint *)&expander_style) == FAILURE) {
		return;
	}

    gtk_paint_expander(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, (gint)x, (gint)y, expander_style);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_extension)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_gap_side = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;
	GtkPositionType gap_side;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiiiV", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height, &php_gap_side))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_gap_side && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_gap_side, (gint *)&gap_side) == FAILURE) {
		return;
	}

    gtk_paint_extension(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height, gap_side);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_flat_box)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_flat_box(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_focus)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_focus(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_handle)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_orientation = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;
	GtkOrientation orientation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiiiV", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height, &php_orientation))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_paint_handle(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height, orientation);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_hline)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x1, x2, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuiii", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x1, &x2, &y))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_hline(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, (gint)x1, (gint)x2, (gint)y);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_layout)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget, *php_layout;
	GtkStateType state_type;
	zend_bool use_text, free_detail = FALSE;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	long x, y;
	PangoLayout *layout = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVbVNuiiN", &window, gdkwindow_ce, &php_state_type, &use_text, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &php_layout, pangolayout_ce))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));
    if (Z_TYPE_P(php_layout) != IS_NULL)
        layout = PANGO_LAYOUT(PHPG_GOBJECT(php_layout));

    gtk_paint_layout(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, (gboolean)use_text, area, widget, detail, (gint)x, (gint)y, layout);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_option)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_option(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}

#line 5 "ext/gtk+/gtkstyle.overrides"
static PHP_METHOD(GtkStyle, paint_polygon)
{
    zval *window, *php_state_type, *php_shadow_type, *php_area;
    GtkStateType state_type;
    GtkShadowType shadow_type;
    GdkRectangle area;
    zval *widget;
    const gchar *detail;
    zval *php_points;
    int npoints, i;
    GdkPoint *points;
    zend_bool fill;
    zval **item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVOs!ab", &window, gdkwindow_ce,
                            &php_state_type, &php_shadow_type, &php_area, &widget,
                            gtkwidget_ce, &detail, &php_points, &fill)) {
        return;
    }

    if (phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
        return;
    }

    if (phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
        return;
    }

    if (phpg_rectangle_from_zval(php_area, &area TSRMLS_CC) == FAILURE) {
        return;
    }

    npoints = zend_hash_num_elements(Z_ARRVAL_P(php_area));
    points = g_new(GdkPoint, npoints);
    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_points)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_points), (void**)&item) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_points)), i++) {

		if (Z_TYPE_PP(item) != IS_ARRAY ||
            !php_gtk_parse_args_hash_quiet(*item, "ii", &points[i].x, &points[i].y)) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "points have to be 2 element integer arrays");
            g_free(points);
            return;
        }
    }

    gtk_paint_polygon(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)),
                      state_type, shadow_type, &area, GTK_WIDGET(PHPG_GOBJECT(widget)),
                      detail, points, npoints, fill);
    g_free(points);
}

#line 19363 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkStyle, paint_resize_grip)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget, *php_edge = NULL;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	GdkWindowEdge edge;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuViiii", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &php_edge, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_edge && phpg_gvalue_get_enum(GDK_TYPE_WINDOW_EDGE, php_edge, (gint *)&edge) == FAILURE) {
		return;
	}

    gtk_paint_resize_grip(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, edge, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_shadow)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_shadow(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_shadow_gap)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_gap_side = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height, gap_x, gap_width;
	GtkPositionType gap_side;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiiiVii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height, &php_gap_side, &gap_x, &gap_width))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_gap_side && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_gap_side, (gint *)&gap_side) == FAILURE) {
		return;
	}

    gtk_paint_shadow_gap(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height, gap_side, (gint)gap_x, (gint)gap_width);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_slider)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget, *php_orientation = NULL;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;
	GtkOrientation orientation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiiiV", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height, &php_orientation))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_paint_slider(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height, orientation);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_string)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail, *string;
	zend_bool free_detail = FALSE, free_string = FALSE;
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuiiu", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &string, &free_string))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_string(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, (gint)x, (gint)y, string);
	if (free_detail) g_free(detail);
	if (free_string) g_free(string);

}


static PHP_METHOD(GtkStyle, paint_tab)
{
	zval *window, *php_state_type = NULL, *php_shadow_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GtkShadowType shadow_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVVNuiiii", &window, gdkwindow_ce, &php_state_type, &php_shadow_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_tab(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, shadow_type, area, widget, detail, (gint)x, (gint)y, (gint)width, (gint)height);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, paint_vline)
{
	zval *window, *php_state_type = NULL, *php_area = NULL, *php_widget;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	GtkWidget *widget = NULL;
	char *detail;
	zend_bool free_detail = FALSE;
	long y1, y2, x;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVNuiii", &window, gdkwindow_ce, &php_state_type, &php_area, &php_widget, gtkwidget_ce, &detail, &free_detail, &y1, &y2, &x))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_paint_vline(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type, area, widget, detail, (gint)y1, (gint)y2, (gint)x);
	if (free_detail) g_free(detail);

}


static PHP_METHOD(GtkStyle, apply_default_background)
{
	zval *window, *php_state_type = NULL, *php_area = NULL;
	zend_bool set_bg;
	GtkStateType state_type;
	GdkRectangle area_arg = { 0, 0, 0, 0 }, *area;
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ObVViiii", &window, gdkwindow_ce, &set_bg, &php_state_type, &php_area, &x, &y, &width, &height))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_area) == IS_NULL) {
        area = NULL;
    } else if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area_arg TSRMLS_CC) == SUCCESS) {
        area = &area_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects area argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    gtk_style_apply_default_background(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), (gboolean)set_bg, state_type, area, (gint)x, (gint)y, (gint)width, (gint)height);

}


static PHP_METHOD(GtkStyle, apply_default_pixmap)
{
	zval *window, *php_area;
	zend_bool set_bg;
	GdkRectangle area = { 0, 0, 0, 0 };
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ObViiii", &window, gdkwindow_ce, &set_bg, &php_area, &x, &y, &width, &height))
		return;

    if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	phpg_warn_deprecated("use GtkStyle::apply_default_background()" TSRMLS_CC);

    gtk_style_apply_default_pixmap(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), (gboolean)set_bg, &area, (gint)x, (gint)y, (gint)width, (gint)height);

}


static PHP_METHOD(GtkStyle, attach)
{
	zval *window;
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gdkwindow_ce))
		return;

    php_retval = gtk_style_attach(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkStyle, copy)
{
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_style_copy(GTK_STYLE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


static PHP_METHOD(GtkStyle, detach)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_style_detach(GTK_STYLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkStyle, get_font)
{
	GdkFont *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_style_get_font(GTK_STYLE(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GDK_TYPE_FONT, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkStyle, lookup_icon_set)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSet *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    php_retval = gtk_style_lookup_icon_set(GTK_STYLE(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);
	phpg_gboxed_new(&return_value, GTK_TYPE_ICON_SET, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkStyle, render_icon)
{
	GtkIconSource *source = NULL;
	zval *php_source, *php_direction = NULL, *php_state = NULL, *php_size = NULL, *php_widget = NULL;
	GtkTextDirection direction;
	GtkStateType state;
	GtkIconSize size;
	GtkWidget *widget = NULL;
	char *detail = NULL;
	zend_bool free_detail = FALSE;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVVV|Nu", &php_source, gboxed_ce, &php_direction, &php_state, &php_size, &php_widget, gtkwidget_ce, &detail, &free_detail))
		return;

    if (phpg_gboxed_check(php_source, GTK_TYPE_ICON_SOURCE, FALSE TSRMLS_CC)) {
        source = (GtkIconSource *) PHPG_GBOXED(php_source);
    } else {
        php_error(E_WARNING, "%s::%s() expects source argument to be a valid GtkIconSource object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}
    if (php_widget) {
        if (Z_TYPE_P(php_widget) == IS_NULL)
            widget = NULL;
        else
            widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));
    }

    php_retval = gtk_style_render_icon(GTK_STYLE(PHPG_GOBJECT(this_ptr)), source, direction, state, size, widget, detail);
	if (free_detail) g_free(detail);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkStyle, set_background)
{
	zval *window, *php_state_type = NULL;
	GtkStateType state_type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &window, gdkwindow_ce, &php_state_type))
		return;

	if (php_state_type && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state_type, (gint *)&state_type) == FAILURE) {
		return;
	}

    gtk_style_set_background(GTK_STYLE(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), state_type);

}


static PHP_METHOD(GtkStyle, set_font)
{
	GdkFont *font = NULL;
	zval *php_font;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_font, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_font, GDK_TYPE_FONT, FALSE TSRMLS_CC)) {
        font = (GdkFont *) PHPG_GBOXED(php_font);
    } else {
        php_error(E_WARNING, "%s::%s() expects font argument to be a valid GdkFont object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_style_set_font(GTK_STYLE(PHPG_GOBJECT(this_ptr)), font);

}


static PHP_METHOD(GtkStyle, lookup_color)
{
	char *color_name;
	zend_bool free_color_name = FALSE;
	GdkColor *color = NULL;
	zval *php_color;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uO", &color_name, &free_color_name, &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_style_lookup_color(GTK_STYLE(PHPG_GOBJECT(this_ptr)), color_name, color);
	if (free_color_name) g_free(color_name);
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_arrow, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, arrow_type)
    ZEND_ARG_INFO(0, fill)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_box, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_box_gap, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, gap_side)
    ZEND_ARG_INFO(0, gap_x)
    ZEND_ARG_INFO(0, gap_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_check, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_diamond, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_expander, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, expander_style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_extension, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, gap_side)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_flat_box, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_focus, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_handle, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_hline, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x1)
    ZEND_ARG_INFO(0, x2)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_layout, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, use_text)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_OBJ_INFO(0, layout, PangoLayout, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_option, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_resize_grip, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, edge)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_shadow, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_shadow_gap, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, gap_side)
    ZEND_ARG_INFO(0, gap_x)
    ZEND_ARG_INFO(0, gap_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_slider, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_string, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, string)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_tab, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_INFO(0, shadow_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_paint_vline, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
    ZEND_ARG_INFO(0, y1)
    ZEND_ARG_INFO(0, y2)
    ZEND_ARG_INFO(0, x)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_apply_default_background, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, set_bg)
    ZEND_ARG_INFO(0, state_type)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_apply_default_pixmap, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, set_bg)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_attach, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_lookup_icon_set, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkstyle_render_icon, 0, 0, 4)
    ZEND_ARG_OBJ_INFO(0, source, GtkIconSource, 1)
    ZEND_ARG_INFO(0, direction)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, detail)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_set_background, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_INFO(0, state_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_set_font, 0)
    ZEND_ARG_OBJ_INFO(0, font, GdkFont, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstyle_lookup_color, 0)
    ZEND_ARG_INFO(0, color_name)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkstyle_methods[] = {
	PHP_ME(GtkStyle, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, apply_default_background, arginfo_gtk_gtkstyle_apply_default_background, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, apply_default_pixmap, arginfo_gtk_gtkstyle_apply_default_pixmap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, attach,               arginfo_gtk_gtkstyle_attach, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, copy,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, detach,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, get_font,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, lookup_color,         arginfo_gtk_gtkstyle_lookup_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, lookup_icon_set,      arginfo_gtk_gtkstyle_lookup_icon_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_arrow,          arginfo_gtk_gtkstyle_paint_arrow, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_box,            arginfo_gtk_gtkstyle_paint_box, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_box_gap,        arginfo_gtk_gtkstyle_paint_box_gap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_check,          arginfo_gtk_gtkstyle_paint_check, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_diamond,        arginfo_gtk_gtkstyle_paint_diamond, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_expander,       arginfo_gtk_gtkstyle_paint_expander, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_extension,      arginfo_gtk_gtkstyle_paint_extension, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_flat_box,       arginfo_gtk_gtkstyle_paint_flat_box, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_focus,          arginfo_gtk_gtkstyle_paint_focus, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_handle,         arginfo_gtk_gtkstyle_paint_handle, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_hline,          arginfo_gtk_gtkstyle_paint_hline, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_layout,         arginfo_gtk_gtkstyle_paint_layout, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_option,         arginfo_gtk_gtkstyle_paint_option, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_polygon,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_resize_grip,    arginfo_gtk_gtkstyle_paint_resize_grip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_shadow,         arginfo_gtk_gtkstyle_paint_shadow, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_shadow_gap,     arginfo_gtk_gtkstyle_paint_shadow_gap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_slider,         arginfo_gtk_gtkstyle_paint_slider, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_string,         arginfo_gtk_gtkstyle_paint_string, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_tab,            arginfo_gtk_gtkstyle_paint_tab, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, paint_vline,          arginfo_gtk_gtkstyle_paint_vline, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, render_icon,          arginfo_gtk_gtkstyle_render_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, set_background,       arginfo_gtk_gtkstyle_set_background, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStyle, set_font,             arginfo_gtk_gtkstyle_set_font, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextBuffer, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "tag_table",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtktexttagtable_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextBuffer);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextBuffer);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextBuffer);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkTextBuffer, tag_table)
{
	GtkTextTagTable* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TEXT_BUFFER(((phpg_gobject_t *)object)->obj)->tag_table;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtktextbuffer_prop_info[] = {
	{ "tag_table", PHPG_PROP_READ_FN(GtkTextBuffer, tag_table), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkTextBuffer, add_selection_clipboard)
{
	zval *clipboard;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &clipboard, gtkclipboard_ce))
		return;

    gtk_text_buffer_add_selection_clipboard(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_CLIPBOARD(PHPG_GOBJECT(clipboard)));

}


static PHP_METHOD(GtkTextBuffer, apply_tag)
{
	zval *tag, *php_start, *php_end;
	GtkTextIter *start = NULL, *end = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOO", &tag, gtktexttag_ce, &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_apply_tag(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_TEXT_TAG(PHPG_GOBJECT(tag)), start, end);

}


static PHP_METHOD(GtkTextBuffer, apply_tag_by_name)
{
	char *name;
	zend_bool free_name = FALSE;
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uOO", &name, &free_name, &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_apply_tag_by_name(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), name, start, end);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkTextBuffer, backspace)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	zend_bool interactive, default_editable;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Obb", &php_iter, gboxed_ce, &interactive, &default_editable))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_backspace(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, (gboolean)interactive, (gboolean)default_editable);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, begin_user_action)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_text_buffer_begin_user_action(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTextBuffer, copy_clipboard)
{
	zval *clipboard;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &clipboard, gtkclipboard_ce))
		return;

    gtk_text_buffer_copy_clipboard(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_CLIPBOARD(PHPG_GOBJECT(clipboard)));

}


static PHP_METHOD(GtkTextBuffer, create_child_anchor)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	GtkTextChildAnchor* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_create_child_anchor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, create_mark)
{
	char *mark_name;
	zend_bool free_mark_name = FALSE, left_gravity = FALSE;
	GtkTextIter *where = NULL;
	zval *php_where;
	GtkTextMark* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uN|b", &mark_name, &free_mark_name, &php_where, gboxed_ce, &left_gravity))
		return;

    if (Z_TYPE_P(php_where) != IS_NULL) {
        if (phpg_gboxed_check(php_where, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
            where = (GtkTextIter *) PHPG_GBOXED(php_where);
        } else {
            php_error(E_WARNING, "%s::%s() expects where argument to be a valid GtkTextIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    php_retval = gtk_text_buffer_create_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), mark_name, where, (gboolean)left_gravity);
	if (free_mark_name) g_free(mark_name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, cut_clipboard)
{
	zval *clipboard;
	zend_bool default_editable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &clipboard, gtkclipboard_ce, &default_editable))
		return;

    gtk_text_buffer_cut_clipboard(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_CLIPBOARD(PHPG_GOBJECT(clipboard)), (gboolean)default_editable);

}


static PHP_METHOD(GtkTextBuffer, delete)
{
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_delete(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), start, end);

}


static PHP_METHOD(GtkTextBuffer, delete_interactive)
{
	GtkTextIter *start_iter = NULL, *end_iter = NULL;
	zval *php_start_iter, *php_end_iter;
	zend_bool default_editable;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOb", &php_start_iter, gboxed_ce, &php_end_iter, gboxed_ce, &default_editable))
		return;

    if (phpg_gboxed_check(php_start_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start_iter = (GtkTextIter *) PHPG_GBOXED(php_start_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects start_iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end_iter = (GtkTextIter *) PHPG_GBOXED(php_end_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects end_iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_delete_interactive(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), start_iter, end_iter, (gboolean)default_editable);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, delete_mark)
{
	zval *mark;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &mark, gtktextmark_ce))
		return;

    gtk_text_buffer_delete_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)));

}


static PHP_METHOD(GtkTextBuffer, delete_mark_by_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_text_buffer_delete_mark_by_name(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkTextBuffer, delete_selection)
{
	zend_bool interactive, default_editable;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "bb", &interactive, &default_editable))
		return;

    php_retval = gtk_text_buffer_delete_selection(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), (gboolean)interactive, (gboolean)default_editable);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, end_user_action)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_text_buffer_end_user_action(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));

}

#line 19 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_bounds)
{
    GtkTextIter start, end;
    zval *php_start = NULL;
    zval *php_end = NULL;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &start, &end);

    phpg_gboxed_new(&php_start, GTK_TYPE_TEXT_ITER, &start, TRUE, TRUE TSRMLS_CC);
    phpg_gboxed_new(&php_end, GTK_TYPE_TEXT_ITER, &end, TRUE, TRUE TSRMLS_CC);

    php_gtk_build_value(&return_value, "(NN)", php_start, php_end);
}

#line 20669 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_char_count)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 75 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_end_iter)
{
    GtkTextIter iter;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20699 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_insert)
{
	GtkTextMark* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_insert(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 127 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_child_anchor)
{
    GtkTextIter iter;
    zval *anchor;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &anchor, gtktextchildanchor_ce))
        return;

    gtk_text_buffer_get_iter_at_child_anchor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter,
                                     GTK_TEXT_CHILD_ANCHOR(PHPG_GOBJECT(anchor)));
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20731 "ext/gtk+/gen_gtk.c"


#line 96 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_line)
{
    GtkTextIter iter;
    int line_number;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &line_number))
        return;

    gtk_text_buffer_get_iter_at_line(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter, line_number);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20748 "ext/gtk+/gen_gtk.c"


#line 174 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_line_index)
{
    GtkTextIter iter;
    int line_number, byte_offset = 0;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &line_number, byte_offset))
        return;

    gtk_text_buffer_get_iter_at_line_index(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter, line_number, byte_offset);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20765 "ext/gtk+/gen_gtk.c"


#line 151 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_line_offset)
{
    GtkTextIter iter;
    int line_number, char_offset;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &line_number, &char_offset))
        return;

    gtk_text_buffer_get_iter_at_line_offset(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter, line_number, char_offset);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20782 "ext/gtk+/gen_gtk.c"


#line 111 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_mark)
{
    GtkTextIter iter;
    zval *mark;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &mark, gtktextmark_ce))
        return;

    gtk_text_buffer_get_iter_at_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter,
                                     GTK_TEXT_MARK(PHPG_GOBJECT(mark)));
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20800 "ext/gtk+/gen_gtk.c"


#line 196 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_iter_at_offset)
{
    GtkTextIter iter;
    int char_offset;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &char_offset))
        return;

    gtk_text_buffer_get_iter_at_offset(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter, char_offset);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20817 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_line_count)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_line_count(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextBuffer, get_mark)
{
	char *name;
	zend_bool free_name = FALSE;
	GtkTextMark* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    php_retval = gtk_text_buffer_get_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, get_modified)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_modified(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, get_selection_bound)
{
	GtkTextMark* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_selection_bound(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 39 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_selection_bounds)
{
    GtkTextIter start, end;
    zval *php_start = NULL;
    zval *php_end = NULL;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (gtk_text_buffer_get_selection_bounds(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &start, &end)) {
        phpg_gboxed_new(&php_start, GTK_TYPE_TEXT_ITER, &start, TRUE, TRUE TSRMLS_CC);
        phpg_gboxed_new(&php_end, GTK_TYPE_TEXT_ITER, &end, TRUE, TRUE TSRMLS_CC);

        php_gtk_build_value(&return_value, "(NN)", php_start, php_end);
    } else {
        RETURN_FALSE;
    }
}

#line 20900 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_slice)
{
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;
	zend_bool include_hidden_chars = TRUE, free_result;
	gchar *php_retval, *cp_ret;
	gsize cp_len;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO|b", &php_start, gboxed_ce, &php_end, gboxed_ce, &include_hidden_chars))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_get_slice(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), start, end, (gboolean)include_hidden_chars);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}

#line 61 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, get_start_iter)
{
    GtkTextIter iter;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), &iter);
    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 20961 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_tag_table)
{
	GtkTextTagTable* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, get_text)
{
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;
	zend_bool include_hidden_chars = TRUE, free_result;
	gchar *php_retval, *cp_ret;
	gsize cp_len;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO|b", &php_start, gboxed_ce, &php_end, gboxed_ce, &include_hidden_chars))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), start, end, (gboolean)include_hidden_chars);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkTextBuffer, insert)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	char *text;
	zend_bool free_text = FALSE;
	long len = -1;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou|i", &php_iter, gboxed_ce, &text, &free_text, &len))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_insert(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, text, (gint)len);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkTextBuffer, insert_at_cursor)
{
	char *text;
	zend_bool free_text = FALSE;
	long len = -1;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u|i", &text, &free_text, &len))
		return;

    gtk_text_buffer_insert_at_cursor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), text, (gint)len);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkTextBuffer, insert_child_anchor)
{
	GtkTextIter *iter = NULL;
	zval *php_iter, *anchor;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_iter, gboxed_ce, &anchor, gtktextchildanchor_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_insert_child_anchor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, GTK_TEXT_CHILD_ANCHOR(PHPG_GOBJECT(anchor)));

}


static PHP_METHOD(GtkTextBuffer, insert_interactive)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	char *text;
	zend_bool free_text = FALSE, default_editable;
	long len;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ouib", &php_iter, gboxed_ce, &text, &free_text, &len, &default_editable))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_insert_interactive(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, text, (gint)len, (gboolean)default_editable);
	if (free_text) g_free(text);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, insert_interactive_at_cursor)
{
	char *text;
	zend_bool free_text = FALSE, default_editable;
	long len;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uib", &text, &free_text, &len, &default_editable))
		return;

    php_retval = gtk_text_buffer_insert_interactive_at_cursor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), text, (gint)len, (gboolean)default_editable);
	if (free_text) g_free(text);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, insert_pixbuf)
{
	GtkTextIter *iter = NULL;
	zval *php_iter, *pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_iter, gboxed_ce, &pixbuf, gdkpixbuf_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_insert_pixbuf(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

}


static PHP_METHOD(GtkTextBuffer, insert_range)
{
	GtkTextIter *iter = NULL, *start = NULL, *end = NULL;
	zval *php_iter, *php_start, *php_end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOO", &php_iter, gboxed_ce, &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_insert_range(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, start, end);

}


static PHP_METHOD(GtkTextBuffer, insert_range_interactive)
{
	GtkTextIter *iter = NULL, *start = NULL, *end = NULL;
	zval *php_iter, *php_start, *php_end;
	zend_bool default_editable;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOOb", &php_iter, gboxed_ce, &php_start, gboxed_ce, &php_end, gboxed_ce, &default_editable))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_insert_range_interactive(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), iter, start, end, (gboolean)default_editable);
	RETVAL_BOOL(php_retval);
}

#line 236 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, insert_with_tags_by_name)
{
    zval *tags = NULL, *php_iter, **tag;
    char *text;
    int text_len, start_offset;
    GtkTextIter *iter, start;
    GtkTextBuffer *buffer;
    zend_bool free_text;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args_quiet(ZEND_NUM_ARGS(), "Ou#a", &php_iter, gtktextiter_ce,
                                  &text, &text_len, &free_text, &tags) &&
        !php_gtk_parse_varargs(ZEND_NUM_ARGS(), 2, &tags, "Ou#", &php_iter, gtktextiter_ce,
                               &text, &text_len, &free_text)) {
        return;
    }

    iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    buffer = GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr));

    start_offset = gtk_text_iter_get_offset(iter);
    gtk_text_buffer_insert(buffer, iter, text, text_len);
    gtk_text_buffer_get_iter_at_offset(buffer, &start, start_offset);

    if (tags) {
        for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(tags));
             zend_hash_get_current_data(Z_ARRVAL_P(tags), (void **) &tag) == SUCCESS;
             zend_hash_move_forward(Z_ARRVAL_P(tags))) {

            convert_to_string_ex(tag);
            gtk_text_buffer_apply_tag_by_name(buffer, Z_STRVAL_PP(tag), &start, iter);
        }
    }
}

#line 21276 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, move_mark)
{
	zval *mark, *php_where;
	GtkTextIter *where = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &mark, gtktextmark_ce, &php_where, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_where, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        where = (GtkTextIter *) PHPG_GBOXED(php_where);
    } else {
        php_error(E_WARNING, "%s::%s() expects where argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_move_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)), where);

}


static PHP_METHOD(GtkTextBuffer, move_mark_by_name)
{
	char *name;
	zend_bool free_name = FALSE;
	GtkTextIter *where = NULL;
	zval *php_where;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uO", &name, &free_name, &php_where, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_where, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        where = (GtkTextIter *) PHPG_GBOXED(php_where);
    } else {
        php_error(E_WARNING, "%s::%s() expects where argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_move_mark_by_name(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), name, where);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkTextBuffer, paste_clipboard)
{
	zval *clipboard, *php_override_location;
	GtkTextIter *override_location = NULL;
	zend_bool default_editable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ONb", &clipboard, gtkclipboard_ce, &php_override_location, gboxed_ce, &default_editable))
		return;

    if (Z_TYPE_P(php_override_location) != IS_NULL) {
        if (phpg_gboxed_check(php_override_location, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
            override_location = (GtkTextIter *) PHPG_GBOXED(php_override_location);
        } else {
            php_error(E_WARNING, "%s::%s() expects override_location argument to be a valid GtkTextIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_text_buffer_paste_clipboard(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_CLIPBOARD(PHPG_GOBJECT(clipboard)), override_location, (gboolean)default_editable);

}


static PHP_METHOD(GtkTextBuffer, place_cursor)
{
	GtkTextIter *where = NULL;
	zval *php_where;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_where, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_where, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        where = (GtkTextIter *) PHPG_GBOXED(php_where);
    } else {
        php_error(E_WARNING, "%s::%s() expects where argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_place_cursor(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), where);

}


static PHP_METHOD(GtkTextBuffer, remove_all_tags)
{
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_remove_all_tags(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), start, end);

}


static PHP_METHOD(GtkTextBuffer, remove_selection_clipboard)
{
	zval *clipboard;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &clipboard, gtkclipboard_ce))
		return;

    gtk_text_buffer_remove_selection_clipboard(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_CLIPBOARD(PHPG_GOBJECT(clipboard)));

}


static PHP_METHOD(GtkTextBuffer, remove_tag)
{
	zval *tag, *php_start, *php_end;
	GtkTextIter *start = NULL, *end = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOO", &tag, gtktexttag_ce, &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_remove_tag(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_TEXT_TAG(PHPG_GOBJECT(tag)), start, end);

}


static PHP_METHOD(GtkTextBuffer, remove_tag_by_name)
{
	char *name;
	zend_bool free_name = FALSE;
	GtkTextIter *start = NULL, *end = NULL;
	zval *php_start, *php_end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uOO", &name, &free_name, &php_start, gboxed_ce, &php_end, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_start, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        start = (GtkTextIter *) PHPG_GBOXED(php_start);
    } else {
        php_error(E_WARNING, "%s::%s() expects start argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_end, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        end = (GtkTextIter *) PHPG_GBOXED(php_end);
    } else {
        php_error(E_WARNING, "%s::%s() expects end argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_remove_tag_by_name(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), name, start, end);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkTextBuffer, select_range)
{
	GtkTextIter *ins = NULL, *bound = NULL;
	zval *php_ins, *php_bound;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_ins, gboxed_ce, &php_bound, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_ins, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        ins = (GtkTextIter *) PHPG_GBOXED(php_ins);
    } else {
        php_error(E_WARNING, "%s::%s() expects ins argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_bound, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        bound = (GtkTextIter *) PHPG_GBOXED(php_bound);
    } else {
        php_error(E_WARNING, "%s::%s() expects bound argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_select_range(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), ins, bound);

}


static PHP_METHOD(GtkTextBuffer, set_modified)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_text_buffer_set_modified(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}

#line 218 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextBuffer, set_text)
{
    char *text;
    int text_len;
    zend_bool free_text;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u#", &text, &text_len, &free_text))
        return;

    gtk_text_buffer_set_text(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), text, text_len);

    if (free_text) g_free(text);
}

#line 21548 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextBuffer, get_has_selection)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_has_selection(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, get_copy_target_list)
{
	GtkTargetList *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_copy_target_list(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_TARGET_LIST, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextBuffer, get_paste_target_list)
{
	GtkTargetList *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_buffer_get_paste_target_list(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_TARGET_LIST, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextBuffer, register_serialize_tagset)
{
	char *tagset_name = NULL;
	zend_bool free_tagset_name = FALSE;
	GdkAtom php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|u", &tagset_name, &free_tagset_name))
		return;

    php_retval = gtk_text_buffer_register_serialize_tagset(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), tagset_name);
	if (free_tagset_name) g_free(tagset_name);
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, register_deserialize_tagset)
{
	char *tagset_name = NULL;
	zend_bool free_tagset_name = FALSE;
	GdkAtom php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|u", &tagset_name, &free_tagset_name))
		return;

    php_retval = gtk_text_buffer_register_deserialize_tagset(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), tagset_name);
	if (free_tagset_name) g_free(tagset_name);
	phpg_gdkatom_new(&return_value, php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextBuffer, unregister_serialize_format)
{
	GdkAtom format;
	zval *php_format = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_format))
		return;

    format = phpg_gdkatom_from_zval(php_format TSRMLS_CC);
    if (format == NULL) {
        php_error(E_WARNING, "%s::%s() expects format argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_unregister_serialize_format(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), format);

}


static PHP_METHOD(GtkTextBuffer, unregister_deserialize_format)
{
	GdkAtom format;
	zval *php_format = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_format))
		return;

    format = phpg_gdkatom_from_zval(php_format TSRMLS_CC);
    if (format == NULL) {
        php_error(E_WARNING, "%s::%s() expects format argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_unregister_deserialize_format(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), format);

}


static PHP_METHOD(GtkTextBuffer, deserialize_set_can_create_tags)
{
	GdkAtom format;
	zval *php_format = NULL;
	zend_bool can_create_tags;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vb", &php_format, &can_create_tags))
		return;

    format = phpg_gdkatom_from_zval(php_format TSRMLS_CC);
    if (format == NULL) {
        php_error(E_WARNING, "%s::%s() expects format argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_deserialize_set_can_create_tags(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), format, (gboolean)can_create_tags);

}


static PHP_METHOD(GtkTextBuffer, deserialize_get_can_create_tags)
{
	GdkAtom format;
	zval *php_format = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_format))
		return;

    format = phpg_gdkatom_from_zval(php_format TSRMLS_CC);
    if (format == NULL) {
        php_error(E_WARNING, "%s::%s() expects format argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_buffer_deserialize_get_can_create_tags(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), format);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextBuffer, add_mark)
{
	zval *mark, *php_where;
	GtkTextIter *where = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &mark, gtktextmark_ce, &php_where, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_where, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        where = (GtkTextIter *) PHPG_GBOXED(php_where);
    } else {
        php_error(E_WARNING, "%s::%s() expects where argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_buffer_add_mark(GTK_TEXT_BUFFER(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)), where);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_gtk_text_buffer_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, tag_table, GtkTextTagTable, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_add_selection_clipboard, 0)
    ZEND_ARG_OBJ_INFO(0, clipboard, GtkClipboard, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_apply_tag, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_apply_tag_by_name, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_backspace, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, interactive)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_copy_clipboard, 0)
    ZEND_ARG_OBJ_INFO(0, clipboard, GtkClipboard, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_create_child_anchor, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_create_mark, 0, 0, 2)
    ZEND_ARG_INFO(0, mark_name)
    ZEND_ARG_OBJ_INFO(0, where, GtkTextIter, 1)
    ZEND_ARG_INFO(0, left_gravity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_cut_clipboard, 0)
    ZEND_ARG_OBJ_INFO(0, clipboard, GtkClipboard, 1)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_delete, 0)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_delete_interactive, 0)
    ZEND_ARG_OBJ_INFO(0, start_iter, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end_iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_delete_mark, 0)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_delete_mark_by_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_delete_selection, 0)
    ZEND_ARG_INFO(0, interactive)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_get_iter_at_line, 0)
    ZEND_ARG_INFO(0, line_number)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_get_iter_at_line_index, 0)
    ZEND_ARG_INFO(0, line_number)
    ZEND_ARG_INFO(0, byte_offset)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_get_iter_at_line_offset, 0)
    ZEND_ARG_INFO(0, line_number)
    ZEND_ARG_INFO(0, char_offset)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_get_iter_at_offset, 0)
    ZEND_ARG_INFO(0, char_offset)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_get_mark, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_get_slice, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
    ZEND_ARG_INFO(0, include_hidden_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_get_text, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
    ZEND_ARG_INFO(0, include_hidden_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_insert, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_insert_at_cursor, 0, 0, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_child_anchor, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, anchor, GtkTextChildAnchor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_interactive, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_interactive_at_cursor, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, len)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_range, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_insert_range_interactive, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_move_mark, 0)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
    ZEND_ARG_OBJ_INFO(0, where, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_move_mark_by_name, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_OBJ_INFO(0, where, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_paste_clipboard, 0)
    ZEND_ARG_OBJ_INFO(0, clipboard, GtkClipboard, 1)
    ZEND_ARG_OBJ_INFO(0, override_location, GtkTextIter, 1)
    ZEND_ARG_INFO(0, default_editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_place_cursor, 0)
    ZEND_ARG_OBJ_INFO(0, where, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_remove_all_tags, 0)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_remove_selection_clipboard, 0)
    ZEND_ARG_OBJ_INFO(0, clipboard, GtkClipboard, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_remove_tag, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_remove_tag_by_name, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_OBJ_INFO(0, start, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, end, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_select_range, 0)
    ZEND_ARG_OBJ_INFO(0, ins, GtkTextIter, 1)
    ZEND_ARG_OBJ_INFO(0, bound, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_set_modified, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_set_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_register_serialize_tagset, 0, 0, 0)
    ZEND_ARG_INFO(0, tagset_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextbuffer_register_deserialize_tagset, 0, 0, 0)
    ZEND_ARG_INFO(0, tagset_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_unregister_serialize_format, 0)
    ZEND_ARG_INFO(0, format)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_unregister_deserialize_format, 0)
    ZEND_ARG_INFO(0, format)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_deserialize_set_can_create_tags, 0)
    ZEND_ARG_INFO(0, format)
    ZEND_ARG_INFO(0, can_create_tags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_deserialize_get_can_create_tags, 0)
    ZEND_ARG_INFO(0, format)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextbuffer_add_mark, 0)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
    ZEND_ARG_OBJ_INFO(0, where, GtkTextIter, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktextbuffer_methods[] = {
	PHP_ME(GtkTextBuffer, __construct,          arginfo_gtk_gtktextbuffer_gtk_text_buffer_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, add_mark,             arginfo_gtk_gtktextbuffer_add_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, add_selection_clipboard, arginfo_gtk_gtktextbuffer_add_selection_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, apply_tag,            arginfo_gtk_gtktextbuffer_apply_tag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, apply_tag_by_name,    arginfo_gtk_gtktextbuffer_apply_tag_by_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, backspace,            arginfo_gtk_gtktextbuffer_backspace, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, begin_user_action,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, copy_clipboard,       arginfo_gtk_gtktextbuffer_copy_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, create_child_anchor,  arginfo_gtk_gtktextbuffer_create_child_anchor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, create_mark,          arginfo_gtk_gtktextbuffer_create_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, cut_clipboard,        arginfo_gtk_gtktextbuffer_cut_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, delete,               arginfo_gtk_gtktextbuffer_delete, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, delete_interactive,   arginfo_gtk_gtktextbuffer_delete_interactive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, delete_mark,          arginfo_gtk_gtktextbuffer_delete_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, delete_mark_by_name,  arginfo_gtk_gtktextbuffer_delete_mark_by_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, delete_selection,     arginfo_gtk_gtktextbuffer_delete_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, deserialize_get_can_create_tags, arginfo_gtk_gtktextbuffer_deserialize_get_can_create_tags, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, deserialize_set_can_create_tags, arginfo_gtk_gtktextbuffer_deserialize_set_can_create_tags, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, end_user_action,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_bounds,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_char_count,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_copy_target_list, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_end_iter,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_has_selection,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_insert,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_child_anchor, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_line,     arginfo_gtk_gtktextbuffer_get_iter_at_line, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_line_index, arginfo_gtk_gtktextbuffer_get_iter_at_line_index, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_line_offset, arginfo_gtk_gtktextbuffer_get_iter_at_line_offset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_mark,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_iter_at_offset,   arginfo_gtk_gtktextbuffer_get_iter_at_offset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_line_count,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_mark,             arginfo_gtk_gtktextbuffer_get_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_modified,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_paste_target_list, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_selection_bound,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_selection_bounds, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_slice,            arginfo_gtk_gtktextbuffer_get_slice, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_start_iter,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_tag_table,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, get_text,             arginfo_gtk_gtktextbuffer_get_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert,               arginfo_gtk_gtktextbuffer_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_at_cursor,     arginfo_gtk_gtktextbuffer_insert_at_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_child_anchor,  arginfo_gtk_gtktextbuffer_insert_child_anchor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_interactive,   arginfo_gtk_gtktextbuffer_insert_interactive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_interactive_at_cursor, arginfo_gtk_gtktextbuffer_insert_interactive_at_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_pixbuf,        arginfo_gtk_gtktextbuffer_insert_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_range,         arginfo_gtk_gtktextbuffer_insert_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_range_interactive, arginfo_gtk_gtktextbuffer_insert_range_interactive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, insert_with_tags_by_name, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, move_mark,            arginfo_gtk_gtktextbuffer_move_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, move_mark_by_name,    arginfo_gtk_gtktextbuffer_move_mark_by_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, paste_clipboard,      arginfo_gtk_gtktextbuffer_paste_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, place_cursor,         arginfo_gtk_gtktextbuffer_place_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, register_deserialize_tagset, arginfo_gtk_gtktextbuffer_register_deserialize_tagset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, register_serialize_tagset, arginfo_gtk_gtktextbuffer_register_serialize_tagset, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, remove_all_tags,      arginfo_gtk_gtktextbuffer_remove_all_tags, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, remove_selection_clipboard, arginfo_gtk_gtktextbuffer_remove_selection_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, remove_tag,           arginfo_gtk_gtktextbuffer_remove_tag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, remove_tag_by_name,   arginfo_gtk_gtktextbuffer_remove_tag_by_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, select_range,         arginfo_gtk_gtktextbuffer_select_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, set_modified,         arginfo_gtk_gtktextbuffer_set_modified, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, set_text,             arginfo_gtk_gtktextbuffer_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, unregister_deserialize_format, arginfo_gtk_gtktextbuffer_unregister_deserialize_format, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextBuffer, unregister_serialize_format, arginfo_gtk_gtktextbuffer_unregister_serialize_format, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextChildAnchor, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextChildAnchor);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextChildAnchor);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTextChildAnchor, get_deleted)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_child_anchor_get_deleted(GTK_TEXT_CHILD_ANCHOR(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 520 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextChildAnchor, get_widgets)
{
    GList *widgets;
    gint i, len;
    zval *php_item = NULL;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    widgets = gtk_text_child_anchor_get_widgets(GTK_TEXT_CHILD_ANCHOR(PHPG_GOBJECT(this_ptr)));
    len = g_list_length(widgets);

    array_init(return_value);
    for (i = 0; i < len; i++) {
        MAKE_STD_ZVAL(php_item);
        phpg_gobject_new(&php_item, (GObject *)g_list_nth_data(widgets, i) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }

    g_list_free(widgets);
}

#line 22154 "ext/gtk+/gen_gtk.c"



static zend_function_entry gtktextchildanchor_methods[] = {
	PHP_ME(GtkTextChildAnchor, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextChildAnchor, get_deleted,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextChildAnchor, get_widgets,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextMark, __construct)
{
	char *name;
	zend_bool free_name = FALSE, left_gravity;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ub", &name, &free_name, &left_gravity)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextMark);
	}

	wrapped_obj = (GObject *) gtk_text_mark_new(name, (gboolean)left_gravity);
	if (free_name) g_free(name);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextMark);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTextMark, get_buffer)
{
	GtkTextBuffer* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_mark_get_buffer(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextMark, get_deleted)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_mark_get_deleted(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextMark, get_left_gravity)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_mark_get_left_gravity(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextMark, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_mark_get_name(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkTextMark, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_mark_get_visible(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextMark, set_visible)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_text_mark_set_visible(GTK_TEXT_MARK(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextmark_gtk_text_mark_new, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, left_gravity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextmark_set_visible, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

static zend_function_entry gtktextmark_methods[] = {
	PHP_ME(GtkTextMark, __construct,          arginfo_gtk_gtktextmark_gtk_text_mark_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, get_buffer,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, get_deleted,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, get_left_gravity,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextMark, set_visible,          arginfo_gtk_gtktextmark_set_visible, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextTag, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "name",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextTag);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextTag);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextTag);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTextTag, event)
{
	zval *event_object, *php_event, *php_iter;
	GdkEvent *event = NULL;
	GtkTextIter *iter = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOO", &event_object, gobject_ce, &php_event, gboxed_ce, &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_tag_event(GTK_TEXT_TAG(PHPG_GOBJECT(this_ptr)), G_OBJECT(PHPG_GOBJECT(event_object)), event, iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextTag, get_priority)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_tag_get_priority(GTK_TEXT_TAG(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextTag, set_priority)
{
	long priority;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &priority))
		return;

    gtk_text_tag_set_priority(GTK_TEXT_TAG(PHPG_GOBJECT(this_ptr)), (gint)priority);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktexttag_gtk_text_tag_new, 0, 0, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttag_event, 0)
    ZEND_ARG_OBJ_INFO(0, event_object, GObject, 1)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttag_set_priority, 0)
    ZEND_ARG_INFO(0, priority)
ZEND_END_ARG_INFO();

static zend_function_entry gtktexttag_methods[] = {
	PHP_ME(GtkTextTag, __construct,          arginfo_gtk_gtktexttag_gtk_text_tag_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTag, event,                arginfo_gtk_gtktexttag_event, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTag, get_priority,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTag, set_priority,         arginfo_gtk_gtktexttag_set_priority, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextTagTable, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextTagTable);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextTagTable);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTextTagTable, add)
{
	zval *tag;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &tag, gtktexttag_ce))
		return;

    gtk_text_tag_table_add(GTK_TEXT_TAG_TABLE(PHPG_GOBJECT(this_ptr)), GTK_TEXT_TAG(PHPG_GOBJECT(tag)));

}


static PHP_METHOD(GtkTextTagTable, get_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_tag_table_get_size(GTK_TEXT_TAG_TABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextTagTable, lookup)
{
	char *name;
	zend_bool free_name = FALSE;
	GtkTextTag* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    php_retval = gtk_text_tag_table_lookup(GTK_TEXT_TAG_TABLE(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextTagTable, remove)
{
	zval *tag;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &tag, gtktexttag_ce))
		return;

    gtk_text_tag_table_remove(GTK_TEXT_TAG_TABLE(PHPG_GOBJECT(this_ptr)), GTK_TEXT_TAG(PHPG_GOBJECT(tag)));

}

#line 556 "ext/gtk+/gtktextview.overrides"
static void phpg_text_tag_table_foreach_func_marshal(GtkTextTagTable *table, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_table = NULL;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_table,       (GObject*)table       TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 1, &n_args);
    args[0] = &php_table;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_table);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkTextTagTable, for_each)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTextTagTableForeach callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    zval_add_ref(&php_callback);
    cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
    callback = (GtkTextTagTableForeach)phpg_text_tag_table_foreach_func_marshal;

    gtk_text_tag_table_foreach(GTK_TEXT_TAG_TABLE(PHPG_GOBJECT(this_ptr)), callback, cb_data);
    phpg_cb_data_destroy(cb_data);
}


#line 22574 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttagtable_add, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttagtable_lookup, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttagtable_remove, 0)
    ZEND_ARG_OBJ_INFO(0, tag, GtkTextTag, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktexttagtable_for_each, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtktexttagtable_methods[] = {
	PHP_ME(GtkTextTagTable, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTagTable, add,                  arginfo_gtk_gtktexttagtable_add, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTagTable, for_each,             arginfo_gtk_gtktexttagtable_for_each, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTagTable, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTagTable, lookup,               arginfo_gtk_gtktexttagtable_lookup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextTagTable, remove,               arginfo_gtk_gtktexttagtable_remove, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToggleAction, __construct)
{
	GParameter params[4];
	zval *php_args[5] = { NULL, };
	char *prop_names[] = { "name", "label", "tooltip", "stock_id",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u^u^u^u", &php_args[0], &php_args[1], &php_args[2], &php_args[3], &php_args[4], &php_args[5], &php_args[6], &php_args[7])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleAction);
	}

	memset(&params, 0, 4 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleAction);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleAction);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToggleAction, get_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_action_get_active(GTK_TOGGLE_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleAction, get_draw_as_radio)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_action_get_draw_as_radio(GTK_TOGGLE_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleAction, set_active)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


static PHP_METHOD(GtkToggleAction, set_draw_as_radio)
{
	zend_bool draw_as_radio;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &draw_as_radio))
		return;

    gtk_toggle_action_set_draw_as_radio(GTK_TOGGLE_ACTION(PHPG_GOBJECT(this_ptr)), (gboolean)draw_as_radio);

}


static PHP_METHOD(GtkToggleAction, toggled)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toggle_action_toggled(GTK_TOGGLE_ACTION(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoggleaction_gtk_toggle_action_new, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, tooltip)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoggleaction_set_active, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoggleaction_set_draw_as_radio, 0)
    ZEND_ARG_INFO(0, draw_as_radio)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoggleaction_methods[] = {
	PHP_ME(GtkToggleAction, __construct,          arginfo_gtk_gtktoggleaction_gtk_toggle_action_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleAction, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleAction, get_draw_as_radio,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleAction, set_active,           arginfo_gtk_gtktoggleaction_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleAction, set_draw_as_radio,    arginfo_gtk_gtktoggleaction_set_draw_as_radio, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleAction, toggled,              NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRadioAction, __construct)
{
	GParameter params[5];
	zval *php_args[6] = { NULL, };
	char *prop_names[] = { "name", "label", "tooltip", "stock_id", "value",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u^u^u^u^i", &php_args[0], &php_args[1], &php_args[2], &php_args[3], &php_args[4], &php_args[5], &php_args[6], &php_args[7], &php_args[8])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioAction);
	}

	memset(&params, 0, 5 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioAction);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioAction);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkRadioAction, get_current_value)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_radio_action_get_current_value(GTK_RADIO_ACTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 4775 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioAction, get_group)
{
    GSList *group, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    group = gtk_radio_action_get_group(GTK_RADIO_ACTION(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = group; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}

#line 22811 "ext/gtk+/gen_gtk.c"


#line 4797 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioAction, set_group)
{
    zval *php_action = NULL;
    GSList *list;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_action, gtkradioaction_ce))
        return;

    list = gtk_radio_action_get_group(GTK_RADIO_ACTION(PHPG_GOBJECT(php_action)));
    if (list == gtk_radio_action_get_group(GTK_RADIO_ACTION(PHPG_GOBJECT(this_ptr)))) {
        return;
    }
    gtk_radio_action_set_group(GTK_RADIO_ACTION(PHPG_GOBJECT(this_ptr)), list);
}

#line 22832 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRadioAction, set_current_value)
{
	long current_value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &current_value))
		return;

    gtk_radio_action_set_current_value(GTK_RADIO_ACTION(PHPG_GOBJECT(this_ptr)), (gint)current_value);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkradioaction_gtk_radio_action_new, 0)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, tooltip)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkradioaction_set_current_value, 0)
    ZEND_ARG_INFO(0, current_value)
ZEND_END_ARG_INFO();

static zend_function_entry gtkradioaction_methods[] = {
	PHP_ME(GtkRadioAction, __construct,          arginfo_gtk_gtkradioaction_gtk_radio_action_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioAction, get_current_value,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioAction, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioAction, set_current_value,    arginfo_gtk_gtkradioaction_set_current_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioAction, set_group,            NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTooltips, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTooltips);
	}


	phpg_warn_deprecated("since GTK+ 2.12, please use new GtkTooltip API" TSRMLS_CC);

	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTooltips);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkTooltips, tip_window)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->tip_window;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTooltips, tip_label)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->tip_label;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}

#line 6540 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTooltips, active_tips_data)
{
	GtkTooltips *tooltips;
	GtkTooltipsData *data;
    zval *tool_widget, *widget;

	tooltips = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj);
    data = (GtkTooltipsData*)tooltips->active_tips_data;

    if (data == NULL) {
        RETVAL_NULL();
    } else {
        phpg_gobject_new(&tool_widget, (GObject *) data->tooltips TSRMLS_CC);
        phpg_gobject_new(&widget, (GObject *) data->widget TSRMLS_CC);
        php_gtk_build_value(&return_value, "(NNss)", tool_widget, widget, data->tip_text, data->tip_private);
    }

	return SUCCESS;
}

#line 22947 "ext/gtk+/gen_gtk.c"


#line 6516 "ext/gtk+/gtk.overrides"
PHPG_PROP_READER(GtkTooltips, tips_data_list)
{
	GtkTooltips *tooltips;
	GList *data, *current;

	tooltips = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj);
    array_init(return_value);
    
    data = tooltips->tips_data_list;
    for (current = data; current; current = current->next) {
		zval *data_array, *item, *tool_widget, *widget;
		GtkTooltipsData *tooltip = (GtkTooltipsData*)current->data;
		MAKE_STD_ZVAL(data_array);
		array_init(data_array);
		phpg_gobject_new(&tool_widget, (GObject *) tooltip->tooltips TSRMLS_CC);
		phpg_gobject_new(&widget, (GObject *) tooltip->widget TSRMLS_CC);
		php_gtk_build_value(&item, "(NNss)", tool_widget, widget, tooltip->tip_text, tooltip->tip_private);
        add_next_index_zval(return_value, data_array);
    }
	return SUCCESS;
}

#line 22973 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkTooltips, delay)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->delay;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTooltips, enabled)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->enabled;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTooltips, use_sticky_delay)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->use_sticky_delay;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkTooltips, timer_tag)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOOLTIPS(((phpg_gobject_t *)object)->obj)->timer_tag;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtktooltips_prop_info[] = {
	{ "tip_window", PHPG_PROP_READ_FN(GtkTooltips, tip_window), NULL },
	{ "tip_label", PHPG_PROP_READ_FN(GtkTooltips, tip_label), NULL },
	{ "active_tips_data", PHPG_PROP_READ_FN(GtkTooltips, active_tips_data), NULL },
	{ "tips_data_list", PHPG_PROP_READ_FN(GtkTooltips, tips_data_list), NULL },
	{ "delay", PHPG_PROP_READ_FN(GtkTooltips, delay), NULL },
	{ "enabled", PHPG_PROP_READ_FN(GtkTooltips, enabled), NULL },
	{ "use_sticky_delay", PHPG_PROP_READ_FN(GtkTooltips, use_sticky_delay), NULL },
	{ "timer_tag", PHPG_PROP_READ_FN(GtkTooltips, timer_tag), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkTooltips, disable)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use new GtkTooltip API" TSRMLS_CC);

    gtk_tooltips_disable(GTK_TOOLTIPS(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTooltips, enable)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use new GtkTooltip API" TSRMLS_CC);

    gtk_tooltips_enable(GTK_TOOLTIPS(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTooltips, force_window)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tooltips_force_window(GTK_TOOLTIPS(PHPG_GOBJECT(this_ptr)));

}

#line 6489 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkTooltips, get_info_from_tip_window)
{
    zval *window;
    GtkTooltips *tooltips;
    GtkWidget *widget;
    gboolean ret;
    zval *php_tooltips = NULL, *php_widget = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gtkwindow_ce)) {
        return;
    }

    ret = gtk_tooltips_get_info_from_tip_window(GTK_WINDOW(PHPG_GOBJECT(window)), &tooltips, &widget);
    if (ret) {
        phpg_gobject_new(&php_tooltips, (GObject *) tooltips TSRMLS_CC);
        phpg_gobject_new(&php_widget, (GObject *) widget TSRMLS_CC);
        php_gtk_build_value(&return_value, "(NN)", tooltips, widget);
    }

    RETURN_FALSE;
}
#line 23110 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTooltips, set_delay)
{
	long delay;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &delay))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use new GtkTooltip API" TSRMLS_CC);

    gtk_tooltips_set_delay(GTK_TOOLTIPS(PHPG_GOBJECT(this_ptr)), (guint)delay);

}


static PHP_METHOD(GtkTooltips, set_tip)
{
	zval *widget;
	char *tip_text, *tip_private = NULL;
	zend_bool free_tip_text = FALSE, free_tip_private = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou|u", &widget, gtkwidget_ce, &tip_text, &free_tip_text, &tip_private, &free_tip_private))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use new GtkTooltip API" TSRMLS_CC);

    gtk_tooltips_set_tip(GTK_TOOLTIPS(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), tip_text, tip_private);
	if (free_tip_text) g_free(tip_text);
	if (free_tip_private) g_free(tip_private);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktooltips_set_delay, 0)
    ZEND_ARG_INFO(0, delay)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktooltips_set_tip, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, tip_text)
    ZEND_ARG_INFO(0, tip_private)
ZEND_END_ARG_INFO();

static zend_function_entry gtktooltips_methods[] = {
	PHP_ME(GtkTooltips, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltips, disable,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltips, enable,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltips, force_window,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltips, get_info_from_tip_window, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkTooltips, set_delay,            arginfo_gtk_gtktooltips_set_delay, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTooltips, set_tip,              arginfo_gtk_gtktooltips_set_tip, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1579 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModelFilter, __construct)
{
    GObject *wrapped_obj;
    GtkTreePath *path = NULL;
    zval *php_model = NULL, *php_path = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|V", &php_model, gtktreemodel_ce, &php_path))
        return;

    if (php_path && phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path argument to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC),
                               "child-model", GTK_TREE_MODEL(PHPG_GOBJECT(php_model)),
                               "virtual-root", path,
                               NULL);

    if (path) {
        gtk_tree_path_free(path);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeModelFilter);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 23203 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModelFilter, clear_cache)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_model_filter_clear_cache(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)));

}

#line 1618 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModelFilter, convert_child_iter_to_iter)
{
    GtkTreeIter filter_iter, *child_iter = NULL;
    zval *php_child_iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_child_iter, gboxed_ce))
        return;

    if (phpg_gboxed_check(php_child_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        child_iter = (GtkTreeIter *) PHPG_GBOXED(php_child_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects child_iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_filter_convert_child_iter_to_iter(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)), &filter_iter, child_iter);

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &filter_iter, TRUE, TRUE TSRMLS_CC);
}

#line 23243 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModelFilter, convert_child_path_to_path)
{
	GtkTreePath *child_path, *php_retval;
	zval *php_child_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_child_path))
		return;

    if (phpg_tree_path_from_zval(php_child_path, &child_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects child_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_filter_convert_child_path_to_path(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)), child_path);

    if (child_path)
        gtk_tree_path_free(child_path);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
    }

}

#line 1650 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModelFilter, convert_iter_to_child_iter)
{
    GtkTreeIter child_iter, *filter_iter = NULL;
    zval *php_filter_iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_filter_iter, gboxed_ce))
        return;

    if (phpg_gboxed_check(php_filter_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        filter_iter = (GtkTreeIter *) PHPG_GBOXED(php_filter_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects filter_iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_filter_convert_iter_to_child_iter(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)), &child_iter, filter_iter);

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &child_iter, TRUE, TRUE TSRMLS_CC);
}

#line 23297 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModelFilter, convert_path_to_child_path)
{
	GtkTreePath *filter_path, *php_retval;
	zval *php_filter_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_filter_path))
		return;

    if (phpg_tree_path_from_zval(php_filter_path, &filter_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects filter_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_filter_convert_path_to_child_path(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)), filter_path);

    if (filter_path)
        gtk_tree_path_free(filter_path);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
    }

}


static PHP_METHOD(GtkTreeModelFilter, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeModelFilter, refilter)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeModelFilter, set_visible_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_tree_model_filter_set_visible_column(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)), (gint)column);

}

#line 1682 "ext/gtk+/gtktreeview.overrides"
static gboolean phpg_tree_model_filter_visible_func_marshal(GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_iter = NULL;
    gboolean show = TRUE;
    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_model;
    args[1] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        show = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        show = TRUE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return show;
}

static PHP_METHOD(GtkTreeModelFilter, set_visible_func)
{
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &callback))
        return;

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(PHPG_GOBJECT(this_ptr)),
                                                           (GtkTreeModelFilterVisibleFunc)phpg_tree_model_filter_visible_func_marshal,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 23431 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreemodelfilter_gtk_tree_model_filter_new, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, model, GtkTreeModel, 1)
    ZEND_ARG_INFO(0, root)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_convert_child_iter_to_iter, 0)
    ZEND_ARG_OBJ_INFO(0, child_iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_convert_child_path_to_path, 0)
    ZEND_ARG_INFO(0, child_path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_convert_iter_to_child_iter, 0)
    ZEND_ARG_OBJ_INFO(0, filter_iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_convert_path_to_child_path, 0)
    ZEND_ARG_INFO(0, filter_path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_set_visible_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelfilter_set_visible_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


static zend_function_entry gtktreemodelfilter_methods[] = {
	PHP_ME(GtkTreeModelFilter, __construct,          arginfo_gtk_gtktreemodelfilter_gtk_tree_model_filter_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, clear_cache,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, convert_child_iter_to_iter, arginfo_gtk_gtktreemodelfilter_convert_child_iter_to_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, convert_child_path_to_path, arginfo_gtk_gtktreemodelfilter_convert_child_path_to_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, convert_iter_to_child_iter, arginfo_gtk_gtktreemodelfilter_convert_iter_to_child_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, convert_path_to_child_path, arginfo_gtk_gtktreemodelfilter_convert_path_to_child_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, refilter,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, set_visible_column,   arginfo_gtk_gtktreemodelfilter_set_visible_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelFilter, set_visible_func,     arginfo_gtk_gtktreemodelfilter_set_visible_func, ZEND_ACC_PUBLIC)

	/***   GtkTreeModel interface implementations   ***/

	PHP_MALIAS(GtkTreeModel, get, get, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_column_type, get_column_type, arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_flags, get_flags, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter, get_iter, arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_first, get_iter_first, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_root, get_iter_root, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_from_string, get_iter_from_string, arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_n_columns, get_n_columns, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_path, get_path, arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_string_from_iter, get_string_from_iter, arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_value, get_value, arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_children, iter_children, arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_has_child, iter_has_child, arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_n_children, iter_n_children, arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_next, iter_next, arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_nth_child, iter_nth_child, arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_parent, iter_parent, arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, ref_node, ref_node, arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_changed, row_changed, arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_deleted, row_deleted, arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_has_child_toggled, row_has_child_toggled, arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_inserted, row_inserted, arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, rows_reordered, rows_reordered, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, unref_node, unref_node, arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, for_each, for_each, arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC)

	/***   GtkTreeDragSource interface implementations   ***/

	PHP_MALIAS(GtkTreeDragSource, drag_data_delete, drag_data_delete, arginfo_gtk_gtktreedragsource_drag_data_delete, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, drag_data_get, drag_data_get, arginfo_gtk_gtktreedragsource_drag_data_get, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, row_draggable, row_draggable, arginfo_gtk_gtktreedragsource_row_draggable, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeModelSort, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "model",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^O", &php_args[0], gtktreemodel_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeModelSort);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeModelSort);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeModelSort);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTreeModelSort, clear_cache)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_model_sort_clear_cache(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)));

}

#line 1757 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModelSort, convert_child_iter_to_iter)
{
    GtkTreeIter sort_iter, *child_iter = NULL;
    zval *php_child_iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_child_iter, gboxed_ce))
        return;

    if (phpg_gboxed_check(php_child_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        child_iter = (GtkTreeIter *) PHPG_GBOXED(php_child_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects child_iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_sort_convert_child_iter_to_iter(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)), &sort_iter, child_iter);

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &sort_iter, TRUE, TRUE TSRMLS_CC);
}

#line 23591 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModelSort, convert_child_path_to_path)
{
	GtkTreePath *child_path, *php_retval;
	zval *php_child_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_child_path))
		return;

    if (phpg_tree_path_from_zval(php_child_path, &child_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects child_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_sort_convert_child_path_to_path(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)), child_path);

    if (child_path)
        gtk_tree_path_free(child_path);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
        gtk_tree_path_free(php_retval);
    }

}

#line 1789 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeModelSort, convert_iter_to_child_iter)
{
    GtkTreeIter child_iter, *sort_iter = NULL;
    zval *php_sort_iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_sort_iter, gboxed_ce))
        return;

    if (phpg_gboxed_check(php_sort_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        sort_iter = (GtkTreeIter *) PHPG_GBOXED(php_sort_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects sort_iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_model_sort_convert_iter_to_child_iter(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)), &child_iter, sort_iter);

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &child_iter, TRUE, TRUE TSRMLS_CC);
}

#line 23646 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeModelSort, convert_path_to_child_path)
{
	GtkTreePath *sorted_path, *php_retval;
	zval *php_sorted_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sorted_path))
		return;

    if (phpg_tree_path_from_zval(php_sorted_path, &sorted_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects sorted_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_sort_convert_path_to_child_path(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)), sorted_path);

    if (sorted_path)
        gtk_tree_path_free(sorted_path);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
        gtk_tree_path_free(php_retval);
    }

}


static PHP_METHOD(GtkTreeModelSort, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeModelSort, iter_is_valid)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_model_sort_iter_is_valid(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeModelSort, reset_default_sort_func)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_model_sort_reset_default_sort_func(GTK_TREE_MODEL_SORT(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_gtk_tree_model_sort_new_with_model, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_convert_child_iter_to_iter, 0)
    ZEND_ARG_OBJ_INFO(0, child_iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_convert_child_path_to_path, 0)
    ZEND_ARG_INFO(0, child_path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_convert_iter_to_child_iter, 0)
    ZEND_ARG_OBJ_INFO(0, sort_iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_convert_path_to_child_path, 0)
    ZEND_ARG_INFO(0, sorted_path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreemodelsort_iter_is_valid, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreemodelsort_methods[] = {
	PHP_ME(GtkTreeModelSort, __construct,          arginfo_gtk_gtktreemodelsort_gtk_tree_model_sort_new_with_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, clear_cache,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, convert_child_iter_to_iter, arginfo_gtk_gtktreemodelsort_convert_child_iter_to_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, convert_child_path_to_path, arginfo_gtk_gtktreemodelsort_convert_child_path_to_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, convert_iter_to_child_iter, arginfo_gtk_gtktreemodelsort_convert_iter_to_child_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, convert_path_to_child_path, arginfo_gtk_gtktreemodelsort_convert_path_to_child_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, iter_is_valid,        arginfo_gtk_gtktreemodelsort_iter_is_valid, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeModelSort, reset_default_sort_func, NULL, ZEND_ACC_PUBLIC)

	/***   GtkTreeModel interface implementations   ***/

	PHP_MALIAS(GtkTreeModel, get, get, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_column_type, get_column_type, arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_flags, get_flags, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter, get_iter, arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_first, get_iter_first, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_root, get_iter_root, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_from_string, get_iter_from_string, arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_n_columns, get_n_columns, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_path, get_path, arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_string_from_iter, get_string_from_iter, arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_value, get_value, arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_children, iter_children, arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_has_child, iter_has_child, arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_n_children, iter_n_children, arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_next, iter_next, arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_nth_child, iter_nth_child, arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_parent, iter_parent, arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, ref_node, ref_node, arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_changed, row_changed, arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_deleted, row_deleted, arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_has_child_toggled, row_has_child_toggled, arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_inserted, row_inserted, arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, rows_reordered, rows_reordered, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, unref_node, unref_node, arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, for_each, for_each, arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC)

	/***   GtkTreeSortable interface implementations   ***/

	PHP_MALIAS(GtkTreeSortable, get_sort_column_id, get_sort_column_id, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, has_default_sort_func, has_default_sort_func, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_default_sort_func, set_default_sort_func, arginfo_gtk_gtktreesortable_set_default_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_column_id, set_sort_column_id, arginfo_gtk_gtktreesortable_set_sort_column_id, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_func, set_sort_func, arginfo_gtk_gtktreesortable_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, sort_column_changed, sort_column_changed, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeSelection, count_selected_rows)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_selection_count_selected_rows(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeSelection, get_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_selection_get_mode(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 1818 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeSelection, get_selected)
{
    GtkTreeSelection *sel;
    GtkTreeIter iter;
    GtkTreeModel *model = NULL;
    zval *php_iter = NULL, *php_model = NULL;
    int result;

    NOT_STATIC_METHOD();

    sel = GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr));
    if (gtk_tree_selection_get_mode(sel) == GTK_SELECTION_MULTIPLE) {
        php_error(E_WARNING, "%s::%s() cannot be used because the selection mode is Gtk::SELECTION_MULTIPLE",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    result = gtk_tree_selection_get_selected(sel, &model, &iter);
    phpg_gobject_new(&php_model, (GObject *)model TSRMLS_CC);
    if (result) {
        phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
        php_gtk_build_value(&return_value, "(NN)", php_model, php_iter);
    } else {
        php_gtk_build_value(&return_value, "(Nn)", php_model);
    }
}

#line 23867 "ext/gtk+/gen_gtk.c"


#line 1850 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeSelection, get_selected_rows)
{
    GtkTreeModel *model = NULL;
    GList *selected, *tmp;
    zval *php_selected = NULL, *php_model = NULL;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    selected = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), &model);
    phpg_gobject_new(&php_model, (GObject *)model TSRMLS_CC);
    if (selected) {
        MAKE_STD_ZVAL(php_selected);
        array_init(php_selected);
        for (tmp = selected; tmp != NULL; tmp = tmp->next) {
            GtkTreePath *path = tmp->data;
            zval *item = NULL;

            phpg_tree_path_to_zval(path, &item TSRMLS_CC);
            add_next_index_zval(php_selected, item);
            gtk_tree_path_free(path);
        }
        php_gtk_build_value(&return_value, "(NN)", php_model, php_selected);
        g_list_free(selected);
    } else {
        php_gtk_build_value(&return_value, "(Nn)", php_model);
    }
}

#line 23901 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSelection, get_tree_view)
{
	GtkTreeView* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_selection_get_tree_view(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeSelection, iter_is_selected)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_selection_iter_is_selected(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeSelection, path_is_selected)
{
	GtkTreePath *path;
	zval *php_path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_selection_path_is_selected(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeSelection, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_selection_select_all(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeSelection, select_iter)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_select_iter(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), iter);

}


static PHP_METHOD(GtkTreeSelection, select_path)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_select_path(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeSelection, select_range)
{
	GtkTreePath *start_path, *end_path;
	zval *php_start_path, *php_end_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_start_path, &php_end_path))
		return;

    if (phpg_tree_path_from_zval(php_start_path, &start_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects start_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_tree_path_from_zval(php_end_path, &end_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects end_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_select_range(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), start_path, end_path);

    if (start_path)
        gtk_tree_path_free(start_path);

    if (end_path)
        gtk_tree_path_free(end_path);

}

#line 1889 "ext/gtk+/gtktreeview.overrides"
static void phpg_tree_selection_foreach_func_marshal (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_path = NULL, *php_iter = NULL;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 3, &n_args);
    args[0] = &php_model;
    args[1] = &php_path;
    args[2] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_path);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkTreeSelection, selected_foreach)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeSelectionForeachFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    zval_add_ref(&php_callback);
    cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
    callback = (GtkTreeSelectionForeachFunc)phpg_tree_selection_foreach_func_marshal;

    gtk_tree_selection_selected_foreach(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), callback, cb_data);
    phpg_cb_data_destroy(cb_data);
}

#line 24120 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSelection, set_mode)
{
	GtkSelectionMode type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SELECTION_MODE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_tree_selection_set_mode(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), type);

}

#line 1961 "ext/gtk+/gtktreeview.overrides"
static gboolean phpg_tree_selection_func_marshal(GtkTreeSelection *selection,
                                                 GtkTreeModel *model,
                                                 GtkTreePath *path,
                                                 gboolean path_currently_selected,
                                                 gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_selection = NULL, *php_model = NULL, *php_path = NULL, *php_path_currently_selected;
    gboolean allow_toggle = TRUE;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_selection,   (GObject*)selection   TSRMLS_CC);
    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_tree_path_to_zval(path,       &php_path             TSRMLS_CC);
    MAKE_STD_ZVAL(php_path_currently_selected);
    ZVAL_BOOL(php_path_currently_selected, path_currently_selected);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 4, &n_args);
    args[0] = &php_selection;
    args[1] = &php_model;
    args[2] = &php_path;
    args[3] = &php_path_currently_selected;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_selection);
    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_path);
    zval_ptr_dtor(&php_path_currently_selected);

    if (retval) {
        allow_toggle = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        allow_toggle = TRUE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return allow_toggle;
}

static PHP_METHOD(GtkTreeSelection, set_select_function)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeSelectionFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTreeSelectionFunc)phpg_tree_selection_func_marshal;
        }
    }

    gtk_tree_selection_set_select_function(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 24227 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeSelection, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_selection_unselect_all(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeSelection, unselect_iter)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_unselect_iter(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), iter);

}


static PHP_METHOD(GtkTreeSelection, unselect_path)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_unselect_path(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeSelection, unselect_range)
{
	GtkTreePath *start_path, *end_path;
	zval *php_start_path, *php_end_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_start_path, &php_end_path))
		return;

    if (phpg_tree_path_from_zval(php_start_path, &start_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects start_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_tree_path_from_zval(php_end_path, &end_path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects end_path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_selection_unselect_range(GTK_TREE_SELECTION(PHPG_GOBJECT(this_ptr)), start_path, end_path);

    if (start_path)
        gtk_tree_path_free(start_path);

    if (end_path)
        gtk_tree_path_free(end_path);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_iter_is_selected, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_path_is_selected, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_select_iter, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_select_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_select_range, 0)
    ZEND_ARG_INFO(0, start_path)
    ZEND_ARG_INFO(0, end_path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_selected_foreach, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_set_mode, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_set_select_function, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_unselect_iter, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_unselect_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeselection_unselect_range, 0)
    ZEND_ARG_INFO(0, start_path)
    ZEND_ARG_INFO(0, end_path)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreeselection_methods[] = {
	PHP_ME(GtkTreeSelection, count_selected_rows,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, get_mode,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, get_selected,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, get_selected_rows,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, get_tree_view,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, iter_is_selected,     arginfo_gtk_gtktreeselection_iter_is_selected, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, path_is_selected,     arginfo_gtk_gtktreeselection_path_is_selected, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, select_all,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, select_iter,          arginfo_gtk_gtktreeselection_select_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, select_path,          arginfo_gtk_gtktreeselection_select_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, select_range,         arginfo_gtk_gtktreeselection_select_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, selected_foreach,     arginfo_gtk_gtktreeselection_selected_foreach, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, set_mode,             arginfo_gtk_gtktreeselection_set_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, set_select_function,  arginfo_gtk_gtktreeselection_set_select_function, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, unselect_all,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, unselect_iter,        arginfo_gtk_gtktreeselection_unselect_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, unselect_path,        arginfo_gtk_gtktreeselection_unselect_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeSelection, unselect_range,       arginfo_gtk_gtktreeselection_unselect_range, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 765 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, __construct)
{
    zval ***args;
    GType *column_types;
    GObject *wrapped_obj;
    int i, argc = ZEND_NUM_ARGS();

    if (argc > 0) {
        args = php_gtk_func_args(argc);
        column_types = emalloc(argc * sizeof(GType));

        for (i = 0; i < argc; i++) {
            column_types[i] = phpg_gtype_from_zval(*args[i] TSRMLS_CC);
            if (column_types[i] == 0) {
                efree(column_types);
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeStore);
            }
        }

        wrapped_obj = g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);
        gtk_tree_store_set_column_types(GTK_TREE_STORE(wrapped_obj), argc, column_types);

        efree(column_types);
        efree(args);
    } else {
        wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeStore);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 24435 "ext/gtk+/gen_gtk.c"


#line 1103 "ext/gtk+/gtktreeview.overrides"

static zval* phpg_gtktreestore_read_dimension_handler(zval *object, zval *offset, int type TSRMLS_DC)
{
    return phpg_gtktreemodel_read_dimension_handler(object, offset, type TSRMLS_CC);
}

#line 24445 "ext/gtk+/gen_gtk.c"


#line 1111 "ext/gtk+/gtktreeview.overrides"

static void phpg_gtktreestore_write_dimension_handler(zval *object, zval *offset, zval *value TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, value TSRMLS_CC);
}

#line 24455 "ext/gtk+/gen_gtk.c"


#line 1095 "ext/gtk+/gtktreeview.overrides"

static int phpg_gtktreestore_has_dimension_handler(zval *object, zval *offset, int check_empty TSRMLS_DC)
{
    return phpg_gtktreemodel_has_dimension_handler(object, offset, check_empty TSRMLS_CC);
}

#line 24465 "ext/gtk+/gen_gtk.c"


#line 1119 "ext/gtk+/gtktreeview.overrides"

static void phpg_gtktreestore_unset_dimension_handler(zval *object, zval *offset TSRMLS_DC)
{
    phpg_gtktreemodel_write_dimension_handler(object, offset, NULL TSRMLS_CC);
}

#line 24475 "ext/gtk+/gen_gtk.c"


#line 1086 "ext/gtk+/gtktreeview.overrides"

static int phpg_gtktreestore_count_elements_handler(zval *object, long *count TSRMLS_DC)
{
    return phpg_gtktreemodel_count_elements_handler(object, count TSRMLS_CC);
}


#line 24486 "ext/gtk+/gen_gtk.c"



static zend_object_handlers gtktreestore_object_handlers;

static zend_object_value phpg_create_gtktreestore(zend_class_entry *ce TSRMLS_DC)
{
    zend_object_value zov;

    zov = phpg_create_gobject(ce TSRMLS_CC);
    zov.handlers = &gtktreestore_object_handlers;

    return zov;
}
#line 849 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, append)
{
    GtkTreeIter iter;
    zval *items = NULL;
    zval *php_parent_iter = NULL;
    GtkTreeIter *parent_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Na", &php_parent_iter, gtktreeiter_ce, &items))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }

    gtk_tree_store_append(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), &iter, parent_iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}


#line 24527 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeStore, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_store_clear(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)));

}

#line 885 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, insert)
{
    GtkTreeIter iter;
    gint position;
    zval *items = NULL;
    zval *php_parent_iter = NULL;
    GtkTreeIter *parent_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i|Na", &position, &php_parent_iter, gtktreeiter_ce, &items))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }

    gtk_tree_store_insert(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), &iter, parent_iter, position);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}


#line 24570 "ext/gtk+/gen_gtk.c"


#line 961 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, insert_after)
{
    GtkTreeIter iter;
    zval *items = NULL;
    zval *php_parent_iter = NULL, *php_sibling_iter = NULL;
    GtkTreeIter *parent_iter = NULL, *sibling_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|NNa", &php_parent_iter, gtktreeiter_ce, &php_sibling_iter, gtktreeiter_ce, &items))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }
    if (php_sibling_iter && IS_OBJECT == Z_TYPE_P(php_sibling_iter)) {
        sibling_iter = (GtkTreeIter *) PHPG_GBOXED(php_sibling_iter);
    }

    gtk_tree_store_insert_after(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), &iter, parent_iter, sibling_iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}


#line 24602 "ext/gtk+/gen_gtk.c"


#line 922 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, insert_before)
{
    GtkTreeIter iter;
    zval *items = NULL;
    zval *php_parent_iter = NULL, *php_sibling_iter = NULL;
    GtkTreeIter *parent_iter = NULL, *sibling_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|NNa", &php_parent_iter, gtktreeiter_ce, &php_sibling_iter, gtktreeiter_ce, &items))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }
    if (php_sibling_iter && IS_OBJECT == Z_TYPE_P(php_sibling_iter)) {
        sibling_iter = (GtkTreeIter *) PHPG_GBOXED(php_sibling_iter);
    }

    gtk_tree_store_insert_before(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), &iter, parent_iter, sibling_iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}


#line 24634 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeStore, is_ancestor)
{
	GtkTreeIter *iter = NULL, *descendant = NULL;
	zval *php_iter, *php_descendant;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_iter, gboxed_ce, &php_descendant, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_descendant, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        descendant = (GtkTreeIter *) PHPG_GBOXED(php_descendant);
    } else {
        php_error(E_WARNING, "%s::%s() expects descendant argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_store_is_ancestor(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter, descendant);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeStore, iter_depth)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_store_iter_depth(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeStore, iter_is_valid)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_store_iter_is_valid(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeStore, move_after)
{
	GtkTreeIter *iter = NULL, *position = NULL;
	zval *php_iter, *php_position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &php_iter, gboxed_ce, &php_position, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_position) != IS_NULL) {
        if (phpg_gboxed_check(php_position, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
            position = (GtkTreeIter *) PHPG_GBOXED(php_position);
        } else {
            php_error(E_WARNING, "%s::%s() expects position argument to be a valid GtkTreeIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_tree_store_move_after(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter, position);

}


static PHP_METHOD(GtkTreeStore, move_before)
{
	GtkTreeIter *iter = NULL, *position = NULL;
	zval *php_iter, *php_position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &php_iter, gboxed_ce, &php_position, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_position) != IS_NULL) {
        if (phpg_gboxed_check(php_position, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
            position = (GtkTreeIter *) PHPG_GBOXED(php_position);
        } else {
            php_error(E_WARNING, "%s::%s() expects position argument to be a valid GtkTreeIter object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_tree_store_move_before(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter, position);

}

#line 999 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, prepend)
{
    GtkTreeIter iter;
    zval *items = NULL;
    zval *php_parent_iter = NULL;
    GtkTreeIter *parent_iter = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Na", &php_parent_iter, gtktreeiter_ce, &items))
        return;

    if (php_parent_iter && IS_OBJECT == Z_TYPE_P(php_parent_iter)) {
        parent_iter = (GtkTreeIter *) PHPG_GBOXED(php_parent_iter);
    }

    gtk_tree_store_prepend(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), &iter, parent_iter);

    if (items && phpg_model_set_row(GTK_TREE_MODEL(PHPG_GOBJECT(this_ptr)), &iter, items TSRMLS_CC) == FAILURE)
        return;

    phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}

#line 24806 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeStore, remove)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_store_remove(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}

#line 1043 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, reorder)
{
    zval *php_parent, *php_order, **elem;
    GtkTreeStore *store;
    GtkTreeIter *parent = NULL;
    gint *order;
    int i, n_elems, n_children;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O!a/", &php_parent, gtktreeiter_ce, &php_order))
        return;

    store = GTK_TREE_STORE(PHPG_GOBJECT(this_ptr));
    n_children = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(store), parent);
    n_elems = zend_hash_num_elements(Z_ARRVAL_P(php_order));

    if (n_elems != n_children) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "new order array size (%d) not the same as the number of children of parent iterator (%d)", n_elems, n_children);
        return;
    }

    order = safe_emalloc(n_children, sizeof(gint), 0);
    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_order)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_order), (void **) &elem) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_order)), i++) {

        convert_to_long(*elem);
        if (Z_LVAL_PP(elem) < 0 || Z_LVAL_PP(elem) >= n_children) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING, "position index out of range 0-%d", n_children);
            efree(order);
            return;
        }
        order[i] = Z_LVAL_PP(elem);
    }

    gtk_tree_store_reorder(store, parent, order);
    efree(order);
}


#line 24875 "ext/gtk+/gen_gtk.c"


#line 1036 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, set)
{
    phpg_store_set_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
}

#line 24884 "ext/gtk+/gen_gtk.c"


#line 802 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeStore, set_column_types)
{
    zval *php_types, **item;
    GType *column_types;
    int i, n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_types)) {
        return;
    }

    n = zend_hash_num_elements(Z_ARRVAL_P(php_types));

    if (n == 0) {
        php_error(E_WARNING, "number of columns has to be > 0");
        return;
    }

    column_types = safe_emalloc(n, sizeof(GType), 0);
    for (i = 0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_types));
         zend_hash_get_current_data(Z_ARRVAL_P(php_types), (void**)&item) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_types)), i++) {

        column_types[i] = phpg_gtype_from_zval(*item TSRMLS_CC);
        if (column_types[i] == 0) {
            efree(column_types);
            php_error(E_WARNING, "could not set column types for GtkTreeStore");
            return;
        }
    }

    gtk_tree_store_set_column_types(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), i, column_types);

    efree(column_types);
}

#line 24925 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeStore, swap)
{
	GtkTreeIter *a = NULL, *b = NULL;
	zval *php_a, *php_b;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_a, gboxed_ce, &php_b, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_a, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        a = (GtkTreeIter *) PHPG_GBOXED(php_a);
    } else {
        php_error(E_WARNING, "%s::%s() expects a argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_b, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        b = (GtkTreeIter *) PHPG_GBOXED(php_b);
    } else {
        php_error(E_WARNING, "%s::%s() expects b argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_store_swap(GTK_TREE_STORE(PHPG_GOBJECT(this_ptr)), a, b);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_gtk_tree_store_new, 0, 0, 1)
    ZEND_ARG_INFO(0, type_col_0)
    ZEND_ARG_INFO(0, type_col_1)
    ZEND_ARG_INFO(0, ...)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_append, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_insert, 0, 0, 1)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_OBJ_INFO(0, parent, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_insert_after, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, sibling, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_insert_before, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, sibling, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_is_ancestor, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, descendant, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_iter_depth, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_iter_is_valid, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_move_after, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, position, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_move_before, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, position, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_prepend, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, items)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_remove, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreestore_set, 0, 0, 3)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreestore_swap, 0)
    ZEND_ARG_OBJ_INFO(0, a, GtkTreeIter, 1)
    ZEND_ARG_OBJ_INFO(0, b, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreestore_methods[] = {
	PHP_ME(GtkTreeStore, __construct,          arginfo_gtk_gtktreestore_gtk_tree_store_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, append,               arginfo_gtk_gtktreestore_append, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, clear,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, insert,               arginfo_gtk_gtktreestore_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, insert_after,         arginfo_gtk_gtktreestore_insert_after, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, insert_before,        arginfo_gtk_gtktreestore_insert_before, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, is_ancestor,          arginfo_gtk_gtktreestore_is_ancestor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, iter_depth,           arginfo_gtk_gtktreestore_iter_depth, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, iter_is_valid,        arginfo_gtk_gtktreestore_iter_is_valid, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, move_after,           arginfo_gtk_gtktreestore_move_after, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, move_before,          arginfo_gtk_gtktreestore_move_before, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, prepend,              arginfo_gtk_gtktreestore_prepend, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, remove,               arginfo_gtk_gtktreestore_remove, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, reorder,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, set,                  arginfo_gtk_gtktreestore_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, set_column_types,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeStore, swap,                 arginfo_gtk_gtktreestore_swap, ZEND_ACC_PUBLIC)

	/***   GtkTreeModel interface implementations   ***/

	PHP_MALIAS(GtkTreeModel, get, get, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_column_type, get_column_type, arginfo_gtk_gtktreemodel_get_column_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_flags, get_flags, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter, get_iter, arginfo_gtk_gtktreemodel_get_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_first, get_iter_first, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_root, get_iter_root, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_iter_from_string, get_iter_from_string, arginfo_gtk_gtktreemodel_get_iter_from_string, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_n_columns, get_n_columns, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_path, get_path, arginfo_gtk_gtktreemodel_get_path, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_string_from_iter, get_string_from_iter, arginfo_gtk_gtktreemodel_get_string_from_iter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, get_value, get_value, arginfo_gtk_gtktreemodel_get_value, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_children, iter_children, arginfo_gtk_gtktreemodel_iter_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_has_child, iter_has_child, arginfo_gtk_gtktreemodel_iter_has_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_n_children, iter_n_children, arginfo_gtk_gtktreemodel_iter_n_children, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_next, iter_next, arginfo_gtk_gtktreemodel_iter_next, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_nth_child, iter_nth_child, arginfo_gtk_gtktreemodel_iter_nth_child, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, iter_parent, iter_parent, arginfo_gtk_gtktreemodel_iter_parent, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, ref_node, ref_node, arginfo_gtk_gtktreemodel_ref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_changed, row_changed, arginfo_gtk_gtktreemodel_row_changed, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_deleted, row_deleted, arginfo_gtk_gtktreemodel_row_deleted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_has_child_toggled, row_has_child_toggled, arginfo_gtk_gtktreemodel_row_has_child_toggled, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, row_inserted, row_inserted, arginfo_gtk_gtktreemodel_row_inserted, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, rows_reordered, rows_reordered, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, unref_node, unref_node, arginfo_gtk_gtktreemodel_unref_node, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeModel, for_each, for_each, arginfo_gtk_gtktreemodel_for_each, ZEND_ACC_PUBLIC)

	/***   GtkTreeDragSource interface implementations   ***/

	PHP_MALIAS(GtkTreeDragSource, drag_data_delete, drag_data_delete, arginfo_gtk_gtktreedragsource_drag_data_delete, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, drag_data_get, drag_data_get, arginfo_gtk_gtktreedragsource_drag_data_get, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragSource, row_draggable, row_draggable, arginfo_gtk_gtktreedragsource_row_draggable, ZEND_ACC_PUBLIC)

	/***   GtkTreeDragDest interface implementations   ***/

	PHP_MALIAS(GtkTreeDragDest, drag_data_received, drag_data_received, arginfo_gtk_gtktreedragdest_drag_data_received, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeDragDest, row_drop_possible, row_drop_possible, arginfo_gtk_gtktreedragdest_row_drop_possible, ZEND_ACC_PUBLIC)

	/***   GtkTreeSortable interface implementations   ***/

	PHP_MALIAS(GtkTreeSortable, get_sort_column_id, get_sort_column_id, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, has_default_sort_func, has_default_sort_func, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_default_sort_func, set_default_sort_func, arginfo_gtk_gtktreesortable_set_default_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_column_id, set_sort_column_id, arginfo_gtk_gtktreesortable_set_sort_column_id, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, set_sort_func, set_sort_func, arginfo_gtk_gtktreesortable_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkTreeSortable, sort_column_changed, sort_column_changed, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 2921 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeViewColumn, __construct)
{
    char *title = NULL;
    zval *php_cell = NULL;
    GtkCellRenderer *cell = NULL;
    GtkTreeViewColumn *tvc = NULL;
    zval ***args = NULL;
    zend_bool free_title = FALSE;
    int i, argc = ZEND_NUM_ARGS();

    if (argc <= 2) {
        if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|uN", &title, &free_title, &php_cell, gtkcellrenderer_ce)) {
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeViewColumn);
        }
    } else {
        if (!php_gtk_parse_args(2, "|uN", &title, &free_title, &php_cell, gtkcellrenderer_ce)) {
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeViewColumn);
        }
    }

    tvc = GTK_TREE_VIEW_COLUMN(g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL));
    if (php_cell && Z_TYPE_P(php_cell) != IS_NULL) {
        cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));
    }

    if (title)
        gtk_tree_view_column_set_title(tvc, title);
    if (cell)
        gtk_tree_view_column_pack_start(tvc, cell, TRUE);

    if (argc > 2) {
        if (argc % 2) {
            php_error(E_WARNING, "%s::%s() requires arguments 3-n to be attribute/column pairs",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeViewColumn);
        }

        args = php_gtk_func_args(argc);
        for (i = 2; i < argc; i += 2) {
            zval *attr = *args[i];
            zval *column = *args[i+1];

            if (Z_TYPE_P(attr) != IS_STRING) {
                php_error(E_WARNING, "%s::%s() requires argument %d to be a string, %s given",
                          get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                          i, zend_zval_type_name(attr));
                efree(args);
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeViewColumn);
            }
            if (Z_TYPE_P(column) != IS_LONG) {
                php_error(E_WARNING, "%s::%s() requires argument %d to be an integer, %s given",
                          get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C),
                          i, zend_zval_type_name(column));
                efree(args);
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeViewColumn);
            }

            gtk_tree_view_column_add_attribute(tvc, cell, Z_STRVAL_P(attr), Z_LVAL_P(column));
        }
        efree(args);
    }

    phpg_gobject_set_wrapper(this_ptr, (GObject *)tvc TSRMLS_CC);
    if (free_title) g_free(title);
}



#line 25185 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeViewColumn, add_attribute)
{
	zval *cell_renderer;
	char *attribute;
	zend_bool free_attribute = FALSE;
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oui", &cell_renderer, gtkcellrenderer_ce, &attribute, &free_attribute, &column))
		return;

    gtk_tree_view_column_add_attribute(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell_renderer)), attribute, (gint)column);
	if (free_attribute) g_free(attribute);

}

#line 2998 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeViewColumn, cell_get_position)
{
    gint start_pos, width;
    zval *cellrenderer;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &cellrenderer, gtkcellrenderer_ce))
        return;

    if (gtk_tree_view_column_cell_get_position(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cellrenderer)), &start_pos, &width)) {
        php_gtk_build_value(&return_value, "(ii)", start_pos, width);
    } else {
        RETURN_FALSE;
    }
}


#line 25225 "ext/gtk+/gen_gtk.c"


#line 3018 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeViewColumn, cell_get_size)
{
    gint x_offset, y_offset, width, height;
    GdkRectangle cell_area;
    zval *php_cell_area = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_tree_view_column_cell_get_size(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), &cell_area, &x_offset, &y_offset, &width, &height);

    phpg_gboxed_new(&php_cell_area, GDK_TYPE_RECTANGLE, &cell_area, TRUE, TRUE TSRMLS_CC);

    php_gtk_build_value(&return_value, "(Niiii)", php_cell_area, x_offset, y_offset, width, height);
}


#line 25249 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeViewColumn, cell_is_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_cell_is_visible(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, cell_set_cell_data)
{
	zval *tree_model, *php_iter;
	GtkTreeIter *iter = NULL;
	zend_bool is_expander, is_expanded;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OObb", &tree_model, gtktreemodel_ce, &php_iter, gboxed_ce, &is_expander, &is_expanded))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_column_cell_set_cell_data(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_TREE_MODEL(PHPG_GOBJECT(tree_model)), iter, (gboolean)is_expander, (gboolean)is_expanded);

}


static PHP_METHOD(GtkTreeViewColumn, clear_attributes)
{
	zval *cell_renderer;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &cell_renderer, gtkcellrenderer_ce))
		return;

    gtk_tree_view_column_clear_attributes(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell_renderer)));

}


static PHP_METHOD(GtkTreeViewColumn, clicked)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_column_clicked(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeViewColumn, focus_cell)
{
	zval *cell;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &cell, gtkcellrenderer_ce))
		return;

    gtk_tree_view_column_focus_cell(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)));

}


static PHP_METHOD(GtkTreeViewColumn, get_alignment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_alignment(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}

#line 209 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkTreeViewColumn, get_cell_renderers)
{
    GList *list, *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    phpg_warn_deprecated("since GTK+ 2.18, please use GtkCellLayout::get_cells()" TSRMLS_CC);

    list = gtk_tree_view_column_get_cell_renderers(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (item = list; item; item = item->next) {
        zval *php_item = NULL;
        GtkWidget *renderer = item->data;
        phpg_gobject_new(&php_item, G_OBJECT(renderer) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }

    g_list_free(list);
}

#line 25370 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeViewColumn, get_clickable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_clickable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_expand)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_expand(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_fixed_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_fixed_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_max_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_max_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_min_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_min_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_reorderable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_reorderable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_resizable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_resizable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_sizing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_sizing(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_sort_column_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_sort_column_id(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_sort_indicator)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_sort_indicator(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_sort_order)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_sort_order(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_spacing(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_title(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkTreeViewColumn, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_visible(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, get_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_widget(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeViewColumn, get_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeViewColumn, pack_end)
{
	zval *cell;
	zend_bool expand = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|b", &cell, gtkcellrenderer_ce, &expand))
		return;

    gtk_tree_view_column_pack_end(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), (gboolean)expand);

}


static PHP_METHOD(GtkTreeViewColumn, pack_start)
{
	zval *cell;
	zend_bool expand = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|b", &cell, gtkcellrenderer_ce, &expand))
		return;

    gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), GTK_CELL_RENDERER(PHPG_GOBJECT(cell)), (gboolean)expand);

}


static PHP_METHOD(GtkTreeViewColumn, set_alignment)
{
	double xalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &xalign))
		return;

    gtk_tree_view_column_set_alignment(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (float)xalign);

}


static PHP_METHOD(GtkTreeViewColumn, set_clickable)
{
	zend_bool active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &active))
		return;

    gtk_tree_view_column_set_clickable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)active);

}


static PHP_METHOD(GtkTreeViewColumn, set_expand)
{
	zend_bool expand;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &expand))
		return;

    gtk_tree_view_column_set_expand(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)expand);

}


static PHP_METHOD(GtkTreeViewColumn, set_fixed_width)
{
	long fixed_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &fixed_width))
		return;

    gtk_tree_view_column_set_fixed_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gint)fixed_width);

}


static PHP_METHOD(GtkTreeViewColumn, set_max_width)
{
	long max_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &max_width))
		return;

    gtk_tree_view_column_set_max_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gint)max_width);

}


static PHP_METHOD(GtkTreeViewColumn, set_min_width)
{
	long min_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &min_width))
		return;

    gtk_tree_view_column_set_min_width(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gint)min_width);

}


static PHP_METHOD(GtkTreeViewColumn, set_reorderable)
{
	zend_bool reorderable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reorderable))
		return;

    gtk_tree_view_column_set_reorderable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)reorderable);

}


static PHP_METHOD(GtkTreeViewColumn, set_resizable)
{
	zend_bool resizable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &resizable))
		return;

    gtk_tree_view_column_set_resizable(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)resizable);

}


static PHP_METHOD(GtkTreeViewColumn, set_sizing)
{
	GtkTreeViewColumnSizing type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_TREE_VIEW_COLUMN_SIZING, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_tree_view_column_set_sizing(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkTreeViewColumn, set_sort_column_id)
{
	long sort_column_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &sort_column_id))
		return;

    gtk_tree_view_column_set_sort_column_id(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gint)sort_column_id);

}


static PHP_METHOD(GtkTreeViewColumn, set_sort_indicator)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_tree_view_column_set_sort_indicator(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkTreeViewColumn, set_sort_order)
{
	GtkSortType order;
	zval *php_order = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_order))
		return;

	if (php_order && phpg_gvalue_get_enum(GTK_TYPE_SORT_TYPE, php_order, (gint *)&order) == FAILURE) {
		return;
	}

    gtk_tree_view_column_set_sort_order(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), order);

}


static PHP_METHOD(GtkTreeViewColumn, set_spacing)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_tree_view_column_set_spacing(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gint)spacing);

}


static PHP_METHOD(GtkTreeViewColumn, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_tree_view_column_set_title(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkTreeViewColumn, set_visible)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_tree_view_column_set_visible(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


static PHP_METHOD(GtkTreeViewColumn, set_widget)
{
	GtkWidget *widget = NULL;
	zval *php_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_widget, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_widget) != IS_NULL)
        widget = GTK_WIDGET(PHPG_GOBJECT(php_widget));

    gtk_tree_view_column_set_widget(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)), widget);

}


static PHP_METHOD(GtkTreeViewColumn, queue_resize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_column_queue_resize(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeViewColumn, get_tree_view)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_column_get_tree_view(GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeviewcolumn_gtk_tree_view_column_new_with_attributes, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_OBJ_INFO(0, cellrenderer, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, attribute)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_add_attribute, 0)
    ZEND_ARG_OBJ_INFO(0, cell_renderer, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, attribute)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_cell_get_position, 0)
    ZEND_ARG_OBJ_INFO(0, cellrenderer, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_cell_set_cell_data, 0)
    ZEND_ARG_INFO(0, tree_model)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
    ZEND_ARG_INFO(0, is_expander)
    ZEND_ARG_INFO(0, is_expanded)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_clear_attributes, 0)
    ZEND_ARG_OBJ_INFO(0, cell_renderer, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_focus_cell, 0)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeviewcolumn_pack_end, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeviewcolumn_pack_start, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_alignment, 0)
    ZEND_ARG_INFO(0, xalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_clickable, 0)
    ZEND_ARG_INFO(0, active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_expand, 0)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_fixed_width, 0)
    ZEND_ARG_INFO(0, fixed_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_max_width, 0)
    ZEND_ARG_INFO(0, max_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_min_width, 0)
    ZEND_ARG_INFO(0, min_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_reorderable, 0)
    ZEND_ARG_INFO(0, reorderable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_resizable, 0)
    ZEND_ARG_INFO(0, resizable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_sizing, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_sort_column_id, 0)
    ZEND_ARG_INFO(0, sort_column_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_sort_indicator, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_sort_order, 0)
    ZEND_ARG_INFO(0, order)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_spacing, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeviewcolumn_set_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreeviewcolumn_methods[] = {
	PHP_ME(GtkTreeViewColumn, __construct,          arginfo_gtk_gtktreeviewcolumn_gtk_tree_view_column_new_with_attributes, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, add_attribute,        arginfo_gtk_gtktreeviewcolumn_add_attribute, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, cell_get_position,    arginfo_gtk_gtktreeviewcolumn_cell_get_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, cell_get_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, cell_is_visible,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, cell_set_cell_data,   arginfo_gtk_gtktreeviewcolumn_cell_set_cell_data, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, clear_attributes,     arginfo_gtk_gtktreeviewcolumn_clear_attributes, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, clicked,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, focus_cell,           arginfo_gtk_gtktreeviewcolumn_focus_cell, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_alignment,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_cell_renderers,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_clickable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_expand,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_fixed_width,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_max_width,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_min_width,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_reorderable,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_resizable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_sizing,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_sort_column_id,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_sort_indicator,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_sort_order,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_spacing,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_tree_view,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_widget,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, get_width,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, pack_end,             arginfo_gtk_gtktreeviewcolumn_pack_end, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, pack_start,           arginfo_gtk_gtktreeviewcolumn_pack_start, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, queue_resize,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_alignment,        arginfo_gtk_gtktreeviewcolumn_set_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_clickable,        arginfo_gtk_gtktreeviewcolumn_set_clickable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_expand,           arginfo_gtk_gtktreeviewcolumn_set_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_fixed_width,      arginfo_gtk_gtktreeviewcolumn_set_fixed_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_max_width,        arginfo_gtk_gtktreeviewcolumn_set_max_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_min_width,        arginfo_gtk_gtktreeviewcolumn_set_min_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_reorderable,      arginfo_gtk_gtktreeviewcolumn_set_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_resizable,        arginfo_gtk_gtktreeviewcolumn_set_resizable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_sizing,           arginfo_gtk_gtktreeviewcolumn_set_sizing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_sort_column_id,   arginfo_gtk_gtktreeviewcolumn_set_sort_column_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_sort_indicator,   arginfo_gtk_gtktreeviewcolumn_set_sort_indicator, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_sort_order,       arginfo_gtk_gtktreeviewcolumn_set_sort_order, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_spacing,          arginfo_gtk_gtktreeviewcolumn_set_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_title,            arginfo_gtk_gtktreeviewcolumn_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_visible,          arginfo_gtk_gtktreeviewcolumn_set_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeViewColumn, set_widget,           arginfo_gtk_gtktreeviewcolumn_set_widget, ZEND_ACC_PUBLIC)

	/***   GtkCellLayout interface implementations   ***/

	PHP_MALIAS(GtkCellLayout, clear, clear, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, reorder, reorder, arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_attributes, set_attributes, arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_cell_data_func, set_cell_data_func, arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, get_cells, get_cells, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkUIManager, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkUIManager);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkUIManager);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkUIManager, add_ui)
{
	long merge_id;
	char *path, *name, *action;
	zend_bool free_path = FALSE, free_name = FALSE, free_action = FALSE, top;
	GtkUIManagerItemType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iuuuVb", &merge_id, &path, &free_path, &name, &free_name, &action, &free_action, &php_type, &top))
		return;

	if (php_type && phpg_gvalue_get_flags(GTK_TYPE_UI_MANAGER_ITEM_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_ui_manager_add_ui(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), (guint)merge_id, path, name, action, type, (gboolean)top);
	if (free_path) g_free(path);
	if (free_name) g_free(name);
	if (free_action) g_free(action);

}


static PHP_METHOD(GtkUIManager, add_ui_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GError *error = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    php_retval = gtk_ui_manager_add_ui_from_file(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), filename, &error);
	if (free_filename) g_free(filename);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
	RETVAL_LONG(php_retval);
}

#line 5683 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkUIManager, add_ui_from_string)
{
    char *text;
    long text_len;
    guint ret;
    GError *error = NULL;
    zend_bool free_text;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u#", &text, &text_len, &free_text))
        return;

    ret = gtk_ui_manager_add_ui_from_string(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), text, text_len, &error);

    if (free_text) g_free(text);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }
    RETURN_LONG(ret);
}

#line 26188 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkUIManager, ensure_update)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_ui_manager_ensure_update(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkUIManager, get_accel_group)
{
	GtkAccelGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_ui_manager_get_accel_group(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkUIManager, get_action)
{
	char *path;
	zend_bool free_path = FALSE;
	GtkAction* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    php_retval = gtk_ui_manager_get_action(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 5743 "ext/gtk+/gtk.overrides"

static PHP_METHOD(GtkUIManager, get_action_groups)
{
    GList *action_groups = 0;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    array_init(return_value);

    for (action_groups = gtk_ui_manager_get_action_groups(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));
         action_groups;
         action_groups = action_groups->next) {

        zval *item = NULL;
        const char *name = NULL;

        name = gtk_action_group_get_name(action_groups->data);
        phpg_gobject_new(&item, G_OBJECT(action_groups->data) TSRMLS_CC);
        add_assoc_zval(return_value, (char *)name, item);
    }
}

#line 26260 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkUIManager, get_add_tearoffs)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_ui_manager_get_add_tearoffs(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 5713 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkUIManager, get_toplevels)
{
    GtkUIManagerItemType types;
    zval *php_types = NULL;
    GSList *toplevels = 0, *tmp;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &types))
        return;

	if (php_types && phpg_gvalue_get_flags(GTK_TYPE_UI_MANAGER_ITEM_TYPE, php_types, (gint *)&types) == FAILURE) {
		return;
	}
    array_init(return_value);

    toplevels = gtk_ui_manager_get_toplevels(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), types);
    for (tmp = toplevels; tmp; tmp = tmp->next) {
        zval *item = NULL;
        const char *name = NULL;

        name = gtk_widget_get_name(tmp->data);
        phpg_gobject_new(&item, G_OBJECT(tmp->data) TSRMLS_CC);
        add_assoc_zval(return_value, (char *)name, item);
    }
    g_slist_free(toplevels);
}


#line 26306 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkUIManager, get_ui)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_ui_manager_get_ui(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkUIManager, get_widget)
{
	char *path;
	zend_bool free_path = FALSE;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &path, &free_path))
		return;

    php_retval = gtk_ui_manager_get_widget(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), path);
	if (free_path) g_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkUIManager, insert_action_group)
{
	zval *action_group;
	long pos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &action_group, gtkactiongroup_ce, &pos))
		return;

    gtk_ui_manager_insert_action_group(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), GTK_ACTION_GROUP(PHPG_GOBJECT(action_group)), (gint)pos);

}


static PHP_METHOD(GtkUIManager, new_merge_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_ui_manager_new_merge_id(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkUIManager, remove_action_group)
{
	zval *action_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &action_group, gtkactiongroup_ce))
		return;

    gtk_ui_manager_remove_action_group(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), GTK_ACTION_GROUP(PHPG_GOBJECT(action_group)));

}


static PHP_METHOD(GtkUIManager, remove_ui)
{
	long merge_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &merge_id))
		return;

    gtk_ui_manager_remove_ui(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), (guint)merge_id);

}


static PHP_METHOD(GtkUIManager, set_add_tearoffs)
{
	zend_bool add_tearoffs;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &add_tearoffs))
		return;

    gtk_ui_manager_set_add_tearoffs(GTK_UI_MANAGER(PHPG_GOBJECT(this_ptr)), (gboolean)add_tearoffs);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_add_ui, 0)
    ZEND_ARG_INFO(0, merge_id)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_INFO(0, name)
    ZEND_ARG_INFO(0, action)
    ZEND_ARG_INFO(0, type)
    ZEND_ARG_INFO(0, top)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_add_ui_from_file, 0)
    ZEND_ARG_INFO(0, filename)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_add_ui_from_string, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_get_action, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_get_toplevels, 0)
    ZEND_ARG_INFO(0, types)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_get_widget, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_insert_action_group, 0)
    ZEND_ARG_OBJ_INFO(0, action_group, GtkActionGroup, 1)
    ZEND_ARG_INFO(0, pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_remove_action_group, 0)
    ZEND_ARG_OBJ_INFO(0, action_group, GtkActionGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_remove_ui, 0)
    ZEND_ARG_INFO(0, merge_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkuimanager_set_add_tearoffs, 0)
    ZEND_ARG_INFO(0, add_tearoffs)
ZEND_END_ARG_INFO();

static zend_function_entry gtkuimanager_methods[] = {
	PHP_ME(GtkUIManager, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, add_ui,               arginfo_gtk_gtkuimanager_add_ui, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, add_ui_from_file,     arginfo_gtk_gtkuimanager_add_ui_from_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, add_ui_from_string,   arginfo_gtk_gtkuimanager_add_ui_from_string, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, ensure_update,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_accel_group,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_action,           arginfo_gtk_gtkuimanager_get_action, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_action_groups,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_add_tearoffs,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_toplevels,        arginfo_gtk_gtkuimanager_get_toplevels, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_ui,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, get_widget,           arginfo_gtk_gtkuimanager_get_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, insert_action_group,  arginfo_gtk_gtkuimanager_insert_action_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, new_merge_id,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, remove_action_group,  arginfo_gtk_gtkuimanager_remove_action_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, remove_ui,            arginfo_gtk_gtkuimanager_remove_ui, ZEND_ACC_PUBLIC)
	PHP_ME(GtkUIManager, set_add_tearoffs,     arginfo_gtk_gtkuimanager_set_add_tearoffs, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

PHPG_PROP_READER(GtkWidget, state)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->state;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, saved_state)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->saved_state;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->name;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, style)
{
	GtkStyle* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->style;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, allocation)
{
	GdkRectangle php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->allocation;
	phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &php_retval, TRUE, TRUE TSRMLS_CC);

    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, window)
{
	GdkWindow* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->window;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWidget, parent)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WIDGET(((phpg_gobject_t *)object)->obj)->parent;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkwidget_prop_info[] = {
	{ "state", PHPG_PROP_READ_FN(GtkWidget, state), NULL },
	{ "saved_state", PHPG_PROP_READ_FN(GtkWidget, saved_state), NULL },
	{ "name", PHPG_PROP_READ_FN(GtkWidget, name), NULL },
	{ "style", PHPG_PROP_READ_FN(GtkWidget, style), NULL },
	{ "allocation", PHPG_PROP_READ_FN(GtkWidget, allocation), NULL },
	{ "window", PHPG_PROP_READ_FN(GtkWidget, window), NULL },
	{ "parent", PHPG_PROP_READ_FN(GtkWidget, parent), NULL },
	{ NULL, NULL, NULL },
};

#line 95 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkWidget, get_allocation)
{
    GtkAllocation allocation;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_get_allocation(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &allocation);

    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &allocation, TRUE, TRUE TSRMLS_CC);
}

#line 26641 "ext/gtk+/gen_gtk.c"


#line 110 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkWidget, set_visible)
{
    zend_bool visible;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible)) {
        return;
    }

    gtk_widget_set_visible(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), visible);
}

#line 26658 "ext/gtk+/gen_gtk.c"


#line 84 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_begin)
{
	zval *php_targets, *php_actions, *event;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	GtkTargetList *tlist;
	GdkDragContext *context;
	gint button, n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "aViO", &php_targets, &php_actions,
							&button, &event, gdkevent_ce)) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
    if (!entries) return;

	tlist = gtk_target_list_new(entries, n);
	efree(entries);
	context = gtk_drag_begin(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), tlist, actions, button, (GdkEvent *)PHPG_GBOXED(event));
	gtk_target_list_unref(tlist);

    phpg_gobject_new(&return_value, (GObject *)context TSRMLS_CC);
}

#line 26693 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, drag_check_threshold)
{
	long start_x, start_y, current_x, current_y;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &start_x, &start_y, &current_x, &current_y))
		return;

    php_retval = gtk_drag_check_threshold(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)start_x, (gint)start_y, (gint)current_x, (gint)current_y);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, drag_dest_add_image_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_dest_add_image_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_dest_add_text_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_dest_add_text_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_dest_add_uri_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_dest_add_uri_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}

#line 187 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_dest_find_target)
{
    zval *php_targets = NULL, *php_context;
    GtkTargetEntry *entries;
    GtkTargetList *tlist = NULL;
    GdkAtom atom;
    gchar *name;
    int n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|a", &php_context, gdkdragcontext_ce, &php_targets)) {
        return;
    }

    if (php_targets) {
        entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
        tlist = gtk_target_list_new(entries, n);
        efree(entries);
    }

    atom = gtk_drag_dest_find_target(GTK_WIDGET(PHPG_GOBJECT(this_ptr)),
                                    GDK_DRAG_CONTEXT(PHPG_GOBJECT(php_context)), tlist);

    if (php_targets) {
        gtk_target_list_unref(tlist);
    }

    name = gdk_atom_name(atom);
    RETVAL_STRING(name, 1);
    g_free(name);
}

#line 26784 "ext/gtk+/gen_gtk.c"


#line 117 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_dest_get_target_list)
{
    GtkTargetList *tlist;
    GtkTargetPair *pair;
    GList *tmp;
    gchar *name;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    tlist = gtk_drag_dest_get_target_list(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (!tlist) {
        return;
    }

    array_init(return_value);
    for (tmp = tlist->list; tmp != NULL; tmp = tmp->next) {
        zval *item = NULL;

        pair = tmp->data;
        name = gdk_atom_name(pair->target);
        php_gtk_build_value(&item, "(sii)", name, pair->flags, pair->info);
        g_free(name);
        add_next_index_zval(return_value, item);
    }
}

#line 26818 "ext/gtk+/gen_gtk.c"


#line 19 "ext/gtk+/gtkdrag.overrides"

static GtkTargetEntry* phpg_parse_target_entries(zval *php_targets, int *n_targets TSRMLS_DC)
{
    GtkTargetEntry *tents;
    zval **item;
    int n;

    *n_targets = 0;
	tents = emalloc(sizeof(GtkTargetEntry) * zend_hash_num_elements(Z_ARRVAL_P(php_targets)));

	for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_targets)), n = 0;
		 zend_hash_get_current_data(Z_ARRVAL_P(php_targets), (void **)&item) == SUCCESS;
		 zend_hash_move_forward(Z_ARRVAL_P(php_targets)), n++) {
		if (Z_TYPE_PP(item) != IS_ARRAY ||
			!php_gtk_parse_args_hash_quiet(*item, "sii", &tents[n].target,
										   &tents[n].flags, &tents[n].info)) {
			php_error(E_WARNING, "%s::%s() was unable to parse target #%d in the list of targets", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C), n+1);
			efree(tents);
			return NULL;
		}
	}

    *n_targets = n;
    return tents;
}

static PHP_METHOD(GtkWidget, drag_dest_set)
{
	zval *php_flags, *php_targets, *php_actions;
	GtkDestDefaults flags;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VaV", &php_flags, &php_targets, &php_actions)) {
		return;
	}

	if (phpg_gvalue_get_flags(GTK_TYPE_DEST_DEFAULTS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
	}

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_drag_dest_set(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), flags, entries, n, actions);
	efree(entries);
}

#line 26875 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, drag_dest_set_proxy)
{
	zval *proxy_window, *php_protocol = NULL;
	GdkDragProtocol protocol;
	zend_bool use_coordinates;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVb", &proxy_window, gdkwindow_ce, &php_protocol, &use_coordinates))
		return;

	if (php_protocol && phpg_gvalue_get_enum(GDK_TYPE_DRAG_PROTOCOL, php_protocol, (gint *)&protocol) == FAILURE) {
		return;
	}

    gtk_drag_dest_set_proxy(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(proxy_window)), protocol, (gboolean)use_coordinates);

}

#line 156 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_dest_set_target_list)
{
    zval *php_targets;
    GtkTargetEntry *entries;
	GtkTargetList *tlist;
    int n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_targets)) {
        return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	tlist = gtk_target_list_new(entries, n);
	efree(entries);

    gtk_drag_dest_set_target_list(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), tlist);
    gtk_target_list_unref(tlist);
}

#line 26920 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, drag_dest_unset)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_dest_unset(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_get_data)
{
	zval *context, *php_target = NULL;
	GdkAtom target;
	long time = GDK_CURRENT_TIME;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV|i", &context, gdkdragcontext_ce, &php_target, &time))
		return;

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_drag_get_data(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_DRAG_CONTEXT(PHPG_GOBJECT(context)), target, (guint32)time);

}


static PHP_METHOD(GtkWidget, drag_highlight)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_highlight(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_source_add_text_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_source_add_text_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_source_add_image_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_source_add_image_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_source_add_uri_targets)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_source_add_uri_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}

#line 260 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_source_get_target_list)
{
    GtkTargetList *tlist;
    GtkTargetPair *pair;
    GList *tmp;
    gchar *name;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    tlist = gtk_drag_source_get_target_list(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (!tlist) {
        return;
    }

    array_init(return_value);
    for (tmp = tlist->list; tmp != NULL; tmp = tmp->next) {
        zval *item = NULL;

        pair = tmp->data;
        name = gdk_atom_name(pair->target);
        php_gtk_build_value(&item, "(sii)", name, pair->flags, pair->info);
        g_free(name);
        add_next_index_zval(return_value, item);
    }
}

#line 27042 "ext/gtk+/gen_gtk.c"


#line 231 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_source_set)
{
	zval *php_sbmask, *php_targets, *php_actions;
	GdkModifierType sbmask;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VaV", &php_sbmask, &php_targets, &php_actions)) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_sbmask, (gint *)&sbmask) == FAILURE) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
        return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_drag_source_set(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), sbmask, entries, n, actions);
	efree(entries);
}

#line 27073 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, drag_source_set_icon)
{
	zval *colormap, *pixmap, *php_mask = NULL;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO|N", &colormap, gdkcolormap_ce, &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;
    if (php_mask) {
        if (Z_TYPE_P(php_mask) == IS_NULL)
            mask = NULL;
        else
            mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));
    }

    gtk_drag_source_set_icon(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_COLORMAP(PHPG_GOBJECT(colormap)), GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);

}


static PHP_METHOD(GtkWidget, drag_source_set_icon_pixbuf)
{
	zval *pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce))
		return;

    gtk_drag_source_set_icon_pixbuf(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

}


static PHP_METHOD(GtkWidget, drag_source_set_icon_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    gtk_drag_source_set_icon_stock(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);

}

#line 299 "ext/gtk+/gtkdrag.overrides"
static PHP_METHOD(GtkWidget, drag_source_set_target_list)
{
    zval *php_targets;
    GtkTargetEntry *entries;
	GtkTargetList *tlist;
    int n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_targets)) {
        return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	tlist = gtk_target_list_new(entries, n);
	efree(entries);

    gtk_drag_source_set_target_list(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), tlist);
    gtk_target_list_unref(tlist);
}

#line 27149 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, drag_source_unset)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_source_unset(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_unhighlight)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_drag_unhighlight(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, grab_add)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_grab_add(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, grab_remove)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_grab_remove(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, rc_get_style)
{
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_rc_get_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, selection_add_target)
{
	GdkAtom selection, target;
	zval *php_selection = NULL, *php_target = NULL;
	long info;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VVi", &php_selection, &php_target, &info))
		return;

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_selection_add_target(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection, target, (guint)info);

}

#line 6103 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, selection_add_targets)
{
    zval *php_selection, *php_targets;
    GdkAtom selection;
    GtkTargetEntry *targets;
    int n_targets;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Va", &php_selection, &php_targets)) {
        return;
    }

    selection = phpg_gdkatom_from_zval(php_targets TSRMLS_CC);
    if (selection == NULL) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "target argument has to be a valid GdkAtom");
        return;
    }

	targets = phpg_parse_target_entries(php_targets, &n_targets TSRMLS_CC);
    if (!targets) return;
    gtk_selection_add_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection, targets, n_targets);
    efree(targets);
    RETURN_TRUE;
}

#line 27275 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, selection_clear)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_selection_clear(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (GdkEventSelection *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, selection_clear_targets)
{
	GdkAtom selection;
	zval *php_selection = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_selection))
		return;

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_selection_clear_targets(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection);

}


static PHP_METHOD(GtkWidget, selection_convert)
{
	GdkAtom selection, target;
	zval *php_selection = NULL, *php_target = NULL;
	long time = GDK_CURRENT_TIME;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV|i", &php_selection, &php_target, &time))
		return;

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    target = phpg_gdkatom_from_zval(php_target TSRMLS_CC);
    if (target == NULL) {
        php_error(E_WARNING, "%s::%s() expects target argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_selection_convert(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection, target, (guint32)time);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, selection_owner_set)
{
	GdkAtom selection;
	zval *php_selection = NULL;
	long time = GDK_CURRENT_TIME;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|i", &php_selection, &time))
		return;

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_selection_owner_set(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection, (guint32)time);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, selection_remove_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_selection_remove_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, activate)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_activate(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, add_accelerator)
{
	char *accel_signal;
	zend_bool free_accel_signal = FALSE;
	zval *accel_group, *php_accel_mods = NULL, *php_accel_flags = NULL;
	long accel_key;
	GdkModifierType accel_mods;
	GtkAccelFlags accel_flags;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uOiVV", &accel_signal, &free_accel_signal, &accel_group, gtkaccelgroup_ce, &accel_key, &php_accel_mods, &php_accel_flags))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

	if (php_accel_flags && phpg_gvalue_get_flags(GTK_TYPE_ACCEL_FLAGS, php_accel_flags, (gint *)&accel_flags) == FAILURE) {
		return;
	}

    gtk_widget_add_accelerator(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), accel_signal, GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)), (guint)accel_key, accel_mods, accel_flags);
	if (free_accel_signal) g_free(accel_signal);

}


static PHP_METHOD(GtkWidget, add_events)
{
	long events;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &events))
		return;

    gtk_widget_add_events(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)events);

}


static PHP_METHOD(GtkWidget, add_mnemonic_label)
{
	zval *label;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &label, gtkwidget_ce))
		return;

    gtk_widget_add_mnemonic_label(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(label)));

}


static PHP_METHOD(GtkWidget, can_activate_accel)
{
	long signal_id;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &signal_id))
		return;

    php_retval = gtk_widget_can_activate_accel(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (guint)signal_id);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, child_focus)
{
	GtkDirectionType direction;
	zval *php_direction = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_direction))
		return;

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_DIRECTION_TYPE, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

    php_retval = gtk_widget_child_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), direction);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, child_notify)
{
	char *child_property;
	zend_bool free_child_property = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &child_property, &free_child_property))
		return;

    gtk_widget_child_notify(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), child_property);
	if (free_child_property) g_free(child_property);

}

#line 5773 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, class_path)
{
    guint length;
    gchar *path;
    gchar *path_reversed;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_class_path(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &length, &path,
                          &path_reversed);
    RETURN_STRING(path, 1);
    g_free(path);
    g_free(path_reversed);
}

#line 27531 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, create_pango_context)
{
	PangoContext* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_create_pango_context(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


static PHP_METHOD(GtkWidget, create_pango_layout)
{
	char *text;
	zend_bool free_text = FALSE;
	PangoLayout* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    php_retval = gtk_widget_create_pango_layout(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    if (php_retval != NULL) {
        g_object_unref(php_retval);
    }
}


static PHP_METHOD(GtkWidget, destroy)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_destroy(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, draw)
{
	GdkRectangle area = { 0, 0, 0, 0 };
	zval *php_area;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_area))
		return;

    if (phpg_rectangle_from_zval(php_area, (GdkRectangle*)&area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
	phpg_warn_deprecated("use GtkWidget::queue_draw_area()" TSRMLS_CC);

    gtk_widget_draw(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &area);

}


static PHP_METHOD(GtkWidget, ensure_style)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_ensure_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, event)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_widget_event(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, freeze_child_notify)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_freeze_child_notify(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, get_accessible)
{
	AtkObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_accessible(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_ancestor)
{
	GType widget_type;
	zval *php_widget_type = NULL;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_widget_type))
		return;
    if ((widget_type = phpg_gtype_from_zval(php_widget_type TSRMLS_CC)) == 0) {
        return;
    }

    php_retval = gtk_widget_get_ancestor(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), widget_type);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 5806 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, get_child_requisition)
{
    GtkRequisition requisition;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_get_child_requisition(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &requisition);
    phpg_gboxed_new(&return_value, GTK_TYPE_REQUISITION, &requisition, TRUE, TRUE TSRMLS_CC);
}

#line 27702 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, get_child_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_child_visible(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_clipboard)
{
	GdkAtom selection;
	zval *php_selection = NULL;
	GtkClipboard* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_selection))
		return;

    selection = phpg_gdkatom_from_zval(php_selection TSRMLS_CC);
    if (selection == NULL) {
        php_error(E_WARNING, "%s::%s() expects selection argument to be a valid GdkAtom object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_widget_get_clipboard(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), selection);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_colormap)
{
	GdkColormap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_colormap(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_composite_name)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_composite_name(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkWidget, get_default_colormap)
{
	GdkColormap* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_default_colormap();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_default_direction)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_default_direction();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, get_default_style)
{
	GtkStyle* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_default_style();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_default_visual)
{
	GdkVisual* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_default_visual();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_direction)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_direction(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, get_display)
{
	GdkDisplay* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_display(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_events)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_events(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, get_extension_events)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_extension_events(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, get_modifier_style)
{
	GtkRcStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_modifier_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_name(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkWidget, get_no_show_all)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_no_show_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_pango_context)
{
	PangoContext* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_pango_context(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_parent)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_parent(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_parent_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_parent_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 5821 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, get_pointer)
{
    gint x, y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_get_pointer(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &x, &y);
    php_gtk_build_value(&return_value, "(ii)", x, y);
}

#line 27999 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, get_root_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_root_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_screen)
{
	GdkScreen* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_screen(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_settings)
{
	GtkSettings* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_settings(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 5836 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, get_size_request)
{
    gint width, height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_get_size_request(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &width, &height);
    php_gtk_build_value(&return_value, "(ii)", width, height);
}


#line 28059 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, get_style)
{
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_toplevel)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_toplevel(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_visual)
{
	GdkVisual* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_visual(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, grab_default)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_grab_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, grab_focus)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_grab_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, has_screen)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_has_screen(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, hide)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_hide(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, hide_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_hide_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, hide_on_delete)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_hide_on_delete(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 5859 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, intersect)
{
    zval *php_area;
    GdkRectangle area, intersection;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_area))
        return;

    if (phpg_rectangle_from_zval(php_area, &area TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects area argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (gtk_widget_intersect(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &area, &intersection)) {
        phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &intersection, TRUE, TRUE TSRMLS_CC);
    } else {
        RETURN_FALSE;
    }
}

#line 28206 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, is_ancestor)
{
	zval *ancestor;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &ancestor, gtkwidget_ce))
		return;

    php_retval = gtk_widget_is_ancestor(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(ancestor)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, is_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_is_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 5882 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, list_mnemonic_labels)
{
    GList *labels;
    gint i, len;
    zval *php_item = NULL;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    labels = gtk_widget_list_mnemonic_labels(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    len = g_list_length(labels);

    array_init(return_value);
    for (i = 0; i < len; i++) {
        MAKE_STD_ZVAL(php_item);
        phpg_gobject_new(&php_item, (GObject *)g_list_nth_data(labels, i) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }

    g_list_free(labels);
}

#line 28262 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, map)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_map(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, mnemonic_activate)
{
	zend_bool group_cycling;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &group_cycling))
		return;

    php_retval = gtk_widget_mnemonic_activate(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)group_cycling);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, modify_base)
{
	GtkStateType state;
	zval *php_state = NULL, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VN", &php_state, &php_color, gboxed_ce))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_color) != IS_NULL) {
        if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            color = (GdkColor *) PHPG_GBOXED(php_color);
        } else {
            php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_base(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), state, color);

}


static PHP_METHOD(GtkWidget, modify_bg)
{
	GtkStateType state;
	zval *php_state = NULL, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VN", &php_state, &php_color, gboxed_ce))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_color) != IS_NULL) {
        if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            color = (GdkColor *) PHPG_GBOXED(php_color);
        } else {
            php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_bg(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), state, color);

}


static PHP_METHOD(GtkWidget, modify_fg)
{
	GtkStateType state;
	zval *php_state = NULL, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VN", &php_state, &php_color, gboxed_ce))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_color) != IS_NULL) {
        if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            color = (GdkColor *) PHPG_GBOXED(php_color);
        } else {
            php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_fg(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), state, color);

}


static PHP_METHOD(GtkWidget, modify_font)
{
	PangoFontDescription *font_desc = NULL;
	zval *php_font_desc;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_font_desc, gboxed_ce))
		return;

    if (Z_TYPE_P(php_font_desc) != IS_NULL) {
        if (phpg_gboxed_check(php_font_desc, PANGO_TYPE_FONT_DESCRIPTION, FALSE TSRMLS_CC)) {
            font_desc = (PangoFontDescription *) PHPG_GBOXED(php_font_desc);
        } else {
            php_error(E_WARNING, "%s::%s() expects font_desc argument to be a valid PangoFontDescription object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_font(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), font_desc);

}


static PHP_METHOD(GtkWidget, modify_style)
{
	zval *style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &style, gtkrcstyle_ce))
		return;

    gtk_widget_modify_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_RC_STYLE(PHPG_GOBJECT(style)));

}


static PHP_METHOD(GtkWidget, modify_text)
{
	GtkStateType state;
	zval *php_state = NULL, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VN", &php_state, &php_color, gboxed_ce))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    if (Z_TYPE_P(php_color) != IS_NULL) {
        if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            color = (GdkColor *) PHPG_GBOXED(php_color);
        } else {
            php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_text(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), state, color);

}

#line 5907 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, path)
{
    guint length;
    gchar *path;
    gchar *path_reversed;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_path(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &length, &path,
                          &path_reversed);
    RETURN_STRING(path, 1);
    g_free(path);
    g_free(path_reversed);
}

#line 28465 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, pop_colormap)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_pop_colormap();

}


static PHP_METHOD(GtkWidget, pop_composite_child)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_pop_composite_child();

}


static PHP_METHOD(GtkWidget, push_colormap)
{
	zval *cmap;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &cmap, gdkcolormap_ce))
		return;

    gtk_widget_push_colormap(GDK_COLORMAP(PHPG_GOBJECT(cmap)));

}


static PHP_METHOD(GtkWidget, push_composite_child)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_push_composite_child();

}


static PHP_METHOD(GtkWidget, queue_clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkWidget::queue_draw()" TSRMLS_CC);

    gtk_widget_queue_clear(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, queue_clear_area)
{
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &x, &y, &width, &height))
		return;

	phpg_warn_deprecated("use GtkWidget::queue_draw_area()" TSRMLS_CC);

    gtk_widget_queue_clear_area(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y, (gint)width, (gint)height);

}


static PHP_METHOD(GtkWidget, queue_draw)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_queue_draw(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, queue_draw_area)
{
	long x, y, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &x, &y, &width, &height))
		return;

    gtk_widget_queue_draw_area(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y, (gint)width, (gint)height);

}


static PHP_METHOD(GtkWidget, queue_resize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_queue_resize(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, queue_resize_no_redraw)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_queue_resize_no_redraw(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, realize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_realize(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, region_intersect)
{
	GdkRegion *region = NULL, *php_retval;
	zval *php_region;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_region, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_region, GDK_TYPE_REGION, FALSE TSRMLS_CC)) {
        region = (GdkRegion *) PHPG_GBOXED(php_region);
    } else {
        php_error(E_WARNING, "%s::%s() expects region argument to be a valid GdkRegion object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_widget_region_intersect(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), region);
	phpg_gboxed_new(&return_value, GDK_TYPE_REGION, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkWidget, remove_accelerator)
{
	zval *accel_group, *php_accel_mods = NULL;
	long accel_key;
	GdkModifierType accel_mods;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiV", &accel_group, gtkaccelgroup_ce, &accel_key, &php_accel_mods))
		return;

	if (php_accel_mods && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_accel_mods, (gint *)&accel_mods) == FAILURE) {
		return;
	}

    php_retval = gtk_widget_remove_accelerator(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)), (guint)accel_key, accel_mods);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, remove_mnemonic_label)
{
	zval *label;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &label, gtkwidget_ce))
		return;

    gtk_widget_remove_mnemonic_label(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(label)));

}


static PHP_METHOD(GtkWidget, render_icon)
{
	char *stock_id, *detail = NULL;
	zend_bool free_stock_id = FALSE, free_detail = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV|u", &stock_id, &free_stock_id, &php_size, &detail, &free_detail))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    php_retval = gtk_widget_render_icon(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), stock_id, size, detail);
	if (free_stock_id) g_free(stock_id);
	if (free_detail) g_free(detail);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, reparent)
{
	zval *new_parent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &new_parent, gtkwidget_ce))
		return;

    gtk_widget_reparent(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(new_parent)));

}


static PHP_METHOD(GtkWidget, reset_rc_styles)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_reset_rc_styles(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, reset_shapes)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_reset_shapes(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, send_expose)
{
	GdkEvent *event = NULL;
	zval *php_event;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_widget_send_expose(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), event);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, set_accel_path)
{
	char *accel_path;
	zend_bool free_accel_path = FALSE;
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uO", &accel_path, &free_accel_path, &accel_group, gtkaccelgroup_ce))
		return;

    gtk_widget_set_accel_path(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), accel_path, GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));
	if (free_accel_path) g_free(accel_path);

}


static PHP_METHOD(GtkWidget, set_app_paintable)
{
	zend_bool app_paintable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &app_paintable))
		return;

    gtk_widget_set_app_paintable(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)app_paintable);

}


static PHP_METHOD(GtkWidget, set_child_visible)
{
	zend_bool is_visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_visible))
		return;

    gtk_widget_set_child_visible(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)is_visible);

}


static PHP_METHOD(GtkWidget, set_colormap)
{
	zval *colormap;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &colormap, gdkcolormap_ce))
		return;

    gtk_widget_set_colormap(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_COLORMAP(PHPG_GOBJECT(colormap)));

}


static PHP_METHOD(GtkWidget, set_composite_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_widget_set_composite_name(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkWidget, set_default_colormap)
{
	zval *colormap;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &colormap, gdkcolormap_ce))
		return;

    gtk_widget_set_default_colormap(GDK_COLORMAP(PHPG_GOBJECT(colormap)));

}


static PHP_METHOD(GtkWidget, set_default_direction)
{
	GtkTextDirection dir;
	zval *php_dir = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_dir))
		return;

	if (php_dir && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_dir, (gint *)&dir) == FAILURE) {
		return;
	}

    gtk_widget_set_default_direction(dir);

}


static PHP_METHOD(GtkWidget, set_direction)
{
	GtkTextDirection dir;
	zval *php_dir = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_dir))
		return;

	if (php_dir && phpg_gvalue_get_enum(GTK_TYPE_TEXT_DIRECTION, php_dir, (gint *)&dir) == FAILURE) {
		return;
	}

    gtk_widget_set_direction(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), dir);

}


static PHP_METHOD(GtkWidget, set_double_buffered)
{
	zend_bool double_buffered;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &double_buffered))
		return;

    gtk_widget_set_double_buffered(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)double_buffered);

}


static PHP_METHOD(GtkWidget, set_events)
{
	long events;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &events))
		return;

    gtk_widget_set_events(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)events);

}


static PHP_METHOD(GtkWidget, set_extension_events)
{
	GdkExtensionMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(GDK_TYPE_EXTENSION_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_widget_set_extension_events(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkWidget, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_widget_set_name(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkWidget, set_no_show_all)
{
	zend_bool no_show_all;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &no_show_all))
		return;

    gtk_widget_set_no_show_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)no_show_all);

}


static PHP_METHOD(GtkWidget, set_parent)
{
	zval *parent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &parent, gtkwidget_ce))
		return;

    gtk_widget_set_parent(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(parent)));

}


static PHP_METHOD(GtkWidget, set_parent_window)
{
	zval *parent_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &parent_window, gdkwindow_ce))
		return;

    gtk_widget_set_parent_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(parent_window)));

}


static PHP_METHOD(GtkWidget, set_redraw_on_allocate)
{
	zend_bool redraw_on_allocate;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &redraw_on_allocate))
		return;

    gtk_widget_set_redraw_on_allocate(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)redraw_on_allocate);

}


static PHP_METHOD(GtkWidget, set_scroll_adjustments)
{
	GtkAdjustment *hadjustment = NULL, *vadjustment = NULL;
	zval *php_hadjustment, *php_vadjustment;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "NN", &php_hadjustment, gtkadjustment_ce, &php_vadjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_hadjustment) != IS_NULL)
        hadjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_hadjustment));
    if (Z_TYPE_P(php_vadjustment) != IS_NULL)
        vadjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_vadjustment));

    php_retval = gtk_widget_set_scroll_adjustments(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), hadjustment, vadjustment);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, set_sensitive)
{
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &sensitive))
		return;

    gtk_widget_set_sensitive(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)sensitive);

}


static PHP_METHOD(GtkWidget, set_size_request)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_widget_set_size_request(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


static PHP_METHOD(GtkWidget, set_state)
{
	GtkStateType state;
	zval *php_state = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_state))
		return;

	if (php_state && phpg_gvalue_get_enum(GTK_TYPE_STATE_TYPE, php_state, (gint *)&state) == FAILURE) {
		return;
	}

    gtk_widget_set_state(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), state);

}


static PHP_METHOD(GtkWidget, set_style)
{
	GtkStyle *style = NULL;
	zval *php_style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_style, gtkstyle_ce))
		return;
    if (Z_TYPE_P(php_style) != IS_NULL)
        style = GTK_STYLE(PHPG_GOBJECT(php_style));

    gtk_widget_set_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), style);

}


static PHP_METHOD(GtkWidget, set_uposition)
{
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    gtk_widget_set_uposition(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkWidget, set_usize)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

	phpg_warn_deprecated("use GtkWidget::set_size_request()" TSRMLS_CC);

    gtk_widget_set_usize(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


static PHP_METHOD(GtkWidget, shape_combine_mask)
{
	zval *shape_mask;
	long offset_x, offset_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &shape_mask, gdkpixmap_ce, &offset_x, &offset_y))
		return;

    gtk_widget_shape_combine_mask(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_PIXMAP(PHPG_GOBJECT(shape_mask)), (gint)offset_x, (gint)offset_y);

}


static PHP_METHOD(GtkWidget, show)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_show(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, show_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_show_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, show_now)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_show_now(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}

#line 5948 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, size_allocate)
{
    GdkRectangle allocation = { 0, 0, 0, 0 };
    zval *php_allocation;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_allocation))
        return;

    if (phpg_rectangle_from_zval(php_allocation, &allocation TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects allocation argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    gtk_widget_size_allocate(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &allocation);

}

#line 29186 "ext/gtk+/gen_gtk.c"


#line 5926 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, size_request)
{
    GtkRequisition requisition;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_widget_size_request(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), &requisition);
    phpg_gboxed_new(&return_value, GTK_TYPE_REQUISITION, &requisition, TRUE, TRUE TSRMLS_CC);
}

#line 29203 "ext/gtk+/gen_gtk.c"


#line 6072 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, style_get_property)
{
    char *param;
    GParamSpec *pspec;
    GValue value = { 0, };

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "s", &param)) {
        return;
    }

    if ((pspec = gtk_widget_class_find_style_property(
                                  GTK_WIDGET_GET_CLASS(PHPG_GOBJECT(this_ptr)), param)) == NULL) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "the widget does not support style property '%s'", param);
        return;
    }

    if (!(pspec->flags & G_PARAM_READABLE)) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "style property '%s is not readable", param);
        return;
    }

    g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
    gtk_widget_style_get_property(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (const gchar *)param, &value);
    phpg_gvalue_to_zval(&value, &return_value, TRUE, TRUE TSRMLS_CC);
    g_value_unset(&value);
}

#line 29236 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, thaw_child_notify)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_thaw_child_notify(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}

#line 5977 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, translate_coordinates)
{
    gint src_x, src_y, dest_x, dest_y;
    zval *dest_widget;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &dest_widget, gtkwidget_ce, &src_x, &src_y))
        return;

    if (gtk_widget_translate_coordinates(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(dest_widget)), src_x, src_y, &dest_x, &dest_y)) {
        php_gtk_build_value(&return_value, "(ii)", dest_x, dest_y);
    } else {
        RETURN_FALSE;
    }
}

#line 29270 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, unmap)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_unmap(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, unparent)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_unparent(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, unrealize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_unrealize(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, drag_set_icon_name)
{
	zval *context;
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	long hot_x, hot_y;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ouii", &context, gdkdragcontext_ce, &icon_name, &free_icon_name, &hot_x, &hot_y))
		return;

    gtk_drag_set_icon_name(GDK_DRAG_CONTEXT(PHPG_GOBJECT(context)), icon_name, (gint)hot_x, (gint)hot_y);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkWidget, drag_source_set_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_drag_source_set_icon_name(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkWidget, get_action)
{
	GtkAction* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, Use GtkActivatable::set_related_action() instead" TSRMLS_CC);

    php_retval = gtk_widget_get_action(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, drag_dest_set_track_motion)
{
	zend_bool track_motion;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &track_motion))
		return;

    gtk_drag_dest_set_track_motion(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)track_motion);

}


static PHP_METHOD(GtkWidget, drag_dest_get_track_motion)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_drag_dest_get_track_motion(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, is_composited)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_is_composited(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, input_shape_combine_mask)
{
	GdkPixmap *shape_mask = NULL;
	zval *php_shape_mask;
	long offset_x, offset_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Nii", &php_shape_mask, gdkpixmap_ce, &offset_x, &offset_y))
		return;
    if (Z_TYPE_P(php_shape_mask) != IS_NULL)
        shape_mask = GDK_PIXMAP(PHPG_GOBJECT(php_shape_mask));

    gtk_widget_input_shape_combine_mask(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), shape_mask, (gint)offset_x, (gint)offset_y);

}


static PHP_METHOD(GtkWidget, keynav_failed)
{
	GtkDirectionType direction;
	zval *php_direction = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_direction))
		return;

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_DIRECTION_TYPE, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

    php_retval = gtk_widget_keynav_failed(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), direction);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, error_bell)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_error_bell(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, set_tooltip_window)
{
	GtkWindow *custom_window = NULL;
	zval *php_custom_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_custom_window, gtkwindow_ce))
		return;
    if (Z_TYPE_P(php_custom_window) != IS_NULL)
        custom_window = GTK_WINDOW(PHPG_GOBJECT(php_custom_window));

    gtk_widget_set_tooltip_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), custom_window);

}


static PHP_METHOD(GtkWidget, get_tooltip_window)
{
	GtkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_tooltip_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, trigger_tooltip_query)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_trigger_tooltip_query(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, set_tooltip_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_widget_set_tooltip_text(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkWidget, get_tooltip_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_tooltip_text(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkWidget, set_tooltip_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup))
		return;

    gtk_widget_set_tooltip_markup(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkWidget, get_tooltip_markup)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_tooltip_markup(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkWidget, modify_cursor)
{
	GdkColor *primary = NULL, *secondary = NULL;
	zval *php_primary, *php_secondary;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "NN", &php_primary, gboxed_ce, &php_secondary, gboxed_ce))
		return;

    if (Z_TYPE_P(php_primary) != IS_NULL) {
        if (phpg_gboxed_check(php_primary, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            primary = (GdkColor *) PHPG_GBOXED(php_primary);
        } else {
            php_error(E_WARNING, "%s::%s() expects primary argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    if (Z_TYPE_P(php_secondary) != IS_NULL) {
        if (phpg_gboxed_check(php_secondary, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
            secondary = (GdkColor *) PHPG_GBOXED(php_secondary);
        } else {
            php_error(E_WARNING, "%s::%s() expects secondary argument to be a valid GdkColor object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_widget_modify_cursor(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), primary, secondary);

}


static PHP_METHOD(GtkWidget, get_has_tooltip)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_has_tooltip(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, set_has_tooltip)
{
	zend_bool has_tooltip;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_tooltip))
		return;

    gtk_widget_set_has_tooltip(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)has_tooltip);

}


static PHP_METHOD(GtkWidget, get_snapshot)
{
	GdkRectangle clip_rect_arg = { 0, 0, 0, 0 }, *clip_rect;
	zval *php_clip_rect = NULL;
	GdkPixmap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|V", &php_clip_rect))
		return;

    if (Z_TYPE_P(php_clip_rect) == IS_NULL) {
        clip_rect = NULL;
    } else if (phpg_rectangle_from_zval(php_clip_rect, (GdkRectangle*)&clip_rect_arg TSRMLS_CC) == SUCCESS) {
        clip_rect = &clip_rect_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects clip_rect argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    php_retval = gtk_widget_get_snapshot(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), clip_rect);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWidget, get_app_paintable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_app_paintable(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_can_default)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_can_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_can_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_can_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_double_buffered)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_double_buffered(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_has_window)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_has_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_receives_default)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_receives_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_sensitive(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_state)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_state(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWidget, get_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_visible(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, has_default)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_has_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, has_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_has_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, has_grab)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_has_grab(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, is_drawable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_is_drawable(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, is_sensitive)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_is_sensitive(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, is_toplevel)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_is_toplevel(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 150 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkWidget, set_allocation)
{
    zval *php_allocation;
    GtkAllocation *allocation, allocation_arg = { 0, 0, 0, 0 };

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_allocation)) {
        return;
    }

    if (Z_TYPE_P(php_allocation) == IS_NULL) {
        allocation = &allocation_arg;
    } else if (phpg_rectangle_from_zval(php_allocation, (GtkAllocation*)&allocation_arg TSRMLS_CC) == SUCCESS) {
        allocation = &allocation_arg;
    } else {
        php_error(E_WARNING, "%s::%s() expects allocation argument to be a 4-element array, a GdkRectangle object, or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_widget_set_allocation(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), allocation);
}

#line 29916 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWidget, set_can_default)
{
	zend_bool can_default;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &can_default))
		return;

    gtk_widget_set_can_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)can_default);

}


static PHP_METHOD(GtkWidget, set_can_focus)
{
	zend_bool can_focus;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &can_focus))
		return;

    gtk_widget_set_can_focus(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)can_focus);

}


static PHP_METHOD(GtkWidget, set_has_window)
{
	zend_bool has_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_window))
		return;

    gtk_widget_set_has_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)has_window);

}


static PHP_METHOD(GtkWidget, set_receives_default)
{
	zend_bool receives_default;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &receives_default))
		return;

    gtk_widget_set_receives_default(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)receives_default);

}


static PHP_METHOD(GtkWidget, set_window)
{
	zval *window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gdkwindow_ce))
		return;

    gtk_widget_set_window(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)));

}


static PHP_METHOD(GtkWidget, set_realized)
{
	zend_bool realized;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &realized))
		return;

    gtk_widget_set_realized(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)realized);

}


static PHP_METHOD(GtkWidget, get_realized)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_realized(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, set_mapped)
{
	zend_bool mapped;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &mapped))
		return;

    gtk_widget_set_mapped(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), (gboolean)mapped);

}


static PHP_METHOD(GtkWidget, get_mapped)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_get_mapped(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, get_requisition)
{
	GtkRequisition *requisition = NULL;
	zval *php_requisition;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_requisition, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_requisition, GTK_TYPE_REQUISITION, FALSE TSRMLS_CC)) {
        requisition = (GtkRequisition *) PHPG_GBOXED(php_requisition);
    } else {
        php_error(E_WARNING, "%s::%s() expects requisition argument to be a valid GtkRequisition object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_widget_get_requisition(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), requisition);

}


static PHP_METHOD(GtkWidget, style_attach)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_widget_style_attach(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWidget, has_rc_style)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_widget_has_rc_style(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWidget, send_focus_change)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_widget_send_focus_change(GTK_WIDGET(PHPG_GOBJECT(this_ptr)), event);
	RETVAL_BOOL(php_retval);
}

#line 6029 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, set_all_visible)
{
    zend_bool visible;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible)) {
        return;
    }

    if (visible) {
        gtk_widget_show_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    } else {
        gtk_widget_hide_all(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
    }
}

#line 30137 "ext/gtk+/gen_gtk.c"


#line 125 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkWidget, is_visible)
{
    zend_bool visible;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }
    phpg_warn_deprecated("since GTK+ 2.18, please use GtkWidget::get_visible()" TSRMLS_CC);

    visible = GTK_WIDGET_VISIBLE(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));

    RETURN_BOOL(visible);
}

#line 30157 "ext/gtk+/gen_gtk.c"


#line 6131 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWidget, get_for_attach_widget)
{
	 GList *list, *item;
	
    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

	list = gtk_menu_get_for_attach_widget(GTK_WIDGET(PHPG_GOBJECT(this_ptr)));
	
	array_init(return_value);
	for (item = list; item; item = item->next) {
        zval *php_item = NULL;
        phpg_gobject_new(&php_item, G_OBJECT(item->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }
}

#line 30180 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_begin, 0)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
    ZEND_ARG_INFO(0, button)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_check_threshold, 0)
    ZEND_ARG_INFO(0, start_x)
    ZEND_ARG_INFO(0, start_y)
    ZEND_ARG_INFO(0, current_x)
    ZEND_ARG_INFO(0, current_y)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_drag_dest_find_target, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, context, GdkDragContext, 1)
    ZEND_ARG_INFO(0, targets)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_dest_set, 0)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_dest_set_proxy, 0)
    ZEND_ARG_OBJ_INFO(0, proxy_window, GdkWindow, 1)
    ZEND_ARG_INFO(0, protocol)
    ZEND_ARG_INFO(0, use_coordinates)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_dest_set_target_list, 0)
    ZEND_ARG_INFO(0, targets)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_drag_get_data, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, context, GdkDragContext, 1)
    ZEND_ARG_INFO(0, target)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_source_set, 0)
    ZEND_ARG_INFO(0, sbmask)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_drag_source_set_icon, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, colormap, GdkColormap, 1)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_source_set_icon_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_source_set_icon_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_source_set_target_list, 0)
    ZEND_ARG_INFO(0, targets)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_selection_add_target, 0)
    ZEND_ARG_INFO(0, selection)
    ZEND_ARG_INFO(0, target)
    ZEND_ARG_INFO(0, info)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_selection_clear, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_selection_clear_targets, 0)
    ZEND_ARG_INFO(0, selection)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_selection_convert, 0, 0, 2)
    ZEND_ARG_INFO(0, selection)
    ZEND_ARG_INFO(0, target)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_selection_owner_set, 0, 0, 1)
    ZEND_ARG_INFO(0, selection)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_add_accelerator, 0)
    ZEND_ARG_INFO(0, accel_signal)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
    ZEND_ARG_INFO(0, accel_flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_add_events, 0)
    ZEND_ARG_INFO(0, events)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_add_mnemonic_label, 0)
    ZEND_ARG_OBJ_INFO(0, label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_can_activate_accel, 0)
    ZEND_ARG_INFO(0, signal_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_child_focus, 0)
    ZEND_ARG_INFO(0, direction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_child_notify, 0)
    ZEND_ARG_INFO(0, child_property)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_create_pango_layout, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_draw, 0)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_event, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_get_ancestor, 0)
    ZEND_ARG_OBJ_INFO(0, widget_type, GType, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_get_clipboard, 0)
    ZEND_ARG_INFO(0, selection)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_intersect, 0)
    ZEND_ARG_OBJ_INFO(0, area, GdkRectangle, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_is_ancestor, 0)
    ZEND_ARG_OBJ_INFO(0, ancestor, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_mnemonic_activate, 0)
    ZEND_ARG_INFO(0, group_cycling)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_base, 0)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_bg, 0)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_fg, 0)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_font, 0)
    ZEND_ARG_OBJ_INFO(0, font_desc, PangoFontDescription, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_style, 0)
    ZEND_ARG_OBJ_INFO(0, style, GtkRcStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_text, 0)
    ZEND_ARG_INFO(0, state)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_push_colormap, 0)
    ZEND_ARG_OBJ_INFO(0, cmap, GdkColormap, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_queue_clear_area, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_queue_draw_area, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_region_intersect, 0)
    ZEND_ARG_OBJ_INFO(0, region, GdkRegion, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_remove_accelerator, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
    ZEND_ARG_INFO(0, accel_key)
    ZEND_ARG_INFO(0, accel_mods)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_remove_mnemonic_label, 0)
    ZEND_ARG_OBJ_INFO(0, label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_render_icon, 0, 0, 2)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_INFO(0, size)
    ZEND_ARG_INFO(0, detail)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_reparent, 0)
    ZEND_ARG_OBJ_INFO(0, new_parent, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_send_expose, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_accel_path, 0)
    ZEND_ARG_INFO(0, accel_path)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_app_paintable, 0)
    ZEND_ARG_INFO(0, app_paintable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_child_visible, 0)
    ZEND_ARG_INFO(0, is_visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_colormap, 0)
    ZEND_ARG_OBJ_INFO(0, colormap, GdkColormap, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_composite_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_default_colormap, 0)
    ZEND_ARG_OBJ_INFO(0, colormap, GdkColormap, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_default_direction, 0)
    ZEND_ARG_INFO(0, dir)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_direction, 0)
    ZEND_ARG_INFO(0, dir)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_double_buffered, 0)
    ZEND_ARG_INFO(0, double_buffered)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_events, 0)
    ZEND_ARG_INFO(0, events)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_extension_events, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_no_show_all, 0)
    ZEND_ARG_INFO(0, no_show_all)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_parent, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_parent_window, 0)
    ZEND_ARG_OBJ_INFO(0, parent_window, GdkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_redraw_on_allocate, 0)
    ZEND_ARG_INFO(0, redraw_on_allocate)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_scroll_adjustments, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
    ZEND_ARG_OBJ_INFO(0, vadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_sensitive, 0)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_size_request, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_state, 0)
    ZEND_ARG_INFO(0, state)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_style, 0)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_uposition, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_usize, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_shape_combine_mask, 0)
    ZEND_ARG_INFO(0, shape_mask)
    ZEND_ARG_INFO(0, offset_x)
    ZEND_ARG_INFO(0, offset_y)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_size_allocate, 0)
    ZEND_ARG_INFO(0, allocation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_style_get_property, 0)
    ZEND_ARG_INFO(0, property_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_translate_coordinates, 0)
    ZEND_ARG_OBJ_INFO(0, dest_widget, GtkWidget, 0)
    ZEND_ARG_INFO(0, src_x)
    ZEND_ARG_INFO(0, src_y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_set_icon_name, 0)
    ZEND_ARG_OBJ_INFO(0, context, GdkDragContext, 1)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, hot_x)
    ZEND_ARG_INFO(0, hot_y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_source_set_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_drag_dest_set_track_motion, 0)
    ZEND_ARG_INFO(0, track_motion)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_input_shape_combine_mask, 0)
    ZEND_ARG_INFO(0, shape_mask)
    ZEND_ARG_INFO(0, offset_x)
    ZEND_ARG_INFO(0, offset_y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_keynav_failed, 0)
    ZEND_ARG_INFO(0, direction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_tooltip_window, 0)
    ZEND_ARG_OBJ_INFO(0, custom_window, GtkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_tooltip_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_tooltip_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_modify_cursor, 0)
    ZEND_ARG_OBJ_INFO(0, primary, GdkColor, 1)
    ZEND_ARG_OBJ_INFO(0, secondary, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_has_tooltip, 0)
    ZEND_ARG_INFO(0, has_tooltip)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwidget_get_snapshot, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, clip_rect, GdkRectangle, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_allocation, 0)
    ZEND_ARG_INFO(0, allocation)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_can_default, 0)
    ZEND_ARG_INFO(0, can_default)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_can_focus, 0)
    ZEND_ARG_INFO(0, can_focus)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_has_window, 0)
    ZEND_ARG_INFO(0, has_window)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_receives_default, 0)
    ZEND_ARG_INFO(0, receives_default)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_window, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_realized, 0)
    ZEND_ARG_INFO(0, realized)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_mapped, 0)
    ZEND_ARG_INFO(0, mapped)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_get_requisition, 0)
    ZEND_ARG_OBJ_INFO(0, requisition, GtkRequisition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_send_focus_change, 0)
    ZEND_ARG_OBJ_INFO(0, event, GdkEvent, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwidget_set_all_visible, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();


static zend_function_entry gtkwidget_methods[] = {
#if ZEND_EXTENSION_API_NO > 220051025
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL, 0)
#else
	PHP_ME_MAPPING(__construct, no_direct_constructor, NULL)
#endif
	PHP_ME(GtkWidget, activate,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, add_accelerator,      arginfo_gtk_gtkwidget_add_accelerator, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, add_events,           arginfo_gtk_gtkwidget_add_events, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, add_mnemonic_label,   arginfo_gtk_gtkwidget_add_mnemonic_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, can_activate_accel,   arginfo_gtk_gtkwidget_can_activate_accel, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, child_focus,          arginfo_gtk_gtkwidget_child_focus, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, child_notify,         arginfo_gtk_gtkwidget_child_notify, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, class_path,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, create_pango_context, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, create_pango_layout,  arginfo_gtk_gtkwidget_create_pango_layout, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, destroy,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_begin,           arginfo_gtk_gtkwidget_drag_begin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_check_threshold, arginfo_gtk_gtkwidget_drag_check_threshold, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_add_image_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_add_text_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_add_uri_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_find_target, arginfo_gtk_gtkwidget_drag_dest_find_target, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_get_target_list, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_get_track_motion, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_set,        arginfo_gtk_gtkwidget_drag_dest_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_set_proxy,  arginfo_gtk_gtkwidget_drag_dest_set_proxy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_set_target_list, arginfo_gtk_gtkwidget_drag_dest_set_target_list, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_set_track_motion, arginfo_gtk_gtkwidget_drag_dest_set_track_motion, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_dest_unset,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_get_data,        arginfo_gtk_gtkwidget_drag_get_data, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_highlight,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_set_icon_name,   arginfo_gtk_gtkwidget_drag_set_icon_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, drag_source_add_image_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_add_text_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_add_uri_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_get_target_list, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set,      arginfo_gtk_gtkwidget_drag_source_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set_icon, arginfo_gtk_gtkwidget_drag_source_set_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set_icon_name, arginfo_gtk_gtkwidget_drag_source_set_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set_icon_pixbuf, arginfo_gtk_gtkwidget_drag_source_set_icon_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set_icon_stock, arginfo_gtk_gtkwidget_drag_source_set_icon_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_set_target_list, arginfo_gtk_gtkwidget_drag_source_set_target_list, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_source_unset,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, drag_unhighlight,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, draw,                 arginfo_gtk_gtkwidget_draw, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, ensure_style,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, error_bell,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, event,                arginfo_gtk_gtkwidget_event, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, freeze_child_notify,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_accessible,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_action,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_allocation,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_ancestor,         arginfo_gtk_gtkwidget_get_ancestor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_app_paintable,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_can_default,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_can_focus,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_child_requisition, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_child_visible,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_clipboard,        arginfo_gtk_gtkwidget_get_clipboard, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_colormap,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_composite_name,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_default_colormap, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, get_default_direction, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, get_default_style,    NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, get_default_visual,   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, get_direction,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_display,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_double_buffered,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_events,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_extension_events, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_for_attach_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_has_tooltip,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_has_window,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_mapped,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_modifier_style,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_no_show_all,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_pango_context,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_parent,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_parent_window,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_pointer,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_realized,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_receives_default, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_requisition,      arginfo_gtk_gtkwidget_get_requisition, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_root_window,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_sensitive,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_settings,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_size_request,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_snapshot,         arginfo_gtk_gtkwidget_get_snapshot, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_state,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_style,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_tooltip_markup,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_tooltip_text,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_tooltip_window,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_toplevel,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_visible,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_visual,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, get_window,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, grab_add,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, grab_default,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, grab_focus,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, grab_remove,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, has_default,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, has_focus,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, has_grab,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, has_rc_style,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, has_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, hide,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, hide_all,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, hide_on_delete,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, input_shape_combine_mask, arginfo_gtk_gtkwidget_input_shape_combine_mask, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, intersect,            arginfo_gtk_gtkwidget_intersect, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_ancestor,          arginfo_gtk_gtkwidget_is_ancestor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_composited,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_drawable,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_focus,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_sensitive,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_toplevel,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, is_visible,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, keynav_failed,        arginfo_gtk_gtkwidget_keynav_failed, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, list_mnemonic_labels, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, map,                  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, mnemonic_activate,    arginfo_gtk_gtkwidget_mnemonic_activate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_base,          arginfo_gtk_gtkwidget_modify_base, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_bg,            arginfo_gtk_gtkwidget_modify_bg, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_cursor,        arginfo_gtk_gtkwidget_modify_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_fg,            arginfo_gtk_gtkwidget_modify_fg, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_font,          arginfo_gtk_gtkwidget_modify_font, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_style,         arginfo_gtk_gtkwidget_modify_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, modify_text,          arginfo_gtk_gtkwidget_modify_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, path,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, pop_colormap,         NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, pop_composite_child,  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, push_colormap,        arginfo_gtk_gtkwidget_push_colormap, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, push_composite_child, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, queue_clear,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, queue_clear_area,     arginfo_gtk_gtkwidget_queue_clear_area, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, queue_draw,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, queue_draw_area,      arginfo_gtk_gtkwidget_queue_draw_area, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, queue_resize,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, queue_resize_no_redraw, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, rc_get_style,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, realize,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, region_intersect,     arginfo_gtk_gtkwidget_region_intersect, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, remove_accelerator,   arginfo_gtk_gtkwidget_remove_accelerator, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, remove_mnemonic_label, arginfo_gtk_gtkwidget_remove_mnemonic_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, render_icon,          arginfo_gtk_gtkwidget_render_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, reparent,             arginfo_gtk_gtkwidget_reparent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, reset_rc_styles,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, reset_shapes,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_add_target, arginfo_gtk_gtkwidget_selection_add_target, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_add_targets, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_clear,      arginfo_gtk_gtkwidget_selection_clear, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_clear_targets, arginfo_gtk_gtkwidget_selection_clear_targets, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_convert,    arginfo_gtk_gtkwidget_selection_convert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_owner_set,  arginfo_gtk_gtkwidget_selection_owner_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, selection_remove_all, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, send_expose,          arginfo_gtk_gtkwidget_send_expose, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, send_focus_change,    arginfo_gtk_gtkwidget_send_focus_change, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_accel_path,       arginfo_gtk_gtkwidget_set_accel_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_all_visible,      arginfo_gtk_gtkwidget_set_all_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_allocation,       arginfo_gtk_gtkwidget_set_allocation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_app_paintable,    arginfo_gtk_gtkwidget_set_app_paintable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_can_default,      arginfo_gtk_gtkwidget_set_can_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_can_focus,        arginfo_gtk_gtkwidget_set_can_focus, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_child_visible,    arginfo_gtk_gtkwidget_set_child_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_colormap,         arginfo_gtk_gtkwidget_set_colormap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_composite_name,   arginfo_gtk_gtkwidget_set_composite_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_default_colormap, arginfo_gtk_gtkwidget_set_default_colormap, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, set_default_direction, arginfo_gtk_gtkwidget_set_default_direction, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWidget, set_direction,        arginfo_gtk_gtkwidget_set_direction, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_double_buffered,  arginfo_gtk_gtkwidget_set_double_buffered, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_events,           arginfo_gtk_gtkwidget_set_events, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_extension_events, arginfo_gtk_gtkwidget_set_extension_events, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_has_tooltip,      arginfo_gtk_gtkwidget_set_has_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_has_window,       arginfo_gtk_gtkwidget_set_has_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_mapped,           arginfo_gtk_gtkwidget_set_mapped, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_name,             arginfo_gtk_gtkwidget_set_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_no_show_all,      arginfo_gtk_gtkwidget_set_no_show_all, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_parent,           arginfo_gtk_gtkwidget_set_parent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_parent_window,    arginfo_gtk_gtkwidget_set_parent_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_realized,         arginfo_gtk_gtkwidget_set_realized, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_receives_default, arginfo_gtk_gtkwidget_set_receives_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_redraw_on_allocate, arginfo_gtk_gtkwidget_set_redraw_on_allocate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_scroll_adjustments, arginfo_gtk_gtkwidget_set_scroll_adjustments, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_sensitive,        arginfo_gtk_gtkwidget_set_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_size_request,     arginfo_gtk_gtkwidget_set_size_request, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_state,            arginfo_gtk_gtkwidget_set_state, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_style,            arginfo_gtk_gtkwidget_set_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_tooltip_markup,   arginfo_gtk_gtkwidget_set_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_tooltip_text,     arginfo_gtk_gtkwidget_set_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_tooltip_window,   arginfo_gtk_gtkwidget_set_tooltip_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_uposition,        arginfo_gtk_gtkwidget_set_uposition, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_usize,            arginfo_gtk_gtkwidget_set_usize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_visible,          arginfo_gtk_gtkwidget_set_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, set_window,           arginfo_gtk_gtkwidget_set_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, shape_combine_mask,   arginfo_gtk_gtkwidget_shape_combine_mask, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, show,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, show_all,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, show_now,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, size_allocate,        arginfo_gtk_gtkwidget_size_allocate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, size_request,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, style_attach,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, style_get_property,   arginfo_gtk_gtkwidget_style_get_property, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, thaw_child_notify,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, translate_coordinates, arginfo_gtk_gtkwidget_translate_coordinates, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, trigger_tooltip_query, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, unmap,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, unparent,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWidget, unrealize,            NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRuler, draw_pos)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_ruler_draw_pos(GTK_RULER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkRuler, draw_ticks)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_ruler_draw_ticks(GTK_RULER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkRuler, get_metric)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_ruler_get_metric(GTK_RULER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 5173 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRuler, get_range)
{
    gdouble lower, upper, position, max_size;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_ruler_get_range(GTK_RULER(PHPG_GOBJECT(this_ptr)), &lower, &upper, &position, &max_size);
    php_gtk_build_value(&return_value, "(dddd)", lower, upper, position, max_size);
}

#line 30987 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRuler, set_metric)
{
	GtkMetricType metric;
	zval *php_metric = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_metric))
		return;

	if (php_metric && phpg_gvalue_get_enum(GTK_TYPE_METRIC_TYPE, php_metric, (gint *)&metric) == FAILURE) {
		return;
	}

    gtk_ruler_set_metric(GTK_RULER(PHPG_GOBJECT(this_ptr)), metric);

}


static PHP_METHOD(GtkRuler, set_range)
{
	double lower, upper, position, max_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dddd", &lower, &upper, &position, &max_size))
		return;

    gtk_ruler_set_range(GTK_RULER(PHPG_GOBJECT(this_ptr)), lower, upper, position, max_size);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkruler_set_metric, 0)
    ZEND_ARG_INFO(0, metric)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkruler_set_range, 0)
    ZEND_ARG_INFO(0, lower)
    ZEND_ARG_INFO(0, upper)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, max_size)
ZEND_END_ARG_INFO();

static zend_function_entry gtkruler_methods[] = {
	PHP_ME(GtkRuler, draw_pos,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRuler, draw_ticks,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRuler, get_metric,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRuler, get_range,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRuler, set_metric,           arginfo_gtk_gtkruler_set_metric, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRuler, set_range,            arginfo_gtk_gtkruler_set_range, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVSeparator, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVSeparator);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVSeparator);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkvseparator_methods[] = {
	PHP_ME(GtkVSeparator, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHSeparator, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHSeparator);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHSeparator);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkhseparator_methods[] = {
	PHP_ME(GtkHSeparator, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVRuler, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVRuler);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVRuler);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkvruler_methods[] = {
	PHP_ME(GtkVRuler, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRange, get_adjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_adjustment(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkRange, get_inverted)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_inverted(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRange, get_update_policy)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_update_policy(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRange, get_value)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_value(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkRange, set_adjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_range_set_adjustment(GTK_RANGE(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkRange, set_increments)
{
	double step, page;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &step, &page))
		return;

    gtk_range_set_increments(GTK_RANGE(PHPG_GOBJECT(this_ptr)), step, page);

}


static PHP_METHOD(GtkRange, set_inverted)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_range_set_inverted(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkRange, set_range)
{
	double min, max;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &min, &max))
		return;

    gtk_range_set_range(GTK_RANGE(PHPG_GOBJECT(this_ptr)), min, max);

}


static PHP_METHOD(GtkRange, set_update_policy)
{
	GtkUpdateType policy;
	zval *php_policy = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_policy))
		return;

	if (php_policy && phpg_gvalue_get_enum(GTK_TYPE_UPDATE_TYPE, php_policy, (gint *)&policy) == FAILURE) {
		return;
	}

    gtk_range_set_update_policy(GTK_RANGE(PHPG_GOBJECT(this_ptr)), policy);

}


static PHP_METHOD(GtkRange, set_value)
{
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

    gtk_range_set_value(GTK_RANGE(PHPG_GOBJECT(this_ptr)), value);

}


static PHP_METHOD(GtkRange, set_lower_stepper_sensitivity)
{
	GtkSensitivityType sensitivity;
	zval *php_sensitivity = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sensitivity))
		return;

	if (php_sensitivity && phpg_gvalue_get_enum(GTK_TYPE_SENSITIVITY_TYPE, php_sensitivity, (gint *)&sensitivity) == FAILURE) {
		return;
	}

    gtk_range_set_lower_stepper_sensitivity(GTK_RANGE(PHPG_GOBJECT(this_ptr)), sensitivity);

}


static PHP_METHOD(GtkRange, get_lower_stepper_sensitivity)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_lower_stepper_sensitivity(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRange, set_upper_stepper_sensitivity)
{
	GtkSensitivityType sensitivity;
	zval *php_sensitivity = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sensitivity))
		return;

	if (php_sensitivity && phpg_gvalue_get_enum(GTK_TYPE_SENSITIVITY_TYPE, php_sensitivity, (gint *)&sensitivity) == FAILURE) {
		return;
	}

    gtk_range_set_upper_stepper_sensitivity(GTK_RANGE(PHPG_GOBJECT(this_ptr)), sensitivity);

}


static PHP_METHOD(GtkRange, get_upper_stepper_sensitivity)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_upper_stepper_sensitivity(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRange, set_show_fill_level)
{
	zend_bool show_fill_level;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_fill_level))
		return;

    gtk_range_set_show_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)show_fill_level);

}


static PHP_METHOD(GtkRange, get_show_fill_level)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_show_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRange, set_restrict_to_fill_level)
{
	zend_bool restrict_to_fill_level;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &restrict_to_fill_level))
		return;

    gtk_range_set_restrict_to_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)restrict_to_fill_level);

}


static PHP_METHOD(GtkRange, get_restrict_to_fill_level)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_restrict_to_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRange, set_fill_level)
{
	double fill_level;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &fill_level))
		return;

    gtk_range_set_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)), fill_level);

}


static PHP_METHOD(GtkRange, get_fill_level)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_fill_level(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkRange, get_flippable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_flippable(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRange, set_flippable)
{
	zend_bool flippable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &flippable))
		return;

    gtk_range_set_flippable(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)flippable);

}


static PHP_METHOD(GtkRange, get_min_slider_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_min_slider_size(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkRange, get_range_rect)
{
	GdkRectangle range_rect = { 0, 0, 0, 0 };
	zval *php_range_rect;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_range_rect))
		return;

    if (phpg_rectangle_from_zval(php_range_rect, (GdkRectangle*)&range_rect TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects range_rect argument to be either a 4-element array or a GdkRectangle object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    gtk_range_get_range_rect(GTK_RANGE(PHPG_GOBJECT(this_ptr)), &range_rect);

}

#line 7 "ext/gtk+/gtk-2.20.overrides"
static PHP_METHOD(GtkRange, get_slider_range)
{
	gint slider_start, slider_end;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_range_get_slider_range(GTK_RANGE(PHPG_GOBJECT(this_ptr)), &slider_start,
            &slider_end);

    array_init(return_value);
    add_next_index_long(return_value, slider_start);
    add_next_index_long(return_value, slider_end);
}

#line 31507 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkRange, get_slider_size_fixed)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_slider_size_fixed(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRange, set_min_slider_size)
{
	zend_bool min_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &min_size))
		return;

    gtk_range_set_min_slider_size(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)min_size);

}


static PHP_METHOD(GtkRange, set_slider_size_fixed)
{
	zend_bool size_fixed;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &size_fixed))
		return;

    gtk_range_set_slider_size_fixed(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gboolean)size_fixed);

}


static PHP_METHOD(GtkRange, set_round_digits)
{
	long round_digits;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &round_digits))
		return;

    gtk_range_set_round_digits(GTK_RANGE(PHPG_GOBJECT(this_ptr)), (gint)round_digits);

}


static PHP_METHOD(GtkRange, get_round_digits)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_range_get_round_digits(GTK_RANGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_adjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_increments, 0)
    ZEND_ARG_INFO(0, step)
    ZEND_ARG_INFO(0, page)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_inverted, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_range, 0)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_update_policy, 0)
    ZEND_ARG_INFO(0, policy)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_lower_stepper_sensitivity, 0)
    ZEND_ARG_INFO(0, sensitivity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_upper_stepper_sensitivity, 0)
    ZEND_ARG_INFO(0, sensitivity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_show_fill_level, 0)
    ZEND_ARG_INFO(0, show_fill_level)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_restrict_to_fill_level, 0)
    ZEND_ARG_INFO(0, restrict_to_fill_level)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_fill_level, 0)
    ZEND_ARG_INFO(0, fill_level)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_flippable, 0)
    ZEND_ARG_INFO(0, flippable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_get_range_rect, 0)
    ZEND_ARG_OBJ_INFO(0, range_rect, GdkRectangle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_min_slider_size, 0)
    ZEND_ARG_INFO(0, min_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_slider_size_fixed, 0)
    ZEND_ARG_INFO(0, size_fixed)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrange_set_round_digits, 0)
    ZEND_ARG_INFO(0, round_digits)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrange_methods[] = {
	PHP_ME(GtkRange, get_adjustment,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_fill_level,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_flippable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_inverted,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_lower_stepper_sensitivity, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_min_slider_size,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_range_rect,       arginfo_gtk_gtkrange_get_range_rect, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_restrict_to_fill_level, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_round_digits,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_show_fill_level,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_slider_range,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_slider_size_fixed, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_update_policy,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_upper_stepper_sensitivity, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, get_value,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_adjustment,       arginfo_gtk_gtkrange_set_adjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_fill_level,       arginfo_gtk_gtkrange_set_fill_level, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_flippable,        arginfo_gtk_gtkrange_set_flippable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_increments,       arginfo_gtk_gtkrange_set_increments, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_inverted,         arginfo_gtk_gtkrange_set_inverted, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_lower_stepper_sensitivity, arginfo_gtk_gtkrange_set_lower_stepper_sensitivity, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_min_slider_size,  arginfo_gtk_gtkrange_set_min_slider_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_range,            arginfo_gtk_gtkrange_set_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_restrict_to_fill_level, arginfo_gtk_gtkrange_set_restrict_to_fill_level, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_round_digits,     arginfo_gtk_gtkrange_set_round_digits, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_show_fill_level,  arginfo_gtk_gtkrange_set_show_fill_level, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_slider_size_fixed, arginfo_gtk_gtkrange_set_slider_size_fixed, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_update_policy,    arginfo_gtk_gtkrange_set_update_policy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_upper_stepper_sensitivity, arginfo_gtk_gtkrange_set_upper_stepper_sensitivity, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRange, set_value,            arginfo_gtk_gtkrange_set_value, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHRuler, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHRuler);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHRuler);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkhruler_methods[] = {
	PHP_ME(GtkHRuler, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkProgress, configure)
{
	double value, min, max;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddd", &value, &min, &max))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_configure(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), value, min, max);

}


static PHP_METHOD(GtkProgress, get_current_percentage)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    php_retval = gtk_progress_get_current_percentage(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkProgress, get_current_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    php_retval = gtk_progress_get_current_text(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkProgress, get_percentage_from_value)
{
	double value, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    php_retval = gtk_progress_get_percentage_from_value(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), value);
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkProgress, get_text_from_value)
{
	double value;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    php_retval = gtk_progress_get_text_from_value(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), value);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkProgress, get_value)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    php_retval = gtk_progress_get_value(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkProgress, set_activity_mode)
{
	zend_bool activity_mode;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &activity_mode))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_activity_mode(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), (gboolean)activity_mode);

}


static PHP_METHOD(GtkProgress, set_adjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_adjustment(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkProgress, set_format_string)
{
	char *format;
	zend_bool free_format = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &format, &free_format))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_format_string(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), format);
	if (free_format) g_free(format);

}


static PHP_METHOD(GtkProgress, set_percentage)
{
	double percentage;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &percentage))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_percentage(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), percentage);

}


static PHP_METHOD(GtkProgress, set_show_text)
{
	zend_bool show_text;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_text))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_show_text(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), (gboolean)show_text);

}


static PHP_METHOD(GtkProgress, set_text_alignment)
{
	double x_align, y_align;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &x_align, &y_align))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_text_alignment(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), (float)x_align, (float)y_align);

}


static PHP_METHOD(GtkProgress, set_value)
{
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

	phpg_warn_deprecated("use GtkProgressBar" TSRMLS_CC);

    gtk_progress_set_value(GTK_PROGRESS(PHPG_GOBJECT(this_ptr)), value);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_configure, 0)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_get_percentage_from_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_get_text_from_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_activity_mode, 0)
    ZEND_ARG_INFO(0, activity_mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_adjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_format_string, 0)
    ZEND_ARG_INFO(0, format)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_percentage, 0)
    ZEND_ARG_INFO(0, percentage)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_show_text, 0)
    ZEND_ARG_INFO(0, show_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_text_alignment, 0)
    ZEND_ARG_INFO(0, x_align)
    ZEND_ARG_INFO(0, y_align)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogress_set_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

static zend_function_entry gtkprogress_methods[] = {
	PHP_ME(GtkProgress, configure,            arginfo_gtk_gtkprogress_configure, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, get_current_percentage, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, get_current_text,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, get_percentage_from_value, arginfo_gtk_gtkprogress_get_percentage_from_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, get_text_from_value,  arginfo_gtk_gtkprogress_get_text_from_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, get_value,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_activity_mode,    arginfo_gtk_gtkprogress_set_activity_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_adjustment,       arginfo_gtk_gtkprogress_set_adjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_format_string,    arginfo_gtk_gtkprogress_set_format_string, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_percentage,       arginfo_gtk_gtkprogress_set_percentage, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_show_text,        arginfo_gtk_gtkprogress_set_show_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_text_alignment,   arginfo_gtk_gtkprogress_set_text_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgress, set_value,            arginfo_gtk_gtkprogress_set_value, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVScrollbar, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "adjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScrollbar);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScrollbar);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScrollbar);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkvscrollbar_gtk_vscrollbar_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkvscrollbar_methods[] = {
	PHP_ME(GtkVScrollbar, __construct,          arginfo_gtk_gtkvscrollbar_gtk_vscrollbar_new, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHScrollbar, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "adjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScrollbar);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScrollbar);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScrollbar);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkhscrollbar_gtk_hscrollbar_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkhscrollbar_methods[] = {
	PHP_ME(GtkHScrollbar, __construct,          arginfo_gtk_gtkhscrollbar_gtk_hscrollbar_new, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkScale, get_digits)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_get_digits(GTK_SCALE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkScale, get_draw_value)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_get_draw_value(GTK_SCALE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkScale, get_layout)
{
	PangoLayout* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_get_layout(GTK_SCALE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 5192 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkScale, get_layout_offsets)
{
    gint x, y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_scale_get_layout_offsets(GTK_SCALE(PHPG_GOBJECT(this_ptr)), &x, &y);
    php_gtk_build_value(&return_value, "(ii)", x, y);
}

#line 32180 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkScale, get_value_pos)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_get_value_pos(GTK_SCALE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkScale, set_digits)
{
	long digits;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &digits))
		return;

    gtk_scale_set_digits(GTK_SCALE(PHPG_GOBJECT(this_ptr)), (gint)digits);

}


static PHP_METHOD(GtkScale, set_draw_value)
{
	zend_bool draw_value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &draw_value))
		return;

    gtk_scale_set_draw_value(GTK_SCALE(PHPG_GOBJECT(this_ptr)), (gboolean)draw_value);

}


static PHP_METHOD(GtkScale, set_value_pos)
{
	GtkPositionType pos;
	zval *php_pos = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_pos))
		return;

	if (php_pos && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_pos, (gint *)&pos) == FAILURE) {
		return;
	}

    gtk_scale_set_value_pos(GTK_SCALE(PHPG_GOBJECT(this_ptr)), pos);

}


static PHP_METHOD(GtkScale, add_mark)
{
	double value;
	GtkPositionType position;
	zval *php_position = NULL;
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dVu", &value, &php_position, &markup, &free_markup))
		return;

	if (php_position && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_position, (gint *)&position) == FAILURE) {
		return;
	}

    gtk_scale_add_mark(GTK_SCALE(PHPG_GOBJECT(this_ptr)), value, position, markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkScale, clear_marks)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_scale_clear_marks(GTK_SCALE(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscale_set_digits, 0)
    ZEND_ARG_INFO(0, digits)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscale_set_draw_value, 0)
    ZEND_ARG_INFO(0, draw_value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscale_set_value_pos, 0)
    ZEND_ARG_INFO(0, pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscale_add_mark, 0)
    ZEND_ARG_INFO(0, value)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

static zend_function_entry gtkscale_methods[] = {
	PHP_ME(GtkScale, add_mark,             arginfo_gtk_gtkscale_add_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, clear_marks,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, get_digits,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, get_draw_value,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, get_layout,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, get_layout_offsets,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, get_value_pos,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, set_digits,           arginfo_gtk_gtkscale_set_digits, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, set_draw_value,       arginfo_gtk_gtkscale_set_draw_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScale, set_value_pos,        arginfo_gtk_gtkscale_set_value_pos, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPreview, __construct)
{
	GtkPreviewType type;
	zval *php_type = NULL;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPreview);
	}

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_PREVIEW_TYPE, php_type, (gint *)&type) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPreview);
	}

	phpg_warn_deprecated("use GtkDrawingArea" TSRMLS_CC);

	wrapped_obj = (GObject *) gtk_preview_new(type);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPreview);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkPreview, draw_row)
{
	char *data;
	zend_bool free_data = FALSE;
	long x, y, w;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uiii", &data, &free_data, &x, &y, &w))
		return;

    gtk_preview_draw_row(GTK_PREVIEW(PHPG_GOBJECT(this_ptr)), data, (gint)x, (gint)y, (gint)w);
	if (free_data) g_free(data);

}


static PHP_METHOD(GtkPreview, get_cmap)
{
	GdkColormap* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    php_retval = gtk_preview_get_cmap();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPreview, get_preview_visual)
{
	GdkVisual* php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    php_retval = gtk_preview_get_visual();
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPreview, put)
{
	zval *window, *gc;
	long srcx, srcy, destx, desty, width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOiiiiii", &window, gdkwindow_ce, &gc, gdkgc_ce, &srcx, &srcy, &destx, &desty, &width, &height))
		return;

    gtk_preview_put(GTK_PREVIEW(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)), GDK_GC(PHPG_GOBJECT(gc)), (gint)srcx, (gint)srcy, (gint)destx, (gint)desty, (gint)width, (gint)height);

}


static PHP_METHOD(GtkPreview, reset)
{

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_reset();

}


static PHP_METHOD(GtkPreview, set_color_cube)
{
	long nred_shades, ngreen_shades, nblue_shades, ngray_shades;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &nred_shades, &ngreen_shades, &nblue_shades, &ngray_shades))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_set_color_cube((guint)nred_shades, (guint)ngreen_shades, (guint)nblue_shades, (guint)ngray_shades);

}


static PHP_METHOD(GtkPreview, set_dither)
{
	GdkRgbDither dither;
	zval *php_dither = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_dither))
		return;

	if (php_dither && phpg_gvalue_get_enum(GDK_TYPE_RGB_DITHER, php_dither, (gint *)&dither) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_set_dither(GTK_PREVIEW(PHPG_GOBJECT(this_ptr)), dither);

}


static PHP_METHOD(GtkPreview, set_expand)
{
	zend_bool expand;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &expand))
		return;

    gtk_preview_set_expand(GTK_PREVIEW(PHPG_GOBJECT(this_ptr)), (gboolean)expand);

}


static PHP_METHOD(GtkPreview, set_gamma)
{
	double gamma;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &gamma))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_set_gamma(gamma);

}


static PHP_METHOD(GtkPreview, set_install_cmap)
{
	long install_cmap;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &install_cmap))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_set_install_cmap((gint)install_cmap);

}


static PHP_METHOD(GtkPreview, set_reserved)
{
	long nreserved;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &nreserved))
		return;

	phpg_warn_deprecated("use GtkPixbuf" TSRMLS_CC);

    gtk_preview_set_reserved((gint)nreserved);

}


static PHP_METHOD(GtkPreview, size)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_preview_size(GTK_PREVIEW(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_gtk_preview_new, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_draw_row, 0)
    ZEND_ARG_INFO(0, data)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
    ZEND_ARG_INFO(0, w)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_put, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
    ZEND_ARG_OBJ_INFO(0, gc, GdkGC, 1)
    ZEND_ARG_INFO(0, srcx)
    ZEND_ARG_INFO(0, srcy)
    ZEND_ARG_INFO(0, destx)
    ZEND_ARG_INFO(0, desty)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_color_cube, 0)
    ZEND_ARG_INFO(0, nred_shades)
    ZEND_ARG_INFO(0, ngreen_shades)
    ZEND_ARG_INFO(0, nblue_shades)
    ZEND_ARG_INFO(0, ngray_shades)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_dither, 0)
    ZEND_ARG_INFO(0, dither)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_expand, 0)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_gamma, 0)
    ZEND_ARG_INFO(0, gamma)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_install_cmap, 0)
    ZEND_ARG_INFO(0, install_cmap)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_set_reserved, 0)
    ZEND_ARG_INFO(0, nreserved)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpreview_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

static zend_function_entry gtkpreview_methods[] = {
	PHP_ME(GtkPreview, __construct,          arginfo_gtk_gtkpreview_gtk_preview_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPreview, draw_row,             arginfo_gtk_gtkpreview_draw_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPreview, get_cmap,             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, get_preview_visual,   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, put,                  arginfo_gtk_gtkpreview_put, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPreview, reset,                NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, set_color_cube,       arginfo_gtk_gtkpreview_set_color_cube, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, set_dither,           arginfo_gtk_gtkpreview_set_dither, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPreview, set_expand,           arginfo_gtk_gtkpreview_set_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPreview, set_gamma,            arginfo_gtk_gtkpreview_set_gamma, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, set_install_cmap,     arginfo_gtk_gtkpreview_set_install_cmap, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, set_reserved,         arginfo_gtk_gtkpreview_set_reserved, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkPreview, size,                 arginfo_gtk_gtkpreview_size, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVScale, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "adjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScale);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScale);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScale);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkVScale, new_with_range)
{
	double min, max, step;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddd", &min, &max, &step)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScale);
	}

	wrapped_obj = (GObject *) gtk_vscale_new_with_range(min, max, step);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVScale);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkvscale_gtk_vscale_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkvscale_gtk_vscale_new_with_range, 0)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
    ZEND_ARG_INFO(0, step)
ZEND_END_ARG_INFO();

static zend_function_entry gtkvscale_methods[] = {
	PHP_ME(GtkVScale, __construct,          arginfo_gtk_gtkvscale_gtk_vscale_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkVScale, new_with_range,       arginfo_gtk_gtkvscale_gtk_vscale_new_with_range, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHScale, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "adjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScale);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScale);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScale);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkHScale, new_with_range)
{
	double min, max, step;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddd", &min, &max, &step)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScale);
	}

	wrapped_obj = (GObject *) gtk_hscale_new_with_range(min, max, step);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHScale);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkhscale_gtk_hscale_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhscale_gtk_hscale_new_with_range, 0)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
    ZEND_ARG_INFO(0, step)
ZEND_END_ARG_INFO();

static zend_function_entry gtkhscale_methods[] = {
	PHP_ME(GtkHScale, __construct,          arginfo_gtk_gtkhscale_gtk_hscale_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHScale, new_with_range,       arginfo_gtk_gtkhscale_gtk_hscale_new_with_range, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};
#line 4741 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkProgressBar, __construct)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_adjustment, gtkadjustment_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkProgressBar);
	}
    if (php_adjustment) {
        if (Z_TYPE_P(php_adjustment) == IS_NULL)
            adjustment = NULL;
        else
            adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));
    }

    if (adjustment) {
        wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "adjustment", adjustment, NULL);
    } else {
        wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkProgressBar);
	}
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 32781 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkProgressBar, get_ellipsize)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_progress_bar_get_ellipsize(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkProgressBar, get_fraction)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_progress_bar_get_fraction(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkProgressBar, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_progress_bar_get_orientation(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkProgressBar, get_pulse_step)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_progress_bar_get_pulse_step(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkProgressBar, get_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_progress_bar_get_text(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkProgressBar, pulse)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_progress_bar_pulse(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkProgressBar, set_activity_blocks)
{
	long blocks;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &blocks))
		return;

    gtk_progress_bar_set_activity_blocks(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), (guint)blocks);

}


static PHP_METHOD(GtkProgressBar, set_activity_step)
{
	long step;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &step))
		return;

    gtk_progress_bar_set_activity_step(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), (guint)step);

}


static PHP_METHOD(GtkProgressBar, set_bar_style)
{
	GtkProgressBarStyle style;
	zval *php_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_style))
		return;

	if (php_style && phpg_gvalue_get_enum(GTK_TYPE_PROGRESS_BAR_STYLE, php_style, (gint *)&style) == FAILURE) {
		return;
	}

    gtk_progress_bar_set_bar_style(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), style);

}


static PHP_METHOD(GtkProgressBar, set_discrete_blocks)
{
	long blocks;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &blocks))
		return;

    gtk_progress_bar_set_discrete_blocks(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), (guint)blocks);

}


static PHP_METHOD(GtkProgressBar, set_ellipsize)
{
	PangoEllipsizeMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(PANGO_TYPE_ELLIPSIZE_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_progress_bar_set_ellipsize(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkProgressBar, set_fraction)
{
	double fraction;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &fraction))
		return;

    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), fraction);

}


static PHP_METHOD(GtkProgressBar, set_orientation)
{
	GtkProgressBarOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_PROGRESS_BAR_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkProgressBar, set_pulse_step)
{
	double fraction;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &fraction))
		return;

    gtk_progress_bar_set_pulse_step(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), fraction);

}


static PHP_METHOD(GtkProgressBar, set_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkprogressbar_gtk_progress_bar_new_with_adjustment, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_activity_blocks, 0)
    ZEND_ARG_INFO(0, blocks)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_activity_step, 0)
    ZEND_ARG_INFO(0, step)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_bar_style, 0)
    ZEND_ARG_INFO(0, style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_discrete_blocks, 0)
    ZEND_ARG_INFO(0, blocks)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_ellipsize, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_fraction, 0)
    ZEND_ARG_INFO(0, fraction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_pulse_step, 0)
    ZEND_ARG_INFO(0, fraction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkprogressbar_set_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

static zend_function_entry gtkprogressbar_methods[] = {
	PHP_ME(GtkProgressBar, __construct,          arginfo_gtk_gtkprogressbar_gtk_progress_bar_new_with_adjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, get_ellipsize,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, get_fraction,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, get_orientation,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, get_pulse_step,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, pulse,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_activity_blocks,  arginfo_gtk_gtkprogressbar_set_activity_blocks, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_activity_step,    arginfo_gtk_gtkprogressbar_set_activity_step, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_bar_style,        arginfo_gtk_gtkprogressbar_set_bar_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_discrete_blocks,  arginfo_gtk_gtkprogressbar_set_discrete_blocks, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_ellipsize,        arginfo_gtk_gtkprogressbar_set_ellipsize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_fraction,         arginfo_gtk_gtkprogressbar_set_fraction, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_orientation,      arginfo_gtk_gtkprogressbar_set_orientation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_pulse_step,       arginfo_gtk_gtkprogressbar_set_pulse_step, ZEND_ACC_PUBLIC)
	PHP_ME(GtkProgressBar, set_text,             arginfo_gtk_gtkprogressbar_set_text, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkOldEditable, changed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_old_editable_changed(GTK_OLD_EDITABLE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkOldEditable, claim_selection)
{
	zend_bool claim;
	long time = GDK_CURRENT_TIME;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b|i", &claim, &time))
		return;

    gtk_old_editable_claim_selection(GTK_OLD_EDITABLE(PHPG_GOBJECT(this_ptr)), (gboolean)claim, (guint32)time);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkoldeditable_claim_selection, 0, 0, 1)
    ZEND_ARG_INFO(0, claim)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();

static zend_function_entry gtkoldeditable_methods[] = {
	PHP_ME(GtkOldEditable, changed,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOldEditable, claim_selection,      arginfo_gtk_gtkoldeditable_claim_selection, ZEND_ACC_PUBLIC)

	/***   GtkEditable interface implementations   ***/

	PHP_MALIAS(GtkEditable, copy_clipboard, copy_clipboard, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, cut_clipboard, cut_clipboard, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, delete_selection, delete_selection, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, delete_text, delete_text, arginfo_gtk_gtkeditable_delete_text, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_chars, get_chars, arginfo_gtk_gtkeditable_get_chars, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_editable, get_editable, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_position, get_position, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_selection_bounds, get_selection_bounds, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, insert_text, insert_text, arginfo_gtk_gtkeditable_insert_text, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, paste_clipboard, paste_clipboard, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, select_region, select_region, arginfo_gtk_gtkeditable_select_region, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, set_editable, set_editable, arginfo_gtk_gtkeditable_set_editable, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, set_position, set_position, arginfo_gtk_gtkeditable_set_position, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4605 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMisc, get_alignment)
{
    gfloat xalign, yalign;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_misc_get_alignment(GTK_MISC(PHPG_GOBJECT(this_ptr)), &xalign, &yalign);
    php_gtk_build_value(&return_value, "(ff)", xalign, yalign);
}


#line 33165 "ext/gtk+/gen_gtk.c"


#line 4621 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMisc, get_padding)
{
    gint xpad, ypad;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_misc_get_padding(GTK_MISC(PHPG_GOBJECT(this_ptr)), &xpad, &ypad);
    php_gtk_build_value(&return_value, "(ii)", xpad, ypad);
}

#line 33182 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkMisc, set_alignment)
{
	double xalign, yalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &xalign, &yalign))
		return;

    gtk_misc_set_alignment(GTK_MISC(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign);

}


static PHP_METHOD(GtkMisc, set_padding)
{
	long xpad, ypad;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &xpad, &ypad))
		return;

    gtk_misc_set_padding(GTK_MISC(PHPG_GOBJECT(this_ptr)), (gint)xpad, (gint)ypad);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmisc_set_alignment, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmisc_set_padding, 0)
    ZEND_ARG_INFO(0, xpad)
    ZEND_ARG_INFO(0, ypad)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmisc_methods[] = {
	PHP_ME(GtkMisc, get_alignment,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMisc, get_padding,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMisc, set_alignment,        arginfo_gtk_gtkmisc_set_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMisc, set_padding,          arginfo_gtk_gtkmisc_set_padding, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkInvisible, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInvisible);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInvisible);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkInvisible, new_for_screen)
{
	zval *screen;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInvisible);
	}

	wrapped_obj = (GObject *) gtk_invisible_new_for_screen(GDK_SCREEN(PHPG_GOBJECT(screen)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInvisible);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkInvisible, get_screen)
{
	GdkScreen* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_invisible_get_screen(GTK_INVISIBLE(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkInvisible, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_invisible_set_screen(GTK_INVISIBLE(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinvisible_gtk_invisible_new_for_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkinvisible_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkinvisible_methods[] = {
	PHP_ME(GtkInvisible, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInvisible, new_for_screen,       arginfo_gtk_gtkinvisible_gtk_invisible_new_for_screen, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkInvisible, get_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkInvisible, set_screen,           arginfo_gtk_gtkinvisible_set_screen, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPixmap, __construct)
{
	zval *pixmap, *php_mask;
	GdkPixmap *mask = NULL;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPixmap);
	}
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

	phpg_warn_deprecated("use GtkImage" TSRMLS_CC);

	wrapped_obj = (GObject *) gtk_pixmap_new(GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPixmap);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 4707 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkPixmap, get)
{
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    zval *php_pixmap = NULL, *php_mask = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    phpg_warn_deprecated("use GtkImage class instead" TSRMLS_CC);

    gtk_pixmap_get(GTK_PIXMAP(PHPG_GOBJECT(this_ptr)), &pixmap, &mask);

    phpg_gobject_new(&php_pixmap, (GObject *)pixmap TSRMLS_CC);
    phpg_gobject_new(&php_mask  , (GObject *)mask TSRMLS_CC);

    php_gtk_build_value(&return_value, "(NN)", php_pixmap, php_mask);
}

#line 33373 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkPixmap, set)
{
	zval *val, *php_mask;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &val, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

    gtk_pixmap_set(GTK_PIXMAP(PHPG_GOBJECT(this_ptr)), GDK_PIXMAP(PHPG_GOBJECT(val)), mask);

}


static PHP_METHOD(GtkPixmap, set_build_insensitive)
{
	zend_bool build;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &build))
		return;

    gtk_pixmap_set_build_insensitive(GTK_PIXMAP(PHPG_GOBJECT(this_ptr)), (gboolean)build);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpixmap_gtk_pixmap_new, 0)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpixmap_set, 0)
    ZEND_ARG_OBJ_INFO(0, val, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpixmap_set_build_insensitive, 0)
    ZEND_ARG_INFO(0, build)
ZEND_END_ARG_INFO();

static zend_function_entry gtkpixmap_methods[] = {
	PHP_ME(GtkPixmap, __construct,          arginfo_gtk_gtkpixmap_gtk_pixmap_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPixmap, get,                  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPixmap, set,                  arginfo_gtk_gtkpixmap_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPixmap, set_build_insensitive, arginfo_gtk_gtkpixmap_set_build_insensitive, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkArrow, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "arrow_type", "shadow_type",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^V^V", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkArrow);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkArrow);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkArrow);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkArrow, set)
{
	GtkArrowType arrow_type;
	zval *php_arrow_type = NULL, *php_shadow_type = NULL;
	GtkShadowType shadow_type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_arrow_type, &php_shadow_type))
		return;

	if (php_arrow_type && phpg_gvalue_get_enum(GTK_TYPE_ARROW_TYPE, php_arrow_type, (gint *)&arrow_type) == FAILURE) {
		return;
	}

	if (php_shadow_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_shadow_type, (gint *)&shadow_type) == FAILURE) {
		return;
	}

    gtk_arrow_set(GTK_ARROW(PHPG_GOBJECT(this_ptr)), arrow_type, shadow_type);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkarrow_gtk_arrow_new, 0)
    ZEND_ARG_INFO(0, arrow_type)
    ZEND_ARG_INFO(0, shadow_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkarrow_set, 0)
    ZEND_ARG_INFO(0, arrow_type)
    ZEND_ARG_INFO(0, shadow_type)
ZEND_END_ARG_INFO();

static zend_function_entry gtkarrow_methods[] = {
	PHP_ME(GtkArrow, __construct,          arginfo_gtk_gtkarrow_gtk_arrow_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkArrow, set,                  arginfo_gtk_gtkarrow_set, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkImage, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkImage, new_from_animation)
{
	zval *animation;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &animation, gdkpixbufanimation_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_animation(GDK_PIXBUF_ANIMATION(PHPG_GOBJECT(animation)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_file(filename);
	if (free_filename) g_free(filename);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &icon_name, &free_icon_name, &php_size)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_icon_name(icon_name, size);
	if (free_icon_name) g_free(icon_name);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_icon_set)
{
	GtkIconSet *icon_set = NULL;
	zval *php_icon_set, *php_size = NULL;
	GtkIconSize size;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &php_icon_set, gboxed_ce, &php_size)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

    if (phpg_gboxed_check(php_icon_set, GTK_TYPE_ICON_SET, FALSE TSRMLS_CC)) {
        icon_set = (GtkIconSet *) PHPG_GBOXED(php_icon_set);
    } else {
        php_error(E_WARNING, "%s::%s() expects icon_set argument to be a valid GtkIconSet object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
    }

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_icon_set(icon_set, size);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_image)
{
	zval *image, *mask;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &image, gdkimage_ce, &mask, gdkpixmap_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_image(GDK_IMAGE(PHPG_GOBJECT(image)), GDK_PIXMAP(PHPG_GOBJECT(mask)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_pixbuf)
{
	zval *pixbuf;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_pixbuf(GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_pixmap)
{
	zval *pixmap, *mask;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &pixmap, gdkpixmap_ce, &mask, gdkpixmap_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_pixmap(GDK_PIXMAP(PHPG_GOBJECT(pixmap)), GDK_PIXMAP(PHPG_GOBJECT(mask)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkImage, new_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &stock_id, &free_stock_id, &php_size)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}

	wrapped_obj = (GObject *) gtk_image_new_from_stock(stock_id, size);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImage);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}

#line 4076 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get)
{
    phpg_warn_deprecated("use GtkImage::get_image()" TSRMLS_CC);
#if ZEND_EXTENSION_API_NO > 220051025
    PHP_MN(GtkImage_get_image)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#else
    PHP_FN(GtkImage_get_image)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#endif
}

#line 33729 "ext/gtk+/gen_gtk.c"


#line 4111 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_animation)
{
    GdkPixbufAnimation *anim;
    GtkImageType type;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_ANIMATION && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GdkPixbufAnimation or be empty");
        return;
    }
    anim = gtk_image_get_animation(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));

    phpg_gobject_new(&return_value, (GObject *) anim TSRMLS_CC);
}

#line 33754 "ext/gtk+/gen_gtk.c"


#line 3948 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_icon_name)
{
    gchar *icon_name = NULL;
    GtkIconSize size = GTK_ICON_SIZE_INVALID;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_image_get_icon_name(GTK_IMAGE(PHPG_GOBJECT(this_ptr)),
                            (G_CONST_RETURN gchar **)&icon_name, &size);

    php_gtk_build_value(&return_value, "(ui)", icon_name, size);
}

#line 33775 "ext/gtk+/gen_gtk.c"


#line 3967 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_icon_set)
{
    GtkIconSet *icon_set;
    GtkIconSize size;
    GtkImageType type;
    zval *php_icon_set = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_ICON_SET && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GtkIconSet or be empty");
        return;
    }
    gtk_image_get_icon_set(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), &icon_set, &size);

    phpg_gobject_new(&php_icon_set, (GObject *) icon_set TSRMLS_CC);
    php_gtk_build_value(&return_value, "(Ni)", php_icon_set, size);
}

#line 33803 "ext/gtk+/gen_gtk.c"


#line 4022 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_image)
{
    GdkImage *image;
    GdkBitmap *mask;
    GtkImageType type;
    zval *php_image = NULL, *php_mask = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_IMAGE && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GdkImage or be empty");
        return;
    }
    gtk_image_get_image(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), &image, &mask);

    phpg_gobject_new(&php_image, (GObject *) image TSRMLS_CC);
    phpg_gobject_new(&php_mask, (GObject *) mask TSRMLS_CC);
    php_gtk_build_value(&return_value, "(NN)", php_image, php_mask);
}

#line 33832 "ext/gtk+/gen_gtk.c"


#line 4088 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_pixbuf)
{
    GdkPixbuf *pixbuf;
    GtkImageType type;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_PIXBUF && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GdkPixbuf or be empty");
        return;
    }
    pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));

    phpg_gobject_new(&return_value, (GObject *) pixbuf TSRMLS_CC);
}

#line 33857 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkImage, get_pixel_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_image_get_pixel_size(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 4049 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_pixmap)
{
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    GtkImageType type;
    zval *php_pixmap = NULL, *php_mask = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_PIXMAP && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GdkPixmap or be empty");
        return;
    }
    gtk_image_get_pixmap(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), &pixmap, &mask);

    phpg_gobject_new(&php_pixmap, (GObject *) pixmap TSRMLS_CC);
    phpg_gobject_new(&php_mask, (GObject *) mask TSRMLS_CC);
    php_gtk_build_value(&return_value, "(NN)", php_pixmap, php_mask);
}

#line 33900 "ext/gtk+/gen_gtk.c"


#line 3993 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImage, get_stock)
{
    gchar *stock_id;
    GtkIconSize size;
    GtkImageType type;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    type = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
    if (type != GTK_IMAGE_STOCK && type != GTK_IMAGE_EMPTY) {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "GtkImage should contain a GtkStock or be empty");
        return;
    }
    gtk_image_get_stock(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), &stock_id, &size);

    if (stock_id) {
        php_gtk_build_value(&return_value, "(ui)", stock_id, size);
    } else {
        zval *dummy = NULL;
        php_gtk_build_value(&return_value, "(ni)", dummy, size);
    }
}

#line 33931 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkImage, get_storage_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_image_get_storage_type(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkImage, set)
{
	zval *val, *mask;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &val, gdkimage_ce, &mask, gdkpixmap_ce))
		return;

    gtk_image_set(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), GDK_IMAGE(PHPG_GOBJECT(val)), GDK_PIXMAP(PHPG_GOBJECT(mask)));

}


static PHP_METHOD(GtkImage, set_from_animation)
{
	zval *animation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &animation, gdkpixbufanimation_ce))
		return;

    gtk_image_set_from_animation(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), GDK_PIXBUF_ANIMATION(PHPG_GOBJECT(animation)));

}


static PHP_METHOD(GtkImage, set_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_image_set_from_file(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkImage, set_from_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &icon_name, &free_icon_name, &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_image_set_from_icon_name(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), icon_name, size);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkImage, set_from_icon_set)
{
	GtkIconSet *icon_set = NULL;
	zval *php_icon_set, *php_size = NULL;
	GtkIconSize size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &php_icon_set, gboxed_ce, &php_size))
		return;

    if (phpg_gboxed_check(php_icon_set, GTK_TYPE_ICON_SET, FALSE TSRMLS_CC)) {
        icon_set = (GtkIconSet *) PHPG_GBOXED(php_icon_set);
    } else {
        php_error(E_WARNING, "%s::%s() expects icon_set argument to be a valid GtkIconSet object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_image_set_from_icon_set(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), icon_set, size);

}


static PHP_METHOD(GtkImage, set_from_image)
{
	GdkImage *gdk_image = NULL;
	zval *php_gdk_image, *php_mask;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "NN", &php_gdk_image, gdkimage_ce, &php_mask, gdkpixmap_ce))
		return;
    if (Z_TYPE_P(php_gdk_image) != IS_NULL)
        gdk_image = GDK_IMAGE(PHPG_GOBJECT(php_gdk_image));
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

    gtk_image_set_from_image(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), gdk_image, mask);

}


static PHP_METHOD(GtkImage, set_from_pixbuf)
{
	GdkPixbuf *pixbuf = NULL;
	zval *php_pixbuf;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_pixbuf, gdkpixbuf_ce))
		return;
    if (Z_TYPE_P(php_pixbuf) != IS_NULL)
        pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));

    gtk_image_set_from_pixbuf(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), pixbuf);

}


static PHP_METHOD(GtkImage, set_from_pixmap)
{
	zval *pixmap, *php_mask;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

    gtk_image_set_from_pixmap(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);

}


static PHP_METHOD(GtkImage, set_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GtkIconSize size;
	zval *php_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &stock_id, &free_stock_id, &php_size))
		return;

	if (php_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_size, (gint *)&size) == FAILURE) {
		return;
	}

    gtk_image_set_from_stock(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), stock_id, size);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkImage, set_pixel_size)
{
	long pixel_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &pixel_size))
		return;

    gtk_image_set_pixel_size(GTK_IMAGE(PHPG_GOBJECT(this_ptr)), (gint)pixel_size);

}


static PHP_METHOD(GtkImage, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_image_clear(GTK_IMAGE(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_animation, 0)
    ZEND_ARG_OBJ_INFO(0, animation, GdkPixbufAnimation, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_file, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_icon_set, 0)
    ZEND_ARG_OBJ_INFO(0, icon_set, GtkIconSet, 1)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_image, 0)
    ZEND_ARG_OBJ_INFO(0, image, GdkImage, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_pixmap, 0)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_image_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set, 0)
    ZEND_ARG_OBJ_INFO(0, val, GdkImage, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_animation, 0)
    ZEND_ARG_OBJ_INFO(0, animation, GdkPixbufAnimation, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_file, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_icon_set, 0)
    ZEND_ARG_OBJ_INFO(0, icon_set, GtkIconSet, 1)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_image, 0)
    ZEND_ARG_OBJ_INFO(0, gdk_image, GdkImage, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_pixmap, 0)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimage_set_pixel_size, 0)
    ZEND_ARG_INFO(0, pixel_size)
ZEND_END_ARG_INFO();

static zend_function_entry gtkimage_methods[] = {
	PHP_ME(GtkImage, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, new_from_animation,   arginfo_gtk_gtkimage_image_new_from_animation, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_file,        arginfo_gtk_gtkimage_image_new_from_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_icon_name,   arginfo_gtk_gtkimage_image_new_from_icon_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_icon_set,    arginfo_gtk_gtkimage_image_new_from_icon_set, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_image,       arginfo_gtk_gtkimage_image_new_from_image, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_pixbuf,      arginfo_gtk_gtkimage_image_new_from_pixbuf, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_pixmap,      arginfo_gtk_gtkimage_image_new_from_pixmap, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, new_from_stock,       arginfo_gtk_gtkimage_image_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkImage, clear,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get,                  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_animation,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_icon_set,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_image,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_pixbuf,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_pixel_size,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_pixmap,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_stock,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, get_storage_type,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set,                  arginfo_gtk_gtkimage_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_animation,   arginfo_gtk_gtkimage_set_from_animation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_file,        arginfo_gtk_gtkimage_set_from_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_icon_name,   arginfo_gtk_gtkimage_set_from_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_icon_set,    arginfo_gtk_gtkimage_set_from_icon_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_image,       arginfo_gtk_gtkimage_set_from_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_pixbuf,      arginfo_gtk_gtkimage_set_from_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_pixmap,      arginfo_gtk_gtkimage_set_from_pixmap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_from_stock,       arginfo_gtk_gtkimage_set_from_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImage, set_pixel_size,       arginfo_gtk_gtkimage_set_pixel_size, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4191 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkLabel, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 0;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLabel);
    }

    if (text) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "label", text,
                                              "use-underline", use_underline, NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLabel);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 34307 "ext/gtk+/gen_gtk.c"


#line 4219 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkLabel, get)
{
    gchar *text = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_label_get(GTK_LABEL(PHPG_GOBJECT(this_ptr)), &text);
    if (text) {
        RETURN_STRING(text, 1);
    } else {
        RETURN_EMPTY_STRING();
    }
}
#line 34327 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkLabel, get_angle)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_angle(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkLabel, get_attributes)
{
	PangoAttrList *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_attributes(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, PANGO_TYPE_ATTR_LIST, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkLabel, get_ellipsize)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_ellipsize(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, get_justify)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_justify(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_label(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkLabel, get_layout)
{
	PangoLayout* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_layout(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 4251 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkLabel, get_layout_offsets)
{
    gint x, y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_label_get_layout_offsets(GTK_LABEL(PHPG_GOBJECT(this_ptr)), &x, &y);
    php_gtk_build_value(&return_value, "(ii)", x, y);
}
#line 34442 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkLabel, get_line_wrap)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_line_wrap(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkLabel, get_max_width_chars)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_max_width_chars(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, get_mnemonic_keyval)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_mnemonic_keyval(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, get_mnemonic_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_mnemonic_widget(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkLabel, get_selectable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_selectable(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 4237 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkLabel, get_selection_bounds)
{
    gint start, end;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_label_get_selection_bounds(GTK_LABEL(PHPG_GOBJECT(this_ptr)), &start, &end);
    php_gtk_build_value(&return_value, "(ii)", start, end);
}
#line 34528 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkLabel, get_single_line_mode)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_single_line_mode(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkLabel, get_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_text(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkLabel, get_use_markup)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_use_markup(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkLabel, get_use_underline)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_use_underline(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkLabel, get_width_chars)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_width_chars(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, parse_uline)
{
	char *string;
	zend_bool free_string = FALSE;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &string, &free_string))
		return;

    php_retval = gtk_label_parse_uline(GTK_LABEL(PHPG_GOBJECT(this_ptr)), string);
	if (free_string) g_free(string);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, select_region)
{
	long start_offset, end_offset;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start_offset, &end_offset))
		return;

    gtk_label_select_region(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gint)start_offset, (gint)end_offset);

}


static PHP_METHOD(GtkLabel, set)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

	phpg_warn_deprecated("use GtkLabel::set_text()" TSRMLS_CC);

    gtk_label_set(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_angle)
{
	long angle;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &angle))
		return;

    gtk_label_set_angle(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gint)angle);

}


static PHP_METHOD(GtkLabel, set_attributes)
{
	PangoAttrList *attrs = NULL;
	zval *php_attrs;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_attrs, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_attrs, PANGO_TYPE_ATTR_LIST, FALSE TSRMLS_CC)) {
        attrs = (PangoAttrList *) PHPG_GBOXED(php_attrs);
    } else {
        php_error(E_WARNING, "%s::%s() expects attrs argument to be a valid PangoAttrList object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_label_set_attributes(GTK_LABEL(PHPG_GOBJECT(this_ptr)), attrs);

}


static PHP_METHOD(GtkLabel, set_ellipsize)
{
	PangoEllipsizeMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(PANGO_TYPE_ELLIPSIZE_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_label_set_ellipsize(GTK_LABEL(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkLabel, set_justify)
{
	GtkJustification jtype;
	zval *php_jtype = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_jtype))
		return;

	if (php_jtype && phpg_gvalue_get_enum(GTK_TYPE_JUSTIFICATION, php_jtype, (gint *)&jtype) == FAILURE) {
		return;
	}

    gtk_label_set_justify(GTK_LABEL(PHPG_GOBJECT(this_ptr)), jtype);

}


static PHP_METHOD(GtkLabel, set_label)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_label_set_label(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_line_wrap)
{
	zend_bool wrap;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &wrap))
		return;

    gtk_label_set_line_wrap(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)wrap);

}


static PHP_METHOD(GtkLabel, set_markup)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_label_set_markup(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_markup_with_mnemonic)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_label_set_markup_with_mnemonic(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_max_width_chars)
{
	long n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n_chars))
		return;

    gtk_label_set_max_width_chars(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gint)n_chars);

}


static PHP_METHOD(GtkLabel, set_mnemonic_widget)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_label_set_mnemonic_widget(GTK_LABEL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkLabel, set_pattern)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_label_set_pattern(GTK_LABEL(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}


static PHP_METHOD(GtkLabel, set_selectable)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_label_set_selectable(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkLabel, set_single_line_mode)
{
	zend_bool single_line_mode;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &single_line_mode))
		return;

    gtk_label_set_single_line_mode(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)single_line_mode);

}


static PHP_METHOD(GtkLabel, set_text)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_label_set_text(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_text_with_mnemonic)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_label_set_text_with_mnemonic(GTK_LABEL(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkLabel, set_use_markup)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_label_set_use_markup(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkLabel, set_use_underline)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_label_set_use_underline(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkLabel, set_width_chars)
{
	long n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n_chars))
		return;

    gtk_label_set_width_chars(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gint)n_chars);

}


static PHP_METHOD(GtkLabel, set_line_wrap_mode)
{
	PangoWrapMode wrap_mode;
	zval *php_wrap_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_wrap_mode))
		return;

	if (php_wrap_mode && phpg_gvalue_get_enum(PANGO_TYPE_WRAP_MODE, php_wrap_mode, (gint *)&wrap_mode) == FAILURE) {
		return;
	}

    gtk_label_set_line_wrap_mode(GTK_LABEL(PHPG_GOBJECT(this_ptr)), wrap_mode);

}


static PHP_METHOD(GtkLabel, get_line_wrap_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_line_wrap_mode(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkLabel, get_current_uri)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_current_uri(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkLabel, set_track_visited_links)
{
	zend_bool track_visited_links;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &track_visited_links))
		return;

    gtk_label_set_track_visited_links(GTK_LABEL(PHPG_GOBJECT(this_ptr)), (gboolean)track_visited_links);

}


static PHP_METHOD(GtkLabel, get_track_visited_links)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_label_get_track_visited_links(GTK_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtklabel_gtk_label_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_parse_uline, 0)
    ZEND_ARG_INFO(0, string)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_select_region, 0)
    ZEND_ARG_INFO(0, start_offset)
    ZEND_ARG_INFO(0, end_offset)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_angle, 0)
    ZEND_ARG_INFO(0, angle)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_attributes, 0)
    ZEND_ARG_OBJ_INFO(0, attrs, PangoAttrList, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_ellipsize, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_justify, 0)
    ZEND_ARG_INFO(0, jtype)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_label, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_line_wrap, 0)
    ZEND_ARG_INFO(0, wrap)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_markup, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_markup_with_mnemonic, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_max_width_chars, 0)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_mnemonic_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_pattern, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_selectable, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_single_line_mode, 0)
    ZEND_ARG_INFO(0, single_line_mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_text, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_text_with_mnemonic, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_use_markup, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_use_underline, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_width_chars, 0)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_line_wrap_mode, 0)
    ZEND_ARG_INFO(0, wrap_mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklabel_set_track_visited_links, 0)
    ZEND_ARG_INFO(0, track_visited_links)
ZEND_END_ARG_INFO();

static zend_function_entry gtklabel_methods[] = {
	PHP_ME(GtkLabel, __construct,          arginfo_gtk_gtklabel_gtk_label_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get,                  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_angle,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_attributes,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_current_uri,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_ellipsize,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_justify,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_layout,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_layout_offsets,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_line_wrap,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_line_wrap_mode,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_max_width_chars,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_mnemonic_keyval,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_mnemonic_widget,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_selectable,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_selection_bounds, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_single_line_mode, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_track_visited_links, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_use_markup,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_use_underline,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, get_width_chars,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, parse_uline,          arginfo_gtk_gtklabel_parse_uline, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, select_region,        arginfo_gtk_gtklabel_select_region, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set,                  arginfo_gtk_gtklabel_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_angle,            arginfo_gtk_gtklabel_set_angle, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_attributes,       arginfo_gtk_gtklabel_set_attributes, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_ellipsize,        arginfo_gtk_gtklabel_set_ellipsize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_justify,          arginfo_gtk_gtklabel_set_justify, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_label,            arginfo_gtk_gtklabel_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_line_wrap,        arginfo_gtk_gtklabel_set_line_wrap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_line_wrap_mode,   arginfo_gtk_gtklabel_set_line_wrap_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_markup,           arginfo_gtk_gtklabel_set_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_markup_with_mnemonic, arginfo_gtk_gtklabel_set_markup_with_mnemonic, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_max_width_chars,  arginfo_gtk_gtklabel_set_max_width_chars, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_mnemonic_widget,  arginfo_gtk_gtklabel_set_mnemonic_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_pattern,          arginfo_gtk_gtklabel_set_pattern, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_selectable,       arginfo_gtk_gtklabel_set_selectable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_single_line_mode, arginfo_gtk_gtklabel_set_single_line_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_text,             arginfo_gtk_gtklabel_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_text_with_mnemonic, arginfo_gtk_gtklabel_set_text_with_mnemonic, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_track_visited_links, arginfo_gtk_gtklabel_set_track_visited_links, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_use_markup,       arginfo_gtk_gtklabel_set_use_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_use_underline,    arginfo_gtk_gtklabel_set_use_underline, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLabel, set_width_chars,      arginfo_gtk_gtklabel_set_width_chars, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 2769 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEntry, __construct)
{
    gchararray text = "";
    int max = 0;
    zend_bool free_text = FALSE;
	GObject *wrapped_obj;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ui", &text, &free_text, &max)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntry);
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC),
                                          "max-length", max, "text", text, NULL);

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntry);
    }
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 35231 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEntry, new_with_buffer)
{
	zval *buffer;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &buffer, gtkentrybuffer_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntry);
	}

	wrapped_obj = (GObject *) gtk_entry_new_with_buffer(GTK_ENTRY_BUFFER(PHPG_GOBJECT(buffer)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEntry);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkEntry, append_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

	phpg_warn_deprecated("use GtkEditable::insert_text()" TSRMLS_CC);

    gtk_entry_append_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkEntry, get_activates_default)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_activates_default(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, get_alignment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_alignment(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkEntry, get_completion)
{
	GtkEntryCompletion* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_completion(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, get_has_frame)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_has_frame(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 2808 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEntry, get_invisible_char)
{
    gunichar ichar;
    gint len;
    gchar ret[8];
    gchar *cp_ret;
    gsize cp_len;
    zend_bool free_result = FALSE;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    ichar = gtk_entry_get_invisible_char(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));

    if (ichar) {
        len = g_unichar_to_utf8(ichar, ret);
        ret[len] = '\0';
        cp_ret = phpg_from_utf8(ret, len, &cp_len, &free_result TSRMLS_CC);

        if (cp_ret) {
             RETVAL_STRINGL((char *)ret, len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }

        if (free_result)
            g_free(cp_ret);
        else
            RETVAL_STRINGL((char *)ichar, len, 1);
    } else {
        RETVAL_NULL();
    }
}

#line 35364 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEntry, get_layout)
{
	PangoLayout* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_layout(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2793 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEntry, get_layout_offsets)
{
    gint x, y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_entry_get_layout_offsets(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), &x, &y);
    php_gtk_build_value(&return_value, "(ii)", x, y);
}

#line 35395 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEntry, get_max_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_max_length(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, get_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkEntry, get_visibility)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_visibility(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, get_width_chars)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_width_chars(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, prepend_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

	phpg_warn_deprecated("use GtkEditable::insert_text()" TSRMLS_CC);

    gtk_entry_prepend_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkEntry, select_region)
{
	long start, end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start, &end))
		return;

    gtk_entry_select_region(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)start, (gint)end);

}


static PHP_METHOD(GtkEntry, set_activates_default)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_entry_set_activates_default(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkEntry, set_alignment)
{
	double xalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &xalign))
		return;

    gtk_entry_set_alignment(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (float)xalign);

}


static PHP_METHOD(GtkEntry, set_completion)
{
	zval *completion;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &completion, gtkentrycompletion_ce))
		return;

    gtk_entry_set_completion(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), GTK_ENTRY_COMPLETION(PHPG_GOBJECT(completion)));

}


static PHP_METHOD(GtkEntry, set_editable)
{
	zend_bool editable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &editable))
		return;

    gtk_entry_set_editable(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gboolean)editable);

}


static PHP_METHOD(GtkEntry, set_has_frame)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_entry_set_has_frame(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}

#line 2853 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkEntry, set_invisible_char)
{
    char *getchar;
    gunichar setchar;
    zend_bool free_char = FALSE;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &getchar, &free_char))
        return;

    setchar = g_utf8_get_char((const gchar*)getchar);
    gtk_entry_set_invisible_char(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), setchar);

    if (free_char)
        g_free(getchar);
}

#line 35589 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkEntry, set_max_length)
{
	long max;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &max))
		return;

    gtk_entry_set_max_length(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)max);

}


static PHP_METHOD(GtkEntry, set_position)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_entry_set_position(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkEntry, set_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_entry_set_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkEntry, set_visibility)
{
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible))
		return;

    gtk_entry_set_visibility(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gboolean)visible);

}


static PHP_METHOD(GtkEntry, set_width_chars)
{
	long n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n_chars))
		return;

    gtk_entry_set_width_chars(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)n_chars);

}


static PHP_METHOD(GtkEntry, set_inner_border)
{
	GtkBorder *border = NULL;
	zval *php_border;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_border, gboxed_ce))
		return;

    if (Z_TYPE_P(php_border) != IS_NULL) {
        if (phpg_gboxed_check(php_border, GTK_TYPE_BORDER, FALSE TSRMLS_CC)) {
            border = (GtkBorder *) PHPG_GBOXED(php_border);
        } else {
            php_error(E_WARNING, "%s::%s() expects border argument to be a valid GtkBorder object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_entry_set_inner_border(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), border);

}


static PHP_METHOD(GtkEntry, get_inner_border)
{
	GtkBorder *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_inner_border(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_BORDER, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkEntry, set_cursor_hadjustment)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_adjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_entry_set_cursor_hadjustment(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), adjustment);

}


static PHP_METHOD(GtkEntry, get_cursor_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_cursor_hadjustment(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, set_overwrite_mode)
{
	zend_bool overwrite;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &overwrite))
		return;

    gtk_entry_set_overwrite_mode(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gboolean)overwrite);

}


static PHP_METHOD(GtkEntry, get_overwrite_mode)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_overwrite_mode(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, get_text_length)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_text_length(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, set_progress_fraction)
{
	double fraction;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &fraction))
		return;

    gtk_entry_set_progress_fraction(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), fraction);

}


static PHP_METHOD(GtkEntry, set_icon_drag_source)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL, *php_target_list, *php_actions = NULL;
	GtkTargetList *target_list = NULL;
	GdkDragAction actions;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VOV", &php_icon_pos, &php_target_list, gboxed_ce, &php_actions))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    if (phpg_gboxed_check(php_target_list, GTK_TYPE_TARGET_LIST, FALSE TSRMLS_CC)) {
        target_list = (GtkTargetList *) PHPG_GBOXED(php_target_list);
    } else {
        php_error(E_WARNING, "%s::%s() expects target_list argument to be a valid GtkTargetList object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

	if (php_actions && phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_drag_source(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, target_list, actions);

}


static PHP_METHOD(GtkEntry, get_current_icon_drag_source)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_current_icon_drag_source(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, get_progress_fraction)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_progress_fraction(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkEntry, set_progress_pulse_step)
{
	double fraction;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &fraction))
		return;

    gtk_entry_set_progress_pulse_step(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), fraction);

}


static PHP_METHOD(GtkEntry, get_progress_pulse_step)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_progress_pulse_step(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkEntry, progress_pulse)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_entry_progress_pulse(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEntry, set_icon_from_pixbuf)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL, *php_pixbuf;
	GdkPixbuf *pixbuf = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VN", &php_icon_pos, &php_pixbuf, gdkpixbuf_ce))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}
    if (Z_TYPE_P(php_pixbuf) != IS_NULL)
        pixbuf = GDK_PIXBUF(PHPG_GOBJECT(php_pixbuf));

    gtk_entry_set_icon_from_pixbuf(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, pixbuf);

}


static PHP_METHOD(GtkEntry, set_icon_from_stock)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	char *stock_id;
	zend_bool free_stock_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu", &php_icon_pos, &stock_id, &free_stock_id))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_from_stock(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, stock_id);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkEntry, set_icon_from_icon_name)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu", &php_icon_pos, &icon_name, &free_icon_name))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_from_icon_name(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkEntry, get_icon_storage_type)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_storage_type(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, get_icon_pixbuf)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_pixbuf(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, get_icon_stock)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_stock(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkEntry, get_icon_name)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_name(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkEntry, set_icon_activatable)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	zend_bool activatable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vb", &php_icon_pos, &activatable))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_activatable(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, (gboolean)activatable);

}


static PHP_METHOD(GtkEntry, get_icon_activatable)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_activatable(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, set_icon_sensitive)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	zend_bool sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vb", &php_icon_pos, &sensitive))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_sensitive(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, (gboolean)sensitive);

}


static PHP_METHOD(GtkEntry, get_icon_sensitive)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_sensitive(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, get_icon_at_pos)
{
	long x, y, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_entry_get_icon_at_pos(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkEntry, set_icon_tooltip_text)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	char *tooltip;
	zend_bool free_tooltip = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu", &php_icon_pos, &tooltip, &free_tooltip))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_tooltip_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, tooltip);
	if (free_tooltip) g_free(tooltip);

}


static PHP_METHOD(GtkEntry, get_icon_tooltip_text)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_tooltip_text(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkEntry, set_icon_tooltip_markup)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	char *tooltip;
	zend_bool free_tooltip = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu", &php_icon_pos, &tooltip, &free_tooltip))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    gtk_entry_set_icon_tooltip_markup(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos, tooltip);
	if (free_tooltip) g_free(tooltip);

}


static PHP_METHOD(GtkEntry, get_icon_tooltip_markup)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_tooltip_markup(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkEntry, unset_invisible_char)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_entry_unset_invisible_char(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkEntry, set_buffer)
{
	zval *buffer;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &buffer, gtkentrybuffer_ce))
		return;

    gtk_entry_set_buffer(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), GTK_ENTRY_BUFFER(PHPG_GOBJECT(buffer)));

}


static PHP_METHOD(GtkEntry, get_buffer)
{
	GtkEntryBuffer* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_buffer(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, get_icon_window)
{
	GtkEntryIconPosition icon_pos;
	zval *php_icon_pos = NULL;
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_pos))
		return;

	if (php_icon_pos && phpg_gvalue_get_enum(GTK_TYPE_ENTRY_ICON_POSITION, php_icon_pos, (gint *)&icon_pos) == FAILURE) {
		return;
	}

    php_retval = gtk_entry_get_icon_window(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), icon_pos);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, get_text_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_entry_get_text_window(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkEntry, im_context_filter_keypress)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_entry_im_context_filter_keypress(GTK_ENTRY(PHPG_GOBJECT(this_ptr)), (GdkEventKey *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEntry, reset_im_context)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_entry_reset_im_context(GTK_ENTRY(PHPG_GOBJECT(this_ptr)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkentry_gtk_entry_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, max_length)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_new_with_buffer, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkEntryBuffer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_append_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_prepend_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_select_region, 0)
    ZEND_ARG_INFO(0, start)
    ZEND_ARG_INFO(0, end)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_activates_default, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_alignment, 0)
    ZEND_ARG_INFO(0, xalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_completion, 0)
    ZEND_ARG_OBJ_INFO(0, completion, GtkEntryCompletion, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_editable, 0)
    ZEND_ARG_INFO(0, editable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_has_frame, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_invisible_char, 0)
    ZEND_ARG_INFO(0, char)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_max_length, 0)
    ZEND_ARG_INFO(0, max)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_visibility, 0)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_width_chars, 0)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_inner_border, 0)
    ZEND_ARG_OBJ_INFO(0, border, GtkBorder, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_cursor_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_overwrite_mode, 0)
    ZEND_ARG_INFO(0, overwrite)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_progress_fraction, 0)
    ZEND_ARG_INFO(0, fraction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_drag_source, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_OBJ_INFO(0, target_list, GtkTargetList, 1)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_progress_pulse_step, 0)
    ZEND_ARG_INFO(0, fraction)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_from_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_from_stock, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_from_icon_name, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_storage_type, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_stock, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_name, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_activatable, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, activatable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_activatable, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_sensitive, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_sensitive, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_tooltip_text, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, tooltip)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_tooltip_text, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_icon_tooltip_markup, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
    ZEND_ARG_INFO(0, tooltip)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_tooltip_markup, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_set_buffer, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkEntryBuffer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_get_icon_window, 0)
    ZEND_ARG_OBJ_INFO(0, icon_pos, GtkEntryIconPosition, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkentry_im_context_filter_keypress, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

static zend_function_entry gtkentry_methods[] = {
	PHP_ME(GtkEntry, __construct,          arginfo_gtk_gtkentry_gtk_entry_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, new_with_buffer,      arginfo_gtk_gtkentry_new_with_buffer, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkEntry, append_text,          arginfo_gtk_gtkentry_append_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_activates_default, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_alignment,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_buffer,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_completion,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_current_icon_drag_source, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_cursor_hadjustment, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_has_frame,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_activatable, arginfo_gtk_gtkentry_get_icon_activatable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_at_pos,      arginfo_gtk_gtkentry_get_icon_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_name,        arginfo_gtk_gtkentry_get_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_pixbuf,      arginfo_gtk_gtkentry_get_icon_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_sensitive,   arginfo_gtk_gtkentry_get_icon_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_stock,       arginfo_gtk_gtkentry_get_icon_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_storage_type, arginfo_gtk_gtkentry_get_icon_storage_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_tooltip_markup, arginfo_gtk_gtkentry_get_icon_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_tooltip_text, arginfo_gtk_gtkentry_get_icon_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_icon_window,      arginfo_gtk_gtkentry_get_icon_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_inner_border,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_invisible_char,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_layout,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_layout_offsets,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_max_length,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_overwrite_mode,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_progress_fraction, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_progress_pulse_step, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_text_length,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_text_window,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_visibility,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, get_width_chars,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, im_context_filter_keypress, arginfo_gtk_gtkentry_im_context_filter_keypress, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, prepend_text,         arginfo_gtk_gtkentry_prepend_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, progress_pulse,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, reset_im_context,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, select_region,        arginfo_gtk_gtkentry_select_region, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_activates_default, arginfo_gtk_gtkentry_set_activates_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_alignment,        arginfo_gtk_gtkentry_set_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_buffer,           arginfo_gtk_gtkentry_set_buffer, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_completion,       arginfo_gtk_gtkentry_set_completion, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_cursor_hadjustment, arginfo_gtk_gtkentry_set_cursor_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_editable,         arginfo_gtk_gtkentry_set_editable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_has_frame,        arginfo_gtk_gtkentry_set_has_frame, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_activatable, arginfo_gtk_gtkentry_set_icon_activatable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_drag_source, arginfo_gtk_gtkentry_set_icon_drag_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_from_icon_name, arginfo_gtk_gtkentry_set_icon_from_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_from_pixbuf, arginfo_gtk_gtkentry_set_icon_from_pixbuf, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_from_stock,  arginfo_gtk_gtkentry_set_icon_from_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_sensitive,   arginfo_gtk_gtkentry_set_icon_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_tooltip_markup, arginfo_gtk_gtkentry_set_icon_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_icon_tooltip_text, arginfo_gtk_gtkentry_set_icon_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_inner_border,     arginfo_gtk_gtkentry_set_inner_border, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_invisible_char,   arginfo_gtk_gtkentry_set_invisible_char, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_max_length,       arginfo_gtk_gtkentry_set_max_length, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_overwrite_mode,   arginfo_gtk_gtkentry_set_overwrite_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_position,         arginfo_gtk_gtkentry_set_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_progress_fraction, arginfo_gtk_gtkentry_set_progress_fraction, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_progress_pulse_step, arginfo_gtk_gtkentry_set_progress_pulse_step, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_text,             arginfo_gtk_gtkentry_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_visibility,       arginfo_gtk_gtkentry_set_visibility, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, set_width_chars,      arginfo_gtk_gtkentry_set_width_chars, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEntry, unset_invisible_char, NULL, ZEND_ACC_PUBLIC)

	/***   GtkEditable interface implementations   ***/

	PHP_MALIAS(GtkEditable, copy_clipboard, copy_clipboard, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, cut_clipboard, cut_clipboard, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, delete_selection, delete_selection, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, delete_text, delete_text, arginfo_gtk_gtkeditable_delete_text, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_chars, get_chars, arginfo_gtk_gtkeditable_get_chars, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_editable, get_editable, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_position, get_position, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, get_selection_bounds, get_selection_bounds, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, insert_text, insert_text, arginfo_gtk_gtkeditable_insert_text, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkEditable, paste_clipboard, paste_clipboard, NULL, ZEND_ACC_PUBLIC)

	/***   GtkCellEditable interface implementations   ***/

	PHP_MALIAS(GtkCellEditable, editing_done, editing_done, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellEditable, remove_widget, remove_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellEditable, start_editing, start_editing, arginfo_gtk_gtkcelleditable_start_editing, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTipsQuery, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTipsQuery);
	}


	phpg_warn_deprecated("don't use this" TSRMLS_CC);

	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTipsQuery);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTipsQuery, set_caller)
{
	zval *caller;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &caller, gtkwidget_ce))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_tips_query_set_caller(GTK_TIPS_QUERY(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(caller)));

}


static PHP_METHOD(GtkTipsQuery, set_labels)
{
	char *label_inactive, *label_no_tip;
	zend_bool free_label_inactive = FALSE, free_label_no_tip = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uu", &label_inactive, &free_label_inactive, &label_no_tip, &free_label_no_tip))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_tips_query_set_labels(GTK_TIPS_QUERY(PHPG_GOBJECT(this_ptr)), label_inactive, label_no_tip);
	if (free_label_inactive) g_free(label_inactive);
	if (free_label_no_tip) g_free(label_no_tip);

}


static PHP_METHOD(GtkTipsQuery, start_query)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_tips_query_start_query(GTK_TIPS_QUERY(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTipsQuery, stop_query)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_tips_query_stop_query(GTK_TIPS_QUERY(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktipsquery_set_caller, 0)
    ZEND_ARG_OBJ_INFO(0, caller, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktipsquery_set_labels, 0)
    ZEND_ARG_INFO(0, label_inactive)
    ZEND_ARG_INFO(0, label_no_tip)
ZEND_END_ARG_INFO();

static zend_function_entry gtktipsquery_methods[] = {
	PHP_ME(GtkTipsQuery, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTipsQuery, set_caller,           arginfo_gtk_gtktipsquery_set_caller, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTipsQuery, set_labels,           arginfo_gtk_gtktipsquery_set_labels, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTipsQuery, start_query,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTipsQuery, stop_query,           NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAccelLabel, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "label",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAccelLabel);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAccelLabel);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAccelLabel);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkAccelLabel, accelerator_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkAccelLabel::get_accel_width()" TSRMLS_CC);

    php_retval = gtk_accel_label_accelerator_width(GTK_ACCEL_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAccelLabel, get_accel_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_label_get_accel_widget(GTK_ACCEL_LABEL(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAccelLabel, get_accel_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_label_get_accel_width(GTK_ACCEL_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkAccelLabel, refetch)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_accel_label_refetch(GTK_ACCEL_LABEL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAccelLabel, set_accel_widget)
{
	zval *accel_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_widget, gtkwidget_ce))
		return;

    gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(accel_widget)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccellabel_gtk_accel_label_new, 0)
    ZEND_ARG_INFO(0, string)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaccellabel_set_accel_widget, 0)
    ZEND_ARG_OBJ_INFO(0, accel_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaccellabel_methods[] = {
	PHP_ME(GtkAccelLabel, __construct,          arginfo_gtk_gtkaccellabel_gtk_accel_label_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelLabel, accelerator_width,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelLabel, get_accel_widget,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelLabel, get_accel_width,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelLabel, refetch,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAccelLabel, set_accel_widget,     arginfo_gtk_gtkaccellabel_set_accel_widget, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSpinButton, __construct)
{
	GParameter params[3];
	zval *php_args[4] = { NULL, };
	char *prop_names[] = { "adjustment", "climb_rate", "digits",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N^d^i", &php_args[0], gtkadjustment_ce, &php_args[1], &php_args[2])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinButton);
	}

	memset(&params, 0, 3 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinButton);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinButton);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkSpinButton, new_with_range)
{
	double min, max, step;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddd", &min, &max, &step)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinButton);
	}

	wrapped_obj = (GObject *) gtk_spin_button_new_with_range(min, max, step);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSpinButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkSpinButton, configure)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;
	double climb_rate;
	long digits;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ndi", &php_adjustment, gtkadjustment_ce, &climb_rate, &digits))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_spin_button_configure(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), adjustment, climb_rate, (guint)digits);

}


static PHP_METHOD(GtkSpinButton, get_adjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkSpinButton, get_digits)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_digits(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 5310 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkSpinButton, get_increments)
{
    gdouble step, page;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_spin_button_get_increments(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), &step, &page);
    php_gtk_build_value(&return_value, "(dd)", step, page);
}

#line 37041 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkSpinButton, get_numeric)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_numeric(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 5325 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkSpinButton, get_range)
{
    gdouble min, max;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_spin_button_get_range(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), &min, &max);
    php_gtk_build_value(&return_value, "(dd)", min, max);
}

#line 37072 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkSpinButton, get_snap_to_ticks)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_snap_to_ticks(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSpinButton, get_update_policy)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_update_policy(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSpinButton, get_value)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_value(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkSpinButton, get_value_as_int)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSpinButton, get_wrap)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_spin_button_get_wrap(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSpinButton, set_adjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_spin_button_set_adjustment(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkSpinButton, set_digits)
{
	long digits;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &digits))
		return;

    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), (guint)digits);

}


static PHP_METHOD(GtkSpinButton, set_increments)
{
	double step, page;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &step, &page))
		return;

    gtk_spin_button_set_increments(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), step, page);

}


static PHP_METHOD(GtkSpinButton, set_numeric)
{
	zend_bool numeric;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &numeric))
		return;

    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)numeric);

}


static PHP_METHOD(GtkSpinButton, set_range)
{
	double min, max;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &min, &max))
		return;

    gtk_spin_button_set_range(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), min, max);

}


static PHP_METHOD(GtkSpinButton, set_snap_to_ticks)
{
	zend_bool snap_to_ticks;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &snap_to_ticks))
		return;

    gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)snap_to_ticks);

}


static PHP_METHOD(GtkSpinButton, set_update_policy)
{
	GtkSpinButtonUpdatePolicy policy;
	zval *php_policy = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_policy))
		return;

	if (php_policy && phpg_gvalue_get_enum(GTK_TYPE_SPIN_BUTTON_UPDATE_POLICY, php_policy, (gint *)&policy) == FAILURE) {
		return;
	}

    gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), policy);

}


static PHP_METHOD(GtkSpinButton, set_value)
{
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), value);

}


static PHP_METHOD(GtkSpinButton, set_wrap)
{
	zend_bool wrap;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &wrap))
		return;

    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)wrap);

}


static PHP_METHOD(GtkSpinButton, spin)
{
	GtkSpinType direction;
	zval *php_direction = NULL;
	double increment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vd", &php_direction, &increment))
		return;

	if (php_direction && phpg_gvalue_get_enum(GTK_TYPE_SPIN_TYPE, php_direction, (gint *)&direction) == FAILURE) {
		return;
	}

    gtk_spin_button_spin(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)), direction, increment);

}


static PHP_METHOD(GtkSpinButton, update)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_spin_button_update(GTK_SPIN_BUTTON(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkspinbutton_gtk_spin_button_new, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
    ZEND_ARG_INFO(0, climb_rate)
    ZEND_ARG_INFO(0, digits)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_gtk_spin_button_new_with_range, 0)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
    ZEND_ARG_INFO(0, step)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_configure, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
    ZEND_ARG_INFO(0, climb_rate)
    ZEND_ARG_INFO(0, digits)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_adjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_digits, 0)
    ZEND_ARG_INFO(0, digits)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_increments, 0)
    ZEND_ARG_INFO(0, step)
    ZEND_ARG_INFO(0, page)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_numeric, 0)
    ZEND_ARG_INFO(0, numeric)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_range, 0)
    ZEND_ARG_INFO(0, min)
    ZEND_ARG_INFO(0, max)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_snap_to_ticks, 0)
    ZEND_ARG_INFO(0, snap_to_ticks)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_update_policy, 0)
    ZEND_ARG_INFO(0, policy)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_set_wrap, 0)
    ZEND_ARG_INFO(0, wrap)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkspinbutton_spin, 0)
    ZEND_ARG_INFO(0, direction)
    ZEND_ARG_INFO(0, increment)
ZEND_END_ARG_INFO();

static zend_function_entry gtkspinbutton_methods[] = {
	PHP_ME(GtkSpinButton, __construct,          arginfo_gtk_gtkspinbutton_gtk_spin_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, new_with_range,       arginfo_gtk_gtkspinbutton_gtk_spin_button_new_with_range, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkSpinButton, configure,            arginfo_gtk_gtkspinbutton_configure, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_adjustment,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_digits,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_increments,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_numeric,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_range,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_snap_to_ticks,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_update_policy,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_value,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_value_as_int,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, get_wrap,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_adjustment,       arginfo_gtk_gtkspinbutton_set_adjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_digits,           arginfo_gtk_gtkspinbutton_set_digits, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_increments,       arginfo_gtk_gtkspinbutton_set_increments, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_numeric,          arginfo_gtk_gtkspinbutton_set_numeric, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_range,            arginfo_gtk_gtkspinbutton_set_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_snap_to_ticks,    arginfo_gtk_gtkspinbutton_set_snap_to_ticks, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_update_policy,    arginfo_gtk_gtkspinbutton_set_update_policy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_value,            arginfo_gtk_gtkspinbutton_set_value, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, set_wrap,             arginfo_gtk_gtkspinbutton_set_wrap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, spin,                 arginfo_gtk_gtkspinbutton_spin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSpinButton, update,               NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkDrawingArea, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDrawingArea);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDrawingArea);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkDrawingArea, size)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

	phpg_warn_deprecated("use GtkWidget::set_size_request()" TSRMLS_CC);

    gtk_drawing_area_size(GTK_DRAWING_AREA(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdrawingarea_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

static zend_function_entry gtkdrawingarea_methods[] = {
	PHP_ME(GtkDrawingArea, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDrawingArea, size,                 arginfo_gtk_gtkdrawingarea_size, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

PHPG_PROP_READER(GtkContainer, focus_child)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->focus_child;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkContainer, border_width)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->border_width;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkContainer, need_resize)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->need_resize;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkContainer, resize_mode)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->resize_mode;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkContainer, reallocate_redraws)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->reallocate_redraws;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkContainer, has_focus_chain)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CONTAINER(((phpg_gobject_t *)object)->obj)->has_focus_chain;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtkcontainer_prop_info[] = {
	{ "focus_child", PHPG_PROP_READ_FN(GtkContainer, focus_child), NULL },
	{ "border_width", PHPG_PROP_READ_FN(GtkContainer, border_width), NULL },
	{ "need_resize", PHPG_PROP_READ_FN(GtkContainer, need_resize), NULL },
	{ "resize_mode", PHPG_PROP_READ_FN(GtkContainer, resize_mode), NULL },
	{ "reallocate_redraws", PHPG_PROP_READ_FN(GtkContainer, reallocate_redraws), NULL },
	{ "has_focus_chain", PHPG_PROP_READ_FN(GtkContainer, has_focus_chain), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkContainer, add)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_container_add(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkContainer, check_resize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_container_check_resize(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkContainer, child_type)
{
	GType php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_child_type(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
    phpg_gtype_new(return_value, php_retval TSRMLS_CC);

}

#line 101 "ext/gtk+/gtkcontainer.overrides"
static PHP_METHOD(GtkContainer, children)
{
	phpg_warn_deprecated("use GtkContainer::get_children()" TSRMLS_CC);
#if ZEND_EXTENSION_API_NO > 220051025
    PHP_MN(GtkContainer_get_children)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#else
    PHP_FN(GtkContainer_get_children)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#endif
}
#line 37610 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkContainer, get_border_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_get_border_width(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 79 "ext/gtk+/gtkcontainer.overrides"
static PHP_METHOD(GtkContainer, get_children)
{
	GList *children, *item;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	children = gtk_container_get_children(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
	for (item = children; item; item = item->next) {
		zval *php_child = NULL;
        phpg_gobject_new(&php_child, G_OBJECT(item->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_child);
	}
    g_list_free(children);
}

#line 37648 "ext/gtk+/gen_gtk.c"


#line 15 "ext/gtk+/gtkcontainer.overrides"
static PHP_METHOD(GtkContainer, get_focus_chain)
{
    GList *widgets, *widget;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

   if (!gtk_container_get_focus_chain(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), &widgets)) {
        RETURN_FALSE;
    }

    array_init(return_value);
    for (widget = widgets; widget; widget = widget->next) {
        zval *php_widget = NULL;
        phpg_gobject_new(&php_widget, G_OBJECT(widget->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_widget);
    }
    g_list_free(widgets);
}

#line 37674 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkContainer, get_focus_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_get_focus_hadjustment(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkContainer, get_focus_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_get_focus_vadjustment(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkContainer, get_resize_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_get_resize_mode(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkContainer, propagate_expose)
{
	zval *child, *php_event;
	GdkEvent *event = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &child, gtkwidget_ce, &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_container_propagate_expose(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (GdkEventExpose *)event);

}


static PHP_METHOD(GtkContainer, remove)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    gtk_container_remove(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkContainer, resize_children)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_container_resize_children(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkContainer, set_border_width)
{
	long border_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &border_width))
		return;

    gtk_container_set_border_width(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), (guint)border_width);

}

#line 46 "ext/gtk+/gtkcontainer.overrides"
static PHP_METHOD(GtkContainer, set_focus_chain)
{
    zval *php_widgets, **php_widget;
    GList *widgets= NULL;
    int n;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_widgets)) {
        return;
    }

    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_widgets)), n = 0;
        zend_hash_get_current_data(Z_ARRVAL_P(php_widgets), (void **)&php_widget) == SUCCESS;
        zend_hash_move_forward(Z_ARRVAL_P(php_widgets)), n++) {

        if (php_gtk_check_class(*php_widget, gtkwidget_ce)) {
            GtkWidget *widget = GTK_WIDGET(PHPG_GOBJECT(*php_widget));
            //prepend is faster than append
            widgets = g_list_prepend(widgets, widget);
        } else {
            php_error(E_WARNING, "%s::%s() requires the focus chain elements to be objects of class GtkWidget", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
    }

    widgets = g_list_reverse(widgets);
    gtk_container_set_focus_chain(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), widgets);

    g_list_free(widgets);
}

#line 37815 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkContainer, set_focus_child)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_container_set_focus_child(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkContainer, set_focus_hadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_container_set_focus_hadjustment(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkContainer, set_focus_vadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_container_set_focus_vadjustment(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkContainer, set_reallocate_redraws)
{
	zend_bool needs_redraws;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &needs_redraws))
		return;

    gtk_container_set_reallocate_redraws(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), (gboolean)needs_redraws);

}


static PHP_METHOD(GtkContainer, set_resize_mode)
{
	GtkResizeMode resize_mode;
	zval *php_resize_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_resize_mode))
		return;

	if (php_resize_mode && phpg_gvalue_get_enum(GTK_TYPE_RESIZE_MODE, php_resize_mode, (gint *)&resize_mode) == FAILURE) {
		return;
	}

    gtk_container_set_resize_mode(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)), resize_mode);

}


static PHP_METHOD(GtkContainer, unset_focus_chain)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_container_unset_focus_chain(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkContainer, get_focus_child)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_container_get_focus_child(GTK_CONTAINER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_add, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_propagate_expose, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_remove, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_border_width, 0)
    ZEND_ARG_INFO(0, border_width)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_focus_chain, 0)
    ZEND_ARG_INFO(0, widgets)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_focus_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_focus_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_focus_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_reallocate_redraws, 0)
    ZEND_ARG_INFO(0, needs_redraws)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcontainer_set_resize_mode, 0)
    ZEND_ARG_INFO(0, resize_mode)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcontainer_methods[] = {
	PHP_ME(GtkContainer, add,                  arginfo_gtk_gtkcontainer_add, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, check_resize,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, child_type,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, children,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_border_width,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_children,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_focus_chain,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_focus_child,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_focus_hadjustment, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_focus_vadjustment, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, get_resize_mode,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, propagate_expose,     arginfo_gtk_gtkcontainer_propagate_expose, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, remove,               arginfo_gtk_gtkcontainer_remove, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, resize_children,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_border_width,     arginfo_gtk_gtkcontainer_set_border_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_focus_chain,      arginfo_gtk_gtkcontainer_set_focus_chain, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_focus_child,      arginfo_gtk_gtkcontainer_set_focus_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_focus_hadjustment, arginfo_gtk_gtkcontainer_set_focus_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_focus_vadjustment, arginfo_gtk_gtkcontainer_set_focus_vadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_reallocate_redraws, arginfo_gtk_gtkcontainer_set_reallocate_redraws, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, set_resize_mode,      arginfo_gtk_gtkcontainer_set_resize_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkContainer, unset_focus_chain,    NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCurve, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCurve);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCurve);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 2509 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCurve, get_vector)
{
    int num, i;
    gfloat *vector = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &num))
        return;

    vector = safe_emalloc(num, sizeof(gfloat *), 0);

    gtk_curve_get_vector(GTK_CURVE(PHPG_GOBJECT(this_ptr)), num, vector);

    array_init(return_value);
    for (i = 0; i < num; i++) {
        add_next_index_double(return_value, vector[i]);
    }
}

#line 38042 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCurve, reset)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_curve_reset(GTK_CURVE(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCurve, set_curve_type)
{
	GtkCurveType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_CURVE_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_curve_set_curve_type(GTK_CURVE(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkCurve, set_gamma)
{
	double gamma;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &gamma))
		return;

    gtk_curve_set_gamma(GTK_CURVE(PHPG_GOBJECT(this_ptr)), (float)gamma);

}


static PHP_METHOD(GtkCurve, set_range)
{
	double min_x, max_x, min_y, max_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dddd", &min_x, &max_x, &min_y, &max_y))
		return;

    gtk_curve_set_range(GTK_CURVE(PHPG_GOBJECT(this_ptr)), (float)min_x, (float)max_x, (float)min_y, (float)max_y);

}

#line 2531 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCurve, set_vector)
{
    int num, i = 0;
    gfloat *vectors = NULL;
    zval *points = NULL;
    zval **data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &points))
        return;

    num = zend_hash_num_elements(Z_ARRVAL_P(points));
    vectors = safe_emalloc(num, sizeof(gfloat *), 0);

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(points));
    while (zend_hash_get_current_data(Z_ARRVAL_P(points), (void **)&data)
            == SUCCESS) {
        if (Z_TYPE_PP(data) != IS_DOUBLE) {
            php_error(E_WARNING, "%s::%s(): each point must be a float value.",
                        get_active_class_name(NULL TSRMLS_CC),
                        get_active_function_name(TSRMLS_C));
            return;
        }
        /* note: the compiler warning thrown here by MSVC6 is due to a compiler bug */
        vectors[i] = Z_DVAL_PP(data);
        i++;
        zend_hash_move_forward(Z_ARRVAL_P(points));
    }

    gtk_curve_set_vector(GTK_CURVE(PHPG_GOBJECT(this_ptr)), num, vectors);
}

#line 38139 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcurve_set_curve_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcurve_set_gamma, 0)
    ZEND_ARG_INFO(0, gamma)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcurve_set_range, 0)
    ZEND_ARG_INFO(0, min_x)
    ZEND_ARG_INFO(0, max_x)
    ZEND_ARG_INFO(0, min_y)
    ZEND_ARG_INFO(0, max_y)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcurve_methods[] = {
	PHP_ME(GtkCurve, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, get_vector,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, reset,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, set_curve_type,       arginfo_gtk_gtkcurve_set_curve_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, set_gamma,            arginfo_gtk_gtkcurve_set_gamma, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, set_range,            arginfo_gtk_gtkcurve_set_range, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCurve, set_vector,           NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTreeView, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "model",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtktreemodel_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeView);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeView);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTreeView);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTreeView, append_column)
{
	zval *column;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &column, gtktreeviewcolumn_ce))
		return;

    php_retval = gtk_tree_view_append_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, collapse_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_collapse_all(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeView, collapse_row)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_collapse_row(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, columns_autosize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_columns_autosize(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeView, create_row_drag_icon)
{
	GtkTreePath *path;
	zval *php_path;
	GdkPixmap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_view_create_row_drag_icon(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2229 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, enable_model_drag_dest)
{
	zval *php_targets, *php_actions;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "aV", &php_targets, &php_actions)) {
		return;
	}

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
	}

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), entries, n, actions);
	efree(entries);
}
#line 38317 "ext/gtk+/gen_gtk.c"


#line 2192 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, enable_model_drag_source)
{
	zval *php_sbmask, *php_targets, *php_actions;
	GdkModifierType sbmask;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VaV", &php_sbmask, &php_targets, &php_actions)) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_sbmask, (gint *)&sbmask) == FAILURE) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
        return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), sbmask, entries, n, actions);
	efree(entries);
}

#line 38348 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, expand_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_expand_all(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeView, expand_row)
{
	GtkTreePath *path;
	zval *php_path;
	zend_bool open_all;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vb", &php_path, &open_all))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_view_expand_row(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, (gboolean)open_all);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, expand_to_path)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_expand_to_path(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}

#line 2536 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_background_area)
{
    GtkTreePath *path;
    zval *php_path;
    zval *php_column;
    GdkRectangle rect;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_column, gtktreeviewcolumn_ce))
        return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_get_background_area(
        GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
        path,
        GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_column)),
        &rect
    );
    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
    gtk_tree_path_free(path);
}

#line 38438 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, get_bin_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_bin_window(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2570 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_cell_area)
{
    GtkTreePath *path;
    zval *php_path;
    zval *php_column;
    GdkRectangle rect;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &php_column, gtktreeviewcolumn_ce))
        return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_get_cell_area(
        GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
        path,
        GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_column)),
        &rect
    );
    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
    gtk_tree_path_free(path);
}

#line 38481 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, get_column)
{
	long n;
	GtkTreeViewColumn* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n))
		return;

    php_retval = gtk_tree_view_get_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gint)n);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2252 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_columns)
{
    GList *columns, *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    columns = gtk_tree_view_get_columns(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (item = columns; item; item = item->next) {
        zval *php_column = NULL;
        phpg_gobject_new(&php_column, G_OBJECT(item->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_column);
    }
    g_list_free(columns);
}

#line 38520 "ext/gtk+/gen_gtk.c"


#line 2597 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_cursor)
{
    GtkTreePath *path = NULL;
    GtkTreeViewColumn *column = NULL;
    zval *php_path = NULL, *php_column = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_tree_view_get_cursor(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), &path, &column);

    array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        ALLOC_INIT_ZVAL(php_path);
    }

    if (column) {
        phpg_gobject_new(&php_column, (GObject*)column TSRMLS_CC);
    } else {
        ALLOC_INIT_ZVAL(php_column);
    }

    php_gtk_build_value(&return_value, "(NN)", php_path, php_column);
}

#line 38555 "ext/gtk+/gen_gtk.c"


#line 2683 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_dest_row_at_pos)
{
    gint x, y;
    GtkTreePath *path;
    GtkTreeViewDropPosition pos;
    zval *php_path = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y)) {
        return;
    }

    if (!gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &path,
                                       &pos)) {
        return;
    }

    array_init(return_value);


    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_path);
    }
    add_next_index_zval(return_value, php_path);

    add_next_index_long(return_value, pos);
}

#line 38590 "ext/gtk+/gen_gtk.c"


#line 2716 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_drag_dest_row)
{
    GtkTreePath *path;
    GtkTreeViewDropPosition pos;
    zval *php_path = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_tree_view_get_drag_dest_row(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), &path, &pos);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
        gtk_tree_path_free(path);
        php_gtk_build_value(&return_value, "(Ni)", php_path, (gint)pos);
    } else {
        RETURN_FALSE;
    }
}

#line 38617 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, get_enable_search)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_enable_search(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_expander_column)
{
	GtkTreeViewColumn* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_expander_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeView, get_fixed_height_mode)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_fixed_height_mode(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_hadjustment(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeView, get_headers_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_headers_visible(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_hover_expand)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_hover_expand(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_hover_selection)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_hover_selection(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_model(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2638 "ext/gtk+/gtktreeview.overrides"

static PHP_METHOD(GtkTreeView, get_path_at_pos)
{
    gint x, y;
    gint cell_x, cell_y;
    GtkTreePath *path;
    GtkTreeViewColumn *column;
    zval *php_path = NULL, *php_column = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y)) {
        return;
    }

    if (!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &path,
                                       &column, &cell_x, &cell_y)) {
        return;
    }

    array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_path);
    }
    add_next_index_zval(return_value, php_path);

    phpg_gobject_new(&php_column, (GObject*)column TSRMLS_CC);
    add_next_index_zval(return_value, php_column);
    add_next_index_long(return_value, cell_x);
    add_next_index_long(return_value, cell_y);
}

#line 38768 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, get_reorderable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_reorderable(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_rules_hint)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_rules_hint(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_search_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_search_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, get_selection)
{
	GtkTreeSelection* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_selection(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeView, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_vadjustment(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2741 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, get_visible_rect)
{
    GdkRectangle rect;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_tree_view_get_visible_rect(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), &rect);
    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
}


#line 38856 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, insert_column)
{
	zval *column;
	long position, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &column, gtktreeviewcolumn_ce, &position))
		return;

    php_retval = gtk_tree_view_insert_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)), (gint)position);
	RETVAL_LONG(php_retval);
}

#line 2454 "ext/gtk+/gtktreeview.overrides"

static void phpg_cell_data_func_marshal(GtkCellLayout *cell_layout, GtkCellRenderer *renderer, GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_cell_layout = NULL, *php_renderer = NULL, *php_model = NULL, *php_iter = NULL;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_cell_layout, (GObject*)cell_layout TSRMLS_CC);
    phpg_gobject_new(&php_renderer,    (GObject*)renderer    TSRMLS_CC);
    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 4, &n_args);
    args[0] = &php_cell_layout;
    args[1] = &php_renderer;
    args[2] = &php_model;
    args[3] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_cell_layout);
    zval_ptr_dtor(&php_renderer);
    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkTreeView, insert_column_with_data_func)
{
    gint position;
    char *title;
    zval *php_cell;
    GtkCellRenderer *cell;
    zval *callback, *extra;
    phpg_cb_data_t *cb_data;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 4, &extra, "isOV", &position, &title, &php_cell, gtkcellrenderer_ce, &callback))
        return;

    cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));

    zval_add_ref(&callback);
    cb_data = phpg_cb_data_new(callback, extra TSRMLS_CC);

    RETURN_LONG(gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
                                                           position, title, cell,
                                                           (GtkTreeCellDataFunc)phpg_cell_data_func_marshal,
                                                           cb_data, phpg_cb_data_destroy));
}


#line 38948 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, move_column_after)
{
	zval *column, *php_base_column;
	GtkTreeViewColumn *base_column = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ON", &column, gtktreeviewcolumn_ce, &php_base_column, gtktreeviewcolumn_ce))
		return;
    if (Z_TYPE_P(php_base_column) != IS_NULL)
        base_column = GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_base_column));

    gtk_tree_view_move_column_after(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)), base_column);

}


static PHP_METHOD(GtkTreeView, remove_column)
{
	zval *column;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &column, gtktreeviewcolumn_ce))
		return;

    php_retval = gtk_tree_view_remove_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, row_activated)
{
	GtkTreePath *path;
	zval *php_path, *column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VO", &php_path, &column, gtktreeviewcolumn_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_row_activated(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)));

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, row_expanded)
{
	GtkTreePath *path;
	zval *php_path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_tree_view_row_expanded(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, scroll_to_cell)
{
	GtkTreePath *path;
	zval *php_path, *php_column = NULL;
	GtkTreeViewColumn *column = NULL;
	zend_bool use_align = FALSE;
	double row_align = 0.0, col_align = 0.0;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|Nbdd", &php_path, &php_column, gtktreeviewcolumn_ce, &use_align, &row_align, &col_align))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (php_column) {
        if (Z_TYPE_P(php_column) == IS_NULL)
            column = NULL;
        else
            column = GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_column));
    }

    gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, column, (gboolean)use_align, (float)row_align, (float)col_align);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, scroll_to_point)
{
	long tree_x, tree_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &tree_x, &tree_y))
		return;

    gtk_tree_view_scroll_to_point(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gint)tree_x, (gint)tree_y);

}

#line 2281 "ext/gtk+/gtktreeview.overrides"
static gboolean phpg_tree_view_column_drop_func_marshal (GtkTreeView *tree_view, GtkTreeViewColumn *column, GtkTreeViewColumn *prev_column, GtkTreeViewColumn *next_column, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_view = NULL, *php_column = NULL, *php_prev_column = NULL, *php_next_column = NULL;
    gboolean may_drop = FALSE;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_view,        (GObject*)tree_view       TSRMLS_CC);
    phpg_gobject_new(&php_column,      (GObject*)column          TSRMLS_CC);
    phpg_gobject_new(&php_prev_column, (GObject*)prev_column     TSRMLS_CC);
    phpg_gobject_new(&php_next_column, (GObject*)next_column     TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 4, &n_args);
    args[0] = &php_view;
    args[1] = &php_column;
    args[2] = &php_prev_column;
    args[3] = &php_next_column;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_view);
    zval_ptr_dtor(&php_column);
    zval_ptr_dtor(&php_prev_column);
    zval_ptr_dtor(&php_next_column);

    if (retval) {
        may_drop = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        may_drop = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return may_drop;
}

static PHP_METHOD(GtkTreeView, set_column_drag_function)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeViewColumnDropFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTreeViewColumnDropFunc)phpg_tree_view_column_drop_func_marshal;
        }
    }

    gtk_tree_view_set_column_drag_function(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 39156 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, set_cursor)
{
	GtkTreePath *path;
	zval *php_path, *php_focus_column = NULL;
	GtkTreeViewColumn *focus_column = NULL;
	zend_bool start_editing = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|Nb", &php_path, &php_focus_column, gtktreeviewcolumn_ce, &start_editing))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (php_focus_column) {
        if (Z_TYPE_P(php_focus_column) == IS_NULL)
            focus_column = NULL;
        else
            focus_column = GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_focus_column));
    }

    gtk_tree_view_set_cursor(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, focus_column, (gboolean)start_editing);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, set_cursor_on_cell)
{
	GtkTreePath *path;
	zval *php_path, *php_focus_column = NULL, *php_focus_cell = NULL;
	GtkTreeViewColumn *focus_column = NULL;
	GtkCellRenderer *focus_cell = NULL;
	zend_bool start_editing = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|NNb", &php_path, &php_focus_column, gtktreeviewcolumn_ce, &php_focus_cell, gtkcellrenderer_ce, &start_editing))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (php_focus_column) {
        if (Z_TYPE_P(php_focus_column) == IS_NULL)
            focus_column = NULL;
        else
            focus_column = GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_focus_column));
    }
    if (php_focus_cell) {
        if (Z_TYPE_P(php_focus_cell) == IS_NULL)
            focus_cell = NULL;
        else
            focus_cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_focus_cell));
    }

    gtk_tree_view_set_cursor_on_cell(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, focus_column, focus_cell, (gboolean)start_editing);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, set_drag_dest_row)
{
	GtkTreePath *path;
	zval *php_path, *php_pos = NULL;
	GtkTreeViewDropPosition pos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_path, &php_pos))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

	if (php_pos && phpg_gvalue_get_enum(GTK_TYPE_TREE_VIEW_DROP_POSITION, php_pos, (gint *)&pos) == FAILURE) {
		return;
	}

    gtk_tree_view_set_drag_dest_row(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), path, pos);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, set_enable_search)
{
	zend_bool enable_search;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enable_search))
		return;

    gtk_tree_view_set_enable_search(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)enable_search);

}


static PHP_METHOD(GtkTreeView, set_expander_column)
{
	zval *column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &column, gtktreeviewcolumn_ce))
		return;

    gtk_tree_view_set_expander_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(column)));

}


static PHP_METHOD(GtkTreeView, set_fixed_height_mode)
{
	zend_bool enable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enable))
		return;

    gtk_tree_view_set_fixed_height_mode(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)enable);

}


static PHP_METHOD(GtkTreeView, set_hadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_tree_view_set_hadjustment(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkTreeView, set_headers_clickable)
{
	zend_bool active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &active))
		return;

    gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)active);

}


static PHP_METHOD(GtkTreeView, set_headers_visible)
{
	zend_bool headers_visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &headers_visible))
		return;

    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)headers_visible);

}


static PHP_METHOD(GtkTreeView, set_hover_expand)
{
	zend_bool expand;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &expand))
		return;

    gtk_tree_view_set_hover_expand(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)expand);

}


static PHP_METHOD(GtkTreeView, set_hover_selection)
{
	zend_bool hover;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &hover))
		return;

    gtk_tree_view_set_hover_selection(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)hover);

}


static PHP_METHOD(GtkTreeView, set_model)
{
	GtkTreeModel *model = NULL;
	zval *php_model = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_model, gtktreemodel_ce))
		return;
    if (php_model) {
        if (Z_TYPE_P(php_model) == IS_NULL)
            model = NULL;
        else
            model = GTK_TREE_MODEL(PHPG_GOBJECT(php_model));
    }

    gtk_tree_view_set_model(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), model);

}


static PHP_METHOD(GtkTreeView, set_reorderable)
{
	zend_bool reorderable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reorderable))
		return;

    gtk_tree_view_set_reorderable(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)reorderable);

}

#line 2369 "ext/gtk+/gtktreeview.overrides"

static gboolean phpg_tree_view_row_separator_func_marshal(GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_iter = NULL;
    gboolean is_separator = FALSE;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_model,       (GObject*)model       TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_model;
    args[1] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        is_separator = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        is_separator = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return is_separator;
}

static PHP_METHOD(GtkTreeView, set_row_separator_func)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeViewRowSeparatorFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTreeViewRowSeparatorFunc)phpg_tree_view_row_separator_func_marshal;
        }
    }

    gtk_tree_view_set_row_separator_func(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 39478 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, set_rules_hint)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkTreeView, set_search_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_tree_view_set_search_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}

#line 2764 "ext/gtk+/gtktreeview.overrides"
static gboolean phpg_tree_view_search_equal_func_marshal (GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *iter, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_model = NULL, *php_column = NULL, *php_iter = NULL, *php_key;
    gboolean matches = FALSE;
    gchar *cp_key;
    gsize cp_len;
    zend_bool free_result;
#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_model,       (GObject*)model              TSRMLS_CC);
    phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, iter, TRUE, TRUE TSRMLS_CC);

    MAKE_STD_ZVAL(php_column);
    ZVAL_LONG(php_column, column);

    cp_key = phpg_from_utf8(key, strlen(key), &cp_len, &free_result TSRMLS_CC);
    if (!cp_key) {
        php_error(E_WARNING, "Could not convert key from UTF-8");
        return FALSE;
    }
    MAKE_STD_ZVAL(php_key);
    ZVAL_STRINGL(php_key, (char*)cp_key, cp_len, 1);
    if (free_result) {
        g_free(cp_key);
    }

    args = php_gtk_hash_as_array_offset(cbd->user_args, 4, &n_args);
    args[0] = &php_model;
    args[1] = &php_column;
    args[2] = &php_key;
    args[3] = &php_iter;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_model);
    zval_ptr_dtor(&php_column);
    zval_ptr_dtor(&php_key);
    zval_ptr_dtor(&php_iter);

    if (retval) {
        matches = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        matches = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return matches;
}

static PHP_METHOD(GtkTreeView, set_search_equal_func)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeViewSearchEqualFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTreeViewSearchEqualFunc)phpg_tree_view_search_equal_func_marshal;
        }
    }

    gtk_tree_view_set_search_equal_func(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 39604 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, set_vadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}

#line 2868 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, tree_to_widget_coords)
{
    gint tx, ty, wx, wy;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &tx, &ty))
        return;

    gtk_tree_view_tree_to_widget_coords(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), tx, ty, &wx, &wy);
    php_gtk_build_value(&return_value, "(ii)", wx, wy);
}


#line 39636 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, unset_rows_drag_dest)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_unset_rows_drag_dest(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkTreeView, unset_rows_drag_source)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tree_view_unset_rows_drag_source(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));

}

#line 2892 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkTreeView, widget_to_tree_coords)
{
    gint wx, wy, tx, ty;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &wx, &wy))
        return;

    gtk_tree_view_widget_to_tree_coords(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), wx, wy, &tx, &ty);
    php_gtk_build_value(&return_value, "(ii)", tx, ty);
}

#line 39679 "ext/gtk+/gen_gtk.c"


#line 262 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkTreeView, get_visible_range)
{
    GtkTreePath *start_path, *end_path;
	zval *php_start = NULL, *php_end = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    if (!gtk_tree_view_get_visible_range(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), &start_path, &end_path))
	{
        return;
    }

    array_init(return_value);
	if (start_path) {
        phpg_tree_path_to_zval(start_path, &php_start TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_start);
    }
	add_next_index_zval(return_value, php_start);

	if (end_path) {
        phpg_tree_path_to_zval(end_path, &php_end TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_end);
    }
	add_next_index_zval(return_value, php_end);
}
#line 39714 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, get_headers_clickable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_headers_clickable(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_search_entry)
{
	GtkEntry* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_search_entry(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTreeView, set_search_entry)
{
	GtkEntry *entry = NULL;
	zval *php_entry = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_entry, gtkentry_ce))
		return;
    if (php_entry) {
        if (Z_TYPE_P(php_entry) == IS_NULL)
            entry = NULL;
        else
            entry = GTK_ENTRY(PHPG_GOBJECT(php_entry));
    }

    gtk_tree_view_set_search_entry(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), entry);

}


static PHP_METHOD(GtkTreeView, set_rubber_banding)
{
	zend_bool enable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enable))
		return;

    gtk_tree_view_set_rubber_banding(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)enable);

}


static PHP_METHOD(GtkTreeView, get_rubber_banding)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_rubber_banding(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, get_grid_lines)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_grid_lines(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, set_grid_lines)
{
	GtkTreeViewGridLines grid_lines;
	zval *php_grid_lines = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_grid_lines))
		return;

	if (php_grid_lines && phpg_gvalue_get_enum(GTK_TYPE_TREE_VIEW_GRID_LINES, php_grid_lines, (gint *)&grid_lines) == FAILURE) {
		return;
	}

    gtk_tree_view_set_grid_lines(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), grid_lines);

}


static PHP_METHOD(GtkTreeView, get_enable_tree_lines)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_enable_tree_lines(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, set_enable_tree_lines)
{
	zend_bool enabled;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enabled))
		return;

    gtk_tree_view_set_enable_tree_lines(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)enabled);

}

#line 539 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_widget_to_tree_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_widget_to_tree_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39872 "ext/gtk+/gen_gtk.c"


#line 487 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_tree_to_widget_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_tree_to_widget_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39892 "ext/gtk+/gen_gtk.c"


#line 513 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_widget_to_bin_window_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_widget_to_bin_window_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39912 "ext/gtk+/gen_gtk.c"


#line 435 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_bin_window_to_widget_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_bin_window_to_widget_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39932 "ext/gtk+/gen_gtk.c"


#line 461 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_tree_to_bin_window_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_tree_to_bin_window_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39952 "ext/gtk+/gen_gtk.c"


#line 409 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, convert_bin_window_to_tree_coords)
{
	gint x, y, wx, wy;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"ii", &x, &y)) {
		return;
	}

	gtk_tree_view_convert_bin_window_to_tree_coords(
		GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &wx, &wy);

	php_gtk_build_value(&return_value, "(ii)", wx, wy);
}

#line 39972 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, set_show_expanders)
{
	zend_bool enabled;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enabled))
		return;

    gtk_tree_view_set_show_expanders(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)enabled);

}


static PHP_METHOD(GtkTreeView, get_show_expanders)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_show_expanders(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTreeView, set_level_indentation)
{
	long indentation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &indentation))
		return;

    gtk_tree_view_set_level_indentation(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gint)indentation);

}


static PHP_METHOD(GtkTreeView, get_level_indentation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_level_indentation(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, is_rubber_banding_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_is_rubber_banding_active(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 373 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkTreeView, get_tooltip_context)
{
	zval *php_model = NULL, *php_iter = NULL, *php_path = NULL;
	gint x, y;
	gboolean keyboard_mode;
	GtkTreeModel *model = NULL;
	GtkTreePath *path = NULL;
	GtkTreeIter iter = {0, };

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"iii", &x, &y, &keyboard_mode)) {
		return;
	}

	if (!gtk_tree_view_get_tooltip_context (GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)),
		&x, &y, keyboard_mode, &model, &path, &iter)) {
		RETURN_FALSE;
	}

	phpg_gobject_new(&php_model, (GObject *) model TSRMLS_CC);
	phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
	phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
	php_gtk_build_value(&return_value, "(NNN)", php_model, php_path, php_iter);
}

#line 40072 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTreeView, set_tooltip_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_tree_view_set_tooltip_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkTreeView, get_tooltip_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tree_view_get_tooltip_column(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTreeView, set_tooltip_row)
{
	zval *tooltip, *php_path;
	GtkTreePath *path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &tooltip, gtktooltip_ce, &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_tree_view_set_tooltip_row(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIP(PHPG_GOBJECT(tooltip)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkTreeView, set_tooltip_cell)
{
	zval *tooltip, *php_path = NULL, *php_column, *php_cell;
	GtkTreePath *path = NULL;
	GtkTreeViewColumn *column = NULL;
	GtkCellRenderer *cell = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVNN", &tooltip, gtktooltip_ce, &php_path, &php_column, gtktreeviewcolumn_ce, &php_cell, gtkcellrenderer_ce))
		return;

    if (php_path && Z_TYPE_P(php_path) != IS_NULL) {
        if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
            php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }
    if (Z_TYPE_P(php_column) != IS_NULL)
        column = GTK_TREE_VIEW_COLUMN(PHPG_GOBJECT(php_column));
    if (Z_TYPE_P(php_cell) != IS_NULL)
        cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));

    gtk_tree_view_set_tooltip_cell(GTK_TREE_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIP(PHPG_GOBJECT(tooltip)), path, column, cell);

    if (path)
        gtk_tree_path_free(path);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_gtk_tree_view_new_with_model, 0, 0, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_append_column, 0)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_collapse_row, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_create_row_drag_icon, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_enable_model_drag_dest, 0)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_enable_model_drag_source, 0)
    ZEND_ARG_INFO(0, sbmask)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_expand_row, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_INFO(0, open_all)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_expand_to_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_get_background_area, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_get_cell_area, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_get_column, 0)
    ZEND_ARG_INFO(0, n)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_get_dest_row_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_get_path_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_insert_column, 0)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_insert_column_with_data_func, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_OBJ_INFO(0, cellrenderer, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_move_column_after, 0)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
    ZEND_ARG_OBJ_INFO(0, base_column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_remove_column, 0)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_row_activated, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_row_expanded, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_scroll_to_cell, 0, 0, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
    ZEND_ARG_INFO(0, use_align)
    ZEND_ARG_INFO(0, row_align)
    ZEND_ARG_INFO(0, col_align)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_scroll_to_point, 0)
    ZEND_ARG_INFO(0, tree_x)
    ZEND_ARG_INFO(0, tree_y)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_column_drag_function, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_set_cursor, 0, 0, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, focus_column, GtkTreeViewColumn, 1)
    ZEND_ARG_INFO(0, start_editing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_set_cursor_on_cell, 0, 0, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, focus_column, GtkTreeViewColumn, 1)
    ZEND_ARG_OBJ_INFO(0, focus_cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, start_editing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_drag_dest_row, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_INFO(0, pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_enable_search, 0)
    ZEND_ARG_INFO(0, enable_search)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_expander_column, 0)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_fixed_height_mode, 0)
    ZEND_ARG_INFO(0, enable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_headers_clickable, 0)
    ZEND_ARG_INFO(0, active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_headers_visible, 0)
    ZEND_ARG_INFO(0, headers_visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_hover_expand, 0)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_hover_selection, 0)
    ZEND_ARG_INFO(0, hover)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_set_model, 0, 0, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_reorderable, 0)
    ZEND_ARG_INFO(0, reorderable)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_row_separator_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_rules_hint, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_search_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_search_equal_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_tree_to_widget_coords, 0)
    ZEND_ARG_INFO(0, tx)
    ZEND_ARG_INFO(0, ty)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_widget_to_tree_coords, 0)
    ZEND_ARG_INFO(0, wx)
    ZEND_ARG_INFO(0, wy)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_set_search_entry, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, entry, GtkEntry, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_rubber_banding, 0)
    ZEND_ARG_INFO(0, enable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_grid_lines, 0)
    ZEND_ARG_INFO(0, grid_lines)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_enable_tree_lines, 0)
    ZEND_ARG_INFO(0, enabled)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_widget_to_tree_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_tree_to_widget_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_widget_to_bin_window_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_bin_window_to_widget_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_tree_to_bin_window_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_convert_bin_window_to_tree_coords, 0)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_show_expanders, 0)
    ZEND_ARG_INFO(0, enabled)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_level_indentation, 0)
    ZEND_ARG_INFO(0, indentation)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktreeview_get_tooltip_context, 0, 0, 3)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
	ZEND_ARG_INFO(0, keyboard_mode)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_tooltip_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_tooltip_row, 0)
    ZEND_ARG_OBJ_INFO(0, tooltip, GtkTooltip, 1)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktreeview_set_tooltip_cell, 0)
    ZEND_ARG_OBJ_INFO(0, tooltip, GtkTooltip, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, column, GtkTreeViewColumn, 1)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtktreeview_methods[] = {
	PHP_ME(GtkTreeView, __construct,          arginfo_gtk_gtktreeview_gtk_tree_view_new_with_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, append_column,        arginfo_gtk_gtktreeview_append_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, collapse_all,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, collapse_row,         arginfo_gtk_gtktreeview_collapse_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, columns_autosize,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_bin_window_to_tree_coords, arginfo_gtk_gtktreeview_convert_bin_window_to_tree_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_bin_window_to_widget_coords, arginfo_gtk_gtktreeview_convert_bin_window_to_widget_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_tree_to_bin_window_coords, arginfo_gtk_gtktreeview_convert_tree_to_bin_window_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_tree_to_widget_coords, arginfo_gtk_gtktreeview_convert_tree_to_widget_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_widget_to_bin_window_coords, arginfo_gtk_gtktreeview_convert_widget_to_bin_window_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, convert_widget_to_tree_coords, arginfo_gtk_gtktreeview_convert_widget_to_tree_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, create_row_drag_icon, arginfo_gtk_gtktreeview_create_row_drag_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, enable_model_drag_dest, arginfo_gtk_gtktreeview_enable_model_drag_dest, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, enable_model_drag_source, arginfo_gtk_gtktreeview_enable_model_drag_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, expand_all,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, expand_row,           arginfo_gtk_gtktreeview_expand_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, expand_to_path,       arginfo_gtk_gtktreeview_expand_to_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_background_area,  arginfo_gtk_gtktreeview_get_background_area, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_bin_window,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_cell_area,        arginfo_gtk_gtktreeview_get_cell_area, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_column,           arginfo_gtk_gtktreeview_get_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_columns,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_cursor,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_dest_row_at_pos,  arginfo_gtk_gtktreeview_get_dest_row_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_drag_dest_row,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_enable_search,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_enable_tree_lines, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_expander_column,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_fixed_height_mode, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_grid_lines,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_headers_clickable, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_headers_visible,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_hover_expand,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_hover_selection,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_level_indentation, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_path_at_pos,      arginfo_gtk_gtktreeview_get_path_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_reorderable,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_rubber_banding,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_rules_hint,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_search_column,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_search_entry,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_selection,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_show_expanders,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_tooltip_column,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_tooltip_context,  arginfo_gtk_gtktreeview_get_tooltip_context, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_visible_range,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, get_visible_rect,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, insert_column,        arginfo_gtk_gtktreeview_insert_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, insert_column_with_data_func, arginfo_gtk_gtktreeview_insert_column_with_data_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, is_rubber_banding_active, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, move_column_after,    arginfo_gtk_gtktreeview_move_column_after, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, remove_column,        arginfo_gtk_gtktreeview_remove_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, row_activated,        arginfo_gtk_gtktreeview_row_activated, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, row_expanded,         arginfo_gtk_gtktreeview_row_expanded, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, scroll_to_cell,       arginfo_gtk_gtktreeview_scroll_to_cell, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, scroll_to_point,      arginfo_gtk_gtktreeview_scroll_to_point, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_column_drag_function, arginfo_gtk_gtktreeview_set_column_drag_function, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_cursor,           arginfo_gtk_gtktreeview_set_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_cursor_on_cell,   arginfo_gtk_gtktreeview_set_cursor_on_cell, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_drag_dest_row,    arginfo_gtk_gtktreeview_set_drag_dest_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_enable_search,    arginfo_gtk_gtktreeview_set_enable_search, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_enable_tree_lines, arginfo_gtk_gtktreeview_set_enable_tree_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_expander_column,  arginfo_gtk_gtktreeview_set_expander_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_fixed_height_mode, arginfo_gtk_gtktreeview_set_fixed_height_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_grid_lines,       arginfo_gtk_gtktreeview_set_grid_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_hadjustment,      arginfo_gtk_gtktreeview_set_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_headers_clickable, arginfo_gtk_gtktreeview_set_headers_clickable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_headers_visible,  arginfo_gtk_gtktreeview_set_headers_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_hover_expand,     arginfo_gtk_gtktreeview_set_hover_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_hover_selection,  arginfo_gtk_gtktreeview_set_hover_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_level_indentation, arginfo_gtk_gtktreeview_set_level_indentation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_model,            arginfo_gtk_gtktreeview_set_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_reorderable,      arginfo_gtk_gtktreeview_set_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_row_separator_func, arginfo_gtk_gtktreeview_set_row_separator_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_rubber_banding,   arginfo_gtk_gtktreeview_set_rubber_banding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_rules_hint,       arginfo_gtk_gtktreeview_set_rules_hint, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_search_column,    arginfo_gtk_gtktreeview_set_search_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_search_entry,     arginfo_gtk_gtktreeview_set_search_entry, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_search_equal_func, arginfo_gtk_gtktreeview_set_search_equal_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_show_expanders,   arginfo_gtk_gtktreeview_set_show_expanders, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_tooltip_cell,     arginfo_gtk_gtktreeview_set_tooltip_cell, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_tooltip_column,   arginfo_gtk_gtktreeview_set_tooltip_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_tooltip_row,      arginfo_gtk_gtktreeview_set_tooltip_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, set_vadjustment,      arginfo_gtk_gtktreeview_set_vadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, tree_to_widget_coords, arginfo_gtk_gtktreeview_tree_to_widget_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, unset_rows_drag_dest, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, unset_rows_drag_source, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTreeView, widget_to_tree_coords, arginfo_gtk_gtktreeview_widget_to_tree_coords, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolbar, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolbar);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolbar);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 5506 "ext/gtk+/gtk.overrides"

static gboolean phpg_toolbar_signal_func_marshal(GtkWidget *widget, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_widget = NULL;
    zend_bool toreturn;

    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return 0;
    }

    phpg_gobject_new(&php_widget, (GObject*)widget TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 1, &n_args);
    args[0] = &php_widget;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_widget);

    if (retval) {
        toreturn = zend_is_true(retval);
        zval_ptr_dtor(&retval);
    } else {
        toreturn = FALSE;
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);

    return toreturn;
}

static PHP_METHOD(GtkToolbar, append_item)
{
    zval *php_callback, *extra, *icon = NULL;
    phpg_cb_data_t *cb_data;
    GtkSignalFunc callback;
    char *text = NULL, *tooltip = NULL, *private_tooltip = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 5, &extra, "sssNV", &text, &tooltip, &private_tooltip, &icon, gtkwidget_ce, &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkSignalFunc)phpg_toolbar_signal_func_marshal;
        }
    }

    phpg_warn_deprecated("use GtkToolbar::insert" TSRMLS_CC);

    if (Z_TYPE_P(icon) != IS_NULL)
        gtk_toolbar_append_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)),
                                    text, tooltip, private_tooltip, GTK_WIDGET(PHPG_GOBJECT(icon)), callback, cb_data);
    else
        gtk_toolbar_append_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)),
                                    text, tooltip, private_tooltip, NULL, callback, cb_data);
}

#line 40682 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkToolbar, append_space)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toolbar_append_space(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolbar, append_widget)
{
	zval *widget;
	char *tooltip_text, *tooltip_private_text;
	zend_bool free_tooltip_text = FALSE, free_tooltip_private_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ouu", &widget, gtkwidget_ce, &tooltip_text, &free_tooltip_text, &tooltip_private_text, &free_tooltip_private_text))
		return;

    gtk_toolbar_append_widget(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), tooltip_text, tooltip_private_text);
	if (free_tooltip_text) g_free(tooltip_text);
	if (free_tooltip_private_text) g_free(tooltip_private_text);

}


static PHP_METHOD(GtkToolbar, get_drop_index)
{
	long x, y, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_toolbar_get_drop_index(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolbar, get_icon_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toolbar_get_icon_size(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolbar, get_item_index)
{
	zval *item;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &item, gtktoolitem_ce))
		return;

    php_retval = gtk_toolbar_get_item_index(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM(PHPG_GOBJECT(item)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolbar, get_n_items)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toolbar_get_n_items(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolbar, get_nth_item)
{
	long n;
	GtkToolItem* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n))
		return;

    php_retval = gtk_toolbar_get_nth_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gint)n);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolbar, get_relief_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toolbar_get_relief_style(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolbar, get_show_arrow)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toolbar_get_show_arrow(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolbar, get_tooltips)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.14, The toolkit-wide \"gtk-enable-tooltips\" property is now used instead." TSRMLS_CC);

    php_retval = gtk_toolbar_get_tooltips(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolbar, insert)
{
	zval *item;
	long pos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &item, gtktoolitem_ce, &pos))
		return;

    gtk_toolbar_insert(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), GTK_TOOL_ITEM(PHPG_GOBJECT(item)), (gint)pos);

}


static PHP_METHOD(GtkToolbar, insert_space)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_toolbar_insert_space(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkToolbar, insert_widget)
{
	zval *widget;
	char *tooltip_text, *tooltip_private_text;
	zend_bool free_tooltip_text = FALSE, free_tooltip_private_text = FALSE;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ouui", &widget, gtkwidget_ce, &tooltip_text, &free_tooltip_text, &tooltip_private_text, &free_tooltip_private_text, &position))
		return;

    gtk_toolbar_insert_widget(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), tooltip_text, tooltip_private_text, (gint)position);
	if (free_tooltip_text) g_free(tooltip_text);
	if (free_tooltip_private_text) g_free(tooltip_private_text);

}

#line 5595 "ext/gtk+/gtk.overrides"

static PHP_METHOD(GtkToolbar, prepend_item)
{
    zval *php_callback, *extra, *icon = NULL;
    phpg_cb_data_t *cb_data;
    GtkSignalFunc callback;
    char *text = NULL, *tooltip = NULL, *private_tooltip = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 5, &extra, "sssNV", &text, &tooltip, &private_tooltip, &icon, gtkwidget_ce, &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkSignalFunc)phpg_toolbar_signal_func_marshal;
        }
    }

    phpg_warn_deprecated("use GtkToolbar::insert" TSRMLS_CC);

    if (Z_TYPE_P(icon) != IS_NULL)
        gtk_toolbar_prepend_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)),
                                    text, tooltip, private_tooltip, GTK_WIDGET(PHPG_GOBJECT(icon)), callback, cb_data);
    else
        gtk_toolbar_prepend_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)),
                                    text, tooltip, private_tooltip, NULL, callback, cb_data);
}

#line 40915 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkToolbar, prepend_space)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toolbar_prepend_space(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolbar, prepend_widget)
{
	zval *widget;
	char *tooltip_text, *tooltip_private_text;
	zend_bool free_tooltip_text = FALSE, free_tooltip_private_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ouu", &widget, gtkwidget_ce, &tooltip_text, &free_tooltip_text, &tooltip_private_text, &free_tooltip_private_text))
		return;

    gtk_toolbar_prepend_widget(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), tooltip_text, tooltip_private_text);
	if (free_tooltip_text) g_free(tooltip_text);
	if (free_tooltip_private_text) g_free(tooltip_private_text);

}


static PHP_METHOD(GtkToolbar, remove_space)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_toolbar_remove_space(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkToolbar, set_drop_highlight_item)
{
	GtkToolItem *tool_item = NULL;
	zval *php_tool_item;
	long index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ni", &php_tool_item, gtktoolitem_ce, &index))
		return;
    if (Z_TYPE_P(php_tool_item) != IS_NULL)
        tool_item = GTK_TOOL_ITEM(PHPG_GOBJECT(php_tool_item));

    gtk_toolbar_set_drop_highlight_item(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), tool_item, (gint)index);

}


static PHP_METHOD(GtkToolbar, set_icon_size)
{
	GtkIconSize icon_size;
	zval *php_icon_size = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_icon_size))
		return;

	if (php_icon_size && phpg_gvalue_get_enum(GTK_TYPE_ICON_SIZE, php_icon_size, (gint *)&icon_size) == FAILURE) {
		return;
	}

    gtk_toolbar_set_icon_size(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), icon_size);

}


static PHP_METHOD(GtkToolbar, set_show_arrow)
{
	zend_bool show_arrow;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_arrow))
		return;

    gtk_toolbar_set_show_arrow(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gboolean)show_arrow);

}


static PHP_METHOD(GtkToolbar, set_tooltips)
{
	zend_bool enable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &enable))
		return;

	phpg_warn_deprecated("since GTK+ 2.14, The toolkit-wide \"gtk-enable-tooltips\" property is now used instead." TSRMLS_CC);

    gtk_toolbar_set_tooltips(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), (gboolean)enable);

}


static PHP_METHOD(GtkToolbar, unset_icon_size)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toolbar_unset_icon_size(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolbar, unset_style)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toolbar_unset_style(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));

}

#line 5631 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkToolbar, get_toolbar_style)
{
    long php_retval;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    php_retval = gtk_toolbar_get_style(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)));
    RETVAL_LONG(php_retval);
}

#line 41070 "ext/gtk+/gen_gtk.c"


#line 5653 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkToolbar, set_toolbar_style)
{
    GtkToolbarStyle style;
    zval *php_style = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_style))
        return;

    if (php_style && phpg_gvalue_get_enum(GTK_TYPE_TOOLBAR_STYLE, php_style, (gint *)&style) == FAILURE) {
        return;
    }

    gtk_toolbar_set_style(GTK_TOOLBAR(PHPG_GOBJECT(this_ptr)), style);
}

#line 41091 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_append_item, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, tooltip_text)
    ZEND_ARG_INFO(0, tooltip_private_text)
    ZEND_ARG_OBJ_INFO(0, icon, GtkWidget, 1)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_append_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, tooltip_text)
    ZEND_ARG_INFO(0, tooltip_private_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_get_drop_index, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_get_item_index, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkToolItem, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_get_nth_item, 0)
    ZEND_ARG_INFO(0, n)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_insert, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkToolItem, 1)
    ZEND_ARG_INFO(0, pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_insert_space, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_insert_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, tooltip_text)
    ZEND_ARG_INFO(0, tooltip_private_text)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_prepend_item, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, tooltip_text)
    ZEND_ARG_INFO(0, tooltip_private_text)
    ZEND_ARG_OBJ_INFO(0, icon, GtkWidget, 1)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_prepend_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, tooltip_text)
    ZEND_ARG_INFO(0, tooltip_private_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_remove_space, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_set_drop_highlight_item, 0)
    ZEND_ARG_OBJ_INFO(0, tool_item, GtkToolItem, 1)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_set_icon_size, 0)
    ZEND_ARG_INFO(0, icon_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_set_show_arrow, 0)
    ZEND_ARG_INFO(0, show_arrow)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_set_tooltips, 0)
    ZEND_ARG_INFO(0, enable)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbar_set_toolbar_style, 0)
    ZEND_ARG_INFO(0, style)
ZEND_END_ARG_INFO();


static zend_function_entry gtktoolbar_methods[] = {
	PHP_ME(GtkToolbar, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, append_item,          arginfo_gtk_gtktoolbar_append_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, append_space,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, append_widget,        arginfo_gtk_gtktoolbar_append_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_drop_index,       arginfo_gtk_gtktoolbar_get_drop_index, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_icon_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_item_index,       arginfo_gtk_gtktoolbar_get_item_index, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_n_items,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_nth_item,         arginfo_gtk_gtktoolbar_get_nth_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_relief_style,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_show_arrow,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_toolbar_style,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, get_tooltips,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, insert,               arginfo_gtk_gtktoolbar_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, insert_space,         arginfo_gtk_gtktoolbar_insert_space, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, insert_widget,        arginfo_gtk_gtktoolbar_insert_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, prepend_item,         arginfo_gtk_gtktoolbar_prepend_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, prepend_space,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, prepend_widget,       arginfo_gtk_gtktoolbar_prepend_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, remove_space,         arginfo_gtk_gtktoolbar_remove_space, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, set_drop_highlight_item, arginfo_gtk_gtktoolbar_set_drop_highlight_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, set_icon_size,        arginfo_gtk_gtktoolbar_set_icon_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, set_show_arrow,       arginfo_gtk_gtktoolbar_set_show_arrow, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, set_toolbar_style,    arginfo_gtk_gtktoolbar_set_toolbar_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, set_tooltips,         arginfo_gtk_gtktoolbar_set_tooltips, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, unset_icon_size,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolbar, unset_style,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTextView, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "buffer",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N", &php_args[0], gtktextbuffer_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextView);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextView);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTextView);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTextView, add_child_at_anchor)
{
	zval *child, *anchor;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &child, gtkwidget_ce, &anchor, gtktextchildanchor_ce))
		return;

    gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), GTK_TEXT_CHILD_ANCHOR(PHPG_GOBJECT(anchor)));

}


static PHP_METHOD(GtkTextView, add_child_in_window)
{
	zval *child, *php_which_window = NULL;
	GtkTextWindowType which_window;
	long xpos, ypos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVii", &child, gtkwidget_ce, &php_which_window, &xpos, &ypos))
		return;

	if (php_which_window && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_which_window, (gint *)&which_window) == FAILURE) {
		return;
	}

    gtk_text_view_add_child_in_window(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), which_window, (gint)xpos, (gint)ypos);

}


static PHP_METHOD(GtkTextView, backward_display_line)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_backward_display_line(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, backward_display_line_start)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_backward_display_line_start(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}

#line 629 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, buffer_to_window_coords)
{
    gint buffer_x, buffer_y, window_x, window_y;
    GtkTextWindowType win_type;
    zval *php_win_type = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vii", &php_win_type, &buffer_x, &buffer_y))
        return;

    if (php_win_type && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_win_type, (gint *)&win_type) == FAILURE)
        return;

    gtk_text_view_buffer_to_window_coords(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), win_type, buffer_x, buffer_y, &window_x, &window_y);
    php_gtk_build_value(&return_value, "(ii)", window_x, window_y);
}


#line 41361 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextView, forward_display_line)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_forward_display_line(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, forward_display_line_end)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_forward_display_line_end(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, get_accepts_tab)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_accepts_tab(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, get_border_window_size)
{
	GtkTextWindowType type;
	zval *php_type = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    php_retval = gtk_text_view_get_border_window_size(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), type);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_buffer)
{
	GtkTextBuffer* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_buffer(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextView, get_cursor_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_cursor_visible(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, get_default_attributes)
{
	GtkTextAttributes *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_default_attributes(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ATTRIBUTES, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkTextView, get_editable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_editable(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, get_indent)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_indent(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 658 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_iter_at_location)
{
    gint x, y;
    GtkTextIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
        return;

    gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), &iter, x, y);

    phpg_gboxed_new(&return_value, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);
}


#line 41534 "ext/gtk+/gen_gtk.c"


#line 684 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_iter_at_position)
{
    gint x, y, trailing;
    zval *php_iter = NULL;
    GtkTextIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
        return;

    gtk_text_view_get_iter_at_position(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), &iter, &trailing, x, y);

    phpg_gboxed_new(&php_iter, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);

    php_gtk_build_value(&return_value, "(Ni)", php_iter, trailing);
}


#line 41557 "ext/gtk+/gen_gtk.c"


#line 712 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_iter_location)
{
    zval *php_iter = NULL;
    GtkTextIter *iter = NULL;
    GdkRectangle rect;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gtktextiter_ce))
        return;

    if (php_iter && IS_OBJECT == Z_TYPE_P(php_iter)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    }

    gtk_text_view_get_iter_location(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter, &rect);
    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
}


#line 41581 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextView, get_justification)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_justification(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_left_margin)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_left_margin(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 741 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_line_at_y)
{
    gint y, line_top;
    zval *php_iter = NULL;
    GtkTextIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &y))
        return;

    gtk_text_view_get_line_at_y(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), &iter, y, &line_top);

    phpg_gboxed_new(&php_iter, GTK_TYPE_TEXT_ITER, &iter, TRUE, TRUE TSRMLS_CC);

    php_gtk_build_value(&return_value, "(Ni)", php_iter, line_top);
}


#line 41632 "ext/gtk+/gen_gtk.c"


#line 769 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_line_yrange)
{
    zval *php_iter = NULL;
    GtkTextIter *iter = NULL;
    gint y, height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gtktextiter_ce))
        return;

    if (php_iter && IS_OBJECT == Z_TYPE_P(php_iter)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    }

    gtk_text_view_get_line_yrange(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter, &y, &height);
    php_gtk_build_value(&return_value, "(ii)", y, height);
}

#line 41655 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextView, get_overwrite)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_overwrite(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, get_pixels_above_lines)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_pixels_above_lines(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_pixels_below_lines)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_pixels_below_lines(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_pixels_inside_wrap)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_pixels_inside_wrap(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_right_margin)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_right_margin(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_tabs)
{
	PangoTabArray *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_tabs(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, PANGO_TYPE_TAB_ARRAY, php_retval, TRUE, TRUE TSRMLS_CC);

}

#line 790 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, get_visible_rect)
{
    GdkRectangle rect;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), &rect);
    phpg_gboxed_new(&return_value, GDK_TYPE_RECTANGLE, &rect, TRUE, TRUE TSRMLS_CC);
}

#line 41757 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextView, get_window)
{
	GtkTextWindowType win;
	zval *php_win = NULL;
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_win))
		return;

	if (php_win && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_win, (gint *)&win) == FAILURE) {
		return;
	}

    php_retval = gtk_text_view_get_window(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), win);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextView, get_window_type)
{
	zval *window;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &window, gdkwindow_ce))
		return;

    php_retval = gtk_text_view_get_window_type(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GDK_WINDOW(PHPG_GOBJECT(window)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, get_wrap_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_wrap_mode(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTextView, move_child)
{
	zval *child;
	long xpos, ypos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &child, gtkwidget_ce, &xpos, &ypos))
		return;

    gtk_text_view_move_child(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)xpos, (gint)ypos);

}


static PHP_METHOD(GtkTextView, move_mark_onscreen)
{
	zval *mark;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &mark, gtktextmark_ce))
		return;

    php_retval = gtk_text_view_move_mark_onscreen(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, move_visually)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	long count;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_iter, gboxed_ce, &count))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_move_visually(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter, (gint)count);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, place_cursor_onscreen)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_place_cursor_onscreen(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, scroll_mark_onscreen)
{
	zval *mark;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &mark, gtktextmark_ce))
		return;

    gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)));

}


static PHP_METHOD(GtkTextView, scroll_to_iter)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	double within_margin, xalign = 0.5, yalign = 0.5;
	zend_bool use_align = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Od|bdd", &php_iter, gboxed_ce, &within_margin, &use_align, &xalign, &yalign))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter, within_margin, (gboolean)use_align, xalign, yalign);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, scroll_to_mark)
{
	zval *mark;
	double within_margin, xalign = 0.5, yalign = 0.5;
	zend_bool use_align = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Od|bdd", &mark, gtktextmark_ce, &within_margin, &use_align, &xalign, &yalign))
		return;

    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TEXT_MARK(PHPG_GOBJECT(mark)), within_margin, (gboolean)use_align, xalign, yalign);

}


static PHP_METHOD(GtkTextView, set_accepts_tab)
{
	zend_bool accepts_tab;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &accepts_tab))
		return;

    gtk_text_view_set_accepts_tab(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)accepts_tab);

}


static PHP_METHOD(GtkTextView, set_border_window_size)
{
	GtkTextWindowType type;
	zval *php_type = NULL;
	long size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vi", &php_type, &size))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_text_view_set_border_window_size(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), type, (gint)size);

}


static PHP_METHOD(GtkTextView, set_buffer)
{
	zval *buffer;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &buffer, gtktextbuffer_ce))
		return;

    gtk_text_view_set_buffer(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TEXT_BUFFER(PHPG_GOBJECT(buffer)));

}


static PHP_METHOD(GtkTextView, set_cursor_visible)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkTextView, set_editable)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_text_view_set_editable(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkTextView, set_indent)
{
	long indent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &indent))
		return;

    gtk_text_view_set_indent(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)indent);

}


static PHP_METHOD(GtkTextView, set_justification)
{
	GtkJustification justification;
	zval *php_justification = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_justification))
		return;

	if (php_justification && phpg_gvalue_get_enum(GTK_TYPE_JUSTIFICATION, php_justification, (gint *)&justification) == FAILURE) {
		return;
	}

    gtk_text_view_set_justification(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), justification);

}


static PHP_METHOD(GtkTextView, set_left_margin)
{
	long left_margin;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &left_margin))
		return;

    gtk_text_view_set_left_margin(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)left_margin);

}


static PHP_METHOD(GtkTextView, set_overwrite)
{
	zend_bool overwrite;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &overwrite))
		return;

    gtk_text_view_set_overwrite(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)overwrite);

}


static PHP_METHOD(GtkTextView, set_pixels_above_lines)
{
	long pixels_above_lines;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &pixels_above_lines))
		return;

    gtk_text_view_set_pixels_above_lines(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)pixels_above_lines);

}


static PHP_METHOD(GtkTextView, set_pixels_below_lines)
{
	long pixels_below_lines;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &pixels_below_lines))
		return;

    gtk_text_view_set_pixels_below_lines(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)pixels_below_lines);

}


static PHP_METHOD(GtkTextView, set_pixels_inside_wrap)
{
	long pixels_inside_wrap;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &pixels_inside_wrap))
		return;

    gtk_text_view_set_pixels_inside_wrap(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)pixels_inside_wrap);

}


static PHP_METHOD(GtkTextView, set_right_margin)
{
	long right_margin;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &right_margin))
		return;

    gtk_text_view_set_right_margin(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (gint)right_margin);

}


static PHP_METHOD(GtkTextView, set_tabs)
{
	PangoTabArray *tabs = NULL;
	zval *php_tabs;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_tabs, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_tabs, PANGO_TYPE_TAB_ARRAY, FALSE TSRMLS_CC)) {
        tabs = (PangoTabArray *) PHPG_GBOXED(php_tabs);
    } else {
        php_error(E_WARNING, "%s::%s() expects tabs argument to be a valid PangoTabArray object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_text_view_set_tabs(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), tabs);

}


static PHP_METHOD(GtkTextView, set_wrap_mode)
{
	GtkWrapMode wrap_mode;
	zval *php_wrap_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_wrap_mode))
		return;

	if (php_wrap_mode && phpg_gvalue_get_enum(GTK_TYPE_WRAP_MODE, php_wrap_mode, (gint *)&wrap_mode) == FAILURE) {
		return;
	}

    gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), wrap_mode);

}


static PHP_METHOD(GtkTextView, starts_display_line)
{
	GtkTextIter *iter = NULL;
	zval *php_iter;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TEXT_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTextIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTextIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_starts_display_line(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), iter);
	RETVAL_BOOL(php_retval);
}

#line 814 "ext/gtk+/gtktextview.overrides"
static PHP_METHOD(GtkTextView, window_to_buffer_coords)
{
    gint window_x, window_y, buffer_x, buffer_y;
    GtkTextWindowType win_type;
    zval *php_win_type = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vii", &php_win_type, &window_x, &window_y))
        return;

    if (php_win_type && phpg_gvalue_get_enum(GTK_TYPE_TEXT_WINDOW_TYPE, php_win_type, (gint *)&win_type) == FAILURE)
        return;

    gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), win_type, window_x, window_y, &buffer_x, &buffer_y);
    php_gtk_build_value(&return_value, "(ii)", buffer_x, buffer_y);
}

#line 42212 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkTextView, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_hadjustment(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextView, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_text_view_get_vadjustment(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkTextView, im_context_filter_keypress)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_text_view_im_context_filter_keypress(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)), (GdkEventKey *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTextView, reset_im_context)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_text_view_reset_im_context(GTK_TEXT_VIEW(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextview_gtk_text_view_new_with_buffer, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkTextBuffer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_add_child_at_anchor, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, anchor, GtkTextChildAnchor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_add_child_in_window, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, which_window)
    ZEND_ARG_INFO(0, xpos)
    ZEND_ARG_INFO(0, ypos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_backward_display_line, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_backward_display_line_start, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_buffer_to_window_coords, 0)
    ZEND_ARG_INFO(0, window_type)
    ZEND_ARG_INFO(0, buffer_x)
    ZEND_ARG_INFO(0, buffer_y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_forward_display_line, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_forward_display_line_end, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_border_window_size, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_iter_at_location, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_iter_at_position, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_iter_location, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 0)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_line_at_y, 0)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_line_yrange, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 0)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_window, 0)
    ZEND_ARG_INFO(0, win)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_get_window_type, 0)
    ZEND_ARG_OBJ_INFO(0, window, GdkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_move_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, xpos)
    ZEND_ARG_INFO(0, ypos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_move_mark_onscreen, 0)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_move_visually, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, count)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_scroll_mark_onscreen, 0)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextview_scroll_to_iter, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
    ZEND_ARG_INFO(0, within_margin)
    ZEND_ARG_INFO(0, use_align)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktextview_scroll_to_mark, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, mark, GtkTextMark, 1)
    ZEND_ARG_INFO(0, within_margin)
    ZEND_ARG_INFO(0, use_align)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_accepts_tab, 0)
    ZEND_ARG_INFO(0, accepts_tab)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_border_window_size, 0)
    ZEND_ARG_INFO(0, type)
    ZEND_ARG_INFO(0, size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_buffer, 0)
    ZEND_ARG_OBJ_INFO(0, buffer, GtkTextBuffer, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_cursor_visible, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_editable, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_indent, 0)
    ZEND_ARG_INFO(0, indent)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_justification, 0)
    ZEND_ARG_INFO(0, justification)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_left_margin, 0)
    ZEND_ARG_INFO(0, left_margin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_overwrite, 0)
    ZEND_ARG_INFO(0, overwrite)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_pixels_above_lines, 0)
    ZEND_ARG_INFO(0, pixels_above_lines)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_pixels_below_lines, 0)
    ZEND_ARG_INFO(0, pixels_below_lines)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_pixels_inside_wrap, 0)
    ZEND_ARG_INFO(0, pixels_inside_wrap)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_right_margin, 0)
    ZEND_ARG_INFO(0, right_margin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_tabs, 0)
    ZEND_ARG_OBJ_INFO(0, tabs, PangoTabArray, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_set_wrap_mode, 0)
    ZEND_ARG_INFO(0, wrap_mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_starts_display_line, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTextIter, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_window_to_buffer_coords, 0)
    ZEND_ARG_INFO(0, window_type)
    ZEND_ARG_INFO(0, window_x)
    ZEND_ARG_INFO(0, window_y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktextview_im_context_filter_keypress, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

static zend_function_entry gtktextview_methods[] = {
	PHP_ME(GtkTextView, __construct,          arginfo_gtk_gtktextview_gtk_text_view_new_with_buffer, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, add_child_at_anchor,  arginfo_gtk_gtktextview_add_child_at_anchor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, add_child_in_window,  arginfo_gtk_gtktextview_add_child_in_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, backward_display_line, arginfo_gtk_gtktextview_backward_display_line, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, backward_display_line_start, arginfo_gtk_gtktextview_backward_display_line_start, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, buffer_to_window_coords, arginfo_gtk_gtktextview_buffer_to_window_coords, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, forward_display_line, arginfo_gtk_gtktextview_forward_display_line, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, forward_display_line_end, arginfo_gtk_gtktextview_forward_display_line_end, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_accepts_tab,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_border_window_size, arginfo_gtk_gtktextview_get_border_window_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_buffer,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_cursor_visible,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_default_attributes, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_editable,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_indent,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_iter_at_location, arginfo_gtk_gtktextview_get_iter_at_location, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_iter_at_position, arginfo_gtk_gtktextview_get_iter_at_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_iter_location,    arginfo_gtk_gtktextview_get_iter_location, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_justification,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_left_margin,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_line_at_y,        arginfo_gtk_gtktextview_get_line_at_y, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_line_yrange,      arginfo_gtk_gtktextview_get_line_yrange, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_overwrite,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_pixels_above_lines, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_pixels_below_lines, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_pixels_inside_wrap, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_right_margin,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_tabs,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_visible_rect,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_window,           arginfo_gtk_gtktextview_get_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_window_type,      arginfo_gtk_gtktextview_get_window_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, get_wrap_mode,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, im_context_filter_keypress, arginfo_gtk_gtktextview_im_context_filter_keypress, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, move_child,           arginfo_gtk_gtktextview_move_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, move_mark_onscreen,   arginfo_gtk_gtktextview_move_mark_onscreen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, move_visually,        arginfo_gtk_gtktextview_move_visually, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, place_cursor_onscreen, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, reset_im_context,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, scroll_mark_onscreen, arginfo_gtk_gtktextview_scroll_mark_onscreen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, scroll_to_iter,       arginfo_gtk_gtktextview_scroll_to_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, scroll_to_mark,       arginfo_gtk_gtktextview_scroll_to_mark, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_accepts_tab,      arginfo_gtk_gtktextview_set_accepts_tab, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_border_window_size, arginfo_gtk_gtktextview_set_border_window_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_buffer,           arginfo_gtk_gtktextview_set_buffer, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_cursor_visible,   arginfo_gtk_gtktextview_set_cursor_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_editable,         arginfo_gtk_gtktextview_set_editable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_indent,           arginfo_gtk_gtktextview_set_indent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_justification,    arginfo_gtk_gtktextview_set_justification, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_left_margin,      arginfo_gtk_gtktextview_set_left_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_overwrite,        arginfo_gtk_gtktextview_set_overwrite, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_pixels_above_lines, arginfo_gtk_gtktextview_set_pixels_above_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_pixels_below_lines, arginfo_gtk_gtktextview_set_pixels_below_lines, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_pixels_inside_wrap, arginfo_gtk_gtktextview_set_pixels_inside_wrap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_right_margin,     arginfo_gtk_gtktextview_set_right_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_tabs,             arginfo_gtk_gtktextview_set_tabs, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, set_wrap_mode,        arginfo_gtk_gtktextview_set_wrap_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, starts_display_line,  arginfo_gtk_gtktextview_starts_display_line, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTextView, window_to_buffer_coords, arginfo_gtk_gtktextview_window_to_buffer_coords, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTable, __construct)
{
	GParameter params[3];
	zval *php_args[4] = { NULL, };
	char *prop_names[] = { "n_rows", "n_columns", "homogeneous",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^i^i^b", &php_args[0], &php_args[1], &php_args[2])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTable);
	}

	memset(&params, 0, 3 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTable);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTable);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkTable, attach)
{
	zval *child, *php_xoptions = NULL, *php_yoptions = NULL;
	long left_attach, right_attach, top_attach, bottom_attach, xpadding = GTK_EXPAND|GTK_FILL, ypadding = GTK_EXPAND|GTK_FILL;
	GtkAttachOptions xoptions = GTK_EXPAND|GTK_FILL, yoptions = GTK_EXPAND|GTK_FILL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oiiii|VVii", &child, gtkwidget_ce, &left_attach, &right_attach, &top_attach, &bottom_attach, &php_xoptions, &php_yoptions, &xpadding, &ypadding))
		return;

	if (php_xoptions && phpg_gvalue_get_flags(GTK_TYPE_ATTACH_OPTIONS, php_xoptions, (gint *)&xoptions) == FAILURE) {
		return;
	}

	if (php_yoptions && phpg_gvalue_get_flags(GTK_TYPE_ATTACH_OPTIONS, php_yoptions, (gint *)&yoptions) == FAILURE) {
		return;
	}

    gtk_table_attach(GTK_TABLE(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (guint)left_attach, (guint)right_attach, (guint)top_attach, (guint)bottom_attach, xoptions, yoptions, (guint)xpadding, (guint)ypadding);

}


static PHP_METHOD(GtkTable, attach_defaults)
{
	zval *widget;
	long left_attach, right_attach, top_attach, bottom_attach;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oiiii", &widget, gtkwidget_ce, &left_attach, &right_attach, &top_attach, &bottom_attach))
		return;

	phpg_warn_deprecated("use attach instead" TSRMLS_CC);

    gtk_table_attach_defaults(GTK_TABLE(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), (guint)left_attach, (guint)right_attach, (guint)top_attach, (guint)bottom_attach);

}


static PHP_METHOD(GtkTable, get_col_spacing)
{
	long column, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    php_retval = gtk_table_get_col_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)column);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTable, get_default_col_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_table_get_default_col_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTable, get_default_row_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_table_get_default_row_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTable, get_homogeneous)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_table_get_homogeneous(GTK_TABLE(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkTable, get_row_spacing)
{
	long row, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
		return;

    php_retval = gtk_table_get_row_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)row);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkTable, resize)
{
	long rows, columns;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &rows, &columns))
		return;

    gtk_table_resize(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)rows, (guint)columns);

}


static PHP_METHOD(GtkTable, set_col_spacing)
{
	long column, spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &column, &spacing))
		return;

    gtk_table_set_col_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)column, (guint)spacing);

}


static PHP_METHOD(GtkTable, set_col_spacings)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_table_set_col_spacings(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)spacing);

}


static PHP_METHOD(GtkTable, set_homogeneous)
{
	zend_bool homogeneous;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &homogeneous))
		return;

    gtk_table_set_homogeneous(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (gboolean)homogeneous);

}


static PHP_METHOD(GtkTable, set_row_spacing)
{
	long row, spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &spacing))
		return;

    gtk_table_set_row_spacing(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)row, (guint)spacing);

}


static PHP_METHOD(GtkTable, set_row_spacings)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_table_set_row_spacings(GTK_TABLE(PHPG_GOBJECT(this_ptr)), (guint)spacing);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktable_gtk_table_new, 0, 0, 0)
    ZEND_ARG_INFO(0, n_rows)
    ZEND_ARG_INFO(0, n_columns)
    ZEND_ARG_INFO(0, homogeneous)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktable_attach, 0, 0, 5)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, left_attach)
    ZEND_ARG_INFO(0, right_attach)
    ZEND_ARG_INFO(0, top_attach)
    ZEND_ARG_INFO(0, bottom_attach)
    ZEND_ARG_INFO(0, xoptions)
    ZEND_ARG_INFO(0, yoptions)
    ZEND_ARG_INFO(0, xpadding)
    ZEND_ARG_INFO(0, ypadding)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_attach_defaults, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, left_attach)
    ZEND_ARG_INFO(0, right_attach)
    ZEND_ARG_INFO(0, top_attach)
    ZEND_ARG_INFO(0, bottom_attach)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_get_col_spacing, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_get_row_spacing, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_resize, 0)
    ZEND_ARG_INFO(0, rows)
    ZEND_ARG_INFO(0, columns)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_set_col_spacing, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_set_col_spacings, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_set_homogeneous, 0)
    ZEND_ARG_INFO(0, homogeneous)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_set_row_spacing, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktable_set_row_spacings, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtktable_methods[] = {
	PHP_ME(GtkTable, __construct,          arginfo_gtk_gtktable_gtk_table_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, attach,               arginfo_gtk_gtktable_attach, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, attach_defaults,      arginfo_gtk_gtktable_attach_defaults, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, get_col_spacing,      arginfo_gtk_gtktable_get_col_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, get_default_col_spacing, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, get_default_row_spacing, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, get_homogeneous,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, get_row_spacing,      arginfo_gtk_gtktable_get_row_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, resize,               arginfo_gtk_gtktable_resize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, set_col_spacing,      arginfo_gtk_gtktable_set_col_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, set_col_spacings,     arginfo_gtk_gtktable_set_col_spacings, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, set_homogeneous,      arginfo_gtk_gtktable_set_homogeneous, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, set_row_spacing,      arginfo_gtk_gtktable_set_row_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkTable, set_row_spacings,     arginfo_gtk_gtktable_set_row_spacings, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSocket, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSocket);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSocket);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkSocket, add_id)
{
	long window_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &window_id))
		return;

    gtk_socket_add_id(GTK_SOCKET(PHPG_GOBJECT(this_ptr)), (GdkNativeWindow)window_id);

}


static PHP_METHOD(GtkSocket, get_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_socket_get_id(GTK_SOCKET(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkSocket, steal)
{
	long wid;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &wid))
		return;

    gtk_socket_steal(GTK_SOCKET(PHPG_GOBJECT(this_ptr)), (GdkNativeWindow)wid);

}


static PHP_METHOD(GtkSocket, get_plug_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_socket_get_plug_window(GTK_SOCKET(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksocket_add_id, 0)
    ZEND_ARG_INFO(0, window_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtksocket_steal, 0)
    ZEND_ARG_INFO(0, wid)
ZEND_END_ARG_INFO();

static zend_function_entry gtksocket_methods[] = {
	PHP_ME(GtkSocket, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSocket, add_id,               arginfo_gtk_gtksocket_add_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSocket, get_id,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSocket, get_plug_window,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSocket, steal,                arginfo_gtk_gtksocket_steal, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkPaned, add1)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_paned_add1(GTK_PANED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkPaned, add2)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_paned_add2(GTK_PANED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkPaned, compute_position)
{
	long allocation, child1_req, child2_req;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iii", &allocation, &child1_req, &child2_req))
		return;

    gtk_paned_compute_position(GTK_PANED(PHPG_GOBJECT(this_ptr)), (gint)allocation, (gint)child1_req, (gint)child2_req);

}


static PHP_METHOD(GtkPaned, get_child1)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paned_get_child1(GTK_PANED(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPaned, get_child2)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paned_get_child2(GTK_PANED(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkPaned, get_position)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paned_get_position(GTK_PANED(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPaned, pack1)
{
	zval *child;
	zend_bool resize = FALSE, shrink = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|bb", &child, gtkwidget_ce, &resize, &shrink))
		return;

    gtk_paned_pack1(GTK_PANED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)resize, (gboolean)shrink);

}


static PHP_METHOD(GtkPaned, pack2)
{
	zval *child;
	zend_bool resize = TRUE, shrink = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|bb", &child, gtkwidget_ce, &resize, &shrink))
		return;

    gtk_paned_pack2(GTK_PANED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)resize, (gboolean)shrink);

}


static PHP_METHOD(GtkPaned, set_position)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_paned_set_position(GTK_PANED(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkPaned, get_handle_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_paned_get_handle_window(GTK_PANED(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpaned_add1, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpaned_add2, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpaned_compute_position, 0)
    ZEND_ARG_INFO(0, allocation)
    ZEND_ARG_INFO(0, child1_req)
    ZEND_ARG_INFO(0, child2_req)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkpaned_pack1, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, resize)
    ZEND_ARG_INFO(0, shrink)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkpaned_pack2, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, resize)
    ZEND_ARG_INFO(0, shrink)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkpaned_set_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

static zend_function_entry gtkpaned_methods[] = {
	PHP_ME(GtkPaned, add1,                 arginfo_gtk_gtkpaned_add1, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, add2,                 arginfo_gtk_gtkpaned_add2, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, compute_position,     arginfo_gtk_gtkpaned_compute_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, get_child1,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, get_child2,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, get_handle_window,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, get_position,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, pack1,                arginfo_gtk_gtkpaned_pack1, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, pack2,                arginfo_gtk_gtkpaned_pack2, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPaned, set_position,         arginfo_gtk_gtkpaned_set_position, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkNotebook, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkNotebook);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkNotebook);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkNotebook, tab_pos)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_NOTEBOOK(((phpg_gobject_t *)object)->obj)->tab_pos;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtknotebook_prop_info[] = {
	{ "tab_pos", PHPG_PROP_READ_FN(GtkNotebook, tab_pos), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkNotebook, append_page)
{
	zval *child, *php_tab_label = NULL;
	GtkWidget *tab_label = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }

    php_retval = gtk_notebook_append_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, append_page_menu)
{
	zval *child, *php_tab_label = NULL, *php_menu_label = NULL;
	GtkWidget *tab_label = NULL, *menu_label = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|NN", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce, &php_menu_label, gtkwidget_ce))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }
    if (php_menu_label) {
        if (Z_TYPE_P(php_menu_label) == IS_NULL)
            menu_label = NULL;
        else
            menu_label = GTK_WIDGET(PHPG_GOBJECT(php_menu_label));
    }

    php_retval = gtk_notebook_append_page_menu(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label, menu_label);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, current_page)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkNotebook::get_current_page()" TSRMLS_CC);

    php_retval = gtk_notebook_current_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, get_current_page)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_current_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, get_menu_label)
{
	zval *child;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_menu_label(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkNotebook, get_menu_label_text)
{
	zval *child;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_menu_label_text(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkNotebook, get_n_pages)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_n_pages(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, get_nth_page)
{
	long page_num;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

    php_retval = gtk_notebook_get_nth_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gint)page_num);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkNotebook, get_scrollable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_scrollable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkNotebook, get_show_border)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_show_border(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkNotebook, get_show_tabs)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_show_tabs(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkNotebook, get_tab_label)
{
	zval *child;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_tab_label(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkNotebook, get_tab_label_text)
{
	zval *child;
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_tab_label_text(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkNotebook, get_tab_pos)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_tab_pos(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, insert_page)
{
	zval *child, *php_tab_label = NULL;
	GtkWidget *tab_label = NULL;
	long position = -1, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|Ni", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce, &position))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }

    php_retval = gtk_notebook_insert_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label, (gint)position);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, insert_page_menu)
{
	zval *child, *php_tab_label = NULL, *php_menu_label = NULL;
	GtkWidget *tab_label = NULL, *menu_label = NULL;
	long position = -1, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|NNi", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce, &php_menu_label, gtkwidget_ce, &position))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }
    if (php_menu_label) {
        if (Z_TYPE_P(php_menu_label) == IS_NULL)
            menu_label = NULL;
        else
            menu_label = GTK_WIDGET(PHPG_GOBJECT(php_menu_label));
    }

    php_retval = gtk_notebook_insert_page_menu(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label, menu_label, (gint)position);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, next_page)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_notebook_next_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkNotebook, page_num)
{
	zval *child;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_page_num(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, popup_disable)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_notebook_popup_disable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkNotebook, popup_enable)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_notebook_popup_enable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkNotebook, prepend_page)
{
	zval *child, *php_tab_label = NULL;
	GtkWidget *tab_label = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }

    php_retval = gtk_notebook_prepend_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, prepend_page_menu)
{
	zval *child, *php_tab_label = NULL, *php_menu_label = NULL;
	GtkWidget *tab_label = NULL, *menu_label = NULL;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|NN", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce, &php_menu_label, gtkwidget_ce))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }
    if (php_menu_label) {
        if (Z_TYPE_P(php_menu_label) == IS_NULL)
            menu_label = NULL;
        else
            menu_label = GTK_WIDGET(PHPG_GOBJECT(php_menu_label));
    }

    php_retval = gtk_notebook_prepend_page_menu(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label, menu_label);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, prev_page)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_notebook_prev_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));

}

#line 4647 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkNotebook, query_tab_label_packing)
{
    zval *php_child;
    gboolean expand, fill;
    GtkPackType pack_type;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_child, gtkwidget_ce))
        return;

    gtk_notebook_query_tab_label_packing(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(php_child)), &expand, &fill, &pack_type);
    php_gtk_build_value(&return_value, "(bbi)", expand, fill, pack_type);
}

#line 43658 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkNotebook, remove_page)
{
	long page_num;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

    gtk_notebook_remove_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gint)page_num);

}


static PHP_METHOD(GtkNotebook, reorder_child)
{
	zval *child;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &position))
		return;

    gtk_notebook_reorder_child(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)position);

}


static PHP_METHOD(GtkNotebook, set_current_page)
{
	long page_num;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

    gtk_notebook_set_current_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gint)page_num);

}


static PHP_METHOD(GtkNotebook, set_homogeneous_tabs)
{
	zend_bool homogeneous;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &homogeneous))
		return;

    gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gboolean)homogeneous);

}


static PHP_METHOD(GtkNotebook, set_menu_label)
{
	zval *child, *php_menu_label = NULL;
	GtkWidget *menu_label = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &child, gtkwidget_ce, &php_menu_label, gtkwidget_ce))
		return;
    if (php_menu_label) {
        if (Z_TYPE_P(php_menu_label) == IS_NULL)
            menu_label = NULL;
        else
            menu_label = GTK_WIDGET(PHPG_GOBJECT(php_menu_label));
    }

    gtk_notebook_set_menu_label(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), menu_label);

}


static PHP_METHOD(GtkNotebook, set_menu_label_text)
{
	zval *child;
	char *menu_text;
	zend_bool free_menu_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &child, gtkwidget_ce, &menu_text, &free_menu_text))
		return;

    gtk_notebook_set_menu_label_text(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), menu_text);
	if (free_menu_text) g_free(menu_text);

}


static PHP_METHOD(GtkNotebook, set_page)
{
	long page_num;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &page_num))
		return;

	phpg_warn_deprecated("use GtkNotebook::set_current_page()" TSRMLS_CC);

    gtk_notebook_set_page(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gint)page_num);

}


static PHP_METHOD(GtkNotebook, set_scrollable)
{
	zend_bool scrollable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &scrollable))
		return;

    gtk_notebook_set_scrollable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gboolean)scrollable);

}


static PHP_METHOD(GtkNotebook, set_show_border)
{
	zend_bool show_border;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_border))
		return;

    gtk_notebook_set_show_border(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gboolean)show_border);

}


static PHP_METHOD(GtkNotebook, set_show_tabs)
{
	zend_bool show_tabs;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_tabs))
		return;

    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gboolean)show_tabs);

}


static PHP_METHOD(GtkNotebook, set_tab_border)
{
	long border_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &border_width))
		return;

    gtk_notebook_set_tab_border(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (guint)border_width);

}


static PHP_METHOD(GtkNotebook, set_tab_hborder)
{
	long tab_hborder;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &tab_hborder))
		return;

    gtk_notebook_set_tab_hborder(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (guint)tab_hborder);

}


static PHP_METHOD(GtkNotebook, set_tab_label)
{
	zval *child, *php_tab_label = NULL;
	GtkWidget *tab_label = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|N", &child, gtkwidget_ce, &php_tab_label, gtkwidget_ce))
		return;
    if (php_tab_label) {
        if (Z_TYPE_P(php_tab_label) == IS_NULL)
            tab_label = NULL;
        else
            tab_label = GTK_WIDGET(PHPG_GOBJECT(php_tab_label));
    }

    gtk_notebook_set_tab_label(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_label);

}


static PHP_METHOD(GtkNotebook, set_tab_label_packing)
{
	zval *child, *php_pack_type = NULL;
	zend_bool expand, fill;
	GtkPackType pack_type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ObbV", &child, gtkwidget_ce, &expand, &fill, &php_pack_type))
		return;

	if (php_pack_type && phpg_gvalue_get_enum(GTK_TYPE_PACK_TYPE, php_pack_type, (gint *)&pack_type) == FAILURE) {
		return;
	}

    gtk_notebook_set_tab_label_packing(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)expand, (gboolean)fill, pack_type);

}


static PHP_METHOD(GtkNotebook, set_tab_label_text)
{
	zval *child;
	char *tab_text;
	zend_bool free_tab_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &child, gtkwidget_ce, &tab_text, &free_tab_text))
		return;

    gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), tab_text);
	if (free_tab_text) g_free(tab_text);

}


static PHP_METHOD(GtkNotebook, set_tab_pos)
{
	GtkPositionType pos;
	zval *php_pos = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_pos))
		return;

	if (php_pos && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_pos, (gint *)&pos) == FAILURE) {
		return;
	}

    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), pos);

}


static PHP_METHOD(GtkNotebook, set_tab_vborder)
{
	long tab_vborder;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &tab_vborder))
		return;

    gtk_notebook_set_tab_vborder(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (guint)tab_vborder);

}


static PHP_METHOD(GtkNotebook, set_group_id)
{
	long group_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &group_id))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkNotebook::set_group()" TSRMLS_CC);

    gtk_notebook_set_group_id(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gint)group_id);

}


static PHP_METHOD(GtkNotebook, get_group_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkNotebook::get_group()" TSRMLS_CC);

    php_retval = gtk_notebook_get_group_id(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, get_tab_reorderable)
{
	zval *child;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_tab_reorderable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkNotebook, set_tab_reorderable)
{
	zval *child;
	zend_bool reorderable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &child, gtkwidget_ce, &reorderable))
		return;

    gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)reorderable);

}


static PHP_METHOD(GtkNotebook, get_tab_detachable)
{
	zval *child;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_notebook_get_tab_detachable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkNotebook, set_tab_detachable)
{
	zval *child;
	zend_bool detachable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &child, gtkwidget_ce, &detachable))
		return;

    gtk_notebook_set_tab_detachable(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)detachable);

}

#line 282 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkNotebook, set_group)
{
	zval *php_pointer = NULL;

	NOT_STATIC_METHOD();

	/* Check for a gtknotebook instance first */
	if (php_gtk_parse_args_quiet(ZEND_NUM_ARGS(),"O", &php_pointer, gtknotebook_ce)) {
	gtk_notebook_set_group(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gpointer) gtk_notebook_get_group(GTK_NOTEBOOK(PHPG_GOBJECT(php_pointer))));
	return;
	}

	/* Otherwise we want any other zval or null to "turn off" the group */
	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"|V", &php_pointer)) {
		return;
	}

	gtk_notebook_set_group(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), (gpointer) php_pointer);
}

#line 44046 "ext/gtk+/gen_gtk.c"


#line 304 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkNotebook, get_group)
{
	zval *php_pointer = NULL;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"")) {
		return;
	}

	php_pointer = (zval *)gtk_notebook_get_group(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETURN_LONG(php_pointer);
}

#line 44064 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkNotebook, get_action_widget)
{
	GtkPackType pack_type;
	zval *php_pack_type = NULL;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_pack_type))
		return;

	if (php_pack_type && phpg_gvalue_get_enum(GTK_TYPE_PACK_TYPE, php_pack_type, (gint *)&pack_type) == FAILURE) {
		return;
	}

    php_retval = gtk_notebook_get_action_widget(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), pack_type);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkNotebook, set_action_widget)
{
	zval *widget, *php_pack_type = NULL;
	GtkPackType pack_type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &widget, gtkwidget_ce, &php_pack_type))
		return;

	if (php_pack_type && phpg_gvalue_get_enum(GTK_TYPE_PACK_TYPE, php_pack_type, (gint *)&pack_type) == FAILURE) {
		return;
	}

    gtk_notebook_set_action_widget(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), pack_type);

}


static PHP_METHOD(GtkNotebook, get_tab_hborder)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_tab_hborder(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, get_tab_vborder)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_tab_vborder(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkNotebook, set_group_name)
{
	char *group_name;
	zend_bool free_group_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &group_name, &free_group_name))
		return;

    gtk_notebook_set_group_name(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)), group_name);
	if (free_group_name) g_free(group_name);

}


static PHP_METHOD(GtkNotebook, get_group_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_notebook_get_group_name(GTK_NOTEBOOK(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_append_page, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_append_page_menu, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, menu_label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_menu_label, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_menu_label_text, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_nth_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_tab_label, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_tab_label_text, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_insert_page, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_insert_page_menu, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, menu_label, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_page_num, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_prepend_page, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_prepend_page_menu, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, menu_label, GtkWidget, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_query_tab_label_packing, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_remove_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_reorder_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_current_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_homogeneous_tabs, 0)
    ZEND_ARG_INFO(0, homogeneous)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_set_menu_label, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, menu_label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_menu_label_text, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, menu_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_page, 0)
    ZEND_ARG_INFO(0, page_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_scrollable, 0)
    ZEND_ARG_INFO(0, scrollable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_show_border, 0)
    ZEND_ARG_INFO(0, show_border)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_show_tabs, 0)
    ZEND_ARG_INFO(0, show_tabs)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_border, 0)
    ZEND_ARG_INFO(0, border_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_hborder, 0)
    ZEND_ARG_INFO(0, tab_hborder)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_set_tab_label, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, tab_label, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_label_packing, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, expand)
    ZEND_ARG_INFO(0, fill)
    ZEND_ARG_INFO(0, pack_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_label_text, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, tab_text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_pos, 0)
    ZEND_ARG_INFO(0, pos)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_vborder, 0)
    ZEND_ARG_INFO(0, tab_vborder)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_group_id, 0)
    ZEND_ARG_INFO(0, group_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_tab_reorderable, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_reorderable, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, reorderable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_tab_detachable, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_tab_detachable, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, detachable)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtknotebook_set_group, 0, 0, 0)
	ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO()


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_get_action_widget, 0)
    ZEND_ARG_INFO(0, pack_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_action_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, pack_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtknotebook_set_group_name, 0)
    ZEND_ARG_INFO(0, group_name)
ZEND_END_ARG_INFO();

static zend_function_entry gtknotebook_methods[] = {
	PHP_ME(GtkNotebook, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, append_page,          arginfo_gtk_gtknotebook_append_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, append_page_menu,     arginfo_gtk_gtknotebook_append_page_menu, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, current_page,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_action_widget,    arginfo_gtk_gtknotebook_get_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_current_page,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_group_id,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_group_name,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_menu_label,       arginfo_gtk_gtknotebook_get_menu_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_menu_label_text,  arginfo_gtk_gtknotebook_get_menu_label_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_n_pages,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_nth_page,         arginfo_gtk_gtknotebook_get_nth_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_scrollable,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_show_border,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_show_tabs,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_detachable,   arginfo_gtk_gtknotebook_get_tab_detachable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_hborder,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_label,        arginfo_gtk_gtknotebook_get_tab_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_label_text,   arginfo_gtk_gtknotebook_get_tab_label_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_pos,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_reorderable,  arginfo_gtk_gtknotebook_get_tab_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, get_tab_vborder,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, insert_page,          arginfo_gtk_gtknotebook_insert_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, insert_page_menu,     arginfo_gtk_gtknotebook_insert_page_menu, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, next_page,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, page_num,             arginfo_gtk_gtknotebook_page_num, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, popup_disable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, popup_enable,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, prepend_page,         arginfo_gtk_gtknotebook_prepend_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, prepend_page_menu,    arginfo_gtk_gtknotebook_prepend_page_menu, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, prev_page,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, query_tab_label_packing, arginfo_gtk_gtknotebook_query_tab_label_packing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, remove_page,          arginfo_gtk_gtknotebook_remove_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, reorder_child,        arginfo_gtk_gtknotebook_reorder_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_action_widget,    arginfo_gtk_gtknotebook_set_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_current_page,     arginfo_gtk_gtknotebook_set_current_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_group,            arginfo_gtk_gtknotebook_set_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_group_id,         arginfo_gtk_gtknotebook_set_group_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_group_name,       arginfo_gtk_gtknotebook_set_group_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_homogeneous_tabs, arginfo_gtk_gtknotebook_set_homogeneous_tabs, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_menu_label,       arginfo_gtk_gtknotebook_set_menu_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_menu_label_text,  arginfo_gtk_gtknotebook_set_menu_label_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_page,             arginfo_gtk_gtknotebook_set_page, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_scrollable,       arginfo_gtk_gtknotebook_set_scrollable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_show_border,      arginfo_gtk_gtknotebook_set_show_border, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_show_tabs,        arginfo_gtk_gtknotebook_set_show_tabs, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_border,       arginfo_gtk_gtknotebook_set_tab_border, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_detachable,   arginfo_gtk_gtknotebook_set_tab_detachable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_hborder,      arginfo_gtk_gtknotebook_set_tab_hborder, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_label,        arginfo_gtk_gtknotebook_set_tab_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_label_packing, arginfo_gtk_gtknotebook_set_tab_label_packing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_label_text,   arginfo_gtk_gtknotebook_set_tab_label_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_pos,          arginfo_gtk_gtknotebook_set_tab_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_reorderable,  arginfo_gtk_gtknotebook_set_tab_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkNotebook, set_tab_vborder,      arginfo_gtk_gtknotebook_set_tab_vborder, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVPaned, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVPaned);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVPaned);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkvpaned_methods[] = {
	PHP_ME(GtkVPaned, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHPaned, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHPaned);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHPaned);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkhpaned_methods[] = {
	PHP_ME(GtkHPaned, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkMenuShell, activate_item)
{
	zval *menu_item;
	zend_bool force_deactivate;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &menu_item, gtkwidget_ce, &force_deactivate))
		return;

    gtk_menu_shell_activate_item(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(menu_item)), (gboolean)force_deactivate);

}


static PHP_METHOD(GtkMenuShell, append)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_menu_shell_append(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkMenuShell, cancel)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_shell_cancel(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuShell, deactivate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_shell_deactivate(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuShell, deselect)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_shell_deselect(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuShell, insert)
{
	zval *child;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &position))
		return;

    gtk_menu_shell_insert(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)position);

}


static PHP_METHOD(GtkMenuShell, prepend)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_menu_shell_prepend(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkMenuShell, select_first)
{
	zend_bool search_sensitive;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &search_sensitive))
		return;

    gtk_menu_shell_select_first(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), (gboolean)search_sensitive);

}


static PHP_METHOD(GtkMenuShell, select_item)
{
	zval *menu_item;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &menu_item, gtkwidget_ce))
		return;

    gtk_menu_shell_select_item(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(menu_item)));

}


static PHP_METHOD(GtkMenuShell, get_take_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_shell_get_take_focus(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkMenuShell, set_take_focus)
{
	zend_bool take_focus;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &take_focus))
		return;

    gtk_menu_shell_set_take_focus(GTK_MENU_SHELL(PHPG_GOBJECT(this_ptr)), (gboolean)take_focus);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_activate_item, 0)
    ZEND_ARG_OBJ_INFO(0, menu_item, GtkWidget, 1)
    ZEND_ARG_INFO(0, force_deactivate)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_append, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_insert, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_prepend, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_select_first, 0)
    ZEND_ARG_INFO(0, search_sensitive)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_select_item, 0)
    ZEND_ARG_OBJ_INFO(0, menu_item, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenushell_set_take_focus, 0)
    ZEND_ARG_INFO(0, take_focus)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmenushell_methods[] = {
	PHP_ME(GtkMenuShell, activate_item,        arginfo_gtk_gtkmenushell_activate_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, append,               arginfo_gtk_gtkmenushell_append, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, cancel,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, deactivate,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, deselect,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, get_take_focus,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, insert,               arginfo_gtk_gtkmenushell_insert, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, prepend,              arginfo_gtk_gtkmenushell_prepend, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, select_first,         arginfo_gtk_gtkmenushell_select_first, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, select_item,          arginfo_gtk_gtkmenushell_select_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuShell, set_take_focus,       arginfo_gtk_gtkmenushell_set_take_focus, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkList, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkList);
	}


	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkList);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkList, child_position)
{
	zval *child;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_list_child_position(GTK_LIST(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkList, clear_items)
{
	long start, end;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &start, &end))
		return;

    gtk_list_clear_items(GTK_LIST(PHPG_GOBJECT(this_ptr)), (gint)start, (gint)end);

}


static PHP_METHOD(GtkList, end_drag_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_end_drag_selection(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, end_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_end_selection(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, extend_selection)
{
	GtkScrollType scroll_type;
	zval *php_scroll_type = NULL;
	double position;
	zend_bool auto_start_selection;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vdb", &php_scroll_type, &position, &auto_start_selection))
		return;

	if (php_scroll_type && phpg_gvalue_get_enum(GTK_TYPE_SCROLL_TYPE, php_scroll_type, (gint *)&scroll_type) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_extend_selection(GTK_LIST(PHPG_GOBJECT(this_ptr)), scroll_type, (float)position, (gboolean)auto_start_selection);

}


static PHP_METHOD(GtkList, scroll_horizontal)
{
	GtkScrollType scroll_type;
	zval *php_scroll_type = NULL;
	double position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vd", &php_scroll_type, &position))
		return;

	if (php_scroll_type && phpg_gvalue_get_enum(GTK_TYPE_SCROLL_TYPE, php_scroll_type, (gint *)&scroll_type) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_scroll_horizontal(GTK_LIST(PHPG_GOBJECT(this_ptr)), scroll_type, (float)position);

}


static PHP_METHOD(GtkList, scroll_vertical)
{
	GtkScrollType scroll_type;
	zval *php_scroll_type = NULL;
	double position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vd", &php_scroll_type, &position))
		return;

	if (php_scroll_type && phpg_gvalue_get_enum(GTK_TYPE_SCROLL_TYPE, php_scroll_type, (gint *)&scroll_type) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_scroll_vertical(GTK_LIST(PHPG_GOBJECT(this_ptr)), scroll_type, (float)position);

}


static PHP_METHOD(GtkList, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_select_all(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, select_child)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_list_select_child(GTK_LIST(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkList, select_item)
{
	long item;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &item))
		return;

    gtk_list_select_item(GTK_LIST(PHPG_GOBJECT(this_ptr)), (gint)item);

}


static PHP_METHOD(GtkList, set_selection_mode)
{
	GtkSelectionMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(GTK_TYPE_SELECTION_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_set_selection_mode(GTK_LIST(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkList, start_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_start_selection(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, toggle_add_mode)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_toggle_add_mode(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, toggle_focus_row)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_toggle_focus_row(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, toggle_row)
{
	zval *item;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &item, gtkwidget_ce))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_toggle_row(GTK_LIST(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(item)));

}


static PHP_METHOD(GtkList, undo_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_undo_selection(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

    gtk_list_unselect_all(GTK_LIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkList, unselect_child)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_list_unselect_child(GTK_LIST(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkList, unselect_item)
{
	long item;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &item))
		return;

    gtk_list_unselect_item(GTK_LIST(PHPG_GOBJECT(this_ptr)), (gint)item);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_child_position, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_clear_items, 0)
    ZEND_ARG_INFO(0, start)
    ZEND_ARG_INFO(0, end)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_extend_selection, 0)
    ZEND_ARG_INFO(0, scroll_type)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, auto_start_selection)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_scroll_horizontal, 0)
    ZEND_ARG_INFO(0, scroll_type)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_scroll_vertical, 0)
    ZEND_ARG_INFO(0, scroll_type)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_select_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_select_item, 0)
    ZEND_ARG_INFO(0, item)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_set_selection_mode, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_toggle_row, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_unselect_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklist_unselect_item, 0)
    ZEND_ARG_INFO(0, item)
ZEND_END_ARG_INFO();

static zend_function_entry gtklist_methods[] = {
	PHP_ME(GtkList, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, child_position,       arginfo_gtk_gtklist_child_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, clear_items,          arginfo_gtk_gtklist_clear_items, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, end_drag_selection,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, end_selection,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, extend_selection,     arginfo_gtk_gtklist_extend_selection, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, scroll_horizontal,    arginfo_gtk_gtklist_scroll_horizontal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, scroll_vertical,      arginfo_gtk_gtklist_scroll_vertical, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, select_all,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, select_child,         arginfo_gtk_gtklist_select_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, select_item,          arginfo_gtk_gtklist_select_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, set_selection_mode,   arginfo_gtk_gtklist_set_selection_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, start_selection,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, toggle_add_mode,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, toggle_focus_row,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, toggle_row,           arginfo_gtk_gtklist_toggle_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, undo_selection,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, unselect_all,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, unselect_child,       arginfo_gtk_gtklist_unselect_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkList, unselect_item,        arginfo_gtk_gtklist_unselect_item, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkMenu, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenu);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenu);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkMenu, attach)
{
	zval *child;
	long left_attach, right_attach, top_attach, bottom_attach;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oiiii", &child, gtkwidget_ce, &left_attach, &right_attach, &top_attach, &bottom_attach))
		return;

    gtk_menu_attach(GTK_MENU(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (guint)left_attach, (guint)right_attach, (guint)top_attach, (guint)bottom_attach);

}


static PHP_METHOD(GtkMenu, detach)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_detach(GTK_MENU(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenu, get_accel_group)
{
	GtkAccelGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_accel_group(GTK_MENU(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMenu, get_active)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_active(GTK_MENU(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMenu, get_attach_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_attach_widget(GTK_MENU(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMenu, get_tearoff_state)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_tearoff_state(GTK_MENU(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkMenu, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_title(GTK_MENU(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkMenu, popdown)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_popdown(GTK_MENU(PHPG_GOBJECT(this_ptr)));

}

#line 4364 "ext/gtk+/gtk.overrides"
static void php_gtk_menu_position(GtkMenu *menu, int *x, int *y, int *push_in, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    char *callback_name;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "unable to call menu position callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    if (cbd->user_args) {
        args = php_gtk_hash_as_array_offset(cbd->user_args, 0, &n_args);
    }

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    if (retval) {
        if (Z_TYPE_P(retval) == IS_ARRAY && !php_gtk_parse_args_hash_quiet(retval, "ii|b", x, y, push_in)) {
            php_error(E_WARNING, "unable to parse result of menu position callback");
        }
        zval_ptr_dtor(&retval);
    }

    if (args) {
        efree(args);
    }

    efree(callback_name);
    phpg_cb_data_destroy(data);
    phpg_handle_marshaller_exception(TSRMLS_C);
}

static PHP_METHOD(GtkMenu, popup)
{
    GtkWidget *pms = NULL, *pmi = NULL;
    zval *php_pms = NULL, *php_pmi = NULL;
    zval *callback = NULL, *user_args = NULL;
    int button = 0, time = 0;
    phpg_cb_data_t *data;

    NOT_STATIC_METHOD();

    if (ZEND_NUM_ARGS() > 5) {
        if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 5, &user_args, "NNVii", &php_pms, gtkwidget_ce, &php_pmi, gtkwidget_ce, &callback, &button, &time)) {
            return;
        }
    } else {
        if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|NNVii", &php_pms, gtkwidget_ce, &php_pmi, gtkwidget_ce, &callback, &button, &time)) {
            return;
        }
    }

    if (php_pms && Z_TYPE_P(php_pms) != IS_NULL)
        pms = GTK_WIDGET(PHPG_GOBJECT(php_pms));
    if (php_pmi && Z_TYPE_P(php_pmi) != IS_NULL)
        pmi = GTK_WIDGET(PHPG_GOBJECT(php_pmi));
    if (!time)
        time = gtk_get_current_event_time();

    if (callback && Z_TYPE_P(callback) != IS_NULL) {

        zval_add_ref(&callback);
        data = phpg_cb_data_new(callback, user_args TSRMLS_CC);

        gtk_menu_popup(GTK_MENU(PHPG_GOBJECT(this_ptr)), pms, pmi,
        (GtkMenuPositionFunc)php_gtk_menu_position, data, button, time);

    } else {

        gtk_menu_popup(GTK_MENU(PHPG_GOBJECT(this_ptr)), pms, pmi, NULL, NULL, button, time);
    }
}
#line 45364 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkMenu, reorder_child)
{
	zval *child;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &position))
		return;

    gtk_menu_reorder_child(GTK_MENU(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)position);

}


static PHP_METHOD(GtkMenu, reposition)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_reposition(GTK_MENU(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenu, set_accel_group)
{
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_group, gtkaccelgroup_ce))
		return;

    gtk_menu_set_accel_group(GTK_MENU(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));

}


static PHP_METHOD(GtkMenu, set_active)
{
	long index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    gtk_menu_set_active(GTK_MENU(PHPG_GOBJECT(this_ptr)), (guint)index);

}


static PHP_METHOD(GtkMenu, set_monitor)
{
	long monitor_num;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &monitor_num))
		return;

    gtk_menu_set_monitor(GTK_MENU(PHPG_GOBJECT(this_ptr)), (gint)monitor_num);

}


static PHP_METHOD(GtkMenu, set_screen)
{
	GdkScreen *screen = NULL;
	zval *php_screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_screen, gdkscreen_ce))
		return;
    if (Z_TYPE_P(php_screen) != IS_NULL)
        screen = GDK_SCREEN(PHPG_GOBJECT(php_screen));

    gtk_menu_set_screen(GTK_MENU(PHPG_GOBJECT(this_ptr)), screen);

}


static PHP_METHOD(GtkMenu, set_tearoff_state)
{
	zend_bool torn_off;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &torn_off))
		return;

    gtk_menu_set_tearoff_state(GTK_MENU(PHPG_GOBJECT(this_ptr)), (gboolean)torn_off);

}


static PHP_METHOD(GtkMenu, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_menu_set_title(GTK_MENU(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkMenu, get_accel_path)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_accel_path(GTK_MENU(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkMenu, get_monitor)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_monitor(GTK_MENU(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkMenu, get_reserve_toggle_size)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_get_reserve_toggle_size(GTK_MENU(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkMenu, set_reserve_toggle_size)
{
	zend_bool reserve_toggle_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reserve_toggle_size))
		return;

    gtk_menu_set_reserve_toggle_size(GTK_MENU(PHPG_GOBJECT(this_ptr)), (gboolean)reserve_toggle_size);

}

#line 4336 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMenu, set_menu_accel_path)
{
    char *accel_path;
    zend_bool free_accel_path = FALSE;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
        return;

    gtk_menu_set_accel_path(GTK_MENU(PHPG_GOBJECT(this_ptr)), accel_path);
    if (free_accel_path) g_free(accel_path);

}

#line 45570 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_attach, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, left_attach)
    ZEND_ARG_INFO(0, right_attach)
    ZEND_ARG_INFO(0, top_attach)
    ZEND_ARG_INFO(0, bottom_attach)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkmenu_popup, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, parent_menu_shell, GtkWidget, 1)
    ZEND_ARG_OBJ_INFO(0, parent_menu_item, GtkWidget, 1)
    ZEND_ARG_INFO(0, position_callback)
    ZEND_ARG_INFO(0, button)
    ZEND_ARG_INFO(0, time)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_reorder_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_accel_group, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_active, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_monitor, 0)
    ZEND_ARG_INFO(0, monitor_num)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_tearoff_state, 0)
    ZEND_ARG_INFO(0, torn_off)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_reserve_toggle_size, 0)
    ZEND_ARG_INFO(0, reserve_toggle_size)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenu_set_menu_accel_path, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();


static zend_function_entry gtkmenu_methods[] = {
	PHP_ME(GtkMenu, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, attach,               arginfo_gtk_gtkmenu_attach, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, detach,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_accel_group,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_accel_path,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_attach_widget,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_monitor,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_reserve_toggle_size, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_tearoff_state,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, popdown,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, popup,                arginfo_gtk_gtkmenu_popup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, reorder_child,        arginfo_gtk_gtkmenu_reorder_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, reposition,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_accel_group,      arginfo_gtk_gtkmenu_set_accel_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_active,           arginfo_gtk_gtkmenu_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_menu_accel_path,  arginfo_gtk_gtkmenu_set_menu_accel_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_monitor,          arginfo_gtk_gtkmenu_set_monitor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_reserve_toggle_size, arginfo_gtk_gtkmenu_set_reserve_toggle_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_screen,           arginfo_gtk_gtkmenu_set_screen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_tearoff_state,    arginfo_gtk_gtkmenu_set_tearoff_state, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenu, set_title,            arginfo_gtk_gtkmenu_set_title, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkMenuBar, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuBar);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuBar);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkMenuBar, get_pack_direction)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_bar_get_pack_direction(GTK_MENU_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkMenuBar, set_pack_direction)
{
	GtkPackDirection pack_dir;
	zval *php_pack_dir = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_pack_dir))
		return;

	if (php_pack_dir && phpg_gvalue_get_enum(GTK_TYPE_PACK_DIRECTION, php_pack_dir, (gint *)&pack_dir) == FAILURE) {
		return;
	}

    gtk_menu_bar_set_pack_direction(GTK_MENU_BAR(PHPG_GOBJECT(this_ptr)), pack_dir);

}


static PHP_METHOD(GtkMenuBar, get_child_pack_direction)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_bar_get_child_pack_direction(GTK_MENU_BAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkMenuBar, set_child_pack_direction)
{
	GtkPackDirection child_pack_dir;
	zval *php_child_pack_dir = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_child_pack_dir))
		return;

	if (php_child_pack_dir && phpg_gvalue_get_enum(GTK_TYPE_PACK_DIRECTION, php_child_pack_dir, (gint *)&child_pack_dir) == FAILURE) {
		return;
	}

    gtk_menu_bar_set_child_pack_direction(GTK_MENU_BAR(PHPG_GOBJECT(this_ptr)), child_pack_dir);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenubar_set_pack_direction, 0)
    ZEND_ARG_INFO(0, pack_dir)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenubar_set_child_pack_direction, 0)
    ZEND_ARG_INFO(0, child_pack_dir)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmenubar_methods[] = {
	PHP_ME(GtkMenuBar, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuBar, get_child_pack_direction, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuBar, get_pack_direction,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuBar, set_child_pack_direction, arginfo_gtk_gtkmenubar_set_child_pack_direction, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuBar, set_pack_direction,   arginfo_gtk_gtkmenubar_set_pack_direction, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentChooserMenu, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentChooserMenu);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentChooserMenu);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkRecentChooserMenu, get_show_numbers)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_recent_chooser_menu_get_show_numbers(GTK_RECENT_CHOOSER_MENU(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkRecentChooserMenu, set_show_numbers)
{
	zend_bool show_numbers;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_numbers))
		return;

    gtk_recent_chooser_menu_set_show_numbers(GTK_RECENT_CHOOSER_MENU(PHPG_GOBJECT(this_ptr)), (gboolean)show_numbers);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkrecentchoosermenu_set_show_numbers, 0)
    ZEND_ARG_INFO(0, show_numbers)
ZEND_END_ARG_INFO();

static zend_function_entry gtkrecentchoosermenu_methods[] = {
	PHP_ME(GtkRecentChooserMenu, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentChooserMenu, get_show_numbers,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRecentChooserMenu, set_show_numbers,     arginfo_gtk_gtkrecentchoosermenu_set_show_numbers, ZEND_ACC_PUBLIC)

	/***   GtkRecentChooser interface implementations   ***/

	PHP_MALIAS(GtkRecentChooser, set_show_private, set_show_private, arginfo_gtk_gtkrecentchooser_set_show_private, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_private, get_show_private, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_not_found, set_show_not_found, arginfo_gtk_gtkrecentchooser_set_show_not_found, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_not_found, get_show_not_found, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkrecentchooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_limit, set_limit, arginfo_gtk_gtkrecentchooser_set_limit, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_limit, get_limit, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_local_only, set_local_only, arginfo_gtk_gtkrecentchooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_tips, set_show_tips, arginfo_gtk_gtkrecentchooser_set_show_tips, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_tips, get_show_tips, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_icons, set_show_icons, arginfo_gtk_gtkrecentchooser_set_show_icons, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_icons, get_show_icons, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_type, set_sort_type, arginfo_gtk_gtkrecentchooser_set_sort_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_sort_type, get_sort_type, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_func, set_sort_func, arginfo_gtk_gtkrecentchooser_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_current_uri, set_current_uri, arginfo_gtk_gtkrecentchooser_set_current_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_uri, get_current_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_item, get_current_item, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_uri, select_uri, arginfo_gtk_gtkrecentchooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkrecentchooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_items, get_items, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, add_filter, add_filter, arginfo_gtk_gtkrecentchooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, remove_filter, remove_filter, arginfo_gtk_gtkrecentchooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_filter, set_filter, arginfo_gtk_gtkrecentchooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkLayout, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "hadjustment", "vadjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N^N", &php_args[0], gtkadjustment_ce, &php_args[1], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLayout);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLayout);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLayout);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkLayout, bin_window)
{
	GdkWindow* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_LAYOUT(((phpg_gobject_t *)object)->obj)->bin_window;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtklayout_prop_info[] = {
	{ "bin_window", PHPG_PROP_READ_FN(GtkLayout, bin_window), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkLayout, freeze)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_layout_freeze(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkLayout, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_layout_get_hadjustment(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 4269 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkLayout, get_size)
{
    guint width, height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_layout_get_size(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), &width, &height);
    php_gtk_build_value(&return_value, "(ii)", width, height);
}

#line 45968 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkLayout, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_layout_get_vadjustment(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkLayout, move)
{
	zval *child_widget;
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &child_widget, gtkwidget_ce, &x, &y))
		return;

    gtk_layout_move(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child_widget)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkLayout, put)
{
	zval *child_widget;
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &child_widget, gtkwidget_ce, &x, &y))
		return;

    gtk_layout_put(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child_widget)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkLayout, set_hadjustment)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_adjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_layout_set_hadjustment(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), adjustment);

}


static PHP_METHOD(GtkLayout, set_size)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_layout_set_size(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), (guint)width, (guint)height);

}


static PHP_METHOD(GtkLayout, set_vadjustment)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_adjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_layout_set_vadjustment(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)), adjustment);

}


static PHP_METHOD(GtkLayout, thaw)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_layout_thaw(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkLayout, get_bin_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_layout_get_bin_window(GTK_LAYOUT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtklayout_gtk_layout_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
    ZEND_ARG_OBJ_INFO(0, vadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklayout_move, 0)
    ZEND_ARG_OBJ_INFO(0, child_widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklayout_put, 0)
    ZEND_ARG_OBJ_INFO(0, child_widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklayout_set_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklayout_set_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklayout_set_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtklayout_methods[] = {
	PHP_ME(GtkLayout, __construct,          arginfo_gtk_gtklayout_gtk_layout_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, freeze,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, get_bin_window,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, move,                 arginfo_gtk_gtklayout_move, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, put,                  arginfo_gtk_gtklayout_put, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, set_hadjustment,      arginfo_gtk_gtklayout_set_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, set_size,             arginfo_gtk_gtklayout_set_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, set_vadjustment,      arginfo_gtk_gtklayout_set_vadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLayout, thaw,                 NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 3827 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconView, __construct)
{
    GObject *wrapped_obj;
    zval *php_model = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_model, gtktreemodel_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconView);
    }

    if (php_model && Z_TYPE_P(php_model) != IS_NULL) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "model",
                                              GTK_TREE_MODEL(PHPG_GOBJECT(php_model)), NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkIconView);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 46166 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, get_column_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_column_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_columns)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_columns(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_item_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_item_width(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_margin)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_margin(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_markup_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_markup_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_model(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconView, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_orientation(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_path_at_pos)
{
	long x, y;
	GtkTreePath *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_icon_view_get_path_at_pos(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
    }

}


static PHP_METHOD(GtkIconView, get_pixbuf_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_pixbuf_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_row_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_row_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 3852 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkIconView, get_selected_items)
{
    GList *selected, *tmp;

    NOT_STATIC_METHOD();
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    selected = gtk_icon_view_get_selected_items(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
    array_init(return_value);
    if (selected) {
        for (tmp = selected; tmp != NULL; tmp = tmp->next) {
            GtkTreePath *path = tmp->data;
            zval *item = NULL;

            phpg_tree_path_to_zval(path, &item TSRMLS_CC);
            add_next_index_zval(return_value, item);
            gtk_tree_path_free(path);
        }
        g_list_free(selected);
    }
}

#line 46338 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, get_selection_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_selection_mode(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_text_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_text_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, item_activated)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_view_item_activated(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkIconView, path_is_selected)
{
	GtkTreePath *path;
	zval *php_path;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_icon_view_path_is_selected(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkIconView, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_view_select_all(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIconView, select_path)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_view_select_path(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}

#line 3884 "ext/gtk+/gtk.overrides"
static void phpg_icon_view_foreach_func_marshal(GtkIconView *icon_view, GtkTreePath *path, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_icon_view = NULL, *php_path = NULL;

    TSRMLS_FETCH();

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_icon_view,       (GObject*)icon_view       TSRMLS_CC);
    phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_icon_view;
    args[1] = &php_path;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_icon_view);
    zval_ptr_dtor(&php_path);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkIconView, selected_foreach)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkIconViewForeachFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    zval_add_ref(&php_callback);
    cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
    callback = (GtkIconViewForeachFunc)phpg_icon_view_foreach_func_marshal;

    gtk_icon_view_selected_foreach(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), callback, cb_data);
    phpg_cb_data_destroy(cb_data);
}

#line 46525 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_column_spacing)
{
	long column_spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column_spacing))
		return;

    gtk_icon_view_set_column_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column_spacing);

}


static PHP_METHOD(GtkIconView, set_columns)
{
	long columns;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &columns))
		return;

    gtk_icon_view_set_columns(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)columns);

}


static PHP_METHOD(GtkIconView, set_item_width)
{
	long item_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &item_width))
		return;

    gtk_icon_view_set_item_width(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)item_width);

}


static PHP_METHOD(GtkIconView, set_margin)
{
	long margin;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &margin))
		return;

    gtk_icon_view_set_margin(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)margin);

}


static PHP_METHOD(GtkIconView, set_markup_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_icon_view_set_markup_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkIconView, set_model)
{
	GtkTreeModel *model = NULL;
	zval *php_model = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_model, gtktreemodel_ce))
		return;
    if (php_model) {
        if (Z_TYPE_P(php_model) == IS_NULL)
            model = NULL;
        else
            model = GTK_TREE_MODEL(PHPG_GOBJECT(php_model));
    }

    gtk_icon_view_set_model(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), model);

}


static PHP_METHOD(GtkIconView, set_orientation)
{
	GtkOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_icon_view_set_orientation(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkIconView, set_pixbuf_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_icon_view_set_pixbuf_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkIconView, set_row_spacing)
{
	long row_spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row_spacing))
		return;

    gtk_icon_view_set_row_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)row_spacing);

}


static PHP_METHOD(GtkIconView, set_selection_mode)
{
	GtkSelectionMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(GTK_TYPE_SELECTION_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_icon_view_set_selection_mode(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkIconView, set_spacing)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_icon_view_set_spacing(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)spacing);

}


static PHP_METHOD(GtkIconView, set_text_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_icon_view_set_text_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkIconView, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_view_unselect_all(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIconView, unselect_path)
{
	GtkTreePath *path;
	zval *php_path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_view_unselect_path(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkIconView, create_drag_icon)
{
	GtkTreePath *path;
	zval *php_path;
	GdkPixmap* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_icon_view_create_drag_icon(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkIconView, unset_model_drag_source)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_view_unset_model_drag_source(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkIconView, unset_model_drag_dest)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_icon_view_unset_model_drag_dest(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));

}

#line 88 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, get_cursor)
{
    GtkTreePath *path = NULL;
    GtkCellRenderer *cell = NULL;
    zval *php_path = NULL, *php_cell = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_icon_view_get_cursor(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), &path, &cell);

    array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        ALLOC_INIT_ZVAL(php_path);
    }

    if (cell) {
        phpg_gobject_new(&php_cell, (GObject*)cell TSRMLS_CC);
    } else {
        ALLOC_INIT_ZVAL(php_cell);
    }

    php_gtk_build_value(&return_value, "(NN)", php_path, php_cell);
}

#line 46831 "ext/gtk+/gen_gtk.c"


#line 129 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, get_dest_item_at_pos)
{
    gint x, y;
    GtkTreePath *path;
    GtkIconViewDropPosition pos;
    zval *php_path = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y)) {
        return;
    }

    if (!gtk_icon_view_get_dest_item_at_pos(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &path,
                                       &pos)) {
        return;
    }

    array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_path);
    }
    add_next_index_zval(return_value, php_path);

    add_next_index_long(return_value, pos);
}
#line 46864 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_drag_dest_item)
{
	GtkTreePath *path = NULL;
	zval *php_path = NULL, *php_pos = NULL;
	GtkIconViewDropPosition pos;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_path, &php_pos))
		return;

    if (php_path && Z_TYPE_P(php_path) != IS_NULL) {
        if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
            php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

	if (php_pos && phpg_gvalue_get_enum(GTK_TYPE_ICON_VIEW_DROP_POSITION, php_pos, (gint *)&pos) == FAILURE) {
		return;
	}

    gtk_icon_view_set_drag_dest_item(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path, pos);

    if (path)
        gtk_tree_path_free(path);

}

#line 233 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, get_drag_dest_item)
{
    GtkTreePath *path;
    GtkIconViewDropPosition pos;
	zval *php_path = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_icon_view_get_drag_dest_item(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), &path, &pos);

   array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_path);
    }
    add_next_index_zval(return_value, php_path);

    add_next_index_long(return_value, pos);
}
#line 46923 "ext/gtk+/gen_gtk.c"


#line 168 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, get_item_at_pos)
{
    gint x, y;
    GtkTreePath *path;
    GtkCellRenderer *renderer;
    zval *php_path = NULL, *php_renderer = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y)) {
        return;
    }

    if (!gtk_icon_view_get_item_at_pos(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), x, y, &path,
                                       &renderer)) {
        return;
    }

    array_init(return_value);

    if (path) {
        phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_path);
    }
    add_next_index_zval(return_value, php_path);

    phpg_gobject_new(&php_renderer, (GObject*)renderer TSRMLS_CC);
    add_next_index_zval(return_value, php_renderer);
}
#line 46957 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_reorderable)
{
	zend_bool reorderable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reorderable))
		return;

    gtk_icon_view_set_reorderable(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gboolean)reorderable);

}


static PHP_METHOD(GtkIconView, get_reorderable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_reorderable(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 28 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, enable_model_drag_source)
{
	zval *php_sbmask, *php_targets, *php_actions;
	GdkModifierType sbmask;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VaV", &php_sbmask, &php_targets, &php_actions)) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_sbmask, (gint *)&sbmask) == FAILURE) {
		return;
    }

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
        return;
    }

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_icon_view_enable_model_drag_source(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), sbmask, entries, n, actions);
	efree(entries);
}

#line 47016 "ext/gtk+/gen_gtk.c"


#line 65 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, enable_model_drag_dest)
{
	zval *php_targets, *php_actions;
	GdkDragAction actions;
	GtkTargetEntry *entries;
	int n;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "aV", &php_targets, &php_actions)) {
		return;
	}

	if (phpg_gvalue_get_flags(GDK_TYPE_DRAG_ACTION, php_actions, (gint *)&actions) == FAILURE) {
		return;
	}

	entries = phpg_parse_target_entries(php_targets, &n TSRMLS_CC);
	gtk_icon_view_enable_model_drag_dest(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), entries, n, actions);
	efree(entries);
}
#line 47041 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_cursor)
{
	GtkTreePath *path;
	zval *php_path, *php_cell = NULL;
	GtkCellRenderer *cell = NULL;
	zend_bool start_editing = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V|Nb", &php_path, &php_cell, gtkcellrenderer_ce, &start_editing))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (php_cell) {
        if (Z_TYPE_P(php_cell) == IS_NULL)
            cell = NULL;
        else
            cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));
    }

    gtk_icon_view_set_cursor(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path, cell, (gboolean)start_editing);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkIconView, scroll_to_path)
{
	GtkTreePath *path;
	zval *php_path;
	zend_bool use_align;
	double row_align, col_align;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vbdd", &php_path, &use_align, &row_align, &col_align))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_view_scroll_to_path(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path, (gboolean)use_align, (float)row_align, (float)col_align);

    if (path)
        gtk_tree_path_free(path);

}

#line 200 "ext/gtk+/gtk-2.8.overrides"
static PHP_METHOD(GtkIconView, get_visible_range)
{
    GtkTreePath *start_path, *end_path;
	zval *php_start = NULL, *php_end = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    if (!gtk_icon_view_get_visible_range(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), &start_path, &end_path))
	{
        return;
    }

    array_init(return_value);
	if (start_path) {
        phpg_tree_path_to_zval(start_path, &php_start TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_start);
    }
	add_next_index_zval(return_value, php_start);

	if (end_path) {
        phpg_tree_path_to_zval(end_path, &php_end TSRMLS_CC);
    } else {
        MAKE_STD_ZVAL(php_end);
    }
	add_next_index_zval(return_value, php_end);
}
#line 47132 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_tooltip_item)
{
	zval *tooltip, *php_path;
	GtkTreePath *path;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OV", &tooltip, gtktooltip_ce, &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_icon_view_set_tooltip_item(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIP(PHPG_GOBJECT(tooltip)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkIconView, set_tooltip_cell)
{
	zval *tooltip, *php_path, *php_cell;
	GtkTreePath *path;
	GtkCellRenderer *cell = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OVN", &tooltip, gtktooltip_ce, &php_path, &php_cell, gtkcellrenderer_ce))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (Z_TYPE_P(php_cell) != IS_NULL)
        cell = GTK_CELL_RENDERER(PHPG_GOBJECT(php_cell));

    gtk_icon_view_set_tooltip_cell(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIP(PHPG_GOBJECT(tooltip)), path, cell);

    if (path)
        gtk_tree_path_free(path);

}

#line 333 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkIconView, get_tooltip_context)
{
	zval *php_model = NULL, *php_iter = NULL, *php_path = NULL;
	gint x, y;
	gboolean keyboard_mode;
	GtkTreeModel *model = NULL;
	GtkTreePath *path = NULL;
	GtkTreeIter iter = {0, };

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"iii", &x, &y, &keyboard_mode)) {
		return;
	}

	if (!gtk_icon_view_get_tooltip_context (GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)),
		&x, &y, keyboard_mode, &model, &path, &iter)) {
		RETURN_FALSE;
	}

	phpg_gobject_new(&php_model, (GObject *) model TSRMLS_CC);
	phpg_gboxed_new(&php_iter, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
	phpg_tree_path_to_zval(path, &php_path TSRMLS_CC);
	php_gtk_build_value(&return_value, "(NNN)", php_model, php_path, php_iter);
}
#line 47210 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkIconView, set_tooltip_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_icon_view_set_tooltip_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkIconView, get_tooltip_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_tooltip_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, set_item_padding)
{
	long item_padding;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &item_padding))
		return;

    gtk_icon_view_set_item_padding(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), (gint)item_padding);

}


static PHP_METHOD(GtkIconView, get_item_padding)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_item_padding(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_item_row)
{
	GtkTreePath *path;
	zval *php_path;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_icon_view_get_item_row(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, get_item_column)
{
	GtkTreePath *path;
	zval *php_path;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_icon_view_get_item_column(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkIconView, set_item_orientation)
{
	GtkOrientation orientation;
	zval *php_orientation = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_orientation))
		return;

	if (php_orientation && phpg_gvalue_get_enum(GTK_TYPE_ORIENTATION, php_orientation, (gint *)&orientation) == FAILURE) {
		return;
	}

    gtk_icon_view_set_item_orientation(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)), orientation);

}


static PHP_METHOD(GtkIconView, get_item_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_icon_view_get_item_orientation(GTK_ICON_VIEW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkiconview_gtk_icon_view_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, model, GtkTreeModel, 0)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_get_path_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_item_activated, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_path_is_selected, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_select_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_selected_foreach, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_column_spacing, 0)
    ZEND_ARG_INFO(0, column_spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_columns, 0)
    ZEND_ARG_INFO(0, columns)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_item_width, 0)
    ZEND_ARG_INFO(0, item_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_margin, 0)
    ZEND_ARG_INFO(0, margin)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_markup_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkiconview_set_model, 0, 0, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_pixbuf_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_row_spacing, 0)
    ZEND_ARG_INFO(0, row_spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_selection_mode, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_spacing, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_text_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_unselect_path, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_create_drag_icon, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_get_dest_item_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_drag_dest_item, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, pos, GtkIconViewDropPosition, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_get_item_at_pos, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_reorderable, 0)
    ZEND_ARG_INFO(0, reorderable)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_enable_model_drag_source, 0)
    ZEND_ARG_INFO(0, sbmask)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_enable_model_drag_dest, 0)
    ZEND_ARG_INFO(0, targets)
    ZEND_ARG_INFO(0, actions)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkiconview_set_cursor, 0, 0, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
    ZEND_ARG_INFO(0, start_editing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_scroll_to_path, 0)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_INFO(0, use_align)
    ZEND_ARG_INFO(0, row_align)
    ZEND_ARG_INFO(0, col_align)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_tooltip_item, 0)
    ZEND_ARG_OBJ_INFO(0, tooltip, GtkTooltip, 1)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_tooltip_cell, 0)
    ZEND_ARG_OBJ_INFO(0, tooltip, GtkTooltip, 1)
    ZEND_ARG_INFO(0, path)
    ZEND_ARG_OBJ_INFO(0, cell, GtkCellRenderer, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkiconview_get_tooltip_context, 0, 0, 3)
	ZEND_ARG_INFO(0, x)
	ZEND_ARG_INFO(0, y)
	ZEND_ARG_INFO(0, keyboard_mode)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_tooltip_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_item_padding, 0)
    ZEND_ARG_INFO(0, item_padding)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_get_item_row, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_get_item_column, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkiconview_set_item_orientation, 0)
    ZEND_ARG_INFO(0, orientation)
ZEND_END_ARG_INFO();

static zend_function_entry gtkiconview_methods[] = {
	PHP_ME(GtkIconView, __construct,          arginfo_gtk_gtkiconview_gtk_icon_view_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, create_drag_icon,     arginfo_gtk_gtkiconview_create_drag_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, enable_model_drag_dest, arginfo_gtk_gtkiconview_enable_model_drag_dest, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, enable_model_drag_source, arginfo_gtk_gtkiconview_enable_model_drag_source, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_column_spacing,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_columns,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_cursor,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_dest_item_at_pos, arginfo_gtk_gtkiconview_get_dest_item_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_drag_dest_item,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_at_pos,      arginfo_gtk_gtkiconview_get_item_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_column,      arginfo_gtk_gtkiconview_get_item_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_orientation, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_padding,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_row,         arginfo_gtk_gtkiconview_get_item_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_item_width,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_margin,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_markup_column,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_orientation,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_path_at_pos,      arginfo_gtk_gtkiconview_get_path_at_pos, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_pixbuf_column,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_reorderable,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_row_spacing,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_selected_items,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_selection_mode,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_spacing,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_text_column,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_tooltip_column,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_tooltip_context,  arginfo_gtk_gtkiconview_get_tooltip_context, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, get_visible_range,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, item_activated,       arginfo_gtk_gtkiconview_item_activated, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, path_is_selected,     arginfo_gtk_gtkiconview_path_is_selected, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, scroll_to_path,       arginfo_gtk_gtkiconview_scroll_to_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, select_all,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, select_path,          arginfo_gtk_gtkiconview_select_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, selected_foreach,     arginfo_gtk_gtkiconview_selected_foreach, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_column_spacing,   arginfo_gtk_gtkiconview_set_column_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_columns,          arginfo_gtk_gtkiconview_set_columns, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_cursor,           arginfo_gtk_gtkiconview_set_cursor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_drag_dest_item,   arginfo_gtk_gtkiconview_set_drag_dest_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_item_orientation, arginfo_gtk_gtkiconview_set_item_orientation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_item_padding,     arginfo_gtk_gtkiconview_set_item_padding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_item_width,       arginfo_gtk_gtkiconview_set_item_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_margin,           arginfo_gtk_gtkiconview_set_margin, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_markup_column,    arginfo_gtk_gtkiconview_set_markup_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_model,            arginfo_gtk_gtkiconview_set_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_orientation,      arginfo_gtk_gtkiconview_set_orientation, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_pixbuf_column,    arginfo_gtk_gtkiconview_set_pixbuf_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_reorderable,      arginfo_gtk_gtkiconview_set_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_row_spacing,      arginfo_gtk_gtkiconview_set_row_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_selection_mode,   arginfo_gtk_gtkiconview_set_selection_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_spacing,          arginfo_gtk_gtkiconview_set_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_text_column,      arginfo_gtk_gtkiconview_set_text_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_tooltip_cell,     arginfo_gtk_gtkiconview_set_tooltip_cell, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_tooltip_column,   arginfo_gtk_gtkiconview_set_tooltip_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, set_tooltip_item,     arginfo_gtk_gtkiconview_set_tooltip_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, unselect_all,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, unselect_path,        arginfo_gtk_gtkiconview_unselect_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, unset_model_drag_dest, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkIconView, unset_model_drag_source, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFixed, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFixed);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFixed);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFixed, get_has_window)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_fixed_get_has_window(GTK_FIXED(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFixed, move)
{
	zval *widget;
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &widget, gtkwidget_ce, &x, &y))
		return;

    gtk_fixed_move(GTK_FIXED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkFixed, put)
{
	zval *widget;
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oii", &widget, gtkwidget_ce, &x, &y))
		return;

    gtk_fixed_put(GTK_FIXED(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkFixed, set_has_window)
{
	zend_bool has_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_window))
		return;

    gtk_fixed_set_has_window(GTK_FIXED(PHPG_GOBJECT(this_ptr)), (gboolean)has_window);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfixed_move, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfixed_put, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfixed_set_has_window, 0)
    ZEND_ARG_INFO(0, has_window)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfixed_methods[] = {
	PHP_ME(GtkFixed, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFixed, get_has_window,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFixed, move,                 arginfo_gtk_gtkfixed_move, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFixed, put,                  arginfo_gtk_gtkfixed_put, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFixed, set_has_window,       arginfo_gtk_gtkfixed_set_has_window, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCellView, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkCellView, new_with_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}

	wrapped_obj = (GObject *) gtk_cell_view_new_with_markup(markup);
	if (free_markup) g_free(markup);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkCellView, new_with_pixbuf)
{
	zval *pixbuf;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &pixbuf, gdkpixbuf_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}

	wrapped_obj = (GObject *) gtk_cell_view_new_with_pixbuf(GDK_PIXBUF(PHPG_GOBJECT(pixbuf)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkCellView, new_with_text)
{
	char *text;
	zend_bool free_text = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}

	wrapped_obj = (GObject *) gtk_cell_view_new_with_text(text);
	if (free_text) g_free(text);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCellView);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}

#line 179 "ext/gtk+/gtk-2.18.overrides"
static PHP_METHOD(GtkCellView, get_cell_renderers)
{
    GList *list, *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    phpg_warn_deprecated("since GTK+ 2.18, please use GtkCellLayout::get_cells()" TSRMLS_CC);

    list = gtk_cell_view_get_cell_renderers(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (item = list; item; item = item->next) {
        zval *php_item = NULL;
        GtkWidget *renderer = item->data;
        phpg_gobject_new(&php_item, G_OBJECT(renderer) TSRMLS_CC);
        add_next_index_zval(return_value, php_item);
    }

    g_list_free(list);
}

#line 47831 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCellView, get_displayed_row)
{
	GtkTreePath *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_view_get_displayed_row(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)));

    if (php_retval) {
        phpg_tree_path_to_zval(php_retval, &return_value TSRMLS_CC);
    }

}

#line 236 "ext/gtk+/gtktreeview.overrides"
static PHP_METHOD(GtkCellView, get_size_of_row)
{
    GtkTreePath *path;
    zval *php_path;
    GtkRequisition requisition;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
        return;

    if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
        php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_cell_view_get_size_of_row(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)), path, &requisition);

    if (path)
        gtk_tree_path_free(path);

    phpg_gboxed_new(&return_value, GTK_TYPE_REQUISITION, &requisition, TRUE, TRUE TSRMLS_CC);
}

#line 47877 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCellView, set_background_color)
{
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_cell_view_set_background_color(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)), color);

}


static PHP_METHOD(GtkCellView, set_displayed_row)
{
	GtkTreePath *path = NULL;
	zval *php_path = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_path))
		return;

    if (php_path && Z_TYPE_P(php_path) != IS_NULL) {
        if (phpg_tree_path_from_zval(php_path, &path TSRMLS_CC) == FAILURE) {
            php_error(E_WARNING, "%s::%s() expects path to be a valid tree path specification", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_cell_view_set_displayed_row(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)), path);

    if (path)
        gtk_tree_path_free(path);

}


static PHP_METHOD(GtkCellView, set_model)
{
	GtkTreeModel *model = NULL;
	zval *php_model;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_model, gtktreemodel_ce))
		return;
    if (Z_TYPE_P(php_model) != IS_NULL)
        model = GTK_TREE_MODEL(PHPG_GOBJECT(php_model));

    gtk_cell_view_set_model(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)), model);

}


static PHP_METHOD(GtkCellView, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_cell_view_get_model(GTK_CELL_VIEW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_new_with_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_new_with_pixbuf, 0)
    ZEND_ARG_OBJ_INFO(0, pixbuf, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_new_with_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_get_size_of_row, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_set_background_color, 0)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_set_displayed_row, 0)
    ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcellview_set_model, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcellview_methods[] = {
	PHP_ME(GtkCellView, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, new_with_markup,      arginfo_gtk_gtkcellview_new_with_markup, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkCellView, new_with_pixbuf,      arginfo_gtk_gtkcellview_new_with_pixbuf, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkCellView, new_with_text,        arginfo_gtk_gtkcellview_new_with_text, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkCellView, get_cell_renderers,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, get_displayed_row,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, get_size_of_row,      arginfo_gtk_gtkcellview_get_size_of_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, set_background_color, arginfo_gtk_gtkcellview_set_background_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, set_displayed_row,    arginfo_gtk_gtkcellview_set_displayed_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCellView, set_model,            arginfo_gtk_gtkcellview_set_model, ZEND_ACC_PUBLIC)

	/***   GtkCellLayout interface implementations   ***/

	PHP_MALIAS(GtkCellLayout, add_attribute, add_attribute, arginfo_gtk_gtkcelllayout_add_attribute, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear, clear, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear_attributes, clear_attributes, arginfo_gtk_gtkcelllayout_clear_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_end, pack_end, arginfo_gtk_gtkcelllayout_pack_end, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_start, pack_start, arginfo_gtk_gtkcelllayout_pack_start, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, reorder, reorder, arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_attributes, set_attributes, arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_cell_data_func, set_cell_data_func, arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, get_cells, get_cells, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

PHPG_PROP_READER(GtkBin, child)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_BIN(((phpg_gobject_t *)object)->obj)->child;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkbin_prop_info[] = {
	{ "child", PHPG_PROP_READ_FN(GtkBin, child), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkBin, get_child)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_bin_get_child(GTK_BIN(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static zend_function_entry gtkbin_methods[] = {
	PHP_ME(GtkBin, get_child,            NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkBox, get_homogeneous)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_box_get_homogeneous(GTK_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkBox, get_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_box_get_spacing(GTK_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkBox, pack_end)
{
	zval *child;
	zend_bool expand = TRUE, fill = TRUE;
	long padding = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|bbi", &child, gtkwidget_ce, &expand, &fill, &padding))
		return;

    gtk_box_pack_end(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)expand, (gboolean)fill, (guint)padding);

}


static PHP_METHOD(GtkBox, pack_end_defaults)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.14, use GtkBox::pack_end() instead" TSRMLS_CC);

    gtk_box_pack_end_defaults(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkBox, pack_start)
{
	zval *child;
	zend_bool expand = TRUE, fill = TRUE;
	long padding = TRUE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|bbi", &child, gtkwidget_ce, &expand, &fill, &padding))
		return;

    gtk_box_pack_start(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)expand, (gboolean)fill, (guint)padding);

}


static PHP_METHOD(GtkBox, pack_start_defaults)
{
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

	phpg_warn_deprecated("since GTK+ 2.14, use GtkBox::pack_start() instead" TSRMLS_CC);

    gtk_box_pack_start_defaults(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));

}

#line 1346 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkBox, query_child_packing)
{
    zval *php_child;
    gboolean expand, fill;
    guint padding;
    GtkPackType pack_type;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_child, gtkwidget_ce))
        return;

    gtk_box_query_child_packing(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(php_child)), &expand, &fill, &padding, &pack_type);
    php_gtk_build_value(&return_value, "(bbii)", expand, fill, padding, pack_type);
}

#line 48169 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkBox, reorder_child)
{
	zval *child;
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &position))
		return;

    gtk_box_reorder_child(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)position);

}


static PHP_METHOD(GtkBox, set_child_packing)
{
	zval *child, *php_pack_type = NULL;
	zend_bool expand, fill;
	long padding;
	GtkPackType pack_type;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ObbiV", &child, gtkwidget_ce, &expand, &fill, &padding, &php_pack_type))
		return;

	if (php_pack_type && phpg_gvalue_get_enum(GTK_TYPE_PACK_TYPE, php_pack_type, (gint *)&pack_type) == FAILURE) {
		return;
	}

    gtk_box_set_child_packing(GTK_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)expand, (gboolean)fill, (guint)padding, pack_type);

}


static PHP_METHOD(GtkBox, set_homogeneous)
{
	zend_bool homogeneous;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &homogeneous))
		return;

    gtk_box_set_homogeneous(GTK_BOX(PHPG_GOBJECT(this_ptr)), (gboolean)homogeneous);

}


static PHP_METHOD(GtkBox, set_spacing)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_box_set_spacing(GTK_BOX(PHPG_GOBJECT(this_ptr)), (gint)spacing);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkbox_pack_end, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, expand)
    ZEND_ARG_INFO(0, fill)
    ZEND_ARG_INFO(0, padding)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_pack_end_defaults, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkbox_pack_start, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, expand)
    ZEND_ARG_INFO(0, fill)
    ZEND_ARG_INFO(0, padding)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_pack_start_defaults, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_query_child_packing, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_reorder_child, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_set_child_packing, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, expand)
    ZEND_ARG_INFO(0, fill)
    ZEND_ARG_INFO(0, padding)
    ZEND_ARG_INFO(0, pack_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_set_homogeneous, 0)
    ZEND_ARG_INFO(0, homogeneous)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbox_set_spacing, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtkbox_methods[] = {
	PHP_ME(GtkBox, get_homogeneous,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, get_spacing,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, pack_end,             arginfo_gtk_gtkbox_pack_end, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, pack_end_defaults,    arginfo_gtk_gtkbox_pack_end_defaults, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, pack_start,           arginfo_gtk_gtkbox_pack_start, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, pack_start_defaults,  arginfo_gtk_gtkbox_pack_start_defaults, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, query_child_packing,  arginfo_gtk_gtkbox_query_child_packing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, reorder_child,        arginfo_gtk_gtkbox_reorder_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, set_child_packing,    arginfo_gtk_gtkbox_set_child_packing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, set_homogeneous,      arginfo_gtk_gtkbox_set_homogeneous, ZEND_ACC_PUBLIC)
	PHP_ME(GtkBox, set_spacing,          arginfo_gtk_gtkbox_set_spacing, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkWindow, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "type",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^V", &php_args[0])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkWindow);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkWindow);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkWindow);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkWindow, title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->title;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, wmclass_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->wmclass_name;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, wmclass_class)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->wmclass_class;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, wm_role)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->wm_role;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, focus_widget)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->focus_widget;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, default_widget)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->default_widget;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, transient_parent)
{
	GtkWindow* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->transient_parent;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, frame)
{
	GdkWindow* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->frame;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, group)
{
	GtkWindowGroup* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->group;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, configure_request_count)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->configure_request_count;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, allow_shrink)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->allow_shrink;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, allow_grow)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->allow_grow;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, configure_notify_received)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->configure_notify_received;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, need_default_position)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->need_default_position;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, need_default_size)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->need_default_size;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, position)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->position;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, type)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->type;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, has_user_ref_count)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->has_user_ref_count;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, has_focus)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->has_focus;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, modal)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->modal;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, destroy_with_parent)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->destroy_with_parent;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, has_frame)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->has_frame;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, iconify_initially)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->iconify_initially;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, stick_initially)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->stick_initially;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, maximize_initially)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->maximize_initially;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, decorated)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->decorated;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, type_hint)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->type_hint;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, gravity)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->gravity;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, frame_left)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->frame_left;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, frame_top)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->frame_top;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, frame_right)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->frame_right;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, frame_bottom)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->frame_bottom;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, keys_changed_handler)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->keys_changed_handler;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkWindow, mnemonic_modifier)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_WINDOW(((phpg_gobject_t *)object)->obj)->mnemonic_modifier;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtkwindow_prop_info[] = {
	{ "title", PHPG_PROP_READ_FN(GtkWindow, title), NULL },
	{ "wmclass_name", PHPG_PROP_READ_FN(GtkWindow, wmclass_name), NULL },
	{ "wmclass_class", PHPG_PROP_READ_FN(GtkWindow, wmclass_class), NULL },
	{ "wm_role", PHPG_PROP_READ_FN(GtkWindow, wm_role), NULL },
	{ "focus_widget", PHPG_PROP_READ_FN(GtkWindow, focus_widget), NULL },
	{ "default_widget", PHPG_PROP_READ_FN(GtkWindow, default_widget), NULL },
	{ "transient_parent", PHPG_PROP_READ_FN(GtkWindow, transient_parent), NULL },
	{ "frame", PHPG_PROP_READ_FN(GtkWindow, frame), NULL },
	{ "group", PHPG_PROP_READ_FN(GtkWindow, group), NULL },
	{ "configure_request_count", PHPG_PROP_READ_FN(GtkWindow, configure_request_count), NULL },
	{ "allow_shrink", PHPG_PROP_READ_FN(GtkWindow, allow_shrink), NULL },
	{ "allow_grow", PHPG_PROP_READ_FN(GtkWindow, allow_grow), NULL },
	{ "configure_notify_received", PHPG_PROP_READ_FN(GtkWindow, configure_notify_received), NULL },
	{ "need_default_position", PHPG_PROP_READ_FN(GtkWindow, need_default_position), NULL },
	{ "need_default_size", PHPG_PROP_READ_FN(GtkWindow, need_default_size), NULL },
	{ "position", PHPG_PROP_READ_FN(GtkWindow, position), NULL },
	{ "type", PHPG_PROP_READ_FN(GtkWindow, type), NULL },
	{ "has_user_ref_count", PHPG_PROP_READ_FN(GtkWindow, has_user_ref_count), NULL },
	{ "has_focus", PHPG_PROP_READ_FN(GtkWindow, has_focus), NULL },
	{ "modal", PHPG_PROP_READ_FN(GtkWindow, modal), NULL },
	{ "destroy_with_parent", PHPG_PROP_READ_FN(GtkWindow, destroy_with_parent), NULL },
	{ "has_frame", PHPG_PROP_READ_FN(GtkWindow, has_frame), NULL },
	{ "iconify_initially", PHPG_PROP_READ_FN(GtkWindow, iconify_initially), NULL },
	{ "stick_initially", PHPG_PROP_READ_FN(GtkWindow, stick_initially), NULL },
	{ "maximize_initially", PHPG_PROP_READ_FN(GtkWindow, maximize_initially), NULL },
	{ "decorated", PHPG_PROP_READ_FN(GtkWindow, decorated), NULL },
	{ "type_hint", PHPG_PROP_READ_FN(GtkWindow, type_hint), NULL },
	{ "gravity", PHPG_PROP_READ_FN(GtkWindow, gravity), NULL },
	{ "frame_left", PHPG_PROP_READ_FN(GtkWindow, frame_left), NULL },
	{ "frame_top", PHPG_PROP_READ_FN(GtkWindow, frame_top), NULL },
	{ "frame_right", PHPG_PROP_READ_FN(GtkWindow, frame_right), NULL },
	{ "frame_bottom", PHPG_PROP_READ_FN(GtkWindow, frame_bottom), NULL },
	{ "keys_changed_handler", PHPG_PROP_READ_FN(GtkWindow, keys_changed_handler), NULL },
	{ "mnemonic_modifier", PHPG_PROP_READ_FN(GtkWindow, mnemonic_modifier), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkWindow, activate_default)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_activate_default(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, activate_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_activate_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, activate_key)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_window_activate_key(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (GdkEventKey *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, add_accel_group)
{
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_group, gtkaccelgroup_ce))
		return;

    gtk_window_add_accel_group(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));

}


static PHP_METHOD(GtkWindow, add_mnemonic)
{
	long keyval;
	zval *target;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &keyval, &target, gtkwidget_ce))
		return;

    gtk_window_add_mnemonic(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (guint)keyval, GTK_WIDGET(PHPG_GOBJECT(target)));

}


static PHP_METHOD(GtkWindow, begin_move_drag)
{
	long button, root_x, root_y, timestamp;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &button, &root_x, &root_y, &timestamp))
		return;

    gtk_window_begin_move_drag(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)button, (gint)root_x, (gint)root_y, (guint32)timestamp);

}


static PHP_METHOD(GtkWindow, begin_resize_drag)
{
	GdkWindowEdge edge;
	zval *php_edge = NULL;
	long button, root_x, root_y, timestamp;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Viiii", &php_edge, &button, &root_x, &root_y, &timestamp))
		return;

	if (php_edge && phpg_gvalue_get_enum(GDK_TYPE_WINDOW_EDGE, php_edge, (gint *)&edge) == FAILURE) {
		return;
	}

    gtk_window_begin_resize_drag(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), edge, (gint)button, (gint)root_x, (gint)root_y, (guint32)timestamp);

}


static PHP_METHOD(GtkWindow, deiconify)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_deiconify(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, fullscreen)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_fullscreen(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, get_accept_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_accept_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_decorated)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_decorated(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 6171 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_default_icon_list)
{
    GList *icons, *item;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    icons = gtk_window_get_default_icon_list();

    array_init(return_value);
    for (item = icons; item; item = item->next) {
        zval *php_icon = NULL;
        phpg_gobject_new(&php_icon, G_OBJECT(item->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_icon);
    }
    g_list_free(icons);
}

#line 49102 "ext/gtk+/gen_gtk.c"


#line 6191 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_default_size)
{
    gint width, height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_window_get_default_size(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), &width, &height);
    php_gtk_build_value(&return_value, "(ii)", width, height);

}

#line 49120 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, get_destroy_with_parent)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_destroy_with_parent(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_focus)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWindow, get_focus_on_map)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_focus_on_map(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 6207 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_frame_dimensions)
{
    gint left, top, right, bottom;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_window_get_frame_dimensions(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), &left, &top, &right, &bottom);
    php_gtk_build_value(&return_value, "(iiii)", left, top, right, bottom);

}

#line 49180 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, get_gravity)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_gravity(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWindow, get_has_frame)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_has_frame(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_icon)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_icon(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 6223 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_icon_list)
{
    GList *icons, *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    icons = gtk_window_get_icon_list(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (item = icons; item; item = item->next) {
        zval *php_icon = NULL;
        phpg_gobject_new(&php_icon, G_OBJECT(item->data) TSRMLS_CC);
        add_next_index_zval(return_value, php_icon);
    }
    g_list_free(icons);
}

#line 49246 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, get_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_icon_name(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkWindow, get_mnemonic_modifier)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_mnemonic_modifier(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWindow, get_modal)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_modal(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 6155 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_position)
{
    gint root_x, root_y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_window_get_position(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), &root_x, &root_y);
    php_gtk_build_value(&return_value, "(ii)", root_x, root_y);

}

#line 49320 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, get_resizable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_resizable(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_role)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_role(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkWindow, get_screen)
{
	GdkScreen* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_screen(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 6245 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, get_size)
{
    gint width, height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_window_get_size(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), &width, &height);
    php_gtk_build_value(&return_value, "(ii)", width, height);

}

#line 49394 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, get_skip_pager_hint)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_skip_pager_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_skip_taskbar_hint)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_skip_taskbar_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_title(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkWindow, get_transient_for)
{
	GtkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_transient_for(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWindow, get_type_hint)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_type_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWindow, has_toplevel_focus)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_has_toplevel_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, iconify)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_iconify(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, is_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_is_active(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 6261 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, list_toplevels)
{
    GList *list, *tmp;
    zval *php_item = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    list = gtk_window_list_toplevels();

    array_init(return_value);
    for (tmp = list; tmp != NULL; tmp = tmp->next) {
        MAKE_STD_ZVAL(php_item);
        phpg_gobject_new(&php_item, G_OBJECT(tmp->data) TSRMLS_CC);
        if (php_item == NULL) {
            g_list_free(list);
            RETURN_NULL();
        }
        add_next_index_zval(return_value, php_item);
    }
    g_list_free(list);
}

#line 49548 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, maximize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_maximize(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, move)
{
	long x, y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    gtk_window_move(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);

}


static PHP_METHOD(GtkWindow, parse_geometry)
{
	char *geometry;
	zend_bool free_geometry = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &geometry, &free_geometry))
		return;

    php_retval = gtk_window_parse_geometry(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), geometry);
	if (free_geometry) g_free(geometry);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, present)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_present(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, propagate_key_event)
{
	GdkEvent *event = NULL;
	zval *php_event;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_event, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_event, GDK_TYPE_EVENT, FALSE TSRMLS_CC)) {
        event = (GdkEvent *) PHPG_GBOXED(php_event);
    } else {
        php_error(E_WARNING, "%s::%s() expects event argument to be a valid GdkEvent object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_window_propagate_key_event(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (GdkEventKey *)event);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, remove_accel_group)
{
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_group, gtkaccelgroup_ce))
		return;

    gtk_window_remove_accel_group(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));

}


static PHP_METHOD(GtkWindow, remove_mnemonic)
{
	long keyval;
	zval *target;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &keyval, &target, gtkwidget_ce))
		return;

    gtk_window_remove_mnemonic(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (guint)keyval, GTK_WIDGET(PHPG_GOBJECT(target)));

}


static PHP_METHOD(GtkWindow, reshow_with_initial_size)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_reshow_with_initial_size(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, resize)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_window_resize(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


static PHP_METHOD(GtkWindow, set_accept_focus)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_accept_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_auto_startup_notification)
{
	zend_bool setting;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_auto_startup_notification((gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_decorated)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_decorated(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_default)
{
	GtkWidget *default_widget = NULL;
	zval *php_default_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_default_widget, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_default_widget) != IS_NULL)
        default_widget = GTK_WIDGET(PHPG_GOBJECT(php_default_widget));

    gtk_window_set_default(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), default_widget);

}


static PHP_METHOD(GtkWindow, set_default_icon)
{
	zval *icon;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &icon, gdkpixbuf_ce))
		return;

    gtk_window_set_default_icon(GDK_PIXBUF(PHPG_GOBJECT(icon)));

}


static PHP_METHOD(GtkWindow, set_default_icon_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GError *error = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_window_set_default_icon_from_file(filename, &error);
	if (free_filename) g_free(filename);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }

}

#line 6295 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, set_default_icon_list)
{
    zval *php_icons, **php_item;
    GList *icons = NULL;
    int n, size;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_icons)) {
        return;
    }

    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_icons)), n = 0;
        zend_hash_get_current_data(Z_ARRVAL_P(php_icons), (void **)&php_item) == SUCCESS;
        zend_hash_move_forward(Z_ARRVAL_P(php_icons)), n++) {

        if (php_gtk_check_class(*php_item, gdkpixbuf_ce)) {
            GdkPixbuf *pixbuf = GDK_PIXBUF(PHPG_GOBJECT(*php_item));
            //prepend is faster than append
            icons = g_list_prepend(icons, pixbuf);
        } else {
            php_error(E_WARNING, "%s::%s() requires the array elements to be objects of class GdkPixbuf", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
    }

    icons = g_list_reverse(icons);
    size = g_list_length(icons);
    gtk_window_set_default_icon_list(icons);

    g_list_free(icons);
}

#line 49806 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, set_default_icon_name)
{
	char *name;
	zend_bool free_name = FALSE;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_window_set_default_icon_name(name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkWindow, set_default_size)
{
	long width, height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &width, &height))
		return;

    gtk_window_set_default_size(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)width, (gint)height);

}


static PHP_METHOD(GtkWindow, set_destroy_with_parent)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_destroy_with_parent(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_focus)
{
	GtkWidget *focus = NULL;
	zval *php_focus;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_focus, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_focus) != IS_NULL)
        focus = GTK_WIDGET(PHPG_GOBJECT(php_focus));

    gtk_window_set_focus(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), focus);

}


static PHP_METHOD(GtkWindow, set_focus_on_map)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_focus_on_map(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_frame_dimensions)
{
	long left, top, right, bottom;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &left, &top, &right, &bottom))
		return;

    gtk_window_set_frame_dimensions(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)left, (gint)top, (gint)right, (gint)bottom);

}

#line 6345 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, set_geometry_hints)
{
    gint min_width;
    gint min_height;
    gint max_width;
    gint max_height;
    gint base_width;
    gint base_height;
    gint width_inc;
    gint height_inc;
    gdouble min_aspect;
    gdouble max_aspect;
    GdkGravity win_gravity;

    zval *php_gravity = NULL;
    zval *php_widget;
    GdkGeometry geometry = { 0 };
    GdkWindowHints geom_mask = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiiiiiiiiddV",
                    &php_widget, gtkwidget_ce,
                    &min_width, &min_height, &max_width, &max_height,
                    &base_width, &base_height, &width_inc, &height_inc,
                    &min_aspect, &max_aspect, &php_gravity)) {
        return;
    }

    if (php_gravity && phpg_gvalue_get_enum(GDK_TYPE_GRAVITY, php_gravity, (gint *)&win_gravity) == FAILURE) {
        return;
    }

    if (min_width >= 0 || min_height >= 0) {
        geometry.min_width = MAX(min_width, 0);
        geometry.min_height = MAX(min_height, 0);
        geom_mask |= GDK_HINT_MIN_SIZE;
    }
    if (max_width >= 0 || max_height >= 0) {
        geometry.max_width = MAX(max_width, 0);
        geometry.max_height = MAX(max_height, 0);
        geom_mask |= GDK_HINT_MAX_SIZE;
    }
    if (base_width >= 0 || base_height >= 0) {
        geometry.base_width = MAX(base_width, 0);
        geometry.base_height = MAX(base_height, 0);
        geom_mask |= GDK_HINT_BASE_SIZE;
    }
    if (width_inc >= 0 || height_inc >= 0) {
        geometry.width_inc = MAX(width_inc, 0);
        geometry.height_inc = MAX(height_inc, 0);
        geom_mask |= GDK_HINT_RESIZE_INC;
    }
    if (min_aspect >= 0.0 || max_aspect >= 0.0) {
        if (min_aspect <= 0.0 || max_aspect <= 0.0) {
            php_error(E_WARNING, "%s::%s() aspect ratios must be positive",
                    get_active_class_name(NULL TSRMLS_CC),
                    get_active_function_name(TSRMLS_C));
            return;
        }
        geometry.min_aspect = min_aspect;
        geometry.max_aspect = max_aspect;
        geom_mask |= GDK_HINT_ASPECT;
    }
    gtk_window_set_geometry_hints(
            GTK_WINDOW(PHPG_GOBJECT(this_ptr)),
            GTK_WIDGET(PHPG_GOBJECT(php_widget)),
            &geometry, geom_mask);

}

#line 49968 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, set_gravity)
{
	GdkGravity gravity;
	zval *php_gravity = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_gravity))
		return;

	if (php_gravity && phpg_gvalue_get_enum(GDK_TYPE_GRAVITY, php_gravity, (gint *)&gravity) == FAILURE) {
		return;
	}

    gtk_window_set_gravity(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), gravity);

}


static PHP_METHOD(GtkWindow, set_has_frame)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_has_frame(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_icon)
{
	GdkPixbuf *icon = NULL;
	zval *php_icon;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_icon, gdkpixbuf_ce))
		return;
    if (Z_TYPE_P(php_icon) != IS_NULL)
        icon = GDK_PIXBUF(PHPG_GOBJECT(php_icon));

    gtk_window_set_icon(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), icon);

}


static PHP_METHOD(GtkWindow, set_icon_from_file)
{
	char *filename;
	zend_bool free_filename = FALSE;
	GError *error = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_window_set_icon_from_file(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), filename, &error);
	if (free_filename) g_free(filename);
    if (phpg_handle_gerror(&error TSRMLS_CC)) {
        return;
    }

}

#line 6425 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, set_icon_list)
{
    zval *php_icons, **php_item;
    GList *icons = NULL;
    int n, size;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_icons)) {
        return;
    }

    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_icons)), n = 0;
        zend_hash_get_current_data(Z_ARRVAL_P(php_icons), (void **)&php_item) == SUCCESS;
        zend_hash_move_forward(Z_ARRVAL_P(php_icons)), n++) {

        if (php_gtk_check_class(*php_item, gdkpixbuf_ce)) {
            GdkPixbuf *pixbuf = GDK_PIXBUF(PHPG_GOBJECT(*php_item));
            icons = g_list_prepend(icons, pixbuf);
        } else {
            php_error(E_WARNING, "%s::%s() requires the array elements to be objects of class GdkPixbuf", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
    }

    icons = g_list_reverse(icons);
    size = g_list_length(icons);
    gtk_window_set_icon_list(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), icons);

    g_list_free(icons);
}


#line 50074 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkWindow, set_icon_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_window_set_icon_name(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkWindow, set_keep_above)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_keep_above(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_keep_below)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_keep_below(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_mnemonic_modifier)
{
	GdkModifierType modifier;
	zval *php_modifier = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_modifier))
		return;

	if (php_modifier && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_modifier, (gint *)&modifier) == FAILURE) {
		return;
	}

    gtk_window_set_mnemonic_modifier(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), modifier);

}


static PHP_METHOD(GtkWindow, set_modal)
{
	zend_bool modal;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &modal))
		return;

    gtk_window_set_modal(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)modal);

}


static PHP_METHOD(GtkWindow, set_policy)
{
	long allow_shrink, allow_grow, auto_shrink;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iii", &allow_shrink, &allow_grow, &auto_shrink))
		return;

	phpg_warn_deprecated("use GtkWindow::set_resizable()" TSRMLS_CC);

    gtk_window_set_policy(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gint)allow_shrink, (gint)allow_grow, (gint)auto_shrink);

}


static PHP_METHOD(GtkWindow, set_position)
{
	GtkWindowPosition position;
	zval *php_position = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_position))
		return;

	if (php_position && phpg_gvalue_get_enum(GTK_TYPE_WINDOW_POSITION, php_position, (gint *)&position) == FAILURE) {
		return;
	}

    gtk_window_set_position(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), position);

}


static PHP_METHOD(GtkWindow, set_resizable)
{
	zend_bool resizable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &resizable))
		return;

    gtk_window_set_resizable(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)resizable);

}


static PHP_METHOD(GtkWindow, set_role)
{
	char *role;
	zend_bool free_role = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &role, &free_role))
		return;

    gtk_window_set_role(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), role);
	if (free_role) g_free(role);

}


static PHP_METHOD(GtkWindow, set_screen)
{
	zval *screen;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &screen, gdkscreen_ce))
		return;

    gtk_window_set_screen(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), GDK_SCREEN(PHPG_GOBJECT(screen)));

}


static PHP_METHOD(GtkWindow, set_skip_pager_hint)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_skip_pager_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_skip_taskbar_hint)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_skip_taskbar_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_window_set_title(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkWindow, set_transient_for)
{
	GtkWindow *parent = NULL;
	zval *php_parent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_parent, gtkwindow_ce))
		return;
    if (Z_TYPE_P(php_parent) != IS_NULL)
        parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));

    gtk_window_set_transient_for(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), parent);

}


static PHP_METHOD(GtkWindow, set_type_hint)
{
	GdkWindowTypeHint hint;
	zval *php_hint = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_hint))
		return;

	if (php_hint && phpg_gvalue_get_enum(GDK_TYPE_WINDOW_TYPE_HINT, php_hint, (gint *)&hint) == FAILURE) {
		return;
	}

    gtk_window_set_type_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), hint);

}


static PHP_METHOD(GtkWindow, set_wmclass)
{
	char *wmclass_name, *wmclass_class;
	zend_bool free_wmclass_name = FALSE, free_wmclass_class = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uu", &wmclass_name, &free_wmclass_name, &wmclass_class, &free_wmclass_class))
		return;

    gtk_window_set_wmclass(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), wmclass_name, wmclass_class);
	if (free_wmclass_name) g_free(wmclass_name);
	if (free_wmclass_class) g_free(wmclass_class);

}


static PHP_METHOD(GtkWindow, stick)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_stick(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, unfullscreen)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_unfullscreen(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, unmaximize)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_unmaximize(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, unstick)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_window_unstick(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkWindow, set_urgency_hint)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_urgency_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, get_urgency_hint)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_urgency_hint(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, present_with_time)
{
	long timestamp;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &timestamp))
		return;

    gtk_window_present_with_time(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (guint32)timestamp);

}


static PHP_METHOD(GtkWindow, set_deletable)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_deletable(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, get_deletable)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_deletable(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_group)
{
	GtkWindowGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_group(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWindow, set_opacity)
{
	double opacity;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &opacity))
		return;

    gtk_window_set_opacity(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), opacity);

}


static PHP_METHOD(GtkWindow, get_opacity)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_opacity(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkWindow, set_startup_id)
{
	char *startup_id;
	zend_bool free_startup_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &startup_id, &free_startup_id))
		return;

    gtk_window_set_startup_id(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), startup_id);
	if (free_startup_id) g_free(startup_id);

}


static PHP_METHOD(GtkWindow, get_default_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_default_widget(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkWindow, set_mnemonics_visible)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_window_set_mnemonics_visible(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkWindow, get_mnemonics_visible)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_mnemonics_visible(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkWindow, get_window_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_get_window_type(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkWindow, has_group)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_window_has_group(GTK_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 6467 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkWindow, window_mnemonic_activate)
{
    long keyval;
    GdkModifierType modifier;
    zval *php_modifier = NULL;
    gboolean php_retval;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &keyval, &php_modifier))
        return;

    if (php_modifier && phpg_gvalue_get_flags(GDK_TYPE_MODIFIER_TYPE, php_modifier, (gint *)&modifier) == FAILURE) {
        return;
    }

    php_retval = gtk_window_mnemonic_activate(GTK_WINDOW(PHPG_GOBJECT(this_ptr)), (guint)keyval, modifier);
    RETVAL_BOOL(php_retval);
}

#line 50601 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkwindow_gtk_window_new, 0, 0, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_activate_key, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_add_accel_group, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_add_mnemonic, 0)
    ZEND_ARG_INFO(0, keyval)
    ZEND_ARG_OBJ_INFO(0, target, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_begin_move_drag, 0)
    ZEND_ARG_INFO(0, button)
    ZEND_ARG_INFO(0, root_x)
    ZEND_ARG_INFO(0, root_y)
    ZEND_ARG_INFO(0, timestamp)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_begin_resize_drag, 0)
    ZEND_ARG_INFO(0, edge)
    ZEND_ARG_INFO(0, button)
    ZEND_ARG_INFO(0, root_x)
    ZEND_ARG_INFO(0, root_y)
    ZEND_ARG_INFO(0, timestamp)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_move, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_parse_geometry, 0)
    ZEND_ARG_INFO(0, geometry)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_propagate_key_event, 0)
    ZEND_ARG_INFO(0, event)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_remove_accel_group, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_remove_mnemonic, 0)
    ZEND_ARG_INFO(0, keyval)
    ZEND_ARG_OBJ_INFO(0, target, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_resize, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_accept_focus, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_auto_startup_notification, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_decorated, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default, 0)
    ZEND_ARG_OBJ_INFO(0, default_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default_icon, 0)
    ZEND_ARG_OBJ_INFO(0, icon, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default_icon_from_file, 0)
    ZEND_ARG_INFO(0, filename)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default_icon_list, 0)
    ZEND_ARG_INFO(0, pixbufs)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default_icon_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_default_size, 0)
    ZEND_ARG_INFO(0, width)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_destroy_with_parent, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_focus, 0)
    ZEND_ARG_OBJ_INFO(0, focus, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_focus_on_map, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_frame_dimensions, 0)
    ZEND_ARG_INFO(0, left)
    ZEND_ARG_INFO(0, top)
    ZEND_ARG_INFO(0, right)
    ZEND_ARG_INFO(0, bottom)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_geometry_hints, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 0)
    ZEND_ARG_INFO(0, min_width)
    ZEND_ARG_INFO(0, min_height)
    ZEND_ARG_INFO(0, max_width)
    ZEND_ARG_INFO(0, max_height)
    ZEND_ARG_INFO(0, base_width)
    ZEND_ARG_INFO(0, base_height)
    ZEND_ARG_INFO(0, width_inc)
    ZEND_ARG_INFO(0, height_inc)
    ZEND_ARG_INFO(0, min_aspect)
    ZEND_ARG_INFO(0, max_aspect)
    ZEND_ARG_INFO(0, win_gravity)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_gravity, 0)
    ZEND_ARG_INFO(0, gravity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_has_frame, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_icon, 0)
    ZEND_ARG_OBJ_INFO(0, icon, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_icon_from_file, 0)
    ZEND_ARG_INFO(0, filename)
    ZEND_ARG_INFO(0, error)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_icon_list, 0)
    ZEND_ARG_INFO(0, pixbufs)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_icon_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_keep_above, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_keep_below, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_mnemonic_modifier, 0)
    ZEND_ARG_INFO(0, modifier)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_modal, 0)
    ZEND_ARG_INFO(0, modal)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_policy, 0)
    ZEND_ARG_INFO(0, allow_shrink)
    ZEND_ARG_INFO(0, allow_grow)
    ZEND_ARG_INFO(0, auto_shrink)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_resizable, 0)
    ZEND_ARG_INFO(0, resizable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_role, 0)
    ZEND_ARG_INFO(0, role)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_screen, 0)
    ZEND_ARG_OBJ_INFO(0, screen, GdkScreen, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_skip_pager_hint, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_skip_taskbar_hint, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_transient_for, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_type_hint, 0)
    ZEND_ARG_INFO(0, hint)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_wmclass, 0)
    ZEND_ARG_INFO(0, wmclass_name)
    ZEND_ARG_INFO(0, wmclass_class)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_urgency_hint, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_present_with_time, 0)
    ZEND_ARG_INFO(0, timestamp)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_deletable, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_opacity, 0)
    ZEND_ARG_INFO(0, opacity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_startup_id, 0)
    ZEND_ARG_INFO(0, startup_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_set_mnemonics_visible, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkwindow_window_mnemonic_activate, 0)
    ZEND_ARG_INFO(0, keyval)
    ZEND_ARG_INFO(0, modifier)
ZEND_END_ARG_INFO();


static zend_function_entry gtkwindow_methods[] = {
	PHP_ME(GtkWindow, __construct,          arginfo_gtk_gtkwindow_gtk_window_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, activate_default,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, activate_focus,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, activate_key,         arginfo_gtk_gtkwindow_activate_key, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, add_accel_group,      arginfo_gtk_gtkwindow_add_accel_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, add_mnemonic,         arginfo_gtk_gtkwindow_add_mnemonic, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, begin_move_drag,      arginfo_gtk_gtkwindow_begin_move_drag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, begin_resize_drag,    arginfo_gtk_gtkwindow_begin_resize_drag, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, deiconify,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, fullscreen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_accept_focus,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_decorated,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_default_icon_list, NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, get_default_size,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_default_widget,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_deletable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_destroy_with_parent, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_focus,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_focus_on_map,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_frame_dimensions, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_gravity,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_has_frame,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_icon,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_icon_list,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_mnemonic_modifier, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_mnemonics_visible, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_modal,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_opacity,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_position,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_resizable,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_role,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_screen,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_skip_pager_hint,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_skip_taskbar_hint, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_transient_for,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_type_hint,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_urgency_hint,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, get_window_type,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, has_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, has_toplevel_focus,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, iconify,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, is_active,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, list_toplevels,       NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, maximize,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, move,                 arginfo_gtk_gtkwindow_move, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, parse_geometry,       arginfo_gtk_gtkwindow_parse_geometry, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, present,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, present_with_time,    arginfo_gtk_gtkwindow_present_with_time, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, propagate_key_event,  arginfo_gtk_gtkwindow_propagate_key_event, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, remove_accel_group,   arginfo_gtk_gtkwindow_remove_accel_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, remove_mnemonic,      arginfo_gtk_gtkwindow_remove_mnemonic, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, reshow_with_initial_size, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, resize,               arginfo_gtk_gtkwindow_resize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_accept_focus,     arginfo_gtk_gtkwindow_set_accept_focus, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_auto_startup_notification, arginfo_gtk_gtkwindow_set_auto_startup_notification, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, set_decorated,        arginfo_gtk_gtkwindow_set_decorated, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_default,          arginfo_gtk_gtkwindow_set_default, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_default_icon,     arginfo_gtk_gtkwindow_set_default_icon, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, set_default_icon_from_file, arginfo_gtk_gtkwindow_set_default_icon_from_file, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, set_default_icon_list, arginfo_gtk_gtkwindow_set_default_icon_list, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, set_default_icon_name, arginfo_gtk_gtkwindow_set_default_icon_name, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkWindow, set_default_size,     arginfo_gtk_gtkwindow_set_default_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_deletable,        arginfo_gtk_gtkwindow_set_deletable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_destroy_with_parent, arginfo_gtk_gtkwindow_set_destroy_with_parent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_focus,            arginfo_gtk_gtkwindow_set_focus, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_focus_on_map,     arginfo_gtk_gtkwindow_set_focus_on_map, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_frame_dimensions, arginfo_gtk_gtkwindow_set_frame_dimensions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_geometry_hints,   arginfo_gtk_gtkwindow_set_geometry_hints, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_gravity,          arginfo_gtk_gtkwindow_set_gravity, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_has_frame,        arginfo_gtk_gtkwindow_set_has_frame, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_icon,             arginfo_gtk_gtkwindow_set_icon, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_icon_from_file,   arginfo_gtk_gtkwindow_set_icon_from_file, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_icon_list,        arginfo_gtk_gtkwindow_set_icon_list, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_icon_name,        arginfo_gtk_gtkwindow_set_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_keep_above,       arginfo_gtk_gtkwindow_set_keep_above, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_keep_below,       arginfo_gtk_gtkwindow_set_keep_below, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_mnemonic_modifier, arginfo_gtk_gtkwindow_set_mnemonic_modifier, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_mnemonics_visible, arginfo_gtk_gtkwindow_set_mnemonics_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_modal,            arginfo_gtk_gtkwindow_set_modal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_opacity,          arginfo_gtk_gtkwindow_set_opacity, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_policy,           arginfo_gtk_gtkwindow_set_policy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_position,         arginfo_gtk_gtkwindow_set_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_resizable,        arginfo_gtk_gtkwindow_set_resizable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_role,             arginfo_gtk_gtkwindow_set_role, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_screen,           arginfo_gtk_gtkwindow_set_screen, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_skip_pager_hint,  arginfo_gtk_gtkwindow_set_skip_pager_hint, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_skip_taskbar_hint, arginfo_gtk_gtkwindow_set_skip_taskbar_hint, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_startup_id,       arginfo_gtk_gtkwindow_set_startup_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_title,            arginfo_gtk_gtkwindow_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_transient_for,    arginfo_gtk_gtkwindow_set_transient_for, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_type_hint,        arginfo_gtk_gtkwindow_set_type_hint, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_urgency_hint,     arginfo_gtk_gtkwindow_set_urgency_hint, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, set_wmclass,          arginfo_gtk_gtkwindow_set_wmclass, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, stick,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, unfullscreen,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, unmaximize,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, unstick,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkWindow, window_mnemonic_activate, arginfo_gtk_gtkwindow_window_mnemonic_activate, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkViewport, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "hadjustment", "vadjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N^N", &php_args[0], gtkadjustment_ce, &php_args[1], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkViewport);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkViewport);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkViewport);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkViewport, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_viewport_get_hadjustment(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkViewport, get_shadow_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_viewport_get_shadow_type(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkViewport, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_viewport_get_vadjustment(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkViewport, set_hadjustment)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_adjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_viewport_set_hadjustment(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)), adjustment);

}


static PHP_METHOD(GtkViewport, set_shadow_type)
{
	GtkShadowType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_viewport_set_shadow_type(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkViewport, set_vadjustment)
{
	GtkAdjustment *adjustment = NULL;
	zval *php_adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_adjustment, gtkadjustment_ce))
		return;
    if (Z_TYPE_P(php_adjustment) != IS_NULL)
        adjustment = GTK_ADJUSTMENT(PHPG_GOBJECT(php_adjustment));

    gtk_viewport_set_vadjustment(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)), adjustment);

}


static PHP_METHOD(GtkViewport, get_bin_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_viewport_get_bin_window(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkViewport, get_view_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_viewport_get_view_window(GTK_VIEWPORT(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkviewport_gtk_viewport_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
    ZEND_ARG_OBJ_INFO(0, vadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkviewport_set_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkviewport_set_shadow_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkviewport_set_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkviewport_methods[] = {
	PHP_ME(GtkViewport, __construct,          arginfo_gtk_gtkviewport_gtk_viewport_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, get_bin_window,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, get_shadow_type,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, get_view_window,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, set_hadjustment,      arginfo_gtk_gtkviewport_set_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, set_shadow_type,      arginfo_gtk_gtkviewport_set_shadow_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkViewport, set_vadjustment,      arginfo_gtk_gtkviewport_set_vadjustment, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 2580 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkDialog, __construct)
{
    char *title = NULL;
    zend_bool free_title;
    GtkWindow *parent = NULL;
    zval *php_parent = NULL, *php_flags = NULL;
    zval *buttons = NULL;
    GtkDialogFlags flags = 0;
    GObject *wrapped_obj;
    zval **text, **response;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|uNVa", &title, &free_title, &php_parent, gtkwindow_ce, &php_flags, &buttons)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDialog);
    }
    if (php_parent) {
        if (Z_TYPE_P(php_parent) == IS_NULL)
            parent = NULL;
        else
            parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
    }

    if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_DIALOG_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDialog);
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDialog);
    }

    if (title) {
        gtk_window_set_title(GTK_WINDOW(wrapped_obj), title);
    }
    if (free_title) g_free(title);

    if (parent) {
        gtk_window_set_transient_for(GTK_WINDOW(wrapped_obj), parent);
    }
    if (flags & GTK_DIALOG_MODAL) {
        gtk_window_set_modal(GTK_WINDOW(wrapped_obj), TRUE);
    }
    if (flags & GTK_DIALOG_DESTROY_WITH_PARENT) {
        gtk_window_set_destroy_with_parent (GTK_WINDOW(wrapped_obj), TRUE);
    }
    if (flags & GTK_DIALOG_NO_SEPARATOR) {
        gtk_dialog_set_has_separator (GTK_DIALOG(wrapped_obj), FALSE);
    }

    if (buttons) {

        if (zend_hash_num_elements(Z_ARRVAL_P(buttons)) % 2) {
            php_error(E_WARNING, "%s::%s(): button list has to contain pairs of items",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            gtk_object_destroy(GTK_OBJECT(wrapped_obj));
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDialog);
        }

        zend_hash_internal_pointer_reset(Z_ARRVAL_P(buttons));
        while (zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&text) == SUCCESS) {
            zend_hash_move_forward(Z_ARRVAL_P(buttons));
            zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&response); /* safe */
            zend_hash_move_forward(Z_ARRVAL_P(buttons));

            if (Z_TYPE_PP(text) != IS_STRING || Z_TYPE_PP(response) != IS_LONG) {
                php_error(E_WARNING, "%s::%s(): each pair in button list has to be string/number",
                          get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                gtk_object_destroy(GTK_OBJECT(wrapped_obj));
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkDialog);
            }

            gtk_dialog_add_button(GTK_DIALOG(wrapped_obj), Z_STRVAL_PP(text), Z_LVAL_PP(response));
        }
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


#line 51285 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkDialog, vbox)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_DIALOG(((phpg_gobject_t *)object)->obj)->vbox;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkDialog, action_area)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_DIALOG(((phpg_gobject_t *)object)->obj)->action_area;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkdialog_prop_info[] = {
	{ "vbox", PHPG_PROP_READ_FN(GtkDialog, vbox), NULL },
	{ "action_area", PHPG_PROP_READ_FN(GtkDialog, action_area), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkDialog, add_action_widget)
{
	zval *child;
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &child, gtkwidget_ce, &response_id))
		return;

    gtk_dialog_add_action_widget(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gint)response_id);

}


static PHP_METHOD(GtkDialog, add_button)
{
	char *button_text;
	zend_bool free_button_text = FALSE;
	long response_id;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ui", &button_text, &free_button_text, &response_id))
		return;

    php_retval = gtk_dialog_add_button(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), button_text, (gint)response_id);
	if (free_button_text) g_free(button_text);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 2668 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkDialog, add_buttons)
{
    zval *buttons = NULL;
    zval **text, **response;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &buttons))
        return;


    if (zend_hash_num_elements(Z_ARRVAL_P(buttons)) % 2) {
        php_error(E_WARNING,
                    "%s::%s(): button list has to contain pairs of items",
                        get_active_class_name(NULL TSRMLS_CC),
                            get_active_function_name(TSRMLS_C));
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(buttons));
    while (zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&text) == SUCCESS) {
        zend_hash_move_forward(Z_ARRVAL_P(buttons));
        zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&response); /* safe */
        zend_hash_move_forward(Z_ARRVAL_P(buttons));

        if (Z_TYPE_PP(text) != IS_STRING || Z_TYPE_PP(response) != IS_LONG) {
            php_error(E_WARNING, "%s::%s(): each pair in button list has to be string/number",
                        get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }

        gtk_dialog_add_button(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), Z_STRVAL_PP(text), Z_LVAL_PP(response));
    }
}

#line 51392 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkDialog, get_has_separator)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_dialog_get_has_separator(GTK_DIALOG(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkDialog, response)
{
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &response_id))
		return;

    gtk_dialog_response(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), (gint)response_id);

}


static PHP_METHOD(GtkDialog, run)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_dialog_run(GTK_DIALOG(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkDialog, set_default_response)
{
	long response_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &response_id))
		return;

    gtk_dialog_set_default_response(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), (gint)response_id);

}


static PHP_METHOD(GtkDialog, set_has_separator)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_dialog_set_has_separator(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkDialog, set_response_sensitive)
{
	long response_id;
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &response_id, &setting))
		return;

    gtk_dialog_set_response_sensitive(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), (gint)response_id, (gboolean)setting);

}


static PHP_METHOD(GtkDialog, get_response_for_widget)
{
	zval *widget;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &widget, gtkwidget_ce))
		return;

    php_retval = gtk_dialog_get_response_for_widget(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(widget)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkDialog, get_action_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_dialog_get_action_area(GTK_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkDialog, get_content_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_dialog_get_content_area(GTK_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkDialog, get_widget_for_response)
{
	long response_id;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &response_id))
		return;

    php_retval = gtk_dialog_get_widget_for_response(GTK_DIALOG(PHPG_GOBJECT(this_ptr)), (gint)response_id);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkdialog_gtk_dialog_new_with_buttons, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_INFO(0, buttons)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_add_action_widget, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_add_button, 0)
    ZEND_ARG_INFO(0, button_text)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_add_buttons, 0)
    ZEND_ARG_INFO(0, buttons)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_response, 0)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_set_default_response, 0)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_set_has_separator, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_set_response_sensitive, 0)
    ZEND_ARG_INFO(0, response_id)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_get_response_for_widget, 0)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkdialog_get_widget_for_response, 0)
    ZEND_ARG_INFO(0, response_id)
ZEND_END_ARG_INFO();

static zend_function_entry gtkdialog_methods[] = {
	PHP_ME(GtkDialog, __construct,          arginfo_gtk_gtkdialog_gtk_dialog_new_with_buttons, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, add_action_widget,    arginfo_gtk_gtkdialog_add_action_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, add_button,           arginfo_gtk_gtkdialog_add_button, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, add_buttons,          arginfo_gtk_gtkdialog_add_buttons, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, get_action_area,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, get_content_area,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, get_has_separator,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, get_response_for_widget, arginfo_gtk_gtkdialog_get_response_for_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, get_widget_for_response, arginfo_gtk_gtkdialog_get_widget_for_response, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, response,             arginfo_gtk_gtkdialog_response, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, run,                  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, set_default_response, arginfo_gtk_gtkdialog_set_default_response, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, set_has_separator,    arginfo_gtk_gtkdialog_set_has_separator, ZEND_ACC_PUBLIC)
	PHP_ME(GtkDialog, set_response_sensitive, arginfo_gtk_gtkdialog_set_response_sensitive, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4676 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkPlug, __construct)
{
    gulong socket_id;
    zval *display = NULL;
    GObject *wrapped_obj = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i|O", &socket_id, &display))
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPlug);

    wrapped_obj = g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkPlug);
    }

    if (display) {
        gtk_plug_construct_for_display(GTK_PLUG(wrapped_obj),
                                       GDK_DISPLAY(PHPG_GOBJECT(display)), (GdkNativeWindow) socket_id);
    } else {
        gtk_plug_construct(GTK_PLUG(wrapped_obj), (GdkNativeWindow) socket_id);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 51638 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkPlug, get_id)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_plug_get_id(GTK_PLUG(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkPlug, get_embedded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_plug_get_embedded(GTK_PLUG(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkPlug, get_socket_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_plug_get_socket_window(GTK_PLUG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkplug_gtk_plug_new, 0, 0, 1)
    ZEND_ARG_INFO(0, socket_id)
    ZEND_ARG_OBJ_INFO(0, display, GdkDisplay, 1)
ZEND_END_ARG_INFO();


static zend_function_entry gtkplug_methods[] = {
	PHP_ME(GtkPlug, __construct,          arginfo_gtk_gtkplug_gtk_plug_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPlug, get_embedded,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPlug, get_id,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkPlug, get_socket_window,    NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkColorSelectionDialog, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "title",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorSelectionDialog);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorSelectionDialog);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorSelectionDialog);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkColorSelectionDialog, colorsel)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COLOR_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->colorsel;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkColorSelectionDialog, ok_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COLOR_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->ok_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkColorSelectionDialog, cancel_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COLOR_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->cancel_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkColorSelectionDialog, help_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COLOR_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->help_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkcolorselectiondialog_prop_info[] = {
	{ "colorsel", PHPG_PROP_READ_FN(GtkColorSelectionDialog, colorsel), NULL },
	{ "ok_button", PHPG_PROP_READ_FN(GtkColorSelectionDialog, ok_button), NULL },
	{ "cancel_button", PHPG_PROP_READ_FN(GtkColorSelectionDialog, cancel_button), NULL },
	{ "help_button", PHPG_PROP_READ_FN(GtkColorSelectionDialog, help_button), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkColorSelectionDialog, get_color_selection)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_dialog_get_color_selection(GTK_COLOR_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcolorselectiondialog_gtk_color_selection_dialog_new, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcolorselectiondialog_methods[] = {
	PHP_ME(GtkColorSelectionDialog, __construct,          arginfo_gtk_gtkcolorselectiondialog_gtk_color_selection_dialog_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelectionDialog, get_color_selection,  NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAboutDialog, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAboutDialog);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAboutDialog);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 942 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, get_artists)
{
    gchar **artists;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    artists = (gchar **)gtk_about_dialog_get_artists(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    if (artists) {
        while (*artists) {
            add_next_index_string(return_value, *artists, 1);
            *artists++;
        }
    }
}

#line 51868 "ext/gtk+/gen_gtk.c"


#line 964 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, get_authors)
{
    gchar **authors;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    authors = (gchar **)gtk_about_dialog_get_authors(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    if (authors) {
        while (*authors) {
            add_next_index_string(return_value, *authors, 1);
            *authors++;
        }
    }
}

#line 51892 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, get_comments)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_comments(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_copyright)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_copyright(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 986 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, get_documenters)
{
    gchar **doccers;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    doccers = (gchar **)gtk_about_dialog_get_documenters(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    if (doccers) {
        while (*doccers) {
            add_next_index_string(return_value, *doccers, 1);
            *doccers++;
        }
    }
}

#line 51972 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, get_license)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_license(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_logo)
{
	GdkPixbuf* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_logo(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkAboutDialog, get_logo_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_logo_icon_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkAboutDialog::get_program_name()" TSRMLS_CC);

    php_retval = gtk_about_dialog_get_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_translator_credits)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_translator_credits(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_version)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_version(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_website)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_website(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, get_website_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_website_label(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 1015 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, set_artists)
{
    gchar **artists;
    zval *arr = NULL;
    zval **data;
    int i, arr_count;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &arr)) {
        return;
    }

    arr_count = zend_hash_num_elements(Z_ARRVAL_P(arr));
    artists = safe_emalloc(arr_count+1, sizeof(gchar *), 0);

    for (i=0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(arr));
         zend_hash_get_current_data(Z_ARRVAL_P(arr), (void **)&data) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(arr))) {

        convert_to_string_ex(data);
        artists[i++] = Z_STRVAL_PP(data);

    }

    artists[i] = NULL;
    gtk_about_dialog_set_artists(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), (const gchar **)artists);
    efree(artists);
}

#line 52218 "ext/gtk+/gen_gtk.c"


#line 1054 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, set_authors)
{
    gchar **authors;
    zval *arr = NULL;
    zval **data;
    int i, arr_count;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &arr)) {
        return;
    }

    arr_count = zend_hash_num_elements(Z_ARRVAL_P(arr));
    authors = safe_emalloc(arr_count+1, sizeof(gchar *), 0);

    for (i=0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(arr));
         zend_hash_get_current_data(Z_ARRVAL_P(arr), (void **)&data) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(arr))) {

        convert_to_string_ex(data);
        authors[i++] = Z_STRVAL_PP(data);

    }

    authors[i] = NULL;
    gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), (const gchar **)authors);
    efree(authors);
}

#line 52252 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, set_comments)
{
	char *comments;
	zend_bool free_comments = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &comments, &free_comments))
		return;

    gtk_about_dialog_set_comments(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), comments);
	if (free_comments) g_free(comments);

}


static PHP_METHOD(GtkAboutDialog, set_copyright)
{
	char *copyright;
	zend_bool free_copyright = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &copyright, &free_copyright))
		return;

    gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), copyright);
	if (free_copyright) g_free(copyright);

}

#line 1093 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, set_documenters)
{
    gchar **doccers;
    zval *arr = NULL;
    zval **data;
    int i, arr_count;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &arr)) {
        return;
    }

    arr_count = zend_hash_num_elements(Z_ARRVAL_P(arr));
    doccers = safe_emalloc(arr_count+1, sizeof(gchar *), 0);

    for (i=0, zend_hash_internal_pointer_reset(Z_ARRVAL_P(arr));
         zend_hash_get_current_data(Z_ARRVAL_P(arr), (void **)&data) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(arr))) {

        convert_to_string_ex(data);
        doccers[i++] = Z_STRVAL_PP(data);

    }

    doccers[i] = NULL;
    gtk_about_dialog_set_documenters(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), (const gchar **)doccers);
    efree(doccers);
}

#line 52318 "ext/gtk+/gen_gtk.c"


#line 1132 "ext/gtk+/gtk.overrides"
static void phpg_about_dialog_activate_link_func_marshal(GtkAboutDialog *about, const gchar *link, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_about = NULL, *php_link = NULL;
    gchar *cp_link;
    gsize cp_len;
    zend_bool free_result;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_about,       (GObject*)about       TSRMLS_CC);
    cp_link = phpg_from_utf8(link, strlen(link), &cp_len, &free_result TSRMLS_CC);
    if (!cp_link) {
        php_error(E_WARNING, "Could not convert link from UTF-8");
        return;
    }
    MAKE_STD_ZVAL(php_link);
    ZVAL_STRINGL(php_link, (char*)cp_link, cp_len, 1);
    if (free_result) {
        g_free(cp_link);
    }


    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_about;
    args[1] = &php_link;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_about);
    zval_ptr_dtor(&php_link);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkAboutDialog, set_email_hook)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkAboutDialogActivateLinkFunc callback;

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (Z_TYPE_P(php_callback) == IS_NULL) {
        cb_data  = NULL;
        callback = NULL;
    } else {
        zval_add_ref(&php_callback);
        cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
        callback = (GtkAboutDialogActivateLinkFunc)phpg_about_dialog_activate_link_func_marshal;
    }

    gtk_about_dialog_set_email_hook(callback, cb_data, phpg_cb_data_destroy);
}

#line 52397 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, set_license)
{
	char *license;
	zend_bool free_license = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &license, &free_license))
		return;

    gtk_about_dialog_set_license(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), license);
	if (free_license) g_free(license);

}


static PHP_METHOD(GtkAboutDialog, set_logo)
{
	GdkPixbuf *logo = NULL;
	zval *php_logo;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_logo, gdkpixbuf_ce))
		return;
    if (Z_TYPE_P(php_logo) != IS_NULL)
        logo = GDK_PIXBUF(PHPG_GOBJECT(php_logo));

    gtk_about_dialog_set_logo(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), logo);

}


static PHP_METHOD(GtkAboutDialog, set_logo_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_about_dialog_set_logo_icon_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkAboutDialog, set_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkAboutDialog::set_program_name()" TSRMLS_CC);

    gtk_about_dialog_set_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}


static PHP_METHOD(GtkAboutDialog, set_translator_credits)
{
	char *translator_credits;
	zend_bool free_translator_credits = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &translator_credits, &free_translator_credits))
		return;

    gtk_about_dialog_set_translator_credits(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), translator_credits);
	if (free_translator_credits) g_free(translator_credits);

}

#line 1216 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, set_url_hook)
{
    zval *php_callback = NULL, *extra;
    phpg_cb_data_t *cb_data;
    GtkAboutDialogActivateLinkFunc callback;

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkAboutDialogActivateLinkFunc)phpg_about_dialog_activate_link_func_marshal;
        }
    }

    gtk_about_dialog_set_url_hook(callback, cb_data, phpg_cb_data_destroy);
}

#line 52507 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, set_version)
{
	char *version;
	zend_bool free_version = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &version, &free_version))
		return;

    gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), version);
	if (free_version) g_free(version);

}


static PHP_METHOD(GtkAboutDialog, set_website)
{
	char *website;
	zend_bool free_website = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &website, &free_website))
		return;

    gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), website);
	if (free_website) g_free(website);

}

#line 1248 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAboutDialog, set_website_label)
{
	const gchar *php_retval = NULL;
	char *website_label;
	zend_bool free_website_label = FALSE;

      NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &website_label, &free_website_label))
		return;

	php_retval = gtk_about_dialog_get_website(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
      
	gtk_about_dialog_set_website_label(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), website_label);
	
	/* Fix for bug #393335 - we can't check for a website hook so instead we try to set the website too */
	if(!php_retval || strlen(php_retval) < 1) {
		gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), website_label);
	}
	if (free_website_label) g_free(website_label);
}
#line 52564 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAboutDialog, get_wrap_license)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_wrap_license(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkAboutDialog, set_wrap_license)
{
	zend_bool wrap_license;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &wrap_license))
		return;

    gtk_about_dialog_set_wrap_license(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), (gboolean)wrap_license);

}


static PHP_METHOD(GtkAboutDialog, get_program_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_about_dialog_get_program_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkAboutDialog, set_program_name)
{
	char *name;
	zend_bool free_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &name, &free_name))
		return;

    gtk_about_dialog_set_program_name(GTK_ABOUT_DIALOG(PHPG_GOBJECT(this_ptr)), name);
	if (free_name) g_free(name);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_artists, 0)
    ZEND_ARG_INFO(0, artists)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_authors, 0)
    ZEND_ARG_INFO(0, authors)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_comments, 0)
    ZEND_ARG_INFO(0, comments)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_copyright, 0)
    ZEND_ARG_INFO(0, copyright)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_documenters, 0)
    ZEND_ARG_INFO(0, documenters)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_email_hook, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_license, 0)
    ZEND_ARG_INFO(0, license)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_logo, 0)
    ZEND_ARG_OBJ_INFO(0, logo, GdkPixbuf, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_logo_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_translator_credits, 0)
    ZEND_ARG_INFO(0, translator_credits)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_url_hook, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_version, 0)
    ZEND_ARG_INFO(0, version)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_website, 0)
    ZEND_ARG_INFO(0, website)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_website_label, 0)
    ZEND_ARG_INFO(0, website_label)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_wrap_license, 0)
    ZEND_ARG_INFO(0, wrap_license)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkaboutdialog_set_program_name, 0)
    ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaboutdialog_methods[] = {
	PHP_ME(GtkAboutDialog, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_artists,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_authors,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_comments,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_copyright,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_documenters,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_license,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_logo,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_logo_icon_name,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_name,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_program_name,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_translator_credits, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_version,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_website,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_website_label,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, get_wrap_license,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_artists,          arginfo_gtk_gtkaboutdialog_set_artists, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_authors,          arginfo_gtk_gtkaboutdialog_set_authors, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_comments,         arginfo_gtk_gtkaboutdialog_set_comments, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_copyright,        arginfo_gtk_gtkaboutdialog_set_copyright, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_documenters,      arginfo_gtk_gtkaboutdialog_set_documenters, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_email_hook,       arginfo_gtk_gtkaboutdialog_set_email_hook, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAboutDialog, set_license,          arginfo_gtk_gtkaboutdialog_set_license, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_logo,             arginfo_gtk_gtkaboutdialog_set_logo, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_logo_icon_name,   arginfo_gtk_gtkaboutdialog_set_logo_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_name,             arginfo_gtk_gtkaboutdialog_set_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_program_name,     arginfo_gtk_gtkaboutdialog_set_program_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_translator_credits, arginfo_gtk_gtkaboutdialog_set_translator_credits, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_url_hook,         arginfo_gtk_gtkaboutdialog_set_url_hook, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkAboutDialog, set_version,          arginfo_gtk_gtkaboutdialog_set_version, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_website,          arginfo_gtk_gtkaboutdialog_set_website, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_website_label,    arginfo_gtk_gtkaboutdialog_set_website_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAboutDialog, set_wrap_license,     arginfo_gtk_gtkaboutdialog_set_wrap_license, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 513 "ext/gtk+/gtk-2.10.overrides"
static PHP_METHOD(GtkRecentChooserDialog, __construct)
{
    char *title = NULL;
    zend_bool free_title;
    GtkWindow *parent = NULL;
    zval *php_parent = NULL;
    zval *buttons = NULL;
    GObject *wrapped_obj;
    zval **text, **response;

    if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"|uNa",&title,&free_title,&php_parent,gtkwindow_ce,&buttons)) {
        return;
    }
    
    if(php_parent) {
        if(Z_TYPE_P(php_parent) == IS_NULL) {
            parent = NULL;
        } else {
            parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
        }
    }

    if(buttons) {
        if(zend_hash_num_elements(Z_ARRVAL_P(buttons)) % 2) {
            // not being able to count is fatal.
            php_error_docref(NULL TSRMLS_CC,E_ERROR,"button list has to contain pairs of items");
        }
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    if(!wrapped_obj) { PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentChooserDialog); }    

    if(buttons) {
        zend_hash_internal_pointer_reset(Z_ARRVAL_P(buttons));
        while(zend_hash_get_current_data(Z_ARRVAL_P(buttons),(void **)&text) == SUCCESS) {
            zend_hash_move_forward(Z_ARRVAL_P(buttons));
            zend_hash_get_current_data(Z_ARRVAL_P(buttons),(void **)&response);
            zend_hash_move_forward(Z_ARRVAL_P(buttons));
            
            // invalid data types cause it to skip the current pair.
            if(Z_TYPE_PP(text) != IS_STRING || Z_TYPE_PP(response) != IS_LONG) {
                php_error_docref(NULL TSRMLS_CC,E_WARNING,"each pair in button list is to be string/integer");
            } else {
	            gtk_dialog_add_button(
	                GTK_DIALOG(wrapped_obj),
	                Z_STRVAL_PP(text),
	                Z_LVAL_PP(response)
	            );
            }
        }
    }
    
    if(title) { gtk_window_set_title(GTK_WINDOW(wrapped_obj), title); }
    if(free_title) { g_free(title); }
    if(parent) { gtk_window_set_transient_for(GTK_WINDOW(wrapped_obj), parent); }
    
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 52821 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkrecentchooserdialog_recent_chooser_dialog_new, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_INFO(0, buttons)
ZEND_END_ARG_INFO();


static zend_function_entry gtkrecentchooserdialog_methods[] = {
	PHP_ME(GtkRecentChooserDialog, __construct,          arginfo_gtk_gtkrecentchooserdialog_recent_chooser_dialog_new, ZEND_ACC_PUBLIC)

	/***   GtkRecentChooser interface implementations   ***/

	PHP_MALIAS(GtkRecentChooser, set_show_private, set_show_private, arginfo_gtk_gtkrecentchooser_set_show_private, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_private, get_show_private, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_not_found, set_show_not_found, arginfo_gtk_gtkrecentchooser_set_show_not_found, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_not_found, get_show_not_found, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkrecentchooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_limit, set_limit, arginfo_gtk_gtkrecentchooser_set_limit, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_limit, get_limit, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_local_only, set_local_only, arginfo_gtk_gtkrecentchooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_tips, set_show_tips, arginfo_gtk_gtkrecentchooser_set_show_tips, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_tips, get_show_tips, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_numbers, set_show_numbers, arginfo_gtk_gtkrecentchooser_set_show_numbers, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_numbers, get_show_numbers, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_icons, set_show_icons, arginfo_gtk_gtkrecentchooser_set_show_icons, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_icons, get_show_icons, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_type, set_sort_type, arginfo_gtk_gtkrecentchooser_set_sort_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_sort_type, get_sort_type, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_func, set_sort_func, arginfo_gtk_gtkrecentchooser_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_current_uri, set_current_uri, arginfo_gtk_gtkrecentchooser_set_current_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_uri, get_current_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_item, get_current_item, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_uri, select_uri, arginfo_gtk_gtkrecentchooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkrecentchooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_items, get_items, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, add_filter, add_filter, arginfo_gtk_gtkrecentchooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, remove_filter, remove_filter, arginfo_gtk_gtkrecentchooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_filter, set_filter, arginfo_gtk_gtkrecentchooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 3541 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileChooserDialog, __construct)
{
    char *title = NULL, *backend = NULL;
    zend_bool free_title, free_backend;
    zval *php_parent = NULL, *php_action = NULL;
    zval *buttons = NULL;
    GtkWindow *parent = NULL;
    GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_OPEN;
    GObject *wrapped_obj;
    int n = 0;
    zval **text, **response;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|uNVau", &title, &free_title, &php_parent,
                            gtkwindow_ce, &php_action, &buttons, &backend, &free_backend)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
    }

    if (php_parent) {
        if (Z_TYPE_P(php_parent) == IS_NULL)
            parent = NULL;
        else
            parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
    }

    if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
    }

    if (buttons) {
        n = zend_hash_num_elements(Z_ARRVAL_P(buttons));
    }
    if (n % 2) {
        php_error(E_WARNING, "%s::%s(): button list has to contain pairs of items",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
    }

    wrapped_obj = (GObject *) g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC),
                                           "title", title,
                                           "action", action,
                                           "file-system-backend", backend, NULL);

    if (free_title) g_free(title);
    if (free_backend) g_free(backend);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
    }

    if (parent) {
        gtk_window_set_transient_for(GTK_WINDOW(wrapped_obj), parent);
    }

    if (buttons) {

        if (zend_hash_num_elements(Z_ARRVAL_P(buttons)) % 2) {
            php_error(E_WARNING, "%s::%s(): button list has to contain pairs of items",
                      get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            gtk_object_destroy(GTK_OBJECT(wrapped_obj));
            PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
        }

        zend_hash_internal_pointer_reset(Z_ARRVAL_P(buttons));
        while (zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&text) == SUCCESS) {
            zend_hash_move_forward(Z_ARRVAL_P(buttons));
            zend_hash_get_current_data(Z_ARRVAL_P(buttons), (void **)&response); /* safe */
            zend_hash_move_forward(Z_ARRVAL_P(buttons));

            if (Z_TYPE_PP(text) != IS_STRING || Z_TYPE_PP(response) != IS_LONG) {
                php_error(E_WARNING, "%s::%s(): each pair in button list has to string/number",
                          get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                gtk_object_destroy(GTK_OBJECT(wrapped_obj));
                PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserDialog);
            }

            gtk_dialog_add_button(GTK_DIALOG(wrapped_obj), Z_STRVAL_PP(text), Z_LVAL_PP(response));
        }
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 52954 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfilechooserdialog_file_chooser_dialog_new_with_backend, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_INFO(0, action)
    ZEND_ARG_INFO(0, buttons)
    ZEND_ARG_INFO(0, backend)
ZEND_END_ARG_INFO();


static zend_function_entry gtkfilechooserdialog_methods[] = {
	PHP_ME(GtkFileChooserDialog, __construct,          arginfo_gtk_gtkfilechooserdialog_file_chooser_dialog_new_with_backend, ZEND_ACC_PUBLIC)

	/***   GtkFileChooser interface implementations   ***/

	PHP_MALIAS(GtkFileChooser, add_filter, add_filter, arginfo_gtk_gtkfilechooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder, add_shortcut_folder, arginfo_gtk_gtkfilechooser_add_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder_uri, add_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_add_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_action, get_action, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder, get_current_folder, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder_uri, get_current_folder_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_extra_widget, get_extra_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filename, get_filename, arginfo_gtk_gtkfilechooser_get_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filenames, get_filenames, arginfo_gtk_gtkfilechooser_get_filenames, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_filename, get_preview_filename, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_uri, get_preview_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget, get_preview_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget_active, get_preview_widget_active, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_show_hidden, get_show_hidden, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uri, get_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_use_preview_label, get_use_preview_label, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folder_uris, list_shortcut_folder_uris, arginfo_gtk_gtkfilechooser_list_shortcut_folder_uris, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folders, list_shortcut_folders, arginfo_gtk_gtkfilechooser_list_shortcut_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_filter, remove_filter, arginfo_gtk_gtkfilechooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder, remove_shortcut_folder, arginfo_gtk_gtkfilechooser_remove_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder_uri, remove_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_remove_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_filename, select_filename, arginfo_gtk_gtkfilechooser_select_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_uri, select_uri, arginfo_gtk_gtkfilechooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_action, set_action, arginfo_gtk_gtkfilechooser_set_action, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder, set_current_folder, arginfo_gtk_gtkfilechooser_set_current_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder_uri, set_current_folder_uri, arginfo_gtk_gtkfilechooser_set_current_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_name, set_current_name, arginfo_gtk_gtkfilechooser_set_current_name, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_extra_widget, set_extra_widget, arginfo_gtk_gtkfilechooser_set_extra_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filename, set_filename, arginfo_gtk_gtkfilechooser_set_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filter, set_filter, arginfo_gtk_gtkfilechooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_local_only, set_local_only, arginfo_gtk_gtkfilechooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget, set_preview_widget, arginfo_gtk_gtkfilechooser_set_preview_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget_active, set_preview_widget_active, arginfo_gtk_gtkfilechooser_set_preview_widget_active, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkfilechooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_show_hidden, set_show_hidden, arginfo_gtk_gtkfilechooser_set_show_hidden, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_uri, set_uri, arginfo_gtk_gtkfilechooser_set_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_use_preview_label, set_use_preview_label, arginfo_gtk_gtkfilechooser_set_use_preview_label, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_filename, unselect_filename, arginfo_gtk_gtkfilechooser_unselect_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkfilechooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_do_overwrite_confirmation, set_do_overwrite_confirmation, arginfo_gtk_gtkfilechooser_set_do_overwrite_confirmation, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_do_overwrite_confirmation, get_do_overwrite_confirmation, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_create_folders, set_create_folders, arginfo_gtk_gtkfilechooser_set_create_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_create_folders, get_create_folders, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFileSelection, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "title",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileSelection);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileSelection);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileSelection);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkFileSelection, dir_list)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->dir_list;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, file_list)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->file_list;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, selection_entry)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->selection_entry;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, selection_text)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->selection_text;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, main_vbox)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->main_vbox;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, ok_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->ok_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, cancel_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->cancel_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, help_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->help_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, history_pulldown)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->history_pulldown;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, history_menu)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->history_menu;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_dialog)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_dialog;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_entry)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_entry;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_file)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_file;
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_c_dir)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_c_dir;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_del_file)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_del_file;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, fileop_ren_file)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->fileop_ren_file;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, button_area)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->button_area;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFileSelection, action_area)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FILE_SELECTION(((phpg_gobject_t *)object)->obj)->action_area;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkfileselection_prop_info[] = {
	{ "dir_list", PHPG_PROP_READ_FN(GtkFileSelection, dir_list), NULL },
	{ "file_list", PHPG_PROP_READ_FN(GtkFileSelection, file_list), NULL },
	{ "selection_entry", PHPG_PROP_READ_FN(GtkFileSelection, selection_entry), NULL },
	{ "selection_text", PHPG_PROP_READ_FN(GtkFileSelection, selection_text), NULL },
	{ "main_vbox", PHPG_PROP_READ_FN(GtkFileSelection, main_vbox), NULL },
	{ "ok_button", PHPG_PROP_READ_FN(GtkFileSelection, ok_button), NULL },
	{ "cancel_button", PHPG_PROP_READ_FN(GtkFileSelection, cancel_button), NULL },
	{ "help_button", PHPG_PROP_READ_FN(GtkFileSelection, help_button), NULL },
	{ "history_pulldown", PHPG_PROP_READ_FN(GtkFileSelection, history_pulldown), NULL },
	{ "history_menu", PHPG_PROP_READ_FN(GtkFileSelection, history_menu), NULL },
	{ "fileop_dialog", PHPG_PROP_READ_FN(GtkFileSelection, fileop_dialog), NULL },
	{ "fileop_entry", PHPG_PROP_READ_FN(GtkFileSelection, fileop_entry), NULL },
	{ "fileop_file", PHPG_PROP_READ_FN(GtkFileSelection, fileop_file), NULL },
	{ "fileop_c_dir", PHPG_PROP_READ_FN(GtkFileSelection, fileop_c_dir), NULL },
	{ "fileop_del_file", PHPG_PROP_READ_FN(GtkFileSelection, fileop_del_file), NULL },
	{ "fileop_ren_file", PHPG_PROP_READ_FN(GtkFileSelection, fileop_ren_file), NULL },
	{ "button_area", PHPG_PROP_READ_FN(GtkFileSelection, button_area), NULL },
	{ "action_area", PHPG_PROP_READ_FN(GtkFileSelection, action_area), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkFileSelection, complete)
{
	char *pattern;
	zend_bool free_pattern = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &pattern, &free_pattern))
		return;

    gtk_file_selection_complete(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)), pattern);
	if (free_pattern) g_free(pattern);

}

#line 3218 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileSelection, get_filename)
{
    const gchar *php_retval;
    gchar *cp_ret;
    gchar *fn = NULL;
    gsize cp_len;
    zend_bool free_result;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    php_retval = gtk_file_selection_get_filename(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        if (convert) {
            fn = g_filename_to_utf8(php_retval, strlen(php_retval), NULL, NULL, NULL);
            cp_ret = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_result TSRMLS_CC);
            if (cp_ret) {
                RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
            } else {
                php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            }
            if (free_result)
                g_free(cp_ret);
        } else {
            RETVAL_STRINGL((char *)php_retval, strlen(php_retval), 1);
        }
    } else {
        RETVAL_NULL();
    }
}

#line 53400 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileSelection, get_select_multiple)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_selection_get_select_multiple(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 3261 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFileSelection, get_selections)
{
    gchar **selections;
    int i = 0;
    gchar *cp = NULL;
    gchar *fn = NULL;
    gsize cp_len = 0;
    zend_bool free_cp = FALSE;
    zend_bool convert = 1;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|b", &convert))
        return;

    selections = gtk_file_selection_get_selections(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)));

    if (selections) {
        array_init(return_value);

        for (i = 0; selections[i]; i++) {
            if (convert) {
                fn = g_filename_to_utf8(selections[i], strlen(selections[i]), NULL, NULL, NULL);
                cp = phpg_from_utf8(fn, strlen(fn), &cp_len, &free_cp TSRMLS_CC);
                if (cp) {
                    add_next_index_string(return_value, (char *)cp, 1);
                } else {
                    php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
                }
                if (free_cp)
                    g_free(cp);
            } else {
                add_next_index_string(return_value, selections[i], 1);
            }
        }

        g_strfreev(selections);
    }
}

#line 53458 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFileSelection, hide_fileop_buttons)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkFileSelection, set_filename)
{
	char *filename;
	zend_bool free_filename = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &filename, &free_filename))
		return;

    gtk_file_selection_set_filename(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)), filename);
	if (free_filename) g_free(filename);

}


static PHP_METHOD(GtkFileSelection, set_select_multiple)
{
	zend_bool select_multiple;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &select_multiple))
		return;

    gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)), (gboolean)select_multiple);

}


static PHP_METHOD(GtkFileSelection, show_fileop_buttons)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_file_selection_show_fileop_buttons(GTK_FILE_SELECTION(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfileselection_gtk_file_selection_new, 0, 0, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfileselection_complete, 0)
    ZEND_ARG_INFO(0, pattern)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfileselection_get_filename, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkfileselection_get_selections, 0, 0, 0)
    ZEND_ARG_INFO(0, convert)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfileselection_set_filename, 0)
    ZEND_ARG_INFO(0, filename)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfileselection_set_select_multiple, 0)
    ZEND_ARG_INFO(0, select_multiple)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfileselection_methods[] = {
	PHP_ME(GtkFileSelection, __construct,          arginfo_gtk_gtkfileselection_gtk_file_selection_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, complete,             arginfo_gtk_gtkfileselection_complete, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, get_filename,         arginfo_gtk_gtkfileselection_get_filename, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, get_select_multiple,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, get_selections,       arginfo_gtk_gtkfileselection_get_selections, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, hide_fileop_buttons,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, set_filename,         arginfo_gtk_gtkfileselection_set_filename, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, set_select_multiple,  arginfo_gtk_gtkfileselection_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileSelection, show_fileop_buttons,  NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFontSelectionDialog, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "title",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontSelectionDialog);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontSelectionDialog);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontSelectionDialog);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkFontSelectionDialog, fontsel)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->fontsel;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFontSelectionDialog, main_vbox)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->main_vbox;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFontSelectionDialog, action_area)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->action_area;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFontSelectionDialog, ok_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->ok_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFontSelectionDialog, apply_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->apply_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkFontSelectionDialog, cancel_button)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_FONT_SELECTION_DIALOG(((phpg_gobject_t *)object)->obj)->cancel_button;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkfontselectiondialog_prop_info[] = {
	{ "fontsel", PHPG_PROP_READ_FN(GtkFontSelectionDialog, fontsel), NULL },
	{ "main_vbox", PHPG_PROP_READ_FN(GtkFontSelectionDialog, main_vbox), NULL },
	{ "action_area", PHPG_PROP_READ_FN(GtkFontSelectionDialog, action_area), NULL },
	{ "ok_button", PHPG_PROP_READ_FN(GtkFontSelectionDialog, ok_button), NULL },
	{ "apply_button", PHPG_PROP_READ_FN(GtkFontSelectionDialog, apply_button), NULL },
	{ "cancel_button", PHPG_PROP_READ_FN(GtkFontSelectionDialog, cancel_button), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkFontSelectionDialog, get_font)
{
	GdkFont *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkFontSelectionDialog::get_font_name()" TSRMLS_CC);

    php_retval = gtk_font_selection_dialog_get_font(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GDK_TYPE_FONT, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkFontSelectionDialog, get_font_name)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFontSelectionDialog, get_preview_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_dialog_get_preview_text(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFontSelectionDialog, set_font_name)
{
	char *fontname;
	zend_bool free_fontname = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &fontname, &free_fontname))
		return;

    php_retval = gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)), fontname);
	if (free_fontname) g_free(fontname);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontSelectionDialog, set_preview_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_font_selection_dialog_set_preview_text(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkFontSelectionDialog, get_ok_button)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_dialog_get_ok_button(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelectionDialog, get_apply_button)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.16, Don't use this function" TSRMLS_CC);

    php_retval = gtk_font_selection_dialog_get_apply_button(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelectionDialog, get_cancel_button)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_dialog_get_cancel_button(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelectionDialog, get_font_selection)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_dialog_get_font_selection(GTK_FONT_SELECTION_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontselectiondialog_gtk_font_selection_dialog_new, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontselectiondialog_set_font_name, 0)
    ZEND_ARG_INFO(0, fontname)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontselectiondialog_set_preview_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfontselectiondialog_methods[] = {
	PHP_ME(GtkFontSelectionDialog, __construct,          arginfo_gtk_gtkfontselectiondialog_gtk_font_selection_dialog_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_apply_button,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_cancel_button,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_font,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_font_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_font_selection,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_ok_button,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, get_preview_text,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, set_font_name,        arginfo_gtk_gtkfontselectiondialog_set_font_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelectionDialog, set_preview_text,     arginfo_gtk_gtkfontselectiondialog_set_preview_text, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkInputDialog, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInputDialog);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkInputDialog);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkinputdialog_methods[] = {
	PHP_ME(GtkInputDialog, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4499 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMessageDialog, __construct)
{
    GtkWindow *parent = NULL;
    zval *php_parent = NULL, *php_flags = NULL, *php_type = NULL, *php_buttons = NULL;
    GtkDialogFlags flags = 0;
    GtkMessageType type = GTK_MESSAGE_INFO;
    GtkButtonsType buttons = GTK_BUTTONS_NONE;
    char *message = NULL;
    zend_bool free_message;
    GObject *wrapped_obj;

    if (phpg_gtype_from_zval(this_ptr TSRMLS_CC) != GTK_TYPE_MESSAGE_DIALOG) {
        PHPG_THROW_EXCEPTION(phpg_construct_exception, "Subclassing GtkMessageDialog requires more than you know");
        return;
    }

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|NVVVu", &php_parent, gtkwindow_ce, &php_flags, &php_type, &php_buttons, &message, &free_message)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMessageDialog);
    }

    if (php_parent) {
        if (Z_TYPE_P(php_parent) == IS_NULL)
            parent = NULL;
        else
            parent = GTK_WINDOW(PHPG_GOBJECT(php_parent));
    }

    if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_DIALOG_FLAGS, php_flags, (gint *)&flags) == FAILURE) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMessageDialog);
    }

    if (php_type && phpg_gvalue_get_enum(GTK_TYPE_MESSAGE_TYPE, php_type, (gint *)&type) == FAILURE) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMessageDialog);
    }

    if (php_buttons && phpg_gvalue_get_enum(GTK_TYPE_BUTTONS_TYPE, php_buttons, (gint *)&buttons) == FAILURE) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMessageDialog);
    }

    if (!message) {
        message = "";
    }

    wrapped_obj = (GObject *) gtk_message_dialog_new(parent, flags, type, buttons, "%s", message);

    if (free_message) g_free(message);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMessageDialog);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 53966 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkMessageDialog, image)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_MESSAGE_DIALOG(((phpg_gobject_t *)object)->obj)->image;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkMessageDialog, label)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_MESSAGE_DIALOG(((phpg_gobject_t *)object)->obj)->label;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkmessagedialog_prop_info[] = {
	{ "image", PHPG_PROP_READ_FN(GtkMessageDialog, image), NULL },
	{ "label", PHPG_PROP_READ_FN(GtkMessageDialog, label), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkMessageDialog, set_markup)
{
	char *str;
	zend_bool free_str = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &str, &free_str))
		return;

    gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(PHPG_GOBJECT(this_ptr)), str);
	if (free_str) g_free(str);

}


static PHP_METHOD(GtkMessageDialog, set_image)
{
	zval *image;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &image, gtkwidget_ce))
		return;

    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(image)));

}


static PHP_METHOD(GtkMessageDialog, get_image)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_message_dialog_get_image(GTK_MESSAGE_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMessageDialog, get_message_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_message_dialog_get_message_area(GTK_MESSAGE_DIALOG(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkmessagedialog_gtk_message_dialog_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, parent, GtkWindow, 1)
    ZEND_ARG_INFO(0, flags)
    ZEND_ARG_INFO(0, type)
    ZEND_ARG_INFO(0, buttons)
    ZEND_ARG_INFO(0, message)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmessagedialog_set_markup, 0)
    ZEND_ARG_INFO(0, str)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmessagedialog_set_image, 0)
    ZEND_ARG_OBJ_INFO(0, image, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmessagedialog_methods[] = {
	PHP_ME(GtkMessageDialog, __construct,          arginfo_gtk_gtkmessagedialog_gtk_message_dialog_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMessageDialog, get_image,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMessageDialog, get_message_area,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMessageDialog, set_image,            arginfo_gtk_gtkmessagedialog_set_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMessageDialog, set_markup,           arginfo_gtk_gtkmessagedialog_set_markup, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolItem, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolItem);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolItem);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToolItem, get_expand)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_expand(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, get_homogeneous)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_homogeneous(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, get_icon_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_icon_size(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_is_important)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_is_important(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, get_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_orientation(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_proxy_menu_item)
{
	char *menu_item_id;
	zend_bool free_menu_item_id = FALSE;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &menu_item_id, &free_menu_item_id))
		return;

    php_retval = gtk_tool_item_get_proxy_menu_item(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), menu_item_id);
	if (free_menu_item_id) g_free(menu_item_id);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolItem, get_relief_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_relief_style(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_toolbar_style)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_toolbar_style(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_use_drag_window)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_use_drag_window(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, get_visible_horizontal)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_visible_horizontal(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, get_visible_vertical)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_visible_vertical(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolItem, rebuild_menu)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tool_item_rebuild_menu(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolItem, retrieve_proxy_menu_item)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_retrieve_proxy_menu_item(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolItem, set_expand)
{
	zend_bool expand;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &expand))
		return;

    gtk_tool_item_set_expand(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)expand);

}


static PHP_METHOD(GtkToolItem, set_homogeneous)
{
	zend_bool homogeneous;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &homogeneous))
		return;

    gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)homogeneous);

}


static PHP_METHOD(GtkToolItem, set_is_important)
{
	zend_bool is_important;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_important))
		return;

    gtk_tool_item_set_is_important(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)is_important);

}


static PHP_METHOD(GtkToolItem, set_proxy_menu_item)
{
	char *menu_item_id;
	zend_bool free_menu_item_id = FALSE;
	GtkWidget *menu_item = NULL;
	zval *php_menu_item;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uN", &menu_item_id, &free_menu_item_id, &php_menu_item, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_menu_item) != IS_NULL)
        menu_item = GTK_WIDGET(PHPG_GOBJECT(php_menu_item));

    gtk_tool_item_set_proxy_menu_item(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), menu_item_id, menu_item);
	if (free_menu_item_id) g_free(menu_item_id);

}


static PHP_METHOD(GtkToolItem, set_tooltip)
{
	zval *tooltips;
	char *tip_text = NULL, *tip_private = NULL;
	zend_bool free_tip_text = FALSE, free_tip_private = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O|uu", &tooltips, gtktooltips_ce, &tip_text, &free_tip_text, &tip_private, &free_tip_private))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkToolItem::set_tooltip_text()" TSRMLS_CC);

    gtk_tool_item_set_tooltip(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIPS(PHPG_GOBJECT(tooltips)), tip_text, tip_private);
	if (free_tip_text) g_free(tip_text);
	if (free_tip_private) g_free(tip_private);

}


static PHP_METHOD(GtkToolItem, set_use_drag_window)
{
	zend_bool use_drag_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_drag_window))
		return;

    gtk_tool_item_set_use_drag_window(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)use_drag_window);

}


static PHP_METHOD(GtkToolItem, set_visible_horizontal)
{
	zend_bool visible_horizontal;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible_horizontal))
		return;

    gtk_tool_item_set_visible_horizontal(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)visible_horizontal);

}


static PHP_METHOD(GtkToolItem, set_visible_vertical)
{
	zend_bool visible_vertical;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible_vertical))
		return;

    gtk_tool_item_set_visible_vertical(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)visible_vertical);

}


static PHP_METHOD(GtkToolItem, set_tooltip_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkToolItem, set_tooltip_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup))
		return;

    gtk_tool_item_set_tooltip_markup(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), markup);
	if (free_markup) g_free(markup);

}


static PHP_METHOD(GtkToolItem, toolbar_reconfigured)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_tool_item_toolbar_reconfigured(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkToolItem, get_ellipsize_mode)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_ellipsize_mode(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_text_alignment)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_text_alignment(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkToolItem, get_text_orientation)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_text_orientation(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkToolItem, get_text_size_group)
{
	GtkSizeGroup* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_item_get_text_size_group(GTK_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_get_proxy_menu_item, 0)
    ZEND_ARG_INFO(0, menu_item_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_expand, 0)
    ZEND_ARG_INFO(0, expand)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_homogeneous, 0)
    ZEND_ARG_INFO(0, homogeneous)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_is_important, 0)
    ZEND_ARG_INFO(0, is_important)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_proxy_menu_item, 0)
    ZEND_ARG_INFO(0, menu_item_id)
    ZEND_ARG_OBJ_INFO(0, menu_item, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktoolitem_set_tooltip, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, tooltips, GtkTooltips, 1)
    ZEND_ARG_INFO(0, tip_text)
    ZEND_ARG_INFO(0, tip_private)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_use_drag_window, 0)
    ZEND_ARG_INFO(0, use_drag_window)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_visible_horizontal, 0)
    ZEND_ARG_INFO(0, visible_horizontal)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_visible_vertical, 0)
    ZEND_ARG_INFO(0, visible_vertical)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_tooltip_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolitem_set_tooltip_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoolitem_methods[] = {
	PHP_ME(GtkToolItem, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_ellipsize_mode,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_expand,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_homogeneous,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_icon_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_is_important,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_orientation,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_proxy_menu_item,  arginfo_gtk_gtktoolitem_get_proxy_menu_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_relief_style,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_text_alignment,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_text_orientation, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_text_size_group,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_toolbar_style,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_use_drag_window,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_visible_horizontal, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, get_visible_vertical, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, rebuild_menu,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, retrieve_proxy_menu_item, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_expand,           arginfo_gtk_gtktoolitem_set_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_homogeneous,      arginfo_gtk_gtktoolitem_set_homogeneous, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_is_important,     arginfo_gtk_gtktoolitem_set_is_important, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_proxy_menu_item,  arginfo_gtk_gtktoolitem_set_proxy_menu_item, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_tooltip,          arginfo_gtk_gtktoolitem_set_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_tooltip_markup,   arginfo_gtk_gtktoolitem_set_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_tooltip_text,     arginfo_gtk_gtktoolitem_set_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_use_drag_window,  arginfo_gtk_gtktoolitem_set_use_drag_window, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_visible_horizontal, arginfo_gtk_gtktoolitem_set_visible_horizontal, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, set_visible_vertical, arginfo_gtk_gtktoolitem_set_visible_vertical, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolItem, toolbar_reconfigured, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVBox, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "homogeneous", "spacing",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^b^i", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVBox);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVBox);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVBox);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkvbox_gtk_vbox_new, 0, 0, 1)
    ZEND_ARG_INFO(0, homogeneous)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtkvbox_methods[] = {
	PHP_ME(GtkVBox, __construct,          arginfo_gtk_gtkvbox_gtk_vbox_new, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkScrolledWindow, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "hadjustment", "vadjustment",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^N^N", &php_args[0], gtkadjustment_ce, &php_args[1], gtkadjustment_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkScrolledWindow);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkScrolledWindow);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkScrolledWindow);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkScrolledWindow, add_with_viewport)
{
	zval *child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));

}


static PHP_METHOD(GtkScrolledWindow, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScrolledWindow, get_placement)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_placement(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 5211 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkScrolledWindow, get_policy)
{
    GtkPolicyType hscrollbar_policy, vscrollbar_policy;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_scrolled_window_get_policy(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), &hscrollbar_policy, &vscrollbar_policy);

    php_gtk_build_value(&return_value, "(ii)", (int)hscrollbar_policy, (int)vscrollbar_policy);
}

#line 54754 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkScrolledWindow, get_shadow_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_shadow_type(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkScrolledWindow, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScrolledWindow, set_hadjustment)
{
	zval *hadjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &hadjustment, gtkadjustment_ce))
		return;

    gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(hadjustment)));

}


static PHP_METHOD(GtkScrolledWindow, set_placement)
{
	GtkCornerType window_placement;
	zval *php_window_placement = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_window_placement))
		return;

	if (php_window_placement && phpg_gvalue_get_enum(GTK_TYPE_CORNER_TYPE, php_window_placement, (gint *)&window_placement) == FAILURE) {
		return;
	}

    gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), window_placement);

}


static PHP_METHOD(GtkScrolledWindow, set_policy)
{
	GtkPolicyType hscrollbar_policy, vscrollbar_policy;
	zval *php_hscrollbar_policy = NULL, *php_vscrollbar_policy = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "VV", &php_hscrollbar_policy, &php_vscrollbar_policy))
		return;

	if (php_hscrollbar_policy && phpg_gvalue_get_enum(GTK_TYPE_POLICY_TYPE, php_hscrollbar_policy, (gint *)&hscrollbar_policy) == FAILURE) {
		return;
	}

	if (php_vscrollbar_policy && phpg_gvalue_get_enum(GTK_TYPE_POLICY_TYPE, php_vscrollbar_policy, (gint *)&vscrollbar_policy) == FAILURE) {
		return;
	}

    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), hscrollbar_policy, vscrollbar_policy);

}


static PHP_METHOD(GtkScrolledWindow, set_shadow_type)
{
	GtkShadowType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkScrolledWindow, set_vadjustment)
{
	zval *hadjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &hadjustment, gtkadjustment_ce))
		return;

    gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(hadjustment)));

}


static PHP_METHOD(GtkScrolledWindow, get_hscrollbar)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_hscrollbar(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScrolledWindow, get_vscrollbar)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scrolled_window_get_vscrollbar(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScrolledWindow, unset_placement)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_scrolled_window_unset_placement(GTK_SCROLLED_WINDOW(PHPG_GOBJECT(this_ptr)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkscrolledwindow_gtk_scrolled_window_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
    ZEND_ARG_OBJ_INFO(0, vadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_add_with_viewport, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_set_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_set_placement, 0)
    ZEND_ARG_INFO(0, window_placement)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_set_policy, 0)
    ZEND_ARG_INFO(0, hscrollbar_policy)
    ZEND_ARG_INFO(0, vscrollbar_policy)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_set_shadow_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscrolledwindow_set_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, hadjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkscrolledwindow_methods[] = {
	PHP_ME(GtkScrolledWindow, __construct,          arginfo_gtk_gtkscrolledwindow_gtk_scrolled_window_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, add_with_viewport,    arginfo_gtk_gtkscrolledwindow_add_with_viewport, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_hscrollbar,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_placement,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_policy,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_shadow_type,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, get_vscrollbar,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, set_hadjustment,      arginfo_gtk_gtkscrolledwindow_set_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, set_placement,        arginfo_gtk_gtkscrolledwindow_set_placement, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, set_policy,           arginfo_gtk_gtkscrolledwindow_set_policy, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, set_shadow_type,      arginfo_gtk_gtkscrolledwindow_set_shadow_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, set_vadjustment,      arginfo_gtk_gtkscrolledwindow_set_vadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScrolledWindow, unset_placement,      NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkRecentChooserWidget, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentChooserWidget);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRecentChooserWidget);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkrecentchooserwidget_methods[] = {
	PHP_ME(GtkRecentChooserWidget, __construct,          NULL, ZEND_ACC_PUBLIC)

	/***   GtkRecentChooser interface implementations   ***/

	PHP_MALIAS(GtkRecentChooser, set_show_private, set_show_private, arginfo_gtk_gtkrecentchooser_set_show_private, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_private, get_show_private, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_not_found, set_show_not_found, arginfo_gtk_gtkrecentchooser_set_show_not_found, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_not_found, get_show_not_found, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkrecentchooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_limit, set_limit, arginfo_gtk_gtkrecentchooser_set_limit, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_limit, get_limit, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_local_only, set_local_only, arginfo_gtk_gtkrecentchooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_tips, set_show_tips, arginfo_gtk_gtkrecentchooser_set_show_tips, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_tips, get_show_tips, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_numbers, set_show_numbers, arginfo_gtk_gtkrecentchooser_set_show_numbers, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_numbers, get_show_numbers, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_show_icons, set_show_icons, arginfo_gtk_gtkrecentchooser_set_show_icons, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_show_icons, get_show_icons, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_type, set_sort_type, arginfo_gtk_gtkrecentchooser_set_sort_type, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_sort_type, get_sort_type, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_sort_func, set_sort_func, arginfo_gtk_gtkrecentchooser_set_sort_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_current_uri, set_current_uri, arginfo_gtk_gtkrecentchooser_set_current_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_uri, get_current_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_current_item, get_current_item, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_uri, select_uri, arginfo_gtk_gtkrecentchooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkrecentchooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_items, get_items, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, add_filter, add_filter, arginfo_gtk_gtkrecentchooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, remove_filter, remove_filter, arginfo_gtk_gtkrecentchooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, set_filter, set_filter, arginfo_gtk_gtkrecentchooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkRecentChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkColorSelection, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorSelection);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorSelection);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 1902 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorSelection, get_color)
{
    gdouble value[4];
    GtkColorSelection *colorsel;
    gboolean has_opacity;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    phpg_warn_deprecated("use get_current_color() instead" TSRMLS_CC);

    colorsel = GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr));
    gtk_color_selection_get_color(colorsel, value);

    g_object_get(colorsel, "has-opacity-control", &has_opacity, NULL);
    if (has_opacity) {
        php_gtk_build_value(&return_value, "(dddd)", value[0], value[1], value[2], value[3]);
    } else {
        php_gtk_build_value(&return_value, "(ddd)", value[0], value[1], value[2]);
    }
}

#line 55086 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkColorSelection, get_current_alpha)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_get_current_alpha(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 1929 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorSelection, get_current_color)
{
    GdkColor color = { 0 , 0 , 0 , 0 };

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), &color);
    phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &color, TRUE, TRUE TSRMLS_CC);
}

#line 55118 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkColorSelection, get_has_opacity_control)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_get_has_opacity_control(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkColorSelection, get_has_palette)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_get_has_palette(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkColorSelection, get_previous_alpha)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_get_previous_alpha(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 1945 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorSelection, get_previous_color)
{

    GdkColor color = { 0 , 0 , 0 , 0 };

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }


    gtk_color_selection_get_previous_color(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), &color);
    phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &color, TRUE, TRUE TSRMLS_CC);

}

#line 55181 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkColorSelection, is_adjusting)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_selection_is_adjusting(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 1999 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorSelection, palette_to_string)
{
    GdkColor *colors = NULL;
    zval *php_colors;
    int n_colors, i;
    zval **elem;
    gchar *ret, *cp_ret;
    gsize cp_len;
    zend_bool free_result;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_colors)) {
        return;
    }

    n_colors = zend_hash_num_elements(Z_ARRVAL_P(php_colors));
    colors = safe_emalloc(n_colors, sizeof(GdkColor), 0);

    for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_colors)), i = 0;
         zend_hash_get_current_data(Z_ARRVAL_P(php_colors), (void**)&elem) == SUCCESS;
         zend_hash_move_forward(Z_ARRVAL_P(php_colors)), i++) {
        if (!phpg_gboxed_check(*elem, GDK_TYPE_COLOR, TRUE TSRMLS_CC)) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING, "colors array should contain only GdkColor objects");
            efree(colors);
            return;
        }

        colors[i] = *(GdkColor *) PHPG_GBOXED(*elem);
    }

    ret = gtk_color_selection_palette_to_string(colors, (gint)n_colors);
    cp_ret = phpg_from_utf8(ret, strlen(ret), &cp_len, &free_result TSRMLS_CC);
    if (cp_ret) {
        RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
    } else {
        php_error_docref(NULL TSRMLS_CC, E_WARNING, "could not convert return value from UTF-8");
    }
    g_free(ret);
    if (free_result)
        g_free(cp_ret);
    efree(colors);
}

#line 55241 "ext/gtk+/gen_gtk.c"


#line 2050 "ext/gtk+/gtk.overrides"
static void phpg_color_selection_change_palette_with_screen_func_marshal(GdkScreen *screen, const GdkColor *colors, gint n_colors)
{

    phpg_cb_data_t *cbd;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_screen = NULL, *php_colors = NULL;

    TSRMLS_FETCH();

    cbd = GTK_G(color_selection_palette_callback);

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_screen, (GObject*)screen TSRMLS_CC);
    phpg_gboxed_new(&php_colors, GDK_TYPE_COLOR, (gpointer) colors, TRUE, TRUE TSRMLS_CC);

    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_screen;
    args[1] = &php_colors;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_screen);
    zval_ptr_dtor(&php_colors);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkColorSelection, set_change_palette_with_screen_hook)
{
    zval *php_callback, *extra;
    GtkColorSelectionChangePaletteWithScreenFunc callback;

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (Z_TYPE_P(php_callback) == IS_NULL) {
        /* set the global var to null */
        GTK_G(color_selection_palette_callback) = NULL;
        callback = NULL;
    } else {
        /* set the global var to our callback struct */
        zval_add_ref(&php_callback);
        GTK_G(color_selection_palette_callback) = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
        callback = (GtkColorSelectionChangePaletteWithScreenFunc)phpg_color_selection_change_palette_with_screen_func_marshal;
    }

    gtk_color_selection_set_change_palette_with_screen_hook(callback);
}

#line 55309 "ext/gtk+/gen_gtk.c"


#line 1974 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorSelection, set_color)
{
    gdouble value[4];

    NOT_STATIC_METHOD();

    phpg_warn_deprecated("use set_current_color() instead" TSRMLS_CC);

    value[3] = 1.0;
    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ddd|d", &value[0], &value[1], &value[2], &value[3])) {
        return;
    }

    gtk_color_selection_set_color(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), value);
}

#line 55329 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkColorSelection, set_current_alpha)
{
	long alpha;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &alpha))
		return;

    gtk_color_selection_set_current_alpha(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), (guint16)alpha);

}


static PHP_METHOD(GtkColorSelection, set_current_color)
{
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_color_selection_set_current_color(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), color);

}


static PHP_METHOD(GtkColorSelection, set_has_opacity_control)
{
	zend_bool has_opacity;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_opacity))
		return;

    gtk_color_selection_set_has_opacity_control(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), (gboolean)has_opacity);

}


static PHP_METHOD(GtkColorSelection, set_has_palette)
{
	zend_bool has_palette;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &has_palette))
		return;

    gtk_color_selection_set_has_palette(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), (gboolean)has_palette);

}


static PHP_METHOD(GtkColorSelection, set_previous_alpha)
{
	long alpha;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &alpha))
		return;

    gtk_color_selection_set_previous_alpha(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), (guint16)alpha);

}


static PHP_METHOD(GtkColorSelection, set_previous_color)
{
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_color_selection_set_previous_color(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), color);

}


static PHP_METHOD(GtkColorSelection, set_update_policy)
{
	GtkUpdateType policy;
	zval *php_policy = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_policy))
		return;

	if (php_policy && phpg_gvalue_get_enum(GTK_TYPE_UPDATE_TYPE, php_policy, (gint *)&policy) == FAILURE) {
		return;
	}

    gtk_color_selection_set_update_policy(GTK_COLOR_SELECTION(PHPG_GOBJECT(this_ptr)), policy);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcolorselection_palette_to_string, 0, 0, 1)
    ZEND_ARG_ARRAY_INFO(0, colors, 0)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_change_palette_with_screen_hook, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcolorselection_set_color, 0, 0, 3)
    ZEND_ARG_INFO(0, red)
    ZEND_ARG_INFO(0, green)
    ZEND_ARG_INFO(0, blue)
    ZEND_ARG_INFO(0, opacity)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_current_alpha, 0)
    ZEND_ARG_INFO(0, alpha)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_current_color, 0)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_has_opacity_control, 0)
    ZEND_ARG_INFO(0, has_opacity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_has_palette, 0)
    ZEND_ARG_INFO(0, has_palette)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_previous_alpha, 0)
    ZEND_ARG_INFO(0, alpha)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_previous_color, 0)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorselection_set_update_policy, 0)
    ZEND_ARG_INFO(0, policy)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcolorselection_methods[] = {
	PHP_ME(GtkColorSelection, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_color,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_current_alpha,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_current_color,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_has_opacity_control, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_has_palette,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_previous_alpha,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, get_previous_color,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, is_adjusting,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, palette_to_string,    arginfo_gtk_gtkcolorselection_palette_to_string, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkColorSelection, set_change_palette_with_screen_hook, arginfo_gtk_gtkcolorselection_set_change_palette_with_screen_hook, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkColorSelection, set_color,            arginfo_gtk_gtkcolorselection_set_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_current_alpha,    arginfo_gtk_gtkcolorselection_set_current_alpha, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_current_color,    arginfo_gtk_gtkcolorselection_set_current_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_has_opacity_control, arginfo_gtk_gtkcolorselection_set_has_opacity_control, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_has_palette,      arginfo_gtk_gtkcolorselection_set_has_palette, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_previous_alpha,   arginfo_gtk_gtkcolorselection_set_previous_alpha, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_previous_color,   arginfo_gtk_gtkcolorselection_set_previous_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorSelection, set_update_policy,    arginfo_gtk_gtkcolorselection_set_update_policy, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFileChooserWidget, __construct)
{
	GtkFileChooserAction action;
	zval *php_action = NULL;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_action)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}

	wrapped_obj = (GObject *) gtk_file_chooser_widget_new(action);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFileChooserWidget, new_with_backend)
{
	GtkFileChooserAction action;
	zval *php_action = NULL;
	char *backend;
	zend_bool free_backend = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Vu", &php_action, &backend, &free_backend)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}

	phpg_warn_deprecated("since GTK+ 2.14, Use GtkFileChooserWidget::__construct() instead" TSRMLS_CC);

	wrapped_obj = (GObject *) gtk_file_chooser_widget_new_with_backend(action, backend);
	if (free_backend) g_free(backend);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserWidget);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserwidget_gtk_file_chooser_widget_new, 0)
    ZEND_ARG_INFO(0, action)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserwidget_file_chooser_widget_new_with_backend, 0)
    ZEND_ARG_INFO(0, action)
    ZEND_ARG_INFO(0, backend)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfilechooserwidget_methods[] = {
	PHP_ME(GtkFileChooserWidget, __construct,          arginfo_gtk_gtkfilechooserwidget_gtk_file_chooser_widget_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserWidget, new_with_backend,     arginfo_gtk_gtkfilechooserwidget_file_chooser_widget_new_with_backend, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)

	/***   GtkFileChooser interface implementations   ***/

	PHP_MALIAS(GtkFileChooser, add_filter, add_filter, arginfo_gtk_gtkfilechooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder, add_shortcut_folder, arginfo_gtk_gtkfilechooser_add_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder_uri, add_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_add_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_action, get_action, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder, get_current_folder, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder_uri, get_current_folder_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_extra_widget, get_extra_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filename, get_filename, arginfo_gtk_gtkfilechooser_get_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filenames, get_filenames, arginfo_gtk_gtkfilechooser_get_filenames, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_filename, get_preview_filename, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_uri, get_preview_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget, get_preview_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget_active, get_preview_widget_active, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_show_hidden, get_show_hidden, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uri, get_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_use_preview_label, get_use_preview_label, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folder_uris, list_shortcut_folder_uris, arginfo_gtk_gtkfilechooser_list_shortcut_folder_uris, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folders, list_shortcut_folders, arginfo_gtk_gtkfilechooser_list_shortcut_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_filter, remove_filter, arginfo_gtk_gtkfilechooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder, remove_shortcut_folder, arginfo_gtk_gtkfilechooser_remove_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder_uri, remove_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_remove_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_filename, select_filename, arginfo_gtk_gtkfilechooser_select_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_uri, select_uri, arginfo_gtk_gtkfilechooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_action, set_action, arginfo_gtk_gtkfilechooser_set_action, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder, set_current_folder, arginfo_gtk_gtkfilechooser_set_current_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder_uri, set_current_folder_uri, arginfo_gtk_gtkfilechooser_set_current_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_name, set_current_name, arginfo_gtk_gtkfilechooser_set_current_name, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_extra_widget, set_extra_widget, arginfo_gtk_gtkfilechooser_set_extra_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filename, set_filename, arginfo_gtk_gtkfilechooser_set_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filter, set_filter, arginfo_gtk_gtkfilechooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_local_only, set_local_only, arginfo_gtk_gtkfilechooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget, set_preview_widget, arginfo_gtk_gtkfilechooser_set_preview_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget_active, set_preview_widget_active, arginfo_gtk_gtkfilechooser_set_preview_widget_active, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkfilechooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_show_hidden, set_show_hidden, arginfo_gtk_gtkfilechooser_set_show_hidden, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_uri, set_uri, arginfo_gtk_gtkfilechooser_set_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_use_preview_label, set_use_preview_label, arginfo_gtk_gtkfilechooser_set_use_preview_label, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_filename, unselect_filename, arginfo_gtk_gtkfilechooser_unselect_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkfilechooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_do_overwrite_confirmation, set_do_overwrite_confirmation, arginfo_gtk_gtkfilechooser_set_do_overwrite_confirmation, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_do_overwrite_confirmation, get_do_overwrite_confirmation, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_create_folders, set_create_folders, arginfo_gtk_gtkfilechooser_set_create_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_create_folders, get_create_folders, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFontSelection, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontSelection);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontSelection);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_font)
{
	GdkFont *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkFontSelection::get_font_name()" TSRMLS_CC);

    php_retval = gtk_font_selection_get_font(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
	phpg_gboxed_new(&return_value, GDK_TYPE_FONT, php_retval, TRUE, TRUE TSRMLS_CC);

}


static PHP_METHOD(GtkFontSelection, get_font_name)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_font_name(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkFontSelection, get_preview_text)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_preview_text(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFontSelection, set_font_name)
{
	char *fontname;
	zend_bool free_fontname = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &fontname, &free_fontname))
		return;

    php_retval = gtk_font_selection_set_font_name(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)), fontname);
	if (free_fontname) g_free(fontname);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontSelection, set_preview_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_font_selection_set_preview_text(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkFontSelection, get_family_list)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_family_list(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_face_list)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_face_list(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_size_entry)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_size_entry(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_size_list)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_size_list(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_preview_entry)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_preview_entry(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_family)
{
	PangoFontFamily* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_family(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_face)
{
	PangoFontFace* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_face(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFontSelection, get_size)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_selection_get_size(GTK_FONT_SELECTION(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontselection_set_font_name, 0)
    ZEND_ARG_INFO(0, fontname)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontselection_set_preview_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfontselection_methods[] = {
	PHP_ME(GtkFontSelection, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_face,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_face_list,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_family,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_family_list,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_font,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_font_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_preview_entry,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_preview_text,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_size,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_size_entry,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, get_size_list,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, set_font_name,        arginfo_gtk_gtkfontselection_set_font_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontSelection, set_preview_text,     arginfo_gtk_gtkfontselection_set_preview_text, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkGammaCurve, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkGammaCurve);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkGammaCurve);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkGammaCurve, table)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_GAMMA_CURVE(((phpg_gobject_t *)object)->obj)->table;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkGammaCurve, curve)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_GAMMA_CURVE(((phpg_gobject_t *)object)->obj)->curve;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkGammaCurve, gamma)
{
	double php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_GAMMA_CURVE(((phpg_gobject_t *)object)->obj)->gamma;
	RETVAL_DOUBLE(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkGammaCurve, gamma_dialog)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_GAMMA_CURVE(((phpg_gobject_t *)object)->obj)->gamma_dialog;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkGammaCurve, gamma_text)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_GAMMA_CURVE(((phpg_gobject_t *)object)->obj)->gamma_text;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkgammacurve_prop_info[] = {
	{ "table", PHPG_PROP_READ_FN(GtkGammaCurve, table), NULL },
	{ "curve", PHPG_PROP_READ_FN(GtkGammaCurve, curve), NULL },
	{ "gamma", PHPG_PROP_READ_FN(GtkGammaCurve, gamma), NULL },
	{ "gamma_dialog", PHPG_PROP_READ_FN(GtkGammaCurve, gamma_dialog), NULL },
	{ "gamma_text", PHPG_PROP_READ_FN(GtkGammaCurve, gamma_text), NULL },
	{ NULL, NULL, NULL },
};


static zend_function_entry gtkgammacurve_methods[] = {
	PHP_ME(GtkGammaCurve, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSeparatorToolItem, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSeparatorToolItem);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSeparatorToolItem);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkSeparatorToolItem, get_draw)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_separator_tool_item_get_draw(GTK_SEPARATOR_TOOL_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkSeparatorToolItem, set_draw)
{
	zend_bool draw;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &draw))
		return;

    gtk_separator_tool_item_set_draw(GTK_SEPARATOR_TOOL_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)draw);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkseparatortoolitem_set_draw, 0)
    ZEND_ARG_INFO(0, draw)
ZEND_END_ARG_INFO();

static zend_function_entry gtkseparatortoolitem_methods[] = {
	PHP_ME(GtkSeparatorToolItem, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSeparatorToolItem, get_draw,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkSeparatorToolItem, set_draw,             arginfo_gtk_gtkseparatortoolitem_set_draw, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToolButton, __construct)
{
	GtkWidget *icon_widget = NULL;
	zval *php_icon_widget = NULL;
	char *label = NULL;
	zend_bool free_label = FALSE;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Nu", &php_icon_widget, gtkwidget_ce, &label, &free_label)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolButton);
	}
    if (php_icon_widget) {
        if (Z_TYPE_P(php_icon_widget) == IS_NULL)
            icon_widget = NULL;
        else
            icon_widget = GTK_WIDGET(PHPG_GOBJECT(php_icon_widget));
    }

	wrapped_obj = (GObject *) gtk_tool_button_new(icon_widget, label);
	if (free_label) g_free(label);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToolButton, new_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolButton);
	}

	wrapped_obj = (GObject *) gtk_tool_button_new_from_stock(stock_id);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToolButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkToolButton, get_icon_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_icon_widget(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolButton, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_label(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkToolButton, get_label_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_label_widget(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkToolButton, get_stock_id)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_stock_id(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkToolButton, get_use_underline)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_use_underline(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToolButton, set_icon_widget)
{
	GtkWidget *icon_widget = NULL;
	zval *php_icon_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_icon_widget, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_icon_widget) != IS_NULL)
        icon_widget = GTK_WIDGET(PHPG_GOBJECT(php_icon_widget));

    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), icon_widget);

}


static PHP_METHOD(GtkToolButton, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_tool_button_set_label(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkToolButton, set_label_widget)
{
	GtkWidget *label_widget = NULL;
	zval *php_label_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_label_widget, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_label_widget) != IS_NULL)
        label_widget = GTK_WIDGET(PHPG_GOBJECT(php_label_widget));

    gtk_tool_button_set_label_widget(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), label_widget);

}


static PHP_METHOD(GtkToolButton, set_stock_id)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id))
		return;

    gtk_tool_button_set_stock_id(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), stock_id);
	if (free_stock_id) g_free(stock_id);

}


static PHP_METHOD(GtkToolButton, set_use_underline)
{
	zend_bool use_underline;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_underline))
		return;

    gtk_tool_button_set_use_underline(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_underline);

}


static PHP_METHOD(GtkToolButton, set_icon_name)
{
	char *icon_name;
	zend_bool free_icon_name = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &icon_name, &free_icon_name))
		return;

    gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), icon_name);
	if (free_icon_name) g_free(icon_name);

}


static PHP_METHOD(GtkToolButton, get_icon_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_tool_button_get_icon_name(GTK_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktoolbutton_gtk_tool_button_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, icon_widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_tool_button_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_icon_widget, 0)
    ZEND_ARG_OBJ_INFO(0, icon_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_label_widget, 0)
    ZEND_ARG_OBJ_INFO(0, label_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_stock_id, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_use_underline, 0)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoolbutton_set_icon_name, 0)
    ZEND_ARG_INFO(0, icon_name)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoolbutton_methods[] = {
	PHP_ME(GtkToolButton, __construct,          arginfo_gtk_gtktoolbutton_gtk_tool_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, new_from_stock,       arginfo_gtk_gtktoolbutton_tool_button_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkToolButton, get_icon_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, get_icon_widget,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, get_label_widget,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, get_stock_id,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, get_use_underline,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_icon_name,        arginfo_gtk_gtktoolbutton_set_icon_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_icon_widget,      arginfo_gtk_gtktoolbutton_set_icon_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_label,            arginfo_gtk_gtktoolbutton_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_label_widget,     arginfo_gtk_gtktoolbutton_set_label_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_stock_id,         arginfo_gtk_gtktoolbutton_set_stock_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToolButton, set_use_underline,    arginfo_gtk_gtktoolbutton_set_use_underline, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkItem, deselect)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_item_deselect(GTK_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkItem, select)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_item_select(GTK_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkItem, toggle)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_item_toggle(GTK_ITEM(PHPG_GOBJECT(this_ptr)));

}


static zend_function_entry gtkitem_methods[] = {
	PHP_ME(GtkItem, deselect,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItem, select,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkItem, toggle,               NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkToggleToolButton, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleToolButton);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleToolButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkToggleToolButton, new_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleToolButton);
	}

	wrapped_obj = (GObject *) gtk_toggle_tool_button_new_from_stock(stock_id);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleToolButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkToggleToolButton, get_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleToolButton, set_active)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

    gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoggletoolbutton_toggle_tool_button_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktoggletoolbutton_set_active, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

static zend_function_entry gtktoggletoolbutton_methods[] = {
	PHP_ME(GtkToggleToolButton, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleToolButton, new_from_stock,       arginfo_gtk_gtktoggletoolbutton_toggle_tool_button_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkToggleToolButton, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleToolButton, set_active,           arginfo_gtk_gtktoggletoolbutton_set_active, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkMenuToolButton, __construct)
{
	zval *icon_widget;
	char *label;
	zend_bool free_label = FALSE;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &icon_widget, gtkwidget_ce, &label, &free_label)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuToolButton);
	}

	wrapped_obj = (GObject *) gtk_menu_tool_button_new(GTK_WIDGET(PHPG_GOBJECT(icon_widget)), label);
	if (free_label) g_free(label);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuToolButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkMenuToolButton, new_from_stock)
{
	char *stock_id;
	zend_bool free_stock_id = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuToolButton);
	}

	wrapped_obj = (GObject *) gtk_menu_tool_button_new_from_stock(stock_id);
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuToolButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkMenuToolButton, get_menu)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_tool_button_get_menu(GTK_MENU_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMenuToolButton, set_arrow_tooltip)
{
	zval *tooltips;
	char *tip_text, *tip_private = NULL;
	zend_bool free_tip_text = FALSE, free_tip_private = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou|u", &tooltips, gtktooltips_ce, &tip_text, &free_tip_text, &tip_private, &free_tip_private))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkMenuToolButton::set_arrow_tooltip_text()" TSRMLS_CC);

    gtk_menu_tool_button_set_arrow_tooltip(GTK_MENU_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), GTK_TOOLTIPS(PHPG_GOBJECT(tooltips)), tip_text, tip_private);
	if (free_tip_text) g_free(tip_text);
	if (free_tip_private) g_free(tip_private);

}


static PHP_METHOD(GtkMenuToolButton, set_menu)
{
	zval *menu;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &menu, gtkwidget_ce))
		return;

    gtk_menu_tool_button_set_menu(GTK_MENU_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(menu)));

}


static PHP_METHOD(GtkMenuToolButton, set_arrow_tooltip_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_menu_tool_button_set_arrow_tooltip_text(GTK_MENU_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkMenuToolButton, set_arrow_tooltip_markup)
{
	char *markup;
	zend_bool free_markup = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &markup, &free_markup))
		return;

    gtk_menu_tool_button_set_arrow_tooltip_markup(GTK_MENU_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), markup);
	if (free_markup) g_free(markup);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenutoolbutton_gtk_menu_tool_button_new, 0)
    ZEND_ARG_OBJ_INFO(0, icon_widget, GtkWidget, 1)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenutoolbutton_gtk_menu_tool_button_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip, 0, 0, 2)
    ZEND_ARG_OBJ_INFO(0, tooltips, GtkTooltips, 1)
    ZEND_ARG_INFO(0, tip_text)
    ZEND_ARG_INFO(0, tip_private)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenutoolbutton_set_menu, 0)
    ZEND_ARG_OBJ_INFO(0, menu, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip_markup, 0)
    ZEND_ARG_INFO(0, markup)
ZEND_END_ARG_INFO();

static zend_function_entry gtkmenutoolbutton_methods[] = {
	PHP_ME(GtkMenuToolButton, __construct,          arginfo_gtk_gtkmenutoolbutton_gtk_menu_tool_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuToolButton, new_from_stock,       arginfo_gtk_gtkmenutoolbutton_gtk_menu_tool_button_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkMenuToolButton, get_menu,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuToolButton, set_arrow_tooltip,    arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuToolButton, set_arrow_tooltip_markup, arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuToolButton, set_arrow_tooltip_text, arginfo_gtk_gtkmenutoolbutton_set_arrow_tooltip_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuToolButton, set_menu,             arginfo_gtk_gtkmenutoolbutton_set_menu, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 5079 "ext/gtk+/gtk.overrides"

static PHP_METHOD(GtkRadioToolButton, __construct)
{
    zval *php_group = NULL;
    GtkRadioToolButton *group = NULL;
    char *stock_id = NULL;
    zend_bool free_stock_id;
    GObject *wrapped_obj;
    GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Nu", &php_group, gtkradiotoolbutton_ce, &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioToolButton);
    }

    if (php_group && Z_TYPE_P(php_group) != IS_NULL) {
        group = GTK_RADIO_TOOL_BUTTON(PHPG_GOBJECT(php_group));
    }

    if (group) {
        if (stock_id) {
            wrapped_obj = (GObject *) g_object_new(gtype, "group", group, "stock_id", stock_id, NULL);
        } else {
            wrapped_obj = (GObject *) g_object_new(gtype, "group", group, NULL);
        }
    } else {
        if (stock_id) {
            wrapped_obj = (GObject *) g_object_new(gtype, "stock_id", stock_id, NULL);
        } else {
            wrapped_obj = (GObject *) g_object_new(gtype, NULL);
        }
    }

    if (free_stock_id) g_free(stock_id);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioToolButton);
    }
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 56785 "ext/gtk+/gen_gtk.c"


#line 5121 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioToolButton, get_group)
{
    GSList *group, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    group = gtk_radio_tool_button_get_group(GTK_RADIO_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = group; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}

#line 56809 "ext/gtk+/gen_gtk.c"


#line 5150 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioToolButton, set_group)
{
    zval *php_button = NULL;
    GSList *list;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_button, gtkradiotoolbutton_ce))
        return;

    list = gtk_radio_tool_button_get_group(GTK_RADIO_TOOL_BUTTON(PHPG_GOBJECT(php_button)));
    if (list == gtk_radio_tool_button_get_group(GTK_RADIO_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)))) {
        return;
    }
    gtk_radio_tool_button_set_group(GTK_RADIO_TOOL_BUTTON(PHPG_GOBJECT(this_ptr)), list);
}

#line 56830 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkradiotoolbutton_gtk_radio_tool_button_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkRadioToolButton, 1)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkradiotoolbutton_set_group, 0, 0, 1)
    ZEND_ARG_OBJ_INFO(0, group, GtkRadioToolButton, 1)
ZEND_END_ARG_INFO();


static zend_function_entry gtkradiotoolbutton_methods[] = {
	PHP_ME(GtkRadioToolButton, __construct,          arginfo_gtk_gtkradiotoolbutton_gtk_radio_tool_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioToolButton, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioToolButton, set_group,            arginfo_gtk_gtkradiotoolbutton_set_group, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHandleBox, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHandleBox);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHandleBox);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkHandleBox, get_handle_position)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_handle_box_get_handle_position(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkHandleBox, get_shadow_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_handle_box_get_shadow_type(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkHandleBox, get_snap_edge)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_handle_box_get_snap_edge(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkHandleBox, set_handle_position)
{
	GtkPositionType position;
	zval *php_position = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_position))
		return;

	if (php_position && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_position, (gint *)&position) == FAILURE) {
		return;
	}

    gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)), position);

}


static PHP_METHOD(GtkHandleBox, set_shadow_type)
{
	GtkShadowType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_handle_box_set_shadow_type(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkHandleBox, set_snap_edge)
{
	GtkPositionType edge;
	zval *php_edge = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_edge))
		return;

	if (php_edge && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_edge, (gint *)&edge) == FAILURE) {
		return;
	}

    gtk_handle_box_set_snap_edge(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)), edge);

}


static PHP_METHOD(GtkHandleBox, get_child_detached)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_handle_box_get_child_detached(GTK_HANDLE_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhandlebox_set_handle_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhandlebox_set_shadow_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhandlebox_set_snap_edge, 0)
    ZEND_ARG_INFO(0, edge)
ZEND_END_ARG_INFO();

static zend_function_entry gtkhandlebox_methods[] = {
	PHP_ME(GtkHandleBox, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, get_child_detached,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, get_handle_position,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, get_shadow_type,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, get_snap_edge,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, set_handle_position,  arginfo_gtk_gtkhandlebox_set_handle_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, set_shadow_type,      arginfo_gtk_gtkhandlebox_set_shadow_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHandleBox, set_snap_edge,        arginfo_gtk_gtkhandlebox_set_snap_edge, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4565 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMenuItem, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 1;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuItem);
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    if (text) {
        GtkWidget *accel_label;
        accel_label = g_object_new(GTK_TYPE_ACCEL_LABEL, NULL);
        gtk_misc_set_alignment(GTK_MISC(accel_label), 0.0, 0.5);
        gtk_container_add(GTK_CONTAINER(wrapped_obj), accel_label);
        if (use_underline)
            gtk_label_set_text_with_mnemonic(GTK_LABEL(accel_label), text);
        else
            gtk_label_set_text(GTK_LABEL(accel_label), text);
        gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(accel_label), GTK_WIDGET(wrapped_obj));
        gtk_widget_show(accel_label);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuItem);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);

}

#line 57050 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkMenuItem, activate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_item_activate(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuItem, deselect)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_item_deselect(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuItem, get_right_justified)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_item_get_right_justified(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkMenuItem, get_submenu)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_item_get_submenu(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkMenuItem, remove_submenu)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("since GTK+ 2.12, please use GtkMenuItem::set_submenu(null);" TSRMLS_CC);

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuItem, right_justify)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

	phpg_warn_deprecated("use GtkMenuItem::set_right_justified()" TSRMLS_CC);

    gtk_menu_item_right_justify(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuItem, select)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_menu_item_select(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkMenuItem, set_right_justified)
{
	zend_bool right_justified;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &right_justified))
		return;

    gtk_menu_item_set_right_justified(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)right_justified);

}


static PHP_METHOD(GtkMenuItem, set_submenu)
{
	zval *submenu;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &submenu, gtkwidget_ce))
		return;

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(submenu)));

}


static PHP_METHOD(GtkMenuItem, toggle_size_allocate)
{
	long allocation;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &allocation))
		return;

    gtk_menu_item_toggle_size_allocate(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gint)allocation);

}

#line 4470 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMenuItem, toggle_size_request)
{
    gint requisition;
    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_menu_item_toggle_size_request(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), &requisition);

    RETURN_LONG(requisition);
}

#line 57206 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkMenuItem, get_accel_path)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_item_get_accel_path(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkMenuItem, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_menu_item_set_label(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkMenuItem, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_item_get_label(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkMenuItem, set_use_underline)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_menu_item_set_use_underline(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkMenuItem, get_use_underline)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_menu_item_get_use_underline(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}

#line 4454 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkMenuItem, set_item_accel_path)
{
    char *accel_path;
    zend_bool free_accel_path = FALSE;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &accel_path, &free_accel_path))
        return;

    gtk_menu_item_set_accel_path(GTK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), accel_path);
    if (free_accel_path) g_free(accel_path);
}

#line 57324 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkmenuitem_gtk_menu_item_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_set_right_justified, 0)
    ZEND_ARG_INFO(0, right_justified)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_set_submenu, 0)
    ZEND_ARG_OBJ_INFO(0, submenu, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_toggle_size_allocate, 0)
    ZEND_ARG_INFO(0, allocation)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_set_use_underline, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkmenuitem_set_item_accel_path, 0)
    ZEND_ARG_INFO(0, accel_path)
ZEND_END_ARG_INFO();


static zend_function_entry gtkmenuitem_methods[] = {
	PHP_ME(GtkMenuItem, __construct,          arginfo_gtk_gtkmenuitem_gtk_menu_item_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, activate,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, deselect,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, get_accel_path,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, get_right_justified,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, get_submenu,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, get_use_underline,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, remove_submenu,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, right_justify,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, select,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, set_item_accel_path,  arginfo_gtk_gtkmenuitem_set_item_accel_path, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, set_label,            arginfo_gtk_gtkmenuitem_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, set_right_justified,  arginfo_gtk_gtkmenuitem_set_right_justified, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, set_submenu,          arginfo_gtk_gtkmenuitem_set_submenu, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, set_use_underline,    arginfo_gtk_gtkmenuitem_set_use_underline, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, toggle_size_allocate, arginfo_gtk_gtkmenuitem_toggle_size_allocate, ZEND_ACC_PUBLIC)
	PHP_ME(GtkMenuItem, toggle_size_request,  NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4294 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkListItem, __construct)
{
    char *label;
    zend_bool free_label;
    GObject *wrapped_obj;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|u", &label, &free_label)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkListItem);
    }

    phpg_warn_deprecated("use GtkTreeView" TSRMLS_CC);

    if (ZEND_NUM_ARGS() == 1) {
        wrapped_obj = (GObject *) gtk_list_item_new_with_label(label);
        if (free_label) g_free(label);
    } else {
        wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkListItem);
    }

    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    if (ZEND_NUM_ARGS() != 1) {
        g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
    }
}

#line 57415 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkListItem, deselect)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_list_item_deselect(GTK_LIST_ITEM(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkListItem, select)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_list_item_select(GTK_LIST_ITEM(PHPG_GOBJECT(this_ptr)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtklistitem_gtk_list_item_new, 0, 0, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();


static zend_function_entry gtklistitem_methods[] = {
	PHP_ME(GtkListItem, __construct,          arginfo_gtk_gtklistitem_gtk_list_item_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListItem, deselect,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkListItem, select,               NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkTearoffMenuItem, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTearoffMenuItem);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkTearoffMenuItem);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtktearoffmenuitem_methods[] = {
	PHP_ME(GtkTearoffMenuItem, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkSeparatorMenuItem, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSeparatorMenuItem);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkSeparatorMenuItem);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkseparatormenuitem_methods[] = {
	PHP_ME(GtkSeparatorMenuItem, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1558 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCheckMenuItem, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 1;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuItem);
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    if (text) {
        GtkWidget *accel_label;
        accel_label = g_object_new(GTK_TYPE_ACCEL_LABEL, NULL);
        gtk_misc_set_alignment(GTK_MISC(accel_label), 0.0, 0.5);
        gtk_container_add(GTK_CONTAINER(wrapped_obj), accel_label);
        if (use_underline)
            gtk_label_set_text_with_mnemonic(GTK_LABEL(accel_label), text);
        else
            gtk_label_set_text(GTK_LABEL(accel_label), text);
        gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(accel_label), GTK_WIDGET(wrapped_obj));
        gtk_widget_show(accel_label);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkMenuItem);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 57547 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkCheckMenuItem, active)
{
	gboolean php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CHECK_MENU_ITEM(((phpg_gobject_t *)object)->obj)->active;
	RETVAL_BOOL(php_retval);
    return SUCCESS;
}


static prop_info_t gtkcheckmenuitem_prop_info[] = {
	{ "active", PHPG_PROP_READ_FN(GtkCheckMenuItem, active), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkCheckMenuItem, get_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCheckMenuItem, get_draw_as_radio)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_check_menu_item_get_draw_as_radio(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCheckMenuItem, get_inconsistent)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_check_menu_item_get_inconsistent(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCheckMenuItem, set_active)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


static PHP_METHOD(GtkCheckMenuItem, set_draw_as_radio)
{
	zend_bool draw_as_radio;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &draw_as_radio))
		return;

    gtk_check_menu_item_set_draw_as_radio(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)draw_as_radio);

}


static PHP_METHOD(GtkCheckMenuItem, set_inconsistent)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_check_menu_item_set_inconsistent(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkCheckMenuItem, set_show_toggle)
{
	zend_bool always;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &always))
		return;

	phpg_warn_deprecated("don't use this method" TSRMLS_CC);

    gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)always);

}


static PHP_METHOD(GtkCheckMenuItem, set_state)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

	phpg_warn_deprecated("use GtkCheckMenuItem::set_active()" TSRMLS_CC);

    gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


static PHP_METHOD(GtkCheckMenuItem, toggled)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_check_menu_item_toggled(GTK_CHECK_MENU_ITEM(PHPG_GOBJECT(this_ptr)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcheckmenuitem_gtk_check_menu_item_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcheckmenuitem_set_active, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcheckmenuitem_set_draw_as_radio, 0)
    ZEND_ARG_INFO(0, draw_as_radio)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcheckmenuitem_set_inconsistent, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcheckmenuitem_set_show_toggle, 0)
    ZEND_ARG_INFO(0, always)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcheckmenuitem_set_state, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcheckmenuitem_methods[] = {
	PHP_ME(GtkCheckMenuItem, __construct,          arginfo_gtk_gtkcheckmenuitem_gtk_check_menu_item_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, get_draw_as_radio,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, get_inconsistent,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, set_active,           arginfo_gtk_gtkcheckmenuitem_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, set_draw_as_radio,    arginfo_gtk_gtkcheckmenuitem_set_draw_as_radio, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, set_inconsistent,     arginfo_gtk_gtkcheckmenuitem_set_inconsistent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, set_show_toggle,      arginfo_gtk_gtkcheckmenuitem_set_show_toggle, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, set_state,            arginfo_gtk_gtkcheckmenuitem_set_state, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCheckMenuItem, toggled,              NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4145 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkImageMenuItem, __construct)
{
	char *stock_id = NULL;
	zend_bool free_stock_id;
	zval *php_accel_group = NULL;
    GtkAccelGroup *accel_group = NULL;
	GObject *wrapped_obj;

    if (phpg_gtype_from_zval(this_ptr TSRMLS_CC) != GTK_TYPE_IMAGE_MENU_ITEM) {
        PHPG_THROW_EXCEPTION(phpg_construct_exception, "Subclassing GtkImageMenuItem requires more than you know");
        return;
    }

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|uO", &stock_id, &free_stock_id, &php_accel_group, gtkaccelgroup_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImageMenuItem);
	}

    if (php_accel_group) {
        accel_group = GTK_ACCEL_GROUP(PHPG_GOBJECT(php_accel_group));
    }

    if (stock_id) {
        wrapped_obj = (GObject *) gtk_image_menu_item_new_from_stock(stock_id, accel_group);
    } else {
        wrapped_obj = (GObject *) gtk_image_menu_item_new();
    }
	if (free_stock_id) g_free(stock_id);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkImageMenuItem);
	}
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 57778 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkImageMenuItem, get_image)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_image_menu_item_get_image(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkImageMenuItem, set_image)
{
	GtkWidget *image = NULL;
	zval *php_image;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "N", &php_image, gtkwidget_ce))
		return;
    if (Z_TYPE_P(php_image) != IS_NULL)
        image = GTK_WIDGET(PHPG_GOBJECT(php_image));

    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)), image);

}


static PHP_METHOD(GtkImageMenuItem, set_always_show_image)
{
	zend_bool always_show;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &always_show))
		return;

    gtk_image_menu_item_set_always_show_image(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)always_show);

}


static PHP_METHOD(GtkImageMenuItem, get_always_show_image)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_image_menu_item_get_always_show_image(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkImageMenuItem, set_use_stock)
{
	zend_bool use_stock;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_stock))
		return;

    gtk_image_menu_item_set_use_stock(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)), (gboolean)use_stock);

}


static PHP_METHOD(GtkImageMenuItem, get_use_stock)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_image_menu_item_get_use_stock(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkImageMenuItem, set_accel_group)
{
	zval *accel_group;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &accel_group, gtkaccelgroup_ce))
		return;

    gtk_image_menu_item_set_accel_group(GTK_IMAGE_MENU_ITEM(PHPG_GOBJECT(this_ptr)), GTK_ACCEL_GROUP(PHPG_GOBJECT(accel_group)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkimagemenuitem_gtk_image_menu_item_new, 0, 0, 0)
    ZEND_ARG_INFO(0, stock_id)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimagemenuitem_set_image, 0)
    ZEND_ARG_OBJ_INFO(0, image, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimagemenuitem_set_always_show_image, 0)
    ZEND_ARG_INFO(0, always_show)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimagemenuitem_set_use_stock, 0)
    ZEND_ARG_INFO(0, use_stock)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkimagemenuitem_set_accel_group, 0)
    ZEND_ARG_OBJ_INFO(0, accel_group, GtkAccelGroup, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkimagemenuitem_methods[] = {
	PHP_ME(GtkImageMenuItem, __construct,          arginfo_gtk_gtkimagemenuitem_gtk_image_menu_item_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, get_always_show_image, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, get_image,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, get_use_stock,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, set_accel_group,      arginfo_gtk_gtkimagemenuitem_set_accel_group, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, set_always_show_image, arginfo_gtk_gtkimagemenuitem_set_always_show_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, set_image,            arginfo_gtk_gtkimagemenuitem_set_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkImageMenuItem, set_use_stock,        arginfo_gtk_gtkimagemenuitem_set_use_stock, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4937 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioMenuItem, __construct)
{
    zval *php_group = NULL;
    char *label = NULL;
    zend_bool free_label, use_underline = 1;
    GtkRadioMenuItem *group = NULL;
    GObject *wrapped_obj;
    GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Nub", &php_group, gtkradiomenuitem_ce, &label, &free_label, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioMenuItem);
    }

    if (php_group && Z_TYPE_P(php_group) != IS_NULL) {
        group = GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(php_group));
    }

    wrapped_obj = (GObject *) g_object_new(gtype, NULL);

    if (label) {
        GtkWidget *accel_label;
        accel_label = g_object_new(GTK_TYPE_ACCEL_LABEL, NULL);
        gtk_misc_set_alignment(GTK_MISC(accel_label), 0.0, 0.5);
        gtk_container_add(GTK_CONTAINER(wrapped_obj), accel_label);
        if (use_underline)
            gtk_label_set_text_with_mnemonic(GTK_LABEL(accel_label), label);
        else
            gtk_label_set_text(GTK_LABEL(accel_label), label);
        gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(accel_label), GTK_WIDGET(wrapped_obj));
        gtk_widget_show(accel_label);
    }
    if (group) {
        gtk_radio_menu_item_set_group(GTK_RADIO_MENU_ITEM(wrapped_obj), gtk_radio_menu_item_get_group(group));
    }

    if (free_label) g_free(label);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioMenuItem);
    }
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);

}

#line 57965 "ext/gtk+/gen_gtk.c"


#line 5046 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioMenuItem, get_group)
{
    GSList *list, *tmp;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    list = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    array_init(return_value);
    for (tmp = list; tmp; tmp = tmp->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(tmp->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}

#line 57989 "ext/gtk+/gen_gtk.c"


#line 4991 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioMenuItem, new_from_widget)
{
    GtkWidget* item;
    char *label;
    zend_bool free_label, use_underline = 1;

    NOT_STATIC_METHOD();

    if (ZEND_NUM_ARGS() == 0) {
        item = gtk_radio_menu_item_new_from_widget(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)));
    } else {
        if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u|b", &label, &free_label, &use_underline))
            return;
        if (!use_underline) {
            item = gtk_radio_menu_item_new_with_label_from_widget(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)), label);
        } else {
            item = gtk_radio_menu_item_new_with_mnemonic_from_widget(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)), label);
        }
        if (free_label) g_free(label);
    }

    phpg_gobject_new(&return_value, (GObject *)item TSRMLS_CC);
}

#line 58017 "ext/gtk+/gen_gtk.c"


#line 5024 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioMenuItem, set_group)
{
    zval *php_group = NULL;
    GtkRadioMenuItem *group = NULL;
    GSList *list;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_group, gtkradiomenuitem_ce)) {
        return;
    }

    group = GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(php_group));
    list = gtk_radio_menu_item_get_group(group);
    if (list == gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)))) {
        return;
    }
    gtk_radio_menu_item_set_group(GTK_RADIO_MENU_ITEM(PHPG_GOBJECT(this_ptr)), list);
}

#line 58041 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkradiomenuitem_gtk_radio_menu_item_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkRadioMenuItem, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkradiomenuitem_new_from_widget, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkradiomenuitem_set_group, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkRadioMenuItem, 1)
ZEND_END_ARG_INFO();


static zend_function_entry gtkradiomenuitem_methods[] = {
	PHP_ME(GtkRadioMenuItem, __construct,          arginfo_gtk_gtkradiomenuitem_gtk_radio_menu_item_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioMenuItem, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioMenuItem, new_from_widget,      arginfo_gtk_gtkradiomenuitem_new_from_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioMenuItem, set_group,            arginfo_gtk_gtkradiomenuitem_set_group, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHBox, __construct)
{
	GParameter params[2];
	zval *php_args[3] = { NULL, };
	char *prop_names[] = { "homogeneous", "spacing",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^b^i", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHBox);
	}

	memset(&params, 0, 2 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHBox);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHBox);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkhbox_gtk_hbox_new, 0, 0, 1)
    ZEND_ARG_INFO(0, homogeneous)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtkhbox_methods[] = {
	PHP_ME(GtkHBox, __construct,          arginfo_gtk_gtkhbox_gtk_hbox_new, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFrame, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "label",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFrame);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFrame);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFrame);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFrame, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_frame_get_label(GTK_FRAME(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}

#line 3628 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkFrame, get_label_align)
{
    gfloat xalign, yalign;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_frame_get_label_align(GTK_FRAME(PHPG_GOBJECT(this_ptr)), &xalign, &yalign);

    php_gtk_build_value(&return_value, "(ff)", xalign, yalign);
}

#line 58196 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkFrame, get_label_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_frame_get_label_widget(GTK_FRAME(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkFrame, get_shadow_type)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_frame_get_shadow_type(GTK_FRAME(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkFrame, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_frame_set_label(GTK_FRAME(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkFrame, set_label_align)
{
	double xalign, yalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &xalign, &yalign))
		return;

    gtk_frame_set_label_align(GTK_FRAME(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign);

}


static PHP_METHOD(GtkFrame, set_label_widget)
{
	zval *label_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &label_widget, gtkwidget_ce))
		return;

    gtk_frame_set_label_widget(GTK_FRAME(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(label_widget)));

}


static PHP_METHOD(GtkFrame, set_shadow_type)
{
	GtkShadowType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_frame_set_shadow_type(GTK_FRAME(PHPG_GOBJECT(this_ptr)), type);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkframe_gtk_frame_new, 0, 0, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkframe_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkframe_set_label_align, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkframe_set_label_widget, 0)
    ZEND_ARG_OBJ_INFO(0, label_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkframe_set_shadow_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

static zend_function_entry gtkframe_methods[] = {
	PHP_ME(GtkFrame, __construct,          arginfo_gtk_gtkframe_gtk_frame_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, get_label_align,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, get_label_widget,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, get_shadow_type,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, set_label,            arginfo_gtk_gtkframe_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, set_label_align,      arginfo_gtk_gtkframe_set_label_align, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, set_label_widget,     arginfo_gtk_gtkframe_set_label_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFrame, set_shadow_type,      arginfo_gtk_gtkframe_set_shadow_type, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkStatusbar, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusbar);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkStatusbar);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkStatusbar, get_context_id)
{
	char *context_description;
	zend_bool free_context_description = FALSE;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &context_description, &free_context_description))
		return;

    php_retval = gtk_statusbar_get_context_id(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), context_description);
	if (free_context_description) g_free(context_description);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkStatusbar, get_has_resize_grip)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_statusbar_get_has_resize_grip(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkStatusbar, pop)
{
	long context_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &context_id))
		return;

    gtk_statusbar_pop(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), (guint)context_id);

}


static PHP_METHOD(GtkStatusbar, push)
{
	long context_id, php_retval;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &context_id, &text, &free_text))
		return;

    php_retval = gtk_statusbar_push(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), (guint)context_id, text);
	if (free_text) g_free(text);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkStatusbar, set_has_resize_grip)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkStatusbar, get_message_area)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_statusbar_get_message_area(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkStatusbar, remove_all)
{
	long context_id;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &context_id))
		return;

    gtk_statusbar_remove_all(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), (guint)context_id);

}

#line 5352 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkStatusbar, remove_message)
{
    long context_id, message_id;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &context_id, &message_id))
        return;

    gtk_statusbar_remove(GTK_STATUSBAR(PHPG_GOBJECT(this_ptr)), (guint)context_id, (guint)message_id);
}

#line 58470 "ext/gtk+/gen_gtk.c"



PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_get_context_id, 0)
    ZEND_ARG_INFO(0, context_description)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_pop, 0)
    ZEND_ARG_INFO(0, context_id)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_push, 0)
    ZEND_ARG_INFO(0, context_id)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_set_has_resize_grip, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_remove_all, 0)
    ZEND_ARG_INFO(0, context_id)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkstatusbar_remove_message, 0)
    ZEND_ARG_INFO(0, context_id)
    ZEND_ARG_INFO(0, message_id)
ZEND_END_ARG_INFO();


static zend_function_entry gtkstatusbar_methods[] = {
	PHP_ME(GtkStatusbar, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, get_context_id,       arginfo_gtk_gtkstatusbar_get_context_id, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, get_has_resize_grip,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, get_message_area,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, pop,                  arginfo_gtk_gtkstatusbar_pop, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, push,                 arginfo_gtk_gtkstatusbar_push, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, remove_all,           arginfo_gtk_gtkstatusbar_remove_all, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, remove_message,       arginfo_gtk_gtkstatusbar_remove_message, ZEND_ACC_PUBLIC)
	PHP_ME(GtkStatusbar, set_has_resize_grip,  arginfo_gtk_gtkstatusbar_set_has_resize_grip, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCombo, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCombo);
	}


	phpg_warn_deprecated("use GtkComboBoxEntry" TSRMLS_CC);

	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCombo);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


PHPG_PROP_READER(GtkCombo, entry)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COMBO(((phpg_gobject_t *)object)->obj)->entry;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


PHPG_PROP_READER(GtkCombo, list)
{
	GtkWidget* php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_COMBO(((phpg_gobject_t *)object)->obj)->list;
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
    return SUCCESS;
}


static prop_info_t gtkcombo_prop_info[] = {
	{ "entry", PHPG_PROP_READ_FN(GtkCombo, entry), NULL },
	{ "list", PHPG_PROP_READ_FN(GtkCombo, list), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkCombo, disable_activate)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_combo_disable_activate(GTK_COMBO(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCombo, set_case_sensitive)
{
	zend_bool val;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &val))
		return;

    gtk_combo_set_case_sensitive(GTK_COMBO(PHPG_GOBJECT(this_ptr)), (gboolean)val);

}


static PHP_METHOD(GtkCombo, set_item_string)
{
	zval *item;
	char *item_value;
	zend_bool free_item_value = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ou", &item, gtkitem_ce, &item_value, &free_item_value))
		return;

    gtk_combo_set_item_string(GTK_COMBO(PHPG_GOBJECT(this_ptr)), GTK_ITEM(PHPG_GOBJECT(item)), item_value);
	if (free_item_value) g_free(item_value);

}

#line 2127 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCombo, set_popdown_strings)
{
    zval *php_strings, **item;
    GList *strings = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_strings)) {
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_strings));

    while (zend_hash_get_current_data(Z_ARRVAL_P(php_strings), (void **)&item) == SUCCESS) {

        gchar *utf8 = NULL;
        gsize utf8_len = 0;
        zend_bool free_utf8 = 0;

        convert_to_string_ex(item);
        utf8 = phpg_to_utf8(Z_STRVAL_PP(item), Z_STRLEN_PP(item), &utf8_len, &free_utf8 TSRMLS_CC);

        if (utf8 && free_utf8) {
            SEPARATE_ZVAL(item);
            zval_dtor(*item);
            ZVAL_STRINGL(*item, utf8, utf8_len, 1);
            g_free(utf8);
        }

        strings = g_list_append(strings, Z_STRVAL_PP(item));
        zend_hash_move_forward(Z_ARRVAL_P(php_strings));
    }

    gtk_combo_set_popdown_strings(GTK_COMBO(PHPG_GOBJECT(this_ptr)), strings);

    g_list_free(strings);
}
#line 58661 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCombo, set_use_arrows)
{
	zend_bool val;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &val))
		return;

    gtk_combo_set_use_arrows(GTK_COMBO(PHPG_GOBJECT(this_ptr)), (gboolean)val);

}


static PHP_METHOD(GtkCombo, set_use_arrows_always)
{
	zend_bool val;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &val))
		return;

    gtk_combo_set_use_arrows_always(GTK_COMBO(PHPG_GOBJECT(this_ptr)), (gboolean)val);

}


static PHP_METHOD(GtkCombo, set_value_in_list)
{
	zend_bool val, ok_if_empty;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "bb", &val, &ok_if_empty))
		return;

    gtk_combo_set_value_in_list(GTK_COMBO(PHPG_GOBJECT(this_ptr)), (gboolean)val, (gboolean)ok_if_empty);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_case_sensitive, 0)
    ZEND_ARG_INFO(0, val)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_item_string, 0)
    ZEND_ARG_OBJ_INFO(0, item, GtkItem, 1)
    ZEND_ARG_INFO(0, item_value)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_popdown_strings, 0)
    ZEND_ARG_INFO(0, strings)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_use_arrows, 0)
    ZEND_ARG_INFO(0, val)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_use_arrows_always, 0)
    ZEND_ARG_INFO(0, val)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombo_set_value_in_list, 0)
    ZEND_ARG_INFO(0, val)
    ZEND_ARG_INFO(0, ok_if_empty)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcombo_methods[] = {
	PHP_ME(GtkCombo, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, disable_activate,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_case_sensitive,   arginfo_gtk_gtkcombo_set_case_sensitive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_item_string,      arginfo_gtk_gtkcombo_set_item_string, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_popdown_strings,  arginfo_gtk_gtkcombo_set_popdown_strings, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_use_arrows,       arginfo_gtk_gtkcombo_set_use_arrows, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_use_arrows_always, arginfo_gtk_gtkcombo_set_use_arrows_always, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCombo, set_value_in_list,    arginfo_gtk_gtkcombo_set_value_in_list, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFileChooserButton, __construct)
{
	char *title;
	zend_bool free_title = FALSE;
	GtkFileChooserAction action;
	zval *php_action = NULL;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uV", &title, &free_title, &php_action)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	wrapped_obj = (GObject *) gtk_file_chooser_button_new(title, action);
	if (free_title) g_free(title);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFileChooserButton, new_with_backend)
{
	char *title, *backend;
	zend_bool free_title = FALSE, free_backend = FALSE;
	GtkFileChooserAction action;
	zval *php_action = NULL;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "uVu", &title, &free_title, &php_action, &backend, &free_backend)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	if (php_action && phpg_gvalue_get_enum(GTK_TYPE_FILE_CHOOSER_ACTION, php_action, (gint *)&action) == FAILURE) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	phpg_warn_deprecated("since GTK+ 2.14, Use GtkFileChooserButton::__construct() instead" TSRMLS_CC);

	wrapped_obj = (GObject *) gtk_file_chooser_button_new_with_backend(title, action, backend);
	if (free_title) g_free(title);
	if (free_backend) g_free(backend);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkFileChooserButton, new_with_dialog)
{
	zval *dialog;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &dialog, gtkwidget_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}

	wrapped_obj = (GObject *) gtk_file_chooser_button_new_with_dialog(GTK_WIDGET(PHPG_GOBJECT(dialog)));

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFileChooserButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkFileChooserButton, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_button_get_title(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFileChooserButton, get_width_chars)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_button_get_width_chars(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkFileChooserButton, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_file_chooser_button_set_title(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkFileChooserButton, set_width_chars)
{
	long n_chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &n_chars))
		return;

    gtk_file_chooser_button_set_width_chars(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)), (gint)n_chars);

}


static PHP_METHOD(GtkFileChooserButton, get_focus_on_click)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_file_chooser_button_get_focus_on_click(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFileChooserButton, set_focus_on_click)
{
	zend_bool focus_on_click;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &focus_on_click))
		return;

    gtk_file_chooser_button_set_focus_on_click(GTK_FILE_CHOOSER_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)focus_on_click);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_INFO(0, action)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new_with_backend, 0)
    ZEND_ARG_INFO(0, title)
    ZEND_ARG_INFO(0, action)
    ZEND_ARG_INFO(0, backend)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new_with_dialog, 0)
    ZEND_ARG_OBJ_INFO(0, dialog, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_set_width_chars, 0)
    ZEND_ARG_INFO(0, n_chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfilechooserbutton_set_focus_on_click, 0)
    ZEND_ARG_INFO(0, focus_on_click)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfilechooserbutton_methods[] = {
	PHP_ME(GtkFileChooserButton, __construct,          arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, new_with_backend,     arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new_with_backend, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkFileChooserButton, new_with_dialog,      arginfo_gtk_gtkfilechooserbutton_file_chooser_button_new_with_dialog, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkFileChooserButton, get_focus_on_click,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, get_width_chars,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, set_focus_on_click,   arginfo_gtk_gtkfilechooserbutton_set_focus_on_click, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, set_title,            arginfo_gtk_gtkfilechooserbutton_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFileChooserButton, set_width_chars,      arginfo_gtk_gtkfilechooserbutton_set_width_chars, ZEND_ACC_PUBLIC)

	/***   GtkFileChooser interface implementations   ***/

	PHP_MALIAS(GtkFileChooser, add_filter, add_filter, arginfo_gtk_gtkfilechooser_add_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder, add_shortcut_folder, arginfo_gtk_gtkfilechooser_add_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, add_shortcut_folder_uri, add_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_add_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_action, get_action, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder, get_current_folder, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_current_folder_uri, get_current_folder_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_extra_widget, get_extra_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filename, get_filename, arginfo_gtk_gtkfilechooser_get_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filenames, get_filenames, arginfo_gtk_gtkfilechooser_get_filenames, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_filter, get_filter, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_local_only, get_local_only, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_filename, get_preview_filename, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_uri, get_preview_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget, get_preview_widget, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_preview_widget_active, get_preview_widget_active, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_select_multiple, get_select_multiple, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_show_hidden, get_show_hidden, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uri, get_uri, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_uris, get_uris, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_use_preview_label, get_use_preview_label, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_filters, list_filters, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folder_uris, list_shortcut_folder_uris, arginfo_gtk_gtkfilechooser_list_shortcut_folder_uris, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, list_shortcut_folders, list_shortcut_folders, arginfo_gtk_gtkfilechooser_list_shortcut_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_filter, remove_filter, arginfo_gtk_gtkfilechooser_remove_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder, remove_shortcut_folder, arginfo_gtk_gtkfilechooser_remove_shortcut_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, remove_shortcut_folder_uri, remove_shortcut_folder_uri, arginfo_gtk_gtkfilechooser_remove_shortcut_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_all, select_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_filename, select_filename, arginfo_gtk_gtkfilechooser_select_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, select_uri, select_uri, arginfo_gtk_gtkfilechooser_select_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_action, set_action, arginfo_gtk_gtkfilechooser_set_action, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder, set_current_folder, arginfo_gtk_gtkfilechooser_set_current_folder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_folder_uri, set_current_folder_uri, arginfo_gtk_gtkfilechooser_set_current_folder_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_current_name, set_current_name, arginfo_gtk_gtkfilechooser_set_current_name, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_extra_widget, set_extra_widget, arginfo_gtk_gtkfilechooser_set_extra_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filename, set_filename, arginfo_gtk_gtkfilechooser_set_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_filter, set_filter, arginfo_gtk_gtkfilechooser_set_filter, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_local_only, set_local_only, arginfo_gtk_gtkfilechooser_set_local_only, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget, set_preview_widget, arginfo_gtk_gtkfilechooser_set_preview_widget, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_preview_widget_active, set_preview_widget_active, arginfo_gtk_gtkfilechooser_set_preview_widget_active, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_select_multiple, set_select_multiple, arginfo_gtk_gtkfilechooser_set_select_multiple, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_show_hidden, set_show_hidden, arginfo_gtk_gtkfilechooser_set_show_hidden, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_uri, set_uri, arginfo_gtk_gtkfilechooser_set_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_use_preview_label, set_use_preview_label, arginfo_gtk_gtkfilechooser_set_use_preview_label, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_all, unselect_all, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_filename, unselect_filename, arginfo_gtk_gtkfilechooser_unselect_filename, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, unselect_uri, unselect_uri, arginfo_gtk_gtkfilechooser_unselect_uri, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_do_overwrite_confirmation, set_do_overwrite_confirmation, arginfo_gtk_gtkfilechooser_set_do_overwrite_confirmation, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_do_overwrite_confirmation, get_do_overwrite_confirmation, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, set_create_folders, set_create_folders, arginfo_gtk_gtkfilechooser_set_create_folders, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkFileChooser, get_create_folders, get_create_folders, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkExpander, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "label",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "^u", &php_args[0], &php_args[1])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkExpander);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkExpander);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkExpander);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkExpander, new_with_mnemonic)
{
	char *label;
	zend_bool free_label = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkExpander);
	}

	wrapped_obj = (GObject *) gtk_expander_new_with_mnemonic(label);
	if (free_label) g_free(label);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkExpander);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkExpander, get_expanded)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_expanded(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkExpander, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_label(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkExpander, get_label_widget)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_label_widget(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkExpander, get_spacing)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_spacing(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkExpander, get_use_markup)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_use_markup(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkExpander, get_use_underline)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_use_underline(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkExpander, set_expanded)
{
	zend_bool expanded;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &expanded))
		return;

    gtk_expander_set_expanded(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), (gboolean)expanded);

}


static PHP_METHOD(GtkExpander, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_expander_set_label(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkExpander, set_label_widget)
{
	zval *label_widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &label_widget, gtkwidget_ce))
		return;

    gtk_expander_set_label_widget(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(label_widget)));

}


static PHP_METHOD(GtkExpander, set_spacing)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_expander_set_spacing(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), (gint)spacing);

}


static PHP_METHOD(GtkExpander, set_use_markup)
{
	zend_bool use_markup;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_markup))
		return;

    gtk_expander_set_use_markup(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), (gboolean)use_markup);

}


static PHP_METHOD(GtkExpander, set_use_underline)
{
	zend_bool use_underline;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_underline))
		return;

    gtk_expander_set_use_underline(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), (gboolean)use_underline);

}


static PHP_METHOD(GtkExpander, set_label_fill)
{
	zend_bool label_fill;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &label_fill))
		return;

    gtk_expander_set_label_fill(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)), (gboolean)label_fill);

}


static PHP_METHOD(GtkExpander, get_label_fill)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_expander_get_label_fill(GTK_EXPANDER(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_gtk_expander_new, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_expander_new_with_mnemonic, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_expanded, 0)
    ZEND_ARG_INFO(0, expanded)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_label_widget, 0)
    ZEND_ARG_OBJ_INFO(0, label_widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_spacing, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_use_markup, 0)
    ZEND_ARG_INFO(0, use_markup)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_use_underline, 0)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkexpander_set_label_fill, 0)
    ZEND_ARG_INFO(0, label_fill)
ZEND_END_ARG_INFO();

static zend_function_entry gtkexpander_methods[] = {
	PHP_ME(GtkExpander, __construct,          arginfo_gtk_gtkexpander_gtk_expander_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, new_with_mnemonic,    arginfo_gtk_gtkexpander_expander_new_with_mnemonic, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkExpander, get_expanded,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_label_fill,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_label_widget,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_spacing,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_use_markup,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, get_use_underline,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_expanded,         arginfo_gtk_gtkexpander_set_expanded, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_label,            arginfo_gtk_gtkexpander_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_label_fill,       arginfo_gtk_gtkexpander_set_label_fill, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_label_widget,     arginfo_gtk_gtkexpander_set_label_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_spacing,          arginfo_gtk_gtkexpander_set_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_use_markup,       arginfo_gtk_gtkexpander_set_use_markup, ZEND_ACC_PUBLIC)
	PHP_ME(GtkExpander, set_use_underline,    arginfo_gtk_gtkexpander_set_use_underline, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAspectFrame, __construct)
{
	GParameter params[5];
	zval *php_args[6] = { NULL, };
	char *prop_names[] = { "label", "xalign", "yalign", "ratio", "obey_child",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^u^d^d^d^b", &php_args[0], &php_args[1], &php_args[2], &php_args[3], &php_args[4], &php_args[5])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAspectFrame);
	}

	memset(&params, 0, 5 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAspectFrame);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAspectFrame);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkAspectFrame, set)
{
	double xalign = 0.0, yalign = 0.0, ratio = 1.0;
	zend_bool obey_child = 1;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|dddb", &xalign, &yalign, &ratio, &obey_child))
		return;

    gtk_aspect_frame_set(GTK_ASPECT_FRAME(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign, (float)ratio, (gboolean)obey_child);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkaspectframe_gtk_aspect_frame_new, 0, 0, 0)
    ZEND_ARG_INFO(0, label)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
    ZEND_ARG_INFO(0, ratio)
    ZEND_ARG_INFO(0, obey_child)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkaspectframe_set, 0, 0, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
    ZEND_ARG_INFO(0, ratio)
    ZEND_ARG_INFO(0, obey_child)
ZEND_END_ARG_INFO();

static zend_function_entry gtkaspectframe_methods[] = {
	PHP_ME(GtkAspectFrame, __construct,          arginfo_gtk_gtkaspectframe_gtk_aspect_frame_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAspectFrame, set,                  arginfo_gtk_gtkaspectframe_set, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkEventBox, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEventBox);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkEventBox);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkEventBox, get_above_child)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_event_box_get_above_child(GTK_EVENT_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEventBox, get_visible_window)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_event_box_get_visible_window(GTK_EVENT_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkEventBox, set_above_child)
{
	zend_bool above_child;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &above_child))
		return;

    gtk_event_box_set_above_child(GTK_EVENT_BOX(PHPG_GOBJECT(this_ptr)), (gboolean)above_child);

}


static PHP_METHOD(GtkEventBox, set_visible_window)
{
	zend_bool visible_window;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visible_window))
		return;

    gtk_event_box_set_visible_window(GTK_EVENT_BOX(PHPG_GOBJECT(this_ptr)), (gboolean)visible_window);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeventbox_set_above_child, 0)
    ZEND_ARG_INFO(0, above_child)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkeventbox_set_visible_window, 0)
    ZEND_ARG_INFO(0, visible_window)
ZEND_END_ARG_INFO();

static zend_function_entry gtkeventbox_methods[] = {
	PHP_ME(GtkEventBox, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEventBox, get_above_child,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEventBox, get_visible_window,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEventBox, set_above_child,      arginfo_gtk_gtkeventbox_set_above_child, ZEND_ACC_PUBLIC)
	PHP_ME(GtkEventBox, set_visible_window,   arginfo_gtk_gtkeventbox_set_visible_window, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkAlignment, __construct)
{
	GParameter params[4];
	zval *php_args[5] = { NULL, };
	char *prop_names[] = { "xalign", "yalign", "xscale", "yscale",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^d^d^d^d", &php_args[0], &php_args[1], &php_args[2], &php_args[3])) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAlignment);
	}

	memset(&params, 0, 4 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAlignment);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkAlignment);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 1319 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkAlignment, get_padding)
{
    guint padding_top, padding_bottom, padding_left, padding_right;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_alignment_get_padding(GTK_ALIGNMENT(PHPG_GOBJECT(this_ptr)), &padding_top, &padding_bottom, &padding_left, &padding_right);

    php_gtk_build_value(&return_value, "(iiii)", padding_top, padding_bottom, padding_left, padding_right);
}

#line 59588 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkAlignment, set)
{
	double xalign, yalign, xscale, yscale;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dddd", &xalign, &yalign, &xscale, &yscale))
		return;

    gtk_alignment_set(GTK_ALIGNMENT(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign, (float)xscale, (float)yscale);

}


static PHP_METHOD(GtkAlignment, set_padding)
{
	long padding_top, padding_bottom, padding_left, padding_right;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &padding_top, &padding_bottom, &padding_left, &padding_right))
		return;

    gtk_alignment_set_padding(GTK_ALIGNMENT(PHPG_GOBJECT(this_ptr)), (guint)padding_top, (guint)padding_bottom, (guint)padding_left, (guint)padding_right);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkalignment_gtk_alignment_new, 0, 0, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
    ZEND_ARG_INFO(0, xscale)
    ZEND_ARG_INFO(0, yscale)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkalignment_set, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
    ZEND_ARG_INFO(0, xscale)
    ZEND_ARG_INFO(0, yscale)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkalignment_set_padding, 0)
    ZEND_ARG_INFO(0, padding_top)
    ZEND_ARG_INFO(0, padding_bottom)
    ZEND_ARG_INFO(0, padding_left)
    ZEND_ARG_INFO(0, padding_right)
ZEND_END_ARG_INFO();

static zend_function_entry gtkalignment_methods[] = {
	PHP_ME(GtkAlignment, __construct,          arginfo_gtk_gtkalignment_gtk_alignment_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAlignment, get_padding,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAlignment, set,                  arginfo_gtk_gtkalignment_set, ZEND_ACC_PUBLIC)
	PHP_ME(GtkAlignment, set_padding,          arginfo_gtk_gtkalignment_set_padding, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1376 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkButton, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 1;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkButton);
    }

    if (text) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "label",
                                              text, "use-underline", use_underline, NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkButton);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 59678 "ext/gtk+/gen_gtk.c"


#line 1411 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkButton, new_from_stock)
{
    GObject *wrapped_obj;
    gchar *stock_id = NULL;
    zend_bool free_stock_id;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &stock_id, &free_stock_id)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkButton);
    }

    wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_class(EG(scope) TSRMLS_CC), "label", stock_id,
                                          "use-stock", TRUE, "use-underline", TRUE, NULL);

    if (free_stock_id) g_free(stock_id);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkButton);
    }

    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


#line 59706 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkButton, clicked)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_button_clicked(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkButton, enter)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_button_enter(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));

}

#line 1437 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkButton, get_alignment)
{
    gfloat xalign, yalign;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    gtk_button_get_alignment(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), &xalign, &yalign);
    //FIXME: bad precision (set 0.1 -> get 0.10000000149012)
    php_gtk_build_value(&return_value, "(ff)", xalign, yalign);
}

#line 59750 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkButton, get_focus_on_click)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_focus_on_click(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkButton, get_image)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_image(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkButton, get_label)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_label(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkButton, get_relief)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_relief(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkButton, get_use_stock)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_use_stock(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkButton, get_use_underline)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_use_underline(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkButton, leave)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_button_leave(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkButton, pressed)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_button_pressed(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkButton, released)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_button_released(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkButton, set_alignment)
{
	double xalign, yalign;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "dd", &xalign, &yalign))
		return;

    gtk_button_set_alignment(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), (float)xalign, (float)yalign);

}


static PHP_METHOD(GtkButton, set_focus_on_click)
{
	zend_bool focus_on_click;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &focus_on_click))
		return;

    gtk_button_set_focus_on_click(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)focus_on_click);

}


static PHP_METHOD(GtkButton, set_image)
{
	zval *image;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &image, gtkwidget_ce))
		return;

    gtk_button_set_image(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(image)));

}


static PHP_METHOD(GtkButton, set_label)
{
	char *label;
	zend_bool free_label = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &label, &free_label))
		return;

    gtk_button_set_label(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), label);
	if (free_label) g_free(label);

}


static PHP_METHOD(GtkButton, set_relief)
{
	GtkReliefStyle newstyle;
	zval *php_newstyle = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_newstyle))
		return;

	if (php_newstyle && phpg_gvalue_get_enum(GTK_TYPE_RELIEF_STYLE, php_newstyle, (gint *)&newstyle) == FAILURE) {
		return;
	}

    gtk_button_set_relief(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), newstyle);

}


static PHP_METHOD(GtkButton, set_use_stock)
{
	zend_bool use_stock;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_stock))
		return;

    gtk_button_set_use_stock(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_stock);

}


static PHP_METHOD(GtkButton, set_use_underline)
{
	zend_bool use_underline;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_underline))
		return;

    gtk_button_set_use_underline(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_underline);

}


static PHP_METHOD(GtkButton, set_image_position)
{
	GtkPositionType position;
	zval *php_position = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_position))
		return;

	if (php_position && phpg_gvalue_get_enum(GTK_TYPE_POSITION_TYPE, php_position, (gint *)&position) == FAILURE) {
		return;
	}

    gtk_button_set_image_position(GTK_BUTTON(PHPG_GOBJECT(this_ptr)), position);

}


static PHP_METHOD(GtkButton, get_image_position)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_image_position(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkButton, get_event_window)
{
	GdkWindow* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_get_event_window(GTK_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkbutton_gtk_button_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_gtk_button_new_from_stock, 0)
    ZEND_ARG_INFO(0, stock_id)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_alignment, 0)
    ZEND_ARG_INFO(0, xalign)
    ZEND_ARG_INFO(0, yalign)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_focus_on_click, 0)
    ZEND_ARG_INFO(0, focus_on_click)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_image, 0)
    ZEND_ARG_OBJ_INFO(0, image, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_label, 0)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_relief, 0)
    ZEND_ARG_INFO(0, newstyle)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_use_stock, 0)
    ZEND_ARG_INFO(0, use_stock)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_use_underline, 0)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbutton_set_image_position, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

static zend_function_entry gtkbutton_methods[] = {
	PHP_ME(GtkButton, __construct,          arginfo_gtk_gtkbutton_gtk_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, new_from_stock,       arginfo_gtk_gtkbutton_gtk_button_new_from_stock, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkButton, clicked,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, enter,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_alignment,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_event_window,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_focus_on_click,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_image,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_image_position,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_label,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_relief,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_use_stock,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, get_use_underline,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, leave,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, pressed,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, released,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_alignment,        arginfo_gtk_gtkbutton_set_alignment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_focus_on_click,   arginfo_gtk_gtkbutton_set_focus_on_click, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_image,            arginfo_gtk_gtkbutton_set_image, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_image_position,   arginfo_gtk_gtkbutton_set_image_position, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_label,            arginfo_gtk_gtkbutton_set_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_relief,           arginfo_gtk_gtkbutton_set_relief, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_use_stock,        arginfo_gtk_gtkbutton_set_use_stock, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButton, set_use_underline,    arginfo_gtk_gtkbutton_set_use_underline, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 2177 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkComboBox, __construct)
{
    GObject *wrapped_obj;
    zval *php_model = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|N", &php_model, gtktreemodel_ce)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBox);
    }

    if (php_model && Z_TYPE_P(php_model) != IS_NULL) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "model",
                                              GTK_TREE_MODEL(PHPG_GOBJECT(php_model)), NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBox);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 60146 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkComboBox, new_text)
{
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBox);
	}

	wrapped_obj = (GObject *) gtk_combo_box_new_text();

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBox);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkComboBox, append_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_combo_box_append_text(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBox, get_active)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_active(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 2202 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkComboBox, get_active_iter)
{
    GtkTreeIter iter;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    if (gtk_combo_box_get_active_iter(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), &iter)) {
        phpg_gboxed_new(&return_value, GTK_TYPE_TREE_ITER, &iter, TRUE, TRUE TSRMLS_CC);
    }
}

#line 60212 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkComboBox, get_active_text)
{
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_active_text(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkComboBox, get_column_span_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_column_span_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkComboBox, get_focus_on_click)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_focus_on_click(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkComboBox, get_model)
{
	GtkTreeModel* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_model(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkComboBox, get_popup_accessible)
{
	AtkObject* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_popup_accessible(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkComboBox, get_row_span_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_row_span_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkComboBox, get_wrap_width)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_wrap_width(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkComboBox, insert_text)
{
	long position;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &position, &text, &free_text))
		return;

    gtk_combo_box_insert_text(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)position, text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBox, popdown)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_combo_box_popdown(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkComboBox, popup)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_combo_box_popup(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkComboBox, prepend_text)
{
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &text, &free_text))
		return;

    gtk_combo_box_prepend_text(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkComboBox, remove_text)
{
	long position;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &position))
		return;

    gtk_combo_box_remove_text(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)position);

}


static PHP_METHOD(GtkComboBox, set_active)
{
	long index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    gtk_combo_box_set_active(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)index);

}


static PHP_METHOD(GtkComboBox, set_active_iter)
{
	GtkTreeIter *iter = NULL;
	zval *php_iter;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_iter, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_iter, GTK_TYPE_TREE_ITER, FALSE TSRMLS_CC)) {
        iter = (GtkTreeIter *) PHPG_GBOXED(php_iter);
    } else {
        php_error(E_WARNING, "%s::%s() expects iter argument to be a valid GtkTreeIter object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_combo_box_set_active_iter(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), iter);

}


static PHP_METHOD(GtkComboBox, set_add_tearoffs)
{
	zend_bool add_tearoffs;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &add_tearoffs))
		return;

    gtk_combo_box_set_add_tearoffs(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gboolean)add_tearoffs);

}


static PHP_METHOD(GtkComboBox, get_add_tearoffs)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_add_tearoffs(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkComboBox, set_column_span_column)
{
	long column_span;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column_span))
		return;

    gtk_combo_box_set_column_span_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)column_span);

}


static PHP_METHOD(GtkComboBox, set_focus_on_click)
{
	zend_bool focus_on_click;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &focus_on_click))
		return;

    gtk_combo_box_set_focus_on_click(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gboolean)focus_on_click);

}


static PHP_METHOD(GtkComboBox, set_model)
{
	zval *model;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &model, gtktreemodel_ce))
		return;

    gtk_combo_box_set_model(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), GTK_TREE_MODEL(PHPG_GOBJECT(model)));

}

#line 2225 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkComboBox, set_row_separator_func)
{
    zval *php_callback, *extra;
    phpg_cb_data_t *cb_data;
    GtkTreeViewRowSeparatorFunc callback;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkTreeViewRowSeparatorFunc)phpg_tree_view_row_separator_func_marshal;
        }
    }

    gtk_combo_box_set_row_separator_func(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)),
                                                           callback,
                                                           cb_data, phpg_cb_data_destroy);
}

#line 60534 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkComboBox, set_row_span_column)
{
	long row_span;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row_span))
		return;

    gtk_combo_box_set_row_span_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)row_span);

}


static PHP_METHOD(GtkComboBox, set_wrap_width)
{
	long width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &width))
		return;

    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)width);

}


static PHP_METHOD(GtkComboBox, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_title(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkComboBox, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_combo_box_set_title(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkComboBox, set_button_sensitivity)
{
	GtkSensitivityType sensitivity;
	zval *php_sensitivity = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sensitivity))
		return;

	if (php_sensitivity && phpg_gvalue_get_enum(GTK_TYPE_SENSITIVITY_TYPE, php_sensitivity, (gint *)&sensitivity) == FAILURE) {
		return;
	}

    gtk_combo_box_set_button_sensitivity(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), sensitivity);

}


static PHP_METHOD(GtkComboBox, get_button_sensitivity)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_button_sensitivity(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkComboBox, get_has_entry)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_has_entry(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkComboBox, set_entry_text_column)
{
	long text_column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &text_column))
		return;

    gtk_combo_box_set_entry_text_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)), (gint)text_column);

}


static PHP_METHOD(GtkComboBox, get_entry_text_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_get_entry_text_column(GTK_COMBO_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcombobox_gtk_combo_box_new, 0, 0, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_append_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_insert_text, 0)
    ZEND_ARG_INFO(0, position)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_prepend_text, 0)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_remove_text, 0)
    ZEND_ARG_INFO(0, position)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_active, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_active_iter, 0)
    ZEND_ARG_OBJ_INFO(0, iter, GtkTreeIter, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_add_tearoffs, 0)
    ZEND_ARG_INFO(0, add_tearoffs)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_column_span_column, 0)
    ZEND_ARG_INFO(0, column_span)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_focus_on_click, 0)
    ZEND_ARG_INFO(0, focus_on_click)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_model, 0)
    ZEND_ARG_INFO(0, model)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_row_separator_func, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_row_span_column, 0)
    ZEND_ARG_INFO(0, row_span)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_wrap_width, 0)
    ZEND_ARG_INFO(0, width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_button_sensitivity, 0)
    ZEND_ARG_INFO(0, sensitivity)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcombobox_set_entry_text_column, 0)
    ZEND_ARG_INFO(0, text_column)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcombobox_methods[] = {
	PHP_ME(GtkComboBox, __construct,          arginfo_gtk_gtkcombobox_gtk_combo_box_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, new_text,             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkComboBox, append_text,          arginfo_gtk_gtkcombobox_append_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_active_iter,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_active_text,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_add_tearoffs,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_button_sensitivity, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_column_span_column, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_entry_text_column, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_focus_on_click,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_has_entry,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_model,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_popup_accessible, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_row_span_column,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, get_wrap_width,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, insert_text,          arginfo_gtk_gtkcombobox_insert_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, popdown,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, popup,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, prepend_text,         arginfo_gtk_gtkcombobox_prepend_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, remove_text,          arginfo_gtk_gtkcombobox_remove_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_active,           arginfo_gtk_gtkcombobox_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_active_iter,      arginfo_gtk_gtkcombobox_set_active_iter, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_add_tearoffs,     arginfo_gtk_gtkcombobox_set_add_tearoffs, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_button_sensitivity, arginfo_gtk_gtkcombobox_set_button_sensitivity, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_column_span_column, arginfo_gtk_gtkcombobox_set_column_span_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_entry_text_column, arginfo_gtk_gtkcombobox_set_entry_text_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_focus_on_click,   arginfo_gtk_gtkcombobox_set_focus_on_click, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_model,            arginfo_gtk_gtkcombobox_set_model, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_row_separator_func, arginfo_gtk_gtkcombobox_set_row_separator_func, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_row_span_column,  arginfo_gtk_gtkcombobox_set_row_span_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_title,            arginfo_gtk_gtkcombobox_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBox, set_wrap_width,       arginfo_gtk_gtkcombobox_set_wrap_width, ZEND_ACC_PUBLIC)

	/***   GtkCellLayout interface implementations   ***/

	PHP_MALIAS(GtkCellLayout, add_attribute, add_attribute, arginfo_gtk_gtkcelllayout_add_attribute, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear, clear, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear_attributes, clear_attributes, arginfo_gtk_gtkcelllayout_clear_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_end, pack_end, arginfo_gtk_gtkcelllayout_pack_end, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_start, pack_start, arginfo_gtk_gtkcelllayout_pack_start, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, reorder, reorder, arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_attributes, set_attributes, arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_cell_data_func, set_cell_data_func, arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, get_cells, get_cells, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 5464 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkToggleButton, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 1;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleButton);
    }

    if (text) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "label",
                                              text, "use-underline", use_underline, NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkToggleButton);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 60846 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkToggleButton, active)
{
	gboolean php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOGGLE_BUTTON(((phpg_gobject_t *)object)->obj)->active;
	RETVAL_BOOL(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkToggleButton, inconsistent)
{
	gboolean php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOGGLE_BUTTON(((phpg_gobject_t *)object)->obj)->inconsistent;
	RETVAL_BOOL(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkToggleButton, draw_indicator)
{
	gboolean php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_TOGGLE_BUTTON(((phpg_gobject_t *)object)->obj)->draw_indicator;
	RETVAL_BOOL(php_retval);
    return SUCCESS;
}


static prop_info_t gtktogglebutton_prop_info[] = {
	{ "active", PHPG_PROP_READ_FN(GtkToggleButton, active), NULL },
	{ "inconsistent", PHPG_PROP_READ_FN(GtkToggleButton, inconsistent), NULL },
	{ "draw_indicator", PHPG_PROP_READ_FN(GtkToggleButton, draw_indicator), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkToggleButton, get_active)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleButton, get_inconsistent)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_button_get_inconsistent(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleButton, get_mode)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_toggle_button_get_mode(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkToggleButton, set_active)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


static PHP_METHOD(GtkToggleButton, set_inconsistent)
{
	zend_bool setting;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &setting))
		return;

    gtk_toggle_button_set_inconsistent(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)setting);

}


static PHP_METHOD(GtkToggleButton, set_mode)
{
	zend_bool draw_indicator;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &draw_indicator))
		return;

    gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)draw_indicator);

}


static PHP_METHOD(GtkToggleButton, set_state)
{
	zend_bool is_active;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &is_active))
		return;

	phpg_warn_deprecated("use GtkToggleButton::set_active()" TSRMLS_CC);

    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)is_active);

}


static PHP_METHOD(GtkToggleButton, toggled)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_toggle_button_toggled(GTK_TOGGLE_BUTTON(PHPG_GOBJECT(this_ptr)));

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtktogglebutton_gtk_toggle_button_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktogglebutton_set_active, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktogglebutton_set_inconsistent, 0)
    ZEND_ARG_INFO(0, setting)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktogglebutton_set_mode, 0)
    ZEND_ARG_INFO(0, draw_indicator)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtktogglebutton_set_state, 0)
    ZEND_ARG_INFO(0, is_active)
ZEND_END_ARG_INFO();

static zend_function_entry gtktogglebutton_methods[] = {
	PHP_ME(GtkToggleButton, __construct,          arginfo_gtk_gtktogglebutton_gtk_toggle_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, get_active,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, get_inconsistent,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, get_mode,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, set_active,           arginfo_gtk_gtktogglebutton_set_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, set_inconsistent,     arginfo_gtk_gtktogglebutton_set_inconsistent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, set_mode,             arginfo_gtk_gtktogglebutton_set_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, set_state,            arginfo_gtk_gtktogglebutton_set_state, ZEND_ACC_PUBLIC)
	PHP_ME(GtkToggleButton, toggled,              NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkOptionMenu, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkOptionMenu);
	}


	phpg_warn_deprecated("use GtkComboBox" TSRMLS_CC);

	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkOptionMenu);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkOptionMenu, get_history)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_option_menu_get_history(GTK_OPTION_MENU(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkOptionMenu, get_menu)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_option_menu_get_menu(GTK_OPTION_MENU(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkOptionMenu, remove_menu)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_option_menu_remove_menu(GTK_OPTION_MENU(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkOptionMenu, set_history)
{
	long index;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &index))
		return;

    gtk_option_menu_set_history(GTK_OPTION_MENU(PHPG_GOBJECT(this_ptr)), (guint)index);

}


static PHP_METHOD(GtkOptionMenu, set_menu)
{
	zval *menu;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &menu, gtkwidget_ce))
		return;

    gtk_option_menu_set_menu(GTK_OPTION_MENU(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(menu)));

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkoptionmenu_set_history, 0)
    ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkoptionmenu_set_menu, 0)
    ZEND_ARG_OBJ_INFO(0, menu, GtkWidget, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkoptionmenu_methods[] = {
	PHP_ME(GtkOptionMenu, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOptionMenu, get_history,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOptionMenu, get_menu,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOptionMenu, remove_menu,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOptionMenu, set_history,          arginfo_gtk_gtkoptionmenu_set_history, ZEND_ACC_PUBLIC)
	PHP_ME(GtkOptionMenu, set_menu,             arginfo_gtk_gtkoptionmenu_set_menu, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1520 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCheckButton, __construct)
{
    GObject *wrapped_obj;
    gchar *text = NULL;
    zend_bool free_text = FALSE, use_underline = 1;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|ub", &text, &free_text, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCheckButton);
    }

    if (text) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "label",
                                              text, "use-underline", use_underline, NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (free_text) g_free(text);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCheckButton);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}
#line 61192 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcheckbutton_gtk_check_button_new, 0, 0, 0)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


static zend_function_entry gtkcheckbutton_methods[] = {
	PHP_ME(GtkCheckButton, __construct,          arginfo_gtk_gtkcheckbutton_gtk_check_button_new, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkFontButton, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontButton);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkFontButton, new_with_font)
{
	char *fontname;
	zend_bool free_fontname = FALSE;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &fontname, &free_fontname)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontButton);
	}

	wrapped_obj = (GObject *) gtk_font_button_new_with_font(fontname);
	if (free_fontname) g_free(fontname);

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkFontButton);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkFontButton, get_font_name)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_font_name(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFontButton, get_show_size)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_show_size(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontButton, get_show_style)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_show_style(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontButton, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_title(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkFontButton, get_use_font)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_use_font(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontButton, get_use_size)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_font_button_get_use_size(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontButton, set_font_name)
{
	char *fontname;
	zend_bool free_fontname = FALSE;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &fontname, &free_fontname))
		return;

    php_retval = gtk_font_button_set_font_name(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), fontname);
	if (free_fontname) g_free(fontname);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkFontButton, set_show_size)
{
	zend_bool show_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_size))
		return;

    gtk_font_button_set_show_size(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)show_size);

}


static PHP_METHOD(GtkFontButton, set_show_style)
{
	zend_bool show_style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_style))
		return;

    gtk_font_button_set_show_style(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)show_style);

}


static PHP_METHOD(GtkFontButton, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_font_button_set_title(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkFontButton, set_use_font)
{
	zend_bool use_font;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_font))
		return;

    gtk_font_button_set_use_font(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_font);

}


static PHP_METHOD(GtkFontButton, set_use_size)
{
	zend_bool use_size;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_size))
		return;

    gtk_font_button_set_use_size(GTK_FONT_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_size);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_gtk_font_button_new_with_font, 0)
    ZEND_ARG_INFO(0, fontname)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_font_name, 0)
    ZEND_ARG_INFO(0, fontname)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_show_size, 0)
    ZEND_ARG_INFO(0, show_size)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_show_style, 0)
    ZEND_ARG_INFO(0, show_style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_use_font, 0)
    ZEND_ARG_INFO(0, use_font)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkfontbutton_set_use_size, 0)
    ZEND_ARG_INFO(0, use_size)
ZEND_END_ARG_INFO();

static zend_function_entry gtkfontbutton_methods[] = {
	PHP_ME(GtkFontButton, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, new_with_font,        arginfo_gtk_gtkfontbutton_gtk_font_button_new_with_font, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkFontButton, get_font_name,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, get_show_size,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, get_show_style,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, get_use_font,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, get_use_size,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_font_name,        arginfo_gtk_gtkfontbutton_set_font_name, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_show_size,        arginfo_gtk_gtkfontbutton_set_show_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_show_style,       arginfo_gtk_gtkfontbutton_set_show_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_title,            arginfo_gtk_gtkfontbutton_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_use_font,         arginfo_gtk_gtkfontbutton_set_use_font, ZEND_ACC_PUBLIC)
	PHP_ME(GtkFontButton, set_use_size,         arginfo_gtk_gtkfontbutton_set_use_size, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 4829 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioButton, __construct)
{
    zval *php_group = NULL;
    char *label = NULL;
    zend_bool free_label, use_underline = 1;
    GtkRadioButton *group = NULL;
    GObject *wrapped_obj;
    GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Nub", &php_group, gtkradiobutton_ce, &label, &free_label, &use_underline)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioButton);
    }

    if (php_group && Z_TYPE_P(php_group) != IS_NULL) {
        group = GTK_RADIO_BUTTON(PHPG_GOBJECT(php_group));
    }

    if (group) {
        if (label) {
            wrapped_obj = (GObject *) g_object_new(gtype, "group", group, "label", label,
                                                   "use-underline", use_underline, NULL);
        } else {
            wrapped_obj = (GObject *) g_object_new(gtype, "group", group, NULL);
        }
    } else {
        if (label) {
            wrapped_obj = (GObject *) g_object_new(gtype, "label", label, "use-underline",
                                                   use_underline, NULL);
        } else {
            wrapped_obj = (GObject *) g_object_new(gtype, NULL);
        }
    }

    if (free_label) g_free(label);

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkRadioButton);
    }
    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 61547 "ext/gtk+/gen_gtk.c"


#line 4872 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioButton, get_group)
{
    GSList *group, *current;
    zval *item;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(PHPG_GOBJECT(this_ptr)));

    array_init(return_value);
    for (current = group; current; current = current->next) {
        MAKE_STD_ZVAL(item);
        phpg_gobject_new(&item, G_OBJECT(current->data) TSRMLS_CC);
        add_next_index_zval(return_value, item);
    }
}

#line 61571 "ext/gtk+/gen_gtk.c"


#line 4913 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioButton, group)
{
    phpg_warn_deprecated("use GtkRadioButton::get_group()" TSRMLS_CC);
#if ZEND_EXTENSION_API_NO > 220051025
    PHP_MN(GtkRadioButton_get_group)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#else
    PHP_FN(GtkRadioButton_get_group)(INTERNAL_FUNCTION_PARAM_PASSTHRU);
#endif
}

#line 61585 "ext/gtk+/gen_gtk.c"


#line 4894 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkRadioButton, set_group)
{
    zval *php_button = NULL;
    GSList *list;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_button, gtkradiobutton_ce))
        return;

    list = gtk_radio_button_get_group(GTK_RADIO_BUTTON(PHPG_GOBJECT(php_button)));
    if (list == gtk_radio_button_get_group(GTK_RADIO_BUTTON(PHPG_GOBJECT(this_ptr)))) {
        return;
    }
    gtk_radio_button_set_group(GTK_RADIO_BUTTON(PHPG_GOBJECT(this_ptr)), list);
}

#line 61606 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkradiobutton_gtk_radio_button_new, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, group, GtkRadioButton, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, use_underline)
ZEND_END_ARG_INFO();


static zend_function_entry gtkradiobutton_methods[] = {
	PHP_ME(GtkRadioButton, __construct,          arginfo_gtk_gtkradiobutton_gtk_radio_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioButton, get_group,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioButton, group,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkRadioButton, set_group,            NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkLinkButton, __construct)
{
	char *uri, *label = NULL;
	zend_bool free_uri = FALSE, free_label = FALSE;
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u|u", &uri, &free_uri, &label, &free_label)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLinkButton);
	}

	wrapped_obj = (GObject *) gtk_link_button_new_with_label(uri, label);
	if (free_uri) g_free(uri);
	if (free_label) g_free(label);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkLinkButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkLinkButton, get_uri)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_link_button_get_uri(GTK_LINK_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkLinkButton, set_uri)
{
	char *uri;
	zend_bool free_uri = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &uri, &free_uri))
		return;

    gtk_link_button_set_uri(GTK_LINK_BUTTON(PHPG_GOBJECT(this_ptr)), uri);
	if (free_uri) g_free(uri);

}

#line 30 "ext/gtk+/gtk-2.10.overrides"
static void phpg_link_button_set_uri_hook_func_marshal(GtkLinkButton *button, const gchar *link, gpointer data)
{
    phpg_cb_data_t *cbd = (phpg_cb_data_t *) data;
    zval *retval = NULL;
    zval ***args = NULL;
    int n_args = 0;
    char *callback_name;
    zval *php_button = NULL, *php_link = NULL;
    gchar *cp_link;
    gsize cp_len;
    zend_bool free_result;

#ifdef ZTS
    TSRMLS_D = cbd->TSRMLS_C;
#endif

    if (!zend_is_callable(cbd->callback, 0, &callback_name PHPGTK_ZEND_IS_CALLABLE)) {
        php_error(E_WARNING, "Unable to invoke callback '%s' specified in %s on line %ld", callback_name, cbd->src_filename, cbd->src_lineno);
        efree(callback_name);
        return;
    }

    phpg_gobject_new(&php_button,       (GObject*)button       TSRMLS_CC);
    cp_link = phpg_from_utf8(link, strlen(link), &cp_len, &free_result TSRMLS_CC);
    if (!cp_link) {
        php_error(E_WARNING, "Could not convert link from UTF-8");
        return;
    }
    MAKE_STD_ZVAL(php_link);
    ZVAL_STRINGL(php_link, (char*)cp_link, cp_len, 1);
    if (free_result) {
        g_free(cp_link);
    }


    args = php_gtk_hash_as_array_offset(cbd->user_args, 2, &n_args);
    args[0] = &php_button;
    args[1] = &php_link;

    call_user_function_ex(EG(function_table), NULL, cbd->callback, &retval, n_args, args, 0, NULL TSRMLS_CC);

    zval_ptr_dtor(&php_button);
    zval_ptr_dtor(&php_link);

    if (retval) {
        zval_ptr_dtor(&retval);
    }

    phpg_handle_marshaller_exception(TSRMLS_C);

    efree(callback_name);
    efree(args);
}

static PHP_METHOD(GtkLinkButton, set_uri_hook)
{
    zval *php_callback = NULL, *extra;
    phpg_cb_data_t *cb_data;
    GtkLinkButtonUriFunc callback;

    if (!php_gtk_parse_varargs(ZEND_NUM_ARGS(), 1, &extra, "V", &php_callback))
        return;

    if (php_callback) {
        if (Z_TYPE_P(php_callback) == IS_NULL) {
            cb_data  = NULL;
            callback = NULL;
        } else {
            zval_add_ref(&php_callback);
            cb_data  = phpg_cb_data_new(php_callback, extra TSRMLS_CC);
            callback = (GtkLinkButtonUriFunc)phpg_link_button_set_uri_hook_func_marshal;
        }
    }

    gtk_link_button_set_uri_hook(callback, cb_data, phpg_cb_data_destroy);
}
#line 61772 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkLinkButton, set_visited)
{
	zend_bool visited;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &visited))
		return;

    gtk_link_button_set_visited(GTK_LINK_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)visited);

}


static PHP_METHOD(GtkLinkButton, get_visited)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_link_button_get_visited(GTK_LINK_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtklinkbutton_link_button_new_with_label, 0, 0, 1)
    ZEND_ARG_INFO(0, uri)
    ZEND_ARG_INFO(0, label)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklinkbutton_set_uri, 0)
    ZEND_ARG_INFO(0, uri)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklinkbutton_set_uri_hook, 0)
    ZEND_ARG_INFO(0, callback)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtklinkbutton_set_visited, 0)
    ZEND_ARG_INFO(0, visited)
ZEND_END_ARG_INFO();

static zend_function_entry gtklinkbutton_methods[] = {
	PHP_ME(GtkLinkButton, __construct,          arginfo_gtk_gtklinkbutton_link_button_new_with_label, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLinkButton, get_uri,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLinkButton, get_visited,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLinkButton, set_uri,              arginfo_gtk_gtklinkbutton_set_uri, ZEND_ACC_PUBLIC)
	PHP_ME(GtkLinkButton, set_uri_hook,         arginfo_gtk_gtklinkbutton_set_uri_hook, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkLinkButton, set_visited,          arginfo_gtk_gtklinkbutton_set_visited, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 576 "ext/gtk+/gtk-2.12.overrides"
static PHP_METHOD(GtkScaleButton, __construct)
{
	gint iconsize;
	gdouble min, max, step;
	GObject *obj;
	zval *iconlist = NULL;
	
	GtkAdjustment *adj;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"iddd|a",&iconsize,&min,&max,&step,&iconlist)) {
		return;
	}

	obj = (GObject *)g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC),0,NULL);
		if(!obj) { PHPG_THROW_CONSTRUCT_EXCEPTION(GtkScaleButton); }
		adj = (GtkAdjustment *)gtk_adjustment_new(min,min,max,step,step,0);
		gtk_scale_button_set_adjustment(GTK_SCALE_BUTTON(obj),GTK_ADJUSTMENT(adj));
	   
	if(iconlist) {
		phpg_scale_button_set_icons_assist(
			GTK_SCALE_BUTTON(obj),
			iconlist TSRMLS_CC
		);
	}

	phpg_gobject_set_wrapper(
		this_ptr,
		G_OBJECT(obj) TSRMLS_CC
	);
}

#line 61868 "ext/gtk+/gen_gtk.c"


#line 611 "ext/gtk+/gtk-2.12.overrides"
static void
phpg_scale_button_set_icons_assist(GtkScaleButton *button, zval *iconlist TSRMLS_DC)
{
	zval **iconname;
	int listlen,iter;
	gchar **list = NULL;

	if(iconlist)
		listlen = zend_hash_num_elements(Z_ARRVAL_P(iconlist));
	else
		listlen = 0;

	if(listlen > 0) {
		list = g_malloc0(sizeof(gchar *) * (listlen + 2));
		/* space for not overwriting the head pointer and leaving a null on
		 the tail end. */
		
		iter = 1;

		/* indexing the array string pointers into a pointer array for the
		   set_icons function. */
		zend_hash_internal_pointer_reset(Z_ARRVAL_P(iconlist));
		while(zend_hash_get_current_data(Z_ARRVAL_P(iconlist),(void **)&iconname) == SUCCESS) {
			if(Z_TYPE_PP(iconname) == IS_STRING) {
				list[iter++] = (gchar *)Z_ARRVAL_PP(iconname);
			} else {
				php_error_docref(
					NULL TSRMLS_CC,
					E_WARNING,
					"each item in icon array is to be a string"
				);
			}
			
			zend_hash_move_forward(Z_ARRVAL_P(iconlist));
		} list[iter] = NULL;
	}
	
	else {
		list = NULL;
	}

	gtk_scale_button_set_icons(
		GTK_SCALE_BUTTON(button),
		(const gchar **)(list+1)
	);

	if(list) {
		g_strfreev(list);
	}
}

static PHP_METHOD(GtkScaleButton, set_icons)
{
	zval *iconlist = NULL;

	NOT_STATIC_METHOD();

	if(!php_gtk_parse_args(ZEND_NUM_ARGS(),"a",&iconlist)) {
		return;
	}

	phpg_scale_button_set_icons_assist(
		GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)),
		iconlist TSRMLS_CC
	);
}

#line 61939 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkScaleButton, get_value)
{
	double php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_button_get_value(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_DOUBLE(php_retval);
}


static PHP_METHOD(GtkScaleButton, set_value)
{
	double value;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "d", &value))
		return;

    gtk_scale_button_set_value(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)), value);

}


static PHP_METHOD(GtkScaleButton, get_adjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_button_get_adjustment(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScaleButton, set_adjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_scale_button_set_adjustment(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkScaleButton, get_plus_button)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_button_get_plus_button(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScaleButton, get_minus_button)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_button_get_minus_button(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkScaleButton, get_popup)
{
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_scale_button_get_popup(GTK_SCALE_BUTTON(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkscalebutton_gtk_scale_button_new, 0, 0, 0)
	ZEND_ARG_INFO(0, iconsize)
	ZEND_ARG_INFO(0, min)
	ZEND_ARG_INFO(0, max)
	ZEND_ARG_INFO(0, step)
	ZEND_ARG_INFO(0, iconlist)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscalebutton_set_value, 0)
    ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkscalebutton_set_adjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

static zend_function_entry gtkscalebutton_methods[] = {
	PHP_ME(GtkScaleButton, __construct,          arginfo_gtk_gtkscalebutton_gtk_scale_button_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, get_adjustment,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, get_minus_button,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, get_plus_button,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, get_popup,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, get_value,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, set_adjustment,       arginfo_gtk_gtkscalebutton_set_adjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, set_icons,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkScaleButton, set_value,            arginfo_gtk_gtkscalebutton_set_value, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkColorButton, __construct)
{
	GParameter params[1];
	zval *php_args[2] = { NULL, };
	char *prop_names[] = { "color",  NULL };
	GObject *wrapped_obj;
	guint i, n_params;
	GType gtype = phpg_gtype_from_zval(this_ptr TSRMLS_CC);

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|^O", &php_args[0], gboxed_ce)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorButton);
	}

	memset(&params, 0, 1 * sizeof(GParameter));
	if (!phpg_parse_ctor_props(gtype, php_args, params, &n_params, prop_names TSRMLS_CC)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorButton);
	}

	wrapped_obj = (GObject *) g_object_newv(gtype, n_params, params);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkColorButton);
	}

	for (i = 0; i < n_params; i++) {
		g_value_unset(&params[i].value);
	}
	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkColorButton, get_alpha)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}

#line 1879 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkColorButton, get_color)
{
    GdkColor color = { 0 , 0 , 0 , 0 };

    NOT_STATIC_METHOD();

    if (ZEND_NUM_ARGS() > 0) {
        php_error(E_WARNING, "%s::%s() requires 0 arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), ZEND_NUM_ARGS());
        return;
    }

    gtk_color_button_get_color(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)), &color);
    phpg_gboxed_new(&return_value, GDK_TYPE_COLOR, &color, TRUE, TRUE TSRMLS_CC);
}

#line 62140 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkColorButton, get_title)
{
	const gchar *php_retval;
	gchar *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_button_get_title(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)));
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        if (free_result)
            g_free(cp_ret);
    } else {
        RETVAL_NULL();
    }
}


static PHP_METHOD(GtkColorButton, get_use_alpha)
{
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_color_button_get_use_alpha(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)));
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkColorButton, set_alpha)
{
	long alpha;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &alpha))
		return;

    gtk_color_button_set_alpha(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)), (guint16)alpha);

}


static PHP_METHOD(GtkColorButton, set_color)
{
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_color_button_set_color(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)), color);

}


static PHP_METHOD(GtkColorButton, set_title)
{
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "u", &title, &free_title))
		return;

    gtk_color_button_set_title(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)), title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkColorButton, set_use_alpha)
{
	zend_bool use_alpha;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_alpha))
		return;

    gtk_color_button_set_use_alpha(GTK_COLOR_BUTTON(PHPG_GOBJECT(this_ptr)), (gboolean)use_alpha);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcolorbutton_color_button_new_with_color, 0, 0, 0)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorbutton_set_alpha, 0)
    ZEND_ARG_INFO(0, alpha)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorbutton_set_color, 0)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorbutton_set_title, 0)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcolorbutton_set_use_alpha, 0)
    ZEND_ARG_INFO(0, use_alpha)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcolorbutton_methods[] = {
	PHP_ME(GtkColorButton, __construct,          arginfo_gtk_gtkcolorbutton_color_button_new_with_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, get_alpha,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, get_color,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, get_title,            NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, get_use_alpha,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, set_alpha,            arginfo_gtk_gtkcolorbutton_set_alpha, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, set_color,            arginfo_gtk_gtkcolorbutton_set_color, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, set_title,            arginfo_gtk_gtkcolorbutton_set_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkColorButton, set_use_alpha,        arginfo_gtk_gtkcolorbutton_set_use_alpha, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVolumeButton, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVolumeButton);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVolumeButton);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static zend_function_entry gtkvolumebutton_methods[] = {
	PHP_ME(GtkVolumeButton, __construct,          NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1604 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, __construct)
{
	GObject *wrapped_obj;
	gint columns, i = 0;
	gchar **titles;
	zval *php_titles = NULL, **temp_title;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i|a", &columns, &php_titles)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCList);
	}

	phpg_warn_deprecated("use GtkListStore/GtkTreeView" TSRMLS_CC);

	if (columns <= 0) {
		PHPG_THROW_EXCEPTION(phpg_construct_exception, "The number of columns is <= 0");
	}

	if (php_titles) {
		if (zend_hash_num_elements(Z_ARRVAL_P(php_titles)) < columns) {
			PHPG_THROW_EXCEPTION(phpg_construct_exception, "The size of the title array does not match the number of columns");
		}

		titles = safe_emalloc(columns, sizeof(gchar *), 0);
		zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_titles));
		while (zend_hash_get_current_data(Z_ARRVAL_P(php_titles), (void **)&temp_title) == SUCCESS) {

			gchar *utf8 = NULL;
			gsize utf8_len = 0;
			zend_bool free_utf8 = 0;

			convert_to_string_ex(temp_title);
			utf8 = phpg_to_utf8(Z_STRVAL_PP(temp_title), Z_STRLEN_PP(temp_title), &utf8_len, &free_utf8 TSRMLS_CC);

			if (!utf8) {
                efree(titles);
                PHPG_THROW_EXCEPTION(phpg_construct_exception, "Could not convert title string to UTF-8");
            }

            if (free_utf8) {
                titles[i++] = utf8;
			} else {
                /* Use GTK+ memory function here since utf8 may also allocated by it */
                titles[i++] = g_strdup(utf8);
            }

			zend_hash_move_forward(Z_ARRVAL_P(php_titles));
		}
		wrapped_obj = (GObject *) gtk_clist_new_with_titles(columns, titles);
        for (--i; i >= 0; i--) {
            g_free(titles[i]);
        }
		efree(titles);

	} else {
		wrapped_obj = (GObject *)gtk_clist_new(columns);
	}

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCList);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 62386 "ext/gtk+/gen_gtk.c"



PHPG_PROP_READER(GtkCList, focus_row)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CLIST(((phpg_gobject_t *)object)->obj)->focus_row;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkCList, rows)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CLIST(((phpg_gobject_t *)object)->obj)->rows;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


PHPG_PROP_READER(GtkCList, columns)
{
	long php_retval;

    if (((phpg_gobject_t *)object)->obj == NULL) {
        return FAILURE;
    }

    php_retval = GTK_CLIST(((phpg_gobject_t *)object)->obj)->columns;
	RETVAL_LONG(php_retval);
    return SUCCESS;
}


static prop_info_t gtkclist_prop_info[] = {
	{ "focus_row", PHPG_PROP_READ_FN(GtkCList, focus_row), NULL },
	{ "rows", PHPG_PROP_READ_FN(GtkCList, rows), NULL },
	{ "columns", PHPG_PROP_READ_FN(GtkCList, columns), NULL },
	{ NULL, NULL, NULL },
};


static PHP_METHOD(GtkCList, clear)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_clear(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}

#line 1672 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, append)
{
    zval *php_strings, **item;
    gchar **text = NULL;
    int size, n = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_strings)) {
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_strings));
    size = zend_hash_num_elements(Z_ARRVAL_P(php_strings));
    text = safe_emalloc(size, sizeof(gchar *), 0);

    while (zend_hash_get_current_data(Z_ARRVAL_P(php_strings),
                                        (void **)&item) == SUCCESS) {

        gchar *utf8 = NULL;
        gsize utf8_len = 0;
        zend_bool free_utf8 = 0;

        convert_to_string_ex(item);
        utf8 = phpg_to_utf8(Z_STRVAL_PP(item), Z_STRLEN_PP(item), &utf8_len, &free_utf8 TSRMLS_CC);

        if (utf8 && free_utf8) {
            SEPARATE_ZVAL(item);
            zval_dtor(*item);
            ZVAL_STRINGL(*item, utf8, utf8_len, 1);
            g_free(utf8);
        }

        text[n++] = Z_STRVAL_PP(item);
        zend_hash_move_forward(Z_ARRVAL_P(php_strings));
    }

    phpg_warn_deprecated(NULL TSRMLS_CC);
    gtk_clist_append(GTK_CLIST(PHPG_GOBJECT(this_ptr)), text);
}

#line 62494 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCList, column_title_active)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_clist_column_title_active(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkCList, column_title_passive)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_clist_column_title_passive(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkCList, column_titles_active)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_column_titles_active(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, column_titles_hide)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_column_titles_hide(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, column_titles_passive)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_column_titles_passive(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, column_titles_show)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_column_titles_show(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, columns_autosize)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clist_columns_autosize(GTK_CLIST(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCList, freeze)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_freeze(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, get_cell_style)
{
	long row, column;
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &column))
		return;

    php_retval = gtk_clist_get_cell_style(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCList, get_cell_type)
{
	long row, column, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &column))
		return;

    php_retval = gtk_clist_get_cell_type(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCList, get_column_title)
{
	long column;
	gchar *php_retval, *cp_ret;
	gsize cp_len;
	zend_bool free_result;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    php_retval = gtk_clist_get_column_title(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);
    if (php_retval) {
        cp_ret = phpg_from_utf8(php_retval, strlen(php_retval), &cp_len, &free_result TSRMLS_CC);
        if (cp_ret) {
            RETVAL_STRINGL((char *)cp_ret, cp_len, 1);
        } else {
            php_error(E_WARNING, "%s::%s(): could not convert return value from UTF-8", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        }
        g_free(php_retval);
        if (free_result)
            g_free(cp_ret);
    } else
        RETVAL_NULL();
}


static PHP_METHOD(GtkCList, get_column_widget)
{
	long column;
	GtkWidget* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    php_retval = gtk_clist_get_column_widget(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCList, get_hadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clist_get_hadjustment(GTK_CLIST(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCList, get_row_style)
{
	long row;
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
		return;

    php_retval = gtk_clist_get_row_style(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCList, get_selectable)
{
	long row;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
		return;

    php_retval = gtk_clist_get_selectable(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row);
	RETVAL_BOOL(php_retval);
}

#line 1830 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, get_selection_info)
{
    gint x, y, row, column;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y)) {
        php_error(E_WARNING, "%s::%s() requires 2 arguments, %d given",
                  get_active_class_name(NULL TSRMLS_CC),
                  get_active_function_name(TSRMLS_C), ZEND_NUM_ARGS());
        return;
    }

    phpg_warn_deprecated(NULL TSRMLS_CC);

    if (gtk_clist_get_selection_info(GTK_CLIST(PHPG_GOBJECT(this_ptr)), x, y, &row, &column) == 0) {
        RETURN_FALSE;
    } else {
        php_gtk_build_value(&return_value, "(ii)", row, column);
    }
}

#line 62743 "ext/gtk+/gen_gtk.c"


#line 1801 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, get_text)
{
    gint row, col;
    gchar *text = NULL;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &col)) {
        return;
    }

    phpg_warn_deprecated(NULL TSRMLS_CC);
    if (gtk_clist_get_text(GTK_CLIST(PHPG_GOBJECT(this_ptr)), row, col, &text) == 0) {
        RETURN_FALSE;
    } else {
        php_gtk_build_value(&return_value, "u", text);
    }
}

#line 62766 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCList, get_vadjustment)
{
	GtkAdjustment* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_clist_get_vadjustment(GTK_CLIST(PHPG_GOBJECT(this_ptr)));
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}

#line 1715 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, insert)
{
    zval *php_strings, **item;
    gchar **text = NULL;
    int size, row, n = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ia", &row, &php_strings)) {
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_strings));
    size = zend_hash_num_elements(Z_ARRVAL_P(php_strings));
    text = safe_emalloc(size, sizeof(gchar *), 0);

    while (zend_hash_get_current_data(Z_ARRVAL_P(php_strings),
                                        (void **)&item) == SUCCESS) {

        gchar *utf8 = NULL;
        gsize utf8_len = 0;
        zend_bool free_utf8 = 0;

        convert_to_string_ex(item);
        utf8 = phpg_to_utf8(Z_STRVAL_PP(item), Z_STRLEN_PP(item), &utf8_len, &free_utf8 TSRMLS_CC);

        if (utf8 && free_utf8) {
            SEPARATE_ZVAL(item);
            zval_dtor(*item);
            ZVAL_STRINGL(*item, utf8, utf8_len, 1);
            g_free(utf8);
        }

        text[n++] = Z_STRVAL_PP(item);
        zend_hash_move_forward(Z_ARRVAL_P(php_strings));
    }

    phpg_warn_deprecated(NULL TSRMLS_CC);
    gtk_clist_insert(GTK_CLIST(PHPG_GOBJECT(this_ptr)), row, text);
}

#line 62825 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCList, moveto)
{
	long row, column;
	double row_align, col_align;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iidd", &row, &column, &row_align, &col_align))
		return;

    gtk_clist_moveto(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, (float)row_align, (float)col_align);

}


static PHP_METHOD(GtkCList, optimal_column_width)
{
	long column, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    php_retval = gtk_clist_optimal_column_width(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);
	RETVAL_LONG(php_retval);
}

#line 1758 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, prepend)
{
    zval *php_strings, **item;
    gchar **text = NULL;
    int size, n = 0;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "a", &php_strings)) {
        return;
    }

    zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_strings));
    size = zend_hash_num_elements(Z_ARRVAL_P(php_strings));
    text = safe_emalloc(size, sizeof(gchar *), 0);

    while (zend_hash_get_current_data(Z_ARRVAL_P(php_strings),
                                        (void **)&item) == SUCCESS) {

        gchar *utf8 = NULL;
        gsize utf8_len = 0;
        zend_bool free_utf8 = 0;

        convert_to_string_ex(item);
        utf8 = phpg_to_utf8(Z_STRVAL_PP(item), Z_STRLEN_PP(item), &utf8_len, &free_utf8 TSRMLS_CC);

        if (utf8 && free_utf8) {
            SEPARATE_ZVAL(item);
            zval_dtor(*item);
            ZVAL_STRINGL(*item, utf8, utf8_len, 1);
            g_free(utf8);
        }

        text[n++] = Z_STRVAL_PP(item);
        zend_hash_move_forward(Z_ARRVAL_P(php_strings));
    }

    phpg_warn_deprecated(NULL TSRMLS_CC);
    gtk_clist_prepend(GTK_CLIST(PHPG_GOBJECT(this_ptr)), text);
}

#line 62899 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCList, row_is_visible)
{
	long row, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
		return;

    php_retval = gtk_clist_row_is_visible(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCList, row_move)
{
	long source_row, dest_row;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &source_row, &dest_row))
		return;

    gtk_clist_row_move(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)source_row, (gint)dest_row);

}


static PHP_METHOD(GtkCList, select_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_select_all(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, select_row)
{
	long row, column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &column))
		return;

    gtk_clist_select_row(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column);

}


static PHP_METHOD(GtkCList, set_auto_sort)
{
	zend_bool auto_sort;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &auto_sort))
		return;

    gtk_clist_set_auto_sort(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gboolean)auto_sort);

}


static PHP_METHOD(GtkCList, set_background)
{
	long row;
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &row, &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_clist_set_background(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, color);

}


static PHP_METHOD(GtkCList, set_button_actions)
{
	long button, button_actions;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &button, &button_actions))
		return;

    gtk_clist_set_button_actions(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (guint)button, (guint8)button_actions);

}


static PHP_METHOD(GtkCList, set_cell_style)
{
	long row, column;
	zval *style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiO", &row, &column, &style, gtkstyle_ce))
		return;

    gtk_clist_set_cell_style(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, GTK_STYLE(PHPG_GOBJECT(style)));

}


static PHP_METHOD(GtkCList, set_column_auto_resize)
{
	long column;
	zend_bool auto_resize;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &column, &auto_resize))
		return;

    gtk_clist_set_column_auto_resize(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gboolean)auto_resize);

}


static PHP_METHOD(GtkCList, set_column_justification)
{
	long column;
	GtkJustification justification;
	zval *php_justification = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iV", &column, &php_justification))
		return;

	if (php_justification && phpg_gvalue_get_enum(GTK_TYPE_JUSTIFICATION, php_justification, (gint *)&justification) == FAILURE) {
		return;
	}

    gtk_clist_set_column_justification(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, justification);

}


static PHP_METHOD(GtkCList, set_column_max_width)
{
	long column, max_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &column, &max_width))
		return;

    gtk_clist_set_column_max_width(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gint)max_width);

}


static PHP_METHOD(GtkCList, set_column_min_width)
{
	long column, min_width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &column, &min_width))
		return;

    gtk_clist_set_column_min_width(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gint)min_width);

}


static PHP_METHOD(GtkCList, set_column_resizeable)
{
	long column;
	zend_bool resizeable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &column, &resizeable))
		return;

    gtk_clist_set_column_resizeable(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gboolean)resizeable);

}


static PHP_METHOD(GtkCList, set_column_title)
{
	long column;
	char *title;
	zend_bool free_title = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iu", &column, &title, &free_title))
		return;

    gtk_clist_set_column_title(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, title);
	if (free_title) g_free(title);

}


static PHP_METHOD(GtkCList, set_column_visibility)
{
	long column;
	zend_bool visible;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &column, &visible))
		return;

    gtk_clist_set_column_visibility(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gboolean)visible);

}


static PHP_METHOD(GtkCList, set_column_widget)
{
	long column;
	zval *widget;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &column, &widget, gtkwidget_ce))
		return;

    gtk_clist_set_column_widget(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, GTK_WIDGET(PHPG_GOBJECT(widget)));

}


static PHP_METHOD(GtkCList, set_column_width)
{
	long column, width;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &column, &width))
		return;

    gtk_clist_set_column_width(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column, (gint)width);

}


static PHP_METHOD(GtkCList, set_foreground)
{
	long row;
	GdkColor *color = NULL;
	zval *php_color;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &row, &php_color, gboxed_ce))
		return;

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_clist_set_foreground(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, color);

}


static PHP_METHOD(GtkCList, set_hadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_clist_set_hadjustment(GTK_CLIST(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkCList, set_pixmap)
{
	long row, column;
	zval *pixmap, *php_mask = NULL;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiO|N", &row, &column, &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;
    if (php_mask) {
        if (Z_TYPE_P(php_mask) == IS_NULL)
            mask = NULL;
        else
            mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));
    }

    gtk_clist_set_pixmap(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);

}


static PHP_METHOD(GtkCList, set_pixtext)
{
	long row, column, spacing;
	char *text;
	zend_bool free_text = FALSE;
	zval *pixmap, *mask;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiuiOO", &row, &column, &text, &free_text, &spacing, &pixmap, gdkpixmap_ce, &mask, gdkpixmap_ce))
		return;

    gtk_clist_set_pixtext(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, text, (guint8)spacing, GDK_PIXMAP(PHPG_GOBJECT(pixmap)), GDK_PIXMAP(PHPG_GOBJECT(mask)));
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkCList, set_reorderable)
{
	zend_bool reorderable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &reorderable))
		return;

    gtk_clist_set_reorderable(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gboolean)reorderable);

}


static PHP_METHOD(GtkCList, set_row_height)
{
	long height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &height))
		return;

    gtk_clist_set_row_height(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (guint)height);

}


static PHP_METHOD(GtkCList, set_row_style)
{
	long row;
	zval *style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iO", &row, &style, gtkstyle_ce))
		return;

    gtk_clist_set_row_style(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, GTK_STYLE(PHPG_GOBJECT(style)));

}


static PHP_METHOD(GtkCList, set_selectable)
{
	long row;
	zend_bool selectable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ib", &row, &selectable))
		return;

    gtk_clist_set_selectable(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gboolean)selectable);

}


static PHP_METHOD(GtkCList, set_selection_mode)
{
	GtkSelectionMode mode;
	zval *php_mode = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_mode))
		return;

	if (php_mode && phpg_gvalue_get_enum(GTK_TYPE_SELECTION_MODE, php_mode, (gint *)&mode) == FAILURE) {
		return;
	}

    gtk_clist_set_selection_mode(GTK_CLIST(PHPG_GOBJECT(this_ptr)), mode);

}


static PHP_METHOD(GtkCList, set_shadow_type)
{
	GtkShadowType type;
	zval *php_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_type))
		return;

	if (php_type && phpg_gvalue_get_enum(GTK_TYPE_SHADOW_TYPE, php_type, (gint *)&type) == FAILURE) {
		return;
	}

    gtk_clist_set_shadow_type(GTK_CLIST(PHPG_GOBJECT(this_ptr)), type);

}


static PHP_METHOD(GtkCList, set_shift)
{
	long row, column, vertical, horizontal;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiii", &row, &column, &vertical, &horizontal))
		return;

    gtk_clist_set_shift(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, (gint)vertical, (gint)horizontal);

}


static PHP_METHOD(GtkCList, set_sort_column)
{
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &column))
		return;

    gtk_clist_set_sort_column(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)column);

}


static PHP_METHOD(GtkCList, set_sort_type)
{
	GtkSortType sort_type;
	zval *php_sort_type = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_sort_type))
		return;

	if (php_sort_type && phpg_gvalue_get_enum(GTK_TYPE_SORT_TYPE, php_sort_type, (gint *)&sort_type) == FAILURE) {
		return;
	}

    gtk_clist_set_sort_type(GTK_CLIST(PHPG_GOBJECT(this_ptr)), sort_type);

}


static PHP_METHOD(GtkCList, set_text)
{
	long row, column;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "iiu", &row, &column, &text, &free_text))
		return;

    gtk_clist_set_text(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column, text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkCList, set_use_drag_icons)
{
	zend_bool use_icons;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &use_icons))
		return;

    gtk_clist_set_use_drag_icons(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gboolean)use_icons);

}


static PHP_METHOD(GtkCList, set_vadjustment)
{
	zval *adjustment;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &adjustment, gtkadjustment_ce))
		return;

    gtk_clist_set_vadjustment(GTK_CLIST(PHPG_GOBJECT(this_ptr)), GTK_ADJUSTMENT(PHPG_GOBJECT(adjustment)));

}


static PHP_METHOD(GtkCList, sort)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_sort(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, swap_rows)
{
	long row1, row2;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row1, &row2))
		return;

    gtk_clist_swap_rows(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row1, (gint)row2);

}


static PHP_METHOD(GtkCList, thaw)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_thaw(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, undo_selection)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_undo_selection(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, unselect_all)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_clist_unselect_all(GTK_CLIST(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCList, unselect_row)
{
	long row, column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &row, &column))
		return;

    gtk_clist_unselect_row(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row, (gint)column);

}

#line 1861 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCList, remove_row)
{
    long row;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
        return;

    gtk_clist_remove(GTK_CLIST(PHPG_GOBJECT(this_ptr)), (gint)row);
}

#line 63522 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkclist_gtk_clist_new, 0, 0, 0)
    ZEND_ARG_INFO(0, columns)
    ZEND_ARG_INFO(0, titles)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_column_title_active, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_column_title_passive, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_cell_style, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_cell_type, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_column_title, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_column_widget, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_row_style, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_selectable, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_get_selection_info, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_moveto, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, row_align)
    ZEND_ARG_INFO(0, col_align)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_optimal_column_width, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_row_is_visible, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_row_move, 0)
    ZEND_ARG_INFO(0, source_row)
    ZEND_ARG_INFO(0, dest_row)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_select_row, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_auto_sort, 0)
    ZEND_ARG_INFO(0, auto_sort)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_background, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_button_actions, 0)
    ZEND_ARG_INFO(0, button)
    ZEND_ARG_INFO(0, button_actions)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_cell_style, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_auto_resize, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, auto_resize)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_justification, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, justification)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_max_width, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, max_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_min_width, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, min_width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_resizeable, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, resizeable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_title, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, title)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_visibility, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, visible)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_widget, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_OBJ_INFO(0, widget, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_column_width, 0)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, width)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_foreground, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_hadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkclist_set_pixmap, 0, 0, 3)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_pixtext, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, spacing)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_reorderable, 0)
    ZEND_ARG_INFO(0, reorderable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_row_height, 0)
    ZEND_ARG_INFO(0, height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_row_style, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_selectable, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, selectable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_selection_mode, 0)
    ZEND_ARG_INFO(0, mode)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_shadow_type, 0)
    ZEND_ARG_INFO(0, type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_shift, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, vertical)
    ZEND_ARG_INFO(0, horizontal)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_sort_column, 0)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_sort_type, 0)
    ZEND_ARG_INFO(0, sort_type)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_text, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_use_drag_icons, 0)
    ZEND_ARG_INFO(0, use_icons)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_set_vadjustment, 0)
    ZEND_ARG_OBJ_INFO(0, adjustment, GtkAdjustment, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_swap_rows, 0)
    ZEND_ARG_INFO(0, row1)
    ZEND_ARG_INFO(0, row2)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_unselect_row, 0)
    ZEND_ARG_INFO(0, row)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkclist_remove_row, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();


static zend_function_entry gtkclist_methods[] = {
	PHP_ME(GtkCList, __construct,          arginfo_gtk_gtkclist_gtk_clist_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, append,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, clear,                NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_title_active,  arginfo_gtk_gtkclist_column_title_active, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_title_passive, arginfo_gtk_gtkclist_column_title_passive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_titles_active, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_titles_hide,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_titles_passive, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, column_titles_show,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, columns_autosize,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, freeze,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_cell_style,       arginfo_gtk_gtkclist_get_cell_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_cell_type,        arginfo_gtk_gtkclist_get_cell_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_column_title,     arginfo_gtk_gtkclist_get_column_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_column_widget,    arginfo_gtk_gtkclist_get_column_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_hadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_row_style,        arginfo_gtk_gtkclist_get_row_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_selectable,       arginfo_gtk_gtkclist_get_selectable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_selection_info,   arginfo_gtk_gtkclist_get_selection_info, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_text,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, get_vadjustment,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, insert,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, moveto,               arginfo_gtk_gtkclist_moveto, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, optimal_column_width, arginfo_gtk_gtkclist_optimal_column_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, prepend,              NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, remove_row,           arginfo_gtk_gtkclist_remove_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, row_is_visible,       arginfo_gtk_gtkclist_row_is_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, row_move,             arginfo_gtk_gtkclist_row_move, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, select_all,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, select_row,           arginfo_gtk_gtkclist_select_row, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_auto_sort,        arginfo_gtk_gtkclist_set_auto_sort, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_background,       arginfo_gtk_gtkclist_set_background, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_button_actions,   arginfo_gtk_gtkclist_set_button_actions, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_cell_style,       arginfo_gtk_gtkclist_set_cell_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_auto_resize, arginfo_gtk_gtkclist_set_column_auto_resize, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_justification, arginfo_gtk_gtkclist_set_column_justification, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_max_width, arginfo_gtk_gtkclist_set_column_max_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_min_width, arginfo_gtk_gtkclist_set_column_min_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_resizeable, arginfo_gtk_gtkclist_set_column_resizeable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_title,     arginfo_gtk_gtkclist_set_column_title, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_visibility, arginfo_gtk_gtkclist_set_column_visibility, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_widget,    arginfo_gtk_gtkclist_set_column_widget, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_column_width,     arginfo_gtk_gtkclist_set_column_width, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_foreground,       arginfo_gtk_gtkclist_set_foreground, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_hadjustment,      arginfo_gtk_gtkclist_set_hadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_pixmap,           arginfo_gtk_gtkclist_set_pixmap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_pixtext,          arginfo_gtk_gtkclist_set_pixtext, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_reorderable,      arginfo_gtk_gtkclist_set_reorderable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_row_height,       arginfo_gtk_gtkclist_set_row_height, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_row_style,        arginfo_gtk_gtkclist_set_row_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_selectable,       arginfo_gtk_gtkclist_set_selectable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_selection_mode,   arginfo_gtk_gtkclist_set_selection_mode, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_shadow_type,      arginfo_gtk_gtkclist_set_shadow_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_shift,            arginfo_gtk_gtkclist_set_shift, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_sort_column,      arginfo_gtk_gtkclist_set_sort_column, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_sort_type,        arginfo_gtk_gtkclist_set_sort_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_text,             arginfo_gtk_gtkclist_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_use_drag_icons,   arginfo_gtk_gtkclist_set_use_drag_icons, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, set_vadjustment,      arginfo_gtk_gtkclist_set_vadjustment, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, sort,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, swap_rows,            arginfo_gtk_gtkclist_swap_rows, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, thaw,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, undo_selection,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, unselect_all,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCList, unselect_row,         arginfo_gtk_gtkclist_unselect_row, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 2266 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkComboBoxEntry, __construct)
{
	zval *model;
	long text_column = 0;
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "|Oi", &model, gtktreemodel_ce, &text_column)) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBoxEntry);
	}

    if (model && Z_TYPE_P(model) != IS_NULL) {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), "model",
                                              GTK_TREE_MODEL(PHPG_GOBJECT(model)),
                                              "text-column", text_column, NULL);
    } else {
        wrapped_obj = (GObject *)g_object_new(phpg_gtype_from_zval(this_ptr TSRMLS_CC), NULL);
    }

    if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBox);
    }

    phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 63895 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkComboBoxEntry, new_text)
{
	GObject *wrapped_obj;

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBoxEntry);
	}

	wrapped_obj = (GObject *) gtk_combo_box_entry_new_text();

	if (!wrapped_obj) {
        PHPG_THROW_CONSTRUCT_EXCEPTION(GtkComboBoxEntry);
	}
    phpg_gobject_new(&return_value, wrapped_obj TSRMLS_CC);
    g_object_unref(wrapped_obj); /* phpg_gobject_new() increments reference count */
}


static PHP_METHOD(GtkComboBoxEntry, get_text_column)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_combo_box_entry_get_text_column(GTK_COMBO_BOX_ENTRY(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkComboBoxEntry, set_text_column)
{
	long text_column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &text_column))
		return;

    gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(PHPG_GOBJECT(this_ptr)), (gint)text_column);

}

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkcomboboxentry_gtk_combo_box_entry_new, 0, 0, 0)
    ZEND_ARG_INFO(0, model)
    ZEND_ARG_INFO(0, text_column)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcomboboxentry_set_text_column, 0)
    ZEND_ARG_INFO(0, text_column)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcomboboxentry_methods[] = {
	PHP_ME(GtkComboBoxEntry, __construct,          arginfo_gtk_gtkcomboboxentry_gtk_combo_box_entry_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxEntry, new_text,             NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkComboBoxEntry, get_text_column,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkComboBoxEntry, set_text_column,      arginfo_gtk_gtkcomboboxentry_set_text_column, ZEND_ACC_PUBLIC)

	/***   GtkCellLayout interface implementations   ***/

	PHP_MALIAS(GtkCellLayout, add_attribute, add_attribute, arginfo_gtk_gtkcelllayout_add_attribute, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear, clear, NULL, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, clear_attributes, clear_attributes, arginfo_gtk_gtkcelllayout_clear_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_end, pack_end, arginfo_gtk_gtkcelllayout_pack_end, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, pack_start, pack_start, arginfo_gtk_gtkcelllayout_pack_start, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, reorder, reorder, arginfo_gtk_gtkcelllayout_reorder, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_attributes, set_attributes, arginfo_gtk_gtkcelllayout_set_attributes, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, set_cell_data_func, set_cell_data_func, arginfo_gtk_gtkcelllayout_set_cell_data_func, ZEND_ACC_PUBLIC)
	PHP_MALIAS(GtkCellLayout, get_cells, get_cells, NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
#line 1474 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkButtonBox, get_child_ipadding)
{
    gint ipad_x, ipad_y;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    phpg_warn_deprecated("use style properties \"child-internal-pad-x/-y\" instead" TSRMLS_CC);

    gtk_button_box_get_child_size(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), &ipad_x, &ipad_y);
    php_gtk_build_value(&return_value, "(ii)", ipad_x, ipad_y);
}

#line 63991 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkButtonBox, get_child_secondary)
{
	zval *child;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &child, gtkwidget_ce))
		return;

    php_retval = gtk_button_box_get_child_secondary(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)));
	RETVAL_BOOL(php_retval);
}

#line 1457 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkButtonBox, get_child_size)
{
    gint min_width, min_height;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
        return;

    phpg_warn_deprecated("use style properties \"child-min-width/-height\" instead" TSRMLS_CC);

    gtk_button_box_get_child_size(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), &min_width, &min_height);
    php_gtk_build_value(&return_value, "(ii)", min_width, min_height);
}

#line 64025 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkButtonBox, get_layout)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_button_box_get_layout(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkButtonBox, set_child_ipadding)
{
	long ipad_x, ipad_y;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &ipad_x, &ipad_y))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_button_box_set_child_ipadding(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), (gint)ipad_x, (gint)ipad_y);

}


static PHP_METHOD(GtkButtonBox, set_child_secondary)
{
	zval *child;
	zend_bool is_secondary;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &child, gtkwidget_ce, &is_secondary))
		return;

    gtk_button_box_set_child_secondary(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), GTK_WIDGET(PHPG_GOBJECT(child)), (gboolean)is_secondary);

}


static PHP_METHOD(GtkButtonBox, set_child_size)
{
	long min_width, min_height;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &min_width, &min_height))
		return;

	phpg_warn_deprecated("don't use this" TSRMLS_CC);

    gtk_button_box_set_child_size(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), (gint)min_width, (gint)min_height);

}


static PHP_METHOD(GtkButtonBox, set_layout)
{
	GtkButtonBoxStyle layout_style;
	zval *php_layout_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_layout_style))
		return;

	if (php_layout_style && phpg_gvalue_get_enum(GTK_TYPE_BUTTON_BOX_STYLE, php_layout_style, (gint *)&layout_style) == FAILURE) {
		return;
	}

    gtk_button_box_set_layout(GTK_BUTTON_BOX(PHPG_GOBJECT(this_ptr)), layout_style);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuttonbox_get_child_secondary, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuttonbox_set_child_ipadding, 0)
    ZEND_ARG_INFO(0, ipad_x)
    ZEND_ARG_INFO(0, ipad_y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuttonbox_set_child_secondary, 0)
    ZEND_ARG_OBJ_INFO(0, child, GtkWidget, 1)
    ZEND_ARG_INFO(0, is_secondary)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuttonbox_set_child_size, 0)
    ZEND_ARG_INFO(0, min_width)
    ZEND_ARG_INFO(0, min_height)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkbuttonbox_set_layout, 0)
    ZEND_ARG_INFO(0, layout_style)
ZEND_END_ARG_INFO();

static zend_function_entry gtkbuttonbox_methods[] = {
	PHP_ME(GtkButtonBox, get_child_ipadding,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, get_child_secondary,  arginfo_gtk_gtkbuttonbox_get_child_secondary, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, get_child_size,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, get_layout,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, set_child_ipadding,   arginfo_gtk_gtkbuttonbox_set_child_ipadding, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, set_child_secondary,  arginfo_gtk_gtkbuttonbox_set_child_secondary, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, set_child_size,       arginfo_gtk_gtkbuttonbox_set_child_size, ZEND_ACC_PUBLIC)
	PHP_ME(GtkButtonBox, set_layout,           arginfo_gtk_gtkbuttonbox_set_layout, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkCalendar, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCalendar);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCalendar);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkCalendar, clear_marks)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_calendar_clear_marks(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCalendar, display_options)
{
	GtkCalendarDisplayOptions flags;
	zval *php_flags = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_flags))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_CALENDAR_DISPLAY_OPTIONS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    gtk_calendar_display_options(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), flags);

}


static PHP_METHOD(GtkCalendar, freeze)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_calendar_freeze(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));

}

#line 1494 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCalendar, get_date)
{
    guint year, month, day;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
        return;
    }

    gtk_calendar_get_date(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), &year, &month, &day);
    php_gtk_build_value(&return_value, "(iii)", year, month, day);
}
#line 64231 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCalendar, get_display_options)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_calendar_get_display_options(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCalendar, mark_day)
{
	long day;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &day))
		return;

    php_retval = gtk_calendar_mark_day(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (guint)day);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCalendar, select_day)
{
	long day;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &day))
		return;

    gtk_calendar_select_day(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (guint)day);

}


static PHP_METHOD(GtkCalendar, select_month)
{
	long month, year;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &month, &year))
		return;

    php_retval = gtk_calendar_select_month(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (guint)month, (guint)year);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCalendar, set_display_options)
{
	GtkCalendarDisplayOptions flags;
	zval *php_flags = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_flags))
		return;

	if (php_flags && phpg_gvalue_get_flags(GTK_TYPE_CALENDAR_DISPLAY_OPTIONS, php_flags, (gint *)&flags) == FAILURE) {
		return;
	}

    gtk_calendar_set_display_options(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), flags);

}


static PHP_METHOD(GtkCalendar, thaw)
{

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    gtk_calendar_thaw(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));

}


static PHP_METHOD(GtkCalendar, unmark_day)
{
	long day;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &day))
		return;

    php_retval = gtk_calendar_unmark_day(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (guint)day);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCalendar, set_detail_width_chars)
{
	long chars;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &chars))
		return;

    gtk_calendar_set_detail_width_chars(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (gint)chars);

}


static PHP_METHOD(GtkCalendar, set_detail_height_rows)
{
	long rows;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &rows))
		return;

    gtk_calendar_set_detail_height_rows(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)), (gint)rows);

}


static PHP_METHOD(GtkCalendar, get_detail_width_chars)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_calendar_get_detail_width_chars(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCalendar, get_detail_height_rows)
{
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_calendar_get_detail_height_rows(GTK_CALENDAR(PHPG_GOBJECT(this_ptr)));
	RETVAL_LONG(php_retval);
}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_display_options, 0)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_mark_day, 0)
    ZEND_ARG_INFO(0, day)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_select_day, 0)
    ZEND_ARG_INFO(0, day)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_select_month, 0)
    ZEND_ARG_INFO(0, month)
    ZEND_ARG_INFO(0, year)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_set_display_options, 0)
    ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_unmark_day, 0)
    ZEND_ARG_INFO(0, day)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_set_detail_width_chars, 0)
    ZEND_ARG_INFO(0, chars)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkcalendar_set_detail_height_rows, 0)
    ZEND_ARG_INFO(0, rows)
ZEND_END_ARG_INFO();

static zend_function_entry gtkcalendar_methods[] = {
	PHP_ME(GtkCalendar, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, clear_marks,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, display_options,      arginfo_gtk_gtkcalendar_display_options, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, freeze,               NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, get_date,             NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, get_detail_height_rows, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, get_detail_width_chars, NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, get_display_options,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, mark_day,             arginfo_gtk_gtkcalendar_mark_day, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, select_day,           arginfo_gtk_gtkcalendar_select_day, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, select_month,         arginfo_gtk_gtkcalendar_select_month, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, set_detail_height_rows, arginfo_gtk_gtkcalendar_set_detail_height_rows, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, set_detail_width_chars, arginfo_gtk_gtkcalendar_set_detail_width_chars, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, set_display_options,  arginfo_gtk_gtkcalendar_set_display_options, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, thaw,                 NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCalendar, unmark_day,           arginfo_gtk_gtkcalendar_unmark_day, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkVButtonBox, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVButtonBox);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkVButtonBox);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkVButtonBox, get_layout_default)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_vbutton_box_get_layout_default();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkVButtonBox, get_spacing_default)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_vbutton_box_get_spacing_default();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkVButtonBox, set_layout_default)
{
	GtkButtonBoxStyle layout;
	zval *php_layout = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_layout))
		return;

	if (php_layout && phpg_gvalue_get_enum(GTK_TYPE_BUTTON_BOX_STYLE, php_layout, (gint *)&layout) == FAILURE) {
		return;
	}

    gtk_vbutton_box_set_layout_default(layout);

}


static PHP_METHOD(GtkVButtonBox, set_spacing_default)
{
	long spacing;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_vbutton_box_set_spacing_default((gint)spacing);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkvbuttonbox_set_layout_default, 0)
    ZEND_ARG_INFO(0, layout)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkvbuttonbox_set_spacing_default, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtkvbuttonbox_methods[] = {
	PHP_ME(GtkVButtonBox, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkVButtonBox, get_layout_default,   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkVButtonBox, get_spacing_default,  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkVButtonBox, set_layout_default,   arginfo_gtk_gtkvbuttonbox_set_layout_default, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkVButtonBox, set_spacing_default,  arginfo_gtk_gtkvbuttonbox_set_spacing_default, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};

static PHP_METHOD(GtkHButtonBox, __construct)
{
	GObject *wrapped_obj;

	phpg_gobject_t *pobj = zend_object_store_get_object(this_ptr TSRMLS_CC);
	if (pobj->obj != NULL) {
		return;
	}

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "")) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHButtonBox);
	}


	wrapped_obj = (GObject *) g_object_newv(phpg_gtype_from_zval(this_ptr TSRMLS_CC), 0, NULL);

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkHButtonBox);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}


static PHP_METHOD(GtkHButtonBox, get_layout_default)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_hbutton_box_get_layout_default();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkHButtonBox, get_spacing_default)
{
	long php_retval;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), ""))
		return;

    php_retval = gtk_hbutton_box_get_spacing_default();
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkHButtonBox, set_layout_default)
{
	GtkButtonBoxStyle layout;
	zval *php_layout = NULL;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_layout))
		return;

	if (php_layout && phpg_gvalue_get_enum(GTK_TYPE_BUTTON_BOX_STYLE, php_layout, (gint *)&layout) == FAILURE) {
		return;
	}

    gtk_hbutton_box_set_layout_default(layout);

}


static PHP_METHOD(GtkHButtonBox, set_spacing_default)
{
	long spacing;

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_hbutton_box_set_spacing_default((gint)spacing);

}


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhbuttonbox_set_layout_default, 0)
    ZEND_ARG_INFO(0, layout)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkhbuttonbox_set_spacing_default, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

static zend_function_entry gtkhbuttonbox_methods[] = {
	PHP_ME(GtkHButtonBox, __construct,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkHButtonBox, get_layout_default,   NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkHButtonBox, get_spacing_default,  NULL, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkHButtonBox, set_layout_default,   arginfo_gtk_gtkhbuttonbox_set_layout_default, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	PHP_ME(GtkHButtonBox, set_spacing_default,  arginfo_gtk_gtkhbuttonbox_set_spacing_default, ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
	{ NULL, NULL, NULL }
};
#line 2306 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCTree, __construct)
{
	GObject *wrapped_obj;
	gint columns, tree_column, i = 0;
	gchar **titles;
	zval *php_titles = NULL, **temp_title;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii|a", &columns, &tree_column, &php_titles)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCTree);
	}

	phpg_warn_deprecated("use GtkTreeStore/GtkTreeView" TSRMLS_CC);

	if (columns <= 0) {
		php_error(E_WARNING, "%s::%s() requires the number of columns to be > 0",
			get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
			PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCTree);
	}

	if (php_titles) {
		if (zend_hash_num_elements(Z_ARRVAL_P(php_titles)) < columns) {
			php_error(E_WARNING, "%s::%s(): the size of the title array needs to match the number of columns",
			get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
			PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCTree);
		}

		titles = safe_emalloc(columns, sizeof(gchar *), 0);
		zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_titles));
		while (zend_hash_get_current_data(Z_ARRVAL_P(php_titles), (void **)&temp_title) == SUCCESS) {

			gchar *utf8 = NULL;
			gsize utf8_len = 0;
			zend_bool free_utf8 = 0;

			convert_to_string_ex(temp_title);
			utf8 = phpg_to_utf8(Z_STRVAL_PP(temp_title), Z_STRLEN_PP(temp_title), &utf8_len, &free_utf8 TSRMLS_CC);

			if (!utf8) {
                efree(titles);
                PHPG_THROW_EXCEPTION(phpg_construct_exception, "Could not convert title string to UTF-8");
            }

            if (free_utf8) {
                titles[i++] = utf8;
			} else {
                /* Use GTK+ memory function here since utf8 may also allocated by it */
                titles[i++] = g_strdup(utf8);
            }

			zend_hash_move_forward(Z_ARRVAL_P(php_titles));
		}

		wrapped_obj = (GObject *) gtk_ctree_new_with_titles(columns, tree_column, titles);
        for (--i; i >= 0; i--) {
            g_free(titles[i]);
        }
		efree(titles);
	} else {
		wrapped_obj = (GObject *)gtk_ctree_new(columns, tree_column);
	}

	if (!wrapped_obj) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCTree);
	}

	phpg_gobject_set_wrapper(this_ptr, wrapped_obj TSRMLS_CC);
}

#line 64719 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCTree, collapse)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_collapse(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, collapse_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_collapse_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, collapse_to_depth)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long depth;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_node, gpointer_ce, &depth))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_collapse_to_depth(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)depth);

}


static PHP_METHOD(GtkCTree, expand)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_expand(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, expand_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_expand_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, expand_to_depth)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long depth;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_node, gpointer_ce, &depth))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_expand_to_depth(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)depth);

}


static PHP_METHOD(GtkCTree, find)
{
	GtkCTreeNode *node = NULL, *child = NULL;
	zval *php_node, *php_child;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_node, gpointer_ce, &php_child, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gpointer_check(php_child, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        child = (GtkCTreeNode *) PHPG_GPOINTER(php_child);
    } else {
        php_error(E_WARNING, "%s::%s() expects child argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_find(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, child);
	RETVAL_BOOL(php_retval);
}

#line 2378 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCTree, insert_node)
{
	zval *php_parent, *php_sibling, *php_text, **temp_text;
	zval *php_pixmap_closed = NULL, *php_pixmap_opened = NULL,
		 *php_mask_closed = NULL, *php_mask_opened = NULL;
	gint columns, spacing = 5, i = 0;
	zend_bool is_leaf = 0, expanded = 0;
	GtkCTreeNode *parent = NULL, *sibling = NULL, *php_retval;
	GdkPixmap *pixmap_closed = NULL, *pixmap_opened = NULL;
	GdkBitmap *mask_closed = NULL, *mask_opened = NULL;
	GtkCTree *tree;
	gchar **text = NULL;
    int array_size;

	NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "NNa|iNNNNbb",  &php_parent, gtkctreenode_ce,
															&php_sibling, gtkctreenode_ce,
															&php_text, &spacing,
															&php_pixmap_closed, gdkpixmap_ce,
															&php_mask_closed, gdkpixmap_ce,
															&php_pixmap_opened, gdkpixmap_ce,
															&php_mask_opened, gdkpixmap_ce,
															&is_leaf, &expanded)) {
		PHPG_THROW_CONSTRUCT_EXCEPTION(GtkCTreeNode);
	}

	tree = GTK_CTREE(PHPG_GOBJECT(this_ptr));
	columns = GTK_CLIST(tree)->columns;
    array_size = zend_hash_num_elements(Z_ARRVAL_P(php_text));
	if (array_size != columns) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "the text array size (%d) does not match the number of columns in the ctree (%d)", array_size, columns);
		return;
	}

	if (Z_TYPE_P(php_parent) != IS_NULL)
		parent			= GTK_CTREE_NODE(PHPG_GPOINTER(php_parent));
	if (Z_TYPE_P(php_sibling) != IS_NULL)
		sibling			= GTK_CTREE_NODE(PHPG_GPOINTER(php_sibling));
	if (php_pixmap_closed && Z_TYPE_P(php_pixmap_closed) != IS_NULL)
		pixmap_closed	= GDK_PIXMAP(PHPG_GOBJECT(php_pixmap_closed));
	if (php_mask_closed && Z_TYPE_P(php_mask_closed) != IS_NULL)
		mask_closed		= GDK_PIXMAP(PHPG_GOBJECT(php_mask_closed));
	if (php_pixmap_opened && Z_TYPE_P(php_pixmap_opened) != IS_NULL)
		pixmap_opened	= GDK_PIXMAP(PHPG_GOBJECT(php_pixmap_opened));
	if (php_mask_opened && Z_TYPE_P(php_mask_opened) != IS_NULL)
		mask_opened		= GDK_PIXMAP(PHPG_GOBJECT(php_mask_opened));

	text = safe_emalloc(columns, sizeof(gchar *), 0);
	zend_hash_internal_pointer_reset(Z_ARRVAL_P(php_text));
	while (zend_hash_get_current_data(Z_ARRVAL_P(php_text), (void **)&temp_text) == SUCCESS) {

		gchar *utf8 = NULL;
		gsize utf8_len = 0;
		zend_bool free_utf8 = 0;

		convert_to_string_ex(temp_text);
		utf8 = phpg_to_utf8(Z_STRVAL_PP(temp_text), Z_STRLEN_PP(temp_text), &utf8_len, &free_utf8 TSRMLS_CC);

        if (!utf8) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING, "Could not convert text string to UTF-8");
            efree(text);
            return;
        }

        if (free_utf8) {
            text[i++] = utf8;
        } else {
            /* Use GTK+ memory function here since utf8 may also allocated by it */
            text[i++] = g_strdup(utf8);
        }

		zend_hash_move_forward(Z_ARRVAL_P(php_text));
	}

	php_retval = gtk_ctree_insert_node(tree, parent, sibling, text, (guint8)spacing,
								pixmap_closed, mask_closed, pixmap_opened,
								mask_opened, is_leaf, expanded);

    for (--i; i >= 0; i--) {
        g_free(text[i]);
    }
	efree(text);
	phpg_gpointer_new(&return_value, GTK_TYPE_CTREE_NODE, php_retval TSRMLS_CC);
}

#line 64973 "ext/gtk+/gen_gtk.c"



static PHP_METHOD(GtkCTree, is_hot_spot)
{
	long x, y;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "ii", &x, &y))
		return;

    php_retval = gtk_ctree_is_hot_spot(GTK_CTREE(PHPG_GOBJECT(this_ptr)), (gint)x, (gint)y);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCTree, is_viewable)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_is_viewable(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCTree, last)
{
	GtkCTreeNode *node = NULL, *php_retval;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_last(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);
	phpg_gpointer_new(&return_value, GTK_TYPE_CTREE_NODE, php_retval TSRMLS_CC);

}


static PHP_METHOD(GtkCTree, move)
{
	GtkCTreeNode *node = NULL, *new_parent = NULL, *new_sibling = NULL;
	zval *php_node, *php_new_parent = NULL, *php_new_sibling = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OOO", &php_node, gpointer_ce, &php_new_parent, gpointer_ce, &php_new_sibling, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (Z_TYPE_P(php_new_parent) != IS_NULL) {
        if (phpg_gpointer_check(php_new_parent, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
            new_parent = (GtkCTreeNode *) PHPG_GPOINTER(php_new_parent);
        } else {
            php_error(E_WARNING, "%s::%s() expects new_parent argument to be a valid GtkCTreeNode object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    if (Z_TYPE_P(php_new_sibling) != IS_NULL) {
        if (phpg_gpointer_check(php_new_sibling, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
            new_sibling = (GtkCTreeNode *) PHPG_GPOINTER(php_new_sibling);
        } else {
            php_error(E_WARNING, "%s::%s() expects new_sibling argument to be a valid GtkCTreeNode object or null", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
            return;
        }
    }

    gtk_ctree_move(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, new_parent, new_sibling);

}


static PHP_METHOD(GtkCTree, node_get_cell_style)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long column;
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_node, gpointer_ce, &column))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_node_get_cell_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCTree, node_get_cell_type)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long column, php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_node, gpointer_ce, &column))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_node_get_cell_type(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCTree, node_get_row_style)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	GtkStyle* php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_node_get_row_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);
    phpg_gobject_new(&return_value, (GObject *)php_retval TSRMLS_CC);
}


static PHP_METHOD(GtkCTree, node_get_selectable)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	gboolean php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_node_get_selectable(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);
	RETVAL_BOOL(php_retval);
}


static PHP_METHOD(GtkCTree, node_is_visible)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_node_is_visible(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);
	RETVAL_LONG(php_retval);
}


static PHP_METHOD(GtkCTree, node_moveto)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long column;
	double row_align, col_align;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oidd", &php_node, gpointer_ce, &column, &row_align, &col_align))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_moveto(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, (float)row_align, (float)col_align);

}


static PHP_METHOD(GtkCTree, node_nth)
{
	long row;
	GtkCTreeNode *php_retval;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &row))
		return;

    php_retval = gtk_ctree_node_nth(GTK_CTREE(PHPG_GOBJECT(this_ptr)), (guint)row);
	phpg_gpointer_new(&return_value, GTK_TYPE_CTREE_NODE, php_retval TSRMLS_CC);

}


static PHP_METHOD(GtkCTree, node_set_background)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_node, gpointer_ce, &php_color, gboxed_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_background(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, color);

}


static PHP_METHOD(GtkCTree, node_set_cell_style)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *style;
	long column;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiO", &php_node, gpointer_ce, &column, &style, gtkstyle_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_cell_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, GTK_STYLE(PHPG_GOBJECT(style)));

}


static PHP_METHOD(GtkCTree, node_set_foreground)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *php_color;
	GdkColor *color = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_node, gpointer_ce, &php_color, gboxed_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gboxed_check(php_color, GDK_TYPE_COLOR, FALSE TSRMLS_CC)) {
        color = (GdkColor *) PHPG_GBOXED(php_color);
    } else {
        php_error(E_WARNING, "%s::%s() expects color argument to be a valid GdkColor object",
                  get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_foreground(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, color);

}


static PHP_METHOD(GtkCTree, node_set_pixmap)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *pixmap, *php_mask;
	long column;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiON", &php_node, gpointer_ce, &column, &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

    gtk_ctree_node_set_pixmap(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);

}


static PHP_METHOD(GtkCTree, node_set_pixtext)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *pixmap, *php_mask;
	long column, spacing;
	char *text;
	zend_bool free_text = FALSE;
	GdkPixmap *mask = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OiuiON", &php_node, gpointer_ce, &column, &text, &free_text, &spacing, &pixmap, gdkpixmap_ce, &php_mask, gdkpixmap_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (Z_TYPE_P(php_mask) != IS_NULL)
        mask = GDK_PIXMAP(PHPG_GOBJECT(php_mask));

    gtk_ctree_node_set_pixtext(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, text, (guint8)spacing, GDK_PIXMAP(PHPG_GOBJECT(pixmap)), mask);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkCTree, node_set_row_style)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *style;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_node, gpointer_ce, &style, gtkstyle_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_row_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, GTK_STYLE(PHPG_GOBJECT(style)));

}


static PHP_METHOD(GtkCTree, node_set_selectable)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	zend_bool selectable;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Ob", &php_node, gpointer_ce, &selectable))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_selectable(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gboolean)selectable);

}


static PHP_METHOD(GtkCTree, node_set_shift)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long column, vertical, horizontal;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oiii", &php_node, gpointer_ce, &column, &vertical, &horizontal))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_shift(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, (gint)vertical, (gint)horizontal);

}


static PHP_METHOD(GtkCTree, node_set_text)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long column;
	char *text;
	zend_bool free_text = FALSE;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oiu", &php_node, gpointer_ce, &column, &text, &free_text))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_node_set_text(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)column, text);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkCTree, real_select_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;
	long state;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "Oi", &php_node, gpointer_ce, &state))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_real_select_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, (gint)state);

}


static PHP_METHOD(GtkCTree, remove_node)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_remove_node(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, select)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_select(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, select_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_select_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, set_expander_style)
{
	GtkCTreeExpanderStyle expander_style;
	zval *php_expander_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_expander_style))
		return;

	if (php_expander_style && phpg_gvalue_get_enum(GTK_TYPE_CTREE_EXPANDER_STYLE, php_expander_style, (gint *)&expander_style) == FAILURE) {
		return;
	}

    gtk_ctree_set_expander_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), expander_style);

}


static PHP_METHOD(GtkCTree, set_indent)
{
	long indent;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &indent))
		return;

    gtk_ctree_set_indent(GTK_CTREE(PHPG_GOBJECT(this_ptr)), (gint)indent);

}


static PHP_METHOD(GtkCTree, set_line_style)
{
	GtkCTreeLineStyle line_style;
	zval *php_line_style = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "V", &php_line_style))
		return;

	if (php_line_style && phpg_gvalue_get_enum(GTK_TYPE_CTREE_LINE_STYLE, php_line_style, (gint *)&line_style) == FAILURE) {
		return;
	}

    gtk_ctree_set_line_style(GTK_CTREE(PHPG_GOBJECT(this_ptr)), line_style);

}


static PHP_METHOD(GtkCTree, set_node_info)
{
	GtkCTreeNode *node = NULL;
	zval *php_node, *php_pixmap_closed, *php_mask_closed, *php_pixmap_opened, *php_mask_opened;
	char *text;
	zend_bool free_text = FALSE, is_leaf, expanded;
	long spacing;
	GdkPixmap *pixmap_closed = NULL, *mask_closed = NULL, *pixmap_opened = NULL, *mask_opened = NULL;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OuiNNNNbb", &php_node, gpointer_ce, &text, &free_text, &spacing, &php_pixmap_closed, gdkpixmap_ce, &php_mask_closed, gdkpixmap_ce, &php_pixmap_opened, gdkpixmap_ce, &php_mask_opened, gdkpixmap_ce, &is_leaf, &expanded))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }
    if (Z_TYPE_P(php_pixmap_closed) != IS_NULL)
        pixmap_closed = GDK_PIXMAP(PHPG_GOBJECT(php_pixmap_closed));
    if (Z_TYPE_P(php_mask_closed) != IS_NULL)
        mask_closed = GDK_PIXMAP(PHPG_GOBJECT(php_mask_closed));
    if (Z_TYPE_P(php_pixmap_opened) != IS_NULL)
        pixmap_opened = GDK_PIXMAP(PHPG_GOBJECT(php_pixmap_opened));
    if (Z_TYPE_P(php_mask_opened) != IS_NULL)
        mask_opened = GDK_PIXMAP(PHPG_GOBJECT(php_mask_opened));

    gtk_ctree_set_node_info(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, text, (guint8)spacing, pixmap_closed, mask_closed, pixmap_opened, mask_opened, (gboolean)is_leaf, (gboolean)expanded);
	if (free_text) g_free(text);

}


static PHP_METHOD(GtkCTree, set_show_stub)
{
	zend_bool show_stub;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "b", &show_stub))
		return;

    gtk_ctree_set_show_stub(GTK_CTREE(PHPG_GOBJECT(this_ptr)), (gboolean)show_stub);

}


static PHP_METHOD(GtkCTree, set_spacing)
{
	long spacing;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "i", &spacing))
		return;

    gtk_ctree_set_spacing(GTK_CTREE(PHPG_GOBJECT(this_ptr)), (gint)spacing);

}


static PHP_METHOD(GtkCTree, sort_node)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_sort_node(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, sort_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_sort_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, toggle_expansion)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_toggle_expansion(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, toggle_expansion_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_toggle_expansion_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, unselect)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_unselect(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}


static PHP_METHOD(GtkCTree, unselect_recursive)
{
	GtkCTreeNode *node = NULL;
	zval *php_node;

    NOT_STATIC_METHOD();

	if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "O", &php_node, gpointer_ce))
		return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    gtk_ctree_unselect_recursive(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node);

}

#line 2474 "ext/gtk+/gtk.overrides"
static PHP_METHOD(GtkCTree, row_is_ancestor)
{
    GtkCTreeNode *node = NULL, *child = NULL;
    zval *php_node, *php_child;
    gboolean php_retval;

    NOT_STATIC_METHOD();

    if (!php_gtk_parse_args(ZEND_NUM_ARGS(), "OO", &php_node, gpointer_ce, &php_child, gpointer_ce))
        return;

    if (phpg_gpointer_check(php_node, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        node = (GtkCTreeNode *) PHPG_GPOINTER(php_node);
    } else {
        php_error(E_WARNING, "%s::%s() expects node argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    if (phpg_gpointer_check(php_child, GTK_TYPE_CTREE_NODE, FALSE TSRMLS_CC)) {
        child = (GtkCTreeNode *) PHPG_GPOINTER(php_child);
    } else {
        php_error(E_WARNING, "%s::%s() expects child argument to be a valid GtkCTreeNode object", get_active_class_name(NULL TSRMLS_CC), get_active_function_name(TSRMLS_C));
        return;
    }

    php_retval = gtk_ctree_is_ancestor(GTK_CTREE(PHPG_GOBJECT(this_ptr)), node, child);
    RETVAL_BOOL(php_retval);
}

#line 65833 "ext/gtk+/gen_gtk.c"


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO_EX(arginfo_gtk_gtkctree_gtk_ctree_new, 0, 0, 0)
    ZEND_ARG_INFO(0, columns)
    ZEND_ARG_INFO(0, tree_column)
    ZEND_ARG_INFO(0, titles)
ZEND_END_ARG_INFO();


PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_collapse, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_collapse_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_collapse_to_depth, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, depth)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_expand, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_expand_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_expand_to_depth, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, depth)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_find, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_is_hot_spot, 0)
    ZEND_ARG_INFO(0, x)
    ZEND_ARG_INFO(0, y)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_is_viewable, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_last, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_move, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, new_parent, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, new_sibling, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_get_cell_style, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_get_cell_type, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_get_row_style, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_get_selectable, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_is_visible, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_moveto, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, row_align)
    ZEND_ARG_INFO(0, col_align)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_nth, 0)
    ZEND_ARG_INFO(0, row)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_background, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_cell_style, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_foreground, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, color, GdkColor, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_pixmap, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_pixtext, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, spacing)
    ZEND_ARG_OBJ_INFO(0, pixmap, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_row_style, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, style, GtkStyle, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_selectable, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, selectable)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_shift, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, vertical)
    ZEND_ARG_INFO(0, horizontal)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_node_set_text, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, column)
    ZEND_ARG_INFO(0, text)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_real_select_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, state)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_remove_node, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_select, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_select_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_expander_style, 0)
    ZEND_ARG_INFO(0, expander_style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_indent, 0)
    ZEND_ARG_INFO(0, indent)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_line_style, 0)
    ZEND_ARG_INFO(0, line_style)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_node_info, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_INFO(0, text)
    ZEND_ARG_INFO(0, spacing)
    ZEND_ARG_OBJ_INFO(0, pixmap_closed, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask_closed)
    ZEND_ARG_OBJ_INFO(0, pixmap_opened, GdkPixmap, 1)
    ZEND_ARG_INFO(0, mask_opened)
    ZEND_ARG_INFO(0, is_leaf)
    ZEND_ARG_INFO(0, expanded)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_show_stub, 0)
    ZEND_ARG_INFO(0, show_stub)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_set_spacing, 0)
    ZEND_ARG_INFO(0, spacing)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_sort_node, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_sort_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_toggle_expansion, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_toggle_expansion_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_unselect, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();

PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_unselect_recursive, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();
PHPGTK_ARG_INFO_STATIC
ZEND_BEGIN_ARG_INFO(arginfo_gtk_gtkctree_row_is_ancestor, 0)
    ZEND_ARG_OBJ_INFO(0, node, GtkCTreeNode, 1)
    ZEND_ARG_OBJ_INFO(0, child, GtkCTreeNode, 1)
ZEND_END_ARG_INFO();


static zend_function_entry gtkctree_methods[] = {
	PHP_ME(GtkCTree, __construct,          arginfo_gtk_gtkctree_gtk_ctree_new, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, collapse,             arginfo_gtk_gtkctree_collapse, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, collapse_recursive,   arginfo_gtk_gtkctree_collapse_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, collapse_to_depth,    arginfo_gtk_gtkctree_collapse_to_depth, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, expand,               arginfo_gtk_gtkctree_expand, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, expand_recursive,     arginfo_gtk_gtkctree_expand_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, expand_to_depth,      arginfo_gtk_gtkctree_expand_to_depth, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, find,                 arginfo_gtk_gtkctree_find, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, insert_node,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, is_hot_spot,          arginfo_gtk_gtkctree_is_hot_spot, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, is_viewable,          arginfo_gtk_gtkctree_is_viewable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, last,                 arginfo_gtk_gtkctree_last, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, move,                 arginfo_gtk_gtkctree_move, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_get_cell_style,  arginfo_gtk_gtkctree_node_get_cell_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_get_cell_type,   arginfo_gtk_gtkctree_node_get_cell_type, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_get_row_style,   arginfo_gtk_gtkctree_node_get_row_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_get_selectable,  arginfo_gtk_gtkctree_node_get_selectable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_is_visible,      arginfo_gtk_gtkctree_node_is_visible, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_moveto,          arginfo_gtk_gtkctree_node_moveto, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_nth,             arginfo_gtk_gtkctree_node_nth, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_background,  arginfo_gtk_gtkctree_node_set_background, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_cell_style,  arginfo_gtk_gtkctree_node_set_cell_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_foreground,  arginfo_gtk_gtkctree_node_set_foreground, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_pixmap,      arginfo_gtk_gtkctree_node_set_pixmap, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_pixtext,     arginfo_gtk_gtkctree_node_set_pixtext, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_row_style,   arginfo_gtk_gtkctree_node_set_row_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_selectable,  arginfo_gtk_gtkctree_node_set_selectable, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_shift,       arginfo_gtk_gtkctree_node_set_shift, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, node_set_text,        arginfo_gtk_gtkctree_node_set_text, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, real_select_recursive, arginfo_gtk_gtkctree_real_select_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, remove_node,          arginfo_gtk_gtkctree_remove_node, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, row_is_ancestor,      arginfo_gtk_gtkctree_row_is_ancestor, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, select,               arginfo_gtk_gtkctree_select, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, select_recursive,     arginfo_gtk_gtkctree_select_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_expander_style,   arginfo_gtk_gtkctree_set_expander_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_indent,           arginfo_gtk_gtkctree_set_indent, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_line_style,       arginfo_gtk_gtkctree_set_line_style, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_node_info,        arginfo_gtk_gtkctree_set_node_info, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_show_stub,        arginfo_gtk_gtkctree_set_show_stub, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, set_spacing,          arginfo_gtk_gtkctree_set_spacing, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, sort_node,            arginfo_gtk_gtkctree_sort_node, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, sort_recursive,       arginfo_gtk_gtkctree_sort_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, toggle_expansion,     arginfo_gtk_gtkctree_toggle_expansion, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, toggle_expansion_recursive, arginfo_gtk_gtkctree_toggle_expansion_recursive, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, unselect,             arginfo_gtk_gtkctree_unselect, ZEND_ACC_PUBLIC)
	PHP_ME(GtkCTree, unselect_recursive,   arginfo_gtk_gtkctree_unselect_recursive, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};

void phpg_gtk0_register_classes(TSRMLS_D)
{

	gtk_ce = phpg_register_class("Gtk", gtk_methods, NULL, 0, NULL, NULL, 0 TSRMLS_CC);

	gtktoolshell_ce = phpg_register_interface("GtkToolShell", gtktoolshell_methods, GTK_TYPE_TOOL_SHELL TSRMLS_CC);

	gtkbuildable_ce = phpg_register_interface("GtkBuildable", gtkbuildable_methods, GTK_TYPE_BUILDABLE TSRMLS_CC);

	gtkprintoperationpreview_ce = phpg_register_interface("GtkPrintOperationPreview", gtkprintoperationpreview_methods, GTK_TYPE_PRINT_OPERATION_PREVIEW TSRMLS_CC);

	gtkrecentchooser_ce = phpg_register_interface("GtkRecentChooser", gtkrecentchooser_methods, GTK_TYPE_RECENT_CHOOSER TSRMLS_CC);

	gtkcelleditable_ce = phpg_register_interface("GtkCellEditable", gtkcelleditable_methods, GTK_TYPE_CELL_EDITABLE TSRMLS_CC);

	gtkcelllayout_ce = phpg_register_interface("GtkCellLayout", gtkcelllayout_methods, GTK_TYPE_CELL_LAYOUT TSRMLS_CC);

	gtkeditable_ce = phpg_register_interface("GtkEditable", gtkeditable_methods, GTK_TYPE_EDITABLE TSRMLS_CC);

	gtkfilechooser_ce = phpg_register_interface("GtkFileChooser", gtkfilechooser_methods, GTK_TYPE_FILE_CHOOSER TSRMLS_CC);

	gtktreedragdest_ce = phpg_register_interface("GtkTreeDragDest", gtktreedragdest_methods, GTK_TYPE_TREE_DRAG_DEST TSRMLS_CC);

	gtktreedragsource_ce = phpg_register_interface("GtkTreeDragSource", gtktreedragsource_methods, GTK_TYPE_TREE_DRAG_SOURCE TSRMLS_CC);

	gtktreemodel_ce = phpg_register_interface("GtkTreeModel", gtktreemodel_methods, GTK_TYPE_TREE_MODEL TSRMLS_CC);

	gtktreesortable_ce = phpg_register_interface("GtkTreeSortable", gtktreesortable_methods, GTK_TYPE_TREE_SORTABLE TSRMLS_CC);

	gtkactivatable_ce = phpg_register_interface("GtkActivatable", gtkactivatable_methods, GTK_TYPE_ACTIVATABLE TSRMLS_CC);

	gtkorientable_ce = phpg_register_interface("GtkOrientable", gtkorientable_methods, GTK_TYPE_ORIENTABLE TSRMLS_CC);

	gtkmountoperation_ce = phpg_register_class("GtkMountOperation", gtkmountoperation_methods, NULL, 0, NULL, NULL, GTK_TYPE_MOUNT_OPERATION TSRMLS_CC);

	gtkbuilder_ce = phpg_register_class("GtkBuilder", gtkbuilder_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_BUILDER TSRMLS_CC);

	gtktooltip_ce = phpg_register_class("GtkTooltip", gtktooltip_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TOOLTIP TSRMLS_CC);

	gtkpagesetup_ce = phpg_register_class("GtkPageSetup", gtkpagesetup_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_PAGE_SETUP TSRMLS_CC);

	gtkprintcontext_ce = phpg_register_class("GtkPrintContext", gtkprintcontext_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_PRINT_CONTEXT TSRMLS_CC);

	gtkprintoperation_ce = phpg_register_class("GtkPrintOperation", gtkprintoperation_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_PRINT_OPERATION TSRMLS_CC);
	zend_class_implements(gtkprintoperation_ce TSRMLS_CC, 1, gtkprintoperationpreview_ce);

	gtkprintsettings_ce = phpg_register_class("GtkPrintSettings", gtkprintsettings_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_PRINT_SETTINGS TSRMLS_CC);

	gtkrecentmanager_ce = phpg_register_class("GtkRecentManager", gtkrecentmanager_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_RECENT_MANAGER TSRMLS_CC);

	gtkstatusicon_ce = phpg_register_class("GtkStatusIcon", gtkstatusicon_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_STATUS_ICON TSRMLS_CC);

	phpgtkcustomtreemodel_ce = phpg_register_class("PhpGtkCustomTreeModel", phpgtkcustomtreemodel_methods, gobject_ce, 0, NULL, phpg_create_phpgtkcustomtreemodel, PHPG_TYPE_CUSTOM_TREE_MODEL TSRMLS_CC);
	phpgtkcustomtreemodel_object_handlers = phpg_gobject_handlers;
	phpgtkcustomtreemodel_object_handlers.read_dimension = phpg_phpgtkcustomtreemodel_read_dimension_handler;
	phpgtkcustomtreemodel_object_handlers.write_dimension = phpg_phpgtkcustomtreemodel_write_dimension_handler;
	phpgtkcustomtreemodel_object_handlers.has_dimension = phpg_phpgtkcustomtreemodel_has_dimension_handler;
	phpgtkcustomtreemodel_object_handlers.unset_dimension = phpg_phpgtkcustomtreemodel_unset_dimension_handler;
	phpgtkcustomtreemodel_object_handlers.count_elements = phpg_phpgtkcustomtreemodel_count_elements_handler;
	zend_class_implements(phpgtkcustomtreemodel_ce TSRMLS_CC, 1, gtktreemodel_ce);

    phpgtkcustomtreemodel_ce->get_iterator = phpg_treemodel_get_iterator;
    zend_class_implements(phpgtkcustomtreemodel_ce TSRMLS_CC, 1, zend_ce_traversable);



	gtkaccelgroup_ce = phpg_register_class("GtkAccelGroup", gtkaccelgroup_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ACCEL_GROUP TSRMLS_CC);

	gtkaccelmap_ce = phpg_register_class("GtkAccelMap", gtkaccelmap_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ACCEL_MAP TSRMLS_CC);

	gtkaccessible_ce = phpg_register_class("GtkAccessible", gtkaccessible_methods, atkobject_ce, 0, NULL, NULL, GTK_TYPE_ACCESSIBLE TSRMLS_CC);

	gtkaction_ce = phpg_register_class("GtkAction", gtkaction_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ACTION TSRMLS_CC);

	gtkactiongroup_ce = phpg_register_class("GtkActionGroup", gtkactiongroup_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ACTION_GROUP TSRMLS_CC);

	gtkrecentaction_ce = phpg_register_class("GtkRecentAction", gtkrecentaction_methods, gtkaction_ce, 0, NULL, NULL, GTK_TYPE_RECENT_ACTION TSRMLS_CC);

	gtkclipboard_ce = phpg_register_class("GtkClipboard", gtkclipboard_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_CLIPBOARD TSRMLS_CC);

	gtkentrycompletion_ce = phpg_register_class("GtkEntryCompletion", gtkentrycompletion_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ENTRY_COMPLETION TSRMLS_CC);
	zend_class_implements(gtkentrycompletion_ce TSRMLS_CC, 1, gtkcelllayout_ce);

	gtkiconfactory_ce = phpg_register_class("GtkIconFactory", gtkiconfactory_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ICON_FACTORY TSRMLS_CC);

	gtkicontheme_ce = phpg_register_class("GtkIconTheme", gtkicontheme_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_ICON_THEME TSRMLS_CC);

	gtkliststore_ce = phpg_register_class("GtkListStore", gtkliststore_methods, gobject_ce, 0, NULL, phpg_create_gtkliststore, GTK_TYPE_LIST_STORE TSRMLS_CC);
	gtkliststore_object_handlers = phpg_gobject_handlers;
	gtkliststore_object_handlers.read_dimension = phpg_gtkliststore_read_dimension_handler;
	gtkliststore_object_handlers.write_dimension = phpg_gtkliststore_write_dimension_handler;
	gtkliststore_object_handlers.has_dimension = phpg_gtkliststore_has_dimension_handler;
	gtkliststore_object_handlers.unset_dimension = phpg_gtkliststore_unset_dimension_handler;
	gtkliststore_object_handlers.count_elements = phpg_gtkliststore_count_elements_handler;
	zend_class_implements(gtkliststore_ce TSRMLS_CC, 4, gtktreemodel_ce, gtktreedragsource_ce, gtktreedragdest_ce, gtktreesortable_ce);

    gtkliststore_ce->get_iterator = phpg_treemodel_get_iterator;
    zend_class_implements(gtkliststore_ce TSRMLS_CC, 1, zend_ce_traversable);


	gtkobject_ce = phpg_register_class("GtkObject", gtkobject_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_OBJECT TSRMLS_CC);

	gtkitemfactory_ce = phpg_register_class("GtkItemFactory", gtkitemfactory_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_ITEM_FACTORY TSRMLS_CC);

	gtkimcontext_ce = phpg_register_class("GtkIMContext", gtkimcontext_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_IM_CONTEXT TSRMLS_CC);

	gtkfilefilter_ce = phpg_register_class("GtkFileFilter", gtkfilefilter_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_FILE_FILTER TSRMLS_CC);

	gtkimmulticontext_ce = phpg_register_class("GtkIMMulticontext", gtkimmulticontext_methods, gtkimcontext_ce, 0, NULL, NULL, GTK_TYPE_IM_MULTICONTEXT TSRMLS_CC);

	gtkimcontextsimple_ce = phpg_register_class("GtkIMContextSimple", gtkimcontextsimple_methods, gtkimcontext_ce, 0, NULL, NULL, GTK_TYPE_IM_CONTEXT_SIMPLE TSRMLS_CC);

	gtkcellrenderer_ce = phpg_register_class("GtkCellRenderer", gtkcellrenderer_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER TSRMLS_CC);

	gtkadjustment_ce = phpg_register_class("GtkAdjustment", gtkadjustment_methods, gtkobject_ce, 0, gtkadjustment_prop_info, NULL, GTK_TYPE_ADJUSTMENT TSRMLS_CC);

	gtkcellrenderertoggle_ce = phpg_register_class("GtkCellRendererToggle", gtkcellrenderertoggle_methods, gtkcellrenderer_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_TOGGLE TSRMLS_CC);

	gtkcellrenderertext_ce = phpg_register_class("GtkCellRendererText", gtkcellrenderertext_methods, gtkcellrenderer_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_TEXT TSRMLS_CC);

	gtkcellrendererprogress_ce = phpg_register_class("GtkCellRendererProgress", gtkcellrendererprogress_methods, gtkcellrenderer_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_PROGRESS TSRMLS_CC);

	gtkcellrendereraccel_ce = phpg_register_class("GtkCellRendererAccel", gtkcellrendereraccel_methods, gtkcellrenderertext_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_ACCEL TSRMLS_CC);

	gtkcellrendererspin_ce = phpg_register_class("GtkCellRendererSpin", gtkcellrendererspin_methods, gtkcellrenderertext_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_SPIN TSRMLS_CC);

	gtkcellrenderercombo_ce = phpg_register_class("GtkCellRendererCombo", gtkcellrenderercombo_methods, gtkcellrenderertext_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_COMBO TSRMLS_CC);

	gtkcellrendererpixbuf_ce = phpg_register_class("GtkCellRendererPixbuf", gtkcellrendererpixbuf_methods, gtkcellrenderer_ce, 0, NULL, NULL, GTK_TYPE_CELL_RENDERER_PIXBUF TSRMLS_CC);

	gtkrecentfilter_ce = phpg_register_class("GtkRecentFilter", gtkrecentfilter_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_RECENT_FILTER TSRMLS_CC);

	gtkrcstyle_ce = phpg_register_class("GtkRcStyle", gtkrcstyle_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_RC_STYLE TSRMLS_CC);

	gtksettings_ce = phpg_register_class("GtkSettings", gtksettings_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_SETTINGS TSRMLS_CC);

	gtksizegroup_ce = phpg_register_class("GtkSizeGroup", gtksizegroup_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_SIZE_GROUP TSRMLS_CC);

	gtkstyle_ce = phpg_register_class("GtkStyle", gtkstyle_methods, gobject_ce, 0, gtkstyle_prop_info, NULL, GTK_TYPE_STYLE TSRMLS_CC);

	gtktextbuffer_ce = phpg_register_class("GtkTextBuffer", gtktextbuffer_methods, gobject_ce, 0, gtktextbuffer_prop_info, NULL, GTK_TYPE_TEXT_BUFFER TSRMLS_CC);

	gtktextchildanchor_ce = phpg_register_class("GtkTextChildAnchor", gtktextchildanchor_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TEXT_CHILD_ANCHOR TSRMLS_CC);

	gtktextmark_ce = phpg_register_class("GtkTextMark", gtktextmark_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TEXT_MARK TSRMLS_CC);

	gtktexttag_ce = phpg_register_class("GtkTextTag", gtktexttag_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TEXT_TAG TSRMLS_CC);

	gtktexttagtable_ce = phpg_register_class("GtkTextTagTable", gtktexttagtable_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TEXT_TAG_TABLE TSRMLS_CC);

	gtktoggleaction_ce = phpg_register_class("GtkToggleAction", gtktoggleaction_methods, gtkaction_ce, 0, NULL, NULL, GTK_TYPE_TOGGLE_ACTION TSRMLS_CC);

	gtkradioaction_ce = phpg_register_class("GtkRadioAction", gtkradioaction_methods, gtktoggleaction_ce, 0, NULL, NULL, GTK_TYPE_RADIO_ACTION TSRMLS_CC);

	gtktooltips_ce = phpg_register_class("GtkTooltips", gtktooltips_methods, gtkobject_ce, 0, gtktooltips_prop_info, NULL, GTK_TYPE_TOOLTIPS TSRMLS_CC);

	gtktreemodelfilter_ce = phpg_register_class("GtkTreeModelFilter", gtktreemodelfilter_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TREE_MODEL_FILTER TSRMLS_CC);
	zend_class_implements(gtktreemodelfilter_ce TSRMLS_CC, 2, gtktreemodel_ce, gtktreedragsource_ce);

	gtktreemodelsort_ce = phpg_register_class("GtkTreeModelSort", gtktreemodelsort_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TREE_MODEL_SORT TSRMLS_CC);
	zend_class_implements(gtktreemodelsort_ce TSRMLS_CC, 2, gtktreemodel_ce, gtktreesortable_ce);

	gtktreeselection_ce = phpg_register_class("GtkTreeSelection", gtktreeselection_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_TREE_SELECTION TSRMLS_CC);

	gtktreestore_ce = phpg_register_class("GtkTreeStore", gtktreestore_methods, gobject_ce, 0, NULL, phpg_create_gtktreestore, GTK_TYPE_TREE_STORE TSRMLS_CC);
	gtktreestore_object_handlers = phpg_gobject_handlers;
	gtktreestore_object_handlers.read_dimension = phpg_gtktreestore_read_dimension_handler;
	gtktreestore_object_handlers.write_dimension = phpg_gtktreestore_write_dimension_handler;
	gtktreestore_object_handlers.has_dimension = phpg_gtktreestore_has_dimension_handler;
	gtktreestore_object_handlers.unset_dimension = phpg_gtktreestore_unset_dimension_handler;
	gtktreestore_object_handlers.count_elements = phpg_gtktreestore_count_elements_handler;
	zend_class_implements(gtktreestore_ce TSRMLS_CC, 4, gtktreemodel_ce, gtktreedragsource_ce, gtktreedragdest_ce, gtktreesortable_ce);

    gtktreestore_ce->get_iterator = phpg_treemodel_get_iterator;
    zend_class_implements(gtktreestore_ce TSRMLS_CC, 1, zend_ce_traversable);



	gtktreeviewcolumn_ce = phpg_register_class("GtkTreeViewColumn", gtktreeviewcolumn_methods, gtkobject_ce, 0, NULL, NULL, GTK_TYPE_TREE_VIEW_COLUMN TSRMLS_CC);
	zend_class_implements(gtktreeviewcolumn_ce TSRMLS_CC, 1, gtkcelllayout_ce);

	gtkuimanager_ce = phpg_register_class("GtkUIManager", gtkuimanager_methods, gobject_ce, 0, NULL, NULL, GTK_TYPE_UI_MANAGER TSRMLS_CC);

	gtkwidget_ce = phpg_register_class("GtkWidget", gtkwidget_methods, gtkobject_ce, 0, gtkwidget_prop_info, NULL, GTK_TYPE_WIDGET TSRMLS_CC);

	gtkseparator_ce = phpg_register_class("GtkSeparator", NULL, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_SEPARATOR TSRMLS_CC);

	gtkruler_ce = phpg_register_class("GtkRuler", gtkruler_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_RULER TSRMLS_CC);

	gtkvseparator_ce = phpg_register_class("GtkVSeparator", gtkvseparator_methods, gtkseparator_ce, 0, NULL, NULL, GTK_TYPE_VSEPARATOR TSRMLS_CC);

	gtkhseparator_ce = phpg_register_class("GtkHSeparator", gtkhseparator_methods, gtkseparator_ce, 0, NULL, NULL, GTK_TYPE_HSEPARATOR TSRMLS_CC);

	gtkvruler_ce = phpg_register_class("GtkVRuler", gtkvruler_methods, gtkruler_ce, 0, NULL, NULL, GTK_TYPE_VRULER TSRMLS_CC);

	gtkrange_ce = phpg_register_class("GtkRange", gtkrange_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_RANGE TSRMLS_CC);

	gtkhruler_ce = phpg_register_class("GtkHRuler", gtkhruler_methods, gtkruler_ce, 0, NULL, NULL, GTK_TYPE_HRULER TSRMLS_CC);

	gtkscrollbar_ce = phpg_register_class("GtkScrollbar", NULL, gtkrange_ce, 0, NULL, NULL, GTK_TYPE_SCROLLBAR TSRMLS_CC);

	gtkprogress_ce = phpg_register_class("GtkProgress", gtkprogress_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_PROGRESS TSRMLS_CC);

	gtkvscrollbar_ce = phpg_register_class("GtkVScrollbar", gtkvscrollbar_methods, gtkscrollbar_ce, 0, NULL, NULL, GTK_TYPE_VSCROLLBAR TSRMLS_CC);

	gtkhscrollbar_ce = phpg_register_class("GtkHScrollbar", gtkhscrollbar_methods, gtkscrollbar_ce, 0, NULL, NULL, GTK_TYPE_HSCROLLBAR TSRMLS_CC);

	gtkscale_ce = phpg_register_class("GtkScale", gtkscale_methods, gtkrange_ce, 0, NULL, NULL, GTK_TYPE_SCALE TSRMLS_CC);

	gtkpreview_ce = phpg_register_class("GtkPreview", gtkpreview_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_PREVIEW TSRMLS_CC);

	gtkvscale_ce = phpg_register_class("GtkVScale", gtkvscale_methods, gtkscale_ce, 0, NULL, NULL, GTK_TYPE_VSCALE TSRMLS_CC);

	gtkhscale_ce = phpg_register_class("GtkHScale", gtkhscale_methods, gtkscale_ce, 0, NULL, NULL, GTK_TYPE_HSCALE TSRMLS_CC);

	gtkprogressbar_ce = phpg_register_class("GtkProgressBar", gtkprogressbar_methods, gtkprogress_ce, 0, NULL, NULL, GTK_TYPE_PROGRESS_BAR TSRMLS_CC);

	gtkoldeditable_ce = phpg_register_class("GtkOldEditable", gtkoldeditable_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_OLD_EDITABLE TSRMLS_CC);
	zend_class_implements(gtkoldeditable_ce TSRMLS_CC, 1, gtkeditable_ce);

	gtkmisc_ce = phpg_register_class("GtkMisc", gtkmisc_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_MISC TSRMLS_CC);

	gtkinvisible_ce = phpg_register_class("GtkInvisible", gtkinvisible_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_INVISIBLE TSRMLS_CC);

	gtkpixmap_ce = phpg_register_class("GtkPixmap", gtkpixmap_methods, gtkmisc_ce, 0, NULL, NULL, GTK_TYPE_PIXMAP TSRMLS_CC);

	gtkarrow_ce = phpg_register_class("GtkArrow", gtkarrow_methods, gtkmisc_ce, 0, NULL, NULL, GTK_TYPE_ARROW TSRMLS_CC);

	gtkimage_ce = phpg_register_class("GtkImage", gtkimage_methods, gtkmisc_ce, 0, NULL, NULL, GTK_TYPE_IMAGE TSRMLS_CC);

	gtklabel_ce = phpg_register_class("GtkLabel", gtklabel_methods, gtkmisc_ce, 0, NULL, NULL, GTK_TYPE_LABEL TSRMLS_CC);

	gtkentry_ce = phpg_register_class("GtkEntry", gtkentry_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_ENTRY TSRMLS_CC);
	zend_class_implements(gtkentry_ce TSRMLS_CC, 2, gtkeditable_ce, gtkcelleditable_ce);

	gtktipsquery_ce = phpg_register_class("GtkTipsQuery", gtktipsquery_methods, gtklabel_ce, 0, NULL, NULL, GTK_TYPE_TIPS_QUERY TSRMLS_CC);

	gtkaccellabel_ce = phpg_register_class("GtkAccelLabel", gtkaccellabel_methods, gtklabel_ce, 0, NULL, NULL, GTK_TYPE_ACCEL_LABEL TSRMLS_CC);

	gtkspinbutton_ce = phpg_register_class("GtkSpinButton", gtkspinbutton_methods, gtkentry_ce, 0, NULL, NULL, GTK_TYPE_SPIN_BUTTON TSRMLS_CC);

	gtkdrawingarea_ce = phpg_register_class("GtkDrawingArea", gtkdrawingarea_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_DRAWING_AREA TSRMLS_CC);

	gtkcontainer_ce = phpg_register_class("GtkContainer", gtkcontainer_methods, gtkwidget_ce, 0, gtkcontainer_prop_info, NULL, GTK_TYPE_CONTAINER TSRMLS_CC);

	gtkcurve_ce = phpg_register_class("GtkCurve", gtkcurve_methods, gtkdrawingarea_ce, 0, NULL, NULL, GTK_TYPE_CURVE TSRMLS_CC);

	gtktreeview_ce = phpg_register_class("GtkTreeView", gtktreeview_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TREE_VIEW TSRMLS_CC);

	gtktoolbar_ce = phpg_register_class("GtkToolbar", gtktoolbar_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TOOLBAR TSRMLS_CC);

	gtktextview_ce = phpg_register_class("GtkTextView", gtktextview_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TEXT_VIEW TSRMLS_CC);

	gtktable_ce = phpg_register_class("GtkTable", gtktable_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_TABLE TSRMLS_CC);

	gtksocket_ce = phpg_register_class("GtkSocket", gtksocket_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_SOCKET TSRMLS_CC);

	gtkpaned_ce = phpg_register_class("GtkPaned", gtkpaned_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_PANED TSRMLS_CC);

	gtknotebook_ce = phpg_register_class("GtkNotebook", gtknotebook_methods, gtkcontainer_ce, 0, gtknotebook_prop_info, NULL, GTK_TYPE_NOTEBOOK TSRMLS_CC);

	gtkvpaned_ce = phpg_register_class("GtkVPaned", gtkvpaned_methods, gtkpaned_ce, 0, NULL, NULL, GTK_TYPE_VPANED TSRMLS_CC);

	gtkhpaned_ce = phpg_register_class("GtkHPaned", gtkhpaned_methods, gtkpaned_ce, 0, NULL, NULL, GTK_TYPE_HPANED TSRMLS_CC);

	gtkmenushell_ce = phpg_register_class("GtkMenuShell", gtkmenushell_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_MENU_SHELL TSRMLS_CC);

	gtklist_ce = phpg_register_class("GtkList", gtklist_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_LIST TSRMLS_CC);

	gtkmenu_ce = phpg_register_class("GtkMenu", gtkmenu_methods, gtkmenushell_ce, 0, NULL, NULL, GTK_TYPE_MENU TSRMLS_CC);

	gtkmenubar_ce = phpg_register_class("GtkMenuBar", gtkmenubar_methods, gtkmenushell_ce, 0, NULL, NULL, GTK_TYPE_MENU_BAR TSRMLS_CC);

	gtkrecentchoosermenu_ce = phpg_register_class("GtkRecentChooserMenu", gtkrecentchoosermenu_methods, gtkmenu_ce, 0, NULL, NULL, GTK_TYPE_RECENT_CHOOSER_MENU TSRMLS_CC);
	zend_class_implements(gtkrecentchoosermenu_ce TSRMLS_CC, 1, gtkrecentchooser_ce);

	gtklayout_ce = phpg_register_class("GtkLayout", gtklayout_methods, gtkcontainer_ce, 0, gtklayout_prop_info, NULL, GTK_TYPE_LAYOUT TSRMLS_CC);

	gtkiconview_ce = phpg_register_class("GtkIconView", gtkiconview_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_ICON_VIEW TSRMLS_CC);

	gtkfixed_ce = phpg_register_class("GtkFixed", gtkfixed_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_FIXED TSRMLS_CC);

	gtkcellview_ce = phpg_register_class("GtkCellView", gtkcellview_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_CELL_VIEW TSRMLS_CC);
	zend_class_implements(gtkcellview_ce TSRMLS_CC, 1, gtkcelllayout_ce);

	gtkbin_ce = phpg_register_class("GtkBin", gtkbin_methods, gtkcontainer_ce, 0, gtkbin_prop_info, NULL, GTK_TYPE_BIN TSRMLS_CC);

	gtkbox_ce = phpg_register_class("GtkBox", gtkbox_methods, gtkcontainer_ce, 0, NULL, NULL, GTK_TYPE_BOX TSRMLS_CC);

	gtkwindow_ce = phpg_register_class("GtkWindow", gtkwindow_methods, gtkbin_ce, 0, gtkwindow_prop_info, NULL, GTK_TYPE_WINDOW TSRMLS_CC);

	gtkviewport_ce = phpg_register_class("GtkViewport", gtkviewport_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_VIEWPORT TSRMLS_CC);

	gtkdialog_ce = phpg_register_class("GtkDialog", gtkdialog_methods, gtkwindow_ce, 0, gtkdialog_prop_info, NULL, GTK_TYPE_DIALOG TSRMLS_CC);

	gtkplug_ce = phpg_register_class("GtkPlug", gtkplug_methods, gtkwindow_ce, 0, NULL, NULL, GTK_TYPE_PLUG TSRMLS_CC);

	gtkcolorselectiondialog_ce = phpg_register_class("GtkColorSelectionDialog", gtkcolorselectiondialog_methods, gtkdialog_ce, 0, gtkcolorselectiondialog_prop_info, NULL, GTK_TYPE_COLOR_SELECTION_DIALOG TSRMLS_CC);

	gtkaboutdialog_ce = phpg_register_class("GtkAboutDialog", gtkaboutdialog_methods, gtkdialog_ce, 0, NULL, NULL, GTK_TYPE_ABOUT_DIALOG TSRMLS_CC);

	gtkrecentchooserdialog_ce = phpg_register_class("GtkRecentChooserDialog", gtkrecentchooserdialog_methods, gtkdialog_ce, 0, NULL, NULL, GTK_TYPE_RECENT_CHOOSER_DIALOG TSRMLS_CC);
	zend_class_implements(gtkrecentchooserdialog_ce TSRMLS_CC, 1, gtkrecentchooser_ce);

	gtkfilechooserdialog_ce = phpg_register_class("GtkFileChooserDialog", gtkfilechooserdialog_methods, gtkdialog_ce, 0, NULL, NULL, GTK_TYPE_FILE_CHOOSER_DIALOG TSRMLS_CC);
	zend_class_implements(gtkfilechooserdialog_ce TSRMLS_CC, 1, gtkfilechooser_ce);

	gtkfileselection_ce = phpg_register_class("GtkFileSelection", gtkfileselection_methods, gtkdialog_ce, 0, gtkfileselection_prop_info, NULL, GTK_TYPE_FILE_SELECTION TSRMLS_CC);

	gtkfontselectiondialog_ce = phpg_register_class("GtkFontSelectionDialog", gtkfontselectiondialog_methods, gtkdialog_ce, 0, gtkfontselectiondialog_prop_info, NULL, GTK_TYPE_FONT_SELECTION_DIALOG TSRMLS_CC);

	gtkinputdialog_ce = phpg_register_class("GtkInputDialog", gtkinputdialog_methods, gtkdialog_ce, 0, NULL, NULL, GTK_TYPE_INPUT_DIALOG TSRMLS_CC);

	gtkmessagedialog_ce = phpg_register_class("GtkMessageDialog", gtkmessagedialog_methods, gtkdialog_ce, 0, gtkmessagedialog_prop_info, NULL, GTK_TYPE_MESSAGE_DIALOG TSRMLS_CC);

	gtktoolitem_ce = phpg_register_class("GtkToolItem", gtktoolitem_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_TOOL_ITEM TSRMLS_CC);

	gtkvbox_ce = phpg_register_class("GtkVBox", gtkvbox_methods, gtkbox_ce, 0, NULL, NULL, GTK_TYPE_VBOX TSRMLS_CC);

	gtkscrolledwindow_ce = phpg_register_class("GtkScrolledWindow", gtkscrolledwindow_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_SCROLLED_WINDOW TSRMLS_CC);

	gtkrecentchooserwidget_ce = phpg_register_class("GtkRecentChooserWidget", gtkrecentchooserwidget_methods, gtkvbox_ce, 0, NULL, NULL, GTK_TYPE_RECENT_CHOOSER_WIDGET TSRMLS_CC);
	zend_class_implements(gtkrecentchooserwidget_ce TSRMLS_CC, 1, gtkrecentchooser_ce);

	gtkcolorselection_ce = phpg_register_class("GtkColorSelection", gtkcolorselection_methods, gtkvbox_ce, 0, NULL, NULL, GTK_TYPE_COLOR_SELECTION TSRMLS_CC);

	gtkfilechooserwidget_ce = phpg_register_class("GtkFileChooserWidget", gtkfilechooserwidget_methods, gtkvbox_ce, 0, NULL, NULL, GTK_TYPE_FILE_CHOOSER_WIDGET TSRMLS_CC);
	zend_class_implements(gtkfilechooserwidget_ce TSRMLS_CC, 1, gtkfilechooser_ce);

	gtkfontselection_ce = phpg_register_class("GtkFontSelection", gtkfontselection_methods, gtkvbox_ce, 0, NULL, NULL, GTK_TYPE_FONT_SELECTION TSRMLS_CC);

	gtkgammacurve_ce = phpg_register_class("GtkGammaCurve", gtkgammacurve_methods, gtkvbox_ce, 0, gtkgammacurve_prop_info, NULL, GTK_TYPE_GAMMA_CURVE TSRMLS_CC);

	gtkseparatortoolitem_ce = phpg_register_class("GtkSeparatorToolItem", gtkseparatortoolitem_methods, gtktoolitem_ce, 0, NULL, NULL, GTK_TYPE_SEPARATOR_TOOL_ITEM TSRMLS_CC);

	gtktoolbutton_ce = phpg_register_class("GtkToolButton", gtktoolbutton_methods, gtktoolitem_ce, 0, NULL, NULL, GTK_TYPE_TOOL_BUTTON TSRMLS_CC);

	gtkitem_ce = phpg_register_class("GtkItem", gtkitem_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_ITEM TSRMLS_CC);

	gtktoggletoolbutton_ce = phpg_register_class("GtkToggleToolButton", gtktoggletoolbutton_methods, gtktoolbutton_ce, 0, NULL, NULL, GTK_TYPE_TOGGLE_TOOL_BUTTON TSRMLS_CC);

	gtkmenutoolbutton_ce = phpg_register_class("GtkMenuToolButton", gtkmenutoolbutton_methods, gtktoolbutton_ce, 0, NULL, NULL, GTK_TYPE_MENU_TOOL_BUTTON TSRMLS_CC);

	gtkradiotoolbutton_ce = phpg_register_class("GtkRadioToolButton", gtkradiotoolbutton_methods, gtktoggletoolbutton_ce, 0, NULL, NULL, GTK_TYPE_RADIO_TOOL_BUTTON TSRMLS_CC);

	gtkhandlebox_ce = phpg_register_class("GtkHandleBox", gtkhandlebox_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_HANDLE_BOX TSRMLS_CC);

	gtkmenuitem_ce = phpg_register_class("GtkMenuItem", gtkmenuitem_methods, gtkitem_ce, 0, NULL, NULL, GTK_TYPE_MENU_ITEM TSRMLS_CC);

	gtklistitem_ce = phpg_register_class("GtkListItem", gtklistitem_methods, gtkitem_ce, 0, NULL, NULL, GTK_TYPE_LIST_ITEM TSRMLS_CC);

	gtktearoffmenuitem_ce = phpg_register_class("GtkTearoffMenuItem", gtktearoffmenuitem_methods, gtkmenuitem_ce, 0, NULL, NULL, GTK_TYPE_TEAROFF_MENU_ITEM TSRMLS_CC);

	gtkseparatormenuitem_ce = phpg_register_class("GtkSeparatorMenuItem", gtkseparatormenuitem_methods, gtkmenuitem_ce, 0, NULL, NULL, GTK_TYPE_SEPARATOR_MENU_ITEM TSRMLS_CC);

	gtkcheckmenuitem_ce = phpg_register_class("GtkCheckMenuItem", gtkcheckmenuitem_methods, gtkmenuitem_ce, 0, gtkcheckmenuitem_prop_info, NULL, GTK_TYPE_CHECK_MENU_ITEM TSRMLS_CC);

	gtkimagemenuitem_ce = phpg_register_class("GtkImageMenuItem", gtkimagemenuitem_methods, gtkmenuitem_ce, 0, NULL, NULL, GTK_TYPE_IMAGE_MENU_ITEM TSRMLS_CC);

	gtkradiomenuitem_ce = phpg_register_class("GtkRadioMenuItem", gtkradiomenuitem_methods, gtkcheckmenuitem_ce, 0, NULL, NULL, GTK_TYPE_RADIO_MENU_ITEM TSRMLS_CC);

	gtkhbox_ce = phpg_register_class("GtkHBox", gtkhbox_methods, gtkbox_ce, 0, NULL, NULL, GTK_TYPE_HBOX TSRMLS_CC);

	gtkframe_ce = phpg_register_class("GtkFrame", gtkframe_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_FRAME TSRMLS_CC);

	gtkstatusbar_ce = phpg_register_class("GtkStatusbar", gtkstatusbar_methods, gtkhbox_ce, 0, NULL, NULL, GTK_TYPE_STATUSBAR TSRMLS_CC);

	gtkcombo_ce = phpg_register_class("GtkCombo", gtkcombo_methods, gtkhbox_ce, 0, gtkcombo_prop_info, NULL, GTK_TYPE_COMBO TSRMLS_CC);

	gtkfilechooserbutton_ce = phpg_register_class("GtkFileChooserButton", gtkfilechooserbutton_methods, gtkhbox_ce, 0, NULL, NULL, GTK_TYPE_FILE_CHOOSER_BUTTON TSRMLS_CC);
	zend_class_implements(gtkfilechooserbutton_ce TSRMLS_CC, 1, gtkfilechooser_ce);

	gtkexpander_ce = phpg_register_class("GtkExpander", gtkexpander_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_EXPANDER TSRMLS_CC);

	gtkaspectframe_ce = phpg_register_class("GtkAspectFrame", gtkaspectframe_methods, gtkframe_ce, 0, NULL, NULL, GTK_TYPE_ASPECT_FRAME TSRMLS_CC);

	gtkeventbox_ce = phpg_register_class("GtkEventBox", gtkeventbox_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_EVENT_BOX TSRMLS_CC);

	gtkalignment_ce = phpg_register_class("GtkAlignment", gtkalignment_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_ALIGNMENT TSRMLS_CC);

	gtkbutton_ce = phpg_register_class("GtkButton", gtkbutton_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_BUTTON TSRMLS_CC);

	gtkcombobox_ce = phpg_register_class("GtkComboBox", gtkcombobox_methods, gtkbin_ce, 0, NULL, NULL, GTK_TYPE_COMBO_BOX TSRMLS_CC);
	zend_class_implements(gtkcombobox_ce TSRMLS_CC, 1, gtkcelllayout_ce);

	gtktogglebutton_ce = phpg_register_class("GtkToggleButton", gtktogglebutton_methods, gtkbutton_ce, 0, gtktogglebutton_prop_info, NULL, GTK_TYPE_TOGGLE_BUTTON TSRMLS_CC);

	gtkoptionmenu_ce = phpg_register_class("GtkOptionMenu", gtkoptionmenu_methods, gtkbutton_ce, 0, NULL, NULL, GTK_TYPE_OPTION_MENU TSRMLS_CC);

	gtkcheckbutton_ce = phpg_register_class("GtkCheckButton", gtkcheckbutton_methods, gtktogglebutton_ce, 0, NULL, NULL, GTK_TYPE_CHECK_BUTTON TSRMLS_CC);

	gtkfontbutton_ce = phpg_register_class("GtkFontButton", gtkfontbutton_methods, gtkbutton_ce, 0, NULL, NULL, GTK_TYPE_FONT_BUTTON TSRMLS_CC);

	gtkradiobutton_ce = phpg_register_class("GtkRadioButton", gtkradiobutton_methods, gtkcheckbutton_ce, 0, NULL, NULL, GTK_TYPE_RADIO_BUTTON TSRMLS_CC);

	gtklinkbutton_ce = phpg_register_class("GtkLinkButton", gtklinkbutton_methods, gtkbutton_ce, 0, NULL, NULL, GTK_TYPE_LINK_BUTTON TSRMLS_CC);

	gtkscalebutton_ce = phpg_register_class("GtkScaleButton", gtkscalebutton_methods, gtkbutton_ce, 0, NULL, NULL, GTK_TYPE_SCALE_BUTTON TSRMLS_CC);

	gtkcolorbutton_ce = phpg_register_class("GtkColorButton", gtkcolorbutton_methods, gtkbutton_ce, 0, NULL, NULL, GTK_TYPE_COLOR_BUTTON TSRMLS_CC);

	gtkvolumebutton_ce = phpg_register_class("GtkVolumeButton", gtkvolumebutton_methods, gtkscalebutton_ce, 0, NULL, NULL, GTK_TYPE_VOLUME_BUTTON TSRMLS_CC);

	gtkclist_ce = phpg_register_class("GtkCList", gtkclist_methods, gtkcontainer_ce, 0, gtkclist_prop_info, NULL, GTK_TYPE_CLIST TSRMLS_CC);

	gtkcomboboxentry_ce = phpg_register_class("GtkComboBoxEntry", gtkcomboboxentry_methods, gtkcombobox_ce, 0, NULL, NULL, GTK_TYPE_COMBO_BOX_ENTRY TSRMLS_CC);
	zend_class_implements(gtkcomboboxentry_ce TSRMLS_CC, 1, gtkcelllayout_ce);

	gtkbuttonbox_ce = phpg_register_class("GtkButtonBox", gtkbuttonbox_methods, gtkbox_ce, 0, NULL, NULL, GTK_TYPE_BUTTON_BOX TSRMLS_CC);

	gtkcalendar_ce = phpg_register_class("GtkCalendar", gtkcalendar_methods, gtkwidget_ce, 0, NULL, NULL, GTK_TYPE_CALENDAR TSRMLS_CC);

	gtkvbuttonbox_ce = phpg_register_class("GtkVButtonBox", gtkvbuttonbox_methods, gtkbuttonbox_ce, 0, NULL, NULL, GTK_TYPE_VBUTTON_BOX TSRMLS_CC);

	gtkhbuttonbox_ce = phpg_register_class("GtkHButtonBox", gtkhbuttonbox_methods, gtkbuttonbox_ce, 0, NULL, NULL, GTK_TYPE_HBUTTON_BOX TSRMLS_CC);

	gtkctree_ce = phpg_register_class("GtkCTree", gtkctree_methods, gtkclist_ce, 0, NULL, NULL, GTK_TYPE_CTREE TSRMLS_CC);
}

#endif /* HAVE_PHP_GTK */
